/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/
/*
Revision: 20.04.2003 E.Sack
*/

/* This module mainly lists generic functions to ensure better
   consistency. The aim is to maintain only one piece of code to fit
   changes in data structures.                                              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <stdarg.h>
#include <float.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"

/* Functions/Methods *******************************************************/

void
init_stats_win(stats_win *sw)
{
  assert(sw);

  sw->display_active = FALSE;
  init_area_def(&sw->wsize);
}

void
default_stats_win(stats_win *sw)
{
  assert(sw);

  sw->display_active = STATS_WIN_DEFACTV;
  sw->wsize.width = sw->wsize.height = STATS_WIN_DEFAULT;
}

void
default_icon_contrast(stats_win *sw)
{
  assert(sw);

  sw->display_active = STATS_ICON_CONTRAST_DEF;
}

void
init_stats_prefs(stats_prefs *se)
{
  assert(se);
                           /* don't init GTK stuff */

  init_stats_win(&se->sapply);
  init_stats_win(&se->srevert);
}

void
default_stats_prefs(stats_prefs *se)
{
  assert(se);

  default_stats_win(&se->sapply);
  default_stats_win(&se->srevert);
}

void
default_stats_icon_contrast(stats_prefs *ic)
{
  assert(ic);

  default_icon_contrast(&ic->sapply);
  default_icon_contrast(&ic->srevert);
}

void 
init_scroll_def(scroll_def *sw)
{
  assert(sw);

  sw->sdrawing = NULL;                                       /* zero pointer */
  sw->v_viewport = NULL;                                     /* zero pointer */
  sw->h_viewport = NULL;                                     /* zero pointer */
  sw->v_adj = NULL;                                          /* zero pointer */
  sw->h_adj = NULL;                                          /* zero pointer */
}

void
init_bc_img(bc_img *bc)
{
  bc->dfmt = STATS_FORMAT_UNDEF;

  bc->topwin = NULL;                                         /* zero pointer */
  bc->preview = NULL;                                        /* zero pointer */
  bc->poproot = NULL;                                        /* zero pointer */
  bc->popmenu = NULL;                                        /* zero pointer */

  bc->key = 0;
  bc->dfmt = STATS_FORMAT_UNDEF;
  bc->img_changed = FALSE;
  bc->magnitude = bc->magnitude_old = 0.0;

  init_img_typ(&bc->timg);
  init_stats_win(&bc->pwin);
  init_scroll_def(&bc->scroll);

  init_pic_def(&bc->magn_img);
  init_pic_def(&bc->bicon);
}

void
init_gbutton_def(gbutton_def *bd)
{
  assert(bd);

  bd->button = NULL;                                         /* zero pointer */
  bd->ebox = NULL;                                           /* zero pointer */
  bd->id = -1;
}

void
init_run_def(run_def *rd)
{
  assert(rd);

  rd->pactive = rd->pstop = NULL;                          /* zero pointer */
  rd->pbar = rd->plabel = NULL;                            /* zero pointer */
  rd->start_time = (time_t)0;
}

void
init_treenode_def(treenode_def *tn)
{
  assert(tn);

  tn->iaux = 0;
  tn->baux = tn->is_valid = FALSE;
}

void
init_foreach_def(foreach_def *fd)
{
  assert(fd);

  fd->txt  = NULL;                                           /* zero pointer */
  fd->iaux = 0;
  fd->baux = FALSE;

  fd->iter = fd->iter;
}

void
init_glist_def(glist_def *gd) 
{
  assert(gd);

  gd->items = NULL;                /* zero pointer: GTK initialising allowed */
  gd->fail = NULL;
}

void
init_entry_def(entry_def *ed, GtkWidget **info) 
{
  assert(ed);

  ed->info_pt = info;                              /* pointer elaboration */

  init_glist_def(&ed->elist);
}

void
init_media_att_set(media_att_set *mas)
{
  assert(mas);

  mas->tps_idx = 0;

  init_media_att_entry(&mas->ety);
}

void
init_cls_att_set(cls_att_set *as)
{
  assert(as);

  as->tps_idx = 0;
  as->class_idx = 0;

  init_cls_att_entry(&as->ety);
}

void
init_tps_id_set(tps_id_set *tis)
{
  assert(tis);

  tis->tps_idx = 0;

  init_tps_id_entry(&tis->ety);
}

void
init_media_att_entry(media_att_entry *mae)
{
  assert(mae);
                           /* don't init GTK stuff */

  mae->parse_status = XPLAB_PARSE_STAT_DEF;

  init_entry_def(&mae->dpi, &mae->infoline);
  init_entry_def(&mae->method, &mae->infoline);
  init_entry_def(&mae->l_phase, &mae->infoline);
  init_entry_def(&mae->h_phase, &mae->infoline);
}

void
init_cls_att_entry(cls_att_entry *ae)
{
  assert(ae);
                           /* don't init GTK stuff */

  ae->parse_status = XPLAB_PARSE_STAT_DEF;

  init_entry_def(&ae->class, &ae->infoline);
  init_entry_def(&ae->translatn, &ae->infoline);
  init_entry_def(&ae->commnt, &ae->infoline);
  init_entry_def(&ae->alignment, &ae->infoline);
  init_entry_def(&ae->printable, &ae->infoline);
  init_entry_def(&ae->fixed, &ae->infoline);
}

void
init_tps_id_entry(tps_id_entry *tie)
{
  assert(tie);
                           /* don't init GTK stuff */

  tie->parse_status = XPLAB_PARSE_STAT_DEF;

  init_entry_def(&tie->id, &tie->infoline);
}

void 
init_widget_lst(widget_lst *wl) 
{
  assert(wl);

  wl->cnt = 0;
  wl->wid = 0;                                               /* zero pointer */
}


void 
init_view_def(view_def *vd)
{
  assert(vd);

  vd->mcol = -1;                                    /* invalid column number */
  vd->title = NULL;                                          /* zero pointer */
}


void 
init_view_lst(view_lst *vl)
{
  assert(vl);

  vl->cnt = 0;
  vl->vdef = NULL;                                           /* zero pointer */
}


void
free_stats_win(stats_win *sw)
{
  assert(sw);

  init_stats_win(sw);
}

void
free_stats_prefs(stats_prefs *se)
{
  assert(se);
                           /* don't init GTK stuff */
  free_stats_win(&se->sapply);
  free_stats_win(&se->srevert);

  init_stats_prefs(se);
}

void 
free_scroll_def(scroll_def *sw)
{
  assert(sw);
                           /* don't free GTK stuff */

  init_scroll_def(sw);
}

void
free_bc_img(bc_img *bc)
{
  free_stats_win(&bc->pwin);
  free_img_typ(&bc->timg);
  free_scroll_def(&bc->scroll);
  free_pic_def(&bc->magn_img);
  free_pic_def(&bc->bicon);

  init_bc_img(bc);
}

void
free_glist_def(glist_def *gd) 
{
  assert(gd);

  g_list_free(gd->items);                       /* GTK deallocating required */
  FREEMEM(gd->fail);                            /* synchron control of GTK   */

  init_glist_def(gd);
}

void
free_entry_def(entry_def *ed)
{
  assert(ed);
                           /* don't free GTK stuff */

  ed->info_pt = NULL;                              /* pointer elaboration */

  free_glist_def(&ed->elist);

  init_entry_def(ed, ed->info_pt);
}


void
free_media_att_set(media_att_set *mas)
{
  assert(mas);

  free_media_att_entry(&mas->ety);

  init_media_att_set(mas);
}


void
free_tps_id_set(tps_id_set *tis)
{
  assert(tis);

  free_tps_id_entry(&tis->ety);

  init_tps_id_set(tis);
}


void
free_cls_att_set(cls_att_set *as)
{
  assert(as);

  free_cls_att_entry(&as->ety);

  init_cls_att_set(as);
}


void
free_media_att_entry(media_att_entry *mae)
{
  assert(mae);
                           /* don't free GTK stuff */

  free_entry_def(&mae->dpi);
  free_entry_def(&mae->method);
  free_entry_def(&mae->l_phase);
  free_entry_def(&mae->h_phase);

  init_media_att_entry(mae);
}


void
free_tps_id_entry(tps_id_entry *tie)
{
  assert(tie);
                           /* don't free GTK stuff */

  free_entry_def(&tie->id);

  init_tps_id_entry(tie);
}


void
free_cls_att_entry(cls_att_entry *ae)
{
  assert(ae);
                           /* don't free GTK stuff */

  free_entry_def(&ae->class);
  free_entry_def(&ae->translatn);
  free_entry_def(&ae->commnt);
  free_entry_def(&ae->alignment);
  free_entry_def(&ae->printable);
  free_entry_def(&ae->fixed);

  init_cls_att_entry(ae);
}


void
free_widget_lst(widget_lst *wl) 
{
  int i;

  assert(wl);

  if (!wl->wid) {                                     /* consistency check */
    assert(!wl->cnt);
  }
  else {
    assert(wl->cnt>0);
  }

  for (i = 0; i < wl->cnt; i++) {
    continue;                                      /* don't free GTK stuff */
  }
  FREEMEM(wl->wid);

  init_widget_lst(wl);
}


void 
free_view_def(view_def *vd)
{
  assert(vd);

  FREEMEM(vd->title);

  init_view_def(vd);
}


void 
free_view_lst(view_lst *vl)
{
  int i;

  assert(vl);

  if (!vl->vdef) {                                       /* consistency check */
    assert(!vl->cnt);
  }
  else {
    assert(vl->cnt>0);
  }

  for (i = 0; i < vl->cnt; i++) {
    free_view_def(&vl->vdef[i]);
  }
  FREEMEM(vl->vdef);

  init_view_lst(vl);
}


/* Puts a widget pointer to list.

Revision: 15.12.2004 E.Sack

Input : - pointer to widget list
        - pointer to widget
Output: - widget list updated
Precondition: - pointer to widget may be NULL.
*/
void
put_widget(widget_lst *wlst, GtkWidget *wid)
{
  int idx;

  assert(wlst);

  if (!wlst->wid) {                                        /* may be empty */
    assert(!wlst->cnt);                              /* must be consistent */
  }
  else {
    assert(wlst->cnt>0);                             /* must be consistent */
  }

  idx = wlst->cnt++;

  if (idx) {
    REALLOC(GtkWidget*, wlst->cnt, wlst->wid);
  }
  else {
    CALLOC(GtkWidget*, wlst->cnt, wlst->wid);
  }

  wlst->wid[idx] = wid;
}



/* Wrapper for 'gtk_entry_get_text'.

Revision: 23.06.2004 E.Sack

Input : - pointer to entry widget
Output: - pointer to entry string
Postcondition: - Pointer returned must be freed.
*/
gchar *
xplab_entry_get_text(GtkEntry *entry)
{
  const gchar *buf = NULL;
  gchar *ret_ptr = NULL;

  if (entry) {
    if (GTK_IS_ENTRY(entry) == TRUE) {
      buf = gtk_entry_get_text(entry);
      add_string(buf, &ret_ptr);                   /* common xplab interface */
    }
  }

  return ret_ptr;
}



/* Wrapper for 'gtk_entry_set_text'.

Revision: 23.06.2004 E.Sack

Input : - pointer to entry widget
        - pointer to string
Output: - entry widget updated
*/
void
xplab_entry_set_text(GtkEntry *entry, const gchar *str)
{
  if (!entry) {
    return;
  }

  if (!str) {
    return;
  }

  if (GTK_IS_ENTRY(entry) == FALSE) {
    return;
  }

  /* PC beeps, e.g. on gtk_widget_show*(), if string is too long */
  gtk_entry_set_text(GTK_ENTRY(entry), str);
}
