/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

/* defines ******************************************************************/

typedef struct {
  gboolean  display_active;                               /* window active */
  area_def  wsize;                                        /* window size   */
} stats_win;

typedef struct {
  GtkWidget *display_button;
  GtkEntry  *wentry_x, *wentry_y;
  stats_win sapply, srevert;
} stats_prefs;                                     /* preference statistic */

typedef struct {
  GtkWidget     *sdrawing, *v_viewport, *h_viewport;
  GtkAdjustment *v_adj, *h_adj;
} scroll_def;

typedef struct {
  GtkWidget    *topwin, *preview;            /* top windows                  */
  GtkWidget    *poproot, *popmenu;           /* root of pop-up menu          */
  gint         key;                          /* request key                  */
  STATS_FORMAT dfmt;                         /* requested destination format */
  gboolean     img_changed;                  /* image changed                */
  gfloat       magnitude, magnitude_old;
  img_typ      timg;                         /* requested image              */
  stats_win    pwin;                         /* window preferences           */
  scroll_def   scroll;                       /* window handling              */
  pic_def      magn_img;               /* displayed buffer for scaling image */
  pic_def      bicon;
} bc_img;                             /* destination image window definition */

typedef struct {
  GtkWidget *pbar, *plabel;
  gboolean  *pactive, *pstop;
  time_t    start_time;
} run_def;

typedef struct {
  GtkWidget *button, *ebox;
  gint       id;
} gbutton_def;

typedef struct {
  GtkTreeIter iter;
  const gchar *txt;
  gint        iaux;
  gboolean    baux;
} foreach_def;

typedef struct {
  GtkTreeIter  piter;
  gint         iaux;
  gboolean     baux, is_valid;
} treenode_def;

typedef struct {
  GList *items;
  gchar *fail;
} glist_def;                                      /* struct to supervise Gtk */

typedef struct {
  GtkWidget *combo;
  GtkWidget **info_pt;                          /* pointer copy to elaborate */
  glist_def elist;
} entry_def;

typedef struct {
  PARSE_STAT parse_status;
  entry_def  class, translatn, commnt, alignment, printable, fixed;
  GtkWidget  *infoline, *statusline3;
} cls_att_entry;

typedef struct {
  PARSE_STAT parse_status;
  entry_def  dpi, method, l_phase, h_phase;
  GtkWidget  *infoline;
} media_att_entry;

typedef struct {
  PARSE_STAT parse_status;
  entry_def  id;
  GtkWidget  *infoline;
} tps_id_entry;

typedef struct {
  int           tps_idx, class_idx;
  cls_att_entry ety;
} cls_att_set;

typedef struct {
  int             tps_idx;
  media_att_entry ety;
} media_att_set;

typedef struct {
  int          tps_idx;
  tps_id_entry ety;
} tps_id_set;

typedef struct {
  int        cnt;                             /* number of widget pointers */
  GtkWidget  **wid;                             /* list of widget pointers */
} widget_lst;                            /* definition widget pointer list */

typedef struct { 
  gint  mcol;                                 /* model column used for view */
  gchar *title;                                       /* column header text */
} view_def;                                       /* definition view column */

typedef struct { 
  int      cnt;                        /* number of view column definitions */
  view_def *vdef;                                 /* view column definition */
} view_lst;                                  /* definition view column list */

/* Prototypes ***************************************************************/

void init_foreach_def(foreach_def *fd);
void init_scroll_def(scroll_def *sw);
void init_bc_img(bc_img *bc);
void init_gbutton_def(gbutton_def *bd);
void init_run_def(run_def *rd);
void init_treenode_def(treenode_def *tn);
void init_glist_def(glist_def *gd);
void init_entry_def(entry_def *ed, GtkWidget **info);
void init_media_att_set(media_att_set *mas);
void init_media_att_entry(media_att_entry *mae);
void init_tps_id_entry(tps_id_entry *tie);
void init_tps_id_set(tps_id_set *tis);
void init_cls_att_set(cls_att_set *as);
void init_cls_att_entry(cls_att_entry *ae);
void init_stats_win(stats_win *sw);
void init_stats_prefs(stats_prefs *se);
void init_widget_lst(widget_lst *wl);
void init_view_def(view_def *vd);
void init_view_lst(view_lst *vl);

void free_scroll_def(scroll_def *sw);
void free_bc_img(bc_img *bc);
void free_cls_att_entry(cls_att_entry *ae);
void free_glist_def(glist_def *gd);
void free_entry_def(entry_def *ed);
void free_cls_att_set(cls_att_set *as);
void free_tps_id_set(tps_id_set *tis);
void free_tps_id_entry(tps_id_entry *tie);
void free_media_att_set(media_att_set *mas);
void free_media_att_entry(media_att_entry *mae);
void free_stats_win(stats_win *sw);
void free_stats_prefs(stats_prefs *se);
void free_widget_lst(widget_lst *wl);
void free_view_def(view_def *vd);
void free_view_lst(view_lst *vl);

void default_stats_prefs(stats_prefs *se);
void default_stats_win(stats_win *sw);
void default_stats_icon_contrast(stats_prefs *ic);
void default_icon_contrast(stats_win *sw);

gchar *xplab_entry_get_text(GtkEntry *entry);
void xplab_entry_set_text(GtkEntry *entry, const gchar *str);
void put_widget(widget_lst *wlst, GtkWidget *wid);

