/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

/* Usage: This header file shall not contain any GTK stuff to avoid problems
          caused by initializing or freeing variables. Use the according 
          XPLAB header file to put GTK stuff in (may be mixed with variables
          declared here).                                                   */

/* defines ******************************************************************/

typedef guint kern_def[CKERNEL_SIZE];

typedef struct { 
  double r, g, b;                            /* coefficients                */
} rgb_coef;                                  /* definition RGB coefficients */

typedef struct { 
  PICT r, g, b;                                     /* color                */
} col_def;                                          /* definition RGB color */

typedef struct {
  col_def bgd_in;                               /* background inside focus  */
  col_def bgd_out;                              /* background outside focus */
  col_def fgd;                                  /* foreground               */
  col_def sel;                                  /* selected                 */
  col_def sbl;                                  /* selectable               */
  col_def tnd;                                  /* trained                  */
  col_def res;                                  /* reserved                 */
  col_def ign;                                  /* ignored                  */
} pattern_color;                       /* pattern marking colors definition */

typedef struct {
  BATCH_PHASE phase;                                          /* what to do */
  BATCH_LOG   blog;                                    /* batch log control */
  FILE        *log_stream;                              /* batch log stream */
} batch_def;                                        /* batch job definition */

typedef struct {
  int help;
  int license;
} cli_def;                                      /* definition parameter CLI */

typedef struct { 
   gint  cnt;                                       /* array length          */
   gchar **tl;                                      /* pointer to line array */
} text_def;                                         /* definition text       */

typedef struct { 
  int  key;                                                          /* key */
  int  wac;                            /* access: 0 read only, 1 read/write */
  char *cmt;                                             /* request comment */
  void *aux;                          /* pointer copy not meant to be freed */
} key_def;

typedef struct { 
  int     cnt;                                            /* number of keys */
  key_def *kar;
} key_lst;

typedef struct { 
   int width;
   int height;
} area_def;                                              /* definition area */

typedef struct { 
  SIGNAL_TYPE sig_type;
  double      aux;
} signal_def;                                          /* definition signal */

typedef struct { 
  int idx; 
  int val;
} sort_int;                    /* definition sort indizes regarding integer */

typedef struct { 
  int      cnt;                             /* array length                 */
  sort_int *il;                             /* pointer to array             */
} sort_ilst;                                /* definition sort integer list */

typedef struct { 
  int    idx; 
  double val;
} sort_dbl;                     /* definition sort indizes regarding double */

typedef struct { 
  int      cnt;                              /* array length                */
  sort_dbl *dl;                              /* pointer to array            */
} sort_dlst;                                 /* definition sort double list */

typedef struct { 
  int    cnt;                                  /* array length              */
  double *dbl;                                 /* pointer to double numbers */
} double_lst;                                  /* definition double list    */

typedef struct { 
  double maxval;                               /* max. histogram value...   */
  int    max_bin;                              /*      ...appurtenant bin   */
  int    th_bin;                               /* threshold                 */
  int    start_bin;                            /* histogram start reference */
  int    set_cnt;                              /* sample size               */
  double_lst hst;
} hst_def;                                     /* definition histogram      */

typedef struct { 
  int      idx;                                    /* index of ring buffer  */
  text_def td;                                     /* text ring buffer      */
} text_ring;                                       /* definition text ring  */

typedef struct { 
  int cnt;                                       /* array length            */
  int *nr;                                       /* pointer to array        */
} int_lst;                                       /* definition integer list */

typedef struct { 
   int     cnt;                        /* array length                      */
   int_lst *il;                        /* pointer to array                  */
} int_lstm;                            /* definition multiple integer lists */

typedef struct { 
   int nr;                                    /* entry number               */
   int aux;                                   /* entry auxiliary            */
} int2;                                       /* definition of two integers */

typedef struct { 
   int  cnt;                             /* number of entries               */
   int2 *n;                              /* pointer to entries              */
} int2_lst;                              /* definition of two integers list */

typedef struct { 
  double  mv;                 /* mean value                                 */
  double  sd;                 /* standard deviation                         */
} stc_def;                    /* definition statistic classification number */

typedef struct { 
  stc_def st;                           /* statistic classification numbers */
  hst_def hd;                           /* histogram                        */
} hst_info;                             /* definition histogram info        */

typedef struct { 
   int x;
   int y;
} c_def;                                           /* definition coordinate */

typedef struct { 
   int tps_idx;
   int cls_idx;
   int nb_idx;
} idx_def;                                         /* definition DB indexes */

typedef struct {
  area_def ar;                                            /* 2-D size       */
  double   *mem;                                          /* memory pointer */
} pic_dbl;                                                /* 2-D definition */

typedef struct {
  c_def    org;                /* coordinate origin left upper image corner */
  area_def ar;                 /* image size                                */
  PICT     *mem;               /* pointer to bitmap                         */
  PICT     *bvl;               /* pointer to bitmap b-values                */
  col_def  *col;               /* pointer to color image                    */
} pic_def;                     /* image definition                          */

typedef struct { 
  int start;                                         /* inclusively         */
  int end;                                           /* inclusively         */
} ival_def;                                          /* definition interval */

typedef struct {
  ival_def line;             
  ival_def column;
} sector_def;                                     /* definition sector list */

typedef struct { 
  gint       cidx, nidx;                                    /* indexes      */
  c_def      afs;                                           /* coordinate   */
  sector_def sec;                                           /* sector       */
} nb_aux;                                           /* definition auxiliary */

typedef struct {
  int    cnt;                                  /* number of entries         */
  nb_aux *dat;                                 /* pointer to array          */
} aux_lst;                                     /* definition auxiliary list */

typedef struct {
  aux_lst il;                  /* image location list                       */
  pic_def eimg;                /* image definition                          */
} img_ext;                     /* extended image definition                 */

typedef struct {
  PNM_FORMAT typ;              /* image type                                */
  pic_def    pimg;             /* image definition                          */
} img_typ;                     /* image definition with type                */

/* RE & IM parts of complex values are stored in consecutive order          */
typedef struct {
  area_def ar;               /* dimension X x Y                             */
  double   **data;           /* complex 2D-spectrum [X][Y]                  */
  double   *spec;            /* complex Fx/2 of spectrum refers to X        */
} fft_def;                   /* definition FFT parameter of real 2-D signal */

typedef struct {
  int     cnt;                               /* number of entries           */
  fft_def *fp;                               /* partial FRB                 */
} part_lst;                                  /* definition partial FRB list */

typedef struct {
  int      c;                                /* class number                */
  int      n;                                /* NB number                   */
  area_def lub;                              /* relevant correlation values */
  double   kwm;                              /* correlation value           */
} lag_def;                                   /* definition lag              */

typedef struct {
  int     cnt;                                       /* number of entries   */
  lag_def *ld;                                       /* KW                  */
} lag_lst;                                           /* definition lag list */

typedef struct {
  area_def ar;                                /* dimension                  */
  lag_lst  **ll;                              /* 2D correlation list [X][Y] */
} kwh_def;                                    /* definition KWH             */

typedef struct {
  int      cnt;                                      /* number of entries   */
  kwh_def *kwh;                                      /* KWH list            */
} kwh_lst;                                           /* definition KWH list */

typedef struct {
  int cdx;                                           /* class index         */
  int ndx;                                           /* NB index            */
  int x, y;                                          /* offset in CNB       */
} nb_chn_def;                                        /* NB chain definition */

typedef struct {
   area_def ar;                                  /* size                    */
   PICT     *patt;                               /* pointer to pattern      */
} cc_def;                                        /* definition class kernel */

typedef struct {
  pic_def pic;                    /* NB with mask frame                     */
  pic_def fld;                    /* filled NB with mask frame              */
  pic_def wtd;                    /* weighted NB with mask frame            */
  pic_def msk;                    /* NB mask with mask frame                */
  pic_def dlt;                    /* dilated NB with mask frame             */
  double  nbm;                    /* NB mass                                */
  double  fldm;                   /* mass of filled NB                      */
  double  dltm;                   /* mass of dilated NB                     */
  double  wtdm;                   /* mass of weighted NB                    */
  double  mskm;                   /* mass of NB mask                        */
  double  cmin;                   /* minimal correlation value              */
} nb_ref;                         /* definition NB references               */
 
typedef struct { 
  int        cnt;                      /* number of chains                  */
  nb_chn_def *cd;                      /* pointer to chain entries          */
} nb_chn_lst;                          /* definition of NB chain entry list */

typedef struct { 
  int    cnt;                            /* number of class kernels         */
  cc_def *cc;                            /* pointer to kernel entries       */
} cc_lst;                                /* definition of kernel entry list */

typedef struct {
  int      cnt;                                 /* number of interval       */
  ival_def *ival;                               /* pointers to intervals    */
} ival_lst;                                     /* definition interval list */

typedef struct {
  int        cnt;                                 /* number of sectors      */
  sector_def *sdl;                                /* pointer to sectors     */
} sector_lst;                                     /* definition sector list */

typedef struct { 
  pic_def    pat;          /* raw binary NB pattern                         */
  nb_chn_lst cnb;  /* NB is a child: CNB chain list (forward ref., parents) */
  nb_chn_lst nnb;  /* NB is a parent: NNB chain list (backward ref.,children)*/
  nb_ref     ref;          /* references, not saved in database             */
  int        aux;          /* auxiliary, not saved in database              */
  nb_aux     dat;          /* auxiliary features, not saved in database     */
} nb_def;                  /* definition named box NB                       */

typedef struct { 
  int    cnt;                            /* number of NBs                   */
  nb_def *nb;                            /* pointer to NB entries           */
} nb_lst;                                /* definition of NB (pattern) list */

typedef struct { 
  int l_phase;                              /* period L Phase               */
  int h_phase;                              /* period H Phase               */
} period_def;                               /* period definition, P = H + L */

typedef struct {
  int        opos;                   /* optimal position of period in list  */
  period_def period;                 /* (estimated) period of position list */
} linespacing_par;                   /* inline method parameter             */

typedef struct { 
  INLINE_METHOD   method;                      /* methods...                */
  linespacing_par linespacing;                 /* ...and their parameter    */
} inline_def;                                  /* inline methods definition */

typedef struct {
  char    *class;                            /* class name                  */
  char    *sign;                             /* class translation           */
  char    *crem;                             /* class comment               */
  ALIGMNT alignment;                         /* class alignment to baseline */
  int     printable;                         /* class printability          */
  int     fixed;                             /* 0 proportional, 1 fixed     */
} cls_att;                                   /* definition class attributes */

typedef struct {
  INH_FLAG   inh;                                 /* inhibition source flag */
  sector_def isd;                                 /* location               */
} inh_def;                                        /* definition inhibition  */

typedef struct {
  int     cnt;                               /* number of inhibitions       */
  inh_def *ind;                              /* pointer to inhibition array */
} inh_lst;                                   /* definition inhibition list  */

typedef struct {
  cls_att att;            /* class attributes                               */
  nb_lst  nbl;            /* NB entry list                                  */
  cc_lst  ccl;            /* class kernel entry list, not saved in database */
  inh_lst inl;            /* inhibition list, not saved in database         */
} cls_def;                /* definition class                               */

typedef struct { 
  int     cnt;                                  /* number of classes        */
  cls_def *cls;                                 /* pointer to classes       */
} cls_lst;                                      /* definition of class list */

typedef struct { 
  int        dpi;                            /* resolution DPI              */
  inline_def in_line;                        /* text line features          */
} media_att;                                 /* definition media attributes */

typedef struct { 
  char      *ident;                   /* pointer to ID string for TPS       */
  media_att media;                    /* media attributes                   */
  cls_lst   cl;                       /* class list                         */
} tps_def;                            /* definition trained pattern set TPS */

typedef struct { 
  int     cnt;                                       /* number of TPSs      */
  tps_def *tps;                                      /* pointer to TPSs     */
} tps_lst;                                           /* definition TPS list */

typedef struct {
  char    *db_fname;                                       /* DB file name  */
  char    *db_version;                                     /* DB version    */
  tps_lst tpl;                                             /* TPS list      */
} db_def;                                                  /* definition DB */

typedef struct { 
  int     lspace;                               /* linespace                */
  int_lst bline;                                /* baselines                */
} baseline_lst;                                 /* definition baseline list */

typedef struct {
  int        raux;
  RB_STATUS  rld;                       /* book status                      */
  sector_def sc;                        /* RB, inclusive                    */
  ival_lst   gridy;                     /* RB vertical subdivision of lines */
} rb_def;                               /* definition raw box RB            */

typedef struct { 
  area_def rar;                                       /* size of RB image   */
  int      cnt;                                       /* number of RBs      */
  rb_def   *rb;                                       /* pointer to RBs     */
} rb_lst;                                             /* definition RB list */

typedef struct { 
  FILE     *report_stream;                         /* report file stream ID */
  char     *xprog;                                        /* called program */
  char     *gtk_rc;                                /* Gtk resource filespec */
  char     *config_filespec;                     /* configuration file name */
  char     *image_filename;                              /* image file name */
  char     *user_options;                        /* CLI user options string */
  int      xpid;                                      /* program process ID */
  gboolean dobatch;                                /* batch job enable flag */
  rb_lst   rblst;                                                /* RB list */
  hst_def  grey_hist;                     /* histogram of grey-scaled image */
  hst_def  th_hist;                 /* method used to get threshold for BMP */
} set_def;                                   /* definition runtime settings */

typedef struct {
  int    mborder;
  int    nborder;
  int    nnbmatch;
  int    maskweight;
  double corrlevel;
  double corrbroken;
  double overlap;
  double nb_lap;
  double interleave;
} match_def;                              /* definition matching parameters */

typedef struct { 
  char            *tps_id;
  MATCHING_METHOD match_method;
  INLINE_METHOD   line_method;
  int             renew_method;
  int             support_dots;
  int             support_dots_auto;
  int             overwrite_image;
  int             single_line;
  int             single_nb;
  int             duration;
  int             dc_store;
  RB_CHAIN_BIT    rb_chain;
  int             report_write;
  int             histo_write;
  int             verbose;
  int             raise;
  int             tips;
  int             cdomain;
  int             timestamp;
  int             writeauto;
  double          syspc;
  int             replay_write;
  int             training_write;
  int             cluster_write;
  CLUSTER_NB_BIT  cluster_nb;
  int             cluster_force;
  double          cluster_step;
  double          cluster_level;
  int             nb_border_x;
  int             nb_border_y;
  int             histo_thold;
  int             rawbit;
  char            *subst_rb;
  char            *subst_nb;
  rgb_coef        rgb_norm;
  kern_def        ckernel;
  int             cluster_thold;
  int2_lst        dc_1d, dc_2d;
  batch_def       batch;                             /* for batch jobs only */
  match_def       mad;                               /* matching parameters */
} cfg_def;                                      /* definition configuration */

typedef struct {
  set_def set;                                          /* dynamic settings */
  cfg_def cfg;                                           /* static settings */
} prog_ctrl;                                      /* definition preferences */

typedef struct { 
  int      yco;                            /* ordinate                      */
  ival_def xil;                            /* abscissa interval x1-x2       */
} h_contour;                               /* definition horizontal contour */

typedef struct { 
  int      xco;                              /* abscissa                    */
  ival_def yil;                              /* ordinate interval y1-y2     */
} v_contour;                                 /* definition vertical contour */

typedef struct {
  CONTOUR_TYPE type;                 /* contour type                        */
  int          cnt;                  /* size of contour list                */
  h_contour    *hcl;                 /* pointer to horizontal contour lines */
} h_contour_lst;                     /* definition horizontal contour list  */

typedef struct { 
  CONTOUR_TYPE type;                   /* contour type                      */
  int       cnt;                       /* size of contour list              */
  v_contour *vcl;                      /* pointer to vertical contour lines */
} v_contour_lst;                       /* definition vertical contour list  */

typedef struct { 
   int good;                                /* conforming values            */
   int kernel;                              /* values of NB pattern         */
} cv_def;                                   /* definition counting features */

typedef struct { 
   double cv;                                          /* correlation value */
   int    x;                                           /* lag               */
   int    y;
} cv_def_i;                  /* definition correlation values (int version) */

typedef struct { 
   double cv;                                          /* correlation value */
  double x;                                            /* lag               */
   double y;
} cv_def_f;                /* definition correlation values (float version) */

typedef struct {
  MATCHING_METHOD mtd;              /* matching method succeeded            */
  cv_def_i        m;                /* correlation features                 */
  int             rbn;              /* RB number (anchor)                   */
  int             c;                /* class number                         */
  int             n;                /* NB number                            */
  ival_def        line;             /* RB line (subdivision)                */
  sector_def      roi;              /* correlation area                     */
  sector_def      ext;              /* extended correlation area            */
  sector_def      knl;              /* NB kernel sector                     */
  sector_def      hdx;              /* hash index                           */
  int_lst         crb;              /* CRB list                             */
  struct cand_lst *alt;             /* candidates with competing alignments */
  int             aux;              /* auxiliary item makes life easier     */
} cand_def;                         /* definition candidate                 */

typedef struct cand_lst {                      /* K&R A.8.3                 */
  int      cnt;                                /* number of candidates      */
  cand_def *c;                                 /* correlation features      */
} cand_lst;                                    /* definition candidate list */

typedef struct { 
  int      cnt;                          /* number of candidate lists       */
  cand_lst *clst;                        /* pointer to candidate list       */
} cand_lst_ary;                          /* definition candidate list array */

typedef struct {
  REGISTER_INDEX use;        /* register flag usability                     */
  ival_def y_range;          /* key describing vertical range of cand. list */
  char     *pic_state;       /* key describing image state of VCL           */
  cand_lst vlst;             /* vertical candidate list VCL                 */
} y_reg_def;                 /* definition candidate list register          */

typedef struct {
  int       cnt;                   /* number of registered candidate lists  */
  y_reg_def *yrd;                  /* pointer to registered candidate lists */
} y_reg_lst;                       /* definition vertical register          */

typedef struct {
  int       cnt;                       /* number of registered columns      */
  y_reg_lst *yrl;                      /* pointer to vertical register list */
} reg_def;                             /* definition horizontal register    */

typedef struct { 
  pic_def  *rpic;             /* RB image                                   */
  pic_def  wpic;              /* weighted RB image                          */
  pic_def  shdw1;             /* working buffer for RB use                  */
  pic_def  shdw2;             /* working buffer for RB use                  */
  kwh_lst  kl;                /* KWH list                                   */
  reg_def  xreg;              /* candidate list register                    */
  char    *pic_state;         /* cut out pattern key describing image state */
} rb_ref;                     /* definition RB references                   */

typedef struct {
  char       *chain_name;         /* pointer to chain image filename        */
  sector_def chain_frame;         /* chain list sector in original document */
  rb_lst     rb_chain;            /* RB chain list is a RB list             */
  cand_lst   bhclst;              /* best horizontal substring              */
  int        caux;                /* auxiliary flag                         */
} rb_chn_def;                     /* RB chain definition                    */

typedef struct { 
  int        cnt;                               /* number of RB chains      */
  rb_chn_def *rcd;                              /* pointer to RB chain list */
} rb_chn_lst;                                   /* RB chain list definition */

typedef struct { 
  rb_chn_def doc;                               /* document image           */
  int        border;                            /* extension of tile frames */
  rb_chn_lst rch;                               /* RB tiles                 */
} tiles_def;                                    /* RB tiles definition      */

typedef struct {
  CLUSTER_NB_BIT ctyp;                             /* type                  */
  int            cwrit;                            /* enable images         */
  int            cforc;                            /* forced                */
  double         clvl;                             /* level                 */
  double         cstp;                             /* step                  */
  kern_def       cknl;                             /* kernel                */
  int            cthd;                             /* threshold             */
} cluster_def;                                     /* definition clustering */

typedef struct {
  int      cnt;         /* number of entries                                */
  int      *c;          /* class index in DB                                */
  int      *n;          /* NB index in DB                                   */
  cand_def *hl;         /* correlation features child -> parent             */
} gen_def;              /* definition NB generation parent[n] -> child[n+1] */

typedef struct {
  int     cnt;                             /* number of entries             */
  gen_def *gd;                             /* generation list               */
} gen_lst;                                 /* definition NB generation list */

typedef struct { 
  int      cnt;                          /* number of entries               */
  int      *cid;                         /* birth ID of child               */
  int      *pid;                         /* birth ID of parent              */
  cand_def *kr;                          /* birth dates of child            */
} hash_def;                              /* definition generation reference */

typedef struct { 
  int      cnt;          /* number of hash entries                          */
  hash_def *nid;         /* pointer to hash list                            */
} hash_lst;              /* definition generation list (history of cluster) */

typedef struct { 
  sector_def sid;           /* hash index (values <0 allowed                */
  double     k;             /* CV of candidate list                         */
  sector_def o;             /* overall sector of candidate list             */
  int        g;             /* gap of most left cand. to the next right one */
  int        lc;            /* class number of most left candidate          */
  int        ln;            /* NB number of most left candidate             */
  sector_def ls;            /* kernel place of most left candidate          */
  cand_lst   r;             /* substring                                    */
} rc_def;                   /* definition string recycling                  */

typedef struct {
  int    cnt;                               /* number of entries            */
  rc_def *s;                                /* pointer to recycling strings */
} rc_lst;                                   /* definition recycling list    */

/* Prototypes ***************************************************************/

float **alloc_fmatrix(int nx, int ny);
fft_def alloc_fft_def(int w, int h);
kwh_def alloc_kwh_def(area_def ar);
cv_def_i **alloc_cv_def_i_matrix(int vector_len);

col_def convert_rgb_coef(rgb_coef in);
int compare_col_def(col_def a, col_def b);

void init_kern_def(kern_def kd);
void init_pattern_color(pattern_color *pc);
void init_batch_def(batch_def *bd);
void init_set_def(set_def *sd);
void init_match_def(match_def *md);
void init_cfg_def(cfg_def *cd);
void init_prog_ctrl(prog_ctrl *ctrl);
void init_cli_def(cli_def *ci);
void init_idx_def(idx_def *id);
void init_area_def(area_def *ad);
void init_key_def(key_def *kd);
void init_key_lst(key_lst *kl);
void init_cluster_def(cluster_def *cd);
void init_baseline_lst(baseline_lst *bl);
void init_c_def(c_def *cd);
void init_sort_ilst(sort_ilst *sl);
void init_sort_dlst(sort_dlst *sl);
void init_nb_lst(nb_lst *nl);
void init_aux_lst(aux_lst *al);
void init_nb_chn_def(nb_chn_def *ncd);
void init_linespacing_par(linespacing_par *lp);
void init_inline_def(inline_def *il);
void init_period_def(period_def *pd);
void init_pic_dbl(pic_dbl *pd);
void init_img_ext(img_ext *ie);
void init_img_typ(img_typ *it);
void init_pic_def(pic_def *pd);
void init_part_lst(part_lst *pl);
void init_lag_lst(lag_lst *ll);
void init_kwh_def(kwh_def *kw);
void init_kwh_lst(kwh_lst *kl);
void init_fft_def(fft_def *fd);
void init_stc_def(stc_def *st);
void init_hst_def(hst_def *hd);
void init_hst_info(hst_info *hi);
void init_int_lst(int_lst *il);
void init_double_lst(double_lst *dl);
void init_text_def(text_def *td);
void init_text_ring(text_ring *tr);
void init_int_lstm(int_lstm *im);
void init_int2_lst(int2_lst *i2l);
void init_col_def(col_def *cl);
void init_tps_def(tps_def *tps);
void init_sector_def(sector_def *sec);
void init_sector_lst(sector_lst *sl);
void init_cls_att(cls_att *ca);
void init_cls_def(cls_def *cls);
void init_cls_lst(cls_lst *cl);
void init_nb_def(nb_def *nb);
void init_nb_aux(nb_aux *na);
void init_cc_def(cc_def *cc);
void init_rb_def(rb_def *rb);
void init_rb_lst(rb_lst *rl);
void init_rb_chn_def(rb_chn_def *rc);
void init_tiles_def(tiles_def *rl);
void init_ival_def(ival_def *ivl);
void init_ival_lst(ival_lst *il);
void init_h_contour(h_contour *cln);
void init_v_contour(v_contour *cln);
void init_h_contour_lst(h_contour_lst *ctr);
void init_v_contour_lst(v_contour_lst *ctr);
void init_cv_def_i(cv_def_i *kd);
void init_cand_def(cand_def *kr);
void init_nb_ref(nb_ref *nref);
void init_rc_def(rc_def *rc);
void init_rc_lst(rc_lst *rl);
void init_cand_lst(cand_lst *rc);
void init_rb_ref(rb_ref *rref);
void init_db_def(db_def *db);
void init_hash_lst(hash_lst *cdef);
void init_hash_def(hash_def *hdef);
void init_gen_lst(gen_lst *gl);
void init_gen_def(gen_def *gd);
void init_media_att(media_att *md);
void init_tps_lst(tps_lst *tpl);
void init_cc_lst(cc_lst *cl);
void init_nb_chn_lst(nb_chn_lst *cl);
void init_rb_chn_lst(rb_chn_lst *rcl);
void init_reg_def(reg_def *rd);
void init_cand_lst_ary(cand_lst_ary *cla);
void init_y_reg_lst(y_reg_lst *yrl);
void init_y_reg_def(y_reg_def *yrd);
void init_inh_lst(inh_lst *il);
void init_inh_def(inh_def *id);
void init_rgb_coef(rgb_coef *rgb);
DATA_STATE state_pic_def(pic_def pd);
DATA_STATE state_inh_lst(inh_lst il);
DATA_STATE state_tps_lst(tps_lst tpl);
DATA_STATE state_db_def(db_def db);

pattern_color default_pattern_color(void);
void xdefault_kern_def(kern_def kd);
char *default_session_filename(char *fname);
char *default_replay_filename(char *fname);
char *default_report_filename(char *fname);
char *default_auto_filename(char *fname);
char *default_config_basename(char *fname);
char *default_subst_rb(void);
char *default_subst_nb(void);
char *default_tps_id(void);
void xdefault_prog_ctrl(prog_ctrl *pc);
void xdefault_set_def(set_def *sd);
void xdefault_cfg_def(cfg_def *cd);
void xdefault_rgb_coef(rgb_coef *rgb);
void xdefault_batch_def(batch_def *bd);
void xdefault_match_def(match_def *md);

int crop_double_lst(double_lst src, double_lst dst, int scr_offset, 
		    int dst_offset, int cnt);
void copy_kern_def(kern_def src, kern_def dst);
void copy_batch_def(batch_def *src, batch_def *dst);
void copy_rb_lst(rb_lst *src, rb_lst *dst);
void copy_prog_ctrl(prog_ctrl *src, prog_ctrl *dst);
void copy_set_def(set_def *src, set_def *dst);
void copy_match_def(match_def *src, match_def *dst);
void copy_cfg_def(cfg_def *src, cfg_def *dst);
void copy_int_array(int *src, int *dst, int len);
void copy_baseline_lst(baseline_lst *src, baseline_lst *dst);
void copy_fft_def(fft_def src, fft_def dst);
int copy_double(double *src, double *dst, int cnt);
void copy_db_def(db_def *src, db_def *dst, gboolean no_chain);
void copy_tps_lst(tps_lst *src, tps_lst *dst, gboolean no_chain);
void copy_tps_def(tps_def *src, tps_def *dst, gboolean no_chain);
void copy_cls_def(cls_def *src, cls_def *dst, gboolean no_chain);
void copy_cls_lst(cls_lst *src, cls_lst *dst, gboolean no_chain);
void copy_nb_def(nb_def *src, nb_def *dst, gboolean no_chain);
void copy_nb_aux(nb_aux *src, nb_aux *dst);
void copy_key_lst(key_lst *src, key_lst *dst);
void copy_cc_def(cc_def *src, cc_def *dst);
void copy_rb_def(rb_def *src, rb_def *dst);
void copy_cand_def(cand_def *src, cand_def *dst);
void copy_cls_att(cls_att src, cls_att *dst);
void copy_img_typ(img_typ *src, img_typ *dst);
void copy_pic_def(pic_def *src, pic_def *dst);
void copy_ival_lst(ival_lst *src, ival_lst *dst);
void copy_sector_def(sector_def *src, sector_def *dst);
void copy_rb_chn_def(rb_chn_def *src, rb_chn_def *dst);
void copy_hst_def(hst_def *src, hst_def *dst);
void copy_media_att(media_att src, media_att *dst);
void copy_nb_lst(nb_lst *src, nb_lst *dst, gboolean no_chain);
void copy_cc_lst(cc_lst *src, cc_lst *dst);
void copy_double_lst(double_lst *src, double_lst *dst);
void copy_int_lst(int_lst *src, int_lst *dst);
void copy_int2_lst(int2_lst *src, int2_lst *dst);
void copy_nb_chn_lst(nb_chn_lst *src, nb_chn_lst *dst);
void copy_nb_chn_def(nb_chn_def *src, nb_chn_def *dst);
void copy_nb_ref(nb_ref *src, nb_ref *dst);
void copy_sector_lst(sector_lst *src, sector_lst *dst);
void copy_rc_def(rc_def *src, rc_def *dst);
void copy_rc_lst(rc_lst *src, rc_lst *dst);
void copy_cand_lst(cand_lst *src, cand_lst *dst);
void copy_inh_lst(inh_lst *src, inh_lst *dst);
void copy_inh_def(inh_def *src, inh_def *dst);
void copy_key_def(key_def *src, key_def *dst);
pic_def xdup_pic_def(pic_def src);
img_typ xdup_img_typ(img_typ src);

void move_rc_lst(rc_lst *src, rc_lst *dst);
void move_nb_def(nb_def *src, nb_def *dst, gboolean no_chain);
void move_cls_def(cls_def *src, cls_def *dst, gboolean no_chain);
void move_cand_lst(cand_lst *src, cand_lst *dst);
void move_cand_def(cand_def *src, cand_def *dst);
void move_double_lst(double_lst *src, double_lst *dst);
void move_key_def(key_def *src, key_def *dst);

void insert_int_lst(int_lst *il, int new_idx);
void insert_int_lstm(int_lstm *im, int new_idx);
int misc_rb_chn_def(rb_chn_lst *rcl, int src_idx, int dst_idx);

int cat_nb_def(nb_def src, nb_lst *dst_lst, gboolean no_chain);
void cat_nb_aux(nb_aux src, aux_lst *dst_lst);
void cat_rb_def(rb_def src, rb_lst *dst_lst);
void cat_rb_lst(rb_lst src, rb_lst *dst);
void cat_cand_def(cand_def src, cand_lst *dst_lst);
void cat_cand_lst(cand_lst *src, cand_lst *dst);

void delete_int2(int2_lst *il, int idx);
void delete_rb_chn_def(rb_chn_lst *rcl, int idx);
void delete_cand_def(cand_lst *cl, int idx);
void delete_key_def(key_lst *kl, int idx);
int erase_key(key_lst *klst, int *key, char **cmt);

void free_batch_def(batch_def *bd);
void free_prog_ctrl(prog_ctrl *ctrl);
void free_set_def(set_def *sd);
void free_match_def(match_def *md);
void free_cfg_def(cfg_def *cd);
void free_key_def(key_def *kd);
void free_key_lst(key_lst *kl);
void free_baseline_lst(baseline_lst *bl);
void free_sort_ilst(sort_ilst *sl);
void free_sort_dlst(sort_dlst *sl);
void free_hst_def(hst_def *hd);
void free_ival_lst(ival_lst *il);
void free_sector_lst(sector_lst *sl);
void free_part_lst(part_lst *pl);
void free_lag_lst_2D(lag_lst ***lag, area_def ar);
void free_kwh_def(kwh_def *kw);
void free_kwh_lst(kwh_lst *kl);
void free_lag_lst(lag_lst *ll);
void free_fft_def(fft_def *fd);
void free_cv_def_i_matrix(cv_def_i ***matrix, int vector_len);
void free_hst_info(hst_info *hi);
void free_int_lst(int_lst *il);
void free_double_lst(double_lst *dl);
void free_int_lstm(int_lstm *im);
void free_int2_lst(int2_lst *i2l);
void free_rc_def(rc_def *rd);
void free_rc_lst(rc_lst *rlist);
void free_tps_def(tps_def *tps);
void free_cls_att(cls_att *ca);
void free_cls_def(cls_def *cls);
void free_cls_lst(cls_lst *cl);
void free_nb_def(nb_def *nb);
void free_cc_def(cc_def *cc);
void free_nb_ref(nb_ref *nr);
void free_pic_dbl(pic_dbl *pd);
void free_pic_def(pic_def *pd);
void free_img_ext(img_ext *ie);
void free_img_typ(img_typ *it);
void free_cand_lst(cand_lst *kl);
void free_cand_def(cand_def *kr);
void free_rc_lst(rc_lst *rlist);
void free_rb_def(rb_def *rd);
void free_rb_lst(rb_lst *rl);
void free_rb_chn_def(rb_chn_def *rc);
void free_tiles_def(tiles_def *rl);
void free_rb_ref(rb_ref *rr);
void free_db_def(db_def *db);
void free_hash_lst(hash_lst *cdef);
void free_hash_def(hash_def *hdef);
void free_gen_lst(gen_lst *gl);
void free_gen_def(gen_def *gd);
void free_fmatrix(int ny, float **m);
void free_tps_lst(tps_lst *tpl);
void free_nb_lst(nb_lst *nl);
void free_nb_aux(nb_aux *na);
void free_aux_lst(aux_lst *al);
void free_cc_lst(cc_lst *cl);
void free_nb_chn_lst(nb_chn_lst *ncl);
void free_rb_chn_lst(rb_chn_lst *rcl);
void free_text_def(text_def *td);
void free_text_ring(text_ring *tr);
void free_reg_def(reg_def *rd);
void free_y_reg_lst(y_reg_lst *yrl);
void free_y_reg_def(y_reg_def *yrd);
void free_cand_lst_ary(cand_lst_ary *cla);
void free_inh_lst(inh_lst *il);
void free_inh_def(inh_def *id);
void free_h_contour_lst(h_contour_lst *cl);
void free_v_contour_lst(v_contour_lst *vl);

void update_int2_lst(int2_lst src, int2_lst *dst, int src_shift_to_aux);
double min_double_left(double_lst dl);
double max_double_left(double_lst dl);
double max_double_right(double_lst dl);
int min_index_double_left(double *dbl, int cnt);
int max_index_double_left(double *dbl, int cnt);
int max_index_double_right(double *dbl, int cnt);
int max_int_lst(int_lst il);
int max_index_int_lst(int_lst il);
int_lst max_index_double_all(double *dbl, int cnt);
int set_int_lst(int_lst il, int to, int nrignore);

void mult_double(double *dbl, int cnt, double mult);
void add_two_pic_dbl(pic_dbl dbl1, pic_dbl dbl2);
void store_pic_dbl(pic_dbl dbl1, pic_dbl dbl2);
void add_two_hst_def(hst_def hd1, hst_def *hd2);
int add_character(char src, char **dst);
int add_string(const char *src, char **dst);
void cat_string(const char *prefix, const char *root, const char *suffix, char **string_out);
void descending_sort_ilst(sort_ilst sil);
void ascending_sort_ilst(sort_ilst sil);
void ascending_sort_dlst(sort_dlst sdl);
int compare_cand_def(const void *cd1, const void *cd2);

int string_to_dpi(const char *in, int *out);
int string_to_l_phase(const char *in, int *out);
int string_to_h_phase(const char *in, int *out);
void int_to_string(char *prefix, int in, char *suffix, char **string_out);
void double_to_string(char *prefix, double in, char *suffix, char **string_out);
int aligmnt_to_string(ALIGMNT enum_in, char **string_out);
int string_to_aligmnt(const char *string_in, ALIGMNT *enum_out);
int aligmnt_to_int(ALIGMNT enum_in, int *int_out);
int int_to_aligmnt(int int_in, ALIGMNT *enum_out);
int inline_method_to_string(INLINE_METHOD enum_in, char **string_out);
int string_to_inline_method(const char *string_in, INLINE_METHOD *enum_out);
int put_key(key_lst *klst, int *key, char *key_comment, int write_access, void *ptr);
void put_sector_lst(sector_lst *slst, sector_def sdef);
void copy_to_text_def(text_def *tdef, const gchar *txt);
int request_key_lst(key_lst klst);
int check_name(const char *name);
int search_aux_nb_def(nb_lst nbl, int item);
pic_dbl pic_def_to_pic_dbl(pic_def ipic);
