/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <time.h>
#include <stdarg.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_algorithm.h"
#include "xplab_output.h"
#include "xplab_report.h"
#include "xplab_maintenance_fe.h"
#include "xplab_image.h"
#include "xplab_tprint.h"
#include "xplab_dchannel.h"
#include "xplab_preferences_fe.h"
#include "xplab_dbio.h"
#include "xplab_training_fe.h"

/* local prototypes ********************************************************/
static void getstr_period_def(period_def pd, char **info);
static void getstr_linespacing_par(linespacing_par lp, char **info);
static void print_period_def(char *prefix, period_def pd, char **info);
static void print_linespacing_par(linespacing_par lpar);
static int calc_line_spacing(char *filename, double_lst plst, 
			     linespacing_par *linespacing, int renew, 
			     run_def *crun);
static int optimal_period_point(char *fname, double_lst plst, 
				period_def period, 
				linespacing_par *optimal_linespace);
static ival_def max_period_sum(double_lst slst, int lag, period_def period, 
			       double *sum_max);
static void print_signal_type(char *prefix, signal_def stype, char *suffix, 
			      char **info);
static int type_to_unipolar_rectangle(double_lst ilst);
static int phase_estimation_list(double_lst ilst, hst_def hist_max, 
				 int h_phase_proposal, int *lag_phase,
				 period_def *pout, pic_dbl *phase_img, 
				 run_def *crun);
static int phase_estimation(double_lst ilst, int period_len, 
			    int h_phase_proposal, int *lag_phase,
			    period_def *period, pic_dbl *phase_img, 
			    run_def *crun);
static void outline_neighbor(cand_lst hlist, cls_def *cls, int_lst *hnr);
static void sort_lines(int_lst blst, int dmax);
static void store_line(int_lst hnr, int_lstm *iln, int bline, int_lst *baselines);
static int seek_start_point(double_lst plst, int seek_range, int *seek_start);
static int topmost_symbol(cand_lst hlist, int_lst *hnr, int no_alt);
static int nearest_left_upper(cand_lst hlist, int_lst *hnr, int dut_i);
static int nearest_right_upper(cand_lst hlist, int_lst *hnr, int dut_i);
static void nearest_left_neighbor(cand_lst hlist, cls_def *cls, int_lst *hnr, 
				  int left_inline, int no_alt);
static void nearest_right_neighbor(cand_lst hlist, cls_def *cls, int_lst *hnr, 
				   int right_inline, int no_alt);
static void outmost_neighbor(cand_lst hlist, int_lst hnr, int *left_inline, 
			     int *right_inline);
static void inline_neighbor(cand_lst hlist, cls_def *cls, int_lst *nr, 
			    int upper_hnr, int no_alt);
static void collect_inline_neighbor(cand_lst hlist, cls_def *cls, int_lst *hnr, 
			            int no_alt);
static int inline_rule(cand_lst hlist, cls_def *cls, int upper_hnr, int lower_hnr);
static char *separate_line(int_lst znr, cls_def *cls, cand_lst hlist, char *fname, 
	 	           int noise_border, int lcnt);
static void xwrite_text(char *suffix, text_def text, GtkWidget *ilog);
static int_lst gen_width(cand_lst hlist, int_lst znr);
static int_lst gen_distance(cand_lst hlist, int_lst znr);
static void separate_groups(int_lst wlst, int_lst dlst);
static void separate_words(double dt, int_lst dlst, hst_info **hi, int noise_border);
static void gen_line(cls_def *cls, char **txt, int spc, int_lst wnr, cand_lst hlist);
static void separate_symbol(int_lst dold, int_lst dnew, int_lst dlst, int start, 
		            double df);
static int_lst positive_int_lst(int_lst ilst);
static char *assembling_line(int_lst dlst, cand_lst hlist, int_lst znr, 
		             cls_def *cls);
static void assembling_group(char *fname, int lcnt, hst_info *info, 
			     int_lst dlst, cand_lst hlist, int_lst znr, 
			     cls_def *cls);
static double_lst position_list(char *filename, cls_def *cls, cand_lst hlist);
static int_lst line_positions(double_lst plst, int baseline_start, 
			      int linespace_start, int height, int_lst *docu_lines);
static int duty_cycle(double_lst ilst, double *dty_cycle);
static int_lstm search_line(cand_lst hlist, cls_lst clst, inline_def in_line, 
			    char *filename, int iheight,
			    INLINE_METHOD forced_line_method, int_lst *docu_lines,
			    int renew, run_def *crun);
static int_lstm method_line_spacing(char *filename, cand_lst hlist, cls_def *cls,
				    linespacing_par linespacing, int iheight, 
				    int_lst *docu_lines,
				    INLINE_METHOD *method_fallback, int renew,
				    run_def *crun);
static baseline_lst calc_baselines(char *filename, cand_lst hlist, 
				   cls_def *cls, linespacing_par linespacing,
				   int height, int_lst *docu_lines, 
				   int renew, run_def *crun);
static int_lstm create_lines(cand_lst hlist, cls_def *cls, baseline_lst blst);
static int valid_candidate(cand_def cand);
static void reset_cand_def_aux(cand_lst cdl);
static int line_border(int baseline, int baseline_above, int baseline_below, 
		       double *line_top, double *line_bot);
static int line_of_point(baseline_lst blst, int nb_ref_point);
static ALIGMNT alignment_of_point(int baseline, int linespace, int ref_point, 
				  int *trend);
static int check_in_line(int baseline, int baseline_above, int baseline_below, 
			 int ref_point);
static int check_in_section(int baseline, int linespace, int ref_point, ALIGMNT alg);
static int best_alignment(cls_def *cls, cand_def *cand, ALIGMNT ref_alg,
			  int trend);
static double centre_of_interval(ival_def ival);
static int distance_alignment(ALIGMNT ref, ALIGMNT aut, int *line_offset);
static int_lstm trim_int_lstm(int_lstm iln);
static int estimated_linespace_from_db(cls_lst clst);
static int_lstm estimated_alignment(cls_def *cls, cand_lst hlist, int_lst hnr, 
				    int *fallback);
static void validate_alignment(int_lstm alst);
static int baseline_estimation(cls_def *cls, cand_lst hlist, int_lst hnr, 
			       int lspace, int *baseline);
static void overlapping_alternatives(cls_def *cls, cand_lst hlist, int_lst *hnr,
				     int lspace, int bline);
static int residual_alternatives(cls_def *cls, cand_lst hlist, int_lstm *ilnr,
				 int lspace, int_lst *blines);
static int best_mid(int_lst blines, int loc, int *new_idx);
static double_lst prepare_tf_set(double_lst signal, int period_len, 
				 double_lst *sig_padded);
static int assert_binary_signal(double_lst dlst, double *dlst_min, double *dlst_max);

/* Functions **************************************************************/

/* Generation and output of the recognized text. No formatting and no layout
   are calculated, only the line orientation, which is inherent, are 
   maintained.

Revision: 02.05.2003 E.Sack

Input : - class list
        - hitlist
        - mini statistic list
        - pointer to input image filename
        - maximal allowed lag
        - pointer to inline features
        - input image size size
        - forced inline method
        - pointer to documentation path
        - pointer to control structure
        - pointer to text view
        - flag (0: use DB period, else calculate period)
Output: - output text file written to harddisc
*/
void
xgen_ascii_text(cls_lst clst, cand_lst hlist, int_lst mst, char *filename, 
	        int noise_border, inline_def in_line, area_def pic_size,
	        INLINE_METHOD forced_line_method, int_lst *docu_lines,
	        run_def *crun, GtkWidget *ilog, int renew)
{
  int      line_cnt = 0, i;
  char     *suffix = 0;
  text_def text;
  int_lstm iln;

  if (!hlist.cnt) {
    return;                                     /* no candidates available */
  }

  init_text_def(&text);

  xreport("%s", "\nOutput Generation\n");

  iln = search_line(hlist, clst, in_line, filename,           /* all lines */
		    pic_size.height, forced_line_method, docu_lines,
		    renew, crun);

  if (!iln.cnt) {
    return;                                               /* nothing found */
  }

  text.cnt = iln.cnt;
  CALLOC(char*, text.cnt, text.tl);                /* allocate text buffer */

  xplab_progress_configure(crun, "line separation");
  xplab_progress_set_value(crun, 1.0, (gdouble)iln.cnt);

  for (i = 0; i < iln.cnt && *crun->pactive==TRUE; i++) {
    xplab_progress_set_value(crun, (gdouble)(i+1), (gdouble)iln.cnt);

    sort_nb(hlist, iln.il[i]);                      /* sort NB number list */
    text.tl[i] = separate_line(iln.il[i], clst.cls, hlist, filename, 
			       noise_border, line_cnt);

    if (text.tl[i]) {
      line_cnt++;                                   /* text line generated */
    }
  }

  suffix = print_mini_statistic(text.tl[0], mst);
  xwrite_text(suffix, text, ilog);

  free_text_def(&text);
  free_int_lstm(&iln); 
  FREEMEM(suffix);
}



/* Assembles DB data to string.

Revision: 11.03.2003 E.Sack

Input : - inline method features
        - pointer to string 
Output: - string updated
*/
static void
getstr_linespacing_par(linespacing_par lp, char **info)
{
  getstr_period_def(lp.period, info);
}



/* Assembles DB data to string.

Revision: 14.09.2002 E.Sack

Input : - period
        - pointer to string 
Output: - string updated
*/
static void
getstr_period_def(period_def pd, char **info)
{
  char *buf = NULL;

  int_to_string("L Phase: ", pd.l_phase, ", ", &buf);
  int_to_string("H Phase: ", pd.h_phase, "  ", &buf);
  int_to_string("(Period ", pd.l_phase+pd.h_phase, ")\n", &buf);
  add_string(buf, info);
  FREEMEM(buf);
}



/* Prints DB data to stream.

Revision: 05.10.2002 E.Sack

Input : - line spacing parameter
        - pointer to streams 
Output: - streams updated
*/
static void
print_linespacing_par(linespacing_par lpar)
{
  char *buf = 0;

  getstr_linespacing_par(lpar, &buf);

  xreport("%s", buf);

  FREEMEM(buf);
}



/* Prints DB data to output.

Revision: 17.10.2004 E.Sack

Input : - period
        - pointer to streams 
Output: - streams updated
*/
static void
print_period_def(char *prefix, period_def pd, char **info)
{
  char *buf = 0;

  add_string(prefix, &buf);

  getstr_period_def(pd, &buf);

  add_string(buf, info);
  FREEMEM(buf);
}



/* Calculating line space features. Some features, which are individual to
   the document, are always calculated, others are independent or constant
   to the medium.

Revision: 17.05.2003 E.Sack

Input : - pointer to input image filename
        - position list
        - pointer to linespace features
        - flag (0: use DB period, else calculate period)
        - pointer to control structure
Output: - linespace features updated as flagged
        - flag (0: no error, 1: error ocurred)        
Postcondition: - position list unchanged
*/
static int
calc_line_spacing(char *filename, double_lst plst, 
		  linespacing_par *linespacing, int renew, run_def *crun)
{
  int             d, pos_start, lag_phase;
  double_lst      h_sig;
  linespacing_par stored_period = *linespacing, optimal_linespace;
  signal_def      stype = {ZERO_MEAN, 0.0}, tf_type = {POLAR, 0.0};

  init_double_lst(&h_sig);

  copy_double_lst(&plst, &h_sig);                           /* signal copy */

  pos_start = adjust_to_phase(&h_sig);     /* eliminate zeros at start/end */

  dc_source_double(filename, h_sig, DC_BIT_POS_HPHASE, "");        /* docu */

  xreport("offset position list is %d\n", pos_start);

  if (renew == 1 ||                       /* ignore database, renew period */
      linespacing->period.h_phase == 0 ||               /* phase undefined */ 
      linespacing->period.l_phase == 0) {
    d = period_estimation(filename, h_sig, &lag_phase, stype, tf_type, 
			  &linespacing->period, crun);
    if (d) {
      free_double_lst(&h_sig);
      return 1;
    }

    xreport("estimated TF lag to signal: %d\n", lag_phase);
    xreport("%s", "estimated line space: ");
    print_linespacing_par(*linespacing);
  }

  xreport("%s", "stored line space: ");
  print_linespacing_par(stored_period);

  d = optimal_period_point(filename, h_sig, linespacing->period, 
			   &optimal_linespace);
  free_double_lst(&h_sig);

  if (d) {
    return 1;
  }

  linespacing->opos = pos_start + optimal_linespace.opos;

  xreport("optimal vertical start: %d\n", linespacing->opos);
  xreport("%s", "optimal line space: ");
  print_linespacing_par(optimal_linespace);

  return 0;
}



/* Finds the signal bin of the best matching H Phase.

Revision: 22.04.2003 E.Sack

Input : - h_phase of signal
        - period
        - pointer to optimal period features
        - verbose mode
Output: - optimal period features updated as flagged
        - flag (0: no error, 1: error ocurred)        
*/
static int
optimal_period_point(char *fname, double_lst h_sig, period_def period, 
		     linespacing_par *optimal_linespace)
{
  int        d, i, opos, linespace = period.h_phase+period.l_phase;
  int        signal_high_max, signal_low_max, signal_high_min;
  int        signal_low_min, signal_high_sum, signal_low_sum;
  double     dutyc;
  ival_def   max_range;
  period_def test_period;
  signal_def tf_type = {UNIPOLAR, 0.0};

  assert(ALIGNMENT_DIVR>=1);

  if (!period.h_phase || !period.l_phase) {
    return 1;
  }

  signal_features(h_sig, &signal_high_max, &signal_low_max, &signal_high_min, 
		  &signal_low_min, &signal_high_sum, &signal_low_sum, 0, 0,
		  &dutyc);

  test_period.h_phase = XPMIN(period.h_phase, signal_high_max);
  test_period.l_phase = linespace-test_period.h_phase;

  d = crosscorrelation_tf(fname, h_sig, tf_type, test_period, &max_range,
			  "optimal period position");
  if (d) {
    return 1;
  }

  if (ALIGNMENT_SIDE>0) {                           /* middle of the range */
    opos = (int)ceil((double)(max_range.end+max_range.start)/2.0);
  }
  else {
    opos = (int)floor((double)(max_range.end+max_range.start)/2.0);
  }

  if (opos >= 0) {
    d = seek_start_point(h_sig, linespace/ALIGNMENT_DIVR, &opos);
  }
  else {
    d = 0;
  }

  if (d == 0) {                                         /* fallback method */
    for (i = 0; i < h_sig.cnt; i++) {     /* seek next base-line-candidate */
      if (h_sig.dbl[i]) {
	opos = i;
	break;
      }
    }
  }

  optimal_linespace->period = test_period;
  optimal_linespace->opos = opos;
  return 0;
}



/* Crosscorrelation of a signal with a periodic binary testfunction. The 
   signal is zero-padded left and right with one period length.

Revision: 16.05.2003 E.Sack

Input : - signal
        - testfunction type
        - period
        - pointer to H Phase range
        - verbose mode
        - pointer to comment
Output: - CCF results dumped
        - range updated as flagged
        - flag (0: no error, 1: error ocurred)        
*/
int
crosscorrelation_tf(char *filename, double_lst signal, signal_def tf_type, 
		    period_def period, ival_def *range, char *subject)
{
  int        d, lag, period_len; 
  double     height, fill_min, sum_max;
  fft_def    flst;
  double_lst testfunction, sig_padded;
  ival_def   max_range;

  init_double_lst(&sig_padded);
 
  period_len = period.l_phase+period.h_phase;

  d = assert_polar_signal(signal, &fill_min, &height);
  if (d) {
    return 1;                                                     /* error */
  }

  testfunction = prepare_tf_set(signal, period_len, &sig_padded);

  if (!testfunction.cnt) {
    return 1;
  }

  fill_with_rectangle(testfunction, 0, period, height, -1);

  signal_transformation(testfunction, tf_type);

  flst = crosscorrelation_of_double_lst(sig_padded, testfunction);

  lag = max_index_double_left(flst.data[0], period_len-1);/* lag in period */

  max_range = max_period_sum(sig_padded, lag, period, &sum_max);

  xreport("CCF of TF: padded array length %d, TF period: %d, "
	 "max %f at lag %d\n",
	 flst.ar.width, period_len, flst.data[0][lag], lag);

  if (max_range.end >= max_range.start) {
    xreport("CCF of TF: max. sum %f in range of padded signal %d - %d \n", 
	   sum_max, max_range.start, max_range.end);

    max_range.start -= period_len;             /* relative to input signal */
    max_range.end -= period_len;               /* relative to input signal */
  }
  else {
    xreport("%s", "CCF of TF: no range proposal for maximal "
            "H Phase sum found\n");
  }

  dc_source_fft(filename, flst, DC_BIT_POS_CCF, subject);
  dc_source_double(filename, sig_padded, DC_BIT_POS_PAD, subject);
  dc_source_double(filename, testfunction, DC_BIT_TFIN, subject);

  fill_with_rectangle(testfunction, lag, period, height, -1);
  dc_source_double(filename, testfunction, DC_BIT_TFLAG, subject);

  free_double_lst(&testfunction);
  free_double_lst(&sig_padded);
  free_fft_def(&flst);

  *range = max_range;
  return 0;
}



/* Prints a signal type.

Revision: 17.10.2004 E.Sack

Input : - signal type
        - pointer to prefix
        - pointer to suffix
Output: - user output
*/
static void
print_signal_type(char *prefix, signal_def stype, char *suffix, char **info)
{
  int  d = -1;
  char *buf = NULL;

  add_string(prefix, &buf);

  switch(stype.sig_type) {
  case NRZ:
    d = add_string("NRZ", &buf);
    break;
  case NRZI:
    d = add_string("NRZI", &buf);
    break;
  case UNIPOLAR:
    d = add_string("UNIPOLAR", &buf);
    break;
  case POLAR:
    d = add_string("POLAR", &buf);
    break;
  case ZERO_MEAN:
    d = add_string("ZERO_MEAN", &buf);
    double_to_string(" (", stype.aux, ")", &buf);
    break;
  case LOWER_PHASE:
    d = add_string("LOWER_PHASE", &buf);
    double_to_string(" (", stype.aux, ")", &buf);
    break;
  default: assert(0); 
    break;
  }

  assert(d);

  add_string(suffix, &buf);
  add_string(buf, info);
  FREEMEM(buf);
}



/* Histogram of period maxima of a period image. A period P=H+L 
   with different phases H,L denotes a diagonal H+L=const (m=-1) in the
   period image. See 'Definition of period image' for further
   descriptions.

Revision: 12.07.2002 E.Sack

Input : - period image
        - start of H and L Phase
Output: - period histogram
*/
hst_def
period_max(pic_dbl cpic, period_def start_phase)
{
  int     w, h;
  hst_def period;

  assert(cpic.mem);
  assert(cpic.ar.width>0);
  assert(cpic.ar.height>0);
  assert(start_phase.h_phase>0);
  assert(start_phase.l_phase>0);

  init_hst_def(&period);

  period.hst.cnt = cpic.ar.width + cpic.ar.height - 1;
  CALLOC(double, period.hst.cnt, period.hst.dbl);

  for (w = 0; w < cpic.ar.width; w++) {                    /* init H Phase */
    period.hst.dbl[w] = cpic.mem[w];        
  }

  for (h = 1; h < cpic.ar.height; h++) {                   /* init L Phase */
    period.hst.dbl[cpic.ar.width-1+h] = 
      cpic.mem[cpic.ar.width-1+h*cpic.ar.width];
  }


  for (w = 0; w < cpic.ar.width-1; w++) {
    for (h = 1; h < cpic.ar.height; h++) {
      period.hst.dbl[w+h] = XPMAX(period.hst.dbl[w+h], 
				  cpic.mem[w+h*cpic.ar.width]);
    }
  }

  period.max_bin = max_index_double_left(period.hst.dbl, period.hst.cnt);
  period.maxval = period.hst.dbl[period.max_bin];
  period.start_bin = start_phase.h_phase + start_phase.l_phase;/* smallest */

  return period;
}



/* Transforms a signal to an unipolar rectangle {0,1}.

Revision: 18.04.2003 E.Sack

Input : - signal list
        - verbose flag (0: no output to STDOUT, 1: output to STDOUT)
Output: - signal list updated
        - flag (0: no error, 1: error occurred)
*/
static int
type_to_unipolar_rectangle(double_lst ilst)
{
  int    i, d;
  double max, min, mid;

  if (!ilst.cnt) {
    xreport("%s", "ERROR: empty list\n");
    return 1;
  }

  d = assert_binary_signal(ilst, &min, &max);

  if (d) {
    return 1;
  }

  mid = (min + max)/2.0;

  for (i = 0; i < ilst.cnt; i++) {
    ilst.dbl[i] = (ilst.dbl[i] > mid) ? UNIPOLAR_MAX : UNIPOLAR_MIN;
  }

  return 0;
}



/* Filling a list with a rectangle type unipolar.

Revision: 15.06.2002 E.Sack

Input : - list
        - rectangle lag
        - period
        - rectangle height
        - repetition of rectangle
Output: - list updated
Precondition:  - positive lag shifts rectangle to the right
Postcondition: - if repetition is negativ, repeat period until end.
*/
void
fill_with_rectangle(double_lst slst, int lag, period_def period, 
		    double height, int repetition)
{
  int        r, s, r_start;
  double_lst rlst;

  init_double_lst(&rlst);

  rlst.cnt = period.h_phase + period.l_phase;
  CALLOC(double, rlst.cnt, rlst.dbl);

  for (r = 0; r < slst.cnt; r++) {                         /* clear output */
    slst.dbl[r] = 0.0;
  }

  for (r = 0; r < period.h_phase; r++) {                        /* H Phase */
    rlst.dbl[r] = height;
  }

  s = 0;
  lag = lag%rlst.cnt;
  r_start = (rlst.cnt-lag)%rlst.cnt;
  do {
    for (r = r_start; r < rlst.cnt && s < slst.cnt; r++) {
      slst.dbl[s++] = rlst.dbl[r];
    }

    r_start = 0;
  } while (s < slst.cnt && repetition--);      /* ensure zero padding ext. */

  free_double_lst(&rlst);
}



/* Searches the interval of a signal according to a H phase with the maximal
   sum.

Revision: 19.07.2002 E.Sack

Input : - list
        - rectangle lag
        - period
Output: - interval
Precondition: - positive lag shifts rectangle to the right
Postcondition: - interval may be empty
*/
static ival_def
max_period_sum(double_lst slst, int lag, period_def period, double *sum_max)
{
  int      r, s = 0, r_start, start, end, period_len;
  double   sum;
  ival_def range;

  init_ival_def(&range); 

  period_len = period.h_phase + period.l_phase;

  lag = lag%period_len;
  r_start = (period_len-lag)%period_len;
  *sum_max = 0.0;
  do {
    start = end = s;
    sum = 0.0;
    for (r = r_start; r < period_len && s < slst.cnt; r++) {
      if (r < period.h_phase) {
        sum += slst.dbl[s];                    /* summation in the H Phase */
	end++;
      }

      s++;
    }

    if (*sum_max < sum) {
      *sum_max = sum;
      range.start = start;
      range.end = end-1;
    }

    r_start = 0;
  } while (s < slst.cnt);

  return range;
}



/* Creates a 'period image' of a binary signal.

   Definition of period image:
   A period image is a 2-D matrix. The horizontal dimension is build from
   the H phase (Hmin <= H <= Hmax), the vertical dimension is build from
   the L phase (Lmin <= L <= Lmax) of testfunctions (see 'Definition of
   testfunction'). There are (Hmax-Hmin+1)*(Lmax-Lmin+1) TFs and
   (Hmax-Hmin+1)+(Lmax-Lmin+1)-1 periods.

   A matrix coordinate therefore denotes a unique testfunction. A matrix
   value is the maximum of the crosscorrelation of a testfunction with the
   binary signal. A period image with at least one phase naught is undefined.

   The period image gives an overview of the best signal approximating TFs,
   however with phase (lag) of the testfunction regarding the signal lost.

Revision: 24.08.2005 E.Sack

Input : - signal list
        - testfunction type
        - start period
        - end period
        - signal height
        - pointer to control structure
Output: - period image 
*/
pic_dbl
period_image(double_lst signal, signal_def tf_type, period_def phase_start, 
	     period_def phase_end, double height, run_def *crun)
{
  int        lag, w, h, tfp, p, pmax, pstart;
  double     cc_max;
  fft_def    flst;
  double_lst testfunction, sig_padded;
  period_def tf_period;
  pic_dbl    pic;

  init_fft_def(&flst);
  init_double_lst(&sig_padded);
  init_pic_dbl(&pic);

  assert (signal.cnt);

  pic.ar.width = phase_end.h_phase-phase_start.h_phase+1;    /* phases image */
  pic.ar.height = phase_end.l_phase-phase_start.l_phase+1;

  CALLOC(double, pic.ar.width*pic.ar.height, pic.mem);

  pmax = pic.ar.width + pic.ar.height - 1;

  xplab_progress_configure(crun, "period image");
  xplab_progress_set_value(crun, 1.0, (gdouble)(pmax+1));

  pstart = phase_start.h_phase + phase_start.l_phase;   /* minimal TF period */

  for (p = 0; p <= pmax && *crun->pactive==TRUE; p++) {  /* period increment */
    xplab_progress_set_value(crun, (gdouble)(p+1), (gdouble)(pmax+1));

    tfp = pstart + p;                                           /* TF period */
    testfunction = prepare_tf_set(signal, tfp, &sig_padded);

    if (!testfunction.cnt) {
      break;
    }

    if (tfp <= phase_end.h_phase + phase_start.l_phase) {
      tf_period.l_phase = phase_start.l_phase;
      tf_period.h_phase = tfp - tf_period.l_phase;
    }
    else {
      tf_period.h_phase = phase_end.h_phase;
      tf_period.l_phase = tfp - tf_period.h_phase;
    }

    while (tf_period.h_phase >= phase_start.h_phase &&
	   tf_period.l_phase <= phase_end.l_phase && 
	   *crun->pactive == TRUE) {
      w = tf_period.h_phase - phase_start.h_phase;
      h = tf_period.l_phase - phase_start.l_phase;

      fill_with_rectangle(testfunction, 0, tf_period, height, -1);
      signal_transformation(testfunction, tf_type);

      flst = crosscorrelation_of_double_lst(sig_padded, testfunction);

      lag = max_index_double_left(flst.data[0], tfp-1);     /* lag in period */

      cc_max = flst.data[0][lag];
      pic.mem[w+h*pic.ar.width] = cc_max;              /* period image value */

      free_fft_def(&flst);

      tf_period.l_phase++;
      tf_period.h_phase--;
    }

    free_double_lst(&testfunction);
    free_double_lst(&sig_padded);
  }

  free_double_lst(&testfunction);
  free_double_lst(&sig_padded);

  return pic;
}



/* A binary signal is transformed to another binary signal.

Revision: 18.04.2003 E.Sack

Input : - signal list
        - signal type
        - verbose mode
Output: - signal list updated
        - transformed signal type
*/
signal_def
signal_transformation(double_lst ilst, signal_def stype)
{
  int        d, i, high_min, low_min, high_max, low_max, high_sum, low_sum;
  double     dutyc;
  signal_def tr_type = stype;

  if (!ilst.cnt) {
    xreport("%s", "ERROR: empty list\n");
    tr_type.sig_type = FIRST_SIGNAL_TYPE;      /* error, no transformation */
    return tr_type;
  }

  d = type_to_unipolar_rectangle(ilst);

  if (d) {
    tr_type.sig_type = FIRST_SIGNAL_TYPE;      /* error, no transformation */
    return tr_type;
  }

  switch(tr_type.sig_type) {
  case NRZ: case NRZI:
    xreport("ERROR: not supported signal type %d\n", tr_type);
    break;

  case UNIPOLAR:                                     /* already functional */
    break;

  case POLAR:
    for (i = 0; i < ilst.cnt; i++) {
      ilst.dbl[i] = (ilst.dbl[i] > UNIPOLAR_MIN) 
 	                         ? UNIPOLAR_MAX : -UNIPOLAR_MAX;
    }
    break;

  case ZERO_MEAN:                                   /* set mean value to 0 */
    signal_features(ilst, &high_max, &low_max, &high_min, &low_min,
		    &high_sum, &low_sum, 0, 0, &dutyc);

    tr_type.aux = -UNIPOLAR_MAX*(double)high_sum/(double)low_sum;

  case LOWER_PHASE:                                     /* set lower phase */
    assert(tr_type.aux < UNIPOLAR_MAX);

    for (i = 0; i < ilst.cnt; i++) {
      ilst.dbl[i] = (ilst.dbl[i] > UNIPOLAR_MIN) ? UNIPOLAR_MAX:tr_type.aux;
    }
    break;

  default: assert(0); 
    break;
  }

  return tr_type;
}



/* Checks for a binary signal.

Revision: 19.04.2003 E.Sack

Input : - signal
        - pointer to minimum and maximum of signal list
        - verbose flag (0: no output to STDOUT, 1: output to STDOUT)
Output: - minimum and maximum of signal updated as flagged
        - flag (0: no error, 1: error occurred)
*/
int
assert_binary_signal(double_lst dlst, double *dlst_min, double *dlst_max)
{
  int    i;
  double min, max;

  if (!dlst.dbl) {
    assert(!dlst.cnt);
    xreport("%s", "\nERROR: empty list\n\n");
    return 1;
  }
  else {
    assert(dlst.cnt>0);
  }

  min = min_double_left(dlst);
  max = max_double_left(dlst);
  assert(min <= max);

  if (fabs(max-min) < 2.0*FLT_EPSILON) {
    xreport("WARNING: not a binary signal because"
	   " single signal level (%G)\n", max);
    return 1;
  }

  for (i = 0; i < dlst.cnt; i++) {
    if (fabs(min-dlst.dbl[i]) > FLT_EPSILON && 
	fabs(max-dlst.dbl[i]) > FLT_EPSILON) {
      xreport("WARNING: not a binary signal because"
             " multiple signal levels (%G, %G, %G)\n",
	     min, max, dlst.dbl[i]);
      return 1;
    }
  }

  *dlst_min = min;
  *dlst_max = max;

  return 0;
}



/* Checks for a polar signal.

Revision: 19.04.2003 E.Sack

Input : - signal
        - pointer to minimum and maximum of signal list
        - verbose flag (0: no output to STDOUT, 1: output to STDOUT)
Output: - minimum and maximum of signal list updated as flagged
        - flag (0: no error, 1: error occurred)
*/
int
assert_polar_signal(double_lst dlst, double *dlst_min, double *dlst_max)
{
  int    d;
  double min, max;

  if (!dlst.dbl) {
    assert(!dlst.cnt);
    xreport("%s", "\nERROR: empty list\n\n");
    return 1;
  }
  else {
    assert(dlst.cnt>0);
  }

  d = assert_binary_signal(dlst, &min, &max);

  if (d) {
    return 1;
  }

  if (max < FLT_EPSILON) {
    xreport("WARNING: not a polar signal because "
	   "negative or zero maximum %f\n", max);
    return 1;
  }

  if (min > FLT_EPSILON) {
    xreport("WARNING: not a polar signal because "
	   "positive minimum %f\n", min);
    return 1;
  }

  *dlst_min = min;
  *dlst_max = max;

  return 0;
}



/* Calculates duty cycle features.

Revision: 15.06.2002 E.Sack

Input : - signal list
        - pointer to biggest H group
        - pointer to biggest L group
        - pointer to smallest H group
        - pointer to smallest L group
        - pointer to H sum
        - pointer to L sum
        - pointer to width histogram list of H groups
        - pointer to width histogram list of L groups
        - pointer to duty cycle
        - verbose flag (0: no output to STDOUT, 1: output to STDOUT)
Output: - pointed values updated as flagged
        - pointed histogram lists updated if not a null pointer
        - flag (0: no error, 1: error occurred)
*/
int
signal_features(double_lst ilst, int *high_max, int *low_max, int *high_min, 
		int *low_min, int *high_sum, int *low_sum, int_lst *hist_low, 
		int_lst *hist_high, double *dty_cycle)
{
  int i, high_phase = 0, low_phase = 0;

  if (!ilst.cnt) {
    xreport("%s", "WARNING: empty list\n");
    return 1;
  }

  if (hist_low) {
    free_int_lst(hist_low);
    hist_low->cnt = ilst.cnt;
    CALLOC(int, hist_low->cnt, hist_low->nr);
  }

  if (hist_high) {
    free_int_lst(hist_high);
    hist_high->cnt = ilst.cnt;
    CALLOC(int, hist_high->cnt, hist_high->nr);
  }

  *high_sum = *low_sum = 0;
  *high_max = *low_max = 0;
  *high_min = *low_min = ilst.cnt+1;

  for (i = 0; i < ilst.cnt; i++) {
    if (ilst.dbl[i] > 0.0) {
      (*high_sum)++;
      high_phase++;

      if (low_phase) {
        *low_max = XPMAX(*low_max, low_phase);
        *low_min = XPMIN(*low_min, low_phase);
	if (hist_low) {
	  hist_low->nr[low_phase]++;
	}
      }

      low_phase = 0;
    }
    else {
      (*low_sum)++;
      low_phase++;

      if (high_phase) {
        *high_max = XPMAX(*high_max, high_phase);
        *high_min = XPMIN(*high_min, high_phase);
	if (hist_high) {
	  hist_high->nr[high_phase]++;
	}
      }

      high_phase = 0;
    }
  }

  if (low_phase) {                                            /* finishing */
    *low_max = XPMAX(*low_max, low_phase);
    *low_min = XPMIN(*low_min, low_phase);
    if (hist_low) {
      hist_low->nr[low_phase]++;
    }
  }

  if (high_phase) {
    *high_max = XPMAX(*high_max, high_phase);
    *high_min = XPMIN(*high_min, high_phase);
    if (hist_high) {
      hist_high->nr[high_phase]++;
    }
  }

  *dty_cycle = (double)*high_sum/(double)(*high_sum+*low_sum);
  assert(*high_sum+*low_sum == ilst.cnt);
  return 0;
}



/* Wrapper for signal_features.

Revision: 08.10.2002 E.Sack

Input : - signal list
        - pointer to duty cycle
        - verbose flag (0: no output to STDOUT, 1: output to STDOUT)
Output: - duty cycle updated as flagged
        - flag (0: no error, 1: error occurred)
*/
static int
duty_cycle(double_lst ilst, double *dty_cycle)
{
  int high_max, low_max, high_min, low_min, high_sum, low_sum, err;

  if (!ilst.cnt) {
    xreport("%s", "WARNING: empty list\n");
    return 1;
  }

  err = signal_features(ilst, &high_max, &low_max, &high_min, &low_min, 
			&high_sum, &low_sum, 0, 0, dty_cycle);
  return err;
}



/* Creates a 'phase image' of a binary signal.

   Definition of phase image: 
   A 'phase image' is a 2-D matrix. Every row is the
   crosscorrelation of a testfunction with the binary signal. The period
   is constant with an increasing H phase in successive rows. The first
   row contains the CCF with TF of phases (H,L)=(1,P-1). There are P-1
   rows if the period is P. The phases are in 0 < H,L < P range. The
   first column of the rows contains the CCF of lag=0, the next column
   the CCF of lag=1 and so on.

   A phase image is always quadratic because the TF-lags are within one
   period.

   The phase image gives an overview of the best signal approximating
   duty cycle arrangement, with all lag positions calculated.

Revision: 24.08.2005 E.Sack

Input : - signal list
        - testfunction type
        - testfunction period
        - signal height
        - pointer to control structure
Output: - phase image
*/
pic_dbl
phase_image(double_lst signal, signal_def tf_type, int period_len, 
	    double height, run_def *crun)
{
  int        d, i;
  fft_def    flst;
  double_lst testfunction, sig_padded;
  period_def period;
  pic_dbl    pic;

  init_pic_dbl(&pic);
  init_fft_def(&flst);
  init_double_lst(&sig_padded);

  testfunction = prepare_tf_set(signal, period_len, &sig_padded);

  if (!testfunction.cnt) {
    return pic;
  }

  pic.ar.width = period_len-1;                          /* lag within period */

  pic.ar.height = period_len-1;
  CALLOC(double, pic.ar.width*pic.ar.height, pic.mem);

  for (period.h_phase = 1, i = 0;                      /* phase image format */
       period.h_phase < period_len && *crun->pactive==TRUE;
       period.h_phase++, i++) {
    period.l_phase = period_len - period.h_phase;
    fill_with_rectangle(testfunction, 0, period, height, -1);
    signal_transformation(testfunction, tf_type);

    flst = crosscorrelation_of_double_lst(sig_padded, testfunction);

    d = copy_double(flst.data[0], &pic.mem[i*pic.ar.width], pic.ar.width);

    assert(d == pic.ar.width);

    free_fft_def(&flst);
  }

  free_double_lst(&testfunction);
  free_double_lst(&sig_padded);

  return pic;
}



/* This function prepares the testfunction set consisting of the allocation
   of a continuous testfunction and a zero padded signal list (see scheme
   below) for crosscorrelation using the FFT. The non-redundant lags are in
   0 <= lag < p. The duty cycle of the TF period is of no interest.

             a)
padded |<-TF period->|<----------input signal----------->|
signal |0000000000000|xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx| lag=
    TF |1111110000000|11111100000001111110000000111111000|  0
       |?111111000000|01111110000000111111000000011111100|  1
       |??11111100000|00111111000000011111100000001111110|  2
       |????????????1|11111000000011111100000001111110000|  p-1
       |?????????????|11111100000001111110000000111111000|  p*

a) zero padding of one TF period to protect aliasing of FFT. This simulates
   the first p lags of an infinite TF. Keep in mind, TFs may be negative.
*  lag >= p is redundant or biased by zero padding of FFT if TF is not
   unipolar.
?  Wrapped values from FFT (periodic repetition) causes aliasing on
   according lag (depends on N and signal lengths).

Usage: The testfunction set is prepared for correlation using the FFT.

Revision: 16.05.2003 E.Sack

Input : - signal
        - period length
        - pointer to zero padded signal
Output: - testfunction (empty if error occurred)
        - zero padded signal updated as flagged
*/
static double_lst
prepare_tf_set(double_lst signal, int period_len, double_lst *sig_padded)
{
  int        d, length;
  double_lst testfunction;

  init_double_lst(&testfunction);

  if (!signal.dbl) {                                  /* consistency check */
    assert(!signal.cnt);
    return testfunction;                                  /* nothing to do */
  }
  else {
    assert(signal.cnt > 0);
  }

  if (period_len <= 0 || !sig_padded) {
    return testfunction;                                 /* makes no sense */
  }

  length = period_len + signal.cnt;

  testfunction.cnt = length;
  CALLOC(double, testfunction.cnt, testfunction.dbl);

  sig_padded->cnt = length;
  CALLOC(double, sig_padded->cnt, sig_padded->dbl);

  d = crop_double_lst(signal, *sig_padded, 0, period_len, signal.cnt);
  assert(d == signal.cnt);

  return testfunction;
}



/* This function is the SECOND STAGE of two stages for the period estimation.
   It calculates a matrix of all reasonable phases of a TF and searches
   for the maximal CCF among multiples which meets a H phase proposal.

Revision: 24.08.2005 E.Sack

Input : - signal
        - period length
        - signal H phase proposal
        - pointer to TF lag
        - pointer to period
        - pointer to phase image
        - pointer to control structure
Output: - TF lag regarding signal updated as flagged
        - period updated as flagged
        - phase image updated if not a zero pointer
        - flag (0: no error, 1: error occurred)
*/
static int
phase_estimation(double_lst signal, int period_len, int h_phase_proposal, 
		 int *lag_phase, period_def *period, pic_dbl *phase_img,
		 run_def *crun)
{
  int        d, lag, i, ii, h_phase, diff_l, diff_h, stop = 0, max_index;
  int        high_lower = 0, high_upper = 0, lag_lower = 0, lag_upper = 0;
  double     fill_height, fill_min, acf_0, cc_max;
  double_lst stemp;
  period_def period_temp;
  signal_def stype = {UNIPOLAR, 0.0};
  pic_dbl    cpic;

  assert(lag_phase);
  assert(period);

  init_pic_dbl(&cpic);
  init_double_lst(&stemp);

  copy_double_lst(&signal, &stemp);                           /* signal copy */

  d = assert_polar_signal(stemp, &fill_min, &fill_height);
  if (d) {
    free_double_lst(&stemp);
    return 1;
  }

  signal_transformation(stemp, stype);

  acf_0 = power_of_signal(stemp);                                  /* ACF(0) */

  cpic = phase_image(stemp, stype, period_len, fill_height, crun);

  max_index = max_index_double_left(cpic.mem, AREA_SIZE(cpic.ar));
  cc_max = cpic.mem[max_index];

  if (fabs(acf_0-cc_max) > FLT_EPSILON) {           /* shouldn't be met if a */
    xreport("WARNING: ACF(0) is %f, max. CCF is %f (difference %G)\n\n",
	   acf_0, cc_max, fabs(acf_0-cc_max));        /* unipolar TF is used */
  }

  diff_l = diff_h = cpic.ar.height+1;
  for (h_phase = 1, i=0; h_phase <= cpic.ar.height && !stop; h_phase++,i++) {
    ii = i*cpic.ar.width;

    for (lag = 0; lag < cpic.ar.width; lag++) {
      if (fabs(cpic.mem[lag+ii]-cc_max) <= FLT_EPSILON) {    /* another max. */
	if (h_phase_proposal >= h_phase) {       /* towards proposed H phase */
	  if (diff_l >= h_phase_proposal-h_phase) {   /* minimize difference */
	    diff_l = h_phase_proposal-h_phase;
	    high_lower = h_phase;                /* lower maximum of H phase */
	    lag_lower = lag;                                /* according lag */
	  }
	}

	if (h_phase_proposal == high_lower) {                /* proposal met */
	  stop = 1;
	  break;
	}

	if (h_phase_proposal < h_phase) {        /* towards proposed H phase */
	  if (diff_h > h_phase-h_phase_proposal) {    /* minimize difference */
	    diff_h = h_phase-h_phase_proposal;
	    high_upper = h_phase;                /* upper minimum of H phase */
	    lag_upper = lag;                                /* according lag */
	  }
	}
      }
    }
  }

  if (stop) {                         /* coordinates of the H phase proposal */
    *lag_phase = lag_lower;
    assert(high_lower>0);
    period_temp.h_phase = high_lower;
  }
  else if (diff_l > diff_h) {/* which maximum is closer to H phase proposal? */
    *lag_phase = lag_upper;
    assert(high_upper>0);
    period_temp.h_phase = high_upper;
  }
  else {
    *lag_phase = lag_lower;
    assert(high_lower>0);
    period_temp.h_phase = high_lower;  /* lowers duty cycle if equal differ. */ 
  }

  if (phase_img) {                                 /* documentation purposes */
    *phase_img = cpic;
  }
  else {
    free_pic_dbl(&cpic);
  }

  free_double_lst(&stemp);

  period_temp.l_phase = period_len-period_temp.h_phase; /*counterpart L phase*/

  *period = period_temp;

  return 0;
}



/* Wrapper for phase_estimation. Provides heuristics to solve the problem
   having several periods with the same maximum of CCF.

Revision: 24.08.2005 E.Sack

Input : - signal
        - histogram of period maxima
        - signal H phase proposal
        - pointer to TF lag
        - pointer to period
        - pointer to phase image
        - pointer to control structure
Output: - TF lag updated as flagged
        - period updated as flagged
        - phase image updated if not zero pointer
        - flag (0: no error, 1: error occurred)
*/
static int
phase_estimation_list(double_lst signal, hst_def hist_max, int h_phase_prop, 
		      int *lag_phase, period_def *pout, pic_dbl *phase_img, 
		      run_def *crun)
{
  int        err = 0, i, lag_phase_temp, period_len, pic_copied;
  double     dutyc_min, dutyc_temp;
  int_lst    max_idx_lst;                         /* list with equal maximas */
  period_def period, period_temp;
  pic_dbl    phase_img_tmp, phase_img_focus;

  init_period_def(&period);
  init_period_def(&period_temp);
  init_int_lst(&max_idx_lst);
  init_pic_dbl(&phase_img_tmp);
  init_pic_dbl(&phase_img_focus);

  max_idx_lst = max_index_double_all(hist_max.hst.dbl, hist_max.hst.cnt);

  dutyc_min = 2.0;                                                   /* init */
  for (i = 0; i < max_idx_lst.cnt && *crun->pactive==TRUE; i++) {/* at least one entry */
    period_len = max_idx_lst.nr[i] + hist_max.start_bin;

    err = phase_estimation(signal, period_len, h_phase_prop, &lag_phase_temp, 
			   &period_temp, &phase_img_tmp, crun);

    if (err) {
      break;
    }

    assert(period_len==period_temp.h_phase+period_temp.l_phase);
    dutyc_temp = (double)period_temp.h_phase/(double)period_len;

    pic_copied = 0;

    if (dutyc_min > dutyc_temp) {          /* always true for the first time */
      dutyc_min = dutyc_temp;                 /* lowest duty cycle preferred */
      period = period_temp;
      *lag_phase = lag_phase_temp;

      free_pic_dbl(&phase_img_focus);
      phase_img_focus = phase_img_tmp;             /* documentation purposes */
      pic_copied = 1;
    }
    else if (dutyc_min == dutyc_temp) {             /* else shortest H phase */
      if (period.h_phase > period_temp.h_phase) {               /* preferred */
	period = period_temp;
	*lag_phase = lag_phase_temp;

	free_pic_dbl(&phase_img_focus);
	phase_img_focus = phase_img_tmp;           /* documentation purposes */
	pic_copied = 1;
      }
      else if (period.h_phase == period_temp.h_phase) {     /* else smallest */
	if (*lag_phase > lag_phase_temp) {                  /* lag preferred */
	  *lag_phase = lag_phase_temp;

	  free_pic_dbl(&phase_img_focus);
	  phase_img_focus = phase_img_tmp;         /* documentation purposes */
	  pic_copied = 1;
	}
      }
    }

    if (!pic_copied) {
      free_pic_dbl(&phase_img_tmp);
    }
  }

  free_int_lst(&max_idx_lst);

  if (err || (period.h_phase + period.l_phase <= 0)) {
    free_pic_dbl(&phase_img_tmp);
    free_pic_dbl(&phase_img_focus);
    return 1;
  }

  *pout = period;

  if (phase_img) {
    *phase_img = phase_img_focus;
  }
  else {
    free_pic_dbl(&phase_img_focus);
  }

  return 0;
}



/* Fills gaps of the H Phase of a binary signal.

Revision: 22.04.2003 E.Sack

Input : - signal
        - gap width
        - fill value
Output: - signal updated
        - number of filled gaps
*/
int
fill_gaps(double_lst ilst, int fill)
{
  int    i, j, jj, f, filled = 0;
  double min, max;

  assert(fill >= 0);

  max = max_double_left(ilst);
  min = min_double_left(ilst);

  for (i = 0; i < ilst.cnt-2-fill; i++) {
    for (j = i+1, jj = 0; 
	 j < ilst.cnt && j <= i+1+fill && ilst.dbl[i] > min; 
	 j++) {
      if (ilst.dbl[j] > min) {
	jj = j;
	break;
      }
    }

    for (f = i+1; f < jj; f++) {
      ilst.dbl[f] = max;
      filled++;
    }
  }

  return filled;
}



/* Estimates the period (H+L) and phase H,L of a binary signal with
   testfunctions (TF), duty cycle = H/(H+L). Periods of pathological
   signals are not estimated.

   Definition of binary signal:
   A binary signal is a real function consisting of only two values, which
   are the minimum and the maximum of the signal requiring minimum < maximum.

   Definition of period P:
   P = H+L. H is the H phase, L is the L phase. This definition is valid
   only for binary signals. A period with at least one
   phase naught is undefined.

   Definition of duty cycle:
   DuC = H/P = H/(H+L), 0 < DuC < 1. A duty cycle with at least one phase
   naught is undefined.

   Definition of testfunction TF:
   A testfunction TF is a periodic binary signal with the period P. In the
   H phase the signal has its maximum, in the L phase the signal has its
   minimum. A TF starts with phase H of its period. A testfunction with
   at least one phase naught is undefined.

   The optimal TF shall approximate the signal with phase H,L and has a lag 
   relativ to signal start and a duty cycle less than 0.5. 

   As a real signal may be disturbed by noise and assuming the S/N is not
   'too bad', a simple signal restoration is made by filling gaps to 
   improve results according following conditions:-
   First condition:  Don't fill signals, if it cause a duty cycle more 
                     than 0.5 otherwise the (unknown) L Phase could be filled.
   Second condition: Don't fill signals, if it would cause a TF with a 
                     duty cycle more than 0.5.

   These conditions implicate that gaps are never filled if the original
   signal has a duty cycle equal to or greater than 0.5.

   This function is the FIRST STAGE of two stages for the period estimation.
   It calculates a matrix of all reasonable TFs.

Revision: 24.08.2005 E.Sack

Input : - pointer to input image filename
        - signal
        - pointer to TF lag
        - signal type
        - TF type
        - pointer to period
        - pointer to control structure
Output: - TF lag updated as flagged
        - period updated as flagged
        - flag (0: no error, 1: error occurred)
*/
int
period_estimation(char *filename, double_lst signal, int *lag_phase, 
		  signal_def signal_type, signal_def tf_type, 
		  period_def *per, run_def *crun)
{
  int        d, i, high_sum, low_sum, period_len, stop = 0;
  int        high_min, low_min, high_max, low_max, fnew, filled = 0;
  char       *buf = NULL;
  double     dutyc, dutyc_max = 0.5, fill_height, fill_min;
  double_lst stemp;
  hst_def    hist_sum, hist_sum_temp, hist_max;
  period_def period, phase_start, phase_end;
  pic_dbl    pic, pic_sum, phase_img;
  signal_def tr_type;

  init_period_def(&phase_start);
  init_period_def(&phase_end);
  init_period_def(&period);
  init_hst_def(&hist_sum);
  init_hst_def(&hist_sum_temp);
  init_hst_def(&hist_max);
  init_pic_dbl(&pic);
  init_pic_dbl(&pic_sum);
  init_pic_dbl(&phase_img);

  init_double_lst(&stemp);

  copy_double_lst(&signal, &stemp);                           /* signal copy */

  signal_features(stemp, &high_max, &low_max, &high_min, &low_min, &high_sum, 
		  &low_sum, 0, 0, &dutyc);

  phase_end.h_phase = XPMAX(high_max, low_max);    /* calculate period range */
  phase_start.h_phase = XPMIN(high_min, phase_end.h_phase);

  phase_end.l_phase = XPMAX(high_max, low_max);
  phase_start.l_phase = XPMIN(low_min, phase_end.l_phase);

  d = assert_polar_signal(stemp, &fill_min, &fill_height);
  if (d) {
    free_double_lst(&stemp);
    return 1;
  }

  tr_type = signal_transformation(stemp, signal_type);               /* init */

  print_signal_type("\nperiod image signal filter: ", tr_type, "\n", &buf);
  print_signal_type("period image TF filter: ", tf_type, "\n", &buf);
  print_period_def("start period image: ", phase_start, &buf);
  print_period_def("  end period image: ", phase_end, &buf);
  xreport("%s", buf);
  FREEMEM(buf);

  pic_sum = period_image(stemp, tf_type, phase_start, phase_end, fill_height,
			 crun);
  hist_sum = period_max(pic_sum, phase_start);                       /* init */
  copy_hst_def(&hist_sum, &hist_sum_temp);                           /* init */

  for (i = 1; i < signal.cnt && *crun->pactive == TRUE; i++) {
    fnew = fill_gaps(stemp, i);                               /* fill signal */

    if (!fnew) {
      continue;                                            /* nothing filled */
    }

    stop = duty_cycle(stemp, &dutyc);         /* duty cycle of filled signal */

    if (stop || 
	dutyc >= dutyc_max) {   /* first cond.- don't promote bad duty cycle */
      break;
    }

    signal_transformation(stemp, signal_type);           /* of filled signal */

    free_pic_dbl(&pic);                     /* period image of filled signal */
    pic = period_image(stemp, tf_type, phase_start, phase_end, fill_height,
		       crun);

    free_hst_def(&hist_max);
    hist_max = period_max(pic, phase_start);        /* maxima of all periods */

    add_two_hst_def(hist_max, &hist_sum_temp);        /* increase confidence */

    stop = phase_estimation_list(signal, hist_sum_temp, high_max, 
				 lag_phase, &period, 0, crun);    /* best TF */

    if (stop) {
      break;                                       /* no estimation possible */
    }

    period_len = period.h_phase+period.l_phase;
    dutyc = (double)period.h_phase/(double)period_len;   /* DuC of estim. TF */

    if (dutyc >= dutyc_max) {        /* second condition - avoid unwanted TF */
      break;
    }

    store_pic_dbl(pic, pic_sum);               /* for documentation purposes */
    copy_hst_def(&hist_sum_temp, &hist_sum);              /* accept forecast */
    filled = i;
  }

  free_pic_dbl(&pic);
  free_double_lst(&stemp);
  free_hst_def(&hist_max);
  free_hst_def(&hist_sum_temp);

  d = 0;

  if (!stop) {
    xreport("filled gaps width: %d\n", filled);

    d = phase_estimation_list(signal, hist_sum, high_max, lag_phase, &period, 
			      &phase_img, crun);

    dc_source_pic_dbl(filename, phase_img, DC_BIT_PEST);
    dc_source_pic_dbl(filename, pic_sum, DC_BIT_PEST_MAX);
  }
  else {
    xreport("%s", "phase estimation stopped\n");
  }

  free_hst_def(&hist_sum);
  free_pic_dbl(&pic_sum);
  free_pic_dbl(&phase_img);

  if (d || stop) {
    return 1;
  }

  *per = period;
  return 0;
}



/* A signal is adjusted to it's H Phase (leading and trailing zeros taken
   out).

Revision: 05.07.2002 E.Sack

Input : - pointer to signal list
Output: - list updated
        - offset discarded
*/
int
adjust_to_phase(double_lst *slst)
{
  int i, j, len, start = 0, stop = slst->cnt-1;

  for (i = 0; i < slst->cnt; i++) {
    if (slst->dbl[i] > 0.0) {                           /* test to H Phase */
      start = i;
      break;
    }
  }

  for (i = slst->cnt-1; i >= 0; i--) {
    if (slst->dbl[i] > 0.0) {
      stop = i;
      break;
    }
  }

  len = stop-start+1;

  if (len == slst->cnt) {
    return start;                   /* equal length as input or no H Phase */
  }
  else if (len > slst->cnt) {
    assert(0);                                           /* algorithm flaw */
  }

  for (i = start, j = 0; i <= stop; i++, j++) {
    slst->dbl[j] = slst->dbl[i];                      /* adjust to H Phase */
  }

  slst->cnt = len;
  REALLOC(double, slst->cnt, slst->dbl);

  return start;
}



/* The inline NB in the same column, defined by the input NB, are searched.

Revision: 15.03.2003 E.Sack

Input : - hitlist
        - pointer to class list
        - pointer to list of indizes of hitlist
        - hit NB book list
        - input NB
Output: - NB number list updated 
        - book list updated
*/
static void
inline_neighbor(cand_lst hlist, cls_def *cls, int_lst *hnr, int upper_hnr,
		int no_alt)
{
  int i, lower_hnr, d, lower_left, lower_right, upper_left, upper_right;
  int old_count;

  upper_left = hlist.c[upper_hnr].hdx.column.start;
  upper_right = hlist.c[upper_hnr].hdx.column.end;

  old_count = hnr->cnt;                            /* only test new entries */

  for (lower_hnr = 0; lower_hnr < hlist.cnt; lower_hnr++) {
    if (hlist.c[lower_hnr].aux == BOOKED) {
      continue;                                                   /* booked */
    }

    if (hlist.c[lower_hnr].alt && no_alt) {
      continue;                                   /* don't use alternatives */
    }

    lower_left = hlist.c[lower_hnr].hdx.column.start;
    lower_right = hlist.c[lower_hnr].hdx.column.end;

    if (lower_left > upper_right || upper_left > lower_right) {
      continue;                              /* no neighbors in same column */
    }

    d = inline_rule(hlist, cls, upper_hnr, lower_hnr);

    if (d > 0) {
      hnr->nr[hnr->cnt++] = lower_hnr;
      hlist.c[lower_hnr].aux = BOOKED;                            /* booked */
    }
  }

  for (i = old_count; i < hnr->cnt; i++) {        /* neighbors of neighbors */
    inline_neighbor(hlist, cls, hnr, hnr->nr[i], no_alt);
  }
}



/* Avoid misleading effects of big pattern, because big pattern can
   merge lines. This is avoided by skipping NBs which have even upper
   neighbors well separated. All NBs are compared to their next
   neighbors.  One cannot test just the nearest neighbor, because one of
   the left or right, respectively, could be a line spanning
   neighbor. If both of them are line spanning neighbors, other
   techniques must be used for placing. Neighbors used far away of the
   DUT can cause problems because of skewing. If all other neighbors are
   line spanning ones, it is a problem, too. The technique used here has
   its limits.

Revision: 21.12.2002 E.Sack

Input : - hitlist
        - pointer to class list
        - pointer to list of indizes of hitlist
        - hit NB book list
Output: - NB number list updated 
        - book list updated 
*/
static void
outline_neighbor(cand_lst hlist, cls_def *cls, int_lst *hnr)
{
  int     dut_i, d, dut_cand, nearest_left_nb, nearest_right_nb, cnt;
  int_lst erased;

  if (!hnr->cnt) {
    return;
  }

  init_int_lst(&erased);
  erased.cnt = hnr->cnt;
  CALLOC(int, erased.cnt, erased.nr);

  for (dut_i = 0; dut_i < hnr->cnt; dut_i++) {
    dut_cand = hnr->nr[dut_i];                             /* NB under test */

    nearest_left_nb = nearest_left_upper(hlist, hnr, dut_i);
    nearest_right_nb = nearest_right_upper(hlist, hnr, dut_i);

    if (nearest_left_nb == nearest_right_nb) {
      nearest_right_nb = -1;                        /* may be the same */
    }

    if (nearest_left_nb >= 0) {
      d = inline_rule(hlist, cls, nearest_left_nb, dut_cand);
      if (d <= 0) {
        erased.nr[dut_i] = -1;                           /* erase hit NB    */
        hlist.c[dut_cand].aux = 0;                       /* cancel bookings */
      }
    }

    if (nearest_right_nb >= 0) {
      d = inline_rule(hlist, cls, nearest_right_nb, dut_cand);
      if (d <= 0) {
        erased.nr[dut_i] = -1;                           /* erase hit NB    */
        hlist.c[dut_cand].aux = 0;                       /* cancel bookings */
      }
    }
  }

  for (dut_i = cnt = 0; dut_i < hnr->cnt; dut_i++) {  /* erase marked cand. */
    if (erased.nr[dut_i] != -1) {             
      hnr->nr[cnt++] = hnr->nr[dut_i];              /* rearrange index list */
    }
  }
  hnr->cnt = cnt;

  free_int_lst(&erased); 
}



/* NB is marked inline or skipped depending of the overlapping rule for
   hold in line. The rule calculates if a candidate is inline referenced 
   to a master NB, which should be the next neighbor. 

Revision: 17.02.2002 E.Sack

Input : - hitlist
        - pointer to class list
        - upper NB
        - lower NB
Output:  Flag (-1: skip, 1: inline)
Precondition: - upper NB may be not below lower NB.
*/
static int
inline_rule(cand_lst hlist, cls_def *cls, int upper_hnr, int lower_hnr)
{
#define P_EX 0

  int lower_nnr, upper_nnr, lower_top, upper_bottom;
  int lower_cnr, upper_cnr, lower_bottom, upper_top;

  assert(lower_hnr >= 0);
  assert(upper_hnr >= 0);
  assert(lower_hnr != upper_hnr);

  lower_top = hlist.c[lower_hnr].hdx.line.start;         /* lower NB frame */
  lower_bottom = hlist.c[lower_hnr].hdx.line.end;
  lower_nnr = hlist.c[lower_hnr].n;
  lower_cnr = hlist.c[lower_hnr].c;

  upper_top = hlist.c[upper_hnr].hdx.line.start;         /* upper NB frame */
  upper_bottom = hlist.c[upper_hnr].hdx.line.end;
  upper_nnr = hlist.c[upper_hnr].n;
  upper_cnr = hlist.c[upper_hnr].c;

  if (lower_top <= upper_top) {
    return 1;                                                     /* inline */
  }

  if (lower_bottom <= upper_bottom) {
    return 1;                                                     /* inline */
  }

  if (lower_cnr >= 0 && upper_cnr >= 0) {          /* could be unrecognized */
    if (cls[lower_cnr].att.alignment == cls[upper_cnr].att.alignment) {
      if (lower_bottom >= upper_bottom+P_EX) {
	if (lower_top >= upper_bottom) {
          return -1;                                       /* skip lower NB */
	}
	else {
          return 1;                                               /* inline */
	}
      }
      else {
	return 1;                                                 /* inline */
      }
    }
  }

  if (lower_cnr >= 0) {                            /* could be unrecognized */
    if (cls[lower_cnr].att.alignment == UPPERLINE &&
        lower_top > upper_bottom+P_EX) {
      return -1;                                           /* skip lower NB */
    }
  }
  else if (lower_top > upper_bottom) {
    return -1;                                             /* skip lower NB */
  }

  if (lower_cnr >= 0) {                            /* could be unrecognized */
    if (cls[lower_cnr].att.alignment == UNDERLINE) {
      if (lower_bottom >= upper_bottom+P_EX) {
        if (lower_top >= upper_bottom) {
	  return -1;                                       /* skip lower NB */
        }
	else {
          return 1;                                               /* inline */
	}

	return 1;                                                 /* inline */
      }
    }
  }

  if (lower_cnr >= 0 && upper_cnr >= 0) {          /* could be unrecognized */
    if (cls[lower_cnr].att.alignment == UPPERLINE &&
	cls[upper_cnr].att.alignment == UNDERLINE) {
      if (lower_bottom > upper_bottom) {
	return -1;                                         /* skip lower NB */
      }

      if (lower_top >= upper_bottom+P_EX) {
	return -1;                                         /* skip lower NB */
      }

      return 1;                                                   /* inline */
    }
  }

  if (lower_top > upper_bottom) {
    return -1;
  }

  return 1;
}



/* The outmost NBs in a list are detected.

Revision: 17.02.2002 E.Sack

Input : - hitlist
        - list of indizes of hitlist
        - pointer to leftmost NB
        - pointer to rightmost NB
Output: - leftmost NB updated
        - rightmost NB updated
*/
static void
outmost_neighbor(cand_lst hlist, int_lst hnr, int *left_inline, 
		 int *right_inline)
{
  int i, hit, snr_left, snr_right;

  hit = hnr.nr[0];
  snr_left = hlist.c[hit].hdx.column.start;               /* initialisation */
  snr_right = hlist.c[hit].hdx.column.end;

  for (i = 0; i < hnr.cnt; i++) {                        /* outmost of list */
    hit = hnr.nr[i];

    if (snr_left >= hlist.c[hit].hdx.column.start) {  /* leftmost inline NB */
      snr_left = hlist.c[hit].hdx.column.start;
      *left_inline = hit;
    }

    if (snr_right <= hlist.c[hit].hdx.column.end) {  /* rightmost inline NB */
      snr_right = hlist.c[hit].hdx.column.end;
      *right_inline = hit;
    }
  }

  assert(*left_inline>=0);
  assert(*right_inline>=0);
}



/* Seeking the nearest left upper neighbor from a NB in a list.

Revision: 22.02.2002 E.Sack

Input : - hitlist
        - pointer to list of indizes of hitlist
        - NB list index
Output: - neighbor NB index (-1: no neighbor found, >=0: neighbor found)
*/
static int
nearest_left_upper(cand_lst hlist, int_lst *hnr, int dut_i)
{
  int top_i, found = 0, next_left_end, next_left_top;
  int nearest_left_top, nearest_left_end = 0, next_left_nb;
  int nearest_left_nb = 0, dut_nb, dut_end, dut_top;

  if (!hnr->cnt) {
    return -1;                                                /* empty list */
  }           

  dut_nb = hnr->nr[dut_i];                                 /* NB under test */
  dut_end = hlist.c[dut_nb].hdx.column.end;
  dut_top = hlist.c[dut_nb].hdx.line.start;

  for (top_i = 0; top_i < hnr->cnt; top_i++) {        /* init. any neighbor */
    if (top_i == dut_i) {
      continue;
    }

    nearest_left_nb = hnr->nr[top_i];                       /* any neighbor */
    nearest_left_top = hlist.c[nearest_left_nb].hdx.line.start;
    nearest_left_end = hlist.c[nearest_left_nb].hdx.column.end;

    if (dut_top >= nearest_left_top &&              /* left upper condition */
	dut_end >= nearest_left_end) {
      found = 1;                           /* found any left upper neighbor */
      break;
    }
  }

  if (!found) {
    return -1;                              /* no left upper neighbor found */
  }

  for (top_i = 0; top_i < hnr->cnt; top_i++) {    /* searching nearest left */
    if (top_i == dut_i) {
      continue;
    }

    next_left_nb = hnr->nr[top_i];                          /* any neighbor */
    next_left_top = hlist.c[next_left_nb].hdx.line.start;
    next_left_end = hlist.c[next_left_nb].hdx.column.end;

    if (dut_top >= next_left_top &&         /* nearest left upper condition */
	dut_end >= next_left_end &&
	nearest_left_end <= next_left_end) {
      nearest_left_end = next_left_end;       /* current nearest left upper */
      nearest_left_nb = next_left_nb;
    }
  }

  return nearest_left_nb;
}



/* Seeking the nearest right upper neighbor from a NB in a list.

Revision: 22.02.2002 E.Sack

Input : - hitlist
        - pointer to list of indizes of hitlist
        - NB list index
Output: - neighbor NB index (-1: no neighbor found, >=0: neighbor found)
*/
static int
nearest_right_upper(cand_lst hlist, int_lst *hnr, int dut_i)
{
  int top_i, found = 0, next_right_start, next_right_top;
  int nearest_right_top, nearest_right_start = 0, next_right_nb;
  int nearest_right_nb = 0, dut_nb, dut_start, dut_top;

  if (!hnr->cnt) return -1;                                   /* empty list */

  dut_nb = hnr->nr[dut_i];                                 /* NB under test */
  dut_start = hlist.c[dut_nb].hdx.column.start;
  dut_top = hlist.c[dut_nb].hdx.line.start;

  for (top_i = 0; top_i < hnr->cnt; top_i++) {        /* init. any neighbor */
    if (top_i == dut_i) continue;

    nearest_right_nb = hnr->nr[top_i];                      /* any neighbor */
    nearest_right_top = hlist.c[nearest_right_nb].hdx.line.start;
    nearest_right_start = hlist.c[nearest_right_nb].hdx.column.start;

    if (dut_top >= nearest_right_top &&            /* right upper condition */
	dut_start <= nearest_right_start) {
      found = 1;                          /* found any right upper neighbor */
      break;
    }
  }

  if (!found) return -1;                   /* no right upper neighbor found */

  for (top_i = 0; top_i < hnr->cnt; top_i++) {   /* searching nearest right */
    if (top_i == dut_i) continue;

    next_right_nb = hnr->nr[top_i];                         /* any neighbor */
    next_right_top = hlist.c[next_right_nb].hdx.line.start;
    next_right_start = hlist.c[next_right_nb].hdx.column.start;

    if (dut_top >= next_right_top &&       /* nearest right upper condition */
	dut_start <= next_right_start &&
	nearest_right_start >= next_right_start) {
      nearest_right_start = next_right_start;/* current nearest right upper */
      nearest_right_nb = next_right_nb;
    }
  }

  return nearest_right_nb;
}



/* Seeking the nearest left neighbor from the input NB in the hitlist.

Revision: 15.03.2003 E.Sack

Input : - hitlist
        - pointer to class list
        - pointer to list of indizes of hitlist
        - book list
        - input NB
Output: - number list updated
        - book list updated
*/
static void
nearest_left_neighbor(cand_lst hlist, cls_def *cls, int_lst *hnr, 
		      int left_inline, int no_alt)
{
  int     d = 0, i, stop, inline_end, next_left_end, nearest_left_end = 0;
  int     nearest_left_nb = -1;
  int_lst touched;

  init_int_lst(&touched);      /* could be saved if hitlist would be sorted */
  touched.cnt = hlist.cnt;
  CALLOC(int, touched.cnt, touched.nr);

  inline_end = hlist.c[left_inline].hdx.column.end;

  do {
    for (i = 0, stop = 1; i < hlist.cnt; i++) { /* init. next left neighbor */
      if (hlist.c[i].aux == BOOKED) {
	continue;                                                 /* booked */
      }

      if (touched.nr[i] == BOOKED) {
	continue;                                                 /* booked */
      }

      if (hlist.c[i].alt && no_alt) {
	continue;                                 /* don't use alternatives */
      }

      nearest_left_end = hlist.c[i].hdx.column.end;
      if (inline_end >= nearest_left_end) {               /* left condition */
	stop = 0;                                /* found any left neighbor */
	break;
      }
    }

    if (stop) {
      break;                                      /* no left neighbor found */
    }

    for (i = 0; i < hlist.cnt; i++) {              /* nearest next neighbor */
      if (hlist.c[i].aux == BOOKED) {
	continue;                                                 /* booked */
      }

      if (touched.nr[i] == BOOKED) {
	continue;                                                 /* booked */
      }

      if (hlist.c[i].alt && no_alt) {
	continue;                                 /* don't use alternatives */
      }

      next_left_end = hlist.c[i].hdx.column.end;

      if (inline_end >= next_left_end &&          /* nearest left condition */
          nearest_left_end <= next_left_end) {
	nearest_left_end = next_left_end;
	nearest_left_nb = i;
      }
    }

    touched.nr[nearest_left_nb] = BOOKED;                      /* test once */
    d = inline_rule(hlist, cls, left_inline, nearest_left_nb);
    if (d > 0) {
      hnr->nr[hnr->cnt++] = nearest_left_nb;
      hlist.c[nearest_left_nb].aux = BOOKED;                      /* booked */
      stop = 1;
    }

    if (stop) {
      break;
    }
  } while (1);

  free_int_lst(&touched);
}



/* Seeking the nearest right neighbor from the input NB in the hitlist.

Revision: 15.03.2003 E.Sack

Input : - hitlist
        - pointer to class list
        - pointer to list of indizes of hitlist
        - book list
        - input NB
Output: - number list updated
        - book list updated
*/
static void
nearest_right_neighbor(cand_lst hlist, cls_def *cls, int_lst *hnr, 
		       int right_inline, int no_alt)
{
  int     d = 0, i, stop, inline_start, next_right_start;
  int     nearest_right_start = 0, nearest_right_nb = -1;
  int_lst touched;

  init_int_lst(&touched);      /* could be saved if hitlist would be sorted */
  touched.cnt = hlist.cnt;
  CALLOC(int, touched.cnt, touched.nr);
 
  inline_start = hlist.c[right_inline].hdx.column.start;

  do {
    for (i = 0, stop = 1; i < hlist.cnt; i++) {/* init. next right neighbor */
      if (hlist.c[i].aux == BOOKED) {
	continue;                                                 /* booked */
      }

      if (touched.nr[i] == BOOKED) {
	continue;                                                 /* booked */
      }

      if (hlist.c[i].alt && no_alt) {
	continue;                                 /* don't use alternatives */
      }

      nearest_right_start = hlist.c[i].hdx.column.start;

      if (inline_start <= nearest_right_start) {         /* right condition */
	stop = 0;                               /* found any right neighbor */
        break;
      }
    }

    if (stop) break;                             /* no right neighbor found */

    for (i = 0; i < hlist.cnt; i++) {              /* nearest next neighbor */
      if (hlist.c[i].aux == BOOKED) {
	continue;                                                 /* booked */
      }

      if (touched.nr[i] == BOOKED) {
	continue;                                                 /* booked */
      }

      if (hlist.c[i].alt && no_alt) {
	continue;                                 /* don't use alternatives */
      }

      next_right_start = hlist.c[i].hdx.column.start;

      if (inline_start <= next_right_start &&    /* nearest right condition */
          nearest_right_start >= next_right_start) {
        nearest_right_start = next_right_start;
        nearest_right_nb = i;
      }
    }

    touched.nr[nearest_right_nb] = BOOKED;                     /* test once */
    d = inline_rule(hlist, cls, right_inline, nearest_right_nb);
    if (d > 0) {
      hnr->nr[hnr->cnt++] = nearest_right_nb;
      hlist.c[nearest_right_nb].aux = BOOKED;                     /* booked */
      stop = 1;
    }

    if (stop) break;
  } while (1);

  free_int_lst(&touched);
}



/* The NBs in the same line are searched starting with the topmost NB. The 
   position of the NBs are compared pairwise successively to the left and
   right direction being inline or not. So the inline NB list is extended 
   with NBs left and right to the topmost one.

Revision: 17.02.2002 E.Sack

Input : - hitlist
        - pointer to class list
        - pointer to list of indizes of hitlist
        - book list
Output: - number list updated
        - book list updated
*/
static void
collect_inline_neighbor(cand_lst hlist, cls_def *cls, int_lst *hnr, 
			int no_alt)
{
  int i, old_count, left_inline = -1, right_inline = -1;

  if (!hnr->cnt) {
    return;                                                 /* no more hits */
  }

  outmost_neighbor(hlist, *hnr, &left_inline, &right_inline);

  old_count = hnr->cnt;                            /* only test new entries */

  nearest_left_neighbor(hlist, cls, hnr, left_inline, no_alt);
  nearest_right_neighbor(hlist, cls, hnr, right_inline, no_alt);

  for (i = old_count; i < hnr->cnt; i++) {        /* neighbors of neighbors */
    inline_neighbor(hlist, cls, hnr, hnr->nr[i], no_alt);
  }

  if (old_count-hnr->cnt) {                        /* no NB inline any more */
    collect_inline_neighbor(hlist, cls, hnr, no_alt);      /* collect cand. */
  }
}



/* NB numbers of every line are gathered. 

Revision: 24.08.2005 E.Sack

Input : - hitlist
        - class list
        - inline method
        - pointer to input image filename
        - input image height
        - forced inline method
        - pointer to documentation path
        - flag (0: use DB period, else calculate period)
        - pointer to control structure
Output: - line number lists
Postcondition: - lines are sorted from top to bottom
*/
static int_lstm
search_line(cand_lst hlist, cls_lst clst, inline_def in_line,
	    char *filename, int iheight, INLINE_METHOD forced_line_method,
	    int_lst *docu_lines, int renew, run_def *crun)
{
  int_lstm      iln;
  INLINE_METHOD method, method_fallback = METHOD_FIRST;

  init_int_lstm(&iln);

  if (*crun->pactive == FALSE) {
    return iln;
  }

  method = XPLAB_IS_METHOD(forced_line_method)?forced_line_method:
                                               in_line.method;

  switch (method) {
  case OVERLAPPING:
    xreport("%s", "line detection method: overlapping\n");

    iln = method_line_overlapping(hlist, clst);

    break;

  case LINE_SPACING:
    xreport("%s", "line detection method: line spacing\n");

    iln = method_line_spacing(filename, hlist, clst.cls, in_line.linespacing,
			      iheight, docu_lines, &method_fallback,
                              renew, crun);

    if (XPLAB_IS_METHOD(method_fallback)) {              /* recursive call */
      assert(method_fallback!=method);                     /* no dead lock */
      xreport("%s", "WARNING: line detection method failed (current "
	     "method is 'line spacing'): ");
      xreport("fallback to 'overlapping'\n");
      iln = search_line(hlist, clst, in_line, filename, iheight,
			method_fallback, docu_lines, renew, crun);
    }

    break;

  default: assert(0); break;                             /* unknown method */
  }

  return iln;                                     /* lines top-down sorted */
}



/* Assembles candidates in lines using a baseline grid.

Revision: 02.05.2003 E.Sack

Input : - pointer to input image filename
        - hitlist
        - pointer to class list
        - inline parameter
        - input image height
        - pointer to documentation path
        - pointer to fallback method
        - flag (0: use DB period, else calculate period)
        - pointer to control structure
Output: - line number lists
        - fallback method updated
*/
static int_lstm
method_line_spacing(char *filename, cand_lst hlist, cls_def *cls, 
		    linespacing_par linespacing, int iheight,
		    int_lst *docu_lines, INLINE_METHOD *method_fallback, 
		    int renew, run_def *crun)
{
  int_lstm     iln;
  baseline_lst blst;

  init_int_lstm(&iln);

  blst = calc_baselines(filename, hlist, cls, linespacing, iheight, docu_lines,
			renew, crun);

  if (blst.lspace == BASELINE_INIT) {
    *method_fallback = OVERLAPPING;
    return iln;
  }

  iln = create_lines(hlist, cls, blst);

  assert(iln.cnt);

  free_baseline_lst(&blst);

  return iln;                           /* lines sorted from top to bottom */
}



/* Assembles candidates in lines. Starting with their NB location, candidates
   are put to the best fitted line regarding their alignment. As result a
   candidate can lie in one of three neighbored lines. For example, if there
   are two alternatives for alignment of the same pattern like '_' and '-', 
   the ideal alignment decides which alignment fits.

   NB location,         ideal alignment         best fit of candidates
   always valid   ==>   regarding         ==>   alignment regarding
                        baseline grid           ideal alignment

Revision: 08.03.2003 E.Sack

Input : - hitlist
        - pointer to class list
        - baseline list
Output: - line number lists
*/
static int_lstm
create_lines(cand_lst hlist, cls_def *cls, baseline_lst blst)
{
  int      bl, ilx, nrx, cand, trend, line_offset = 0;
  int_lstm iln, trimmed;
  ALIGMNT  ideal_alg = UDEFLINE;

  init_int_lstm(&iln);

  iln.cnt = blst.bline.cnt + 2;    /* initially one line at both ends more */
  CALLOC(int_lst, iln.cnt, iln.il);                         /* number list */

  for (bl = 0; bl < iln.cnt; bl++) {
    init_int_lst(&iln.il[bl]);
  }

  for (cand = 0; cand < hlist.cnt; cand++) {
    bl = line_of_point(blst, hlist.c[cand].hdx.line.end);/*baseline of line*/
    assert(bl>=0);                            /* candidate within document */

    ilx = bl+1;                               /* line index in number list */

    if (hlist.c[cand].c >= 0) {                      /* classified pattern */
      ideal_alg = alignment_of_point(blst.bline.nr[bl], blst.lspace, 
				     hlist.c[cand].hdx.line.end, &trend);
      assert(ideal_alg>UDEFLINE);

      line_offset = best_alignment(cls, &hlist.c[cand], ideal_alg, trend);

      ilx -= line_offset;                               /* line correction */
    }

    nrx = iln.il[ilx].cnt++;

    if (nrx) {
      REALLOC(int, iln.il[ilx].cnt, iln.il[ilx].nr);
    }
    else {
      CALLOC(int, iln.il[ilx].cnt, iln.il[ilx].nr);
    }

    iln.il[ilx].nr[nrx] = cand;                /* enroll candidate's index */
  }

  trimmed = trim_int_lstm(iln);     /* erase empty lines at top and bottom */

  free_int_lstm(&iln);

  return trimmed;      /* lines sorted from top to bottom as baseline list */
}



/* Erases empty lines at the beginning and the end of a line list.

Revision: 09.03.2003 E.Sack

Input : - pointer to line number list
Output: - line number list updated
Postcondition: - line number list may be empty; if empty it is initialized
               - order of input list retained
*/
static int_lstm
trim_int_lstm(int_lstm iln)
{
  int      line, tdx, end = -1, first;
  int_lstm trimmed;

  init_int_lstm(&trimmed);

  for (line = iln.cnt-1; line >= 0; line--) {
    if (!iln.il[line].cnt) {
      continue;
    }

    end = line;                       /* last non-empty line of input list */
    break;
  }

  for (line = first = 0; line <= end; line++) {
    if (!iln.il[line].cnt && !first) {
      continue;
    }

    first = 1;                       /* first non-empty line of input list */

    tdx = trimmed.cnt++;

    if (tdx) {
      REALLOC(int_lst, trimmed.cnt, trimmed.il);
    }
    else {
      CALLOC(int_lst, trimmed.cnt, trimmed.il);
    }

    init_int_lst(&trimmed.il[tdx]);

    copy_int_lst(&iln.il[line], &trimmed.il[tdx]);
  }

  return trimmed;
}



/* Seeks baseline number for a reference point.

Revision: 19.04.2003 E.Sack

Input : - baseline list
        - reference point
Output: - baseline number
*/
static int
line_of_point(baseline_lst blst, int ref_point)
{
  int bl, in_line, baseline_above, baseline_below;

  assert(blst.bline.cnt > 0);

  baseline_above = blst.bline.nr[0] - blst.lspace;
  baseline_below = blst.bline.cnt == 1 ? blst.bline.nr[0] + blst.lspace : 
                                         blst.bline.nr[1];

  in_line = check_in_line(blst.bline.nr[0], baseline_above, baseline_below,
			  ref_point);                /* check topmost line */
  if (in_line) {
    return 0;                              /* reference point lies in line */
  }

  baseline_below = blst.bline.nr[blst.bline.cnt-1] + blst.lspace;
  baseline_above = blst.bline.cnt == 1 ? 
                            blst.bline.nr[blst.bline.cnt-1] - blst.lspace :
                            blst.bline.nr[blst.bline.cnt-2];

  in_line = check_in_line(blst.bline.nr[blst.bline.cnt-1], baseline_above,
			  baseline_below, ref_point); /* check lowest line */
  if (in_line) {
    return blst.bline.cnt-1;               /* reference point lies in line */
  }

  for (bl = 1; bl < blst.bline.cnt-1; bl++) {      /* check residual lines */
    baseline_above = blst.bline.nr[bl-1];
    baseline_below = blst.bline.nr[bl+1];

    in_line = check_in_line(blst.bline.nr[bl], baseline_above, 
			    baseline_below, ref_point);
    if (in_line) {
      return bl;                           /* reference point lies in line */
    }
  }

  if (ref_point < blst.bline.nr[0]) {
    return -UPPERLINE;              /* reference point lies above top line */
  }

  return -UNDERLINE;             /* reference point lies below bottom line */
}



/* Calculates the smallest distance of alignments and their position in
   lines.

Revision: 08.03.2003 E.Sack

Input : - reference alignment
        - alignment under test
	- pointer to line offset
Output: - line offset updated (-1: line below,
                                0: line of reference,
                               +1: line above)
        - distance (negative: alignment below,
                           0: alignment like reference,
                    positive: alignment above)
*/
static int
distance_alignment(ALIGMNT ref, ALIGMNT aut, int *line_offset)
{
  int distance, distance_min;

  distance_min = aut - ref;
  *line_offset = 0;                  /* alignment sector in reference line */

  distance = aut + UPPERLINE - ref;

  if (abs(distance_min) > abs(distance)) {
    distance_min = distance;
    *line_offset = 1;                    /* alignment sector in line above */
  }

  distance = aut - UPPERLINE - ref;

  if (abs(distance_min) > abs(distance)) {
    distance_min = distance;
    *line_offset = -1;                   /* alignment sector in line below */
  }

  return distance_min;
}



/* Calculates line offset regarding a candidate relative to his reference
   alignment. 

Usage: Reference shall be the section for the candidate's pattern
       regardless of his alignment. This function decides for the best
       real alignment.

Revision: 08.03.2003 E.Sack

Input : - pointer to class list
        - candidate
        - reference
        - trend of pattern (-1: to lower, +1: to upper)
Output: - flag (-1: line below reference,
                 0: line of reference,
                +1: line above reference)
*/
static int
best_alignment(cls_def *cls, cand_def *cand, ALIGMNT ref_alg, int trend)
{
  int     class, alt, distance, distance_min = 0, line_offset;
  int     who = 0, update_best, line_offset_min = 0, distance_abs = 0;
  ALIGMNT type, type_min;

  assert(trend == -1 || trend == +1);

  class = cand->c;                                       /* main candidate */
  assert(class >= 0);                           /* to be handled elsewhere */

  type = cls[class].att.alignment;

  alt = -1;
  do {
    update_best = 0;

    distance = distance_alignment(ref_alg, type, &line_offset);

    if (alt == -1) {
      update_best = 1;
    }
    else if (distance_abs > abs(distance)) {
      update_best = 1;
    }
    else if (distance_abs == abs(distance)) {
      if (line_offset_min == line_offset) {
	if (trend > 0) {                         /* pattern tends to upper */
	  if (distance_min < distance) {
	    update_best = 1;
	  }                                      /* else nothing to change */
	}
	else {                                   /* pattern tends to lower */
	  if (distance_min > distance) {
	    update_best = 1;
	  }                                      /* else nothing to change */
	}
      }
      else {
	if (trend > 0) {                         /* pattern tends to upper */
	  if (line_offset_min < line_offset) {
	    update_best = 1;
	  }                                      /* else nothing to change */
	}
	else {                                   /* pattern tends to lower */
	  if (line_offset_min > line_offset) {
	    update_best = 1;
	  }                                      /* else nothing to change */
	}
      }
    }

    if (update_best) {
      distance_abs = abs(distance);
      distance_min = distance;
      line_offset_min = line_offset;
      type_min = type;
      who = alt;                                        /* store candidate */
    }

    if (!cand->alt) {                             /* look for alternatives */
      break;                                            /* no alternatives */
    }

    if (++alt >= cand->alt->cnt) {               /* check all alternatives */
      break;                                       /* no more alternatives */
    }

    class = cand->alt->c[alt].c;                  /* alternative candidate */
    assert(class >= 0);                         /* no unclassified pattern */

    type = cls[class].att.alignment;
  } while (1);

  if (who >= 0) {
    cand_def altc;

    init_cand_def(&altc);           /* shift alternative to main candidate */

    copy_cand_def(&cand->alt->c[who], &altc);
    copy_cand_def(&altc, cand);

    free_cand_def(&altc);
  }

  return line_offset_min;
}



/* Seeks alignment of a reference point.

Revision: 19.03.2003 E.Sack

Input : - baseline list
        - reference point
        - pointer to trend
Output: - baseline number (negativ: no alignment possible)
        - trend updated (0: fail, +1: to upper, -1: to lower)
*/
static ALIGMNT
alignment_of_point(int baseline, int linespace, int ref_point, int *trend)
{
  ALIGMNT alg;

  for (alg = UNDERLINE; alg <= UPPERLINE; alg++) {
    *trend = check_in_section(baseline, linespace, ref_point, alg);

    if (*trend) {
      return alg;                          /* reference point lies in line */
    }
  }

  if (ref_point < baseline) {
    *trend = 1;
    return -UPPERLINE;         /* reference point lies above top alignment */
  }

  *trend = -1;
  return -UNDERLINE;        /* reference point lies below bottom alignment */
}



/* Checks alignment for reference point.

Revision: 08.03.2003 E.Sack

Input : - baseline
        - line space
        - reference point
        - alignment
Output: - flag (0: not inline, +1: in upper section, -1: in lower section)
*/
static int
check_in_section(int baseline, int linespace, int ref_point, ALIGMNT alg)
{
  int      d, int_alg;
  double   bot, top, section, point, centre;
  ival_def ival;

  assert(linespace > 0);

  d = aligmnt_to_int(alg, &int_alg);
  assert(d);

  section = linespace/4.0;                      /* four alignment sections */

  top = (double)(1-int_alg)*section;
  bot = top + section;

  point = (double)(ref_point - baseline);

  if (top < point && point <= bot) {
    ival.start = (int)ceil(top);
    ival.end = (int)bot;

    centre = centre_of_interval(ival);
    if (point <= centre) {
      return 1;              /* reference point lies in upper section half */
    }
    else {
      return -1;             /* reference point lies in lower section half */
    }
  }

  return 0;                     /* reference point lies not within section */
}



/* Calculates the center of an interval.

Revision: 08.03.2003 E.Sack

Input : - interval
Output: - center
*/
static double
centre_of_interval(ival_def ival)
{
  double len = (double)INTERVAL_SIZE(ival);

  assert(ival.start<=ival.end);

  return (double)(ival.start - 1) + len/2.0;
}



/* Checks reference point in line.

Revision: 19.04.2003 E.Sack

Input : - baseline
        - linespace to baseline above
        - linespace to baseline below
        - reference point
Output: - flag (0: not inline, 1: in line)
*/
static int
check_in_line(int baseline, int baseline_above, int baseline_below, 
	      int ref_point)
{
  int    d;
  double line_bot, line_top;

  d = line_border(baseline, baseline_above, baseline_below, &line_top, 
		  &line_bot);
  assert(d);

  if (ref_point > line_top && ref_point <= line_bot) {          /* in line */
    return 1;                              /* reference point lies in line */
  }

  return 0;
}



/* Resets candidates in a list.

Revision: 07.03.2003 E.Sack

Input : - candidate list
Output: - candidate list updated
*/
static void
reset_cand_def_aux(cand_lst cdl)
{
  int cand;

  for (cand = 0; cand < cdl.cnt; cand++) {
    cdl.c[cand].aux = XPLAB_INIT_AUX_CAND_DEF;
  }
}



/* Calculates the baselines from a candidate list.

Revision: 21.02.2003 E.Sack

Input : - pointer to input image filename
        - candidate list
        - pointer to class list
        - inline parameter
        - input image height
        - pointer to documentation path
        - flag (0: use DB period, else calculate period)
        - pointer to control structure
Output: - baseline list
Postcondition: baseline list in init-state if calculation is not possible.
*/
static baseline_lst
calc_baselines(char *filename, cand_lst hlist, cls_def *cls, 
	       linespacing_par linespacing, int iheight, int_lst *docu_lines,
	       int renew, run_def *crun)
{
  int          d;
  double_lst   plst;
  baseline_lst blst;

  init_baseline_lst(&blst);

  plst = position_list(filename, cls, hlist);

  if (!plst.cnt) {
    free_double_lst(&plst);
    return blst;
  }

  d = calc_line_spacing(filename, plst, &linespacing, renew, crun);

  if (d) {
    free_double_lst(&plst);
    return blst;
  }

  blst.lspace = linespacing.period.h_phase + linespacing.period.l_phase;

  blst.bline = line_positions(plst, linespacing.opos, blst.lspace, iheight, 
			      docu_lines);
  free_double_lst(&plst);

  assert(blst.lspace);
  return blst;
}



/* Assembles candidates in lines. Candidates are assumed to be inline if
   they have overlapping heights with their next neighbor. A complex
   overlapping rule is used.

Revision: 15.03.2003 E.Sack

Input : - hitlist
        - pointer to class list
Output: - line number lists
*/
int_lstm
method_line_overlapping(cand_lst hlist, cls_lst clst)
{
  int      d, found, bline = 0, lspace, no_alt = 1;
  int_lst  hnr, blines;
  int_lstm ilnr;

  init_int_lst(&blines);
  init_int_lstm(&ilnr);

  init_int_lst(&hnr);
  hnr.cnt = hlist.cnt;
  CALLOC(int, hnr.cnt, hnr.nr);                          /* working buffer */

  lspace = estimated_linespace_from_db(clst);
  reset_cand_def_aux(hlist);

  do {
    found = topmost_symbol(hlist, &hnr, no_alt);

    if (!found) {
      break;
    }

    collect_inline_neighbor(hlist, clst.cls, &hnr, no_alt);

    outline_neighbor(hlist, clst.cls, &hnr);        /* big pattern effects */

    baseline_estimation(clst.cls, hlist, hnr, lspace, &bline);

    if (no_alt) {
      overlapping_alternatives(clst.cls, hlist, &hnr, lspace, bline);
    }

    store_line(hnr, &ilnr, bline, &blines);
  } while (1);

  d = residual_alternatives(clst.cls, hlist, &ilnr, lspace, &blines);
  assert(!d);

  free_int_lst(&hnr); 
  free_int_lst(&blines); 

  return ilnr;                          /* lines sorted from top to bottom */
}



/* Extracts the adjusted line positions of a candidates position list. The
   precalculated line positions are an estimation because of unsatisfying
   precision which is based on a pixel unit. Instead of using subpixel 
   (which reasonably should be a tenth part of a pixel or less and needs much
   more calculation time) we use the candidates real position. This runs 
   fast and have the advantage of the real (averaged) place. Calculation
   starts with the best integer estimation point and is working towards its
   neighbored lines. Beware using the fixed value calculated for linespace
   for calculating neighbored baselines of a certain baseline. Instead 
   always use the baseline list.
   

Revision: 19.08.2002 E.Sack

Input : - estimated position list
        - optimal position start
        - optimal linespace
        - document image height
        - pointer to documentation path
Output: - adjusted line position list
*/
int_lst
line_positions(double_lst plst, int baseline_start, int linespace_start, 
	       int height, int_lst *docu_lines)
{
  int     pos_old, baseline, linespace_step, safe_range, upper_bound;
  int     found, countdown = 0, lower_bound, linespace;
  int_lst lines;

  init_int_lst(&lines);

  if (docu_lines) {                                /* documentation wanted */
    free_int_lst(docu_lines); 
    init_int_lst(docu_lines);
    docu_lines->cnt = height;
    CALLOC(int, docu_lines->cnt, docu_lines->nr);
    docu_lines->cnt = 0;                         /* documentation purposes */
  }

  safe_range = 2*linespace_start;                   /* state machine setup */
  upper_bound = height+safe_range;
  lower_bound = -safe_range;
  lines.cnt = upper_bound-lower_bound-1;
  CALLOC(int, lines.cnt, lines.nr);                        /* maximal list */
  lines.cnt = 0;

  linespace = linespace_start;
  linespace_step = linespace_start;                   /* state: lines down */
  baseline = baseline_start;             /* estimation best start position */
  pos_old = baseline_start;
  while (1) {                                       /* state machine entry */
    if (docu_lines &&                              /* documentation wanted */
	(baseline >= 0) && (baseline < height)) {/* visible line positions */
      docu_lines->nr[docu_lines->cnt++] = baseline;
    }

    lines.nr[lines.cnt++] = baseline;           /* all used line positions */

    if (baseline >= upper_bound) {                         /* state change */
      linespace = linespace_start;
      linespace_step = -linespace_start;                /* state: lines up */
      baseline = baseline_start;                   /* reset start position */
      pos_old = baseline_start;
      countdown = lines.cnt;                /* help for line index sorting */
    }
    else if (baseline <= lower_bound){                      /* state: exit */
      break;
    }

    baseline += linespace_step;                    /* expect next baseline */

    found = seek_start_point(plst, linespace/ALIGNMENT_DIVR, &baseline);

    if (found) {                                   /* line synchronisation */
      linespace = linespace_start;                     /* reset line space */
      pos_old = baseline;
    }
    else {    /* widen line space because period has integer unsafety of 1 */
      double delta;
      delta = fabs((double)(pos_old-baseline))/(double)(linespace_start-1);
      linespace = linespace_start + (int)delta;
    }
  }

  sort_lines(lines, countdown-1);             /* sort strange index scheme */

  return lines;
}



/* Sorts entries from index b to index a regarding following scheme:
   a  0 1 2 3 4      with N=5 and
   b  4 3 2 0 1      with b=1 as maximal down index.
   up <----|--> down
 
Revision: 19.08.2002 E.Sack

Input : - line number list
        - maximal down index
Output: - line array updated
*/
static void
sort_lines(int_lst blst, int dmax)
{
  int     a, b;
  int_lst exchange;

  assert(dmax >= 0);
  assert(dmax < blst.cnt);

  init_int_lst(&exchange);
  exchange.cnt = blst.cnt;
  CALLOC(int, exchange.cnt, exchange.nr);

  for (b = 0; b < blst.cnt; b++) {
    if (0 <= b && b <= dmax) {                           /* sorting scheme */
      a = blst.cnt - dmax - 1 + b;
    }
    else {
      a = blst.cnt - 1 - b;
    }

    exchange.nr[a] = blst.nr[b];                             /* sort lines */
  }

  for (a = 0; a < blst.cnt; a++) {
    blst.nr[a] = exchange.nr[a];                         /* exchange lines */
  }

  free_int_lst(&exchange);
}



/* Seeks the nearest h value within a given range and a given center of
   a binary signal.

Revision: 21.07.2002 E.Sack

Input : - binary signal
        - seek range
        - pointer to seek start
Output: - seek start updated as flagged
        - flag (0: no signal found, 1: signal found)
*/
static int
seek_start_point(double_lst plst, int seek_range, int *seek_start)
{
  int i, bin, opos = -1;

  assert(seek_range>=0);
  assert(ALIGNMENT_SIDE);

  for (i = 0; i <= seek_range; i++) {  /* seek nearest base-line-candidate */
    bin = *seek_start+i*(ALIGNMENT_SIDE);
    if (bin >= 0 && bin < plst.cnt) {           /* alternating up and down */
      if (plst.dbl[bin] > 0.0) {
	opos = bin;
	break;
      }
    }

    bin = *seek_start-i*(ALIGNMENT_SIDE);
    if (bin >= 0 && bin < plst.cnt) {           /* alternating up and down */
      if (plst.dbl[bin] > 0.0) {
	opos = bin;
	break;
      }
    }
  }

  if (opos < 0) {
    return 0;
  }

  *seek_start = opos;

  return 1;
}



/* Stores a candidate number line in a line array with baseline location.

Revision: 14.03.2003 E.Sack

Input : - number array
        - pointer to line array
        - pointer to baseline list
Output: - line array updated
        - baseline list updated
*/
static void 
store_line(int_lst hnr, int_lstm *iln, int bline, int_lst *baselines)
{
  if (!hnr.cnt) {
    return;
  }

  if (!iln->il) {
    CALLOC(int_lst, iln->cnt+1, iln->il);
    CALLOC(int, baselines->cnt+1, baselines->nr);
  }
  else {
    REALLOC(int_lst, iln->cnt+1, iln->il);
    REALLOC(int, baselines->cnt+1, baselines->nr);
  }

  init_int_lst(&iln->il[iln->cnt]);
  copy_int_lst(&hnr, &iln->il[iln->cnt]);
  iln->cnt++;

  baselines->nr[baselines->cnt] = bline;
  baselines->cnt++;
}



/* Decides wether a candidate is ready to serve as a validated one.
   Motivation is, that several classes can have the same NB and differs
   only by their alignment, which is not a feature of a NB. All these classes
   are enrolled in a list and need further information for filtering. As 
   long as the list contains more than one alignment, the candidates pattern 
   cannot used as a source for a certain alignment.

Revision: 01.03.2003 E.Sack

Input : - candidate
Output: - flag (0: alternatives, 1: no alternatives)
*/
static int
valid_candidate(cand_def cand)
{
  if (!cand.alt) {
    return 1;                                          /* single candidate */
  }

  if (!cand.alt->cnt) {
    return 1;                                          /* single candidate */
  }

  return 0;                                                /* alternatives */
}



/* Calculates the upper and lower line borders regarding a baseline grid.

Revision: 19.04.2003 E.Sack

Input : - absolute baseline
        - linespace to baseline above
        - linespace to baseline below
        - pointer to top line border
        - pointer to bottom line border
Output: - top and bottom updated as flagged
        - flag (0: calculation failed, 1: calculation succeeded)
*/
int
line_border(int baseline, int baseline_above, int baseline_below, 
	    double *line_top, double *line_bot)
{
  double section_above, section_below;
  double lspace_below = (double)(baseline_below-baseline);
  double lspace_above = (double)(baseline-baseline_above);

  if (lspace_below <= 0.0) {
    return 0;
  }

  section_above = 3.0*lspace_above/4.0;         /* four alignment sections */
  section_below = lspace_below/4.0;             /* four alignment sections */

  *line_bot = (double)baseline + section_below;               /* inclusive */
  *line_top = (double)baseline - section_above;               /* exclusive */

  return 1;
}



/* The topmost sector in the hitlist is searched.

Revision: 15.03.2003 E.Sack

Input : - hitlist
        - pointer to list of indizes of hitlist
        - book list
Output: - hit number list updated
        - Flag (0: no NB found, 1: NB found)
Precondition: - allocated hit number list to avoid reallocation
Postcondition: - sector may contain unrecognized pattern
*/
static int 
topmost_symbol(cand_lst hlist, int_lst *hnr, int no_alt)
{
  int i, upper_hnr = -1, zstart, top = 0, stop;

  for (i = 0, stop = 1; i < hlist.cnt; i++) {      /* initialisation values */
    if (hlist.c[i].aux == BOOKED) {
      continue;                                                   /* booked */
    }

    if (hlist.c[i].alt && no_alt) {
      continue;                                   /* don't use alternatives */
    }

    top = hlist.c[i].hdx.line.start;   /* reasonable initialisation values */
    stop = 0;                                              /* any hit found */
    break;
  }

  if (stop) {
    return 0;                                   /* hit list empty or booked */
  }

  for (i = 0; i < hlist.cnt; i++) {          /* looking for the topmost hit */
    if (hlist.c[i].aux == BOOKED) {
      continue;                                                   /* booked */
    }

    if (hlist.c[i].alt && no_alt) {
      continue;                                   /* don't use alternatives */
    }

    zstart = hlist.c[i].hdx.line.start;

    if (top >= zstart) {
      top = zstart;
      upper_hnr = i;
    }
  }                                                 /* topmost hit NB found */

  hnr->cnt = 0;                                /* initialize NB number list */
  assert(upper_hnr>=0);
  hnr->nr[hnr->cnt++] = upper_hnr;                      /* store hit number */
  hlist.c[upper_hnr].aux = BOOKED;                              /* book hit */

  return 1;
}



/* Candidates are sorted from left to right. Candidates are allowed
   superimposing each other.

Revision: 14.02.2002 E.Sack

Input : - hitlist
        - candidate number list
Output: - candidate number list sorted
Precondition: Die candidate number list contains numbers of the hitlist.
*/
void
sort_nb(cand_lst hlist, int_lst cand) 
{
  int i, ii, j, n, start;

  for (i = 0; i < cand.cnt-1; i++) {
    start = hlist.c[cand.nr[i]].hdx.column.start;

    for (j = i+1, ii = i; j < cand.cnt; j++) {
      if (start > hlist.c[cand.nr[j]].hdx.column.start) {
	start = hlist.c[cand.nr[j]].hdx.column.start;
	ii = j;
      }
    }

    if (ii == i) {
      continue;
    }

    XPSWAP(n, cand.nr[ii], cand.nr[i]);             /* candidates exchange */
  }
}



/* Generation of a text line.

Input : - number list of NB candidates in line
        - pointer to class list
        - pointer ot text buffer
        - hitlist
        - pointer to filename
        - max. allowed lag
Output: - text buffer updated
*/
static char*
separate_line(int_lst znr, cls_def *cls, cand_lst hlist, char *fname,
	      int noise_border, int lcnt)
{
  int      i;
  double   mv = 0.0, sd = 0.0;
  char     *ltxt, *txt = 0;
  int_lst  dlst, wlst, plst;
  hst_info *info = 0;

  assert(cls);
  assert(fname);

  if (!znr.cnt) {
    return 0;
  }

  init_int_lst(&dlst);
  init_int_lst(&wlst);

  CALLOC(hst_info, znr.cnt, info);
  for (i = 0; i < znr.cnt; i++) {
    init_hst_info(&info[i]);
  }

  wlst = gen_width(hlist, znr);                   /* generate width list    */
  dlst = gen_distance(hlist, znr);                /* generate distance list */
  assert(znr.cnt-1==dlst.cnt);

  ltxt = assembling_line(dlst, hlist, znr, cls);   /* initial no separation */
  histo_of_line(dlst, lcnt, ltxt, fname, xctrl.cfg.histo_write);    /* docu */
  FREEMEM(ltxt);

  separate_groups(wlst, dlst);                 /* separate string in groups */

  plst = positive_int_lst(dlst);        /* nonnegatives, separate histogram */
  if (plst.cnt) {
    mv = mean_value(plst.nr, plst.cnt, SKIP_ZERO, 0);    /* line statistics */
    sd = standard_deviation(plst.nr, plst.cnt, mv, SKIP_ZERO, 0);
    ltxt = assembling_line(dlst, hlist, znr, cls);
    histo_of_line(plst, lcnt, ltxt, fname, xctrl.cfg.histo_write);  /* docu */
    FREEMEM(ltxt);
  }
  free_int_lst(&plst);

  if (mv < sd) {
    separate_words(XPMAX(mv+sd, (double)(2*noise_border)),/*don't esti. noise */
		   dlst, &info, noise_border);
  }

  write_sep_info(dlst, znr, hlist, cls, lcnt, fname, xctrl.cfg.histo_write);

  txt = assembling_line(dlst, hlist, znr, cls);

  assembling_group(fname, lcnt, info, dlst, hlist, znr, cls);       /* docu */

  plst = positive_int_lst(dlst);                                    /* docu */
  if (plst.cnt) {
    histo_of_line(plst, lcnt, txt, fname, xctrl.cfg.histo_write);
  }
  free_int_lst(&plst);

  free_int_lst(&dlst);
  free_int_lst(&wlst);

  for (i = 0; i < znr.cnt; i++) {
    free_hst_info(&info[i]);
  }
  FREEMEM(info);

  return txt;
}



/* Documentation of groups statistic.

Input : - pointer to filename
        - line number
        - pointer to histogram infos
        - distance list
        - hitlist
        - number list of NB candidates in line
        - pointer to class list
Output: - histogram data file
*/
static void
assembling_group(char *fname, int lcnt, hst_info *info, 
		 int_lst dlst, cand_lst hlist, int_lst znr, cls_def *cls)
{
  int     i, j, iter = 0;
  char    *htxt = 0;
  int_lst rel;

  init_int_lst(&rel);

  rel.cnt = znr.cnt;
  CALLOC(int, rel.cnt, rel.nr);

  for (i = 0, j = 0, rel.cnt = 0; i < dlst.cnt; i++) {
    rel.nr[rel.cnt++] = znr.nr[i];                       /* assembling word */

    if (dlst.nr[i] == SEP_SR2 || dlst.nr[i] == SEP_WMV) {
      gen_line(cls, &htxt, 1, rel, hlist);           /* add word with blank */
      rel.cnt = 0;
    }

    if (dlst.nr[i] == SEP_SR1) {
      gen_line(cls, &htxt, 0, rel, hlist);        /* add word without blank */
      rel.cnt = 0;
      write_hst_line(WRDNAMH_DEFAULT, info[j++], lcnt, htxt, fname, /* docu */
		     iter++, xctrl.cfg.histo_write);
      htxt[0] = '\0';
    }
  }

  rel.nr[rel.cnt++] = znr.nr[dlst.cnt];           /* assembling last symbol */
  gen_line(cls, &htxt, 0, rel, hlist);            /* add word without blank */
  write_hst_line(WRDNAMH_DEFAULT, info[j], lcnt, htxt, fname,       /* docu */
		 iter++, xctrl.cfg.histo_write);

  free_int_lst(&rel);
  FREEMEM(htxt);
}



/* Assembling line with gaps indicating separation.

Input : - distance list
        - hitlist
        - number list of NB candidates in line
        - pointer to class list
Output: - pointer to line buffer
*/
static char*
assembling_line(int_lst dlst, cand_lst hlist, int_lst znr, cls_def *cls)
{
  int     i;
  char    *line = 0;
  int_lst wnr;

  if (!znr.cnt) return line;

  init_int_lst(&wnr);

  wnr.cnt = znr.cnt;
  CALLOC(int, wnr.cnt, wnr.nr);

  for (i = wnr.cnt = 0; i < dlst.cnt; i++) {
    wnr.nr[wnr.cnt++] = znr.nr[i];                       /* assembling word */

    if (dlst.nr[i] < 0) {                                      /* up to gap */
      gen_line(cls, &line, 1, wnr, hlist);           /* add word with blank */
      wnr.cnt = 0;
    }
  }

  wnr.nr[wnr.cnt++] = znr.nr[dlst.cnt];           /* assembling last symbol */
  gen_line(cls, &line, 0, wnr, hlist);            /* add word without blank */

  free_int_lst(&wnr);

  return line;
}



/* The recognized text line are written to a file in ISO 8859-1 character set.

Revision: 12.04.2004 E.Sack

Input : - pointer to document image file name
        - number of lines
        - pointer to line buffer array
Output: - text file dumped
*/
static void 
xwrite_text(char *suffix, text_def text, GtkWidget *ilog)
{
  int  i, d;
  char *bufa = NULL, *nam = NULL;
  FILE *fin;

  for (i = 0; i < text.cnt; i++) {                     /* setup whole text */
    cat_string(text.tl[i]?text.tl[i]:"", "\n", "", &bufa);
  }                                     /* empty lines only with line feed */

  if (xctrl.set.dobatch == TRUE) {   /* dump to file directly if batch job */
    cat_string(xctrl.set.image_filename, FILE_TXT, "", &nam);

    fin = xplab_open_file(nam, "w");

    if (fin == NULL) {
      FREEMEM(bufa);
      FREEMEM(nam);
      return;
    }

    d = fprintf(fin, "%s", bufa); 
    assert(d >= 0);

    xplab_close_file(fin, nam);

    FREEMEM(bufa);
    int_to_string("", text.cnt, " line", &bufa);
    cat_string(text.cnt==1?"":"s", " dumped to '", nam, &bufa);
    cat_string("'", suffix, "\n", &bufa);
  }
  else if (GTK_IS_TEXT_VIEW(ilog) == TRUE) {          /* dump to log if GUI */
    clear_gtk_text(ilog);

    xplab_text_insert(ilog, bufa);

    FREEMEM(bufa);
    int_to_string("", text.cnt, " line", &bufa);
    cat_string(text.cnt==1?"":"s", " generated ", suffix, &bufa);
    add_string("\n", &bufa);
  }

  xreport("%s", bufa);

  FREEMEM(bufa);
  FREEMEM(nam);
}



/* Eine Liste von Symbolbreiten der NB-Muster wird erzeugt.

Input:  - hitlist
        - number list of NB candidates
Output: - width list
Precondition: - Die Zeilennummern sind sortiert und die Liste enthlt 
                mindestens einen Eintrag.
Postcondition: - Die Breitenliste enthlt nur Werte grer Null.
*/
static int_lst
gen_width(cand_lst hlist, int_lst znr)
{
  int     i;
  int_lst wlst;

  assert(znr.cnt);

  init_int_lst(&wlst);

  wlst.cnt = znr.cnt;
  CALLOC(int, wlst.cnt, wlst.nr);

   /* Beachte: sort_nb */
  for (i = 0; i < wlst.cnt; i++) {
    wlst.nr[i] = hlist.c[znr.nr[i]].knl.column.end -
                 hlist.c[znr.nr[i]].knl.column.start + 1;
  }

  return wlst;
}



/* Eine Abstandsliste von Symbolen wird erzeugt.

Input:  - hitlist
        - number list of NB candidates
        - mask border size
Output: - distance list
Precondition: Die Zeilennummern sind sortiert.
Postcondition: Die Abstandsliste enthlt keine negative Werte.
*/
static int_lst
gen_distance(cand_lst hlist, int_lst znr)
{
  int     i, end;
  int_lst dlst;

  init_int_lst(&dlst);

  if (znr.cnt > 1) {
    dlst.cnt = znr.cnt-1;
    CALLOC(int, dlst.cnt, dlst.nr);               /* allocate distance-list */

    for (i = 1, end = hlist.c[znr.nr[0]].knl.column.end;/* Beachte: sort_nb */
         i < znr.cnt; 
         i++) {
      /*                         d(i,i+1) = [start(i+1)-1] - [end(i)+1] + 1 */
      dlst.nr[i-1] = XPMAX(hlist.c[znr.nr[i]].knl.column.start-end-1, 0);

      end = hlist.c[znr.nr[i]].knl.column.end;
    }
  }

  return dlst;
}



/* The symbols of the line will be separated in groups with separation rule 
   SR1.

Input:  - width list
        - distance list
Output: - distance list updated
*/
static void
separate_groups(int_lst wlst, int_lst dlst) 
{
  int i;

  assert(dlst.cnt == wlst.cnt-1);

  if (!dlst.cnt) return;                                      /* empty list */

  for (i = 0; i < dlst.cnt; i++) {
    if (((double)dlst.nr[i] >= (double)wlst.nr[i]*xctrl.cfg.syspc) &&
        ((double)dlst.nr[i] >= (double)wlst.nr[i+1]*xctrl.cfg.syspc)) {
      dlst.nr[i] = SEP_SR1;
    }
  }
}



/* Symbolgroups are separated with relaxation.

Input:  - relaxation point
        - distance list
        - documentation info
        - maximal allowed lag
Output: - distance list updated
        - histogram output
*/
static void
separate_words(double dt, int_lst dlst, hst_info **hi, int noise_border)
{
  int     i, j, first;
  int_lst dold, dnew;
  double  mv, sd;

  if (!dlst.cnt) {
    return;
  }

  init_int_lst(&dold);
  init_int_lst(&dnew);

  dold.cnt = dlst.cnt;
  CALLOC(int, dlst.cnt, dold.nr);

  for (i = 0, j = 0, dold.cnt = 0, first = 0; i < dlst.cnt; i++) {
    if (dlst.nr[i] >= 0) {                                 /* up to old gap */
      dold.nr[dold.cnt++] = dlst.nr[i];                  /* gathering group */
    }
    else {                                                    /* separation */
      if (!dold.cnt) {                           /* no group, nothing to do */
        first = i+1;                             /* may be next group start */
        j++;
        continue;                                 /* looking for next group */
      }                                       /* group found for separation */

      mv = mean_value(dold.nr, dold.cnt, SKIP_ZERO, 0);
      sd = standard_deviation(dold.nr, dold.cnt, mv, SKIP_ZERO, 0);
      (*hi)[j++] = grab_infos(dold, mv, sd);         /* grab infos for docu */

      dnew = relax_list(dold, dt, 1);               /* update distance-list */

      separate_symbol(dold, dnew, dlst, first,      /* don't estimate noise */
		      XPMAX(mv+sd, (double)(2*noise_border)));

      dold.cnt = 0;
      first = i+1;                                     /* may be next start */
      free_int_lst(&dnew);
    }
  }

  if (dold.cnt) {                                             /* last group */
    mv = mean_value(dold.nr, dold.cnt, SKIP_ZERO, 0);
    sd = standard_deviation(dold.nr, dold.cnt, mv, SKIP_ZERO, 0);
    (*hi)[j] = grab_infos(dold, mv, sd);             /* grab infos for docu */

    dnew = relax_list(dold, dt, 1);                 /* update distance-list */

    separate_symbol(dold, dnew, dlst, first,        /* don't estimate noise */
		    XPMAX(mv+sd, (double)(2*noise_border)));

    free_int_lst(&dnew);
  }

  free_int_lst(&dold);
}



/* A new list is build consisting the non-negative values of a source list.

Input:  - input list
Output: - output list
*/
static int_lst
positive_int_lst(int_lst ilst)
{
  int     i;
  int_lst olst;

  init_int_lst(&olst);

  if (!ilst.cnt) {
    return olst;
  }

  CALLOC(int, ilst.cnt, olst.nr);                /* allocate distance-list */

  for (i = olst.cnt = 0; i < ilst.cnt; i++) {
    if (ilst.nr[i] >= 0) {
      olst.nr[olst.cnt++] = ilst.nr[i];    /* gathering non-negative values */
    }
  }

  return olst;
}



/* Ein Wort wird zusammengesetzt und der Zeile hinzugefgt.

Input : - Klassenliste
        - Zeiger auf Textzeile
        - Flag (0: keine Blank, 1: Blank)
        - NB-Nummernliste
        - Trefferliste
Output: - Textzeile auf Stand
Precondition: - die NB sind (vlnr) sortiert.
*/
static void 
gen_line(cls_def *cls, char **txt, int spc, int_lst wnr, cand_lst hlist) 
{
  int  i, n, c, wlen, tlen;
  char *sig = NULL, *word = NULL;

  CALLOC(char, 1, word);

  for (i = 0, word[0] = '\0', wlen = 1; i < wnr.cnt; i++) {
    n = hlist.c[wnr.nr[i]].n;
    c = hlist.c[wnr.nr[i]].c;

    FREEMEM(sig);
    switch (n) {
      case SUBSTNB_FLAG: add_string(xctrl.cfg.subst_nb, &sig); break;
      case SUBSTRB_FLAG: add_string(xctrl.cfg.subst_rb, &sig); break;
      default:           add_string(cls[c].att.sign, &sig); break;
    }

    wlen += strlen(sig);
    REALLOC(char, wlen, word);
    strcat(word, sig);
  }

  if (*txt) {
    tlen = strlen(*txt);
    REALLOC(char, tlen + wlen, *txt);
  }
  else {
    tlen = 0;
    CALLOC(char, tlen + wlen, *txt);
  }

  strcat(*txt, word);               /* Wort an das Zeilentextende kopieren */

  if (spc) {
    tlen = strlen(*txt);
    REALLOC(char, tlen + 2, *txt);
    strcat(*txt, " ");                             /* Leerzeichen einfgen */
  }

  FREEMEM(sig);
  FREEMEM(word);
}



/* Die Symbole werden nach Trennregel SR2 und SR3 in Worte getrennt.

Input:  - Abstandsliste der Gruppe
        - Abstandsliste der Gruppe nach Relaxation
        - Abstandsliste der Zeile
Output: - Abstandsliste der Zeile auf Stand
Precondition: - Die Gruppen-Listen sind Ausschnitte aus der Zeilenliste.
              - Erwartungswerte der Abstnde sind nicht negativ.
*/
static void
separate_symbol(int_lst dold, int_lst dnew, int_lst dlst, int start, double df)
{
  int i;

  assert(dold.cnt == dnew.cnt);
  assert(dlst.cnt >= dnew.cnt);

  for (i = 0; i < dold.cnt; i++) {     /* looking for enlarged differences */
    if (dold.nr[i] < dnew.nr[i]) {         /* difference shows word bounds */
      assert(dlst.nr[i+start] != SEP_SR1);
      assert(dlst.nr[i+start] != SEP_SR2);
      assert(dlst.nr[i+start] != SEP_WMV);
      dlst.nr[i+start] = SEP_SR2;                       /* SR2             */
    }
    else if ((double)dnew.nr[i] >= df) {                /* gap point found */
      assert(dlst.nr[i+start] != SEP_SR1);
      assert(dlst.nr[i+start] != SEP_SR2);
      assert(dlst.nr[i+start] != SEP_WMV);
      dlst.nr[i+start] = SEP_WMV;                       /* SR3             */
    }
  }
}



/* Generates the position list of the hitlist as a unipolar binary signal.
   Trailing zeros are discarded.

Revision: 21.04.2003 E.Sack

Input:  - pointer to class list
        - hitlist
Output: - position list 
        - position list dumped to file
Postcondition: - position list may be empty
*/
static double_lst
position_list(char *filename, cls_def *cls, cand_lst hlist)
{
  int        d, i, j, class, bline;
  double_lst plst;

  assert(cls);

  init_double_lst(&plst);

  for (i = 0; i < hlist.cnt; i++) {     
    d = valid_candidate(hlist.c[i]);

    if (!d) {
      continue;                        /* deferred as source for alignment */
    }

    class = hlist.c[i].c;

    if (class < 0) {                                 /* unrecognized class */
      continue;
    }

    if (cls[class].att.alignment == BASELINE) {/*search baseline candidate */
      bline = hlist.c[i].hdx.line.end;               /* baseline of symbol */

      if (!plst.dbl) {
	plst.cnt = bline+1;
	CALLOC(double, plst.cnt, plst.dbl);

	for (j = 0; j < bline; j++) {
          plst.dbl[j] = UNIPOLAR_MIN;                /* init position list */
	}
      }
      else if (plst.cnt < bline+1) {
	REALLOC(double, bline+1, plst.dbl);

	for (j = plst.cnt; j < bline; j++) {
          plst.dbl[j] = UNIPOLAR_MIN;                /* init position list */
	}

	plst.cnt = bline+1;
      }

      plst.dbl[bline] = UNIPOLAR_MAX;              /* update position list */
    }
  }

  dc_source_double(filename, plst, DC_BIT_POS, "");

  return plst;
}



/* Picks up alternative candidates in a line.

Usage: This routine shall be used for residual alternatives which could not
       assembled in lines elsewhere. Therefore alternatives found here
       always are put in new lines.

Revision: 15.03.2003 E.Sack

Input : - pointer to class list
        - hitlist
        - pointer to candidate number list
        - linespace
        - pointer to baseline list
Output: - line number list updated
        - baseline list updated
        - flag (0: no candidates left, 1: number of candidates left)
Precondition: number list and baseline list coincides.
Postcondition: left candidates are booked.
*/
static int
residual_alternatives(cls_def *cls, cand_lst hlist, int_lstm *ilnr,
		      int lspace, int_lst *blines)
{
  int     cand, found, latest, latest_bot, trend, new_idx, bline;
  int     candidates_left = 0;
  int_lst hnr_alt;
  ALIGMNT ideal_alg;

  assert(cls);
  assert(lspace>0);
  assert(ilnr->cnt==blines->cnt);

  init_int_lst(&hnr_alt);
  hnr_alt.cnt = 1;
  CALLOC(int, hnr_alt.cnt, hnr_alt.nr);                  /* working buffer */

  do {
    found = topmost_symbol(hlist, &hnr_alt, 0);  /* the only in local list */

    if (!found) {
      break;
    }

    latest = hnr_alt.cnt-1;
    assert(!latest);
    cand = hnr_alt.nr[latest];

    if (hlist.c[cand].alt) {
      assert(hlist.c[cand].c >= 0);    /* unclassified cannot have altern. */

      latest_bot = hlist.c[cand].hdx.line.end;

      bline = best_mid(*blines, latest_bot, &new_idx);

      ideal_alg = alignment_of_point(bline, lspace, latest_bot, &trend);
      assert(ideal_alg>UDEFLINE);

      best_alignment(cls, &hlist.c[cand], ideal_alg, trend); /* force line */

      insert_int_lstm(ilnr, new_idx);

      ilnr->il[new_idx].cnt = 1;
      CALLOC(int, 1, ilnr->il[new_idx].nr);
      ilnr->il[new_idx].nr[0] = cand;/* store candidate number in new list */
      hlist.c[cand].aux = BOOKED;                        /* book candidate */

      collect_inline_neighbor(hlist, cls, &ilnr->il[new_idx], 0);

      outline_neighbor(hlist, cls, &ilnr->il[new_idx]);/* big pattern eff. */

      baseline_estimation(cls, hlist, ilnr->il[new_idx], lspace, &bline);

      insert_int_lst(blines, new_idx);
      blines->nr[new_idx] = bline;          /* store baseline in new entry */
    }
    else {
      hlist.c[cand].aux = BOOKED;                        /* book candidate */
      candidates_left++;                           /* candidates untreated */
    }
  } while (1);

  free_int_lst(&hnr_alt);

  return candidates_left;
}



/* Calculates the center of baselines surrounding a location.

Usage: Assuming a missing line between two lines their center should be
       an estimation of the center of the missing line.

Revision: 15.03.2003 E.Sack

Input:  - baseline list
        - location
        - pointer to new index
Output: - baseline
        - new index updated
*/
static int
best_mid(int_lst blines, int loc, int *new_idx)
{
  int l, bot, top, top_1;

  assert(new_idx);

  if (!blines.cnt) {
    *new_idx = 0;                                              /* new line */
    return loc;
  }

  top = top_1 = blines.nr[0];

  if (blines.cnt == 1) {
    *new_idx = loc < top ? 0:1;    /* new line at the top or at the bottom */
    return loc;
  }

  assert(top < blines.nr[blines.cnt-1]);              /* order consistency */

  bot = blines.nr[1];
  for (l = 1; l < blines.cnt; l++) {
    bot = blines.nr[l];

    if (top <= loc &&  loc < bot) {
      *new_idx = l;
      return (top+bot) >> 1;                                     /* center */
    }

    top_1 = top;
    top = bot;
  }

  *new_idx = blines.cnt;                         /* new line at the bottom */
  return bot + XPMAX(1, ((top-1+bot) >> 1));                     /* center */
}



/* Assembles alternative candidates in a line.

Revision: 15.03.2003 E.Sack

Input : - pointer to class list
        - hitlist
        - pointer to candidate number list
        - linespace
        - baseline
Output: - line number list updated */
static void
overlapping_alternatives(cls_def *cls, cand_lst hlist, int_lst *hnr,
			 int lspace, int bline)
{
  int     cand, trend, line_offset;
  ALIGMNT ideal_alg;

  assert(cls);

  for (cand = 0; cand < hlist.cnt; cand++) {
    if (hlist.c[cand].aux == BOOKED) {
      continue;                                                  /* booked */
    }

    if (!hlist.c[cand].alt) {
      continue;                                         /* no alternatives */
    }

    assert(hlist.c[cand].c >= 0); /* unclassified cannot have alternatives */

    ideal_alg = alignment_of_point(bline, lspace, 
				   hlist.c[cand].hdx.line.end, &trend);
    assert(ideal_alg>UDEFLINE);

    line_offset = best_alignment(cls, &hlist.c[cand], ideal_alg, trend);

    if (!line_offset) {                                       /* same line */
      hnr->nr[hnr->cnt++] = cand;                /* store candidate number */
      hlist.c[cand].aux = BOOKED;                        /* book candidate */
    }
  }
}



/* Estimates baseline location of candidates whose numbers are stored in
   a list.

Revision: 15.03.2003 E.Sack

Input : - pointer to class list
        - hitlist
        - candidate number list
        - linespace
        - pointer to baseline
Output: - baseline updated
        - flag (0: coarse baseline estimation, 1: statistic estimation)
Bug: No algebraic solution.
*/
static int
baseline_estimation(cls_def *cls, cand_lst hlist, int_lst hnr, int lspace,
		    int *baseline)
{
  int      und, mid, estimation = 1, fallback;
  int_lstm alst;

  alst = estimated_alignment(cls, hlist, hnr, &fallback);

  assert(lspace > 0);
  assert(UPPERLINE > UNDERLINE);
  assert(UPPERLINE - UNDERLINE == 3);      /* ensure fixed line separation */
  assert(alst.cnt == 4);                                  /* ensure preset */
  assert(alst.il);                                        /* ensure preset */

  if (alst.il[1].nr) {                     /* does 'base' alignment exist? */
    assert(alst.il[1].cnt = 1);                           /* ensure preset */
    *baseline = alst.il[1].nr[0];
  }
  else if (alst.il[0].nr && alst.il[2].nr) {/*does 'under'&'middle' exist? */
    assert(alst.il[0].cnt = 1);                           /* ensure preset */
    assert(alst.il[2].cnt = 1);                           /* ensure preset */
    und = alst.il[0].nr[0];
    mid = alst.il[2].nr[0];
    assert(und < mid);                                  /* ensure validity */
    *baseline = (und + mid) >> 1;
  }
  else if (alst.il[0].nr) {                         /* does 'under' exist? */
    assert(alst.il[0].cnt = 1);                           /* ensure preset */
    *baseline = alst.il[0].nr[0] - XPMAX(1, lspace>>2);
  }
  else if (alst.il[2].nr) {                        /* does 'middle' exist? */
    assert(alst.il[2].cnt = 1);                           /* ensure preset */
    *baseline = alst.il[2].nr[0] + XPMAX(1, lspace>>2);
  }
  else if (alst.il[3].nr) {                         /* does 'upper' exist? */
    assert(alst.il[3].cnt = 1);                           /* ensure preset */
    *baseline = alst.il[3].nr[0] + XPMAX(1, lspace>>1);
  }
  else {
    *baseline = fallback;/*really a coarse estimation of all bottom points */
    estimation = 0;               /* list empty - fallback estimation used */
  }

  free_int_lstm(&alst);

  return estimation;
}



/* Calculates an estimation for a linespace of a given class list.

Revision: 13.03.2003 E.Sack

Input:  - class list
Output: - linespace
*/
static int
estimated_linespace_from_db(cls_lst clst)
{
  int     nnr, cnr, max_height = 0, max_cnr = 0, linespace = 0, height;
  ALIGMNT aln;

  assert(UPPERLINE - UNDERLINE == 3);      /* ensure fixed line separation */

  for (cnr = 0; cnr < clst.cnt; cnr++) {/* maximal NB height in class list */
    aln = clst.cls[cnr].att.alignment;

    if (aln < UNDERLINE || aln > UPPERLINE) {
      continue;
    }

    for (nnr = 0; nnr < clst.cls[cnr].nbl.cnt; nnr++) {
      height = clst.cls[cnr].nbl.nb[nnr].pat.ar.height;

      if (max_height < height) {
	max_height = height;
	max_cnr = cnr;
      }
    }
  }

  if (max_height) {
    switch (clst.cls[max_cnr].att.alignment) {
    case UPPERLINE: 
      linespace = 4*max_height;
      break;
    case MIDDLELINE:
      linespace = 2*max_height;
      break;
    case BASELINE:
      linespace = (int)ceil((double)(4*max_height)/3.0);
      break;
    case UNDERLINE:
      linespace = max_height;
      break;
    default: 
      assert(0); 
      break;
    }
  }

  /* Minimum of 1 pixel for every interval including 1 pixel of space      */
  linespace = XPMAX(linespace, 5);

  return linespace;
}



/* Calculates estimations for alignments.

Revision: 13.03.2003 E.Sack

Input:  - class list
        - candidate list
        - number list
Output: - alignment list
*/
static int_lstm
estimated_alignment(cls_def *cls, cand_lst hlist, int_lst hnr, int *fallback)
{
  int      cand, cnt, hcnt, nr, cnr, adx, unc_cnt = 0, unc_sum = 0;
  int_lstm alst;
  ALIGMNT  ifs, aln;

  assert(cls);
  assert(UPPERLINE > UNDERLINE);

  cnt = UPPERLINE - UNDERLINE + 1;
  assert(cnt == 4);                        /* ensure fixed line separation */

  ifs = UNDERLINE;

  init_int_lstm(&alst);
  alst.cnt = cnt;
  CALLOC(int_lst, alst.cnt, alst.il);                    /* all alignments */

  for (adx = 0; adx < cnt; adx++) {
    init_int_lst(&alst.il[adx]);
    alst.il[adx].cnt = 0;                                    /* fake count */
    CALLOC(int, hnr.cnt, alst.il[adx].nr);          /* allocate max. count */
  }

  for (nr = 0; nr < hnr.cnt; nr++) {
    cand = hnr.nr[nr];
    cnr = hlist.c[cand].c;

    if (cnr < 0) {
      unc_cnt++;                                    /* fallback estimation */
      unc_sum += hlist.c[cand].hdx.line.end;
      continue;                                    /* unclassified pattern */
    }

    aln = cls[cnr].att.alignment;

    if (aln < UNDERLINE || aln > UPPERLINE) {        /* special alignments */
      continue;
    }

    hcnt = alst.il[aln-ifs].cnt++;               /* histogram of alignment */
    alst.il[aln-ifs].nr[hcnt] = hlist.c[cand].hdx.line.end;
  }

  for (adx = 0; adx < cnt; adx++) {
    if (!alst.il[adx].cnt) {
      free_int_lst(&alst.il[adx]);                 /* make list consistent */
    }
  }

  for (adx = 0; adx < cnt; adx++) {   /* mean value of alignments location */
    if (!alst.il[adx].cnt) {
      continue;
    }

    alst.il[adx].nr[0] = (int)mean_value(alst.il[adx].nr, alst.il[adx].cnt, 
					 SKIP_ZERO, 0);
    alst.il[adx].cnt = 1;                       /* meaning of list changed */
  }

  validate_alignment(alst);                /* valid location of alignments */

  *fallback = (int)((double)unc_sum/(double)XPMAX(1,unc_cnt));

  return alst;
}



/* Validates alignments.

Revision: 14.03.2003 E.Sack

Input:  - alignments location list
Output: - alignment location list updated
Bug: No algebraic solution.
*/
static void
validate_alignment(int_lstm alst)
{
  int und = 0, bas = 0, mid = 0, upp = 0;
  int und_err = 0, mid_err = 0, upp_err = 0;
  int und_val = 0, bas_val = 1, mid_val = 2, upp_val = 3;

  assert(UPPERLINE > UNDERLINE);
  assert(UPPERLINE - UNDERLINE == 3);      /* ensure fixed line separation */
  assert(alst.cnt == 4);                                  /* ensure preset */
  assert(alst.il);                                        /* ensure preset */

  if (alst.il[0].nr) {                            /* does alignment exist? */
    assert(alst.il[0].cnt = 1);                           /* ensure preset */
    und = 1;                                            /* alignment exist */
    und_val = alst.il[0].nr[0];
  }

  if (alst.il[1].nr) {                            /* does alignment exist? */
    assert(alst.il[1].cnt = 1);                           /* ensure preset */
    bas = 1;                                            /* alignment exist */
    bas_val = alst.il[1].nr[0];
  }

  if (alst.il[2].nr) {                            /* does alignment exist? */
    assert(alst.il[2].cnt = 1);                           /* ensure preset */
    mid = 1;                                            /* alignment exist */
    mid_val = alst.il[2].nr[0];
  }

  if (alst.il[3].nr) {                            /* does alignment exist? */
    assert(alst.il[3].cnt = 1);                           /* ensure preset */
    upp = 1;                                            /* alignment exist */
    upp_val = alst.il[3].nr[0];
  }

  if (und && bas && (und_val <= bas_val)) {             /* |  UPPER lose   */
    und_err = 1;                                        /* |    |          */
  }                                                     /* |    V          */
                                                        /* |  MIDDLE       */
  if (und && mid && (und_val <= mid_val)) {             /* |    |          */
    und_err = 1;                                        /* |    V          */
  }                                                     /* |  BASE wins    */
                                                        /* |    |          */
  if (und && upp && (und_val <= upp_val)) {             /* |    V          */
    upp_err = 1;                                        /* |  UNDER        */
  }                                                     /* V               */
                                                        /* y ordinate incr.*/
  if (bas && mid && (bas_val <= mid_val)) {
    mid_err = 1;
  }

  if (bas && upp && (bas_val <= upp_val)) {
    upp_err = 1;
  }

  if (mid && upp && (mid_val <= upp_val)) {
    upp_err = 1;
  }

  if (und_err) {
    free_int_lst(&alst.il[0]);
  }

  if (mid_err) {
    free_int_lst(&alst.il[2]);
  }

  if (upp_err) {
    free_int_lst(&alst.il[3]);
  }
}

