/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <assert.h>
#include <math.h>
#include <unistd.h>
#include <time.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_preferences_fe.h"
#include "xplab_preferences_be.h"
#include "xplab_dchannel.h"
#include "xplab_algorithm.h"
#include "xplab_image.h"
#include "xplab_report.h"
#include "xplab_maintenance_be.h"

/* local prototypes ********************************************************/

static rb_lst rb_seek(pic_def pic);
static void merge_rb(rb_lst rbl, pic_def pic);
static void create_rb(rb_def *rb, pic_def bmp_bild);
static int pixel_on_frame(pic_def pic, sector_def frame, c_def *point);
static int evaluate_support_dots(db_def edb, char *tps_id, int force_support_dots, int support_dots_auto);
static rb_lst bounding_box(char *image_filename, pic_def rimg, int support_dots);
static void check_rb_frame(rb_lst rblst);
static int calculate_has_dots(cls_lst clst);
static void stats_size_setting(GtkWidget *cbox, stats_prefs *sentr);

static void cb_stats_check(GTK_CONVENTION_ENTRY_ACTIVATE_RTS_VOID);
static void cb_stats_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_stats_default(GTK_CONVENTION_BUTTON_RTS_VOID);

/* Functions ***************************************************************/

/* Assembles current settings to a string.

Revision: 20.08.2005 E.Sack

Input : - program settings
        - pointer to header string
        - pointer to DB filename string
Output: - pointer to string
Usage:  - only procedural settings are to be printed.
*/
char *
print_settings(prog_ctrl ctrl, char *header, char *db_name)
{
  int    d;
  time_t timep = time((time_t *)NULL);
  char   *buf0 = NULL, buf1[MAXCHAR], buf2[MAXCHAR], *p = NULL, *settings = 0;
  char   formats[] = "%13s: %-18s\n", formatd[] = "%13s: %-18d\n";
  char   formatf[] = "%13s: %-18.4f\n";

  if (!ctrl.cfg.timestamp) {
    add_string(NO_TIMESTAMP, &buf0);
  }
  else {
    add_string(ctime(&timep), &buf0);
    p = strchr(buf0, '\n');

    if (p != NULL) {
      *p = 0;
    }
  }

  cat_string(XPLAB_VERSION "  ", header, "    ", &settings);
  add_string(buf0, &settings);
  FREEMEM(buf0);

  cat_string("\n\n", ctrl.set.user_options, "\n\nparameter set:\n", 
	     &settings);

  d = check_name(ctrl.set.image_filename);
  g_snprintf(buf1, sizeof(buf1), formats, PINPUT_KEY, 
	     d?NO_INPUT:ctrl.set.image_filename);
  add_string(buf1, &settings);

  d = check_name(db_name);
  assert(!d);

  g_snprintf(buf1, sizeof(buf1), formats, DBNAME_KEY, db_name);
  add_string(buf1, &settings);

  d = check_name(ctrl.cfg.tps_id);
  if (!d) {
    g_snprintf(buf1, sizeof(buf1), formats, TPSID_KEY, ctrl.cfg.tps_id);
    add_string(buf1, &settings);
  }

  g_snprintf(buf1, sizeof(buf1), formatf, LEVEL_KEY, ctrl.cfg.mad.corrlevel);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatf, BROKEN_LEVEL_KEY, ctrl.cfg.mad.corrbroken);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatf, CLUSTER_LVL_KEY, ctrl.cfg.cluster_level);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatf, CLUSTER_STEP_KEY, ctrl.cfg.cluster_step);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, NNBMATCH_KEY, ctrl.cfg.mad.nnbmatch);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatf, INTERLEAVE_KEY, ctrl.cfg.mad.interleave);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatf, OVERLAP_KEY, ctrl.cfg.mad.overlap);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatf, NB_LAP_KEY, ctrl.cfg.mad.nb_lap);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatf, SYSPC_KEY, ctrl.cfg.syspc);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, MASKWEIGHT_KEY, ctrl.cfg.mad.maskweight);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, MBORDER_KEY, ctrl.cfg.mad.mborder);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, NBORDER_KEY, ctrl.cfg.mad.nborder);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, DC_STORE_KEY, ctrl.cfg.dc_store);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, THOLD_KEY, ctrl.cfg.histo_thold);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, RAWBIT_KEY, ctrl.cfg.rawbit);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, REPORT_WRITE_KEY, ctrl.cfg.report_write);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, RB_CHAIN_KEY, ctrl.cfg.rb_chain);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, HISTO_WRITE_KEY, ctrl.cfg.histo_write);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, DURATION_KEY, ctrl.cfg.duration);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, SUPPORT_DOTS_KEY, ctrl.cfg.support_dots);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, SUPPORT_DOTS_AUTO_KEY, ctrl.cfg.support_dots_auto);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, OVERWRITE_IMAGE_KEY, ctrl.cfg.overwrite_image);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, REPLAY_WRITE_KEY, ctrl.cfg.replay_write);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, TRAINING_WRITE_KEY, ctrl.cfg.training_write);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, SINGLE_LINE_KEY, ctrl.cfg.single_line);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, SINGLE_NB_KEY, ctrl.cfg.single_nb);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, BROKEN_NB_KEY, 
	   (ctrl.cfg.match_method == MATCHING_METHOD_2) ? 1 : 0);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, RENEW_METHOD_KEY, ctrl.cfg.renew_method);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, CLUSTER_FORCE_KEY, ctrl.cfg.cluster_force);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, CLUSTER_WRITE_KEY, ctrl.cfg.cluster_write);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, CLUSTER_NB_KEY, ctrl.cfg.cluster_nb);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, DOMAIN_KEY, ctrl.cfg.cdomain);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, DC_1D_KEY, ctrl.cfg.dc_1d);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, DC_2D_KEY, ctrl.cfg.dc_2d);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, VERBOSE_KEY, ctrl.cfg.verbose);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, BATCH_PHASE_KEY, ctrl.cfg.batch.phase);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, BATCH_LOG_KEY, ctrl.cfg.batch.blog);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, TIPS_KEY, ctrl.cfg.tips);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, RAISE_KEY, ctrl.cfg.raise);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, LINE_METHOD_KEY, ctrl.cfg.line_method);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, TIMESTAMP_KEY, ctrl.cfg.timestamp);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formats, SUBSTRB_KEY, ctrl.cfg.subst_rb);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formats, SUBSTNB_KEY, ctrl.cfg.subst_nb);
  add_string(buf1, &settings);

  g_snprintf(buf2, sizeof(buf2), "%d %d", ctrl.cfg.nb_border_x, ctrl.cfg.nb_border_y);
  g_snprintf(buf1, sizeof(buf1), formats, NB_BORDER_KEY, buf2);
  add_string(buf1, &settings);

  g_snprintf(buf2, sizeof(buf2), "%.3g %.3g %.3g", ctrl.cfg.rgb_norm.r, 
	     ctrl.cfg.rgb_norm.g, ctrl.cfg.rgb_norm.b);
  g_snprintf(buf1, sizeof(buf1), formats, RGB_NORM_KEY, buf2);
  add_string(buf1, &settings);

  g_snprintf(buf2, sizeof(buf2), "%d %d %d %c", 
	     ctrl.cfg.ckernel[0], 
	     ctrl.cfg.ckernel[1], 
	     ctrl.cfg.ckernel[2], 
	     LINE_CONT_SIGN);
  g_snprintf(buf1, sizeof(buf1), formats, CKERNEL_KEY, buf2);
  add_string(buf1, &settings);

  g_snprintf(buf2, sizeof(buf2), "%d %d %d %c", 
	     ctrl.cfg.ckernel[0 + CKERNEL_W], 
	     ctrl.cfg.ckernel[1 + CKERNEL_W], 
	     ctrl.cfg.ckernel[2 + CKERNEL_W], 
	     LINE_CONT_SIGN);
  g_snprintf(buf1, sizeof(buf1), "%13s  %-18s\n", "", buf2);
  add_string(buf1, &settings);

  g_snprintf(buf2, sizeof(buf2), "%d %d %d", 
	     ctrl.cfg.ckernel[0 + 2*CKERNEL_W], 
	     ctrl.cfg.ckernel[1 + 2*CKERNEL_W], 
	     ctrl.cfg.ckernel[2 + 2*CKERNEL_W]);
  g_snprintf(buf1, sizeof(buf1), "%13s  %-18s\n", "", buf2);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), formatd, CHOLD_KEY, ctrl.cfg.cluster_thold);
  add_string(buf1, &settings);

  g_snprintf(buf1, sizeof(buf1), "\n");
  add_string(buf1, &settings);

  return settings;
}



/* Reads image from hard disc and converts it to bitmap.

Revision: 01.11.2004 E.Sack

Input : - database
        - pointer to image file name
        - settings
        - verbose flag
        - pointer to RB list
        - pointer to original image
        - pointer to session image
Output: - binary image
        - original image updated
        - session image updated
Precondition:  - if database is empty, some features cannot be calculated.
               - pointer to original image may be NULL
               - pointer to session image may be NULL
               - pointer to RB list may be NULL
Postcondition: - whenever a session image was loaded, it is converted to
                 its binary counterpart, the bit-mapped original image.
*/
img_typ
setup_input_image(db_def idb, char *filename, cfg_def cfg, rb_lst *rblst,
		  hst_def *grey_hist, hst_def *th_hist, img_typ *porig, 
		  img_typ *ses)
{
  int     mgv, support_dots, verbose = cfg.verbose > BE_QUIET;
  img_typ rimg, tmp_pic;

  init_img_typ(&rimg);

  tmp_pic = read_pnm_file(filename, &mgv, verbose); /* read image from HDD */

  if (tmp_pic.typ == PNM_FORMAT_UNDEF) {                  /* no image read */
    assert(AREA_SIZE(tmp_pic.pimg.ar) == 0);          /* consistency check */
    free_img_typ(&tmp_pic);                                     /* why not */
    return rimg;
  }

  if (porig) {                                   /* retain new input image */
    free_img_typ(porig);                           /* lose old input image */
    *porig = xdup_img_typ(tmp_pic);           /* copy from new input image */
  }

  rimg = xdup_img_typ(tmp_pic);               /* copy from new input image */

  free_img_typ(&tmp_pic);                     /* free temporary image ASAP */

  dc_source_pic(filename, rimg.pimg, DC_BIT_ORG);                    /* DC */

  pnm_to_bitmap(&rimg, mgv, filename, cfg.histo_thold, ses,  /* PreProc. I */ 
		grey_hist, th_hist);                                

  dc_source_pic(filename, rimg.pimg, DC_BIT_BMP);                    /* DC */

  verify_bitmap(rimg.pimg, rimg.pimg.ar, 0, 0);

  support_dots = evaluate_support_dots(idb, cfg.tps_id,     /* PreProc. II */
				       cfg.support_dots,
				       cfg.support_dots_auto);

  if (rblst) {
    free_rb_lst(rblst);                                    /* PreProc. III */
    *rblst = bounding_box(filename, rimg.pimg, support_dots);
  }

  return rimg;                                   /* bit-mapped input image */
}



/* Generates RB list and dumps image files.

Revision: 20.10.2004 E.Sack

Input : - image
        - settings
Output: - image updated
        - RB list
*/
static rb_lst
bounding_box(char *image_filename, pic_def rimg, int support_dots)
{
  int    rb_cnt, dots;
  rb_lst rblst;

  assert(image_filename);

  init_rb_lst(&rblst);

  if (!AREA_SIZE(rimg.ar)) {
    return rblst;
  }

  dots = pixel_erase(rimg, NBH8, support_dots);          /* eliminate dots */ 
  xreport("number of dots: %d (dots %s)\n", dots, 
	  support_dots?"supported":"deleted");

  dc_source_pic(image_filename, rimg, DC_BIT_CLEAN);                 /* DC */

  rb_cnt = create_rb_list(&rblst, rimg);

  if (rb_cnt) {
    dc_source_rb(image_filename, rblst);                             /* DC */

    check_rb_frame(rblst);                           /*check image quality */
  }
  else {
    xreport("image '%s' empty\n\n", image_filename);
  }

  return rblst;
}



/* Check RB frames, for documentation only.

Revision: 30.03.2002 E.Sack

Input : - RB list
Output: - warnings
*/
static void 
check_rb_frame(rb_lst rblst) 
{
  int errorsum = 0, number = rblst.cnt;

  while (number--) {
    assert(rblst.rb[number].sc.line.start >= 0);
    assert(rblst.rb[number].sc.line.end <= rblst.rar.height-1);
    assert(rblst.rb[number].sc.column.start >= 0);
    assert(rblst.rb[number].sc.column.end <= rblst.rar.width-1);
  }

  number = rblst.cnt;

  while (number--) {
    if (rblst.rb[number].sc.line.start == 0) {
      xreport(errorsum?"":"\n");
      xreport("WARNING: RB No. %d touching upper document border\n", 
	     number);
      errorsum++;
    }

    if (rblst.rb[number].sc.line.end == rblst.rar.height-1) {
      xreport(errorsum?"":"\n");
      xreport("WARNING: RB No. %d touching lower document border\n", 
	     number);
      errorsum++;
    }

    if (rblst.rb[number].sc.column.start == 0) {
      xreport(errorsum?"":"\n");
      xreport("WARNING: RB No. %d touching left document border\n", 
	     number);
      errorsum++;
    }
      
    if (rblst.rb[number].sc.column.end == rblst.rar.width-1) {
      xreport(errorsum?"":"\n");
      xreport("WARNING: RB No. %d touching right document border\n", 
	     number);
      errorsum++;
    }
  }

  if (errorsum) {
    xreport("WARNING: %d image border touch%s\n", errorsum, 
	   errorsum==1?"":"es");
  }

  xreport(errorsum?"\n":"");
}



/* Evaluates the current support dots flag.

Revision: 22.10.2004 E.Sack

Input : - database
        - pointer to TPS ID
        - flag (0: don't force support, 1: force support)
        - flag (0: don't support auto, 1: support auto)
Output: - settings updated
Postcondition: - force bit overwrites auto bit
               - auto bit enables database driven support
*/
static int
evaluate_support_dots(db_def edb, char *tps_id, int force_support_dots, 
		      int support_dots_auto)
{
  int tps_idx, support_dots = 0, tps_has_dots = 0;

  tps_idx = search_tps_id(&tps_id,         /* pointer change has no effect */
			  edb.tpl);

  if (tps_idx >= 0) {
    tps_has_dots = calculate_has_dots(edb.tpl.tps[tps_idx].cl);
  }

  if (force_support_dots) {
    support_dots = force_support_dots;              /* support dots always */
  }
  else if (support_dots_auto) {                    /* support driven by DB */
    support_dots = tps_has_dots;
  }

  if (tps_has_dots && !support_dots) {
    assert(tps_idx>=0);
    int_to_log(main_log, "INFO: TPS[", tps_idx, "] ");
    string_to_log(main_log, "(", edb.tpl.tps[tps_idx].ident,
		  ") has unsupported dot-NBs\n");
  }

  return support_dots;
}



/* Returns the flag wether there are NBs in the class list which are just
   one pixel.

Revision: 22.10.2004 E.Sack

Input : - class list
Output: - flag (0: all NBs are greater than on pixel, 
                1: there are NBs of one pixel)
*/
static int
calculate_has_dots(cls_lst clst)
{
  int cnr, nnr;

  for (cnr = 0; cnr < clst.cnt; cnr++) {
    for (nnr = 0; nnr < clst.cls[cnr].nbl.cnt; nnr++) {
      if (clst.cls[cnr].nbl.nb[nnr].pat.ar.width > 1) {/* avoid multiplication */
	continue;
      }

      if (clst.cls[cnr].nbl.nb[nnr].pat.ar.height > 1) {
	continue;
      }

      return 1;                                          /* one pixel size */
    }
  }

  return 0;
}



/* The RB image and the RB list are build from a bitmap image.

Revision: 11.01.2003 E.Sack

Input : - pointer to RB list
        - document image
Output: - RB list updated
        - RB number
*/
int
create_rb_list(rb_lst *rblst, pic_def document) 
{
  int     rb_cnt_old = 0;
  pic_def zpic;

  init_pic_def(&zpic);

  assert(document.mem && !document.col);                  /* ensure bitmap */

  copy_pic_def(&document, &zpic);

  while (1) {                  /* merge RB until all overlappings are found */
    free_rb_lst(rblst);

    *rblst = rb_seek(zpic);                                /* search for RB */

    if (rb_cnt_old == rblst->cnt) {
      break;
    }
    else {
      rb_cnt_old = rblst->cnt;
    }
   
    merge_rb(*rblst, zpic);
  }

  free_pic_def(&zpic);

  if (rblst->cnt) {
    REALLOC(rb_def, rblst->cnt, (*rblst).rb);              /* adapt RB list */

    rb_sort(rblst);                                           /* RB sorting */
  }
  else {
    free_rb_lst(rblst);                                    /* RB list empty */
  }

  xreport("number of RBs: %d\n", rblst->cnt);

  if(xctrl.cfg.report_write & REPORT_WRITE_RBS) {
    char *buf = print_report_rbs(*rblst);
    xreport(buf);
    FREEMEM(buf);
  }

  return rblst->cnt;
}



/* RB frames are filled with foreground color and plotted into a image.
   Serves to merge RBs.

Revision: 30.03.2002 E.Sack

Input : - RB list
        - image
Output: - image updated
Precondition: RB list must be consistent with image.
*/
static void 
merge_rb(rb_lst rbl, pic_def pic) 
{
  int  r, z, s_start, s_end, z_end, slen, w = pic.ar.width;
  PICT *dst;

  set_bitmap(AREA_SIZE(pic.ar), pic.mem, BACKGROUND);

  for (r = 0; r < rbl.cnt; r++) {
    s_start = rbl.rb[r].sc.column.start;
    s_end   = rbl.rb[r].sc.column.end;
    z_end   = rbl.rb[r].sc.line.end;

    for (z = rbl.rb[r].sc.line.start; z <= z_end; z++) {
      dst = &pic.mem[s_start+z*w];
      slen = s_end-s_start+1;

      while (slen--) {
	*dst++ = FOREGROUND;
      }
    }
  }
}



/* The RB list is sorted according to area size.

Revision: 20.02.2003 E.Sack

Input : - pointer to RB list
Output: - RB list updated
*/
void 
rb_sort(rb_lst *rbl) {
  int    i, k, j, max, z, *idx = 0, *siz = 0;
  rb_lst zrb;

  if (rbl->cnt <= 1) {
    return;                                           /* no sorting needed */
  }

  CALLOC(int, rbl->cnt, idx);
  CALLOC(int, rbl->cnt, siz);

  init_rb_lst(&zrb);
  zrb.cnt = rbl->cnt;
  CALLOC(rb_def, zrb.cnt, zrb.rb);

  for (i = 0; i < rbl->cnt; i++) {              /* prepare auxiliary lists */
    idx[i] = i;                                            /* saving index */
    siz[i] = SECTOR_SIZE(rbl->rb[i].sc);               /* calculating size */

    init_rb_def(&zrb.rb[i]);
    copy_rb_def(&rbl->rb[i], &zrb.rb[i]);                       /* copy RB */
  }

  for (j = 0; j < rbl->cnt-1; j++) {             /* sorting regarding size */
    z   = idx[j];
    max = siz[j];
    k = j;

    for (i = j+1; i < rbl->cnt; i++) {
      if (max < siz[i]) {
	z   = idx[i];
	max = siz[i];
	k = i;
      }
    }
   
    if (k == j) {
      continue;
    }

    idx[k] = idx[j];                                /* simple sorting algo */
    siz[k] = siz[j];                                           /* exchange */
    idx[j] = z;
    siz[j] = max;
  }

  for (j = 0; j < rbl->cnt; j++) {                /* copy back sorted list */
    copy_rb_def(&zrb.rb[idx[j]], &rbl->rb[j]);
  }

  FREEMEM(idx);
  FREEMEM(siz);
  free_rb_lst(&zrb);
}



/* Seeks all RB in an image.

Revision: 25.12.2002 E.Sack

Input : - image
Output: - RB list
Postcondition: - image cleared 
*/
static rb_lst
rb_seek(pic_def pic) 
{
  int        d, oflag = 1;
  sector_def clip;
  c_def      point;
  rb_lst     rbl;

  init_rb_lst(&rbl);

  AREA_TO_SECTOR(pic.ar, clip);

  while (1) {
    d = xscan_bmp(clip, &point, pic, FOREGROUND, &oflag, 0);

    if (!d) {
      break;
    }

    if (rbl.cnt) {
      REALLOC(rb_def, rbl.cnt+1, rbl.rb);
    }
    else {
      CALLOC(rb_def, rbl.cnt+1, rbl.rb);
    }

    init_rb_def(&rbl.rb[rbl.cnt]);

    rbl.rb[rbl.cnt].sc.column.start = point.x;
    rbl.rb[rbl.cnt].sc.column.end   = point.x;
    rbl.rb[rbl.cnt].sc.line.start  = point.y;
    rbl.rb[rbl.cnt].sc.line.end    = point.y;

    create_rb(&rbl.rb[rbl.cnt], pic);                          /* build RB */
 
    rbl.cnt++;                                             /* seek next RB */
  }

  rbl.rar = pic.ar;                                        /* general info */

  return rbl;
}



/* Builds a RB and clears it from image.

Revision: 25.12.2002 E.Sack

Input : - pointer to RB
        - image
Output: - RB updated
        - image updated
*/
void 
create_rb(rb_def *rb, pic_def bmp_bild) 
{
  int   d, y, yy, x;
  c_def point;

  while (1) {                                                  /* build RB */
    d = pixel_on_frame(bmp_bild, rb->sc, &point);

    if (!d) {
      break;                                                /* RB finished */
    }
      
    rb->sc.column.end   = XPMAX(point.x, rb->sc.column.end);  /* fit frame */
    rb->sc.line.end     = XPMAX(point.y, rb->sc.line.end);
    rb->sc.column.start = XPMIN(point.x, rb->sc.column.start);
    rb->sc.line.start   = XPMIN(point.y, rb->sc.line.start);
  }

  for (y = rb->sc.line.start; y <= rb->sc.line.end; y++) {     /* clear RB */
    yy = y*bmp_bild.ar.width;

    for (x = rb->sc.column.start; x <= rb->sc.column.end; x++) {
      bmp_bild.mem[x + yy] = BACKGROUND;            
    }
  }
}



/* Seeks a foreground pixel on a frame clockwise.

Revision: 25.12.2002 E.Sack

Input : - pointer to image
        - frame
        - pointer to point
Output: - point updated
        - flag (0: no pixel found, 1: pixel found)
Postcondition: - the top side of the frame isn't tested assuming the start 
                 point is on the right frame side.
*/
static int
pixel_on_frame(pic_def pic, sector_def frame, c_def *point) 
{
  int col, row, end, start;

  if (frame.column.end != pic.ar.width-1) {       /* test right frame side */
    end = frame.column.end + 1;
   
    for (row = frame.line.start; row <= frame.line.end; row++)
      if (pic.mem[end + row*pic.ar.width] == FOREGROUND) {  /*top to bottom*/
	point->y = row;
	point->x = end;
	return 1;                                           /* pixel found */
      }
  }
  else {
    end = frame.column.end;
  }
   
  if (frame.column.start != 0) {                   /* test left frame side */
    start = frame.column.start-1;
  }
  else {
    start = frame.column.start;
  }

  if (frame.line.end != pic.ar.height-1) {
    row = frame.line.end + 1;                    /* test bottom frame side */

    for (col = end; col >= start; col--) {                /* right to left */
      if (pic.mem[col + row*pic.ar.width] == FOREGROUND) {
	point->y = row;
	point->x = col;
	return 1;                                           /* pixel found */
      }
    }
  }
   
  if (frame.column.start != 0) {                   /* test left frame side */
    start = frame.column.start-1;
   
    for (row = frame.line.end; row >= frame.line.start; row--) {
      if (pic.mem[start + row*pic.ar.width] == FOREGROUND) {/*bottom to top*/
	point->y = row;
	point->x = start;
	return 1;                                           /* pixel found */
      }
    }
  }

  return 0;                                     /* no pixel on frame found */
}



/* Creates a preference setting box.

Revision: 10.11.2004 E.Sack

Input : - pointer to container
        - pointer to key string
        - pointer to remark string
        - pointer to settings
Output: - container updated
*/
void
stats_window_setting(GtkWidget *cbox, char *key, char *rem,
		     stats_prefs *sentr)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 0);

  setting_key(hbox, key, rem, 0);

  stats_size_setting(hbox, sentr);
}



/* Creates a preference setting box.

Revision: 21.12.2004 E.Sack

Input : - pointer to container
        - pointer to key string
        - pointer to remark string
        - pointer to settings
Output: - container updated
*/
void
stats_icon_contrast_setting(GtkWidget *cbox, char *key, char *rem,
			    stats_prefs *sentr)
{
  GtkWidget *hbox, *gbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 0);

  setting_key(hbox, key, rem, 0);

  gbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), gbox, FALSE, FALSE, 2);

  /* set check button */
  sentr->display_button = gtk_check_button_new_with_label("maximise");
  gtk_box_pack_start(GTK_BOX(gbox), sentr->display_button, FALSE, FALSE, 10);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sentr->display_button),
			       sentr->sapply.display_active);
}



/* Creates a preference setting box.

Revision: 12.11.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
stats_size_setting(GtkWidget *cbox, stats_prefs *sentr)
{
  int       len;
  char      *buf1 = NULL;
  GtkWidget *pbox, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_end(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  /* set check button */
  sentr->display_button = gtk_check_button_new_with_label("display");
  gtk_box_pack_start(GTK_BOX(hbox), sentr->display_button, FALSE, FALSE, 10);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sentr->display_button),
			       sentr->sapply.display_active);

  setting_key(hbox, "", STATS_WIN_REM, 0);                /* no key needed */
  setting_integer_limits(hbox, STATS_WIN_DEFAULT, STATS_WIN_MIN, STATS_WIN_MAX);

  int_to_string("", STATS_WIN_MAX, "", &buf1);
  len = strlen(buf1);
  FREEMEM(buf1);

  /* height */
  int_to_string("", sentr->sapply.wsize.height, "", &buf1);/* current setting */

  pbox = pentry(&sentr->wentry_y, TRUE, 
		GTK_SIGNAL_FUNC(cb_stats_check), buf1, len, 
		XPLAB_ENTRY_LEN_INT3,
		GTK_SIGNAL_FUNC(cb_stats_revert), "R", 
		GTK_SIGNAL_FUNC(cb_stats_default), "D", 
		NULL, NULL, (gpointer)sentr);

  gtk_box_pack_end(GTK_BOX(hbox), pbox, FALSE, FALSE, 2);

  FREEMEM(buf1);

  check_entry_limits_integer(sentr->wentry_y, STATS_WIN_MIN, STATS_WIN_MAX, 
			     STATS_WIN_DEFAULT);

  /* widht */
  int_to_string("", sentr->sapply.wsize.width, "", &buf1);/* current setting */

  pbox = pentry(&sentr->wentry_x, TRUE, 
		GTK_SIGNAL_FUNC(cb_stats_check), buf1, len, 
		XPLAB_ENTRY_LEN_INT3,
		NULL, NULL, NULL, NULL, NULL, NULL, (gpointer)sentr);

  gtk_box_pack_end(GTK_BOX(hbox), pbox, FALSE, FALSE, 2);

  check_entry_limits_integer(sentr->wentry_x, STATS_WIN_MIN, STATS_WIN_MAX, 
			     STATS_WIN_DEFAULT);

  FREEMEM(buf1);
}



/* Reverts a setting.

Revision: 12.11.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_stats_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  stats_prefs *sentr;

  assert(user_data);
  assert(button);

  sentr = (stats_prefs *)user_data;

  sentr->sapply.wsize = sentr->srevert.wsize;          /* only window size */

  update_statistic_entry(*sentr);
}



/* Copies the old statistic settings into the current working set.

Revision: 12.11.2004 E.Sack

Input : - pointer to statistic settings
Output: - statistic settings updated from old
Usage: - nothing else than a copy.
*/
void
stats_revert_all(stats_prefs *sentr) 
{
  sentr->sapply = sentr->srevert;
}



/* Sets a setting to default.

Revision: 12.11.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_stats_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  stats_prefs *sentr;
  gchar       *buf1 = NULL;

  assert(user_data);
  assert(button);

  sentr = (stats_prefs *)user_data;

  int_to_string("", STATS_WIN_DEFAULT, "", &buf1);
  xplab_entry_set_text(sentr->wentry_x, buf1);
  xplab_entry_set_text(sentr->wentry_y, buf1);
  FREEMEM(buf1);
}



/* Updates statistic page.

Revision: 12.11.2004 E.Sack

Input : - statistic setting
Output: - statistic entry updated from setting
Usage: - Sets entry to the current working settings.
*/
void
update_statistic_entry(stats_prefs stenty)
{
  gchar *buf1 = NULL;

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(stenty.display_button),
			       stenty.sapply.display_active!=STATS_WIN_DEFACTV);

  int_to_string("", stenty.sapply.wsize.width, "", &buf1);
  xplab_entry_set_text(stenty.wentry_x, buf1);
  FREEMEM(buf1);

  int_to_string("", stenty.sapply.wsize.height, "", &buf1);
  xplab_entry_set_text(stenty.wentry_y, buf1);
  FREEMEM(buf1);
}



/* Applies statistic page.

Revision: 12.11.2004 E.Sack

Input : - statistic setting
Output: - statistic setting updated from entry
*/
void
apply_statistic_entry(stats_prefs *sentr)
{
  sentr->sapply.wsize.width = check_entry_limits_integer(sentr->wentry_x,
							 STATS_WIN_MIN,
							 STATS_WIN_MAX, 
							 STATS_WIN_DEFAULT);

  sentr->sapply.wsize.height = check_entry_limits_integer(sentr->wentry_y,
							  STATS_WIN_MIN,
							  STATS_WIN_MAX, 
							  STATS_WIN_DEFAULT);

  sentr->sapply.display_active =                  /* widget state conversion */
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sentr->display_button));
}



/* Checks setting entry.

Revision: 12.11.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_stats_check(GTK_CONVENTION_ENTRY_ACTIVATE_RTS_VOID)
{
  stats_prefs *sentr;

  STSFY_GCC(editable);

  assert(user_data);

  sentr = (stats_prefs *)user_data;

  check_entry_limits_integer(sentr->wentry_x, STATS_WIN_MIN, STATS_WIN_MAX, 
			     STATS_WIN_DEFAULT);

  check_entry_limits_integer(sentr->wentry_y, STATS_WIN_MIN, STATS_WIN_MAX, 
			     STATS_WIN_DEFAULT);
}



/* Creates a preference label.

Revision: 22.10.2004 E.Sack

Input : - pointer to horizontal box
        - pointer to key string
        - pointer to remark string
        - flag (0: pack to begin of container, 1: pack to end of container)
Output: - container updated
*/
void
setting_key(GtkWidget *hbox, char *key, char *rem, int end)
{
  int       err;
  char      *buf = NULL;
  GtkWidget *label;

  add_string(key, &buf);

  err = check_name(rem);

  if (!err) {
    cat_string(" (", rem, ")", &buf);
  }

  add_string(":", &buf);

  label = gtk_label_new(buf);

  FREEMEM(buf);

  if (end) {
    gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 2);
  }
  else {
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);
  }
}



/* Creates a preference label for integer limits.

Revision: 05.10.2003 E.Sack

Input : - pointer to container
        - default integer
        - minimum integer
        - maximum integer
Output: - container updated
*/
void
setting_integer_limits(GtkWidget *cbox, int def, int min, int max)
{
  char      *buf = NULL;
  GtkWidget *label;

  int_to_string("", min, "...", &buf);
  int_to_string("", max, " [", &buf);
  int_to_string("", def, "]", &buf);

  label = gtk_label_new(buf);
  gtk_box_pack_start(GTK_BOX(cbox), label, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a general text entry widget.

Revision: 12.11.2004 E.Sack

Input : - pointer to text entry widget pointer
        - flag (0: text entry is not editable, 1: text entry is editable)
        - text entry handler
        - pointer to default text entry text string
        - text entry length
        - button1 handler
        - button2 handler
        - button3 handler
        - pointer to handler data
Output: - pointer to container of text entry widget
        - text entry widget pointer updated
Precondition: - handlers may be NULL
*/
GtkWidget*
pentry(GtkEntry **entry, gboolean entry_is_editable, GtkSignalFunc entry_cb,
       gchar *default_entry_txt, guint16 entry_max_length, guint box_len,
       GtkSignalFunc button1_cb, gchar *button1_label,
       GtkSignalFunc button2_cb, gchar *button2_label,
       GtkSignalFunc button3_cb, gchar *button3_label, gpointer gp)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);

  if (entry) {
    assert(entry_max_length > 0);
    assert(box_len > 5);

    *entry = GTK_ENTRY(gtk_entry_new());
    gtk_editable_set_editable(GTK_EDITABLE(*entry), entry_is_editable);
    gtk_entry_set_max_length(GTK_ENTRY(*entry), entry_max_length);
    gtk_entry_set_width_chars(GTK_ENTRY(*entry), entry_max_length);
    gtk_entry_set_alignment(GTK_ENTRY(*entry), 0.5);

    if (default_entry_txt != NULL) {
      xplab_entry_set_text(*entry, default_entry_txt);
    }

    gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(*entry), FALSE, FALSE, 0);

    if (entry_cb) {
      g_signal_connect(G_OBJECT(*entry), "activate", G_CALLBACK(entry_cb), gp);
    }
  }

  if (button1_label != NULL) {
    assert(button1_cb);
    xset_clicked_button(hbox, button1_cb, button1_label, NULL, NULL, gp);
  }

  if (button2_label != NULL) {
    assert(button2_cb);
    xset_clicked_button(hbox, button2_cb, button2_label, NULL, NULL, gp);
  }

  if (button3_label != NULL) {
    assert(button3_cb);
    xset_clicked_button(hbox, button3_cb, button3_label, NULL, NULL, gp);
  }

  return hbox;
}



/* Checks setting entry.

Revision: 12.11.2004 E.Sack

Input : - pointer to entry widget
        - minimum
        - maximum
        - default
Output: - entry updated
        - entry value
*/
int
check_entry_limits_integer(GtkEntry *entry, int min, int max, int def)
{
  int   d, entry_value;
  gchar *bufc = NULL, *buf1 = NULL;

  bufc = xplab_entry_get_text(entry);

  d = check_name(bufc);

  if (!d) {                                            /* check user input */
    d = sscanf(bufc, "%d", &entry_value);

    if (!d || d == EOF) {                           /* test string content */
      entry_value = def;                               /* no integer found */
    }
    else if (entry_value > max) {
      entry_value = max;                           /* limit to upper bound */
    }
    else if (entry_value < min) {
      entry_value = min;                           /* limit to lower bound */
    }
  }
  else {
    entry_value = def;                              /* no entry text found */
  }

  int_to_string("", entry_value, "", &buf1);
  xplab_entry_set_text(entry, buf1);
  FREEMEM(bufc);
  FREEMEM(buf1);

  return entry_value;
}
