/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <unistd.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_replay.h"
#include "xplab_preferences_fe.h"
#include "xplab_preferences_be.h"
#include "xplab_dbio.h"
#include "xplab_report.h"
#include "xplab_recall_be.h"
#include "xplab_maintenance_be.h"
#include "xplab_selection.h"
#include "xplab_batch.h"
#include "xplab_dchannel.h"

/* local definitions *******************************************************/
#define XPLAB_REPLAY_TITLE_ITEM "replay"
#define XPLAB_REPLAY_WARNING "WARNING: no replay file selected"
#define XPLAB_REPLAY_WINDOW_WIDTH  511
#define XPLAB_REPLAY_WINDOW_HEIGHT 171

typedef struct {
  GtkWidget *top_replay, *menu_bar, *rlog;
  GtkWidget *log_dump, *log_clear, **caller;
  gchar     *replay_log_filename;
  gboolean  cactive, cstop;
  run_def   crun;
} rplay_def;

rplay_def ry;                                             /* replay dialog */

/* local prototypes ********************************************************/
static void init_rplay_def(rplay_def *rd);
static void free_rplay_def(rplay_def *rd);

static void replay_setup(char *replay);
static int check_replay_resources(char *replay_filename);
static int replay_recognition(rb_lst rblst, char *db_name, char **ident, 
			      char *image_filename, cfg_def cfg, 
			      cand_lst hit_lst);

/* Functions assembling widgets */
static void replay_window(void);
static void replay_menu_bar(GtkWidget *vbox);
static void replay_menu_entry_file(GtkWidget *menu_bar);
static void replay_file_menu_item_quit(GtkWidget *file_menu);
static void replay_control(GtkWidget *vbox);

/* Handlers/Callbacks */
static void cb_replay_stop(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_replay_start(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_replay_dump(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_replay_clear(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_check_replay_log(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_replay_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID);

/* Functions assembling handlers/callbacks */

/* Functions ***************************************************************/

static void
init_rplay_def(rplay_def *rd)
{
  assert(rd);

  rd->top_replay = NULL;                                     /* zero pointer */
  rd->menu_bar = NULL;                                       /* zero pointer */
  rd->rlog = NULL;                                           /* zero pointer */
  rd->caller = NULL;                                         /* zero pointer */
  rd->log_dump = NULL;                                       /* zero pointer */
  rd->log_clear = NULL;                                      /* zero pointer */

  rd->replay_log_filename = NULL;                            /* zero pointer */

  rd->cactive = rd->cstop = FALSE;

  init_run_def(&rd->crun);
}

static void
free_rplay_def(rplay_def *rd)
{
  assert(rd);
                           /* don't free GTK stuff */

  FREEMEM(rd->replay_log_filename);

  init_rplay_def(rd);
}



/* Initializes the replay.

Revision: 08.11.2003 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_init_replay(void)
{
  init_rplay_def(&ry);

  ry.crun.pactive = &ry.cactive;
  ry.crun.pstop = &ry.cstop;
}



/* Frees the replay.

Revision: 08.11.2003 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_free_replay(void)
{
  free_rplay_def(&ry);

  xplab_init_replay();
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - menu
Output: - menu item
*/
void
xplab_replay_menu(GtkWidget *match_menu, GtkWidget **caller)
{
  GtkWidget *match_menu_replay;

  match_menu_replay = gtk_menu_item_new_with_label(XPLAB_REPLAY_TITLE_ITEM XPLAB_MENU_EXT);
  gtk_menu_shell_append(GTK_MENU_SHELL(match_menu), match_menu_replay);

  g_signal_connect(match_menu_replay, "activate", 
		   G_CALLBACK(cb_replay_dialog), (gpointer)caller);
}



/* Starts replay dialog.

Revision: 08.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_replay_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gboolean raised;

  assert(user_data);
  assert(menuitem);

  raised = xplab_window_raise(&ry.top_replay, 1);

  if (raised == FALSE) {
    ry.caller = (GtkWidget **)user_data;              /* last succeeded call */
    replay_window();                                           /* top window */
  }
}



/* Creates top window for replay.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - top window
*/
static void
replay_window(void)
{
  char      *buf = NULL;
  GtkWidget *vbox;

  FREEMEM(ry.replay_log_filename);
  add_string(XPLAB_REPLAY_LOGNAME_DEFAULT, &ry.replay_log_filename);

  /* set the replay top level widget */
  ry.top_replay = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(ry.top_replay), XPLAB_REPLAY_TITLE_ITEM);
  gtk_window_set_default_size(GTK_WINDOW(ry.top_replay), 
			      XPLAB_REPLAY_WINDOW_WIDTH, 
			      XPLAB_REPLAY_WINDOW_HEIGHT);

 /* put into top level widget a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(ry.top_replay), vbox);

  /* put into vertical box a menubar */
  replay_menu_bar(vbox);

  /* put into vertical box a scrolled window */
  ry.rlog = scrolled_text(vbox);

  /* put into vertical box the control format */
  replay_control(vbox);

  gtk_widget_show_all(ry.top_replay);

  g_signal_connect(G_OBJECT(ry.top_replay), "delete-event", 
		   G_CALLBACK(cb_replay_top_delete), NULL);

  xplab_set_window_icon(ry.top_replay);                       /* window icon */

  if (replay_names.idx >= 0) {
    add_string(replay_names.td.tl[replay_names.idx], &buf);
  }
  else {
    add_string(XPLAB_REPLAY_WARNING, &buf);
  }

  xplab_label_set_text(ry.crun.plabel, buf);

  FREEMEM(buf);
}



/* Deletes top window.

Revision: 08.11.2003 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
gboolean
cb_replay_top_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);

  assert(event);
  assert(widget);

  cb_replay_quit(NULL, NULL);

  return FALSE;
}



/* Builds a menu bar.

Revision: 01.10.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
replay_menu_bar(GtkWidget *vbox)
{
  ry.menu_bar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox), ry.menu_bar, FALSE, FALSE, 0);

  replay_menu_entry_file(ry.menu_bar);                          /* menu file */
  xplab_menu_help(ry.menu_bar, &ry.top_replay);                 /* menu help */
}



/* Builds an entry for a menu bar.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu bar
Output: - menu bar updated
*/
static void
replay_menu_entry_file(GtkWidget *menu_bar)
{
  GtkWidget *file_entry, *file_menu, *sep;

  file_entry = gtk_menu_item_new_with_label(MENU_BUTTON_FILE);

  file_menu = gtk_menu_new();

  xplab_preferences_menu(file_menu, &ry.top_replay);   /* item preferences */
  xplab_selection_menu(file_menu, &ry.top_replay);       /* item selection */
  file_menu_item_log(file_menu,                                /* item log */
		     GTK_SIGNAL_FUNC(cb_replay_dump), &ry.log_dump,
		     GTK_SIGNAL_FUNC(cb_replay_clear), &ry.log_clear, NULL);

  sep = gtk_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), sep);

  replay_file_menu_item_quit(file_menu);                      /* item quit */

  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_entry), file_menu);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_entry);

  g_signal_connect(G_OBJECT(file_entry), "activate", 
		   G_CALLBACK(cb_check_replay_log), NULL);
}



/* Checks log menu entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_check_replay_log(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  int           len;
  GtkTextBuffer *buffer;

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  if (GTK_IS_TEXT_VIEW(ry.rlog) == TRUE) {
    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(ry.rlog));/* fetch default buffer */
    len = gtk_text_buffer_get_char_count(buffer);

    gtk_widget_set_sensitive(ry.log_clear, len ? TRUE : FALSE);
    gtk_widget_set_sensitive(ry.log_dump, len ? TRUE : FALSE);
  }
}



/* Clears recall log.

Revision: 02.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_replay_clear(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  clear_gtk_text(ry.rlog);
}



/* Dumps replay log.

Revision: 12.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_replay_dump(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  char *txt_file = NULL;

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  cat_string(ry.replay_log_filename, XPLAB_REPLAY_LOGNAME_EXT, "", &txt_file);

  dump_gtk_text(ry.rlog, txt_file);

  FREEMEM(txt_file);
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
replay_file_menu_item_quit(GtkWidget *file_menu)
{
  GtkWidget *file_menu_quit;

  file_menu_quit = gtk_menu_item_new_with_label(DIALOG_BUTTON_QUIT);

  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_quit);

  g_signal_connect(G_OBJECT(file_menu_quit), "activate", 
		   G_CALLBACK(cb_replay_quit), NULL);
}



/* Quits GTK main loop.

Revision: 08.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_replay_quit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  cb_replay_stop(NULL, NULL);                             /* sudden run stop */

  if (GTK_IS_WIDGET(ry.top_replay) == TRUE) {
    xplab_window_raise(ry.caller, xctrl.cfg.raise);
    xplab_widget_destroy(&ry.top_replay);

    /* cannot free data here because of sudden run stop                      */

  }
}



/* Stops replay.

Revision: 08.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_replay_stop(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  stop_run_def(&ry.crun);
}



/* Builds a progress bar.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
replay_control(GtkWidget *vbox)
{
  GtkWidget *hbox;

  /* put into inbox a horizontal box */
  ry.crun.plabel = gtk_label_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), ry.crun.plabel, FALSE, FALSE, 0);

  /* put into inbox a horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  /* put into horizontal box a start button */
  xset_clicked_button(hbox, GTK_SIGNAL_FUNC(cb_replay_start), 
		      SINGLE_BUTTON_START, 
		      DESC1_REPLAY_START, DESC2_REPLAY_START, NULL);

  /* put into horizontal box a stop button */
  xset_clicked_button(hbox, GTK_SIGNAL_FUNC(cb_replay_stop), 
		      SINGLE_BUTTON_STOP, 
		      DESC1_STOP_ACTION, DESC2_STOP_ACTION, NULL);

  /* put into horizontal box a progress bar */
  ry.crun.pbar = xplab_progress(hbox);
}



/* Starts replay.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_replay_start(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(button);

  if (*ry.crun.pactive == TRUE) {
    return;
  }

  start_run_def(&ry.crun);

  if (replay_names.idx >= 0) {                 /* latest name from selection */
    replay_setup(replay_names.td.tl[replay_names.idx]);
  }
  else {
    xplab_label_set_text(ry.crun.plabel, XPLAB_REPLAY_WARNING);
  }

  stop_run_def(&ry.crun);
}



/* Forces to fetch latest replay file name from selection.

Revision: 01.10.2005 E.Sack

Input : - -
Output: - log updated
*/
void
xplab_input_replay_file(void)
{
  if (GTK_IS_WIDGET(ry.top_replay) == TRUE) {
    if (GTK_IS_WIDGET(ry.crun.plabel) == TRUE) {
      xplab_label_set_text(ry.crun.plabel, 
			   replay_names.td.tl[replay_names.idx]);
    }
  }

  string_to_log(main_log, "selected replay file '", 
		replay_names.td.tl[replay_names.idx], "' \n");
}



/* Automatic replay of recognition phase results stored in a file.

Revision: 05.08.2005 E.Sack

Input : - programm control
Output: - Flag (0: no error, 1: error)
Precondition: - the replay procedure is ment to be used for debugging 
                purposes mainly. To keep consistence the database, 
                preferences and image should be identic as been as for 
                the initial run.
*/
static void
replay_setup(char *replay_name)
{
  int       err;
  char      *db_filename = NULL;
  cand_lst  hit_lst;
  prog_ctrl ctrl;

  init_prog_ctrl(&ctrl);
  init_cand_lst(&hit_lst);

  err = check_replay_resources(replay_name);

  if (err) {
    return;
  }

  xplab_dc_exit();                                          /* reset old DC */
  xplab_dc_init();

  err = xplab_read_recognition(replay_name, &ctrl.set.rblst, &db_filename, 
			       &ctrl.cfg.tps_id, &ctrl.set.image_filename, 
			       &ctrl.cfg.mad.mborder, &ctrl.cfg.mad.nborder, 
			       &ctrl.cfg.mad.maskweight, 
                               &ctrl.cfg.histo_thold, 
			       &hit_lst, &ctrl.set.user_options, 1);
  if(err) {
    string_to_log(main_log, "WARNING: can't read replay file '", replay_name, "' \n");
  }
  else {
    char *buf = NULL;

    FREEMEM(ry.replay_log_filename);           /* reinit filename for output */
    add_string(ctrl.set.image_filename, &ry.replay_log_filename);

    xreport("\n*** START REPLAY PHASE ***\n");

    cat_string("replay '", replay_name, "'", &buf);
    xplab_label_set_text(ry.crun.plabel, buf);
    FREEMEM(buf);

    err = check_name(ctrl.set.user_options);
    if (!err) {
      add_string(ctrl.set.user_options, &buf);
    }
    else {
      add_string("- none -", &buf);
    }

    xreport("replay options: %s\n", buf);
    FREEMEM(buf);

    if(xctrl.cfg.report_write & REPORT_WRITE_RBS) {
      buf = print_report_rbs(ctrl.set.rblst);
      xreport(buf);
      FREEMEM(buf);
    }

    err = replay_recognition(ctrl.set.rblst, db_filename, &ctrl.cfg.tps_id, 
			     ctrl.set.image_filename, ctrl.cfg, hit_lst);
    if (err) {
      cat_string("could not replay '", replay_name, "'", &buf);
    }
    else {
      cat_string("'", replay_name, "' replayed", &buf);
    }

    xplab_label_set_text(ry.crun.plabel, buf);
    FREEMEM(buf);

    time_to_log(main_log, XPLAB_REPLAY_TITLE_ITEM " finished after ", 
		ry.crun.start_time, "\n");

    xreport("*** END REPLAY PHASE ***\n\n");
  }

  FREEMEM(db_filename);
  free_cand_lst(&hit_lst);
  free_prog_ctrl(&ctrl);
}



/* Wrapper: Checks the replay resources.

Revision: 08.11.2003 E.Sack

Input : - pointer to replay file name string
        - image
        - RB list
Output: - flag (0: no error, 1: error)
        - message output
*/
static int
check_replay_resources(char *replay_filename)
{
  int err;

  err = check_name(replay_filename);

  if (err) {
    xplab_text_insert(main_log, 
		      "WARNING: replay file name undefined\n");
    return 1;
  }

  return 0;
}



/* Replay recognition phase with dumped phase results.

Revision: 01.11.2004 E.Sack

Input : - RB list
        - pointer to RB image
        - pointer to DB filename
        - pointer to TPS ID
        - pointer to input image filename
        - mask border
        - maximal allowed lag
        - forced inline method
        - hitlist
Output: - recognition text
        - flag (0: replayed, 1: not replayed)
*/
static int
replay_recognition(rb_lst rblst, char *db_name, char **ident, 
		   char *image_filename, cfg_def cfg, cand_lst hit_lst)
{
  int     nbl_nr, err = 1, serious_error;
  db_def  db;
  pic_def rpic;                                            /* binary image */

  init_pic_def(&rpic);

  db = xplab_read_db(db_name, 1, 1, &serious_error);

  if (!serious_error) {
    err = check_name(db.db_fname);

    if (!err) {                                      /* database available */
      nbl_nr = search_tps_id(ident, db.tpl);

      if (nbl_nr >= 0) {                       /* ID in database available */
	gen_class_references(db.tpl.tps[nbl_nr].cl, cfg.mad.mborder, 
			     cfg.mad.maskweight);

	rpic =                      /* as last as possible read image file */
	  setup_input_image(db, image_filename, cfg, NULL, NULL, NULL, NULL, 
			    NULL).pimg;

	xrecognition_out(rblst, hit_lst, db.tpl.tps[nbl_nr], cfg.mad.mborder, 
			 cfg.mad.nborder, rpic, image_filename, 
			 cfg.line_method, &ry.crun, ry.rlog, 
			 cfg.renew_method);
	err = 0;
      }
      else {
	string_to_log(main_log, "WARNING: replayed TPS ID '", *ident, 
		      "' not found\n");
      }
    }
  }

  free_db_def(&db);
  free_pic_def(&rpic);

  return err;
}



