/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_algorithm.h"
#include "xplab_report.h"
#include "xplab_image.h"
#include "xplab_maintenance_fe.h"
#include "xplab_preferences_fe.h"
#include "xplab_dbio.h"

/* Functions **************************************************************/

/* Investigates a candidate list.

Revision: 20.01.2006 E.Sack

Input : - RB list
        - class list
        - candidate list
        - flag (0: all infos, 1: terse infos)
Output: - output to screen
        - mini statistic
*/
int_lst
check_quality(rb_lst rblst, cls_lst cl, cand_lst hlist, int terse) 
{
  int     d, i, j, bad_nb, good_nb, good_rb, used_cls, invalid;
  int     *r_nr = NULL;                               /* list for checked RB */
  int     *c_nr = NULL;                                  /* list for classes */
  char    *buf = NULL, *tbuf1 = NULL, *tbuf2 = NULL;
  int_lst mini_statistic;                              /* #good/#bad objects */

  init_int_lst(&mini_statistic);

  if (!hlist.cnt) {
    return mini_statistic;
  }

  CALLOC(int, 2, mini_statistic.nr);
  CALLOC(int, rblst.cnt, r_nr);
  CALLOC(int, cl.cnt, c_nr);

  for (i = 0, bad_nb = 0, good_nb = 0; i < hlist.cnt; i++) {
    if (hlist.c[i].n >= 0) {
      r_nr[hlist.c[i].rbn] = 1;      /* incrementable for further statistics */
      c_nr[hlist.c[i].c]++;                                  /* only classes */
      good_nb++;

      for (j = 0; j < hlist.c[i].crb.cnt; j++) {
	r_nr[hlist.c[i].crb.nr[j]] = 1;
      }
    }
    else {
      bad_nb++;                         /* marks unrecognzed NB or RB, resp. */
    }
  }

  for (i = good_rb = invalid = 0; i < rblst.cnt; i++) {
    if (rblst.rb[i].rld == INVALID) {
      invalid++;
    }

    if (r_nr[i]) {
      good_rb++;
    }
  }

  tbuf1 = xplab_vsnprintf("Number of (partial) recognized RB: %d/%d (%f%%)\n",
			  good_rb, rblst.cnt, 
			  100*(float)good_rb/(float)rblst.cnt);
  add_string(tbuf1, &tbuf2);
  FREEMEM(tbuf1);

  int_to_string("Number of invalid RB: ", invalid, "\n", &tbuf2);

  for (i = 0, used_cls = 0; i < cl.cnt; i++) {
    if (c_nr[i]) used_cls++;
  }

  int_to_string("Number of used classes: ", used_cls, "\n", &tbuf2);
  int_to_string("Number of used NB: ", good_nb, "\n", &tbuf2);
  int_to_string("Number of unrecognized objects: ", bad_nb, "\n", &tbuf2);

  if (used_cls && !terse) {                            /* summary of classes */
    add_string("\nSummary of classes:\n<number of classes> x <class ",&tbuf2);
    add_string("name> (<translation>) [alignment]\n",&tbuf2);

    for (i = 0; i < cl.cnt; i++) {
      if (c_nr[i]) {
	int_to_string("", c_nr[i], " x ", &tbuf2);
	add_string(cl.cls[i].att.class, &tbuf2);

	cat_string(" (", cl.cls[i].att.sign, ")", &tbuf2);

	add_string(" [", &tbuf2);

	d = aligmnt_to_string(cl.cls[i].att.alignment, &buf);
	assert(d);

	cat_string(buf, "]\n", "", &tbuf2);
	buf[0] = '\0';                                /* reset string buffer */
      }
    }
  }

  xreport("%s", tbuf2);                          /* finally write whole text */

  assert(tbuf1 == NULL);
  FREEMEM(tbuf2);

  FREEMEM(c_nr);
  FREEMEM(r_nr);
  FREEMEM(buf);

  mini_statistic.cnt = 2;
  mini_statistic.nr[0] = good_nb;
  mini_statistic.nr[1] = bad_nb;

  return mini_statistic;
}



/* Dumps the decisive factors (separation rules) of word separations within
   a line.

Revision: 05.08.2005 E.Sack

Input:  - distance list
        - list of candidates index of the line
        - candidates list
        - pointer to class list
        - line number
        - pointer to image filename
        - documentation flag
Output: - separation rules dumped to file
*/
void
write_sep_info(int_lst dlst, int_lst znr, cand_lst hlist, cls_def *cls, 
	       int lcnt, char *filename, int docu)
{
  int  d, j, i, n, c, w;
  char nam1[MAXCHAR], buf[MAXCHAR], txt[MAXCHAR];
  FILE *file;

  if (!docu) {
    return;
  }

  g_snprintf(nam1, sizeof(nam1), INFNAMD_FORMAT(1), filename, lcnt);

  file = xplab_open_file(nam1, "w");
  if (!file) {
    return;
  }

  g_snprintf(txt, sizeof(txt), 
	   "# File: %s\n"                                  /* build remark */
	   "# Content: separation results\n"
	   "# Symbols: %d\n",
	     filename, dlst.cnt+1);

  d = fprintf(file, txt);
  assert(d >= 0);

  if (dlst.cnt == 0) {
    d = fprintf(file, "#\n# no list found\n#\n");
    assert(d >= 0);
  }
  else {
    assert(dlst.cnt == znr.cnt-1);

    for (i = 0, w = 0; i < dlst.cnt; i++) {
      if (dlst.nr[i] < 0) w++;
    }

    d = fprintf(file, "# Words: %d\n#\n", w+1);
    assert(d >= 0);

    d = fprintf(file, "Dist.\nNbr.   Rule       Separation\n");
    assert(d >= 0);
  }

  for (i = 0; i < dlst.cnt; i++) {
    switch (dlst.nr[i]) {
    case SEP_SR1: strncpy(txt, "SEP_SR1", MAXCHAR-1); break;
    case SEP_SR2: strncpy(txt, "SEP_SR2", MAXCHAR-1); break;
    case SEP_WMV: strncpy(txt, "SEP_WMV", MAXCHAR-1); break;
    default: continue; break;
    }

    txt[MAXCHAR-1] = 0;

    for (j = i, buf[0] = '\0'; j <= i+1; j++) {
      n = hlist.c[znr.nr[j]].n;
      c = hlist.c[znr.nr[j]].c;

      switch (n) {
      case SUBSTNB_FLAG: strcat(buf, xctrl.cfg.subst_nb); break;
      case SUBSTRB_FLAG: strcat(buf, xctrl.cfg.subst_rb); break;
      default:           strcat(buf, cls[c].att.sign); break;
      }

      if (j == i) {
	strcat(buf, " - ");
      }
    }

    d = fprintf(file, "%-5d  %s    (%-5s)\n", i, txt, buf);
    assert(d >= 0);
  }

  if (xctrl.cfg.verbose > BE_QUIET) {
    string_to_log(main_log, "separation results dumped to '", nam1, "'\n");
  }

  xplab_close_file(file, nam1);
}



/* Dumps line statistics.

Revision: 01.11.2003 E.Sack

Input:  - line distance list
        - line number
        - pointer to line string
        - pointer to image filename
        - documentation flag
Output: - distance histogram dumped to file
*/
void
histo_of_line(int_lst dlst, int line, char *ltxt, char *filename, int docu)
{
  int      i;
  double   mv = 0.0, sd = 0.0;
  hst_info hi;

  if (!docu || !dlst.cnt) {
    return;
  }

  init_hst_info(&hi);

  for (i = 0; i < dlst.cnt; i++) {
    assert(dlst.nr[i] >= 0);
  }

  mv = mean_value(dlst.nr, dlst.cnt, SKIP_ZERO, 0);   /* statistics of list */
  sd = standard_deviation(dlst.nr, dlst.cnt, mv, SKIP_ZERO, 0);

  hi = grab_infos(dlst, mv, sd);                              /* grab infos */

  write_hst_line(LINNAMH_DEFAULT, hi, line, ltxt, filename, 0, 1);

  free_hst_info(&hi);
}



/* Dumps line statistics.

Revision: 05.08.2005 E.Sack

Input:  - histogram statistics
        - line number
        - pointer to line string
        - pointer to image filename
        - documentation flag
Output: - file
*/
void 
write_hst_line(char *id, hst_info info, int line, char *patt, char *filename,
	       int iter, int docu) 
{
  char fnam[MAXCHAR], *txt = 0;

  if (!docu) {
    return;
  }

  g_snprintf(fnam, sizeof(fnam), LINNAMH_FORMAT(1), filename, id, line, iter);

  string_to_log(main_log, "# Input: ", filename, "\n");     /* build remark */
  add_string("# Content: histogram\n", &txt);
  int_to_log(main_log, "# Line: ", line, "");
  int_to_log(main_log, "  Iteration: ", iter, "\n");
  string_to_log(main_log, "# Pattern: ", patt, "\n");
  double_to_log("# MV = ", info.st.mv, "");
  double_to_log(", SD = ", info.st.sd, "");
  double_to_log(", MV+SD = ", info.st.mv+info.st.sd, "\n");

  write_histo(info.hd, 1, fnam, txt);

  FREEMEM(txt);
}



/* Histogramm einer Stichprobe berechnen und Erwartungswerte sammeln.

Input:  - Liste
        - Mittelwert
        - Standardabweichung
Output: - Infos auf Stand
*/
hst_info
grab_infos(int_lst old, double mv, double sd)
{
  hst_info hi;

  init_hst_info(&hi);
  hi.hd = histogram_auto(old.nr, old.cnt);

  hi.st.mv = mv;
  hi.st.sd = sd;

  return hi;
}



/* Assembles RB list to a string.

Revision: 05.08.2005 E.Sack

Input : - RB list
Output: - string
Postcondition: - output string to be freed.
*/
char *
print_report_rbs(rb_lst rblst)
{
  int  i, j; 
  char buf1[MAXCHAR], buf2[MAXCHAR], *buf3 = 0;

  if (!rblst.cnt) {
    add_string("\n*** no RB found ***\n", &buf3);
  }
  else {
    add_string("\nRB list:\n", &buf3);

    for (i = 0; i < rblst.cnt; i+=3) {                    /* three columns */
      for (j = i; j<rblst.cnt && j<i+3; j++) {
	g_snprintf(buf1, sizeof(buf1), "R[%d] %d-%d x %d-%d", j, 
		 rblst.rb[j].sc.column.start,
		 rblst.rb[j].sc.column.end,
		 rblst.rb[j].sc.line.start,
		   rblst.rb[j].sc.line.end);

	g_snprintf(buf2, sizeof(buf2), "%-25s", buf1);
	add_string(buf2, &buf3);
      }

      add_string("\n", &buf3);
    }
  }

  add_string("\n", &buf3);
  return buf3;
}



/* Dump of candidate hit list.

Revision: 05.08.2005 E.Sack

Input : - image size
        - maximal allowed lag
        - mask border size
        - hit list
        - pointer to class list
Output: - dump of RB list
*/
void
print_hitlist(area_def par, int noise_border, int mask_border, cand_lst hlist,
	      cls_def *cls)
{
  int  i;
  char *tbuf1 = NULL, *tbuf2 = NULL, buf1[MAXCHAR], buf2[MAXCHAR];

  tbuf1 = xplab_vsnprintf("\nHitlist (%d entr%s, p = %d, m = %d, document size %dx%d):\n",
			  hlist.cnt, hlist.cnt==1?"y":"ies", noise_border, 
			  mask_border, par.width, par.height);
  add_string(tbuf1, &tbuf2);
  FREEMEM(tbuf1);

  add_string("entry  method  RB     Cls/NB   Trs    Corr.     lag      "
	     "RB sector             kernel\n", &tbuf2);

  for (i = 0; i < hlist.cnt; i++) {
    tbuf1 = xplab_vsnprintf("%-5d    %-2d    %-5d ", i, hlist.c[i].mtd, 
                                                        hlist.c[i].rbn);
    add_string(tbuf1, &tbuf2);
    FREEMEM(tbuf1);

    if (hlist.c[i].n >= 0) {
      g_snprintf(buf1, sizeof(buf1), "%4d/%-4d %-5s", hlist.c[i].c, 
                                                    hlist.c[i].n, 
		 cls[hlist.c[i].c].att.sign);

      tbuf1 = xplab_vsnprintf("%-15s  %f  %-3d %-3d  ", buf1, hlist.c[i].m.cv, 
			      hlist.c[i].m.x, hlist.c[i].m.y);
      add_string(tbuf1, &tbuf2);
      FREEMEM(tbuf1);
    }
    else {
      tbuf1 = xplab_vsnprintf("%-34s  ", "not recognized....................");
      add_string(tbuf1, &tbuf2);
      FREEMEM(tbuf1);
    }

    g_snprintf(buf1, sizeof(buf1), "(%d-%d)x(%d-%d)",
	   hlist.c[i].roi.column.start, hlist.c[i].roi.column.end,
	       hlist.c[i].roi.line.start, hlist.c[i].roi.line.end);

    g_snprintf(buf2, sizeof(buf2), "(%d-%d)x(%d-%d)",
	   hlist.c[i].knl.column.start, hlist.c[i].knl.column.end,
	       hlist.c[i].knl.line.start, hlist.c[i].knl.line.end);

    tbuf1 = xplab_vsnprintf("%-21.21s %-21.21s\n", buf1, buf2);
    add_string(tbuf1, &tbuf2);
    FREEMEM(tbuf1);
  }

  xreport("%s", tbuf2);                   /* finally write whole report text */

  assert(tbuf1 == NULL);
  FREEMEM(tbuf2);
}



/* Mini Statistik ausgeben

Revision: 05.08.2005 E.Sack

Input : - file name
        - number of lines
        - string
        - mini statistic
Output: - line with the inputs
*/
char *
print_mini_statistic(char *preview, int_lst mst)
{
  char *buf = 0;

  CALLOC(char, MAXCHAR, buf);

  if (mst.cnt) {
    g_snprintf(buf, MAXCHAR, " [%d][%d] %s", mst.nr[0],     /* good */
	                                     mst.nr[1],     /* bad  */
	       preview?preview:"<empty>");
   }

  return buf;
}



/* Dumps the histogram to disk, corrected to start bin.

Revision: 22.12.2004 E.Sack

Input : - histogram
        - flag (0: no offset, 1: index consistent to start-bin)
        - filename
        - comment
Output: - file
*/
void 
write_histo(hst_def histo, int start_bin_index, char *hname, char *rem) 
{
  int  i, d;
  FILE *file;

  if (histo.hst.cnt == 0) {
    if (xctrl.cfg.verbose > BE_QUIET) {
      string_to_log(main_log, "histogram is empty\n", "", "");
    }

    return;
  }

  file = xplab_open_file(hname, "w");
  if (!file) {
    return;
  }

  d = check_name(rem);
  if (!d) {
    d = fprintf(file, rem);
    assert(d > 0);
  }

  d = fprintf(file, "# maximum is %5f at %d\n", histo.maxval,
	      histo.max_bin+histo.start_bin);
  assert(d > 0);

  d = fprintf(file, "# threshold is %d\n", histo.th_bin);
  assert(d > 0); 

  d = fprintf(file, "# sample size %d\n", histo.set_cnt);
  assert(d > 0); 

  d = fprintf(file, "# startbin reference %d\n", histo.start_bin);
  assert(d > 0);

  for (i = 0; i < histo.start_bin && start_bin_index; i++) {
    d = fprintf(file, "%d %f\n", i, 0.0); /* index correction to start bin */
    assert(d > 0);
  }

  for (i = 0; i < histo.hst.cnt; i++) {
    d = fprintf(file, "%d %f\n", i+histo.start_bin, histo.hst.dbl[i]);
    assert(d > 0);
  }

  if (xctrl.cfg.verbose > BE_QUIET) {
    string_to_log(main_log, "histogram dumped to '", hname, "'\n");
  }

  xplab_close_file(file, hname);
}



/* An integer array is dumped to disk as a textfile with indices and values.

Revision: 13.04.2002 E.Sack

Input : - array
Output: - file dumped to disk
*/
void 
dump_int(int *in, int cnt, char *fname) 
{
  int  i, d;
  FILE *file;

  assert(cnt);

  file = xplab_open_file(fname, "w");
  if (!file) {
    return;
  }

  for (i = 0; i < cnt; i++) {
    d = fprintf(file, "%d %d\n", i, in[i]);
    assert(d > 0);
  }

  if (xctrl.cfg.verbose > BE_QUIET) {
    string_to_log(main_log, "array dumped to '", fname, "'\n");
  }

  xplab_close_file(file, fname);
}



/* Outputs an array in a text file.

Input : - array
Output: - file
*/
void 
dump_double(double *in, int cnt, char *fname, char *msg, const char *rem, 
	    const char *subject) 
{
  int  i, d;
  FILE *file;

  assert(cnt>0);

  file = xplab_open_file(fname, "w");
  if (!file) {
    return;
  }

  if (subject) {
    if (strlen(subject)) {
      d = fprintf(file, "# subject: ");
      d += fprintf(file, subject);
      d += fprintf(file, "\n");
      assert(d > 0);
    }
  }

  if (rem) {
    if (strlen(rem)) {
      d = fprintf(file, rem);
      assert(d > 0);
    }
  }

  for (i = 0; i < cnt; i++) {
    d = fprintf(file, "%d %f\n", i, in[i]);
    assert(d > 0);
  }

  if (xctrl.cfg.verbose > BE_QUIET) {
    string_to_log(main_log, msg?msg:"array", " dumped to ", "");
    string_to_log(main_log, "'", fname, "'\n");
  }

  xplab_close_file(file, fname);
}
