/****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_algorithm.h"
#include "xplab_recall_be.h"
#include "xplab_maintenance_fe.h"
#include "xplab_maintenance_be.h"
#include "xplab_dbio.h"
#include "xplab_image.h"
#include "xplab_tprint.h"
#include "xplab_training_be.h"
#include "xplab_preferences_fe.h"
#include "xplab_selection.h"
#include "xplab_batch.h"

/* local definitions *******************************************************/
#define XPLAB_MAINTENANCE_TITLE_ITEM   "maintenance"
#define XPLAB_MAINTENANCE_EDIT_TITLE   "maintenance edit"
#define XPLAB_MAINTENANCE_WINDOW_WIDTH  511
#define XPLAB_MAINTENANCE_WINDOW_HEIGHT 342

#define XPLAB_MAINTENANCE_NODE_ID_PATTERN_SIZE_FORMAT_SEP "x"
#define XPLAB_MAINTENANCE_EDIT_IF GtkTreeIter piter, GtkWidget *fbox, int do_action

#define XPM_ICON_BGND_CHAR "+"                        /* XPM character token */
#define XPM_NB_BGND_CHAR   "."                        /* XPM character token */
#define XPM_NB_FGND_CHAR   "#"                        /* XPM character token */
#define XPM_NB_FGND_COLOR  "000000"               /* BMP foreground is black */
#define XPM_NB_BGND_COLOR XPM_COLOR_ICON_FILL_READ
#define XPM_PATTERN_FLAG_ICON  "icon"
#define XPM_PATTERN_VALU_ICON  16
#define XPM_PATTERN_FLAG_16x16 "16x16"
#define XPM_PATTERN_VALU_16x16 16
#define XPM_PATTERN_FLAG_32x32 "32x32"
#define XPM_PATTERN_VALU_32x32 32
#define XPM_PATTERN_FLAG_SCALE "scale"
#define XPM_PATTERN_FLAG_ORIGINAL "original"
#define CCF_INTRA_BUTTON "intra class"
#define CCF_INTER_BUTTON "inter classes"
#define CCF_INDEX_BUTTON "sort by index"
#define CCF_VALUE_BUTTON "sort by value"
#define MENU_BUTTON_AUX  "auxiliary items"
#define MENU_BUTTON_DUMP "dump pattern"
#define MENU_BUTTON_IMAGE        "Image"
#define MENU_BUTTON_IMAGE_STATIC "static"
#define MENU_BUTTON_IMAGE_ON     "on"

#define XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC_TITLE " XPlab database values "
#define XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC_TXT_TITLE ""          /* not used*/
#define XPLAB_MAINTENANCE_TREE_VIEW_COMMENT_TITLE " [unit] (comment) "
#define XPLAB_MAINTENANCE_TREE_VIEW_SUMMARY_TITLE " {summary of main items} "

enum {
  XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC = 0,   /* contains the graphical tree */
  XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT,   /* contains the graphic text   */
  XPLAB_MAINTENANCE_TREE_STORE_COMMENT,       /* contains parents spec       */
  XPLAB_MAINTENANCE_TREE_STORE_SUMMARY,       /* contains children summary   */
  XPLAB_MAINTENANCE_TREE_STORE_NODE_ID,       /* auxiliary ID for easy ctrl  */
  XPLAB_MAINTENANCE_TREE_STORE_NODE_FLAG,    /* auxiliary flag for easy ctrl */
  XPLAB_MAINTENANCE_TREE_STORE_COLUMNS                      /* store columns */
};

enum {
  XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC = 0,
  XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC_TXT,
  XPLAB_MAINTENANCE_TREE_VIEW_COMMENT,
  XPLAB_MAINTENANCE_TREE_VIEW_SUMMARY,
  XPLAB_MAINTENANCE_TREE_VIEW_COLUMNS                   /* displayed columns */
};

enum {
  XPLAB_MAINTENANCE_NODE_ID_UNDEF   = 0,
  XPLAB_MAINTENANCE_NODE_ID_DEFAULT = 0x8000,
  XPLAB_MAINTENANCE_NODE_ID_DB_ATTRIBUTES,        /* first root sibling node */
  XPLAB_MAINTENANCE_NODE_ID_DB_NAME,      
  XPLAB_MAINTENANCE_NODE_ID_DB_VERSION,   
  XPLAB_MAINTENANCE_NODE_ID_TPS_LIST,            /* second root sibling node */
  XPLAB_MAINTENANCE_NODE_ID_TPS_INDEX,
  XPLAB_MAINTENANCE_NODE_ID_TPS_ID,
  XPLAB_MAINTENANCE_NODE_ID_MEDIA_ATTRIBUTES,
  XPLAB_MAINTENANCE_NODE_ID_DPI,
  XPLAB_MAINTENANCE_NODE_ID_LINE_METHOD,
  XPLAB_MAINTENANCE_NODE_ID_OPOS,
  XPLAB_MAINTENANCE_NODE_ID_PERIOD_LENGTH,   
  XPLAB_MAINTENANCE_NODE_ID_L_PHASE,
  XPLAB_MAINTENANCE_NODE_ID_H_PHASE,
  XPLAB_MAINTENANCE_NODE_ID_CLASS_LIST,
  XPLAB_MAINTENANCE_NODE_ID_CLASS_INDEX,   
  XPLAB_MAINTENANCE_NODE_ID_CLASS_ATTRIBUTES,
  XPLAB_MAINTENANCE_NODE_ID_CLASS_NAME,
  XPLAB_MAINTENANCE_NODE_ID_CLASS_TRANSLATION,
  XPLAB_MAINTENANCE_NODE_ID_CLASS_COMMENT,
  XPLAB_MAINTENANCE_NODE_ID_CLASS_ALIGNMENT,
  XPLAB_MAINTENANCE_NODE_ID_CLASS_PRINTABLE,
  XPLAB_MAINTENANCE_NODE_ID_CLASS_FIXED,
  XPLAB_MAINTENANCE_NODE_ID_NB_LIST,
  XPLAB_MAINTENANCE_NODE_ID_NB_INDEX,
  XPLAB_MAINTENANCE_NODE_ID_CNB_LIST,                  /* contained in a NNB */
  XPLAB_MAINTENANCE_NODE_ID_NNB_LIST,                  /* contained in a CNB */
  XPLAB_MAINTENANCE_NODE_ID_CNB_INDEX,                 /* contained in a NNB */
  XPLAB_MAINTENANCE_NODE_ID_NNB_INDEX,                 /* contained in a CNB */
  XPLAB_MAINTENANCE_NODE_ID_CHAIN_X_OFFSET,
  XPLAB_MAINTENANCE_NODE_ID_CHAIN_Y_OFFSET,
  XPLAB_MAINTENANCE_NODE_ID_CHAIN_CLASS_INDEX,
  XPLAB_MAINTENANCE_NODE_ID_CHAIN_NB_INDEX,
  XPLAB_MAINTENANCE_NODE_ID_PATTERN
};

enum {
  XPLAB_MAINTENANCE_NODE_FLAG_DEFAULT = 0,
  XPLAB_MAINTENANCE_NODE_FLAG_FND     = 0x1,           /* bit exclusive flag */
  XPLAB_MAINTENANCE_NODE_FLAG_NEW     = 0x2,           /* bit exclusive flag */
  XPLAB_MAINTENANCE_NODE_FLAG_SEL     = 0x4            /* bit exclusive flag */
};

enum {
  XPLAB_MAINTENANCE_EDITABLE                = 1,       /* bit exclusive flag */
  XPLAB_MAINTENANCE_HASENTRY                = 2,       /* bit exclusive flag */
  XPLAB_MAINTENANCE_EDIT_DB_ATTRIBUTES      = 0,
  XPLAB_MAINTENANCE_EDIT_DB_NAME            = 0,      
  XPLAB_MAINTENANCE_EDIT_DB_VERSION         = 0,   
  XPLAB_MAINTENANCE_EDIT_TPS_LIST           = 0,
  XPLAB_MAINTENANCE_EDIT_TPS_INDEX          = 0,      
  XPLAB_MAINTENANCE_EDIT_TPS_ID             = XPLAB_MAINTENANCE_EDITABLE|XPLAB_MAINTENANCE_HASENTRY,
  XPLAB_MAINTENANCE_EDIT_MEDIA_ATTRIBUTES   = 0,
  XPLAB_MAINTENANCE_EDIT_DPI                = XPLAB_MAINTENANCE_EDITABLE|XPLAB_MAINTENANCE_HASENTRY,
  XPLAB_MAINTENANCE_EDIT_LINE_METHOD        = XPLAB_MAINTENANCE_EDITABLE,
  XPLAB_MAINTENANCE_EDIT_OPOS               = 0,
  XPLAB_MAINTENANCE_EDIT_PERIOD_LENGTH      = 0,   
  XPLAB_MAINTENANCE_EDIT_L_PHASE            = XPLAB_MAINTENANCE_EDITABLE|XPLAB_MAINTENANCE_HASENTRY,
  XPLAB_MAINTENANCE_EDIT_H_PHASE            = XPLAB_MAINTENANCE_EDITABLE|XPLAB_MAINTENANCE_HASENTRY,
  XPLAB_MAINTENANCE_EDIT_CLASS_LIST         = 0,
  XPLAB_MAINTENANCE_EDIT_CLASS_INDEX        = 0,   
  XPLAB_MAINTENANCE_EDIT_CLASS_ATTRIBUTES   = 0,
  XPLAB_MAINTENANCE_EDIT_CLASS_NAME         = XPLAB_MAINTENANCE_EDITABLE|XPLAB_MAINTENANCE_HASENTRY,
  XPLAB_MAINTENANCE_EDIT_CLASS_TRANSLATION  = XPLAB_MAINTENANCE_EDITABLE|XPLAB_MAINTENANCE_HASENTRY,
  XPLAB_MAINTENANCE_EDIT_CLASS_COMMENT      = XPLAB_MAINTENANCE_EDITABLE|XPLAB_MAINTENANCE_HASENTRY,
  XPLAB_MAINTENANCE_EDIT_CLASS_ALIGNMENT    = XPLAB_MAINTENANCE_EDITABLE,
  XPLAB_MAINTENANCE_EDIT_CLASS_PRINTABLE    = XPLAB_MAINTENANCE_EDITABLE,
  XPLAB_MAINTENANCE_EDIT_CLASS_FIXED        = XPLAB_MAINTENANCE_EDITABLE,
  XPLAB_MAINTENANCE_EDIT_NB_LIST            = 0,
  XPLAB_MAINTENANCE_EDIT_NB_INDEX           = 0,
  XPLAB_MAINTENANCE_EDIT_CNB_LIST           = 0,
  XPLAB_MAINTENANCE_EDIT_NNB_LIST           = 0,
  XPLAB_MAINTENANCE_EDIT_CNB_INDEX          = 0,
  XPLAB_MAINTENANCE_EDIT_NNB_INDEX          = 0,
  XPLAB_MAINTENANCE_EDIT_CHAIN_X_OFFSET     = 0,
  XPLAB_MAINTENANCE_EDIT_CHAIN_Y_OFFSET     = 0,
  XPLAB_MAINTENANCE_EDIT_CHAIN_CLASS_INDEX  = 0,
  XPLAB_MAINTENANCE_EDIT_CHAIN_NB_INDEX     = 0,
  XPLAB_MAINTENANCE_EDIT_PATTERN            = 0
};                                       /* edit macros and associated flags */

enum {
  XPLAB_EDIT_ACTION_ERASE  = 1,                        /* bit exclusive flag */
  XPLAB_EDIT_ACTION_CREATE = 2,                        /* bit exclusive flag */
  XPLAB_EDIT_ACTION_COPY   = 4,                        /* bit exclusive flag */
  XPLAB_EDIT_ACTION_PASTE  = 8,                        /* bit exclusive flag */
  XPLAB_EDIT_ACTION_SORT_BY_NAME = 16,                 /* bit exclusive flag */
  XPLAB_EDIT_ACTION_SORT_BY_TRANS = 32,                /* bit exclusive flag */
  XPLAB_EDIT_ACTION_WHERE_REFERENCED,                        /* integer flag */
  XPLAB_EDIT_ACTION_EXTRACT,                                 /* integer flag */
  XPLAB_EDIT_ACTION_STOP,                                    /* integer flag */
  XPLAB_EDIT_ACTION_DUMP,                                    /* integer flag */
  XPLAB_EDIT_DIALOG,                                         /* integer flag */
  XPLAB_EDIT_DIALOG_UPPER,                                   /* integer flag */
  XPLAB_EDIT_DIALOG_LOWER,                                   /* integer flag */
  XPLAB_EDIT_DIALOG_FRAME,                                   /* integer flag */
  XPLAB_QUERY_ACTION_DUMP,                                   /* integer flag */
  XPLAB_QUERY_SENSITIVITY,                                   /* integer flag */
  XPLAB_QUERY_IMAGE,                                         /* integer flag */
  XPLAB_EDIT_MENU                                            /* integer flag */
};

typedef struct {
  GtkWidget *intra_button, *inter_button, *sort_cv_button, *sort_nb_button;
  GtkWidget *dump_stop_button;
  guint     dump_id, stop_id;
  run_def   crun;
} ccf_def;

typedef struct {
  gint     tps_idx, cls_idx, nb_idx;
  gboolean aux;
} query_mimg;

typedef struct {
  gint is_cnb, tdx, cdx, ndx, sum, aux, aux2;
} query_chain;

typedef struct {
  GdkPixbuf *closed_read, *open_read, *open_write, *toedit_read, *unedit;
  GdkPixbuf *closed_write, *toedit_write;
} pixbuf_def;           /* global pixbuf refs for unreferring at program end */

typedef struct {
  GtkTreeIter piter;
  void        (*func)(XPLAB_MAINTENANCE_EDIT_IF);
} edit_set;

typedef struct {
  GtkWidget *edialog, *einfoline, *cancel_button;
  edit_set  eset;
} dialog_edit_def;

typedef struct {
  gint     tps_idx, cls_idx;
  cand_lst features;
} ccf_mat;

typedef struct {
  dialog_edit_def edt;
  ccf_def cbt;
  ccf_mat matrix;
} dialog_ccf_def;

typedef struct {
  dialog_edit_def edt;
  dialog_ccf_def  ccf;
} maint_dialog;

typedef struct {
  GtkTreeStore *tps_id, *dpi, *method, *l_phase, *h_phase, *class_name;
  GtkTreeStore *class_translation,  *class_comment, *alignment, *printable;
  GtkTreeStore *fixed;
} combo_store_def;                          /* stores used for input history */

typedef struct {
  GtkWidget  *mtop, *status_bar;   
  gint       tps_idx, cls_idx, nb_idx;                            /* indexes */
  gfloat     magn, magn_old;                                    /* magnitude */
  scroll_def scroll;                                       /* image features */
  img_ext    pimg;                                  /* original image buffer */
  pic_def    magn_img;                              /*   scaled image buffer */
} mimg_def;                           /* maintenance image window definition */

typedef struct {
  GtkWidget       *top_maintenance, *mlabel, *lock_button, **caller, *treeview;
  GtkWidget       *quit_button, *revert_button, *revert_quit_button, *ccf_menu;
  GtkWidget       *image_static, *image_on;
  GtkWidget       *expand_selected, *tree_menu_delete, *collapse_selected;
  GtkWidget       *row_extract, *row_edit, *where_referenced, *scale_button;
  GtkWidget       *row_dump, *x16_button, *x32_button, *orig_button, *icon_button;
  GtkSignalFunc   cancel_hook_func;
  GdkEvent        *popevent;
  gpointer        *cancel_hook_data;
  gint            items_deleted, items_created, items_changed, cursor_changed_id;
  gint            db_key_lock;/* separate keys to keep independent functionality */
  gint            db_key_read;
  gboolean        cactive, cstop;
  maint_dialog    dialog;
  GList           *mimg;                   /* stores image windows, to unref */
  nb_lst          copy_nbl;
  pattern_color   pc;                                       /* common colors */
  pixbuf_def      mpixbuf;                                       /* to unref */
  combo_store_def mcombo_store;                                  /* to unref */
  db_def          mdb;                                           /* to free  */
} maint_def;

/* local prototypes ********************************************************/
static void init_dialog_ccf_def(dialog_ccf_def *ccfd);
static void init_ccf_def(ccf_def *ccfd);
static void init_ccf_mat(ccf_mat *cm);
static void init_query_chain(query_chain *qc);
static void init_query_mimg(query_mimg *qm);
static void init_maint_dialog(maint_dialog *md);
static void init_maint_def(maint_def *mi);
static void init_dialog_edit_def(dialog_edit_def *edt);
static void init_edit_set(edit_set *es);
static void init_combo_store_def(combo_store_def *cs);
static void init_mimg_def(mimg_def *md);
static void free_mimg_def(mimg_def *md);
static void free_edit_set(edit_set *es);
static void free_dialog_edit_def(dialog_edit_def *edt);
static void free_maint_dialog(maint_dialog *md);
static void free_maint_def(maint_def *mi);
static void free_dialog_ccf_def(dialog_ccf_def *ccfd);
static void free_ccf_def(ccf_def *ccfd);
static void free_ccf_mat(ccf_mat *cm);

static void xplab_maintenance(void);
static void xplab_rebuild_db_tree(void);
static img_ext tps_image(cls_lst cl, int form_width, int space_x, int space_y);
static int check_tps_id(tps_lst tpl);
static int remove_identities(tps_lst tpl);
static int identity_rule(cls_lst cl, int c1, int n1, int c2, int n2);
static int identic_pattern(cls_lst cl, int *c, int *n1, int *n2);
static int copy_chains(cls_def *cls, int *n1, int n2);
static int check_identities(tps_lst tpl);
static int check_all_bval(tps_lst tpl);
static int check_all_pattern(tps_lst tpl);
static int check_empty_items(tps_lst tpl);
static int check_classes(tps_lst tpl);
static int check_bval(cls_lst cl, int tps_idx);
static int check_empty_classes(cls_lst cl, int tps_idx, int *sum);
static int check_nb_chn_lst(cls_lst clst, nb_chn_lst ncl, int clnr, int nbnr, int tps_idx);
static int check_pattern(cls_lst cl, int tps_idx);
static int check_cnb_index(cls_def cls, int cnr, int tps_idx);
static int check_cnb_alignments(tps_lst tpl);
static int check_cnb_alignment(cls_lst clst, int tps_idx);
static cand_lst intra_corr(nb_lst nbl, int *pairs, run_def *crun);
static cand_lst inter_corr(cls_lst cl, int *pairs, run_def *crun);
static void gen_intra_corr(ccf_mat *corr_features, db_def idb, int tnr, int cnr, int sort, run_def *crun);
static void gen_inter_corr(ccf_mat *corr_features, db_def idb, int tnr, int sort, run_def *crun);
static void string_ccf(cand_lst corr_features, char **prefix, int sort, gboolean intra, run_def *crun);
static int check_nb_identities(cls_lst cl, int tps_idx);
static void xdump_all_classes(db_def db, int tnr, int space_x, int space_y);
static void xdump_all_nb(db_def db, int tnr, int cnr, int space_x, int space_y);
static void xdump_nb_def(db_def db, int tnr, int cnr, int nnr, int space_x, int space_y);
static void ref_class(cls_lst clst, int cls_from, int cls_to);
static void cref_nb_chn_lst(nb_chn_lst ncl, int cls_from, int cls_to);
static int erase_cls_def(cls_lst *clst, int cnr, GtkWidget *label);
static void erase_tps_def(tps_lst *tpl, int tps_nr, GtkWidget *label);
static int class_sort(int cnt, cls_def *cls, int what);
static int getstr_printable_classes(cls_lst clst);
static void xextract_tps_def(db_def db, int tps_nr);
static void xextract_cls_def(db_def db, int tps_nr, int cls_nr);
static void xextract_nb_def(db_def db, int tps_nr, int cls_nr, int nb_nr);
static int empty_class(cls_lst clst);
static int remove_empty_classes(tps_lst tpl);
static void erase_nb_lst(cls_lst clst, int cnr, int vflag);
static void erase_nb_chn_lst(cls_lst clst, int cnr, int nnr, nb_chn_def *rptr, int vflag);
static void erase_nb_chn_def(cls_lst clst, int cnr, int nnr, int hnr, nb_chn_def *rptr, int vflag);
static void xdefault_cls_att(cls_att *catt, char *cname);
static int create_cls_def(cls_lst *clst, char *cname);
static idx_def paste_nb(GtkTreeModel *model, nb_lst copy_nbl, GtkTreeIter dst_iter, tps_lst tpl);
static void db_consistency(db_def db);
static int check_cls_lst(cls_lst cl, int tps_idx);
static void maintenance_tree_store_fill(GtkWidget *tview, gboolean revertable);
static void fill_combo_store_dpi(GtkTreeStore **combo_store);
static void fill_combo_store_method(GtkTreeStore **combo_store);
static void fill_combo_store_l_phase(GtkTreeStore **combo_store);
static void fill_combo_store_h_phase(GtkTreeStore **combo_store);
static void fill_combo_store_alignment(GtkTreeStore **combo_store);
static void fill_combo_store_printable(GtkTreeStore **combo_store);
static void fill_combo_store_fixed(GtkTreeStore **combo_store);
static void fill_combo_store_tps_id(GtkTreeStore **combo_store);
static void fill_combo_store_class_name(GtkTreeStore **combo_store);
static void fill_combo_store_class_translation(GtkTreeStore **combo_store);
static void fill_combo_store_class_comment(GtkTreeStore **combo_store);

/* Functions assembling widgets */
static text_def bmp_to_xpm(pic_def pattern, int square);
static pixbuf_def ref_pixbufs(void);
static void unref_pixbufs(pixbuf_def pixbuf);
static combo_store_def ref_combo_stores(void);
static void unref_combo_stores(combo_store_def combo_store);
static void maintenance_window(void);
static void maintenance_menu_bar(GtkWidget *vbox);
static void maintenance_menu_entry_file(GtkWidget *menu_bar);
static void maintenance_menu_entry_revert(GtkWidget *menu_bar);
static void maintenance_menu_entry_save(GtkWidget *menu_bar);
static void maintenance_menu_entry_tree(GtkWidget *menu_bar);
static void maintenance_menu_entry_tools(GtkWidget *menu_bar);
static void maintenance_menu_entry_image(GtkWidget *menu_bar);
static void maintenance_menu_image_static(GtkWidget *menu);
static void maintenance_menu_image_on(GtkWidget *menu);
static void maintenance_menu_item_quit(GtkWidget *menu);
static void maintenance_menu_item_save_quit(GtkWidget *menu);
static void maintenance_menu_item_revert_quit(GtkWidget *menu);
static void maintenance_menu_item_lock(GtkWidget *menu);
static void maintenance_menu_consistency(GtkWidget *menu);
static void maintenance_menu_consistency_check(GtkWidget *menu);
static void maintenance_menu_consistency_delete(GtkWidget *menu);
static void maintenance_menu_delete_identities(GtkWidget *menu);
static void maintenance_menu_delete_empty_classes(GtkWidget *menu);
static void maintenance_menu_delete_auxiliary_items(GtkWidget *tree_menu);
static  GtkWidget *maintenance_menu_where_referenced(GtkWidget *menu);
static void maintenance_menu_item_edit(GtkWidget *menu);
static void maintenance_menu_item_ccf(GtkWidget *menu);
static void maintenance_menu_item_extract(GtkWidget *menu);
static void maintenance_menu_item_dump(GtkWidget *menu);
static void maintenance_menu_item_collapse(GtkWidget *menu);
static void maintenance_menu_item_collapse_all(GtkWidget *menu);
static void maintenance_menu_item_collapse_selected(GtkWidget *menu);
static void maintenance_menu_item_expand(GtkWidget *menu);
static void maintenance_menu_item_expand_all(GtkWidget *menu);
static void maintenance_menu_item_expand_classes(GtkWidget *menu);
static void maintenance_menu_item_expand_nb(GtkWidget *menu);
static void maintenance_menu_item_expand_selected(GtkWidget *menu);
static void maintenance_menu_item_pattern(GtkWidget *menu);
static void maintenance_menu_pattern(GtkWidget *menu);
static void maintenance_menu_size(GtkWidget *tree_menu);
static void maintenance_menu_scale(GtkWidget *tree_menu);
static void maintenance_scrolled_window(GtkWidget *inbox);
static GtkWidget *maintenance_scrolled_treeview(void);
static GtkTreeStore *maintenance_tree_store_init(void);
static GtkWidget *maintenance_tree_view(GtkTreeStore *tstore);
static void cell_func_combo(GTK_CONVENTION_TREE_CELL_DATA_FUNC_RTS_VOID);
static void put_history_to_tree_store(gchar *htext, GtkTreeStore *tstore);

static gboolean set_nodes_flag(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN);
static gboolean expand_to_id(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN);
static gboolean delete_nodes_flag(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN);
static void treestore_db_def(GtkTreeView *tview, db_def db);
static void treestore_db_att(GtkTreeView *tview, GtkTreeIter piter, db_def db, gchar *path_string);
static GtkTreeIter treestore_node_db_att(GtkTreeView *tview, GtkTreeIter piter, db_def db, gchar *path_string);
static void treestore_db_name(GtkTreeView *tview, GtkTreeIter piter, char *db_name);
static void treestore_db_version(GtkTreeView *tview, GtkTreeIter piter, char *db_version);
static GtkTreeIter treestore_node_tps_lst(GtkTreeView *tview, GtkTreeIter piter, tps_lst tpl, gchar *path_string);
static void treestore_tps_lst(GtkTreeView *tview, GtkTreeIter piter, tps_lst tpl, gchar *path_string);
static GtkTreeIter treestore_node_tps_index(GtkTreeView *tview, GtkTreeIter piter, tps_def tps, gint tidx);
static void treestore_tps_def(GtkTreeView *tview, GtkTreeIter piter, tps_def tps, gint tidx);
static void treestore_media_att(GtkTreeView *tview, GtkTreeIter piter, media_att media);
static GtkTreeIter treestore_node_media_att(GtkTreeView *tview, GtkTreeIter piter, media_att media);
static void treestore_dpi(GtkTreeView *tview, GtkTreeIter piter, int dpi);
static void treestore_tps_ident(GtkTreeView *tview, GtkTreeIter piter, gchar *tps_ident);
static void treestore_cls_lst(GtkTreeView *tview, GtkTreeIter piter, cls_lst clst);
static GtkTreeIter treestore_node_cls_lst(GtkTreeView *tview, GtkTreeIter piter, gint cls_cnt);
static void treestore_nb_lst(GtkTreeView *tview, GtkTreeIter piter, nb_lst nbl);
static GtkTreeIter treestore_node_nb_lst(GtkTreeView *tview, GtkTreeIter piter, int nbl_cnt);
static void treestore_nb_def(GtkTreeView *tview, GtkTreeIter piter, nb_def nb, gint nidx);
static GtkTreeIter treestore_node_nb_index(GtkTreeView *tview, GtkTreeIter piter, nb_def nb, gint nidx);
static void treestore_pic_def(GtkTreeView *tview, GtkTreeIter piter, pic_def img);
static void treestore_nb_chn_lst(GtkTreeView *tview, GtkTreeIter piter, nb_chn_lst chn, int is_cnb);
static GtkTreeIter treestore_node_nb_chn_lst(GtkTreeView *tview, GtkTreeIter piter, int chn_cnt, int is_cnb);
static void treestore_nb_chn_def(GtkTreeView *tview, GtkTreeIter piter, nb_chn_def cd, int is_cnb, int node_id, int chidx);
static GtkTreeIter treestore_node_nb_chn_def(GtkTreeView *tview, GtkTreeIter piter, gint node_id, gint chidx, nb_chn_def cd, gint is_cnb);
static void treestore_cls_def(GtkTreeView *tview, GtkTreeIter piter, cls_def cls, gint cidx);
static GtkTreeIter treestore_node_cls_index(GtkTreeView *tview, GtkTreeIter piter, cls_def cls, gint cidx);
static void treestore_class_attributes(GtkTreeView *tview, GtkTreeIter piter, cls_def cls);
static void treestore_class_name(GtkTreeView *tview, GtkTreeIter piter, gchar *class_name);
static void treestore_class_translation(GtkTreeView *tview, GtkTreeIter piter, gchar *class_translation);
static void treestore_class_comment(GtkTreeView *tview, GtkTreeIter piter, gchar *class_comment);
static void treestore_class_alignment(GtkTreeView *tview, GtkTreeIter piter, ALIGMNT class_alignment);
static void treestore_class_printable(GtkTreeView *tview, GtkTreeIter piter, int printable);
static void treestore_class_fixed(GtkTreeView *tview, GtkTreeIter piter, int fixed);
static gboolean scan_rows(GtkTreeModel *tmodel, GtkTreeIter *parent);
static void maintenance_store_set(GtkTreeView *tview, GtkTreeIter piter, gchar *graphic_txt, gchar *comment_txt, gchar *summary_txt, gint node_id, GdkPixbuf *pixbuf);
static GtkTreeIter treestore_node_string(GtkTreeView *tview, GtkTreeIter piter, char *graphic_txt, char *comment_txt, char *summary_txt, int node_id, gchar *path_string);
static void treestore_leaf_string(GtkTreeView *tview, GtkTreeIter piter, char *graphic_txt, char *comment_txt, char *summary_txt, int node_id, int is_selectable);
static void treestore_inline_method(GtkTreeView *tview, GtkTreeIter piter, INLINE_METHOD method);
static void treestore_inline_def(GtkTreeView *tview, GtkTreeIter piter, inline_def ild);
static void treestore_linespacing_par(GtkTreeView *tview, GtkTreeIter piter, linespacing_par lpar);
static void treestore_opos(GtkTreeView *tview, GtkTreeIter piter, gint opos);
static GtkTreeIter treestore_node_period_length(GtkTreeView *tview, GtkTreeIter piter, period_def period);
static void treestore_period_def(GtkTreeView *tview, GtkTreeIter piter, period_def period);
static void treestore_l_phase(GtkTreeView *tview, GtkTreeIter piter, gint l_phase);
static gchar *treestore_comment_l_phase(gint l_phase);
static void treestore_h_phase(GtkTreeView *tview, GtkTreeIter piter, gint h_phase);
static gchar *treestore_comment_h_phase(gint h_phase);
static gchar *treestore_comment_period_length(period_def period);
static gchar *treestore_summary_period_length(period_def period);
static gchar *treestore_summary_db_att(char *db_name, char *db_version);
static gchar *treestore_summary_media_att(int dpi, INLINE_METHOD method);
static gchar *treestore_summary_tps_lst(int tps_cnt);
static gchar *treestore_summary_tps_index(gchar *tps_id, int class_cnt);
static gchar *treestore_summary_cls_lst(int cls_cnt);
static gchar *treestore_summary_cls_index(gchar *class_name, ALIGMNT class_alignment, gint nb_cnt);
static gchar *treestore_summary_nb_lst(int nb_cnt);
static gchar *treestore_summary_nb_index(area_def ad, gint cnb_cnt, gint nnb_cnt);
static gchar *treestore_summary_nb_chn_lst(gint chn_cnt);
static gchar *treestore_summary_nb_chn_def(gint cd_cdx, gint cd_ndx, gint cd_x, gint cd_y, gint is_cnb);
static gboolean set_nb_icon(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN);
static gboolean set_node_icon(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN);
static GdkPixbuf *node_icon(GtkTreeView *tview, GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter);
static void remove_node(GtkTreeModel *model);
static gboolean maintenance_row_2click(GtkWidget *widget, GdkEvent *event);
static void maintenance_image_on(gboolean on);
static void delete_image_window(GTK_CONVENTION_GFunc_RTS_VOID);
static void search_image_window(GTK_CONVENTION_GFunc_RTS_VOID);
static void update_image_window(GTK_CONVENTION_GFunc_RTS_VOID);
static void color_image_window(GTK_CONVENTION_GFunc_RTS_VOID);
static void update_image_window_all(void);
static void create_image_window(gint tps_idx, gint cls_idx, gint nb_idx);
static img_ext assemble_image(gint tps_idx, gint cls_idx, gint nb_idx, gchar **window_title);
static void rebuild_image_window(mimg_def *image_data_path, GtkSignalFunc mouse_cb, GtkSignalFunc expose_cb);
static XPLAB_INDEX_TYPE db_indexes(tps_lst tpl, gint tps_idx, gint cls_idx, gint nb_idx);

static void maintenance_save(void);
static void maintenance_revert(void);
static GdkPixbuf *pixbuf_nb(int row_height, int tps_idx, int cls_idx, int nb_idx);
static gboolean search_nb_index(GtkTreeModel *model, GtkTreeIter *io_iter, gint nb_idx);
static gboolean search_class_index(GtkTreeModel *model, GtkTreeIter *io_iter, gint cls_idx);
static gboolean search_tps_index(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN);
static void maintenance_reset_parse_status(void);
static void maintenance_row_edit(int do_item);
static void assert_node_id(GtkTreeModel *cmodel, GtkTreeIter iter, int id_query);
static void maintenance_statusbar_default(void);
static gint cursor_get_node_id(GtkTreeView *tview, GtkTreeIter *iter);
static void area_def_to_string(area_def ad, char **pattern_size);
static void dialog_delete_all(void);
static int unique_cls_name(cls_lst clst, char **cname);
static treenode_def xsearch_nb_index(GtkTreeModel *cmodel, gint tps_idx, gint cls_idx, gint nb_idx);
static treenode_def xsearch_class_index(GtkTreeModel *cmodel, int tps_idx, int cls_idx);
static treenode_def xsearch_tps_index(GtkTreeModel *cmodel, int tps_index);
static void index_to_cursor(gint tps_idx, gint cls_idx, gint nb_idx);
static int child_has_nb_index(GtkTreeModel *model, GtkTreeIter child, GtkTreeIter *parent);
static int child_has_class_index(GtkTreeModel *model, GtkTreeIter child, GtkTreeIter *parent);
static int child_has_tps_index(GtkTreeModel *model, GtkTreeIter child, GtkTreeIter *parent);
static int child_has_nnb_index(GtkTreeModel *model, GtkTreeIter child, GtkTreeIter *parent);
static int child_has_cnb_index(GtkTreeModel *model, GtkTreeIter child, GtkTreeIter *parent);
static int parent_has_class_index(GtkTreeModel *model, GtkTreeIter parent, GtkTreeIter *child);
static int parent_has_chain_class_index(GtkTreeModel *model, GtkTreeIter parent);
static int parent_has_chain_nb_index(GtkTreeModel *model, GtkTreeIter iter);
static gboolean search_below(GtkTreeModel *model, GtkTreeIter *io_iter, gint search_id);
static gboolean search_above(GtkTreeModel *model, GtkTreeIter *io_iter, gint search_id);
static pic_def scale_bmp(pic_def in_img, double vs);
static void mimage_to_state(mimg_def *image_data_path, gboolean redraw_buf);
static void image_window_popup(GdkEventButton *event, gpointer *image_data_path);
static void image_cursor(GdkEventButton *event, mimg_def *mdef);

static gboolean reference_nb_node(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN);
static gboolean reference_chain_node(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN);
static int reference_nb(GtkTreeModel *model, gint tdx, gint cdx, gint ndx, gint *moveto, gint *link_reset);
static int reference_class_nb(GtkTreeModel *model, int tps_idx, int cls_idx, int nb_idx, int *moveto, int *link_reset);
static void reference_nb_chn_lst(GtkTreeModel *model, int tps_idx, nb_chn_lst chain_list, int *moveto, int *link_reset);
static void reference_nb_lst(GtkTreeModel *model, gint tps_idx, cls_lst clst, int cls_idx, int *moveto, int *link_reset);
static void reference_cls_lst(GtkTreeModel *model, gint tps_idx, cls_lst clst, int *moveto, gint *link_reset);
static int reference_nb_chain(GtkTreeModel *model, int tps_idx, int cls_idx, int nb_idx, int *moveto, int *link_reset);
static int reference_chain(GtkTreeModel *model, int is_cnb, int tdx, int cdx, int ndx, int *moveto, int *link_reset);
static void expand_to_root(GtkTreeView *tree_view, GtkTreeIter iter, gboolean including);
static void tree_view_set_cursor(GtkTreeView *tree_view, GtkTreeIter iter);

static ccf_def ccf_setting(GtkWidget *fbox);
static void ccf_sensitivity(void);
static void dump_ccf(run_def *crun);
static void renew_ccf(ccf_mat *corr_features, int tnr, int cnr);

static void dialog_row_edit(dialog_edit_def *edt, int do_item);
static void dialog_row_edit_upper(dialog_edit_def *edt);
static void dialog_row_edit_lower(dialog_edit_def *edt);
static void dialog_row_edit_action(dialog_edit_def edt, int do_action);

static void edit_func_tps_lst(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_tps_def(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_cls_lst(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_cls_def(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_nb_lst(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_nb_def(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_cnb_lst(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_cnb_def(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_nnb_lst(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_nnb_def(XPLAB_MAINTENANCE_EDIT_IF);
static void edit_func_ccf(XPLAB_MAINTENANCE_EDIT_IF);

/* Handlers/Callbacks */
static void cb_cell_edited(GTK_CONVENTION_GtkCellRendererText_RTS_VOID);
static void cb_maintenance_row_select(GTK_CONVENTION_TREEVIEW_CURSOR_CHANGED_RTS_VOID);
static void cb_maintenance_row_expanded_collapsed(GTK_CONVENTION_TREEVIEW_ROW_EXPANDED_COLLAPSED_RTS_VOID);
static void cb_consistency_check(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_delete_identities(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_delete_auxiliary_items(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_delete_empty_classes(GTK_CONVENTION_MENUITEM_RTS_VOID);
static gboolean cb_maintenance_row_click(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN);
static gboolean cb_maintenance_save(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN);
static void cb_maintenance_revert(GTK_CONVENTION_MENUITEM_RTS_VOID);
static gboolean cb_maintenance_revert_event(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN);
static void cb_maintenance_revert_quit(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_maintenance_sensitivity(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_maintenance_lock(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static void cb_row_edit(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_ccf(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_collapse_selected(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_collapse_all(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_expand_all(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_expand_selected(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_expand_to_id(GTK_CONVENTION_MENUITEM_RTS_VOID);
static gboolean cb_maintenance_phase(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN); 
static void cb_maintenance_image_on(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);

static void cb_dialog_edit_sort_trans(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dialog_edit_sort_name(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dialog_edit_paste(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dialog_edit_copy(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dialog_edit_create(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dialog_edit_erase(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dialog_edit_cancel(GTK_CONVENTION_BUTTON_RTS_VOID);
static gboolean cb_dialog_delete_edit(GTK_CONVENTION_DIALOG_RTS_GBOOLEAN);
static void cb_dialog_ccf_dump(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dialog_ccf_stop(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dialog_ccf_setting(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID);
static void cb_pattern_icon(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);

static void cb_image_quit(GTK_CONVENTION_BUTTON_RTS_VOID);
static gboolean cb_image_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN);
static gboolean cb_image_expose(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN);
static gboolean cb_image_actions(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN);
static void cb_image_magnitude(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);

/* local data **************************************************************/

static maint_def mt;

/* Functions ***************************************************************/

static void
init_dialog_ccf_def(dialog_ccf_def *ccfd)
{
  assert(ccfd);

  init_dialog_edit_def(&ccfd->edt);
  init_ccf_def(&ccfd->cbt);
  init_ccf_mat(&ccfd->matrix);
}

static void
init_ccf_mat(ccf_mat *cm)
{
  assert(cm);

  cm->tps_idx = cm->cls_idx = -1;                   /* init to invalid index */
  init_cand_lst(&cm->features);
}

static void
init_ccf_def(ccf_def *ccfd)
{
  assert(ccfd);

  ccfd->dump_id = ccfd->stop_id = 0;

  init_run_def(&ccfd->crun);
}

static void
init_query_mimg(query_mimg *qm)
{
  assert(qm);

  qm->tps_idx = qm->cls_idx = qm->nb_idx = XPLAB_IDX_NONE;
  qm->aux = FALSE;
}

static void
init_query_chain(query_chain *qc)
{
  assert(qc);

  qc->is_cnb = qc->ndx = qc->sum = qc->aux = qc->aux2 = 0;
  qc->tdx = qc->cdx = -1;
}

static void
init_maint_dialog(maint_dialog *md)
{
  assert(md);

  init_dialog_edit_def(&md->edt);
  init_dialog_ccf_def(&md->ccf);
}

static void
init_edit_set(edit_set *es)
{
  assert(es);

  es->func = NULL;                                          /* zero pointer */
}

static void
init_dialog_edit_def(dialog_edit_def *edt)
{
  assert(edt);

  edt->edialog = NULL;                                      /* zero pointer */
  edt->einfoline = NULL;                                    /* zero pointer */
  edt->cancel_button = NULL;                                /* zero pointer */

  init_edit_set(&edt->eset);
}

static void
init_pixbuf_def(pixbuf_def *pb)
{
  assert(pb);

  pb->closed_read = NULL;                                    /* zero pointer */
  pb->open_read = NULL;                                      /* zero pointer */
  pb->open_write = NULL;                                     /* zero pointer */
  pb->toedit_read = NULL;                                    /* zero pointer */
  pb->unedit = NULL;                                         /* zero pointer */
  pb->closed_write = NULL;                                   /* zero pointer */
  pb->toedit_write = NULL;                                   /* zero pointer */
}

static void 
init_mimg_def(mimg_def *md)
{
  assert(md);

  md->mtop = NULL;                                           /* zero pointer */
  md->status_bar = NULL;                                     /* zero pointer */

  md->tps_idx = md->cls_idx = md->nb_idx = XPLAB_IDX_NONE;
  md->magn = md->magn_old = XPLAB_MAGN_X1_VALUE;

  init_scroll_def(&md->scroll);
  init_img_ext(&md->pimg);
  init_pic_def(&md->magn_img);
}

static void
init_combo_store_def(combo_store_def *cs)
{
  assert(cs);

  cs->tps_id = NULL;                                         /* zero pointer */
  cs->dpi = NULL;                                            /* zero pointer */
  cs->method = NULL;                                         /* zero pointer */
  cs->l_phase = NULL;                                        /* zero pointer */
  cs->h_phase = NULL;                                        /* zero pointer */
  cs->class_name = NULL;                                     /* zero pointer */
  cs->class_translation = NULL;                              /* zero pointer */
  cs->class_comment = NULL;                                  /* zero pointer */
  cs->alignment = NULL;                                      /* zero pointer */
  cs->printable = NULL;                                      /* zero pointer */
  cs->fixed = NULL;                                          /* zero pointer */
}

static void
init_maint_def(maint_def *mi)
{
  assert(mi);

  mi->top_maintenance = NULL;                                 /* zero pointer */
  mi->mlabel = NULL;                                          /* zero pointer */
  mi->lock_button = NULL;                                     /* zero pointer */
  mi->caller = NULL;                                          /* zero pointer */
  mi->treeview = NULL;                                        /* zero pointer */
  mi->quit_button = NULL;                                     /* zero pointer */
  mi->revert_button = NULL;                                   /* zero pointer */
  mi->revert_quit_button = NULL;                              /* zero pointer */
  mi->ccf_menu = NULL;                                        /* zero pointer */
  mi->image_static = NULL;                                    /* zero pointer */
  mi->image_on = NULL;                                        /* zero pointer */
  mi->expand_selected = NULL;                                 /* zero pointer */
  mi->tree_menu_delete = NULL;                                /* zero pointer */
  mi->collapse_selected = NULL;                               /* zero pointer */
  mi->row_extract = NULL;                                     /* zero pointer */
  mi->row_edit = NULL;                                        /* zero pointer */
  mi->where_referenced = NULL;                                /* zero pointer */
  mi->scale_button = NULL;                                    /* zero pointer */
  mi->row_dump = NULL;                                        /* zero pointer */
  mi->x16_button = NULL;                                      /* zero pointer */
  mi->x32_button = NULL;                                      /* zero pointer */
  mi->orig_button = NULL;                                     /* zero pointer */
  mi->icon_button = NULL;                                     /* zero pointer */

  mi->cancel_hook_func = NULL;                                /* zero pointer */
  mi->cancel_hook_data = NULL;                                /* zero pointer */

  mi->mimg = NULL;                                            /* zero pointer */

  mi->items_deleted = mi->items_created = mi->items_changed = 0;
  mi->cursor_changed_id = 0;

  mi->db_key_lock = mi->db_key_read = 0;

  mi->cactive = mi->cstop = FALSE;

  init_nb_lst(&mi->copy_nbl);
  init_pattern_color(&mi->pc);
  init_maint_dialog(&mi->dialog);
  init_pixbuf_def(&mi->mpixbuf);
  init_combo_store_def(&mi->mcombo_store);
  init_db_def(&mi->mdb);
}

static void 
free_mimg_def(mimg_def *md)
{
  assert(md);

  free_scroll_def(&md->scroll);
  free_img_ext(&md->pimg);
  free_pic_def(&md->magn_img);

  init_mimg_def(md);
}

static void
free_edit_set(edit_set *es)
{
  assert(es);

  init_edit_set(es);
}

static void
free_maint_def(maint_def *mi)
{
  assert(mi);
                                            /* don't free GTK stuff...       */

  free_maint_dialog(&mi->dialog);
  free_nb_lst(&mi->copy_nbl);
  free_db_def(&mi->mdb);

  init_maint_def(mi);                       /* ...however clear old pointers */
}

static void
free_ccf_def(ccf_def *ccfd)
{
  assert(ccfd);

  init_ccf_def(ccfd);
}

static void
free_maint_dialog(maint_dialog *md)
{
  assert(md);

  free_dialog_edit_def(&md->edt);
  free_dialog_ccf_def(&md->ccf);

  init_maint_dialog(md);
}

static void
free_dialog_ccf_def(dialog_ccf_def *ccfd)
{
  assert(ccfd);

  free_dialog_edit_def(&ccfd->edt);
  free_ccf_def(&ccfd->cbt);
  free_ccf_mat(&ccfd->matrix);
}

static void
free_ccf_mat(ccf_mat *cm)
{
  assert(cm);

  free_cand_lst(&cm->features);
  init_ccf_mat(cm);
}

static void
free_dialog_edit_def(dialog_edit_def *edt)
{
  assert(edt);

  free_edit_set(&edt->eset);

  init_dialog_edit_def(edt);
}



/* Initializes the maintenance phase.

Revision: 21.08.2005 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_init_maintenance(void)
{
  init_maint_def(&mt);
}



/* Frees the maintenance phase.

Revision: 21.08.2005 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_free_maintenance(void)
{
  free_maint_def(&mt);
}



/* Calculates number of printable classes.

Revision: 14.09.2002 E.Sack

Input : - class list
Output: - number of printable classes
*/
static int
getstr_printable_classes(cls_lst clst) 
{
  int c, printable = 0;

  for (c = 0; c < clst.cnt; c++) {
    if (clst.cls[c].att.printable) {
      printable++;
    }
  }

  return printable;
}



/* Stores a pure pattern to a class. Neither attributes nor references
   are considered. Hence the pattern is treated as an anonym, unclassified
   pattern.

Revision: 10.12.2005 E.Sack

Input : - pointer to tree model
        - pointer to NB source iter
        - pointer to class destination iter
        - TPS list
Output: - TPS list updated as flagged
        - new indexes (-1: NB not pasted)
*/
static idx_def
paste_nb(GtkTreeModel *model, nb_lst copy_nbl, GtkTreeIter dst_iter, 
	 tps_lst tpl)
{
  gint        dst_cnr, dst_nnr, dst_tps;
  gchar       *nbs = NULL;
  idx_def     idef;
  GtkTreeIter tps_iter, cls_iter;

  init_idx_def(&idef);

  assert_node_id(model, dst_iter, XPLAB_MAINTENANCE_NODE_ID_NB_LIST);
  assert(tpl.tps);
  assert(tpl.cnt>0);

  dst_tps = child_has_tps_index(model, dst_iter, &tps_iter);
  assert(dst_tps>=0 && dst_tps<tpl.cnt);

  dst_cnr = child_has_class_index(model, dst_iter, &cls_iter);
  assert(dst_cnr>=0 && dst_cnr<tpl.tps[dst_tps].cl.cnt);

  dst_nnr = cat_nb_def(copy_nbl.nb[0], 
		       &tpl.tps[dst_tps].cl.cls[dst_cnr].nbl, FALSE);

  string_to_log(main_log, "", "pattern pasted to ", "");
  int_to_string("NB[", dst_tps, "]", &nbs);
  int_to_string("[", dst_cnr, "]", &nbs);
  int_to_string("[", dst_nnr, "] (class name '", &nbs);
  string_to_log(main_log, nbs, tpl.tps[dst_tps].cl.cls[dst_cnr].att.class, 
		"')\n");

  FREEMEM(nbs);

  idef.tps_idx = dst_tps;
  idef.cls_idx = dst_cnr;
  idef.nb_idx = dst_nnr;

  return idef;
}



/* Investigates the DB of inconsistencies.

Revision: 17.09.2002 E.Sack

Input : - DB
        - pointer to DB file name
Output: - result dumped
*/
static void 
db_consistency(db_def db) 
{
  int err;

  string_to_log(main_log, "\nconsistency check of database", "", "");

  err = check_name(db.db_fname);
  if (err) {
    string_to_log(main_log, "\ndatabase file name undefined", "", "");
  }
  else {
    string_to_log(main_log, " '", db.db_fname, "'");
  }

  string_to_log(main_log, "", "", "\n");

  err = check_name(db.db_version);
  if (err) {
    string_to_log(main_log, "database version undefined", "", "\n");
  }

  if (!db.tpl.tps) {                                  /* consistency check */
    if (db.tpl.cnt) {
      int_to_log(main_log, 
		 "ERROR: consistency: TPS list pointer NULL, count is ", 
		 db.tpl.cnt, "\n");
      return;
    }
  }
  else {
    if (db.tpl.cnt<=0) {
      int_to_log(main_log, 
		 "ERROR: consistency: TPS list defined, count is ", 
		 db.tpl.cnt, "\n");
      return;
    }
  }

  string_to_log(main_log, "", "TPS ID check ... ", "");
  err = check_tps_id(db.tpl);
  if (!err) {
    string_to_log(main_log, "", "ok\n", "");
  }

  string_to_log(main_log, "", "class check ... ", "");
  err = check_classes(db.tpl);
  if (!err) {
    string_to_log(main_log, "", "ok\n", "");
  }
   
  string_to_log(main_log, "", "check empty items ... ", "");
  err = check_empty_items(db.tpl);
  if (!err) {
    string_to_log(main_log, "", "ok\n", "");
  }
   
  string_to_log(main_log, "", "pattern check ... ", "");
  err = check_all_pattern(db.tpl);
  if (!err) {
    string_to_log(main_log, "", "ok\n", "");
  }
   
  string_to_log(main_log, "", "B-value check ... ", "");
  err = check_all_bval(db.tpl);
  if (!err) {
    string_to_log(main_log, "", "ok\n", "");
  }
   
  string_to_log(main_log, "", "identity check ... ", "");
  err = check_identities(db.tpl);
  if (!err) {
    string_to_log(main_log, "", "ok\n", "");
  }

  string_to_log(main_log, "", "CNB check ... ", "");
  err = check_cnb_alignments(db.tpl);
  if (!err) {
    string_to_log(main_log, "", "ok\n", "");
  }

  string_to_log(main_log, "", "auxiliary references ... ", "");
  err = check_aux(db.tpl, 0);
  if (err) {
    string_to_log(main_log, "INFO: auxiliary references occurred", "\n", "");
  }
  else {
    string_to_log(main_log, "", "ok\n", "");
  }
}



/* Checks identities.

Revision: 10.06.2004 E.Sack

Input : - TPS list
Output: - error sum
*/
static int
check_identities(tps_lst tpl)
{
  int i, err = 0, errsum = 0;

  for (i = 0; i < tpl.cnt; i++) {
    err = check_nb_identities(tpl.tps[i].cl, i);

    if (err) {
      string_to_log(main_log, err>0?"INFO: ":"WARNING: ", "", "");
      int_to_log(main_log, "identities occurred for TPS[", i, "]");
      string_to_log(main_log, " (ID '", tpl.tps[i].ident, "')\n\n");
      errsum++;
    }
  }

  return errsum;
}



/* Checks B-values.

Revision: 10.06.2004 E.Sack

Input : - TPS list
Output: - error sum
*/
static int 
check_all_bval(tps_lst tpl)
{
  int i, d, err = 0;

  for (i = 0; i < tpl.cnt; i++) {
    d = check_bval(tpl.tps[i].cl, i);

    if (d) {
      string_to_log(main_log, "Errors occurred for TPS ID '", 
		    tpl.tps[i].ident, "'\n\n");
      err = 1;
    }
  }

  return err;
}



/* Checks pattern.

Revision: 10.06.2004 E.Sack

Input : - TPS list
Output: - error sum
*/
static int 
check_all_pattern(tps_lst tpl)
{
  int i, d, err = 0;

  for (i = 0; i < tpl.cnt; i++) {
    d = check_pattern(tpl.tps[i].cl, i);

    if (d) {
      string_to_log(main_log, "Errors occurred for TPS ID '", 
		    tpl.tps[i].ident, "'\n\n");
      err = 1;
    }
  }

  return err;
}



/* Checks empty items.

Revision: 01.01.2006 E.Sack

Input : - TPS list
Output: - error sum
*/
static int 
check_empty_items(tps_lst tpl)
{
  int  i, err = 0, errsum = 0;
  char *check_msg = NULL;

  for (i = 0; i < tpl.cnt; i++) {
    if (tpl.tps[i].cl.cnt) {
      err = check_empty_classes(tpl.tps[i].cl, i, &errsum);

      if (err) {
	int_to_log(main_log, "WARNING: ", err, " empty class");
	string_to_log(main_log, err==1?"":"es", " occurred in TPS ID '", "");
	string_to_log(main_log, "", tpl.tps[i].ident, "'\n");
      }
    }
    else {
      check_message_tps(errsum?"WARNING: ":"\nWARNING: ", i, &check_msg);
      string_to_log(main_log, check_msg, " - no classes in TPS ID '", "");
      string_to_log(main_log, "", tpl.tps[i].ident, "'\n");
      errsum++;
      FREEMEM(check_msg);
    }
  }

  return errsum;
}



/* Checks classes.

Revision: 10.06.2004 E.Sack

Input : - TPS list
Output: - error sum
*/
static int 
check_classes(tps_lst tpl)
{
  int i, d, err = 0;

  for (i = 0; i < tpl.cnt; i++) {
    d = check_cls_lst(tpl.tps[i].cl, i);

    if (d) {
      string_to_log(main_log, "Errors occurred for TPS ID '", 
		    tpl.tps[i].ident, "'\n\n");
      err = 1;
    }
  }

  return err;
}



/* Checks CNB-NNB consistency of alignment for all classes.

Revision: 05.10.2004 E.Sack

Input : - TPS list
Output: - error sum
*/
static int 
check_cnb_alignments(tps_lst tpl)
{
  int i, d, err = 0;

  for (i = 0; i < tpl.cnt; i++) {
    d = check_cnb_alignment(tpl.tps[i].cl, i);

    if (d) {
      string_to_log(main_log, "Warnings occurred for TPS ID '", 
		    tpl.tps[i].ident, "'\n\n");
      err = 1;
    }
  }

  return err;
}



/* Checks CNB-NNB consistency of alignment.

Revision: 10.12.2005 E.Sack

Input : - class list
        - TPS index
Output: - output results in the log
        - flag (0: no errors detected, else: errors detected)
*/
static int 
check_cnb_alignment(cls_lst clst, int tps_idx) 
{
  int     c, i, err = 0, common_alignment, nnb, nnb_cnt, nnb_class_idx;
  char    *check_msg = NULL;
  ALIGMNT parent_alignment, child_alignment;
   
  if (!clst.cls) {                                    /* consistency check */
    assert(!clst.cnt);
  }
  else {
    assert(clst.cnt>=0);
  }

  for (c = 0; c < clst.cnt; c++) {                          /* all classes */
    if (!clst.cls[c].nbl.nb) {                        /* consistency check */
      assert(!clst.cls[c].nbl.cnt);
    }
    else {
      assert(clst.cls[c].nbl.cnt>=0);
    }

    parent_alignment = clst.cls[c].att.alignment;

    for (i = 0; i < clst.cls[c].nbl.cnt; i++) {                  /* all NB */
      nnb_cnt = clst.cls[c].nbl.nb[i].nnb.cnt;

      if (!nnb_cnt) {
	continue;                     /* not a CNB having a NNB child list */
      }

      common_alignment = 0;
      for (nnb = 0; nnb < nnb_cnt; nnb++) {
	nnb_class_idx = clst.cls[c].nbl.nb[i].nnb.cd[nnb].cdx;

	if (nnb_class_idx<0 || nnb_class_idx>=clst.cnt) {
	  common_alignment = 2;
	  break;
	}

	child_alignment = clst.cls[nnb_class_idx].att.alignment;

	if (child_alignment == parent_alignment) {
	  common_alignment = 1;
	  break;
	}
      }

      switch (common_alignment) {
      case 0:
	FREEMEM(check_msg);
	check_message_nb(err?"WARNING: ":"\nWARNING: ", tps_idx, c, i, 
			 &check_msg);
	string_to_log(main_log, check_msg, 
		      " - alignment of CNB not found in its NNB\n", "");
	err++;

	break;

      case 2:
	FREEMEM(check_msg);
	check_message_nb(err?"WARNING: ":"\nWARNING: ", tps_idx, c, i, 
			 &check_msg);
	string_to_log(main_log, check_msg, 
		      " - NNB of CNB not found in TPS\n", "");
	err++;

	break;
      }
    }
  }

  FREEMEM(check_msg);
  return err;
}



/* Checks identic TPS IDs in list.

Revision: 11.09.2004 E.Sack

Input : - document list
        - verbose flag
Output: - flag (0: no identic TPS IDs, >0 number of identic TPS IDs)
*/
static int 
check_tps_id(tps_lst tpl)
{
  int  d, i, j, err = 0;
  char *check_msg = NULL;

  for (i = 0; i < tpl.cnt-1; i++) {
    for (j = i+1; j < tpl.cnt; j++) {
      d = strcmp(tpl.tps[i].ident, tpl.tps[j].ident);

      if (!d) {
	FREEMEM(check_msg);
	check_message_tps(err?"WARNING: ":"\nWARNING: ", i, &check_msg);
	string_to_log(main_log, check_msg, "", "");

	FREEMEM(check_msg);
	check_message_tps(" and ", j, &check_msg);
	string_to_log(main_log, check_msg, " - identic TPS ID '", 
		      tpl.tps[i].ident);
	string_to_log(main_log, "", "", "'\n");

	err++;
      }
    }
  }

  FREEMEM(check_msg);
  return err;
}



/* Deletes identic pattern in a document list. Chains are saved.

Revision: 14.09.2002 E.Sack

Input : - document list
Output: - output results on the screen
        - flag (0 DB unchanged, 1 DB changed)
*/
static int
remove_identities(tps_lst tpl) 
{
  int err1sum = 0, err2sum = 0, i = 0, err1, err2, c, n1, n2;

  if (!tpl.tps) {                                     /* consistency check */
    assert(!tpl.cnt);
    return 0;                                                   /* no data */
  }
  else {
    assert(tpl.cnt>0);
  }

  string_to_log(main_log, "\n", "delete identities", "\n");

  do {
    err1 = identic_pattern(tpl.tps[i].cl, &c, &n1, &n2);

    if (err1) {
      err2 = copy_chains(&tpl.tps[i].cl.cls[c], &n1, n2);
      if (err2) {
	err2sum++;
      }

      erase_nb_def(tpl.tps[i].cl, c, n1, NULL, 1);     /* erase identic NB */
      err1sum++;                                             /* repetition */
      continue;
    }
    else i++;
  } while (i < tpl.cnt);

  if (!err1sum) {
    string_to_log(main_log, "", "no identities intra classes found\n", "");
    return 0;
  }
  else {
    int_to_log(main_log, "", err2sum, " chain");
    string_to_log(main_log, err2sum==1?"":"s", " erased\n", "");

    int_to_log(main_log, "", err1sum, " identit");
    string_to_log(main_log, err1sum==1?"y":"ies", " erased\n", "");
    return 1;
  }
}



/* Deletes classes which have no NB pattern. Chains concerned are erased

Revision: 19.01.2003 E.Sack

Input : - document list
Output: - output results on log
        - flag (0 DB unchanged, 1 DB changed)
*/
static int
remove_empty_classes(tps_lst tpl) 
{
  int d, erased = 0, i = 0, cnr;

  if (!tpl.tps) {                                     /* consistency check */
    assert(!tpl.cnt);
    return 0;                                                   /* no data */
  }
  else {
    assert(tpl.cnt>0);
  }

  string_to_log(main_log, "\n", "delete empty classes", "\n");

  do {
    cnr = empty_class(tpl.tps[i].cl);

    if (cnr >= 0) {
      d = erase_cls_def(&tpl.tps[i].cl, cnr, NULL);

      if (d) {
	erased++;
      }

      continue;                                    /* repeat investigation */
    }
    else {
      i++;
    }
  } while (i < tpl.cnt);

  if (!erased) {
    string_to_log(main_log, "", "no empty classes found\n", "");
    return 0;
  }
  else {
    int_to_log(main_log, "", erased, " empty class");
    string_to_log(main_log, erased==1?"":"es", " erased\n", "");
    return 1;
  }
}



/* Takes over chains.

Input : - pointer to class list
        - pointer to first NB index
        - second NB index
Output: - class list updated
        - first NB index updated
        - flag (0: no chains, 1: chains taken over)
Postcondition: - NBx possesses all chains of NBy,
               - NBy unchanged
               - first NB index is index of NBy
Usage: - first NB ready for deletion
*/
static int
copy_chains(cls_def *cls, int *n1, int n2)
{
  int max, min, i, j, identic, cnt;

  assert(*n1!=n2);

  /* which chain list is greater?                                          */
  if (cls->nbl.nb[*n1].cnb.cnt > cls->nbl.nb[n2].cnb.cnt) {
    max = *n1;
    min = n2;
  }
  else {
    max = n2;
    min = *n1;
  }

  *n1 = min;

  if (!cls->nbl.nb[min].cnb.cnt) return 0;          /* NB not concatenated */

  cnt = cls->nbl.nb[max].cnb.cnt;                             /* short cut */
  for (i = 0; i < cls->nbl.nb[min].cnb.cnt; i++) {        /* chain on both */
    for (j = 0, identic = 0; j < cnt; j++) {
      if (cls->nbl.nb[max].cnb.cd[j].cdx == cls->nbl.nb[min].cnb.cd[i].cdx &&
          cls->nbl.nb[max].cnb.cd[j].ndx == cls->nbl.nb[min].cnb.cd[i].ndx &&
          cls->nbl.nb[max].cnb.cd[j].x   == cls->nbl.nb[min].cnb.cd[i].x   &&
          cls->nbl.nb[max].cnb.cd[j].y   == cls->nbl.nb[min].cnb.cd[i].y) {
	identic = 1;                          /* identic chain, already in */
	break;
      }
    }

    if (!identic) {                                    /* chain not yet in */
      REALLOC(nb_chn_def, cnt+1, cls->nbl.nb[max].cnb.cd);    /* take over */
      cls->nbl.nb[max].cnb.cd[cnt].cdx = cls->nbl.nb[min].cnb.cd[i].cdx;
      cls->nbl.nb[max].cnb.cd[cnt].ndx = cls->nbl.nb[min].cnb.cd[i].ndx;
      cls->nbl.nb[max].cnb.cd[cnt].x   = cls->nbl.nb[min].cnb.cd[i].x;
      cls->nbl.nb[max].cnb.cd[cnt].y   = cls->nbl.nb[min].cnb.cd[i].y;
      cnt++;
    }
  }

  cls->nbl.nb[max].cnb.cnt = cnt;

  return 1;                                       /* all chains taken over */
}



/* Investigates a class list.

Revision: 11.09.2004 E.Sack

Input : - class list
        - TPS index
Output: - output results in the log
        - flag (0: no errors detected, else: errors detected)
*/
static int 
check_cls_lst(cls_lst clst, int tps_idx) 
{
  int  d, c, i, err = 0;
  char *check_msg = NULL;
   
  if (!clst.cls) {                                    /* consistency check */
    assert(!clst.cnt);
  }
  else {
    assert(clst.cnt>=0);
  }

  for (c = 0; c < clst.cnt-1; c++) {
    for (i = c+1; i < clst.cnt; i++) {
      if (strcmp(clst.cls[c].att.class, clst.cls[i].att.class)) {
	continue;
      }

      FREEMEM(check_msg);
      check_message_class("\nERROR: ", tps_idx, c, &check_msg);
      check_message_class(" and ", tps_idx, i, &check_msg);
      string_to_log(main_log, check_msg, "", "");
      string_to_log(main_log, " - same class name '", clst.cls[i].att.class,
		    "'\n");
      FREEMEM(check_msg);
      return 1;
    }
  }

  for (c = 0; c < clst.cnt; c++) {
    for (i = 0; i < clst.cls[c].nbl.cnt; i++) {
      if (!(AREA_SIZE(clst.cls[c].nbl.nb[i].pat.ar))) {
	FREEMEM(check_msg);
	check_message_nb("\nERROR: ", tps_idx, c, i, &check_msg);
	string_to_log(main_log, check_msg, " - empty pattern store\n", "");
	FREEMEM(check_msg);
	return 1;
      }
    }
  }

  for (c = 0; c < clst.cnt; c++) {                          /* all classes */
    if (!clst.cls[c].nbl.nb) {                        /* consistency check */
      assert(!clst.cls[c].nbl.cnt);
    }
    else {
      assert(clst.cls[c].nbl.cnt>=0);
    }

    d = check_cnb_index(clst.cls[c], c, tps_idx);

    if (d) {
      FREEMEM(check_msg);
      return 1;
    }

    for (i = 0; i < clst.cls[c].nbl.cnt; i++) {                  /* all NB */
      if (clst.cls[c].att.printable == 0 && !clst.cls[c].nbl.nb[i].cnb.cnt) {
	FREEMEM(check_msg);
	check_message_nb(err?"WARNING: ":"\nWARNING: ", tps_idx, c, i, 
			 &check_msg);
	string_to_log(main_log, check_msg, 
		      " - neither chained nor printable\n", "");
        err++;
      }

      err += check_nb_chn_lst(clst, clst.cls[c].nbl.nb[i].cnb, c, i, tps_idx);
      err += check_nb_chn_lst(clst, clst.cls[c].nbl.nb[i].nnb, c, i, tps_idx);
    }
  }

  FREEMEM(check_msg);
  return err;
}



/* Investigates a chain list.

Revision: 11.09.2004 E.Sack

Input : - class list
        - chain list
        - class index
        - NB index
        - TPS index
Output: - output results in the log
        - flag (0: no errors detected, 1: errors detected)
*/
static int
check_nb_chn_lst(cls_lst clst, nb_chn_lst ncl, int clnr, int nbnr, 
		 int tps_idx)
{
  int  j, cc, nn;
  char *check_msg = NULL;

  if (!ncl.cd) {                                      /* consistency check */
    assert(!ncl.cnt);
  }
  else {
    assert(ncl.cnt>=0);
  }

  for (j = 0; j < ncl.cnt; j++) {                            /* all chains */
    cc = ncl.cd[j].cdx;                                       /* short cut */
    nn = ncl.cd[j].ndx;

    if (0 > cc || clst.cnt <= cc) {
      FREEMEM(check_msg);
      check_message_nb("\nERROR: ", tps_idx, clnr, nbnr, &check_msg);
      check_message_class(" - untrained index ", tps_idx, cc, &check_msg);
      string_to_log(main_log, check_msg, " in chain list\n", "");
      FREEMEM(check_msg);
      return 1;
    }

    if (0 > nn || clst.cls[cc].nbl.cnt <= nn) {
      FREEMEM(check_msg);
      check_message_nb("\nERROR: ", tps_idx, clnr, nbnr, &check_msg);
      check_message_nb(" - untrained index ", tps_idx, cc, nn, &check_msg);
      string_to_log(main_log, check_msg, " in chain list\n", "");
      FREEMEM(check_msg);
      return 1;
    }
  }

  return 0;
}



/* Investigates a class list for empty classes.

Revision: 01.01.2006 E.Sack

Input : - class list
        - TPS index
Output: - output results on the log
        - sum of empty classes
*/
static int 
check_empty_classes(cls_lst clst, int tps_idx, int *sum) 
{
  int  c, err = 0;
  char *check_msg = NULL;

  assert(sum);

  for (c = 0; c < clst.cnt; c++) {
    if (!clst.cls[c].nbl.nb) {
      assert(!clst.cls[c].nbl.cnt);

      FREEMEM(check_msg);
      check_message_class((*sum)?"WARNING: ":"\nWARNING: ", tps_idx, c, 
			  &check_msg);

      string_to_log(main_log, check_msg, " - no pattern\n", "");
      err++;
      (*sum)++;
    }
    else {
      assert(clst.cls[c].nbl.cnt>0);
    }
  }

  FREEMEM(check_msg);
  return err;
}



/* Detects in a class list an empty class.

Revision: 19.01.2003 E.Sack

Input : - class list
Output: - class index (-1: no empty class found, >=0: index of empty class)
*/
static int 
empty_class(cls_lst clst) 
{
  int cnr;
   
  for (cnr = 0; cnr < clst.cnt; cnr++) {
    if (!clst.cls[cnr].nbl.cnt) {
      assert(!clst.cls[cnr].nbl.nb);
      return cnr;
    }
    else {
      assert(clst.cls[cnr].nbl.cnt>0);
      assert(clst.cls[cnr].nbl.nb);
    }
  }

  return -1;
}



/* Checks the CNBs of a class. NBs are not allowed chained within their own
   class.

Revision: 11.09.2004 E.Sack

Input : - class
        - class number
        - TPS index
Output: - flag (0: no error, 1: error)
*/
static int
check_cnb_index(cls_def cls, int cnr, int tps_idx) 
{
  int i, j;

  for (i = 0; i < cls.nbl.cnt; i++) {
    for (j = 0; j < cls.nbl.nb[i].cnb.cnt; j++) {
      if (cls.nbl.nb[i].cnb.cd[j].cdx == cnr) {
	char *check_msg = NULL;

	check_message_nb("\nERROR: ", tps_idx, cnr, i, &check_msg);
	check_message_nb(" and ", tps_idx, cnr, cls.nbl.nb[i].cnb.cd[j].ndx,
			 &check_msg);
	string_to_log(main_log, check_msg, " - chained\n", "");
	FREEMEM(check_msg);
        return 1;
      }
    }
  }

  return 0;
}



/* Tests the b-values, which must be covered with foreground pixels.

Revision: 10.09.2004 E.Sack

Input : - class list
        - TPS index
Output: - error sum
*/
static int 
check_bval(cls_lst clst, int tps_idx) 
{
  int  i, j, c, n, err, h, w;
  char *check_msg = NULL;
  PICT m, b;

  for (c = 0, err = 0; c < clst.cnt; c++) {
    for (n = 0; n < clst.cls[c].nbl.cnt; n++) {
      if (!clst.cls[c].nbl.nb[n].pat.bvl) {
	continue;                                           /* no b-values */
      }

      w = clst.cls[c].nbl.nb[n].pat.ar.width;                 /* short cut */
      h = clst.cls[c].nbl.nb[n].pat.ar.height;                /* short cut */

      for (i = 0; i < h; i++) {
	for (j = 0; j < w; j++) {
          m = clst.cls[c].nbl.nb[n].pat.mem[j + i*w];         /* short cut */
          b = clst.cls[c].nbl.nb[n].pat.bvl[j + i*w];         /* short cut */

          if (b && !m) {
	    FREEMEM(check_msg);
	    string_to_log(main_log, err?"":"\n", "", "");
	    check_message_nb("ERROR: ", tps_idx, c, n, &check_msg);
	    string_to_log(main_log, "", check_msg, " - loosed b-value ");
	    int_to_log(main_log, "(", j, ",");
	    int_to_log(main_log, "", i, ")\n");
	    err++;
	  }
	}
      }
    }
  }

  FREEMEM(check_msg);
  return err;
}



/* Tests the patterns of a class list.

Revision: 10.09.2004 E.Sack

Input : - class list
        - TPS index
Output: - flag (0: no error, 1: error)
*/
static int 
check_pattern(cls_lst clst, int tps_idx) 
{
  int     c, i, s, z, links, rechts, oben, unten, err = 0, size;
  char    *buf = NULL, *check_msg = NULL;
  pic_def a;

  init_pic_def(&a);

  if (!clst.cnt) {
    return 0;
  }

  a.ar = max_cls_area(clst);                            /* maximal NB size */

  size = AREA_SIZE(a.ar);

  if (!size) {
    return 0;                                               /* empty class */
  }

  CALLOC(PICT, size, a.mem);

  for (c = 0; c < clst.cnt; c++) {      
    for (i = 0; i < clst.cls[c].nbl.cnt; i++) {
      rechts = links = oben = unten = 0;
      a.ar = clst.cls[c].nbl.nb[i].pat.ar;

      copy_bmps(clst.cls[c].nbl.nb[i].pat, a, clst.cls[c].nbl.nb[i].pat.ar, 
		0, 0, 0, 0);

      for (z = 0; z < a.ar.height; z++) {
	if (a.mem[z*a.ar.width] == FOREGROUND) links++;
	if (a.mem[a.ar.width-1 + z*a.ar.width] == FOREGROUND) rechts++;
      }

      for (s = 0; s < a.ar.width; s++) {
	if (a.mem[s] == FOREGROUND) oben++;
	if (a.mem[s + (a.ar.height-1)*a.ar.width] == FOREGROUND) unten++;
      }

      if (!rechts) {
	add_string("right", &buf);
	err = 1;
      }

      if (!links) {
	add_string("left", &buf);
	err = 1;
      }

      if (!oben) {
	add_string("upper", &buf);
	err = 1;
      }

      if (!unten) {
	add_string("lower", &buf);
	err = 1;
      }

      if (err) {
	check_message_nb("\nERROR: ", tps_idx, c, i, &check_msg);
	cat_string(" - no pixel on ", buf, " frame border\n", &check_msg);
	string_to_log(main_log, "", "", check_msg);

	free_pic_def(&a);
	FREEMEM(buf);
	FREEMEM(check_msg);
	return 1;
      }
    }
  }

  free_pic_def(&a);
  FREEMEM(buf);
  FREEMEM(check_msg);
  return 0;
}



/* Investigates the NB list for identic NB, independent of classes and chains.

Revision: 10.09.2004 E.Sack

Input : - class list
        - TPS index
Output: - number of identities
        - output results on the screen
*/
static int 
check_nb_identities(cls_lst clst, int tps_idx) 
{
  int  c1, n1, c2, n2, err, errsum, neg_is_warning = 1;
  char *check_msg = NULL;

  for (c1 = 0, errsum = 0; c1 < clst.cnt; c1++) {
    for (n1 = 0; n1 < clst.cls[c1].nbl.cnt; n1++) {
      for (c2 = c1; c2 < clst.cnt; c2++) {
        for (n2 = c2==c1?n1+1:0; n2 < clst.cls[c2].nbl.cnt; n2++) {
	  err = identity_rule(clst, c1, n1, c2, n2);
	  if (err) {
	    neg_is_warning = c1==c2 ? -1 : neg_is_warning;

	    errsum++;

	    FREEMEM(check_msg);
	    check_message_nb(c1==c2?"\nWARNING: ":"\nINFO: ", tps_idx, c1, n1,
			     &check_msg);

	    string_to_log(main_log, "", "", check_msg);

	    FREEMEM(check_msg);
	    check_message_nb(" and ", tps_idx, c2, n2, &check_msg);

	    string_to_log(main_log, check_msg, "", " - identic pattern");
	  }
	}
      }
    }
  }

  if (errsum) {
    string_to_log(main_log, "", "\n", "");
  }

  FREEMEM(check_msg);
  return neg_is_warning * errsum;
}



/* Detects in a NB list identic NB within their class, independent of their
   chains.

Revision: 27.07.2002 E.Sack

Input : - class list
        - pointer to indexes
Output: - flag (0: no identities, 1: identities detected)
        - indexes updated
*/
static int 
identic_pattern(cls_lst clst, int *c, int *n1, int *n2) 
{
  int identic;

  for (*c = 0; *c < clst.cnt; (*c)++) {
    for (*n1 = 0; *n1 < clst.cls[*c].nbl.cnt-1; (*n1)++) { 
      for (*n2 = *n1+1; *n2 < clst.cls[*c].nbl.cnt; (*n2)++) {
	identic = identity_rule(clst, *c, *n1, *c, *n2);
	if (identic) {
	  return 1;
	}
      }
    }
  }

  return 0;
}



/* Investigates the difference of two NB, independent of their class and
   chains.

Revision: 17.09.2002 E.Sack

Input : - class list
        - indexes
Output: - flag (0: NBs not identic, 1: NBs identic)
*/
static int
identity_rule(cls_lst clst, int c1, int n1, int c2, int n2)
{
  int d;

  assert(c1==c2?n1!=n2:1);
  assert(0<=c1 && c1<clst.cnt);
  assert(0<=c2 && c2<clst.cnt);

  if (!clst.cls) {                                    /* consistency check */
    assert(!clst.cnt);
    return 0;                        /* not identic because of empty class */
  }
  else {
    assert(clst.cnt>=0);
  }

  assert(0<=n1 && n1<clst.cls[c1].nbl.cnt);
  assert(0<=n2 && n2<clst.cls[c2].nbl.cnt);

  if (clst.cls[c1].att.alignment != clst.cls[c2].att.alignment) {
    return 0;               /* not identic because of different alignment */
  }

  d = compare_bmp(clst.cls[c1].nbl.nb[n1], clst.cls[c2].nbl.nb[n2]);

  return d;
}



/* Calculates the correlation values of the NBs of all classes.

Revision: 30.07.2005 E.Sack

Input : - class list
        - pointer to preview pairs number
Output: - correlation matrix
        - pointer updated
*/
static cand_lst
inter_corr(cls_lst clst, int *pairs, run_def *crun)
{
  int      c1, c2, n1, n, n2, r, cnt;
  cand_lst corr_list;

  init_cand_lst(&corr_list);

  if (clst.cnt <= 1) {
    return corr_list;                                  /* not enough classes */
  }

  /*(N-1)*(N-2) class permutations without repetition and with different classes */
  for (c1 = 0, cnt = 0; c1 < clst.cnt-1; c1++) {
    for (c2 = c1+1; c2 < clst.cnt; c2++) {
      cnt += clst.cls[c1].nbl.cnt * clst.cls[c2].nbl.cnt;        /* c1 != c2 */
    }
  }

  if (pairs) {
    *pairs = cnt;
    return corr_list;                                       /* dummy classes */
  }

  corr_list.cnt = cnt;
  CALLOC(cand_def, corr_list.cnt, corr_list.c);

  xplab_progress_configure(crun, NULL);
  xplab_progress_set_value(crun, 1.0, (gdouble)corr_list.cnt);

  /* correlation of all class permutations                                   */
  for (c1 = n = r = 0; c1 < clst.cnt-1; c1++) {
    for (c2 = c1+1; c2 < clst.cnt; c2++) {
      for (n1 = 0; n1 < clst.cls[c1].nbl.cnt; n1++) {
	for (n2 = 0; n2 < clst.cls[c2].nbl.cnt; n2++, n++) {
	  xplab_progress_set_value(crun, (gdouble)(++r), (gdouble)corr_list.cnt);

	  if (*crun->pactive == FALSE) {
	    return corr_list;
	  }

	  corr_list.c[n].c = c1;                             /* class1 index */
	  corr_list.c[n].aux = n1;                              /* NB1 index */
	  corr_list.c[n].rbn = c2;                           /* class2 index */
	  corr_list.c[n].n = n2;                                /* NB2 index */

	  corr_list.c[n].m = full_correlation1(clst.cls[c2].nbl.nb[n2].pat,
					       clst.cls[c1].nbl.nb[n1].pat,
					       crun);
	}
      }
    }
  }

  return corr_list;
}



/* Calculates the CV-matrix of all NB of a class.

Usage: this can be used to check eq. 3.2-13 [Tou/Gonzalez] for justifying
the 1-NN distance usage, else the q-NN distance should be used.

Revision: 30.07.2005 E.Sack

Input : - NB list
        - pointer to preview pairs number
        - pointer to control structure
Output: - correlation matrix
        - pointer updated
Usage:  If pairs pointer is not NULL, only the pairs number is calculated
        and neither the crun pointer is considered nor any further 
        calculation is done. This feature is for a quick preview for the
        amount of possible CCF calculations.
*/
static cand_lst
intra_corr(nb_lst nbl, int *pairs, run_def *crun)
{
  int      n1, n2, n, r, cnt;
  cand_lst corr_features;

  init_cand_lst(&corr_features);

  if (nbl.cnt <= 1) {
    return corr_features;                              /* not enough pattern */
  }

  cnt = nbl.cnt * nbl.cnt;

  if (pairs) {                             /* only calculate number of pairs */
    *pairs = cnt;
    return corr_features;                                   /* dummy classes */
  }

  assert(crun);

  corr_features.cnt = cnt;                       /* linear ordered CC matrix */
  CALLOC(cand_def, corr_features.cnt, corr_features.c);

  xplab_progress_configure(crun, NULL);
  xplab_progress_set_value(crun, 1.0, (gdouble)corr_features.cnt);

  for (n1 = n = r = 0; n1 < nbl.cnt; n1++) {               /* enroll indexes */
    for (n2 = 0; n2 < nbl.cnt; n2++, n++) {                     /* ACF & CCF */
      xplab_progress_set_value(crun, (gdouble)(++r), (gdouble)corr_features.cnt);

      if (*crun->pactive == FALSE) {
	break;
      }

      corr_features.c[n].rbn = n1;               /* ACF: n1==n2, CCF: n1!=n2 */
      corr_features.c[n].n = n2;
      corr_features.c[n].c = 0;                     /* no class index needed */
      corr_features.c[n].aux = 0;

      corr_features.c[n].m = full_correlation1(nbl.nb[n2].pat, 
					       nbl.nb[n1].pat,
					       crun);
    }
  }

  return corr_features;
}



/* Dumps correlation features of NBs containing correlation matrix and offsets.

Revision: 04.01.2006 E.Sack

Input : - pointer to correlation features
        - DB
        - TPS index
        - class index
        - flag (0: sorting by index, 1: sorting by CV)
        - pointer to DB name
        - pointer to control structure
Output: - correlation features updated
        - correlation features dumped
Precondition: - if candidate list of correlation features is non-empty it must
                be consistent with DB parameters.
Usage: The candidate list is only updated if it is empty. That has the advantage
       of low runtime consumption, as data can be re-used without re-calculation.
*/
static void 
gen_intra_corr(ccf_mat *corr_features, db_def idb, int tnr, int cnr, int sort,
	       run_def *crun) 
{
  char    *buf = NULL;
  cls_def cls;

  assert(corr_features);
  assert(tnr>=0);
  assert(cnr>=0);

  if (!idb.tpl.cnt) {
    return;
  }

  if (!idb.tpl.tps[tnr].cl.cnt) {
    return;
  }

  if (!idb.tpl.tps[tnr].cl.cls[cnr].nbl.cnt) {
    return;
  }

  cls = idb.tpl.tps[tnr].cl.cls[cnr];

  if (corr_features->features.cnt == 0) {        /* only update on empty list */
    assert(corr_features->features.c == NULL);           /* consistency check */
    corr_features->features = intra_corr(cls.nbl, NULL, crun); /*update matrix*/
    corr_features->tps_idx = tnr;                   /* update current indexes */
    corr_features->cls_idx = cnr;
  }

  if (!corr_features->features.cnt) {
    string_to_log(main_log,"",
		  "WARNING: class has not enaugh pattern for CCF","\n\n");
    return;
  }
  else {
    assert(corr_features->features.cnt == cls.nbl.cnt*cls.nbl.cnt);
  }

  int_to_string("\nlist of intra class correlation matrix (", cls.nbl.cnt, "x", &buf);
  int_to_string("", cls.nbl.cnt, ") ", &buf);
  int_to_string("of class [", tnr, "]", &buf);
  int_to_string("[", cnr, "] ´", &buf);
  cat_string("", cls.att.class, "´\n", &buf);

  string_ccf(corr_features->features, &buf, sort, TRUE, crun);

  string_to_log(main_log, "", buf, "\n");

  FREEMEM(buf);
}



/* Dumps correlation matrix of all classes.

Revision: 04.01.2006 E.Sack

Input : - correlation matrix
        - pointer to prefix string
        - flag (0: sorting by index, 1: sorting by CV)
        - flag (TRUE: intra class matrix, FALSE: inter class matrix)
        - pointer to control structure
Output: - pointer to prefix string updated
*/
static void
string_ccf(cand_lst corr_features, char **prefix, int sort, gboolean intra,
	   run_def *crun)
{
  gint     i;
  char     num[256];
  cand_lst corr_tmp;

  init_cand_lst(&corr_tmp);
  copy_cand_lst(&corr_features, &corr_tmp);                 /* protect input */

  if (sort == 1) {               /* sorting by CV, order of matrix disturbed */
    qsort((void*)corr_tmp.c, corr_tmp.cnt, sizeof(cand_def),
	  compare_cand_def);                      /* sort in ascending order */
  }

  cat_string("list sorted by ", sort?"correlation value":"NB index", "\n", prefix);

  if (intra == TRUE) {                                        /* intra class */
    add_string("entry   Ref - NB     CV         Offset (x,y)\n", prefix);
  }
  else {                                                    /* inter classes */
    add_string("entry   Class/NB  - Class/NB       CV         Offset (x,y)\n", prefix);
  }

  xplab_progress_configure(crun, NULL);
  xplab_progress_set_value(crun, 1.0, (gdouble)corr_tmp.cnt);

  for (i = 0; i < corr_tmp.cnt && *crun->pactive == TRUE; i++) {
    xplab_progress_set_value(crun, (gdouble)(i+1), (gdouble)corr_tmp.cnt);

    g_snprintf(num, sizeof(num), "%5d", i);
    cat_string(num, (intra == FALSE)?":    ":":  ", "", prefix);

    g_snprintf(num, sizeof(num), "%3d", corr_tmp.c[i].rbn);

    if (intra == FALSE) {                                   /* inter classes */
      cat_string(num, "/", "", prefix);
      g_snprintf(num, sizeof(num), "%3d", corr_tmp.c[i].n);
    }

    cat_string(num, " - ", "", prefix);

    if (intra == FALSE) {                                   /* inter classes */
      g_snprintf(num, sizeof(num), "  %3d", corr_tmp.c[i].c);
      cat_string(num, "/", "", prefix);

      g_snprintf(num, sizeof(num), "%3d", corr_tmp.c[i].aux);
    }
    else {                                                    /* intra class */
      g_snprintf(num, sizeof(num), "%-3d", corr_tmp.c[i].n);
    }

    cat_string(num, (intra == FALSE)?"     ":"   ", "", prefix);

    double_to_string("", corr_tmp.c[i].m.cv, " ", prefix);

    add_string("     (", prefix);
    int_to_string("", corr_tmp.c[i].m.x, ",", prefix);
    int_to_string("", corr_tmp.c[i].m.y, ")\n", prefix);
  }

  if (*crun->pactive == FALSE) {
    add_string("*** CCF interrupted - not complete ***\n", prefix);
  }

  free_cand_lst(&corr_tmp);
}



/* Dumps correlation matrix of all classes.

Revision: 04.01.2006 E.Sack

Input : - pointer to correlation features
        - DB
        - TPS index
        - flag (0: sorting by index, 1: sorting by CV)
        - pointer to control structure
Output: - correlation features updated
        - correlation features dumped
Precondition: - if candidate list of correlation features is non-empty it must
                be consistent with DB parameters.
Usage: The candidate list is only updated if it is empty. That has the advantage
       of low runtime consumption, as data can be re-used without re-calculation.
*/
static void 
gen_inter_corr(ccf_mat *corr_features, db_def idb, int tnr, int sort, 
	       run_def *crun) 
{
  char *buf = NULL;

  assert(corr_features);
  assert(tnr>=0);

  if (!idb.tpl.cnt) {
    return;
  }

  if (!idb.tpl.tps[tnr].cl.cnt) {
    return;
  }

  if (corr_features->features.cnt == 0) {        /* only update on empty list */
    assert(corr_features->features.c == NULL);           /* consistency check */
    corr_features->features = inter_corr(idb.tpl.tps[tnr].cl, NULL, crun);/*update matrix*/
    corr_features->tps_idx = tnr;                     /* update current index */
  }

  if (!corr_features->features.cnt) {
    string_to_log(main_log,"",
		  "WARNING: class has not enaugh pattern for CCF","\n\n");
    return;
  }

  add_string("\nlist of inter class correlation matrix", &buf);
  int_to_string(" of TPS[", tnr, "]", &buf);
  cat_string(" '", idb.tpl.tps[tnr].ident, "'", &buf);
  int_to_string(" (NB pairs: ", corr_features->features.cnt, ")\n", &buf);

  string_ccf(corr_features->features, &buf, sort, FALSE, crun);

  string_to_log(main_log, "", buf, "\n");

  FREEMEM(buf);
}



/* Builds an image from trained pattern set with constant linespace, 
   inclusive space left. The format is: left to the right, top to bottom.

Revision: 14.01.2006 E.Sack

Input : - class list
        - preferred form width
        - vertical and horizontal space
Output: - image
Postcondition: - only printable NBs are used.
Usage: The algorithm uses a pseudo class (built from the whole TPS) and
       the class image algorithm.
*/
static img_ext
tps_image(cls_lst clst, int form_width, int space_x, int space_y) 
{
  int     c, i, n;
  cls_def pcls;
  img_ext img_tmp;

  init_img_ext(&img_tmp);
  init_cls_def(&pcls);

  for (c = 0; c < clst.cnt; c++) {                  /* assemble pseudo class */
    if (!clst.cls[c].att.printable) {
      continue;                                    /* only printable classes */
    }

    for (i = 0; i < clst.cls[c].nbl.cnt; i++) {            /* preset indexes */
      clst.cls[c].nbl.nb[i].dat.cidx = c;
      clst.cls[c].nbl.nb[i].dat.nidx = i;
    }

    if (pcls.att.class == NULL) {                                    /* init */
      copy_cls_def(&clst.cls[c], &pcls, TRUE);                  /* no chains */
    }
    else {                                                 /* extend NB list */
      for (i = 0; i < clst.cls[c].nbl.cnt; i++) {
	cat_nb_def(clst.cls[c].nbl.nb[i], &pcls.nbl, TRUE);
      }
    }
  }

  if (pcls.att.class == NULL) {
    return img_tmp;                                 /* no patterns assembled */
  }

  img_tmp.eimg = xplab_sample_image(pcls, form_width, space_x, space_y);/* image of pseudo class */

  for (c = n = 0; c < clst.cnt; c++) {           /* save image location list */
    if (!clst.cls[c].att.printable) {
      continue;                                    /* only printable classes */
    }

    for (i = 0; i < clst.cls[c].nbl.cnt; i++, n++) {
      cat_nb_aux(pcls.nbl.nb[n].dat, &img_tmp.il);
    }
  }

  free_cls_def(&pcls);                                 /* free pseudo class  */

  return img_tmp;
}



/* Dumps TPS class list in a single image file.

Revision: 14.01.2006 E.Sack

Input : - database
        - TPS index
        - horizontal and vertical frame width
Output: - -
*/
static void 
xdump_all_classes(db_def db, int tnr, int space_x, int space_y) 
{
  char    *fname = NULL;
  img_ext img_tmp;

  assert(tnr>=0);

  init_img_ext(&img_tmp);

  if (!db.tpl.cnt) {
    return;
  }

  if (!db.tpl.tps[tnr].cl.cnt) {
    return;
  }

  img_tmp = tps_image(db.tpl.tps[tnr].cl, 0, space_x, space_y);

  if (AREA_SIZE(img_tmp.eimg.ar)) {
    add_string(db.db_fname, &fname);
    int_to_string(TPSNAME_DEFAULT, tnr, CLASS_NAME_DEFAULT, &fname);

    dump_pic_def(img_tmp.eimg, fname, img_tmp.eimg.ar, 0, 0, 1, PNM_FORMAT_UNDEF);
  }
  else {
    string_to_log(main_log, "TPS image not printable\n", "", "");
  }

  free_img_ext(&img_tmp);
  FREEMEM(fname);
}



/* Dumps images of all NB of a class in a single file.

Revision: 25.07.2002 E.Sack

Input : - NB list
        - class number
        - pointer to DB name
        - NB horizontal and vertical free space
Output: - image file
*/
static void 
xdump_all_nb(db_def db, int tnr, int cnr, int space_x, int space_y) 
{
  char    *fname = NULL;
  pic_def img_tmp;

  assert(tnr>=0);
  assert(cnr>=0);

  if (!db.tpl.cnt) {
    return;
  }

  if (!db.tpl.tps[tnr].cl.cnt) {
    return;
  }

  if (!db.tpl.tps[tnr].cl.cls[cnr].nbl.cnt) {
    return;
  }

  img_tmp = xplab_sample_image(db.tpl.tps[tnr].cl.cls[cnr], 0, space_x, space_y);

  if (AREA_SIZE(img_tmp.ar)) {
    add_string(db.db_fname, &fname);
    int_to_string(TPSNAME_DEFAULT, tnr, CLASS_NAME_DEFAULT, &fname);
    int_to_string(FILE_EXT, cnr, "", &fname);

    dump_pic_def(img_tmp, fname, img_tmp.ar, 0, 0, 1, PNM_FORMAT_UNDEF);
  }
  else {
    int_to_log(main_log, "NB image of class[", cnr, "] not printable\n");
  }

  free_pic_def(&img_tmp);
  FREEMEM(fname);
}



/* Dumps an image containing one NB of a class in a file.

Revision: 22.02.2004 E.Sack

Input : - class number
        - NB number
        - NB list
        - pointer to DB file name
        - horizontal and vertical free space (image border)
Output: - image file
*/
static void
xdump_nb_def(db_def db, int tnr, int cnr, int nnr, int space_x, int space_y)
{
  char    *fname = NULL;
  pic_def img_tmp;

  assert(tnr>=0);
  assert(cnr>=0);
  assert(nnr>=0);

  if (!db.tpl.cnt) {
    return;
  }

  if (!db.tpl.tps[tnr].cl.cnt) {
    return;
  }

  if (!db.tpl.tps[tnr].cl.cls[cnr].nbl.cnt) {
    return;
  }

  img_tmp = xextend_image(db.tpl.tps[tnr].cl.cls[cnr].nbl.nb[nnr].pat, space_x, 
			  space_y);

  add_string(db.db_fname, &fname);
  int_to_string(TPSNAME_DEFAULT, tnr, CLASS_NAME_DEFAULT, &fname);
  int_to_string(FILE_EXT, cnr, NBNAME_DEFAULT, &fname);
  int_to_string("", nnr, "", &fname);

  dump_pic_def(img_tmp, fname, img_tmp.ar, 0, 0, 1, PNM_FORMAT_UNDEF);

  free_pic_def(&img_tmp);
  FREEMEM(fname);
}



/* Builds an entry for a menu bar.

Revision: 30.07.2005 E.Sack

Input : - pointer to menu bar
        - pointer to caller widget pointer
Output: - menu bar updated
*/
void
xplab_maintenance_menu(GtkWidget *menu_bar, GtkWidget **caller)
{
  GtkWidget *maint_entry;

  maint_entry = gtk_menu_item_new_with_label(XPLAB_MAINTENANCE_TITLE_ITEM);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), maint_entry);

  g_signal_connect(G_OBJECT(maint_entry), "event", 
		   G_CALLBACK(cb_maintenance_phase), (gpointer)caller);
}



/* Deletes a class.

Revision: 01.09.2002 E.Sack

Input : - pointer to class list
        - class number
Output: - class list updated
        - flag (0: no changes, 1: changes)
*/
static int
erase_cls_def(cls_lst *clst, int cnr, GtkWidget *label) 
{
  int  last;
  char *class_name = NULL;
   
  assert(clst);

  if (!clst->cls) {                                   /* consistency check */
    assert(!clst->cnt);
    return 0;                                                   /* no data */
  }
  else {
    assert(clst->cnt>0);
  }

  last = clst->cnt-1;
  assert(cnr >= 0);
  assert(cnr <= last);

  add_string(clst->cls[cnr].att.class, &class_name);        /* save string */

  ref_class(*clst, cnr, -1);
  ref_class(*clst, last, cnr);                             /* index change */

  if (cnr < last) {                                  /* class index change */
    copy_cls_def(&clst->cls[last], &clst->cls[cnr], FALSE);
  }

  free_cls_def(&clst->cls[last]);

  clst->cnt--;                                           /* one class less */

  string_to_label(label, "class '", class_name, "' deleted");
  FREEMEM(class_name);

  if (clst->cnt) {
    REALLOC(cls_def, clst->cnt, clst->cls);
  }
  else {
    free_cls_lst(clst);
    return 1;
  }

  unref_indexes(*clst, NULL);

  return 1;
}



/* Changes a class reference.

Revision: 14.09.2002 E.Sack

Input : - document class list
        - old class number
        - new class number
Output: - document class list updated
*/
static void
ref_class(cls_lst clst, int cls_from, int cls_to) 
{
  int     cls_nr, nb_nr;
  nb_def  nb;
  cls_def class;

  if (cls_from == cls_to) {                           /* nothing to change */
    return;
  }

  if (!clst.cls) {                                    /* consistency check */
    assert(!clst.cnt);
    return;                                                     /* no data */
  }
  else {
    assert(clst.cnt>0);
  }

  for (cls_nr = 0; cls_nr < clst.cnt; cls_nr++) {
    class = clst.cls[cls_nr];                         /* shortcut to class */

    if (!class.nbl.nb) {                              /* consistency check */
      assert(!class.nbl.cnt);
    }
    else {
      assert(class.nbl.cnt>0);
    }

    for (nb_nr = 0; nb_nr < class.nbl.cnt; nb_nr++) {         
      nb = class.nbl.nb[nb_nr];                          /* shortcut to NB */

      cref_nb_chn_lst(nb.cnb, cls_from, cls_to);                /* parents */
      cref_nb_chn_lst(nb.nnb, cls_from, cls_to);               /* children */
    }  
  }
}



/* Changes class chain references.

Revision: 14.09.2002 E.Sack

Input : - NB chain list
        - old class number
        - new class number
Output: - NB chain list updated
*/
static void
cref_nb_chn_lst(nb_chn_lst ncl, int cls_from, int cls_to)
{
  int        chn_nr;
  nb_chn_def *chain = 0;

  if (!ncl.cd) {                                      /* consistency check */
    assert(!ncl.cnt);
    return;                                                     /* no data */
  }
  else {
    assert(ncl.cnt>0);
  }

  for (chn_nr = 0; chn_nr < ncl.cnt; chn_nr++) {
    chain = &ncl.cd[chn_nr];                          /* shortcut to chain */

    if (chain->cdx == cls_from) {
      chain->cdx = cls_to;
    }
  }
}



/* Deletes a NB list and updates the class list.

Revision: 24.01.2004 E.Sack

Input : - class list
        - class number
        - NB number
        - verbose flag
Output: - class list updated
*/
static void
erase_nb_lst(cls_lst clst, int cnr, int vflag)
{
  int last;

  while (clst.cls[cnr].nbl.cnt > 0) {
    last = clst.cls[cnr].nbl.cnt-1; /* better performance erasing last one */

    erase_nb_def(clst, cnr, last, NULL, vflag);
  }
}



/* Deletes a CNB/NNB list and updates the class list parents/children. The 
   CNB/NNB reference lists are kept consistent.

Revision: 24.01.2004 E.Sack

Input : - class list
        - class index
        - NB index
        - pointer to CNB/NNB list
        - verbose flag
Output: - class list updated
*/
static void
erase_nb_chn_lst(cls_lst clst, int cnr, int nnr, nb_chn_def *rptr, int vflag)
{
  int        d, i, deleted_refs, deleted_nb = 0;
  int        chain_cls_nr, chain_nb_nr, is_cnb;
  nb_chn_lst *chain_list;

  /* switch between CNB and NNB keeps consistency */
  if (rptr == clst.cls[cnr].nbl.nb[nnr].cnb.cd) {
    is_cnb = 1;
    chain_list = &clst.cls[cnr].nbl.nb[nnr].cnb;/* referenced in NNB lists */
  }
  else if (rptr == clst.cls[cnr].nbl.nb[nnr].nnb.cd) {
    is_cnb = 0;
    chain_list = &clst.cls[cnr].nbl.nb[nnr].nnb;/* referenced in CNB lists */
  }
  else {
    assert(0);
  }

  for (i = 0; i < chain_list->cnt; i++) {
    chain_cls_nr = chain_list->cd[i].cdx;
    chain_nb_nr  = chain_list->cd[i].ndx;

    assert(chain_cls_nr >= 0 && chain_cls_nr < clst.cnt);
    assert(chain_nb_nr >= 0 && chain_nb_nr < clst.cls[chain_cls_nr].nbl.cnt);

    d = pref_nb_chn_lst(is_cnb?                   /* mark referenced lists */
			clst.cls[chain_cls_nr].nbl.nb[chain_nb_nr].nnb:
			clst.cls[chain_cls_nr].nbl.nb[chain_nb_nr].cnb,
			cnr, -1, nnr, -1);
    assert(d>0);                               /* there must be references */
  }

  free_nb_chn_lst(chain_list);                               /* erase list */

  deleted_refs = unref_indexes(clst, &deleted_nb);
  assert(deleted_refs>0);                      /* there must be references */

  if (vflag) {
    int_to_log(main_log, "", deleted_refs, " NB reference");
    string_to_log(main_log, deleted_refs==1?"":"s", " deleted", "\n");

    int_to_log(main_log, "", deleted_nb, " NB");
    string_to_log(main_log, deleted_nb==1?"":"s", " deleted", "\n");
  }
}



/* Deletes a CNB/NNB and updates the class list parents/children. The 
   CNB/NNB reference lists are kept consistent.

Revision: 25.01.2004 E.Sack

Input : - class list
        - class index
        - NB index
        - pointer to CNB/NNB list
        - verbose flag
Output: - class list updated
*/
static void
erase_nb_chn_def(cls_lst clst, int cnr, int nnr, int hnr, nb_chn_def *rptr, 
		 int vflag)
{
  int        d, deleted_refs, deleted_nb = 0;
  int        chain_cls_nr, chain_nb_nr, is_cnb;
  nb_chn_lst *chain_list;

  /* switch between CNB and NNB keeps consistency */
  if (rptr == clst.cls[cnr].nbl.nb[nnr].cnb.cd) {
    is_cnb = 1;
    chain_list = &clst.cls[cnr].nbl.nb[nnr].cnb;/* referenced in NNB lists */
  }
  else if (rptr == clst.cls[cnr].nbl.nb[nnr].nnb.cd) {
    is_cnb = 0;
    chain_list = &clst.cls[cnr].nbl.nb[nnr].nnb;/* referenced in CNB lists */
  }
  else {
    assert(0);
  }

  chain_cls_nr = chain_list->cd[hnr].cdx;
  chain_nb_nr  = chain_list->cd[hnr].ndx;

  d = pref_nb_chn_lst(is_cnb?                     /* mark referenced lists */
		      clst.cls[chain_cls_nr].nbl.nb[chain_nb_nr].nnb:
		      clst.cls[chain_cls_nr].nbl.nb[chain_nb_nr].cnb,
		      cnr, -1, nnr, -1);
  assert(d>0);                                 /* there must be references */

  delete_nb_chn_def(chain_list, hnr);                      /* delete chain */

  deleted_refs = unref_indexes(clst, &deleted_nb);
  assert(deleted_refs>0);                      /* there must be references */

  if (vflag) {
    int_to_log(main_log, "", deleted_refs, " NB reference");
    string_to_log(main_log, deleted_refs==1?"":"s", " deleted", "\n");

    int_to_log(main_log, "", deleted_nb, " NB");
    string_to_log(main_log, deleted_nb==1?"":"s", " deleted", "\n");
  }
}



/* Erases document data from the DB.

Revision: 14.09.2002 E.Sack

Input : - pointer to document list
        - document number
Output: - document list updated
Postcondition: - document list may be empty
*/
static void
erase_tps_def(tps_lst *tpl, int tps_nr, GtkWidget *label) 
{
  int  last;
  char *ident = NULL;

  assert(tpl);

  if (!tpl->tps) {                                    /* consistency check */
    assert(!tpl->cnt);
    return;                                                     /* no data */
  }
  else {
    assert(tpl->cnt>0);
  }

  add_string(tpl->tps[tps_nr].ident, &ident);               /* save string */

  last = tpl->cnt-1;
  assert(tps_nr >= 0);
  assert(tps_nr <= last);

  if (tps_nr < last) {                               /* exchange with last */
    copy_tps_def(&tpl->tps[last], &tpl->tps[tps_nr], FALSE);
  }

  free_tps_def(&tpl->tps[last]);

  tpl->cnt--;                                   /* one document less in DB */

  if (tpl->cnt) {
    REALLOC(tps_def, tpl->cnt, tpl->tps);
  }
  else {
    free_tps_lst(tpl);                              /* all documents erased */
  }

  string_to_label(label, "TPS with ID '", ident, "' deleted");
  FREEMEM(ident);
}



/* Sorts a TPS.

Revision: 30.12.2005 E.Sack

Input : - length of class list
        - pointer to class list
        - Flag (0: sort by name, 1: sort by translation)
Output: - class list updated
        - flag (0: list unchanged, 1: list sorted)
Usage: The database is sorted directly.
*/
static int 
class_sort(int cnt, cls_def *cls, int what) 
{
  guint   len = 0;
  int     i, k, n, j, z, flag;
  int     *cdx = NULL;                                 /* class index list */
  cls_def *zcls = NULL;                            /* temporary class list */
  char    * max = NULL;             /* contains string with greatest value */
  char    **txt = NULL;

  assert(cls);

  if (cnt <= 1) {
    return 0;                                     /* one entry not to sort */
  }

  for (i = 0; i < cnt; i++) {                     /* maximal string length */
    len = XPMAX(len, strlen(cls[i].att.sign));
    len = XPMAX(len, strlen(cls[i].att.class));
  }

  CALLOC(char, len+1, max);

  CALLOC(char*, cnt, txt);                                   /* line count */
  for (i = 0; i < cnt; i++) {
    txt[i] = NULL;                                         /* zero pointer */
    CALLOC(char, len+1, txt[i]);
    txt[i][0] = '\0';
  }

  CALLOC(int, cnt, cdx);
  CALLOC(cls_def, cnt, zcls);

  for (i = 0; i < cnt; i++) {
    cdx[i] = i;                                          /* setup index list */

    strncpy(txt[i], what ? cls[i].att.sign : cls[i].att.class, len);

    init_cls_def(&zcls[i]);
    move_cls_def(&cls[i], &zcls[i], FALSE);               /* class buffering */
  }

  for (j = 0, z = -1; j < cnt-1; j++) {                    /* string sorting */
    strncpy(max, txt[j], len);
    k = j;

    for (i = j+1; i < cnt; i++) {
      if (strcmp(max, txt[i]) > 0) {
	z = cdx[i];
	strncpy(max, txt[i], len);
	k = i;
      }
    }
               
    if (k == j) {
      continue;
    }

    assert(z >= 0);
    cdx[k] = cdx[j];                                       /* exchange index */
    cdx[j] = z;
    strncpy(txt[k], txt[j], len);
    strncpy(txt[j], max, len);
  }                                                     /* index list sorted */

  for (i = 0; i < cnt; i++) { /* chain re-referencing in unsorted class list */
    for (n = 0; n < zcls[i].nbl.cnt; n++) {
      for (k = 0; k < zcls[i].nbl.nb[n].cnb.cnt; k++) {        /* CNB chains */
	for (j = 0; j < cnt; j++) {
	  if (zcls[i].nbl.nb[n].cnb.cd[k].cdx == cdx[j]) {    /* class index */
	    zcls[i].nbl.nb[n].cnb.cd[k].cdx = j;
	    break;
	  }
	}
      }

      for (k = 0; k < zcls[i].nbl.nb[n].nnb.cnt; k++) {        /* NNB chains */
	for (j = 0; j < cnt; j++) {
	  if (zcls[i].nbl.nb[n].nnb.cd[k].cdx == cdx[j]) {    /* class index */
	    zcls[i].nbl.nb[n].nnb.cd[k].cdx = j;
	    break;
	  }
	}
      }
    }
  }

  if (z == -1) {
    flag = 0;
    string_to_log(main_log, "class list already sorted\n", "", "");
  }
  else {
    flag = 1;
  }

  for (i = 0; i < cnt; i++) {                 /* copy back class list sorted */
    move_cls_def(&zcls[cdx[i]], &cls[i], FALSE);
  }

  FREEMEM(zcls);
  FREEMEM(cdx);
  FREEMEM(max);

  for (i = 0; i < cnt; i++) {
    FREEMEM(txt[i]);
  }
  FREEMEM(txt);

  return flag;
}



/* Extracts a TPS to a new DB.

Revision: 22.02.2004 E.Sack

Input : - DB
        - TPS index
Output: - TPS extracted
*/
static void
xextract_tps_def(db_def old, int tps_nr)
{
  int    err;
  db_def new;

  assert(tps_nr>=0 && tps_nr<old.tpl.cnt);
  assert(old.db_fname);

  err = check_name(old.db_fname);
  assert(!err);

  init_db_def(&new);

  new.tpl.cnt++;
  CALLOC(tps_def, new.tpl.cnt, new.tpl.tps);
  copy_tps_def(&old.tpl.tps[tps_nr], &new.tpl.tps[new.tpl.cnt-1], FALSE);

  cat_string(old.db_fname, TPSNAME_DEFAULT, "", &new.db_fname);
  int_to_string("", tps_nr, EXTRACT_DEFAULT, &new.db_fname);

  write_db_file(new);

  free_db_def(&new);
}



/* Extracts class definition to a new DB without NB Chain references.

Revision: 22.12.2005 E.Sack

Input : - DB
        - document definition number
        - class definition number
Output: - class definition extracted
*/
static void
xextract_cls_def(db_def old, int tps_nr, int cls_nr)
{
  db_def  new;
  cls_lst *old_clst = 0, *new_clst = 0;

  if (old.tpl.tps) {
    assert(old.tpl.cnt>0);
  }
  else {
    assert(!old.tpl.cnt);
  }

  assert(tps_nr>=0 && tps_nr<old.tpl.cnt);

  old_clst = &old.tpl.tps[tps_nr].cl;

  if (old_clst->cls) {
    assert(old_clst->cnt>0);
  }
  else {
    assert(!old_clst->cnt);
  }

  assert(cls_nr>=0 && cls_nr<old_clst->cnt);

  assert(old.db_fname);
  assert(strlen(old.db_fname));

  init_db_def(&new);

  new.tpl.cnt++;
  CALLOC(tps_def, new.tpl.cnt, new.tpl.tps);

  add_string(old.tpl.tps[tps_nr].ident, &new.tpl.tps[new.tpl.cnt-1].ident);

  copy_media_att(old.tpl.tps[tps_nr].media, 
		 &new.tpl.tps[new.tpl.cnt-1].media);

  new_clst = &new.tpl.tps[new.tpl.cnt-1].cl;

  new_clst->cnt++;
  CALLOC(cls_def, new_clst->cnt, new_clst->cls);

  copy_cls_def(&old_clst->cls[cls_nr], &new_clst->cls[new_clst->cnt-1],
	       TRUE);                                        /* no NB chains */

  cat_string(old.db_fname, TPSNAME_DEFAULT, "", &new.db_fname);
  int_to_string("", tps_nr, CLASS_NAME_DEFAULT, &new.db_fname);
  int_to_string(FILE_EXT, cls_nr, EXTRACT_DEFAULT, &new.db_fname);

  write_db_file(new);

  free_db_def(&new);
}



/* Extracts NB definition to a new DB.

Revision: 22.12.2005 E.Sack

Input : - DB
        - document definition number
        - class definition number
        - NB definition number
Output: - NB definition extracted
Postcondition: - class kernel not extracted
*/
static void
xextract_nb_def(db_def old, int tps_nr, int cls_nr, int nb_nr)
{
  db_def  new;
  tps_def *old_tps = 0, *new_tps = 0;
  cls_def *old_cls = 0;
  cls_lst *old_clst = 0, *new_clst = 0;
  nb_lst  *old_nbl = 0, *new_nbl = 0;

  assert(tps_nr>=0 && tps_nr<old.tpl.cnt);

  if (!old.tpl.cnt) {
    assert(!old.tpl.tps);
    string_to_log(main_log, "TPS list empty\n", "", "");
    return;
  }
  else {
    assert(old.tpl.cnt>0);
    assert(old.tpl.tps);
  }

  old_tps = &old.tpl.tps[tps_nr];

  if (!old_tps->cl.cls) {
    assert(!old_tps->cl.cnt);
    string_to_log(main_log, "TPS empty\n", "", "");
    return;
  }
  else {
    assert(old_tps->cl.cnt>0);
  }

  old_clst = &old_tps->cl;
  assert(cls_nr>=0 && cls_nr<old_clst->cnt);

  if (!old_clst->cls) {
    assert(!old_clst->cnt);
    string_to_log(main_log, "class list empty\n", "", "");
    return;
  }
  else {
    assert(old_clst->cnt>0);
  }

  old_cls = &old_clst->cls[cls_nr];

  if (!old_cls->nbl.nb) {
    assert(!old_cls->nbl.cnt);
    string_to_log(main_log, "class empty\n", "", "");
    return;
  }
  else {
    assert(old_cls->nbl.cnt>0);
  }

  old_nbl = &old_cls->nbl;
  assert(nb_nr>=0 && nb_nr<old_nbl->cnt);

  assert(old.db_fname);
  assert(strlen(old.db_fname));

  init_db_def(&new);

  new.tpl.cnt++;
  CALLOC(tps_def, new.tpl.cnt, new.tpl.tps);

  new_tps = &new.tpl.tps[new.tpl.cnt-1];

  add_string(old_tps->ident, &new_tps->ident);
  copy_media_att(old_tps->media, &new_tps->media);

  new_clst = &new_tps->cl;

  new_clst->cnt++;
  CALLOC(cls_def, new_clst->cnt, new_clst->cls);

  copy_cls_att(old_cls->att, &new_clst->cls[new_clst->cnt-1].att);

  new_nbl = &new_clst->cls[new_clst->cnt-1].nbl;

  new_nbl->cnt++;
  CALLOC(nb_def, new_nbl->cnt, new_nbl->nb);

  copy_nb_def(&old_nbl->nb[nb_nr], &new_nbl->nb[new_nbl->cnt-1], TRUE);/* NB looses all chains */

  cat_string(old.db_fname, TPSNAME_DEFAULT, "", &new.db_fname);
  int_to_string("", tps_nr, CLASS_NAME_DEFAULT, &new.db_fname);
  int_to_string(FILE_EXT, cls_nr, NBNAME_DEFAULT, &new.db_fname);
  int_to_string("", nb_nr, EXTRACT_DEFAULT, &new.db_fname);

  write_db_file(new);

  free_db_def(&new);
}



/* Adds a default empty class definition to a class list.

Revision: 14.02.2004 E.Sack

Input : - pointer to class list
        - pointer to class name
Output: - class list updated
        - index of class (-1: no class added)
Precondition: - class name may be NULL
Postcondition: - the class name is made unique.
*/
static int 
create_cls_def(cls_lst *clst, char *cname)
{
  int err, found;

  if (!clst->cls) {                                   /* consistency check */
    assert(!clst->cnt);
  }
  else {
    assert(clst->cnt>0);
  }

  err = unique_cls_name(*clst, &cname);          /* make class name unique */
  if (err) {
    return -1;
  }

  found = clst->cnt++;

  if (found) {                                              /* list exists */
    REALLOC(cls_def, clst->cnt, clst->cls);
  }
  else {                                             /* list doesn't exist */
    CALLOC(cls_def, clst->cnt, clst->cls);
  }

  init_cls_def(&clst->cls[found]);
  xdefault_cls_att(&clst->cls[found].att, cname);

  FREEMEM(cname);

  return found;
}



/* Creates an unique, non empty TPS ID.

Revision: 30.07.2005 E.Sack

Input : - class list
        - pointer to class name
Output: - class name updated
        - flag (0: string updated, 1: string not updated)
*/
static int
unique_cls_name(cls_lst clst, char **cname)
{
  int  d, len, err = 0, uni = 0;
  char *base = NULL, *ustr = NULL;

  assert(cname);

  add_string(*cname, &ustr);                                   /* copy input */

  d = check_name(ustr);

  if (d) {
    add_string(XPLAB_DEF_CLASS_NAME, &ustr);                      /* default */
  }

  add_string(ustr, &base);                                 /* initial string */

  len = strlen(base);

  while (1) {
    if (len >= MAX_LEN_CLASS_NAME) {                /* limited string length */
      int_to_log(main_log, "WARNING: class name length (", len+1, ") ");/*incl. LF */
      int_to_log(main_log, "exceeds maximum (", MAX_LEN_CLASS_NAME, ")\n");
      err = 1;
      break;
    }

    d = search_class(ustr, clst);          /* search string among the others */

    if (d < 0) {
      FREEMEM(*cname);
      add_string(ustr, cname);
      break;                                                /* unique string */
    }

    FREEMEM(ustr);                                           /* reset string */
    int_to_string(base, uni++, "", &ustr);                  /* create string */
    len = strlen(ustr);
  }

  FREEMEM(ustr);
  FREEMEM(base);

  return err;
}



/* Starts maintenance phase.

Revision: 17.08.2005 E.Sack

Input : - GTK convention
Output: - GTK convention
*/
static gboolean
cb_maintenance_phase(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN)
{
  gboolean filtered, raised;

  assert(widget);
  assert(user_data);

  filtered = filter_event_type(event, GDK_BUTTON_PRESS);

  if (filtered == FALSE) {
    return FALSE;                      /* propagate further event processing */
  }

  raised = xplab_window_raise(&mt.top_maintenance, 1);

  if (raised == FALSE) {
    free_maint_def(&mt);        /* must free here because of sudden run stop */
    mt.caller = (GtkWidget **)user_data;              /* last succeeded call */

    xplab_maintenance();                                       /* top window */
  }

  return TRUE;                      /* prevent from further event processing */
}



/* Maintenance phase.

Revision: 21.08.2005 E.Sack

Input : - pointer to caller window
Output: - maintenance dialog
*/
static void
xplab_maintenance(void)
{
  DATA_STATE db_state;

  mt.mpixbuf = ref_pixbufs();
  mt.mcombo_store = ref_combo_stores();
  mt.pc = default_pattern_color();

  xplab_release_db(&mt.db_key_read, 0);                         /* erase key */

  free_db_def(&mt.mdb);
  mt.mdb = xplab_request_db(NULL, &mt.db_key_read, XPLAB_DB_RO,
			    &mt.revert_button);                   /* DB copy */

  db_state = state_db_def(mt.mdb);

  if (db_state == STATE_IS_INIT) {
    xplab_text_insert(main_log, "WARNING: could not request database\n");
  }
  else {
    maintenance_window();                                   /* setup widgets */
  }
}



/* Creates top window for maintenance.

Revision: 30.07.2005 E.Sack

Input : - -
Output: - top window
*/
static void
maintenance_window(void)
{
  GtkWidget *vbox;

  /* set the maintenance top level widget */
  mt.top_maintenance = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(mt.top_maintenance), 
		       XPLAB_MAINTENANCE_TITLE_ITEM);
  gtk_window_set_default_size(GTK_WINDOW(mt.top_maintenance), 
			      XPLAB_MAINTENANCE_WINDOW_WIDTH, 
			      XPLAB_MAINTENANCE_WINDOW_HEIGHT);

  /* put into top level widget a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(mt.top_maintenance), vbox);

  /* put into vertical box a menubar */
  maintenance_menu_bar(vbox);

  /* put into vertical box a scrolled window */
  maintenance_scrolled_window(vbox);

  /* put into vertical box a label */
  mt.mlabel = gtk_label_new(NULL);

  gtk_box_pack_start(GTK_BOX(vbox), mt.mlabel, FALSE, FALSE, 0);

  g_signal_connect(G_OBJECT(mt.top_maintenance), "delete-event", 
		   G_CALLBACK(cb_maintenance_top_delete), NULL);

  xplab_set_window_icon(mt.top_maintenance);                  /* window icon */

  maintenance_statusbar_default();

  gtk_widget_show_all(mt.top_maintenance);
}



/* Builds a menu bar.

Revision: 30.07.2005 E.Sack

Input : - container
Output: - container updated
*/
static void
maintenance_menu_bar(GtkWidget *vbox)
{
  GtkWidget *menu_bar;

  menu_bar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox), menu_bar, FALSE, FALSE, 0);

  maintenance_menu_entry_file(menu_bar);                      /* menu file   */
  maintenance_menu_entry_tree(menu_bar);                      /* menu tree   */
  maintenance_menu_entry_tools(menu_bar);                     /* menu tools  */
  maintenance_menu_entry_save(menu_bar);                      /* menu save   */
  maintenance_menu_entry_revert(menu_bar);                    /* menu revert */
  maintenance_menu_entry_image(menu_bar);                     /* menu image  */
  xplab_menu_help(menu_bar, &mt.top_maintenance);             /* menu help   */
}



/* Builds an entry for a menu bar.

Revision: 30.07.2005 E.Sack

Input : - menu bar
Output: - menu bar updated
*/
static void
maintenance_menu_entry_file(GtkWidget *menu_bar)
{
  GtkWidget *file_entry, *file_menu, *sep;

  file_entry = gtk_menu_item_new_with_label(MENU_BUTTON_FILE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_entry);

  file_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_entry), file_menu);

  xplab_preferences_menu(file_menu, &mt.top_maintenance);  /*item preferences*/
  xplab_selection_menu(file_menu, &mt.top_maintenance);    /* item selection */

  sep = gtk_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), sep);

  maintenance_menu_item_revert_quit(file_menu);        /* item revert & quit */
  maintenance_menu_item_save_quit(file_menu);            /* item save & quit */
  maintenance_menu_item_quit(file_menu);                        /* item quit */

  g_signal_connect(G_OBJECT(file_entry), "activate",
		   G_CALLBACK(cb_maintenance_sensitivity), NULL);
}



/* Builds an entry for a menu bar.

Revision: 30.07.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
static void
maintenance_menu_entry_tools(GtkWidget *menu_bar)
{
  GtkWidget *tools_entry, *tools_menu;

  tools_entry = gtk_menu_item_new_with_label(MENU_BUTTON_TOOLS);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), tools_entry);

  tools_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tools_entry), tools_menu);

  maintenance_menu_consistency(tools_menu);              /* item consistency */
  mt.where_referenced = maintenance_menu_where_referenced(tools_menu);/* item referenced  */
  maintenance_menu_item_extract(tools_menu);             /* item extract     */
  maintenance_menu_item_dump(tools_menu);                /* item dump        */
  maintenance_menu_item_edit(tools_menu);                /* item edit        */
  maintenance_menu_item_ccf(tools_menu);                 /* item CCF         */

  g_signal_connect(G_OBJECT(tools_entry), "activate",
		   G_CALLBACK(cb_maintenance_sensitivity), NULL);
}



/* Builds an entry for a menu bar.

Revision: 30.07.2005 E.Sack

Input : - menu bar
Output: - menu bar updated
*/
static void
maintenance_menu_entry_tree(GtkWidget *menu_bar)
{
  GtkWidget *tree_entry, *tree_menu;

  tree_entry = gtk_menu_item_new_with_label(MENU_BUTTON_TREE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), tree_entry);

  tree_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tree_entry), tree_menu);

  maintenance_menu_item_lock(tree_menu);                    /* item lock     */
  maintenance_menu_item_collapse(tree_menu);                /* item collapse */
  maintenance_menu_item_expand(tree_menu);                  /* item expand   */
  maintenance_menu_item_pattern(tree_menu);                 /* item pattern  */

  g_signal_connect(G_OBJECT(tree_entry), "activate",
		   G_CALLBACK(cb_maintenance_sensitivity), NULL);
}



/* Checks menu entry.

Revision: 04.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_maintenance_sensitivity(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gint              pstatus;
  gboolean          node_ok;
  GtkTreePath       *path = NULL;
  GtkTreeViewColumn *fcolumn;

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  if (GTK_IS_WIDGET(mt.top_maintenance) == TRUE) {
    if (GTK_IS_TREE_VIEW(mt.treeview) == TRUE) {
      pstatus = mt.items_deleted + mt.items_created + mt.items_changed;

      if (pstatus) {                             /* there are changed items */
	gtk_widget_set_sensitive(mt.revert_button, TRUE);
	gtk_widget_set_sensitive(mt.revert_quit_button, TRUE);
	gtk_widget_set_sensitive(mt.quit_button, FALSE);
	gtk_widget_set_sensitive(mt.lock_button, FALSE);
      }
      else {                                  /* only unchanged/saved items */
	gtk_widget_set_sensitive(mt.revert_quit_button, FALSE);
	gtk_widget_set_sensitive(mt.revert_button, FALSE);
	gtk_widget_set_sensitive(mt.quit_button, TRUE);
	gtk_widget_set_sensitive(mt.lock_button, TRUE);
      }

      gtk_tree_view_get_cursor(GTK_TREE_VIEW(mt.treeview), &path, &fcolumn);

      gtk_widget_set_sensitive(mt.tree_menu_delete,
			       gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button)));
      gtk_widget_set_sensitive(mt.ccf_menu,
			       gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button)) == TRUE?FALSE:TRUE);

      node_ok = (path == NULL) ? FALSE : TRUE;
      gtk_widget_set_sensitive(mt.collapse_selected, node_ok);
      gtk_widget_set_sensitive(mt.expand_selected, node_ok);

      gtk_widget_set_sensitive(mt.row_extract, FALSE);             /* preset */
      gtk_widget_set_sensitive(mt.row_dump, FALSE);                /* preset */
      gtk_widget_set_sensitive(mt.row_edit, FALSE);                /* preset */
      gtk_widget_set_sensitive(mt.where_referenced, FALSE);        /* preset */
      maintenance_row_edit(XPLAB_QUERY_SENSITIVITY);                  /* set */

      maintenance_row_edit(XPLAB_QUERY_ACTION_DUMP);               /* update */

      ccf_sensitivity();
    }
  }

  gtk_tree_path_free(path);
}



/* Builds an entry for a menu bar.

Revision: 30.07.2005 E.Sack

Input : - menu bar
Output: - menu bar updated
*/
static void
maintenance_menu_entry_revert(GtkWidget *menu_bar)
{
  mt.revert_button = gtk_menu_item_new_with_label(DIALOG_BUTTON_REVERT);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), mt.revert_button);

  g_signal_connect(G_OBJECT(mt.revert_button), "event", 
		   G_CALLBACK(cb_maintenance_revert_event), NULL);

  g_signal_connect(G_OBJECT(mt.revert_button), "activate",
		   G_CALLBACK(cb_maintenance_revert), NULL);

  gtk_widget_set_sensitive(mt.revert_button, FALSE);
}



/* Builds an entry for a menu bar.

Revision: 07.01.2006 E.Sack

Input : - menu bar
Output: - menu bar updated
*/
static void
maintenance_menu_entry_image(GtkWidget *menu_bar)
{
  GtkWidget *image_entry, *image_menu;

  image_entry = gtk_menu_item_new_with_label(MENU_BUTTON_IMAGE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), image_entry);

  image_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(image_entry), image_menu);

  maintenance_menu_image_on(image_menu);
  maintenance_menu_image_static(image_menu);
}



/* Adds an item to a menu.

Revision: 07.01.2006 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
maintenance_menu_image_on(GtkWidget *menu)
{
  mt.image_on = gtk_check_menu_item_new_with_label(MENU_BUTTON_IMAGE_ON);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), mt.image_on);

  g_signal_connect(G_OBJECT(mt.image_on), "toggled", 
		   G_CALLBACK(cb_maintenance_image_on), NULL);
}



/* Adds an item to a menu.

Revision: 07.01.2006 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
maintenance_menu_image_static(GtkWidget *menu)
{
  mt.image_static = gtk_check_menu_item_new_with_label(MENU_BUTTON_IMAGE_STATIC);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), mt.image_static);
}



/* Builds an entry for a menu bar.

Revision: 30.07.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
static void
maintenance_menu_entry_save(GtkWidget *menu_bar)
{
  GtkWidget *save_button;

  save_button = gtk_menu_item_new_with_label(DIALOG_BUTTON_SAVE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), save_button);

  g_signal_connect(G_OBJECT(save_button), "event", 
		   G_CALLBACK(cb_maintenance_save), NULL);
}



/* Triggers image window build.

Revision: 07.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_maintenance_image_on(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  gboolean on;

  STSFY_GCC(user_data);

  on = gtk_check_menu_item_get_active(checkmenuitem);

  if (on == TRUE) {
    assert(mt.mimg == NULL);                            /* consistency check */
  }

  maintenance_image_on(on);
}



/* Sets write access to DB.

Revision: 04.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_maintenance_lock(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  int err;

  STSFY_GCC(user_data);

  if (gtk_check_menu_item_get_active(checkmenuitem) == TRUE) {  /* reread DB */
    db_def db_tmp;

    db_tmp = xplab_request_db(XPLAB_MAINTENANCE_TITLE_ITEM,       /* lock DB */
			      &mt.db_key_lock, XPLAB_DB_RW, NULL);

    err = check_name(db_tmp.db_fname);
    if (err) {
      xplab_text_insert(main_log, "WARNING: could not request database\n");
      gtk_check_menu_item_set_active(checkmenuitem, FALSE);
    }
    else {                                               /* renew tree store */
      copy_db_def(&db_tmp, &mt.mdb, FALSE);                       /* DB copy */
      maintenance_tree_store_fill(mt.treeview, FALSE);
    }

    free_db_def(&db_tmp);
  }
  else {
    maintenance_reset_parse_status();

    xplab_release_db(&mt.db_key_lock, 1);                       /* unlock DB */
    xplab_rebuild_db_tree();
  }

  dialog_delete_all();
}



/* Updates maintenance status bar.

Revision: 01.04.2003 E.Sack

Input : - -
Output: - maintenance status updated
*/
static void
maintenance_statusbar_default(void)
{
  if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button)) == TRUE) {
    string_to_label(mt.mlabel, "database '", mt.mdb.db_fname,
		    "' opened for write");
  }
  else {
    string_to_label(mt.mlabel, "database '", mt.mdb.db_fname, 
		    "' opened for read only");
  }
}



/* Deletes all dialogs.

Revision: 08.01.2006 E.Sack

Input : - -
Output: - dialogs deleted
*/
static void
dialog_delete_all(void)
{
  if (GTK_IS_OBJECT(mt.dialog.ccf.edt.cancel_button)) {
    g_signal_emit_by_name((gpointer)mt.dialog.ccf.edt.cancel_button, 
			  "clicked");
  }
}



/* Sets tree node icon. Sets the node icon according to its expanded/collapsed
   state.

Revision: 17.12.2005 E.Sack

Input : - GTK convention
Output: - GTK convention
*/
static gboolean
set_node_icon(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN)
{
  GdkPixbuf *icon;

  if (gtk_tree_model_iter_has_child(model, iter) == TRUE) {     /* node only */
    assert(data);
    assert(GTK_IS_TREE_VIEW(data) == TRUE);
    assert(GTK_IS_TREE_STORE(model) == TRUE);

    icon = node_icon(GTK_TREE_VIEW(data), model, path, iter);

    gtk_tree_store_set(GTK_TREE_STORE(model), iter, 
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC, icon,
		       -1);                        /* Pixbuf to unref by GTK */
  }

  return FALSE;                                            /* treat all rows */
}



/* Chooses the node icon. 

Usage: Used for nodes only, if used for leafs, icons may be unwanted.

Revision: 17.12.2005 E.Sack

Input : - pointer to tree view
        - pointer to model
        - pointer to path
        - pointer to iter
Output: - pointer to icon
*/
static GdkPixbuf *
node_icon(GtkTreeView *tview, GtkTreeModel *model, GtkTreePath *path,
	    GtkTreeIter *iter)
{
  gint      node_id;
  gboolean  locked, expanded;
  GdkPixbuf *open, *closed, *icon;

  assert(tview);
  assert(model);
  assert(path);
  assert(iter);

  gtk_tree_model_get(model, iter, 
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

  expanded = gtk_tree_view_row_expanded(tview, path);

  if (node_id == XPLAB_MAINTENANCE_NODE_ID_DB_ATTRIBUTES) {
    open = mt.mpixbuf.open_read;         /* node is neither editable nor has */
    closed = mt.mpixbuf.closed_read;     /* editable children.               */
  }
  else {
    locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

    if (locked == TRUE) {
      open = mt.mpixbuf.open_write;
      closed = mt.mpixbuf.closed_write;
    }
    else {
      open = mt.mpixbuf.open_read;
      closed = mt.mpixbuf.closed_read;
    }
  }

  if (expanded == TRUE) {
    icon = open;
  }
  else {
    icon = closed;
  }

  return icon;
}



/* Sets tree NB icons.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - GTK convention
*/
static gboolean
set_nb_icon(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN)
{
  gint        node_id, nb_idx, cls_idx, tps_idx, row_height;
  gboolean    do_unref = FALSE, is_icon, is_orig, is_scal, is_x16, is_x32;
  GdkPixbuf   *set_pixbuf, *icon_pixbuf = mt.mpixbuf.unedit;
  GtkTreeIter titer;

  STSFY_GCC(path);
  STSFY_GCC(data);

  gtk_tree_model_get(model, iter, XPLAB_MAINTENANCE_TREE_STORE_NODE_ID,
		     &node_id, -1);

  if (node_id != XPLAB_MAINTENANCE_NODE_ID_PATTERN) {
    return FALSE;
  }

  /* scan radio buttons                                                      */
  is_icon = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.icon_button));
  is_orig = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.orig_button));
  is_scal = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.scale_button));
  is_x16 = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.x16_button));
  is_x32 = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.x32_button));

  if (is_icon == TRUE) {
    row_height = XPM_PATTERN_VALU_ICON;
    set_pixbuf = icon_pixbuf;                     /* icon instead of pattern */
    do_unref = FALSE;        /* ref'd multiple, so we need a global to unref */
  }
  else {                                          /* pattern instead of icon */
    if (is_orig == TRUE || is_scal == TRUE) {
      if (is_x32 == TRUE) {
	row_height = XPM_PATTERN_VALU_32x32;
      }
      else if (is_x16 == TRUE) {
	row_height = XPM_PATTERN_VALU_16x16;
      }
      else {
	row_height = XPM_PATTERN_VALU_ICON;
      }
    }
    else if (is_x16 == TRUE) {
      row_height = XPM_PATTERN_VALU_16x16;
    }
    else if (is_x32 == TRUE) {
      row_height = XPM_PATTERN_VALU_32x32;
    }
    else {                                   /* don't exaggerate iconisation */
      assert(0);                                          /* unknown request */
    }

    nb_idx = child_has_nb_index(model, *iter, &titer);        /* get indexes */
    cls_idx = child_has_class_index(model, *iter, &titer);
    tps_idx = child_has_tps_index(model, *iter, &titer);
    assert(nb_idx >= 0);
    assert(cls_idx >= 0);
    assert(tps_idx >= 0);

    set_pixbuf = pixbuf_nb(row_height, tps_idx, cls_idx, nb_idx);
    do_unref = TRUE;                  /* individual NB icons only once ref'd */
  }

  assert(GTK_IS_TREE_STORE(model) == TRUE);

  gtk_tree_store_set(GTK_TREE_STORE(model), iter, 
		     XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC, set_pixbuf, -1);

  if (do_unref == TRUE) {
    g_object_unref(set_pixbuf);          /* unref here as there is no global */
  }

  return FALSE;                                            /* treat all rows */
}



/* Creates an icon from a NB.

Revision: 30.07.2005 E.Sack

Input : - row height
        - TPS index
        - class index
        - NB index
Output: - pointer to pixbuf
*/
static GdkPixbuf *
pixbuf_nb(int row_height, int tps_idx, int cls_idx, int nb_idx)
{
  gint      hext, wext, width, height;
  pic_def   img_tmp1, img_tmp2, scut;
  text_def  xpm;
  area_def  maxar;
  GdkPixbuf *set_pixbuf;                                   /* preset default */ 

  init_pic_def(&img_tmp1);
  init_pic_def(&img_tmp2);

  assert(row_height > 0);

  maxar = max_cls_area(mt.mdb.tpl.tps[tps_idx].cl);          /* max. NB size */

  scut = mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].pat;

  if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.scale_button)) == FALSE) {    /* crop NB */
    width = XPMIN(row_height, maxar.width);            /* limit NB to square */
    height = XPMIN(row_height, maxar.height);

    img_tmp2.ar.width = XPMIN(scut.ar.width, width);
    img_tmp2.ar.height = XPMIN(scut.ar.height, height);

    CALLOC(PICT, AREA_SIZE(img_tmp2.ar), img_tmp2.mem);   /* start with left */
    copy_bmps(scut, img_tmp2, img_tmp2.ar, 0, 0, 0, 0);   /* upper corner    */
  }
  else {                                                         /* scale NB */
    double vs;

    if (maxar.width >= maxar.height) {
      assert(maxar.width > 0.0);
      vs = (double)row_height/(double)maxar.width;
      width = row_height;                                      /* w' = v * w */
      height = (int)(vs*(double)maxar.height);                 /* h' = v * h */
    }
    else {
      assert(maxar.height > 0.0);
      vs = (double)row_height/(double)maxar.height;
      width = (int)(vs*(double)maxar.width);                   /* w' = v * w */
      height = row_height;                                     /* h' = v * h */
    }

    img_tmp2 = scale_bmp(scut, vs);                              /* scale NB */
  }

  wext = XPMAX(0, width - img_tmp2.ar.width);           /* maximal extension */
  wext = wext - (wext>>1);

  hext = XPMAX(0, height - img_tmp2.ar.height);
  hext = hext - (hext>>1);

  img_tmp1 = xextend_image(img_tmp2, wext, hext);  /* extended symmetrically */
  free_pic_def(&img_tmp2);

  img_tmp2.ar.width = width;        /* may be 1 pixel greater, therefore cut */
  img_tmp2.ar.height = height;                        /* needed for coloring */
  CALLOC(PICT, AREA_SIZE(img_tmp2.ar), img_tmp2.mem);
  copy_bmps(img_tmp1, img_tmp2, img_tmp2.ar, 0, 0, 0, 0);
  free_pic_def(&img_tmp1);

  xpm = bmp_to_xpm(img_tmp2, row_height);

  set_pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)xpm.tl);

  free_text_def(&xpm);
  free_pic_def(&img_tmp2);

  return set_pixbuf;
}



/* Converts BMP to squared XPM.

Revision: 01.02.2004 E.Sack

Input : - image
        - square size
Output: - XPM string array
Usage:  - should be used only for small images
*/
static text_def
bmp_to_xpm(pic_def pattern, int square)
{
  int      colnum, x, y, hext, wext, idx, len = AREA_SIZE(pattern.ar);
  PICT     *bmp = pattern.mem;
  text_def xpm;

  init_text_def(&xpm);

  assert(len > 0);

  wext = XPMAX(0, square - pattern.ar.width) >> 1;
  hext = XPMAX(0, square - pattern.ar.height) >> 1;

  colnum = 3;                                          /* number of colors */
  xpm.cnt = 1 + colnum;
  CALLOC(char *, xpm.cnt, xpm.tl);
  xpm.tl[0] = xpm.tl[1] = xpm.tl[2] = xpm.tl[3] = NULL;

  /* image spec in first string */
  int_to_string("", square, " ", &xpm.tl[0]);        /* square icon format */
  int_to_string("", square, " ", &xpm.tl[0]);
  int_to_string("", colnum, " 1", &xpm.tl[0]);  /* colors, chars per pixel */

  /* image colors in successive strings */
  add_string(XPM_ICON_BGND_CHAR " c #" XPM_COLOR_ICON_BGND, &xpm.tl[1]);
  add_string(XPM_NB_BGND_CHAR   " c #" XPM_NB_BGND_COLOR,   &xpm.tl[2]);
  add_string(XPM_NB_FGND_CHAR   " c #" XPM_NB_FGND_COLOR,   &xpm.tl[3]);

  for (y = 0; y < square; y++) { /* image definition in successive strings */
    idx = xpm.cnt++;
    REALLOC(char *, xpm.cnt, xpm.tl);
    xpm.tl[idx] = NULL;

    if (y < hext || y >= pattern.ar.height+hext) {
      for (x = 0; x < square; x++) {
	add_string(XPM_ICON_BGND_CHAR, &xpm.tl[idx]);
      }
    }
    else {
      for (x = 0; x < wext; x++) {
	add_string(XPM_ICON_BGND_CHAR, &xpm.tl[idx]);
      }

      for (; x < pattern.ar.width+wext; x++) {
	add_string((*bmp++) == BACKGROUND?XPM_NB_BGND_CHAR: XPM_NB_FGND_CHAR, 
		   &xpm.tl[idx]);
      }

      for (; x < square; x++) {
	add_string(XPM_ICON_BGND_CHAR, &xpm.tl[idx]);
      }
    }
  }

  return xpm;
}



/* Deletes top window.

Revision: 05.12.2003 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
gboolean
cb_maintenance_top_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);

  assert(event);
  assert(widget);

  cb_maintenance_quit(NULL, NULL);                      /* event to signal */

  return FALSE;
}



/* Receives a node selection and starts the appropriate edit dialog.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_maintenance_row_select(GTK_CONVENTION_TREEVIEW_CURSOR_CHANGED_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(treeview);

  cb_maintenance_sensitivity(NULL, NULL);            /* direct handler call */
}



/* Receives a node expansion/collapse.

Revision: 17.12.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_maintenance_row_expanded_collapsed(GTK_CONVENTION_TREEVIEW_ROW_EXPANDED_COLLAPSED_RTS_VOID)
{
  GtkTreeModel *model;

  STSFY_GCC(user_data);
  STSFY_GCC(arg1);
  STSFY_GCC(arg2);

  assert(treeview);
  assert(GTK_IS_TREE_VIEW(treeview) == TRUE);

  model = gtk_tree_view_get_model(treeview);

  gtk_tree_model_foreach(model, (GtkTreeModelForeachFunc)set_node_icon,
			 (gpointer)treeview);
}



/* Edits node.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_row_edit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  assert(user_data);
  assert(menuitem);

  maintenance_row_edit(GPOINTER_TO_INT(user_data));
}



/* Checks nodes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_ccf_setting(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(togglebutton);

  if (gtk_toggle_button_get_active(togglebutton) == TRUE) {
    ccf_sensitivity();
  }
}



/* Edits node.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_ccf(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(menuitem);

  mt.dialog.ccf.edt.eset.func = edit_func_ccf;

  dialog_row_edit(&mt.dialog.ccf.edt, XPLAB_EDIT_DIALOG);
}



/* Creates a tools dialog for correlation.

Revision: 06.03.2004 E.Sack

Input : - pointer to dialog settings
Output: - -
*/
static void
edit_func_ccf(XPLAB_MAINTENANCE_EDIT_IF)
{
  GtkWidget *hbox;

  STSFY_GCC(piter);

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    mt.dialog.ccf.cbt = ccf_setting(fbox);
    ccf_sensitivity();

    free_ccf_mat(&mt.dialog.ccf.matrix);   /* free old CCF data (DB editing) */

    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gtk_table_resize(GTK_TABLE(fbox), 1, 2);
      gtk_table_set_homogeneous(GTK_TABLE(fbox), TRUE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_ccf_dump),
			       DIALOG_BUTTON_DUMP, DESC1_MAINTENANCE_CCF_DUMP, 
			       DESC2_MAINTENANCE_CCF_DUMP, 
			       (gpointer)&mt.dialog.ccf.edt);
      mt.dialog.ccf.cbt.dump_id = gd.id;
      mt.dialog.ccf.cbt.dump_stop_button = gd.button;

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox,
		       0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

      mt.dialog.ccf.cbt.stop_id = 
	g_signal_connect(G_OBJECT(mt.dialog.ccf.cbt.dump_stop_button), 
			 "clicked", G_CALLBACK(cb_dialog_ccf_stop), 
			 (gpointer)&mt.dialog.ccf.edt);

      /* block stop handler */
      g_signal_handler_block((gpointer)mt.dialog.ccf.cbt.dump_stop_button, 
			     mt.dialog.ccf.cbt.stop_id);

      hbox = gtk_hbox_new(FALSE, 0);                   /* set a progress bar */

      mt.dialog.ccf.cbt.crun.pbar = xplab_progress(hbox);
      mt.dialog.ccf.cbt.crun.pactive = &mt.cactive;
      mt.dialog.ccf.cbt.crun.pstop = &mt.cstop;
      gtk_table_attach(GTK_TABLE(fbox), hbox, 
		       1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

      mt.cancel_hook_func = GTK_SIGNAL_FUNC(cb_dialog_ccf_stop);

      mt.cancel_hook_data = (gpointer)&mt.dialog.ccf.edt;

      gtk_window_set_modal(GTK_WINDOW(mt.dialog.ccf.edt.edialog), FALSE);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "CCF");
    break;

  case XPLAB_EDIT_ACTION_STOP:
    stop_run_def(&mt.dialog.ccf.cbt.crun);
    break;

  case XPLAB_EDIT_ACTION_DUMP:
    if (*mt.dialog.ccf.cbt.crun.pactive == TRUE) {      /* already running */
      return;
    }

    start_run_def(&mt.dialog.ccf.cbt.crun);                       /* start */
    dump_ccf(&mt.dialog.ccf.cbt.crun);
    stop_run_def(&mt.dialog.ccf.cbt.crun);                     /* finished */

    break;

  case XPLAB_QUERY_SENSITIVITY:
    break;

  default: assert(0); break;
  }

  return;
}



/* Dumps CCF features with run-time control.

Revision: 04.01.2006 E.Sack

Input : - pointer to control structure
Output: - features dumped
*/
static void
dump_ccf(run_def *crun)
{
  gint         node_id, tps_idx, cls_idx, sort;
  GtkTreeIter  iter, titer;
  GtkTreeModel *cmodel;

  /* set button to stop */
  xplab_label_set_text(GTK_BIN(mt.dialog.ccf.cbt.dump_stop_button)->child,
		       SINGLE_BUTTON_STOP);

  gtk_button_leave(GTK_BUTTON(GTK_OBJECT(mt.dialog.ccf.cbt.dump_stop_button)));/*just for display taste*/

  /* block dump handler */
  g_signal_handler_block((gpointer)mt.dialog.ccf.cbt.dump_stop_button, 
			 mt.dialog.ccf.cbt.dump_id);

  /* unblock stop handler */
  g_signal_handler_unblock((gpointer)mt.dialog.ccf.cbt.dump_stop_button,
			   mt.dialog.ccf.cbt.stop_id);

  sort = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mt.dialog.ccf.cbt.sort_cv_button)) == TRUE;

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));
  node_id = cursor_get_node_id(GTK_TREE_VIEW(mt.treeview), &iter);

  switch (node_id) {
  case XPLAB_MAINTENANCE_NODE_ID_CLASS_INDEX:
    assert(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mt.dialog.ccf.cbt.intra_button)) == TRUE);
    cls_idx = child_has_class_index(cmodel, iter, &titer);
    tps_idx = child_has_tps_index(cmodel, iter, &titer);

    if (cls_idx >= 0 && tps_idx >= 0) {
      renew_ccf(&mt.dialog.ccf.matrix, tps_idx, cls_idx);
      gen_intra_corr(&mt.dialog.ccf.matrix, mt.mdb, tps_idx, cls_idx, sort, crun);
    }

    break;

  case XPLAB_MAINTENANCE_NODE_ID_TPS_INDEX:
    assert(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mt.dialog.ccf.cbt.inter_button)) == TRUE);
    tps_idx = child_has_tps_index(cmodel, iter, &titer);

    if (tps_idx >= 0) {
      renew_ccf(&mt.dialog.ccf.matrix, tps_idx, -1);
      gen_inter_corr(&mt.dialog.ccf.matrix, mt.mdb, tps_idx, sort, crun);
    }

    break;

  default: break;
  }

  /* set button to dump assuming button and signals are living consistently */
  if (GTK_IS_LABEL(GTK_BIN(mt.dialog.ccf.cbt.dump_stop_button)->child) == TRUE) {
    xplab_label_set_text(GTK_BIN(mt.dialog.ccf.cbt.dump_stop_button)->child,
		         DIALOG_BUTTON_DUMP);

    /* block stop handler */
    g_signal_handler_block((gpointer)mt.dialog.ccf.cbt.dump_stop_button, 
			   mt.dialog.ccf.cbt.stop_id);

    /* unblock dump handler */
    g_signal_handler_unblock((gpointer)mt.dialog.ccf.cbt.dump_stop_button,
			     mt.dialog.ccf.cbt.dump_id);
  }
}



/* Frees candidate list if indexes have been changed.

Revision: 06.01.2006 E.Sack

Input : - pointer to correlation features
        - TPS index
        - class index
Output: - candidate list updated
*/
static void
renew_ccf(ccf_mat *corr_features, int tnr, int cnr)
{
  assert(corr_features);

  if (tnr < 0 || cnr < 0) {
    return;
  }

  if (corr_features->tps_idx != tnr || corr_features->cls_idx != cnr) {
    free_ccf_mat(corr_features);
  }
}



/* Creates the edit dialog according selected row.

Revision: 30.07.2005 E.Sack

Input : - action
Output: - dialog created

*/
static void
maintenance_row_edit(int do_action)
{
  gint              node_id;
  gboolean          is_path;
  GtkTreePath       *path;
  GtkTreeIter       piter;
  GtkTreeModel      *cmodel;
  GtkTreeViewColumn *fcolumn;

  mt.dialog.edt.eset.func = NULL;

  if (GTK_IS_TREE_VIEW(mt.treeview) == FALSE) {
    return;
  }

  gtk_tree_view_get_cursor(GTK_TREE_VIEW(mt.treeview), &path, &fcolumn);

  if (path == NULL) {
    return;
  }

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));
  is_path = gtk_tree_model_get_iter(cmodel, &piter, path);
  gtk_tree_path_free(path);

  assert(is_path == TRUE);

  gtk_tree_model_get(cmodel, &piter,
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

  mt.dialog.edt.eset.piter = piter;

  switch (node_id) {                              /* call appropriate dialog */
  case XPLAB_MAINTENANCE_NODE_ID_TPS_LIST:
    mt.dialog.edt.eset.func = edit_func_tps_lst;
  break;

  case XPLAB_MAINTENANCE_NODE_ID_TPS_INDEX:
    mt.dialog.edt.eset.func = edit_func_tps_def;
  break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_LIST:
    mt.dialog.edt.eset.func = edit_func_cls_lst;
  break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_INDEX:
    mt.dialog.edt.eset.func = edit_func_cls_def;
  break;

  case XPLAB_MAINTENANCE_NODE_ID_NB_LIST:
    mt.dialog.edt.eset.func = edit_func_nb_lst;
  break;

  case XPLAB_MAINTENANCE_NODE_ID_NB_INDEX:
    mt.dialog.edt.eset.func = edit_func_nb_def;
  break;

  case XPLAB_MAINTENANCE_NODE_ID_CNB_LIST:             /* contained in a NNB */
    mt.dialog.edt.eset.func = edit_func_cnb_lst;
  break;

  case XPLAB_MAINTENANCE_NODE_ID_CNB_INDEX:            /* contained in a NNB */
    mt.dialog.edt.eset.func = edit_func_cnb_def;
  break;

  case XPLAB_MAINTENANCE_NODE_ID_NNB_LIST:             /* contained in a CNB */
    mt.dialog.edt.eset.func = edit_func_nnb_lst;
  break;

  case XPLAB_MAINTENANCE_NODE_ID_NNB_INDEX:            /* contained in a CNB */
    mt.dialog.edt.eset.func = edit_func_nnb_def;
  break;

  /* following IDs can be edited only in their tree view row */
  case XPLAB_MAINTENANCE_NODE_ID_DPI:
  case XPLAB_MAINTENANCE_NODE_ID_LINE_METHOD:
  case XPLAB_MAINTENANCE_NODE_ID_L_PHASE:
  case XPLAB_MAINTENANCE_NODE_ID_H_PHASE:
  case XPLAB_MAINTENANCE_NODE_ID_CLASS_ATTRIBUTES:
  case XPLAB_MAINTENANCE_NODE_ID_CLASS_NAME:
  case XPLAB_MAINTENANCE_NODE_ID_CLASS_TRANSLATION:
  case XPLAB_MAINTENANCE_NODE_ID_CLASS_COMMENT:
  case XPLAB_MAINTENANCE_NODE_ID_CLASS_ALIGNMENT:
  case XPLAB_MAINTENANCE_NODE_ID_CLASS_PRINTABLE:
  case XPLAB_MAINTENANCE_NODE_ID_CLASS_FIXED:
  case XPLAB_MAINTENANCE_NODE_ID_TPS_ID:
    break;

  /* following IDs cannot be edited */
  case XPLAB_MAINTENANCE_NODE_ID_CHAIN_CLASS_INDEX:
  case XPLAB_MAINTENANCE_NODE_ID_CHAIN_NB_INDEX:
  case XPLAB_MAINTENANCE_NODE_ID_CHAIN_X_OFFSET:
  case XPLAB_MAINTENANCE_NODE_ID_CHAIN_Y_OFFSET:
  case XPLAB_MAINTENANCE_NODE_ID_MEDIA_ATTRIBUTES:
  case XPLAB_MAINTENANCE_NODE_ID_PERIOD_LENGTH:
  case XPLAB_MAINTENANCE_NODE_ID_DB_ATTRIBUTES:
  case XPLAB_MAINTENANCE_NODE_ID_DB_NAME:
  case XPLAB_MAINTENANCE_NODE_ID_DB_VERSION:
  case XPLAB_MAINTENANCE_NODE_ID_PATTERN:
  case XPLAB_MAINTENANCE_NODE_ID_OPOS:
  break;

  default: assert(0); break;
  }

  if (mt.dialog.edt.eset.func) {
    dialog_row_edit(&mt.dialog.edt, do_action);
  }
}



/* Ensures node ID.

Revision: 30.07.2005 E.Sack

Input : - pointer to  model
        - iter
        - ID
Output: - -
*/
static void
assert_node_id(GtkTreeModel *cmodel, GtkTreeIter iter, int id_query)
{
  gint node_id;

  gtk_tree_model_get(cmodel, &iter,
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

  assert(id_query == node_id);
}



/* Quits maintenance phase. 

Revision: 20.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_maintenance_quit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  if (GTK_IS_WIDGET(mt.top_maintenance) == TRUE) {
    xplab_release_db(&mt.db_key_lock, 1);                       /* unlock DB */
    xplab_release_db(&mt.db_key_read, 0);                       /* erase key */
    dialog_delete_all();
    maintenance_image_on(FALSE);                     /* delete image windows */

    xplab_window_raise(mt.caller, xctrl.cfg.raise);
    xplab_widget_destroy(&mt.top_maintenance);          /* includes treeview */

    unref_combo_stores(mt.mcombo_store);             /* unref single widgets */
    unref_pixbufs(mt.mpixbuf);

    /* cannot free data here because of sudden run stop                      */

  }
}



/* Activates actions on row clicks.

Revision: 31.12.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static gboolean
cb_maintenance_row_click(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN)
{
  gboolean propagate = FALSE, filtered;

  STSFY_GCC(user_data);

  if (event == NULL) {
    return TRUE;
  }

  if (GTK_IS_TREE_VIEW(widget) == FALSE) {
    return TRUE;
  }

  switch (event->button.button) {
  case 1:                                               /* left mouse button */
    propagate = maintenance_row_2click(widget, event);

    break;

  case 2:                                                /* mid mouse button */
    filtered = filter_event_type(event, GDK_BUTTON_RELEASE);

    if (filtered == TRUE) {
      maintenance_image_on(gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.image_on)));
    }

    break;

  case 3:                                              /* right mouse button */
    if (event->type == GDK_BUTTON_RELEASE) {     /* after cursor positioning */
      mt.popevent = event;
      maintenance_row_edit(XPLAB_EDIT_MENU);
    }

    break;

  default:
    break;
  }

  return propagate;
}



/* Expands/collapses current selected row.

Revision: 31.12.2005 E.Sack

Input : - pointer to widget
        - pointer to event
Output: - propagation flag
*/
static gboolean
maintenance_row_2click(GtkWidget *widget, GdkEvent *event)
{
  gboolean          filtered, expanded;
  GtkTreePath       *path;
  GtkTreeViewColumn *fcolumn;

  filtered = filter_event_type(event, GDK_2BUTTON_PRESS);

  if (filtered == FALSE) {
    return FALSE;                      /* propagate further event processing */
  }

  gtk_tree_view_get_cursor(GTK_TREE_VIEW(widget), &path, &fcolumn);

  if (path == NULL) {
    return TRUE;
  }

  expanded = gtk_tree_view_row_expanded(GTK_TREE_VIEW(widget), path);

  if (expanded == TRUE) {
    gtk_tree_view_collapse_row(GTK_TREE_VIEW(widget), path);
  }
  else {
    gtk_tree_view_expand_row(GTK_TREE_VIEW(widget), path, FALSE);
  }

  gtk_tree_path_free(path);

  return TRUE;                      /* prevent from further event processing */
}



/* Generates/deletes maintenance image windows.

Revision: 07.01.2006 E.Sack

Input : - pointer to widget
Output: - -
*/
static void
maintenance_image_on(gboolean on)
{
  if (on == TRUE) {
    maintenance_row_edit(XPLAB_QUERY_IMAGE);              /* fill image list */
  }
  else {                                             /* delete image windows */
    if (mt.mimg) {
      g_list_foreach(mt.mimg, delete_image_window, NULL);
    }

    g_list_free(mt.mimg);                       /* GTK deallocating required */
    mt.mimg = NULL;                                         /* clear pointer */
  }
}



/* Deletes a maintenance image window.

Revision: 07.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
delete_image_window(GTK_CONVENTION_GFunc_RTS_VOID) 
{
  mimg_def *mdef = (mimg_def *)data;                    /* member definition */

  STSFY_GCC(user_data);

  if (GTK_IS_WIDGET(mdef->mtop) == TRUE) {
    xplab_widget_destroy(&mdef->mtop);                 /* destroy top widget */
  }

  free_mimg_def(mdef);                          /* free list member contents */
  FREEMEM(mdef);                                         /* free list member */
}



/* Reverts to old DB loosing all changes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
Usage : Handler for events.
*/
static gboolean
cb_maintenance_revert_event(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN)
{
  gboolean filtered;

  STSFY_GCC(widget);
  STSFY_GCC(user_data);

  filtered = filter_event_type(event, GDK_BUTTON_PRESS);

  if (filtered == FALSE) {
    return FALSE;                      /* propagate further event processing */
  }

  maintenance_revert();

  return TRUE;                      /* prevent from further event processing */
}



/* Reverts to old DB loosing all changes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
Usage : Handler for signals.
*/
static void
cb_maintenance_revert(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(menuitem);
  STSFY_GCC(user_data);

  maintenance_revert();
}



/* Reverts maintenance tree to old DB loosing all changes.

Revision: 17.08.2005 E.Sack

Input : - -
Output: - DB updated
*/
static void
maintenance_revert(void)
{
  maintenance_reset_parse_status();

  xplab_rebuild_db_tree();                         /* read DB from hard disk */
  gtk_widget_set_sensitive(mt.revert_button, FALSE);
}



/* Looses DB changes and quits maintenance phase.

Revision: 21.12.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_maintenance_revert_quit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  maintenance_revert();
  cb_maintenance_quit(NULL, NULL);
}



/* Saves changed DB to disk and quits maintenance phase.

Revision: 21.12.2003 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_maintenance_save_quit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  if (GTK_IS_WIDGET(mt.top_maintenance) == TRUE) {
    maintenance_save();
    cb_maintenance_quit(NULL, NULL);
  }
}



/* Handler to update changed DB.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static gboolean
cb_maintenance_save(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN)
{
  gboolean filtered;

  STSFY_GCC(widget);
  STSFY_GCC(user_data);

  filtered = filter_event_type(event, GDK_BUTTON_PRESS);

  if (filtered == FALSE) {
    return FALSE;                      /* propagate further event processing */
  }

  maintenance_save();

  return TRUE;                           /* prevent further event processing */
}



/* Updates changed DB.

Revision: 17.08.2005 E.Sack

Input : - -
Output: - DB updated
*/
static void
maintenance_save(void)
{
  int err;

  maintenance_reset_parse_status();

  err = xplab_update_db(mt.mdb, mt.db_key_lock);                /* update DB */

  if (err) {
    xplab_text_insert(main_log, 
		      "WARNING: could not update database by maintenance\n");
  }
  else {
    xplab_text_insert(main_log, "database updated by maintenance\n");
    gtk_widget_set_sensitive(mt.revert_button, FALSE);
  }
}



/* Resets DB parse status.

Revision: 30.07.2005 E.Sack

Input : - -
Output: - -
*/
static void
maintenance_reset_parse_status(void)
{
  mt.items_deleted = 0;
  mt.items_created = 0;
  mt.items_changed = 0;
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_consistency(GtkWidget *tree_menu)
{
  GtkWidget *tree_menu_consistency, *tree_submenu_consistency;

  tree_menu_consistency = gtk_menu_item_new_with_label("consistency");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), tree_menu_consistency);

  tree_submenu_consistency = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tree_menu_consistency), 
			    tree_submenu_consistency);

  maintenance_menu_consistency_check(tree_submenu_consistency);
  maintenance_menu_consistency_delete(tree_submenu_consistency);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_consistency_check(GtkWidget *tree_menu)
{
  GtkWidget *tree_menu_consistency;

  tree_menu_consistency = gtk_menu_item_new_with_label("check");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), tree_menu_consistency);

  g_signal_connect(G_OBJECT(tree_menu_consistency), "activate",
		   G_CALLBACK(cb_consistency_check), (gpointer)&mt.mdb);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static GtkWidget *
maintenance_menu_where_referenced(GtkWidget *tree_menu)
{
  GtkWidget *where_referenced;

  where_referenced = gtk_menu_item_new_with_label("where referenced");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), where_referenced);

  g_signal_connect(G_OBJECT(where_referenced), "activate", 
		   G_CALLBACK(cb_row_edit), 
		   GINT_TO_POINTER(XPLAB_EDIT_ACTION_WHERE_REFERENCED));

  return where_referenced;
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_expand(GtkWidget *tree_menu)
{
  GtkWidget *tree_menu_expand, *tree_submenu_expand;

  tree_menu_expand = gtk_menu_item_new_with_label("expand");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), tree_menu_expand);

  tree_submenu_expand = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tree_menu_expand), 
			    tree_submenu_expand);

  maintenance_menu_item_expand_all(tree_submenu_expand);
  maintenance_menu_item_expand_selected(tree_submenu_expand);
  maintenance_menu_item_expand_classes(tree_submenu_expand);
  maintenance_menu_item_expand_nb(tree_submenu_expand);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_pattern(GtkWidget *tree_menu)
{
  GtkWidget *tree_menu_pattern, *tree_submenu_pattern;

  tree_menu_pattern = gtk_menu_item_new_with_label("pattern");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), tree_menu_pattern);

  tree_submenu_pattern = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tree_menu_pattern), 
			    tree_submenu_pattern);

  maintenance_menu_pattern(tree_submenu_pattern);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_pattern(GtkWidget *tree_menu)
{
  GtkWidget *sep;

  maintenance_menu_size(tree_menu);

  sep = gtk_menu_item_new();                              /* menu separator */
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), sep);

  maintenance_menu_scale(tree_menu);

  gtk_widget_set_sensitive(mt.scale_button, FALSE);
  gtk_widget_set_sensitive(mt.orig_button, FALSE);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu updated
*/
static void
maintenance_menu_size(GtkWidget *tree_menu) 
{
  GSList *item_group = NULL;

  mt.icon_button = gtk_radio_menu_item_new_with_label(item_group,
						      XPM_PATTERN_FLAG_ICON);
  item_group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(mt.icon_button));

  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), mt.icon_button);

  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt.icon_button), TRUE);
  g_signal_connect(mt.icon_button, "toggled", G_CALLBACK(cb_pattern_icon),
		   NULL);

  mt.x16_button = gtk_radio_menu_item_new_with_label(item_group,
						     XPM_PATTERN_FLAG_16x16);
  item_group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(mt.x16_button));

  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), mt.x16_button);

  g_signal_connect(mt.x16_button, "toggled", G_CALLBACK(cb_pattern_icon),
		   NULL);

  mt.x32_button = gtk_radio_menu_item_new_with_label(item_group, 
						     XPM_PATTERN_FLAG_32x32);

  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), mt.x32_button);

  g_signal_connect(mt.x32_button, "toggled", G_CALLBACK(cb_pattern_icon),
		   NULL);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_scale(GtkWidget *tree_menu)
{
  GSList *item_group = NULL;

  mt.scale_button = gtk_radio_menu_item_new_with_label(item_group, 
						       XPM_PATTERN_FLAG_SCALE);

  item_group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(mt.scale_button));

  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), mt.scale_button);

  g_signal_connect(mt.scale_button, "toggled", G_CALLBACK(cb_pattern_icon),
		   NULL);

  mt.orig_button = gtk_radio_menu_item_new_with_label(item_group, 
					   XPM_PATTERN_FLAG_ORIGINAL);

  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), mt.orig_button);

  g_signal_connect(mt.orig_button, "toggled", G_CALLBACK(cb_pattern_icon),
		   NULL);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_collapse(GtkWidget *tree_menu)
{
  GtkWidget *tree_menu_collapse, *tree_submenu_collapse;

  tree_menu_collapse = gtk_menu_item_new_with_label("collapse");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), tree_menu_collapse);

  tree_submenu_collapse = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tree_menu_collapse), 
			    tree_submenu_collapse);

  maintenance_menu_item_collapse_all(tree_submenu_collapse);
  maintenance_menu_item_collapse_selected(tree_submenu_collapse);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_collapse_selected(GtkWidget *tree_menu)
{
  mt.collapse_selected = gtk_menu_item_new_with_label("selected");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), mt.collapse_selected);

  g_signal_connect(G_OBJECT(mt.collapse_selected), "activate",
		   G_CALLBACK(cb_collapse_selected), NULL);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_expand_selected(GtkWidget *tree_menu)
{
  mt.expand_selected = gtk_menu_item_new_with_label("selected");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), mt.expand_selected);

  g_signal_connect(G_OBJECT(mt.expand_selected), "activate",
		   G_CALLBACK(cb_expand_selected), NULL);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
maintenance_menu_item_extract(GtkWidget *menu)
{
  mt.row_extract = gtk_menu_item_new_with_label("extract");
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), mt.row_extract);

  g_signal_connect(G_OBJECT(mt.row_extract), "activate", 
		   G_CALLBACK(cb_row_edit), 
		   GINT_TO_POINTER(XPLAB_EDIT_ACTION_EXTRACT));
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
maintenance_menu_item_dump(GtkWidget *menu)
{
  mt.row_dump = gtk_menu_item_new_with_label(MENU_BUTTON_DUMP);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), mt.row_dump);

  g_signal_connect(G_OBJECT(mt.row_dump), "activate", 
		   G_CALLBACK(cb_row_edit), 
		   GINT_TO_POINTER(XPLAB_EDIT_ACTION_DUMP));
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_edit(GtkWidget *menu)
{
  mt.row_edit = gtk_menu_item_new_with_label(MENU_BUTTON_EDIT XPLAB_MENU_EXT);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), mt.row_edit);

  g_signal_connect(G_OBJECT(mt.row_edit), "activate", 
		   G_CALLBACK(cb_row_edit), GINT_TO_POINTER(XPLAB_EDIT_DIALOG));
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_ccf(GtkWidget *menu)
{
  mt.ccf_menu = gtk_menu_item_new_with_label("CCF" XPLAB_MENU_EXT);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu), mt.ccf_menu);

  g_signal_connect(G_OBJECT(mt.ccf_menu), "activate", 
		   G_CALLBACK(cb_ccf), NULL);
}



/* Collapses selected tree view nodes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_collapse_selected(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  GtkTreePath       *path;
  GtkTreeViewColumn *fcolumn;

  STSFY_GCC(user_data);

  assert(menuitem);

  if (GTK_IS_TREE_VIEW(mt.treeview) == FALSE) {
    return;
  }

  gtk_tree_view_get_cursor(GTK_TREE_VIEW(mt.treeview), &path, &fcolumn);

  if (path == NULL) {
    return;
  }

  gtk_tree_view_collapse_row(GTK_TREE_VIEW(mt.treeview), path);

  gtk_tree_path_free(path);
}



/* Expands selected tree view nodes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_expand_selected(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  GtkTreePath       *path;
  GtkTreeViewColumn *fcolumn;

  STSFY_GCC(user_data);

  assert(menuitem);

  if (GTK_IS_TREE_VIEW(mt.treeview) == FALSE) {
    return;
  }

  gtk_tree_view_get_cursor(GTK_TREE_VIEW(mt.treeview), &path, &fcolumn);

  if (path == NULL) {
    return;
  }

  gtk_tree_view_expand_row(GTK_TREE_VIEW(mt.treeview), path, TRUE);

  gtk_tree_path_free(path);
}



/* Sets NB patterns as icons into the tree view.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_pattern_icon(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  gboolean     is_icon;
  GtkTreeModel *cmodel;

  STSFY_GCC(user_data);

  assert(checkmenuitem);

  if (gtk_check_menu_item_get_active(checkmenuitem) == TRUE) {
    is_icon = 
      gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.icon_button));

    if (is_icon == TRUE) {                      /* icon selected for display */
      gtk_widget_set_sensitive(mt.scale_button, FALSE);
      gtk_widget_set_sensitive(mt.orig_button, FALSE);
    }
    else {                                /* NB pattern selected for display */
      gtk_widget_set_sensitive(mt.scale_button, TRUE);
      gtk_widget_set_sensitive(mt.orig_button, TRUE);
    }

    cmodel = gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview));

    assert(GTK_IS_TREE_MODEL(cmodel) == TRUE);
    gtk_tree_model_foreach(cmodel, (GtkTreeModelForeachFunc)set_nb_icon, NULL);
  }
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_collapse_all(GtkWidget *tree_menu)
{
  GtkWidget *tree_menu_collapse;

  tree_menu_collapse = gtk_menu_item_new_with_label("all");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), tree_menu_collapse);

  g_signal_connect(G_OBJECT(tree_menu_collapse), "activate",
		   G_CALLBACK(cb_collapse_all), NULL);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_expand_classes(GtkWidget *tree_menu)
{
  GtkWidget *tree_menu_expand;

  tree_menu_expand = gtk_menu_item_new_with_label("to classes");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), tree_menu_expand);

  g_signal_connect(G_OBJECT(tree_menu_expand), "activate",
		   G_CALLBACK(cb_expand_to_id), 
		   GINT_TO_POINTER(XPLAB_MAINTENANCE_NODE_ID_CLASS_INDEX));
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_expand_nb(GtkWidget *tree_menu)
{
  GtkWidget *tree_menu_expand;

  tree_menu_expand = gtk_menu_item_new_with_label("to patterns");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), tree_menu_expand);

  g_signal_connect(G_OBJECT(tree_menu_expand), "activate",
		   G_CALLBACK(cb_expand_to_id), 
		   GINT_TO_POINTER(XPLAB_MAINTENANCE_NODE_ID_PATTERN));
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_expand_all(GtkWidget *tree_menu)
{
  GtkWidget *tree_menu_expand;

  tree_menu_expand = gtk_menu_item_new_with_label("all");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), tree_menu_expand);

  g_signal_connect(G_OBJECT(tree_menu_expand), "activate",
		   G_CALLBACK(cb_expand_all), NULL);
}



/* Collapses all tree view nodes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_collapse_all(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  assert(GTK_IS_TREE_VIEW(mt.treeview) == TRUE);

  gtk_tree_view_collapse_all(GTK_TREE_VIEW(mt.treeview));
}



/* Expands to id nodes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_expand_to_id(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  GtkTreeModel *cmodel;

  assert(user_data);
  assert(menuitem);
  assert(GTK_IS_TREE_VIEW(mt.treeview) == TRUE);

  gtk_tree_view_collapse_all(GTK_TREE_VIEW(mt.treeview));

  cmodel = gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview));

  assert(GTK_IS_TREE_MODEL(cmodel) == TRUE);

  gtk_tree_model_foreach(cmodel, (GtkTreeModelForeachFunc)expand_to_id,
			 user_data);
}



/* This function expands the rows to the ID.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - GTK convention
Usage:  To be called by the GTK's foreach-Function.
*/
static gboolean
expand_to_id(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN)
{
  gint node_id;

  STSFY_GCC(path);

  gtk_tree_model_get(GTK_TREE_MODEL(model), iter,
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

  if (node_id == GPOINTER_TO_INT(data)) {
    expand_to_root(GTK_TREE_VIEW(mt.treeview), *iter, FALSE);
  }

  return FALSE;                                           /* treat all rows */
}



/* Expands all tree view nodes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_expand_all(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(menuitem);
  assert(GTK_IS_TREE_VIEW(mt.treeview) == TRUE);

  gtk_tree_view_expand_all(GTK_TREE_VIEW(mt.treeview));
}



/* Checks DB consistency.

Revision: 02.01.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_consistency_check(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  db_def *db;

  assert(user_data);
  assert(menuitem);

  db = (db_def *)user_data;

  db_consistency(*db);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_consistency_delete(GtkWidget *tree_menu)
{
  GtkWidget *tree_submenu_delete;

  mt.tree_menu_delete = gtk_menu_item_new_with_label(MENU_BUTTON_DELETE);
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), mt.tree_menu_delete);

  gtk_widget_set_sensitive(mt.tree_menu_delete, FALSE);

  tree_submenu_delete = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(mt.tree_menu_delete), 
			    tree_submenu_delete);

  maintenance_menu_delete_identities(tree_submenu_delete);
  maintenance_menu_delete_empty_classes(tree_submenu_delete);
  maintenance_menu_delete_auxiliary_items(tree_submenu_delete);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_delete_identities(GtkWidget *tree_menu)
{
  GtkWidget *item_del;

  item_del = gtk_menu_item_new_with_label("identities");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), item_del);

  g_signal_connect(G_OBJECT(item_del), "activate", 
		   G_CALLBACK(cb_delete_identities), NULL);
}



/* Deletes NB identities in classes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_delete_identities(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  int d;

  STSFY_GCC(user_data);

  assert(menuitem);

  d = remove_identities(mt.mdb.tpl);

  if (d) {
    mt.items_deleted++;
  }

  maintenance_tree_store_fill(mt.treeview, TRUE);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_delete_empty_classes(GtkWidget *tree_menu)
{
  GtkWidget *item_del;

  item_del = gtk_menu_item_new_with_label("empty classes");
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), item_del);

  g_signal_connect(G_OBJECT(item_del), "activate", 
		   G_CALLBACK(cb_delete_empty_classes), NULL);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_delete_auxiliary_items(GtkWidget *tree_menu)
{
  GtkWidget *item_del;

  item_del = gtk_menu_item_new_with_label(MENU_BUTTON_AUX);
  gtk_menu_shell_append(GTK_MENU_SHELL(tree_menu), item_del);

  g_signal_connect(G_OBJECT(item_del), "activate", 
		   G_CALLBACK(cb_delete_auxiliary_items), NULL);
}



/* Deletes empty classes.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_delete_empty_classes(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  int d;

  STSFY_GCC(user_data);

  assert(menuitem);

  d = remove_empty_classes(mt.mdb.tpl);

  if (d) {
    mt.items_deleted++;
  }

  maintenance_tree_store_fill(mt.treeview, TRUE);
}



/* Deletes auxiliary items. That are values used for some actions but 
   are of no interest for the database at all. For instance, they are not
   stored on disc or displayed in the tree store.

Revision: 08.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_delete_auxiliary_items(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  int d;

  STSFY_GCC(user_data);

  assert(menuitem);

  d = check_aux(mt.mdb.tpl, 1);

  if (d) {
    mt.items_deleted++;
  }
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_lock(GtkWidget *file_menu)
{
  mt.lock_button = gtk_check_menu_item_new_with_label("lock DB");
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), mt.lock_button);

  g_signal_connect(G_OBJECT(mt.lock_button), "toggled", 
		   G_CALLBACK(cb_maintenance_lock), NULL);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_revert_quit(GtkWidget *file_menu)
{
  mt.revert_quit_button = gtk_menu_item_new_with_label("revert & quit");
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), mt.revert_quit_button);

  g_signal_connect(G_OBJECT(mt.revert_quit_button), "activate", 
		   G_CALLBACK(cb_maintenance_revert_quit), NULL);

  gtk_widget_set_sensitive(mt.revert_quit_button, FALSE);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_save_quit(GtkWidget *file_menu)
{
  GtkWidget *save_quit_button;

  save_quit_button = gtk_menu_item_new_with_label(DIALOG_BUTTON_SAVE_QUIT);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), save_quit_button);

  g_signal_connect(G_OBJECT(save_quit_button), "activate", 
		   G_CALLBACK(cb_maintenance_save_quit), NULL);
}



/* Adds an item to a menu.

Revision: 30.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
maintenance_menu_item_quit(GtkWidget *file_menu)
{
  mt.quit_button = gtk_menu_item_new_with_label(DIALOG_BUTTON_QUIT);

  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), mt.quit_button);

  g_signal_connect(G_OBJECT(mt.quit_button), "activate",
		   G_CALLBACK(cb_maintenance_quit), NULL);
}



/* Builds a scrolled window.

Revision: 05.12.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
maintenance_scrolled_window(GtkWidget *vbox)
{
  GtkWidget *scrolled_tree;

  /* put into box one scrolled tree                                          */
  scrolled_tree = maintenance_scrolled_treeview();
  gtk_box_pack_start(GTK_BOX(vbox), scrolled_tree, TRUE, TRUE, 0);
}



/* Creates the basic tree store for maintenance.

Revision: 30.07.2005 E.Sack

Input : - -
Output: - pointer to sort model
Postcondition: - the reference count of some GValues are increased by
                 filling the store and accordingly decreased when deleting 
                 the store, hence leaving the original reference count.
                 Be aware to unref the original instances after finalizing
                 GTK's main loop.
*/
static GtkTreeStore *
maintenance_tree_store_init(void)
{
  gint         cols;
  GType        gtypes[XPLAB_MAINTENANCE_TREE_STORE_COLUMNS];
  GtkTreeStore *tstore;

  for (cols = 0; cols < XPLAB_MAINTENANCE_TREE_STORE_COLUMNS; cols++) {
    switch (cols) {                               /* init store column types */
    case XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC:
      gtypes[cols] = GDK_TYPE_PIXBUF;  /* in-/decreasing its reference count */
      break;

    case XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT:
    case XPLAB_MAINTENANCE_TREE_STORE_COMMENT:
    case XPLAB_MAINTENANCE_TREE_STORE_SUMMARY:
      gtypes[cols] = G_TYPE_STRING;
      break;

    case XPLAB_MAINTENANCE_TREE_STORE_NODE_ID:
    case XPLAB_MAINTENANCE_TREE_STORE_NODE_FLAG:
      gtypes[cols] = G_TYPE_INT;
      break;

    default:
      assert(0);                                     /* something is missing */
      break;
    }
  }
 
  tstore = gtk_tree_store_newv(XPLAB_MAINTENANCE_TREE_STORE_COLUMNS, gtypes);

  return tstore;
}



/* Creates a scrolled tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to scrolled window definition
Output: - scrolled window definition updated
*/
static GtkWidget *
maintenance_scrolled_treeview(void)
{
  GtkWidget    *scrolled_window;
  GtkTreeStore *tstore;

  if (GTK_IS_TREE_VIEW(mt.treeview) == FALSE) {
    tstore = maintenance_tree_store_init();

    mt.treeview = maintenance_tree_view(tstore);

    maintenance_tree_store_fill(mt.treeview, FALSE);
  }

  /* create a scrolled window                                                */
  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window), 
				 GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);

  gtk_container_add(GTK_CONTAINER(scrolled_window), GTK_WIDGET(mt.treeview));

  return scrolled_window;
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_dpi(GtkTreeStore **combo_store)
{
  GtkTreeIter iter; 
  gchar       *string_out = NULL;

  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 

  /* propose some values */
  int_to_string("", DPI_MIN, "", &string_out);
  gtk_tree_store_append(*combo_store, &iter, NULL); 
  gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 
  FREEMEM(string_out);

  int_to_string("", DPI_DEF, "", &string_out);
  gtk_tree_store_append(*combo_store, &iter, NULL); 
  gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 
  FREEMEM(string_out);

  int_to_string("", DPI_MAX, "", &string_out);
  gtk_tree_store_append(*combo_store, &iter, NULL); 
  gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 
  FREEMEM(string_out);
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_method(GtkTreeStore **combo_store)
{
  gint          d;
  INLINE_METHOD i;
  char          *string_out = NULL;
  GtkTreeIter   iter; 

  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 

  for (i = OVERLAPPING; i <= LINE_SPACING; i++) {
    d = inline_method_to_string(i, &string_out);
    assert(d == 1);

    gtk_tree_store_append(*combo_store, &iter, NULL); 
    gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 

    FREEMEM(string_out);
  }
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_l_phase(GtkTreeStore **combo_store)
{
  gchar       *string_out = NULL;
  GtkTreeIter iter; 

  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 

  /* propose some values */
  int_to_string("", XPLAB_L_PHASE_MIN, "", &string_out);
  gtk_tree_store_append(*combo_store, &iter, NULL); 
  gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 
  FREEMEM(string_out);

  int_to_string("", XPLAB_L_PHASE_DEF, "", &string_out);
  gtk_tree_store_append(*combo_store, &iter, NULL); 
  gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 
  FREEMEM(string_out);

  int_to_string("", XPLAB_L_PHASE_MAX, "", &string_out);
  gtk_tree_store_append(*combo_store, &iter, NULL); 
  gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 
  FREEMEM(string_out);
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_h_phase(GtkTreeStore **combo_store)
{
  gchar       *string_out = NULL;
  GtkTreeIter iter; 

  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 

  /* propose some values */
  int_to_string("", XPLAB_H_PHASE_MIN, "", &string_out);
  gtk_tree_store_append(*combo_store, &iter, NULL); 
  gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 
  FREEMEM(string_out);

  if (XPLAB_H_PHASE_DEF != XPLAB_H_PHASE_MIN) {
    int_to_string("", XPLAB_H_PHASE_DEF, "", &string_out);
    gtk_tree_store_append(*combo_store, &iter, NULL); 
    gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 
    FREEMEM(string_out);
  }

  int_to_string("", XPLAB_H_PHASE_MAX, "", &string_out);
  gtk_tree_store_append(*combo_store, &iter, NULL); 
  gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 
  FREEMEM(string_out);
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_alignment(GtkTreeStore **combo_store)
{
  gint        d;
  ALIGMNT     al;
  char        *string_out = NULL;
  GtkTreeIter iter; 

  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 

  for (al = UPPERLINE; al >= UNDERLINE; al--) {
    d = aligmnt_to_string(al, &string_out);
    assert(d == 1);

    gtk_tree_store_append(*combo_store, &iter, NULL); 
    gtk_tree_store_set(*combo_store, &iter, 0, string_out, -1); 

    FREEMEM(string_out);
  }
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_printable(GtkTreeStore **combo_store)
{
  GtkTreeIter iter; 

  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 

  gtk_tree_store_append(*combo_store, &iter, NULL);
  gtk_tree_store_set(*combo_store, &iter, 0, "FALSE", -1);

  gtk_tree_store_append(*combo_store, &iter, NULL);
  gtk_tree_store_set(*combo_store, &iter, 0, "TRUE", -1);
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_fixed(GtkTreeStore **combo_store)
{
  GtkTreeIter iter; 

  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 

  gtk_tree_store_append(*combo_store, &iter, NULL);
  gtk_tree_store_set(*combo_store, &iter, 0, "FALSE", -1);

  gtk_tree_store_append(*combo_store, &iter, NULL);
  gtk_tree_store_set(*combo_store, &iter, 0, "TRUE", -1);
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_tps_id(GtkTreeStore **combo_store)
{
  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_class_name(GtkTreeStore **combo_store)
{
  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_class_translation(GtkTreeStore **combo_store)
{
  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 
}



/* Creates and fills a combo store.

Revision: 30.07.2005 E.Sack

Input : - pointer to combo store pointer
Output: - combo store updated
*/
static void
fill_combo_store_class_comment(GtkTreeStore **combo_store)
{
  assert(combo_store);

  if (GTK_IS_TREE_STORE(*combo_store) == TRUE) {
    return;                                  /* static store already filled */
  }

  *combo_store = gtk_tree_store_new(1, G_TYPE_STRING); 
}



/* Creates and fills the tree view for maintenance.

Revision: 03.12.2005 E.Sack

Input : - pointer to store
Output: - pointer to tree view
*/
static GtkWidget *
maintenance_tree_view(GtkTreeStore *tstore)
{
  gint              cols;
  view_def          vcols[XPLAB_MAINTENANCE_TREE_STORE_COLUMNS];
  GtkWidget         *tview;
  GtkTreeSelection  *tselctn;
  GtkCellRenderer   *crenderer;
  GtkTreeViewColumn *tv_column;

  for (cols = 0; cols < XPLAB_MAINTENANCE_TREE_VIEW_COLUMNS; cols++) {
    init_view_def(&vcols[cols]);
  }

  /* assign view headers                                                     */
  add_string(XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC_TITLE, &vcols[XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC].title);
  add_string(XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC_TXT_TITLE, &vcols[XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC_TXT].title);
  add_string(XPLAB_MAINTENANCE_TREE_VIEW_COMMENT_TITLE, &vcols[XPLAB_MAINTENANCE_TREE_VIEW_COMMENT].title);
  add_string(XPLAB_MAINTENANCE_TREE_VIEW_SUMMARY_TITLE, &vcols[XPLAB_MAINTENANCE_TREE_VIEW_SUMMARY].title);

  /* sort-column-ID translation table for the view                           */
  vcols[XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC].mcol = XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC;
  vcols[XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC_TXT].mcol = XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT;
  vcols[XPLAB_MAINTENANCE_TREE_VIEW_COMMENT].mcol = XPLAB_MAINTENANCE_TREE_STORE_COMMENT;
  vcols[XPLAB_MAINTENANCE_TREE_VIEW_SUMMARY].mcol = XPLAB_MAINTENANCE_TREE_STORE_SUMMARY;

  tview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(tstore));

  /* The view now holds a reference.  We can get rid of our own reference.   */
  g_object_unref(G_OBJECT(tstore));

  tselctn = gtk_tree_view_get_selection(GTK_TREE_VIEW(tview));
  gtk_tree_selection_set_mode(GTK_TREE_SELECTION(tselctn), GTK_SELECTION_SINGLE);

  for (cols = 0; cols < XPLAB_MAINTENANCE_TREE_VIEW_COLUMNS; cols++) {
    switch (cols) {
    case XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC:/* two store columns into one view column */
      tv_column = gtk_tree_view_column_new();
      gtk_tree_view_column_set_title(tv_column, vcols[cols].title);

      crenderer = gtk_cell_renderer_pixbuf_new();/* icon in first model column */

      gtk_tree_view_column_pack_start(tv_column, crenderer, FALSE);

      gtk_tree_view_column_set_attributes(tv_column, crenderer,
					  "pixbuf", vcols[cols].mcol,
					  NULL);

      /* the combo-renderer serves as an intermediate buffer showing strings to
         be choosed. The strings have to be stored to the backing store by a CB
         because the renderer doesn't do it himself. The strings for the combo
         box are stored in an own model.                                     */

      crenderer = gtk_cell_renderer_combo_new();      /* second model column */

      gtk_tree_view_column_pack_start(tv_column, crenderer, TRUE);

      gtk_tree_view_column_set_attributes
	(tv_column, crenderer,
	 "text", vcols[XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC_TXT].mcol,
	 NULL);

      /* we need individual properties for certain cells                     */
      gtk_tree_view_column_set_cell_data_func
	(tv_column, crenderer, (GtkTreeCellDataFunc)cell_func_combo,
	 GINT_TO_POINTER(cols),            /* assuming columns not reordered */
	 NULL);

      g_signal_connect(G_OBJECT(crenderer), "edited", 
		       G_CALLBACK(cb_cell_edited), (gpointer)tview);

      gtk_tree_view_append_column(GTK_TREE_VIEW(tview), tv_column);

      break;

    case XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC_TXT:           /* handled above */
      break;

    case XPLAB_MAINTENANCE_TREE_VIEW_COMMENT:
    case XPLAB_MAINTENANCE_TREE_VIEW_SUMMARY:
      crenderer = gtk_cell_renderer_text_new();

      tv_column = gtk_tree_view_column_new_with_attributes
	(vcols[cols].title,
	 crenderer,
	 "text", vcols[cols].mcol,/* model column: the real link to the store */
	 NULL);

      gtk_tree_view_append_column(GTK_TREE_VIEW(tview), tv_column);

      break;

    default:
      assert(0);
      break;
    }
  }

  gtk_tree_view_set_enable_search(GTK_TREE_VIEW(tview), FALSE);

  mt.cursor_changed_id =
    g_signal_connect(G_OBJECT(tview), "cursor-changed",
		     G_CALLBACK(cb_maintenance_row_select), NULL);

  g_signal_connect(G_OBJECT(tview), "event",
		   G_CALLBACK(cb_maintenance_row_click), NULL);

  g_signal_connect(G_OBJECT(tview), "row-expanded",
		   G_CALLBACK(cb_maintenance_row_expanded_collapsed), NULL);

  g_signal_connect(G_OBJECT(tview), "row-collapsed",
		   G_CALLBACK(cb_maintenance_row_expanded_collapsed), NULL);

  for (cols = 0; cols < XPLAB_MAINTENANCE_TREE_VIEW_COLUMNS; cols++) {
    free_view_def(&vcols[cols]);
  }

  return tview;
}



/* Updates the DB model with the user input after parsing.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cell_edited(GTK_CONVENTION_GtkCellRendererText_RTS_VOID)
{
  gint          msuccess, node_id, tps_idx, cls_idx, dpi, l_phase, h_phase;
  gint          cidx, err;
  gchar         *temp_buf = NULL, *ret_buf = NULL;
  gboolean      valid_path;
  INLINE_METHOD method_out;
  ALIGMNT       aligmnt_out;
  GtkTreeView   *tree_view = (GtkTreeView *)user_data;
  GtkTreeModel  *tree_model;
  GtkTreeIter   iter, titer;

  STSFY_GCC(cell);

  assert(user_data);
  assert(GTK_IS_TREE_VIEW(tree_view) == TRUE);

  tree_model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree_view));

  valid_path = gtk_tree_model_get_iter_from_string(tree_model, &iter,
						   path_string);
  assert(valid_path == TRUE);

  gtk_tree_model_get(tree_model, &iter, 
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

  switch (node_id) {
  case XPLAB_MAINTENANCE_NODE_ID_DPI:
    msuccess = string_to_dpi(new_text, &dpi);                       /* parse */

    tps_idx = child_has_tps_index(tree_model, iter, &titer);
    assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

    mt.mdb.tpl.tps[tps_idx].media.dpi = dpi;

    maintenance_tree_store_fill(mt.treeview, TRUE);
    mt.items_changed++;

    if (msuccess) {
      put_history_to_tree_store(new_text, mt.mcombo_store.dpi);
      string_to_log(main_log, XPLAB_DPI_CONVERSION_SUCCESS_MSG, "", "\n");
    }
    else {
      string_to_log(main_log, XPLAB_DPI_CONVERSION_CORRECTED_MSG, "", "\n");
    }

    break;

  case XPLAB_MAINTENANCE_NODE_ID_LINE_METHOD:
    msuccess = string_to_inline_method(new_text, &method_out);      /* parse */

    if (msuccess) {
      tps_idx = child_has_tps_index(tree_model, iter, &titer);
      assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

      mt.mdb.tpl.tps[tps_idx].media.in_line.method = method_out;

      maintenance_tree_store_fill(mt.treeview, TRUE);
      mt.items_changed++;

      string_to_log(main_log, XPLAB_METHOD_CONVERSION_SUCCESS_MSG, "", "\n");
    }
    else {
      string_to_log(main_log, XPLAB_METHOD_CONVERSION_FAILED_MSG, "", "\n");
    }

    break;

  case XPLAB_MAINTENANCE_NODE_ID_L_PHASE:
    msuccess = string_to_l_phase(new_text, &l_phase);               /* parse */

    tps_idx = child_has_tps_index(tree_model, iter, &titer);
    assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

    mt.mdb.tpl.tps[tps_idx].media.in_line.linespacing.period.l_phase = l_phase;

    maintenance_tree_store_fill(mt.treeview, TRUE);
    mt.items_changed++;

    if (msuccess) {
      put_history_to_tree_store(new_text, mt.mcombo_store.l_phase);
      string_to_log(main_log, XPLAB_L_PHASE_CONVERSION_SUCCESS_MSG, "", "\n");
    }
    else {
      string_to_log(main_log, XPLAB_L_PHASE_CONVERSION_CORRECTED_MSG, "", "\n");
    }

    break;

  case XPLAB_MAINTENANCE_NODE_ID_H_PHASE:
    msuccess = string_to_h_phase(new_text, &h_phase);               /* parse */

    tps_idx = child_has_tps_index(tree_model, iter, &titer);
    assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

    mt.mdb.tpl.tps[tps_idx].media.in_line.linespacing.period.h_phase = h_phase;

    maintenance_tree_store_fill(mt.treeview, TRUE);
    mt.items_changed++;

    if (msuccess) {
      put_history_to_tree_store(new_text, mt.mcombo_store.h_phase);
      string_to_log(main_log, XPLAB_H_PHASE_CONVERSION_SUCCESS_MSG, "", "\n");
    }
    else {
      string_to_log(main_log, XPLAB_H_PHASE_CONVERSION_CORRECTED_MSG, "", "\n");
    }

    break;

  case XPLAB_MAINTENANCE_NODE_ID_TPS_ID:
    ret_buf = xplab_parse_string(new_text, &msuccess, MAX_LEN_TPS_ID);

    switch (msuccess) {
    case XPLAB_STRING_CONVERSION_SUCCESS:
      msuccess = 1;
      break;

    case XPLAB_STRING_TRUNCATED:
      cat_string(" [", "STRING_TRUNCATED", "]", &temp_buf);
      msuccess = 1;
      break;

    case XPLAB_STRING_ZERO_LENGTH:
      cat_string(" [", "STRING_ZERO_LENGTH", "]", &temp_buf);
      msuccess = 0;
      break;

    case XPLAB_STRING_CONVERSION_FAILED:
      msuccess = 0;
      break;

    default:
      assert(0);
      break;
    }

    tps_idx = child_has_tps_index(tree_model, iter, &titer);
    assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

    if (msuccess) {
      FREEMEM(mt.mdb.tpl.tps[tps_idx].ident);
      add_string(ret_buf, &mt.mdb.tpl.tps[tps_idx].ident);
      put_history_to_tree_store(ret_buf, mt.mcombo_store.tps_id);

      maintenance_tree_store_fill(mt.treeview, TRUE);
      mt.items_changed++;

      string_to_log(main_log, XPLAB_TPS_ID_CONVERSION_SUCCESS_MSG, temp_buf, "\n");
    }
    else {
      string_to_log(main_log, XPLAB_TPS_ID_PARSING_ERROR, temp_buf, "\n");
    }

    FREEMEM(ret_buf);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_NAME:
    ret_buf = xplab_parse_string(new_text, &msuccess, MAX_LEN_CLASS_NAME);

    switch (msuccess) {
    case XPLAB_STRING_CONVERSION_SUCCESS:
      msuccess = 1;
      break;

    case XPLAB_STRING_TRUNCATED:
      cat_string(" [", "STRING_TRUNCATED", "]", &temp_buf);
      msuccess = 1;
      break;

    case XPLAB_STRING_ZERO_LENGTH:
      cat_string(" [", "STRING_ZERO_LENGTH", "]", &temp_buf);
      msuccess = 0;
      break;

    case XPLAB_STRING_CONVERSION_FAILED:
      msuccess = 0;
      break;

    default:
      assert(0);
      break;
    }

    tps_idx = child_has_tps_index(tree_model, iter, &titer);
    assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

    if (msuccess) {                       /* search class name in class list */
      cidx = search_class(ret_buf, mt.mdb.tpl.tps[tps_idx].cl);

      if (cidx >= 0) {                              /* same class name found */
	msuccess = 0;
      }
      else {
	msuccess = 1;
      }
    }

    if (msuccess) {
      cls_idx = child_has_class_index(tree_model, iter, &titer);
      assert(cls_idx >= 0 && tps_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

      FREEMEM(mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.class);
      add_string(ret_buf, &mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.class);
      put_history_to_tree_store(ret_buf, mt.mcombo_store.class_name);

      maintenance_tree_store_fill(mt.treeview, TRUE);
      mt.items_changed++;

      string_to_log(main_log, XPLAB_CLASS_NAME_CONVERSION_SUCCESS_MSG, temp_buf, "\n");
    }
    else {
      string_to_log(main_log, XPLAB_CLASS_NAME_PARSING_ERROR, temp_buf, "\n");
    }

    FREEMEM(ret_buf);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_TRANSLATION:
    ret_buf = xplab_parse_string(new_text, &msuccess, MAX_LEN_CLASS_TRANS);

    switch (msuccess) {
    case XPLAB_STRING_CONVERSION_SUCCESS:
      msuccess = 1;
      break;

    case XPLAB_STRING_TRUNCATED:
      cat_string(" [", "STRING_TRUNCATED", "]", &temp_buf);
      msuccess = 1;
      break;

    case XPLAB_STRING_ZERO_LENGTH:
      cat_string(" [", "STRING_ZERO_LENGTH", "]", &temp_buf);
      msuccess = 0;
      break;

    case XPLAB_STRING_CONVERSION_FAILED:
      msuccess = 0;
      break;

    default:
      assert(0);
      break;
    }

    if (msuccess) {
      tps_idx = child_has_tps_index(tree_model, iter, &titer);
      assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

      cls_idx = child_has_class_index(tree_model, iter, &titer);
      assert(cls_idx >= 0 && tps_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

      FREEMEM(mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.sign);
      add_string(ret_buf, &mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.sign);
      put_history_to_tree_store(ret_buf, mt.mcombo_store.class_translation);

      maintenance_tree_store_fill(mt.treeview, TRUE);
      mt.items_changed++;

      string_to_log(main_log, XPLAB_CLASS_TRANSLATION_CONVERSION_SUCCESS_MSG, temp_buf, "\n");
    }
    else {
      string_to_log(main_log, XPLAB_CLASS_TRANSLATION_PARSING_ERROR, temp_buf, "\n");
    }

    FREEMEM(ret_buf);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_COMMENT:
    ret_buf = xplab_parse_string(new_text, &msuccess, MAX_LEN_CLASS_COMMENT);

    switch (msuccess) {
    case XPLAB_STRING_CONVERSION_SUCCESS:
      msuccess = 1;
      break;

    case XPLAB_STRING_TRUNCATED:
      cat_string(" [", "STRING_TRUNCATED", "]", &temp_buf);
      msuccess = 1;
      break;

    case XPLAB_STRING_ZERO_LENGTH:
      cat_string(" [", "STRING_ZERO_LENGTH", "]", &temp_buf);
      msuccess = 0;
      break;

    case XPLAB_STRING_CONVERSION_FAILED:
      msuccess = 0;
      break;

    default:
      assert(0);
      break;
    }

    if (msuccess) {
      tps_idx = child_has_tps_index(tree_model, iter, &titer);
      assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

      cls_idx = child_has_class_index(tree_model, iter, &titer);
      assert(cls_idx >= 0 && tps_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

      FREEMEM(mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.crem);
      add_string(ret_buf, &mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.crem);
      put_history_to_tree_store(ret_buf, mt.mcombo_store.class_comment);

      maintenance_tree_store_fill(mt.treeview, TRUE);
      mt.items_changed++;

      string_to_log(main_log, XPLAB_CLASS_COMMENT_CONVERSION_SUCCESS_MSG, temp_buf, "\n");
    }
    else {
      string_to_log(main_log, XPLAB_CLASS_COMMENT_PARSING_ERROR, temp_buf, "\n");
    }

    FREEMEM(ret_buf);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_ALIGNMENT:
    msuccess = string_to_aligmnt(new_text, &aligmnt_out);

    if (msuccess) {
      tps_idx = child_has_tps_index(tree_model, iter, &titer);
      assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

      cls_idx = child_has_class_index(tree_model, iter, &titer);
      assert(cls_idx >= 0 && tps_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

      mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.alignment = aligmnt_out;

      maintenance_tree_store_fill(mt.treeview, TRUE);
      mt.items_changed++;

      string_to_log(main_log, XPLAB_ALIGNMENT_CONVERSION_SUCCESS_MSG, "", "\n");
    }
    else {
      string_to_log(main_log, XPLAB_ALIGNMENT_CONVERSION_FAILED_MSG, "", "\n");
    }

    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_PRINTABLE:
    err = check_name(new_text);

    msuccess = err ? 0 : 1;

    if (!msuccess || (strcmp(new_text, "TRUE") && strcmp(new_text, "FALSE"))) {
      string_to_log(main_log, XPLAB_CLASS_PRINTABLE_PARSING_ERROR, "", "\n");
    }
    else {
      tps_idx = child_has_tps_index(tree_model, iter, &titer);
      assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

      cls_idx = child_has_class_index(tree_model, iter, &titer);
      assert(cls_idx >= 0 && tps_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

      if (!strcmp(new_text, "TRUE")) {
	mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.printable = 1;
      }
      else {
	mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.printable = 0;
      }

      maintenance_tree_store_fill(mt.treeview, TRUE);
      mt.items_changed++;

      string_to_log(main_log, XPLAB_CLASS_PRINTABLE_CONVERSION_SUCCESS_MSG, "", "\n");
    }

    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_FIXED:
    err = check_name(new_text);

    msuccess = err ? 0 : 1;

    if (!msuccess || (strcmp(new_text, "TRUE") && strcmp(new_text, "FALSE"))) {
      string_to_log(main_log, XPLAB_CLASS_FIXED_PARSING_ERROR, "", "\n");
    }
    else {
      tps_idx = child_has_tps_index(tree_model, iter, &titer);
      assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

      cls_idx = child_has_class_index(tree_model, iter, &titer);
      assert(cls_idx >= 0 && tps_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

      if (!strcmp(new_text, "TRUE")) {
	mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.fixed = 1;
      }
      else {
	mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.fixed = 0;
      }

      maintenance_tree_store_fill(mt.treeview, TRUE);
      mt.items_changed++;

      string_to_log(main_log, XPLAB_CLASS_FIXED_CONVERSION_SUCCESS_MSG, "", "\n");
    }

    break;

  default: assert(0);
    break;
  }

  cb_maintenance_sensitivity(NULL, NULL);

  FREEMEM(temp_buf);
}



/* Puts string to a tree store.

Revision: 01.07.2005 E.Sack

Input : - pointer to string
        - pointer to store
Output: - -
Usage:  The store is meant to contain input strings, which can be used
        later on. Therefore identic strings are filtered. The store has
        a limited size.
Bug:    The store cannot be cleared.
*/
static void
put_history_to_tree_store(gchar *htext, GtkTreeStore *tstore)
{
  gint        err, nodes;
  gchar       *sum;
  gboolean    node_valid, valid_utf;
  foreach_def fdef;
  GtkTreeIter iter;

  init_foreach_def(&fdef);

  if (XPLAB_HISTORY_LENGTH <= 0) {
    return;                                                    /* no history */
  }

  err = check_name(htext);

  if (err) {
    return;
  }

  if (GTK_IS_TREE_STORE(tstore) == FALSE) {
    return;
  }

  /* filter identic strings for combo store */
  fdef.iaux = XPLAB_COMBO_STORE_COLUMN_TXT;
  fdef.txt = htext;                             /* don't free pointer copy */

  gtk_tree_model_foreach(GTK_TREE_MODEL(tstore),
			 (GtkTreeModelForeachFunc)tree_model_search_strings,
			 (gpointer)&fdef);

  if (fdef.baux == TRUE) {
    return;
  }

  nodes = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(tstore), NULL);

  if (nodes < XPLAB_HISTORY_LENGTH) {
    gtk_tree_store_append(tstore, &iter, NULL);
  }
  else {                        /* if buffer is full, only change last row */
    gtk_tree_store_insert(tstore, &iter, NULL, XPLAB_HISTORY_LENGTH);
  }

  valid_utf = g_utf8_validate(htext, -1, NULL);           /* check for UTF */

  sum = (valid_utf == TRUE) ? g_strdup(htext) : xplab_to_utf(htext);

  gtk_tree_store_set(tstore, &iter, 0, sum, -1);              /* store UTF */
  g_free(sum);

  node_valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(tstore), &iter);

  while (node_valid == TRUE) {
    node_valid = gtk_tree_store_remove(tstore, &iter);
  }
}



/* Cell function sets the properties for all renderers of the maintenance 
   tree view.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cell_func_combo(GTK_CONVENTION_TREE_CELL_DATA_FUNC_RTS_VOID)
{
  gint node_id;

  assert(tree_column);
  assert(GTK_IS_CELL_RENDERER(cell) == TRUE);
  assert(GTK_IS_TREE_MODEL(tree_model) == TRUE);

  assert(GPOINTER_TO_INT(data) == XPLAB_MAINTENANCE_TREE_VIEW_GRAPHIC);

  if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button)) == FALSE) {
    return;
  }

  gtk_tree_model_get(GTK_TREE_MODEL(tree_model), iter, 
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

  switch (node_id) {
  case XPLAB_MAINTENANCE_NODE_ID_TPS_ID:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.tps_id) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_TPS_ID&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE,
		 "has_entry", XPLAB_MAINTENANCE_EDIT_TPS_ID&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE,
		 "model", GTK_TREE_MODEL(mt.mcombo_store.tps_id),
 	         "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_DPI:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.dpi) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_DPI&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE,
		 "has_entry", XPLAB_MAINTENANCE_EDIT_DPI&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE,
		 "model", GTK_TREE_MODEL(mt.mcombo_store.dpi),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_LINE_METHOD:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.method) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_LINE_METHOD&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE, 
		 "has_entry", XPLAB_MAINTENANCE_EDIT_LINE_METHOD&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE, 
		 "model", GTK_TREE_MODEL(mt.mcombo_store.method),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_L_PHASE:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.l_phase) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_L_PHASE&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE, 
		 "has_entry", XPLAB_MAINTENANCE_EDIT_L_PHASE&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE, 
		 "model", GTK_TREE_MODEL(mt.mcombo_store.l_phase),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_H_PHASE:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.h_phase) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_H_PHASE&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE, 
		 "has_entry", XPLAB_MAINTENANCE_EDIT_H_PHASE&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE, 
		 "model", GTK_TREE_MODEL(mt.mcombo_store.h_phase),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_NAME:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.class_name) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_CLASS_NAME&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE, 
		 "has_entry", XPLAB_MAINTENANCE_EDIT_CLASS_NAME&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE, 
		 "model", GTK_TREE_MODEL(mt.mcombo_store.class_name),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_TRANSLATION:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.class_translation) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_CLASS_TRANSLATION&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE, 
		 "has_entry", XPLAB_MAINTENANCE_EDIT_CLASS_TRANSLATION&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE, 
		 "model", GTK_TREE_MODEL(mt.mcombo_store.class_translation),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_COMMENT:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.class_comment) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_CLASS_COMMENT&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE, 
		 "has_entry", XPLAB_MAINTENANCE_EDIT_CLASS_COMMENT&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE, 
		 "model", GTK_TREE_MODEL(mt.mcombo_store.class_comment),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_ALIGNMENT:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.alignment) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_CLASS_ALIGNMENT&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE, 
		 "has_entry", XPLAB_MAINTENANCE_EDIT_CLASS_ALIGNMENT&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE, 
		 "model", GTK_TREE_MODEL(mt.mcombo_store.alignment),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_PRINTABLE:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.printable) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_CLASS_PRINTABLE&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE, 
		 "has_entry", XPLAB_MAINTENANCE_EDIT_CLASS_PRINTABLE&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE, 
		 "model", GTK_TREE_MODEL(mt.mcombo_store.printable),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  case XPLAB_MAINTENANCE_NODE_ID_CLASS_FIXED:
    assert(GTK_IS_TREE_MODEL(mt.mcombo_store.fixed) == TRUE);
    g_object_set(cell, 
		 "editable", XPLAB_MAINTENANCE_EDIT_CLASS_FIXED&XPLAB_MAINTENANCE_EDITABLE?TRUE:FALSE, 
		 "has_entry", XPLAB_MAINTENANCE_EDIT_CLASS_FIXED&XPLAB_MAINTENANCE_HASENTRY?TRUE:FALSE, 
		 "model", GTK_TREE_MODEL(mt.mcombo_store.fixed),
		 "text-column", XPLAB_COMBO_STORE_COLUMN_TXT,
		 NULL);
    break;

  default:
    g_object_set(cell, 
		 "editable", FALSE, 
		 "has_entry", FALSE, 
		 NULL);
    break;
  }
}



/* References the history tree view stores.

Revision: 22.08.2005 E.Sack

Input : - -
Output: - icons referenced
Postcondition: - stores must be unreferenced explicitly at progam end.
*/
static combo_store_def
ref_combo_stores(void)
{
  combo_store_def combo_store;

  init_combo_store_def(&combo_store);

  fill_combo_store_tps_id(&combo_store.tps_id);/* fill static pop-down lists */
  assert(G_OBJECT(combo_store.tps_id)->ref_count == 1);

  fill_combo_store_dpi(&combo_store.dpi);
  assert(G_OBJECT(combo_store.dpi)->ref_count == 1);

  fill_combo_store_method(&combo_store.method);
  assert(G_OBJECT(combo_store.method)->ref_count == 1);

  fill_combo_store_l_phase(&combo_store.l_phase);
  assert(G_OBJECT(combo_store.l_phase)->ref_count == 1);

  fill_combo_store_h_phase(&combo_store.h_phase);
  assert(G_OBJECT(combo_store.h_phase)->ref_count == 1);

  fill_combo_store_class_name(&combo_store.class_name);
  assert(G_OBJECT(combo_store.class_name)->ref_count == 1);

  fill_combo_store_class_translation(&combo_store.class_translation);
  assert(G_OBJECT(combo_store.class_translation)->ref_count == 1);

  fill_combo_store_class_comment(&combo_store.class_comment);
  assert(G_OBJECT(combo_store.class_comment)->ref_count == 1);

  fill_combo_store_alignment(&combo_store.alignment);
  assert(G_OBJECT(combo_store.alignment)->ref_count == 1);

  fill_combo_store_printable(&combo_store.printable);
  assert(G_OBJECT(combo_store.printable)->ref_count == 1);

  fill_combo_store_fixed(&combo_store.fixed);
  assert(G_OBJECT(combo_store.fixed)->ref_count == 1);

  return combo_store;
}



/* Dereferences the maintenance tree view default icons.

Revision: 30.07.2005 E.Sack

Input : - icons definition
Output: - icons dereferenced
Postcondition: - output must be unreferenced explicitly at progam end.
*/
static void
unref_combo_stores(combo_store_def combo_store)
{
  assert(G_OBJECT(combo_store.tps_id)->ref_count == 1);
  g_object_unref (combo_store.tps_id);

  assert(G_OBJECT(combo_store.dpi)->ref_count == 1);
  g_object_unref (combo_store.dpi);

  assert(G_OBJECT(combo_store.method)->ref_count == 1);
  g_object_unref (combo_store.method);

  assert(G_OBJECT(combo_store.l_phase)->ref_count == 1);
  g_object_unref (combo_store.l_phase);

  assert(G_OBJECT(combo_store.h_phase)->ref_count == 1);
  g_object_unref (combo_store.h_phase);

  assert(G_OBJECT(combo_store.class_name)->ref_count == 1);
  g_object_unref (combo_store.class_name);

  assert(G_OBJECT(combo_store.class_translation)->ref_count == 1);
  g_object_unref (combo_store.class_translation);

  assert(G_OBJECT(combo_store.class_comment)->ref_count == 1);
  g_object_unref (combo_store.class_comment);

  assert(G_OBJECT(combo_store.alignment)->ref_count == 1);
  g_object_unref (combo_store.alignment);

  assert(G_OBJECT(combo_store.printable)->ref_count == 1);
  g_object_unref (combo_store.printable);

  assert(G_OBJECT(combo_store.fixed)->ref_count == 1);
  g_object_unref (combo_store.fixed);
}



/* References the maintenance tree view default icons.

Revision: 30.07.2005 E.Sack

Input : - -
Output: - icons referenced
Postcondition: - icons must be unreferenced explicitly at progam end.
*/
static pixbuf_def
ref_pixbufs(void)
{
  pixbuf_def pixbuf;

  static gchar *xpm_opened_read[] = {
    "16 16 3 1",
    "       c #" XPM_COLOR_ICON_BGND,
    ".      c #" XPM_COLOR_ICON_FGND,
    "X      c #" XPM_COLOR_ICON_FILL_READ,
    "                ",
    "   ......       ",
    "   .XXX.X.      ",
    "   .XXX.XX.     ",
    "   .XXX.XXX.    ",
    "   .XXX.....    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .........    ",
    "                ",
    "                "};

  static gchar *xpm_opened_write[] = {
    "16 16 3 1",
    "       c #" XPM_COLOR_ICON_BGND,
    ".      c #" XPM_COLOR_ICON_FGND,
    "X      c #" XPM_COLOR_ICON_FILL_WRITE,
    "                ",
    "   ......       ",
    "   .XXX.X.      ",
    "   .XXX.XX.     ",
    "   .XXX.XXX.    ",
    "   .XXX.....    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .........    ",
    "                ",
    "                "};

  static gchar *xpm_closed_read[] = {
    "16 16 3 1",
    "       c #" XPM_COLOR_ICON_BGND,
    ".      c #" XPM_COLOR_ICON_FGND,
    "X      c #" XPM_COLOR_ICON_FILL_READ,
    "                ",
    "   .........    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .........    ",
    "                ",
    "                "};

  static gchar *xpm_closed_write[] = {
    "16 16 3 1",
    "       c #" XPM_COLOR_ICON_BGND,
    ".      c #" XPM_COLOR_ICON_FGND,
    "X      c #" XPM_COLOR_ICON_FILL_WRITE,
    "                ",
    "   .........    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .XXXXXXX.    ",
    "   .........    ",
    "                ",
    "                "};

  static gchar *xpm_editable_read[] = {
    "16 16 3 1",
    "       c #" XPM_COLOR_ICON_BGND,
    ".      c #" XPM_COLOR_ICON_FGND,
    "X      c #" XPM_COLOR_ICON_FILL_READ,
    "                ",
    "                ",
    "................",
    ".XXXXXXXXXXXXXX.",
    ".XXXXXXXXXXXXXX.",
    ".XX..........XX.",
    ".XXXXXXXXXXXXXX.",
    ".XX..........XX.",
    ".XXXXXXXXXXXXXX.",
    ".XX......XXXXXX.",
    ".XXXXXXXXXXXXXX.",
    ".XXXXXXXXXXXXXX.",
    ".XXXXXXXXXXXXXX.",
    "................",
    "                ",
    "                "};

  static gchar *xpm_editable_write[] = {
    "16 16 3 1",
    "       c #" XPM_COLOR_ICON_BGND,
    ".      c #" XPM_COLOR_ICON_FGND,
    "X      c #" XPM_COLOR_ICON_FILL_WRITE,
    "                ",
    "                ",
    "................",
    ".XXXXXXXXXXXXXX.",
    ".XXXXXXXXXXXXXX.",
    ".XX..........XX.",
    ".XXXXXXXXXXXXXX.",
    ".XX..........XX.",
    ".XXXXXXXXXXXXXX.",
    ".XX......XXXXXX.",
    ".XXXXXXXXXXXXXX.",
    ".XXXXXXXXXXXXXX.",
    ".XXXXXXXXXXXXXX.",
    "................",
    "                ",
    "                "};

  static gchar *xpm_not_editable[] = {
    "16 16 3 1",
    "       c #" XPM_COLOR_ICON_BGND,
    ".      c #" XPM_COLOR_ICON_FGND,
    "X      c #" XPM_COLOR_ICON_FILL_READ,
    "                ",
    "                ",
    "................",
    "..XXXXXXXXXXXX..",
    ".X..XXXXXXXX..X.",
    ".XXX.XXXXXX.XXX.",
    ".XXXX..XX..XXXX.",
    ".XXXXXX..XXXXXX.",
    ".XXXXXX..XXXXXX.",
    ".XXXX..XX..XXXX.",
    ".XXX.XXXXXX.XXX.",
    ".X..XXXXXXXX..X.",
    "..XXXXXXXXXXXX..",
    "................",
    "                ",
    "                "};

  init_pixbuf_def(&pixbuf);

  pixbuf.closed_read = gdk_pixbuf_new_from_xpm_data((const char **)xpm_closed_read);
  assert(G_OBJECT(pixbuf.closed_read)->ref_count == 1);

  pixbuf.open_read = gdk_pixbuf_new_from_xpm_data((const char **)xpm_opened_read);
  assert(G_OBJECT(pixbuf.open_read)->ref_count == 1);

  pixbuf.open_write = gdk_pixbuf_new_from_xpm_data((const char **)xpm_opened_write);
  assert(G_OBJECT(pixbuf.open_write)->ref_count == 1);

  pixbuf.closed_write = gdk_pixbuf_new_from_xpm_data((const char **)xpm_closed_write);
  assert(G_OBJECT(pixbuf.closed_write)->ref_count == 1);

  pixbuf.toedit_read = gdk_pixbuf_new_from_xpm_data((const char **)xpm_editable_read);
  assert(G_OBJECT(pixbuf.toedit_read)->ref_count == 1);

  pixbuf.toedit_write = gdk_pixbuf_new_from_xpm_data((const char **)xpm_editable_write);
  assert(G_OBJECT(pixbuf.toedit_write)->ref_count == 1);

  pixbuf.unedit = gdk_pixbuf_new_from_xpm_data((const char **)xpm_not_editable);
  assert(G_OBJECT(pixbuf.unedit)->ref_count == 1);

  return pixbuf;
}



/* Dereferences the maintenance tree view default icons.

Revision: 30.07.2005 E.Sack

Input : - icons definition
Output: - icons dereferenced
Postcondition: - output must be unreferenced explicitly at progam end.
*/
static void
unref_pixbufs(pixbuf_def pixbuf)
{
  assert(G_OBJECT(pixbuf.closed_read)->ref_count == 1);
  g_object_unref (pixbuf.closed_read);

  assert(G_OBJECT(pixbuf.open_read)->ref_count == 1);
  g_object_unref (pixbuf.open_read);

  assert(G_OBJECT(pixbuf.open_write)->ref_count == 1);
  g_object_unref (pixbuf.open_write);

  assert(G_OBJECT(pixbuf.closed_write)->ref_count == 1);
  g_object_unref (pixbuf.closed_write);

  assert(G_OBJECT(pixbuf.toedit_read)->ref_count == 1);
  g_object_unref (pixbuf.toedit_read);

  assert(G_OBJECT(pixbuf.toedit_write)->ref_count == 1);
  g_object_unref (pixbuf.toedit_write);

  assert(G_OBJECT(pixbuf.unedit)->ref_count == 1);
  g_object_unref (pixbuf.unedit);
}



/* Fills up the maintenance tree store.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree view
        - flag (tree store is revertable)
Output: - tree view updated
Usage : the flag should be set whenever the store has been changed.
*/
static void 
maintenance_tree_store_fill(GtkWidget *tview, gboolean revertable)
{
  assert(GTK_IS_TREE_VIEW(tview) == TRUE);

  treestore_db_def(GTK_TREE_VIEW(tview), mt.mdb);

  gtk_widget_set_sensitive(mt.revert_button, revertable);

  ccf_sensitivity();
  update_image_window_all();

  maintenance_statusbar_default();
}



/* Rereads the maintenance tree store from hard disc.

Revision: 30.07.2005 E.Sack

Input : - -
Output: - -
Usage : Access function for local data.
*/
static void 
xplab_rebuild_db_tree(void)
{
  int err;

  if (GTK_IS_WIDGET(mt.top_maintenance) == TRUE) {
    if (GTK_IS_TREE_VIEW(mt.treeview) == TRUE) {                 /* re-using */
      free_db_def(&mt.mdb);
      xplab_release_db(&mt.db_key_read, 0);                     /* erase key */
      mt.mdb = xplab_request_db(NULL, &mt.db_key_read, XPLAB_DB_RO,
				&mt.revert_button);               /* DB copy */

      err = check_name(mt.mdb.db_fname);
      assert(!err);

      maintenance_tree_store_fill(mt.treeview, FALSE);
    }
  }                                    /* else no maintenance dialog running */
}



/* Creates the database tree store.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree view
        - DB
Output: - tree store updated
*/
static void
treestore_db_def(GtkTreeView *tview, db_def db)
{
  GtkTreeIter  piter;
  GtkTreeStore *tstore;

  tstore = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(tview)));
  assert(GTK_IS_TREE_MODEL(tstore) == TRUE);

  gtk_tree_model_foreach(GTK_TREE_MODEL(tstore),        /* set deletion flag */
			 (GtkTreeModelForeachFunc)set_nodes_flag, NULL);

  treestore_db_att(tview, piter, db, "0");        /* first root node in tree */
  treestore_tps_lst(tview, piter, db.tpl, "1");  /* second root node in tree */

  remove_node(GTK_TREE_MODEL(tstore));                   /* delete old iters */

  gtk_tree_model_foreach(GTK_TREE_MODEL(tstore),     /* set icons to current */
			 (GtkTreeModelForeachFunc)set_nb_icon, NULL);
}



/* This function sets the flag bit in all node IDs.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - GTK convention
Usage :  To be called by the GTK's foreach-Function.
*/
static gboolean
set_nodes_flag(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN)
{
  gint node_flag;

  STSFY_GCC(path);
  STSFY_GCC(data);

  gtk_tree_model_get(GTK_TREE_MODEL(model), iter,
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_FLAG, &node_flag, -1);

  gtk_tree_store_set(GTK_TREE_STORE(model), 
		     iter, XPLAB_MAINTENANCE_TREE_STORE_NODE_FLAG, 
		     node_flag | XPLAB_MAINTENANCE_NODE_FLAG_SEL,
		     -1);

  return FALSE;                                            /* treat all rows */
}



/* This function detects an iter ment for deletion, based on a flagged
   node ID.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - GTK convention
Usage:  As an iter deletion doesn't work in the GTK's foreach-Function,
        a detected iter is returned for deletion out of the GTK's loop.
*/
static gboolean
delete_nodes_flag(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN)
{
  gint         node_flag;
  treenode_def *raux = (treenode_def *)data;

  STSFY_GCC(path);
  STSFY_GCC(data);

  assert(raux);

  gtk_tree_model_get(GTK_TREE_MODEL(model), iter,
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_FLAG, &node_flag, -1);

  if (node_flag & XPLAB_MAINTENANCE_NODE_FLAG_SEL) {           /* flag set */
    raux->piter = *iter;
    raux->iaux = TRUE;                       /* one row for deletion found */
  }
  else {
    raux->iaux = FALSE;                                /* breaks all loops */
  }

  return raux->iaux;
}



/* Removes old nodes from the tree model.

Revision: 30.07.2005 E.Sack

Input : - GTK convention
Output: - tree model updated
*/
static void
remove_node(GtkTreeModel *model)
{
  treenode_def aux;

  assert(GTK_IS_TREE_MODEL(model) == TRUE);

  while (1) {
    gtk_tree_model_foreach(GTK_TREE_MODEL(model),
			   (GtkTreeModelForeachFunc)delete_nodes_flag,
			   (gpointer)&aux);

    if (aux.iaux == TRUE) {
      gtk_tree_store_remove(GTK_TREE_STORE(model), &aux.piter);
    }
    else {
      break;
    }
  }
}



/* Creates the database attributes node.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - database
        - path string to use
Output: - tree store updated
*/
static void
treestore_db_att(GtkTreeView *tview, GtkTreeIter piter, db_def db, 
		 gchar *path_string)
{
  piter = treestore_node_db_att(tview, piter, db, path_string);

  treestore_db_name(tview, piter, db.db_fname);
  treestore_db_version(tview, piter, db.db_version);
}



/* Creates the database attributes node.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - database
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_db_att(GtkTreeView *tview, GtkTreeIter piter, db_def db, 
		      gchar *path_string)
{
  char *summary;

  summary = treestore_summary_db_att(db.db_fname, db.db_version);

  piter = treestore_node_string(tview, piter, "", "(DB attributes)", summary,
				XPLAB_MAINTENANCE_NODE_ID_DB_ATTRIBUTES,
				path_string);

  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - pointer to DB file name
        - pointer to DB version string
Output: - pointer to summary text
*/
static gchar *
treestore_summary_db_att(char *db_name, char *db_version)
{
  gchar *summary = NULL;

  cat_string("{", db_name, ", ", &summary);
  cat_string(db_version, "}", "", &summary);

  return summary;
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - pointer to DB file name
Output: - tree store updated
*/
static void
treestore_db_name(GtkTreeView *tview, GtkTreeIter piter, char *db_name)
{
  treestore_leaf_string(tview, piter, db_name, "(DB file name)", "",
			XPLAB_MAINTENANCE_NODE_ID_DB_NAME, 
			XPLAB_MAINTENANCE_EDIT_DB_NAME &
			XPLAB_MAINTENANCE_EDITABLE);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - pointer to version string
Output: - tree store updated
*/
static void
treestore_db_version(GtkTreeView *tview, GtkTreeIter piter, char *db_version)
{
  treestore_leaf_string(tview, piter, db_version, "(DB version)", "", 
			XPLAB_MAINTENANCE_NODE_ID_DB_VERSION, 
			XPLAB_MAINTENANCE_EDIT_DB_VERSION &
			XPLAB_MAINTENANCE_EDITABLE);
}



/* Creates the TPS list tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - TPS list
        - pointer to path string
Output: - tree store updated
*/
static void
treestore_tps_lst(GtkTreeView *tview, GtkTreeIter piter, tps_lst tpl, 
		  gchar *path_string)
{
  gint i;

  piter = treestore_node_tps_lst(tview, piter, tpl, path_string);

  for (i = 0; i < tpl.cnt; i++) {                /* successive nodes in root */
    treestore_tps_def(tview, piter, tpl.tps[i], i);
  }
}



/* Creates the TPS list node.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - test pattern set list
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_tps_lst(GtkTreeView *tview, GtkTreeIter piter, tps_lst tpl, 
		       gchar *path_string)
{
  char *summary;

  summary = treestore_summary_tps_lst(tpl.cnt);

  piter = treestore_node_string(tview, piter, "", "(TPS list)", summary, 
				XPLAB_MAINTENANCE_NODE_ID_TPS_LIST, path_string);

  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - number of test pattern sets
Output: - pointer to summary text
*/
static gchar *
treestore_summary_tps_lst(int tps_cnt)
{
  gchar *summary = NULL;

  int_to_string("{", tps_cnt, "}", &summary);

  return summary;
}



/* Creates the TPS tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - TPS
        - TPS index
Output: - tree store updated
*/
static void
treestore_tps_def(GtkTreeView *tview, GtkTreeIter piter, tps_def tps, 
		  gint tidx)
{
  piter = treestore_node_tps_index(tview, piter, tps, tidx);

  treestore_tps_ident(tview, piter, tps.ident);
  treestore_media_att(tview, piter, tps.media);
  treestore_cls_lst(tview, piter, tps.cl);
}



/* Creates the TPS index tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - TPS
        - TPS index
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_tps_index(GtkTreeView *tview, GtkTreeIter piter, tps_def tps, 
			 gint tidx)
{
  char *index = NULL;
  char *summary;

  int_to_string("", tidx, "", &index);

  summary = treestore_summary_tps_index(tps.ident, tps.cl.cnt);

  piter = treestore_node_string(tview, piter, index, "(TPS index)", summary, 
				XPLAB_MAINTENANCE_NODE_ID_TPS_INDEX, NULL);

  FREEMEM(index);
  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - pointer to TPS ID
        - number of classes
Output: - pointer to summary text
*/
static gchar *
treestore_summary_tps_index(gchar *tps_id, gint class_cnt)
{
  char *summary = NULL;

  cat_string("{", tps_id, "", &summary);
  int_to_string(", ", class_cnt, "}", &summary);

  return summary;
}



/* Creates the TPS media tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - media
        - query set
Output: - tree store updated
*/
static void
treestore_media_att(GtkTreeView *tview, GtkTreeIter piter, media_att media)
{
  piter = treestore_node_media_att(tview, piter, media);

  treestore_dpi(tview, piter, media.dpi);
  treestore_inline_def(tview, piter, media.in_line);
}



/* Creates the media node.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - media attributes
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_media_att(GtkTreeView *tview, GtkTreeIter piter, 
			 media_att media)
{
  char *summary;

  summary = treestore_summary_media_att(media.dpi, media.in_line.method);

  piter = treestore_node_string(tview, piter, "", "(media attributes)", summary, 
				XPLAB_MAINTENANCE_NODE_ID_MEDIA_ATTRIBUTES, NULL);

  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - dpi value
        - inline detection method
Output: - pointer to summary text
*/
static gchar *
treestore_summary_media_att(gint dpi, INLINE_METHOD method)
{
  int  d;
  char *tmp = NULL, *summary = NULL;

  d = inline_method_to_string(method, &tmp);
  assert(d);

  int_to_string("{", dpi, ", ", &summary);
  cat_string(tmp, "}", "", &summary);

  FREEMEM(tmp);

  return summary;
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - dpi
Output: - tree store updated
*/
static void
treestore_dpi(GtkTreeView *tview, GtkTreeIter piter, int dpi)
{
  char *buf = NULL;

  int_to_string("", dpi, "", &buf);

  treestore_leaf_string(tview, piter, buf, "[dpi] (resolution)", "",
			XPLAB_MAINTENANCE_NODE_ID_DPI,
			XPLAB_MAINTENANCE_EDIT_DPI & 
			XPLAB_MAINTENANCE_EDITABLE);

  FREEMEM(buf);
}



/* Creates the inline methods tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - inline method
Output: - tree store updated
*/
static void
treestore_inline_def(GtkTreeView *tview, GtkTreeIter piter, inline_def ild)
{
  treestore_inline_method(tview, piter, ild.method);
  treestore_linespacing_par(tview, piter, ild.linespacing);
}



/* Creates the linespacing tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - linespacing
Output: - tree store updated
*/
static void
treestore_linespacing_par(GtkTreeView *tview, GtkTreeIter piter, 
			  linespacing_par lpar)
{
  treestore_opos(tview, piter, lpar.opos);
  treestore_period_def(tview, piter, lpar.period);
}



/* Creates the linespacing tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - period
Output: - tree store updated
*/
static void
treestore_period_def(GtkTreeView *tview, GtkTreeIter piter, period_def period)
{
  piter = treestore_node_period_length(tview, piter, period);

  treestore_l_phase(tview, piter, period.l_phase);
  treestore_h_phase(tview, piter, period.h_phase);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - L phase
Output: - tree store updated
*/
static void
treestore_l_phase(GtkTreeView *tview, GtkTreeIter piter, gint l_phase)
{
  char *graphic = NULL, *comment;

  int_to_string("", l_phase, "", &graphic);

  comment = treestore_comment_l_phase(l_phase);

  treestore_leaf_string(tview, piter, graphic, comment, "", 
			XPLAB_MAINTENANCE_NODE_ID_L_PHASE,
			XPLAB_MAINTENANCE_EDIT_L_PHASE &
			XPLAB_MAINTENANCE_EDITABLE);

  FREEMEM(comment);
  FREEMEM(graphic);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - H phase
Output: - tree store updated
*/
static void
treestore_h_phase(GtkTreeView *tview, GtkTreeIter piter, gint h_phase)
{
  char *graphic = NULL, *comment;

  int_to_string("", h_phase, "", &graphic);

  comment = treestore_comment_h_phase(h_phase);

  treestore_leaf_string(tview, piter, graphic, comment, "", 
			XPLAB_MAINTENANCE_NODE_ID_H_PHASE,
			XPLAB_MAINTENANCE_EDIT_H_PHASE &
			XPLAB_MAINTENANCE_EDITABLE);

  FREEMEM(comment);
  FREEMEM(graphic);
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - period
Output: - pointer to summary text
*/
static gchar *
treestore_comment_l_phase(gint l_phase)
{
  gchar *comment = NULL;

  cat_string((l_phase >= XPLAB_L_PHASE_MIN) ? "[pixel]":"undefined",
	     " (L phase)", "", &comment);

  return comment;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - period
Output: - pointer to summary text
*/
static gchar *
treestore_comment_h_phase(gint h_phase)
{
  gchar *comment = NULL;

  cat_string((h_phase >= XPLAB_H_PHASE_MIN) ? "[pixel]":"undefined",
	     " (H phase)", "", &comment);

  return comment;
}



/* Creates the period node.

Revision: 30.07.2005 E.Sack

Input : - pointer to trees store
        - period
        - query set
Output: - pointer to node
*/
static GtkTreeIter
treestore_node_period_length(GtkTreeView *tview, GtkTreeIter piter,
			     period_def period)
{
  char *comment, *summary;

  comment = treestore_comment_period_length(period);
  summary = treestore_summary_period_length(period);

  piter = treestore_node_string(tview, piter, "", comment, summary,
				XPLAB_MAINTENANCE_NODE_ID_PERIOD_LENGTH, NULL);

  FREEMEM(comment);
  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - period
Output: - pointer to comment text
*/
static gchar *
treestore_comment_period_length(period_def period)
{
  gchar *comment = NULL;

  if (period.l_phase > 0 && period.h_phase > 0) {
    int_to_string("", period.l_phase + period.h_phase, " [pixel]", &comment);
  }
  else {
    add_string("undefined", &comment);
  }

  add_string(" (period length)", &comment);

  return comment;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - period
Output: - pointer to summary text
*/
static gchar *
treestore_summary_period_length(period_def period)
{
  gchar *summary = NULL;

  int_to_string("{", period.l_phase, ", ", &summary);
  int_to_string("", period.h_phase, "}", &summary);

  return summary;
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - optimal position
Output: - tree store updated
*/
static void
treestore_opos(GtkTreeView *tview, GtkTreeIter piter, gint opos)
{
  char *buf = NULL, *tmp = NULL;

  int_to_string("", opos, "", &buf);

  cat_string(opos<0?"undefined":"[pixel]", " (optimal line start position)", 
	     "", &tmp);

  treestore_leaf_string(tview, piter, buf, tmp, "", 
			XPLAB_MAINTENANCE_NODE_ID_OPOS,
			XPLAB_MAINTENANCE_EDIT_OPOS &
			XPLAB_MAINTENANCE_EDITABLE);

  FREEMEM(buf);
  FREEMEM(tmp);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - inline method
Output: - tree store updated
*/
static void
treestore_inline_method(GtkTreeView *tview, GtkTreeIter piter, 
			INLINE_METHOD method)
{
  char *buf = NULL;

  inline_method_to_string(method, &buf);

  treestore_leaf_string(tview, piter, buf, "(line detection method)", "", 
			XPLAB_MAINTENANCE_NODE_ID_LINE_METHOD,
			XPLAB_MAINTENANCE_EDIT_LINE_METHOD &
			XPLAB_MAINTENANCE_EDITABLE);

  FREEMEM(buf);
}



/* Creates a TPS tree node.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - pointer to TPS ID
Output: - tree store updated
*/
static void
treestore_tps_ident(GtkTreeView *tview, GtkTreeIter piter, gchar *tps_ident)
{
  treestore_leaf_string(tview, piter, tps_ident, "(TPS ID)", "", 
			XPLAB_MAINTENANCE_NODE_ID_TPS_ID,
			XPLAB_MAINTENANCE_EDIT_TPS_ID &
			XPLAB_MAINTENANCE_EDITABLE);
}



/* Creates the class list tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - class list
Output: - tree store updated
*/
static void
treestore_cls_lst(GtkTreeView *tview, GtkTreeIter piter, cls_lst clst)
{
  int i;

  piter = treestore_node_cls_lst(tview, piter, clst.cnt);

  for (i = 0; i < clst.cnt; i++) {
    treestore_cls_def(tview, piter, clst.cls[i], i);
  }
}



/* Creates the class list node.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - number of classes
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_cls_lst(GtkTreeView *tview, GtkTreeIter piter, gint cls_cnt)
{
  char *summary;

  summary = treestore_summary_cls_lst(cls_cnt);

  piter = treestore_node_string(tview, piter, "", "(class list)", summary, 
				XPLAB_MAINTENANCE_NODE_ID_CLASS_LIST, NULL);

  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - number of classes
Output: - pointer to summary text
*/
static gchar *
treestore_summary_cls_lst(int cls_cnt)
{
  char *summary = NULL;

  int_to_string("{", cls_cnt, "}", &summary);

  return summary;
}



/* Creates the class tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - class
        - class index
Output: - tree store updated
*/
static void
treestore_cls_def(GtkTreeView *tview, GtkTreeIter piter, cls_def cls, 
		  gint cidx)
{
  piter = treestore_node_cls_index(tview, piter, cls, cidx);

  treestore_class_attributes(tview, piter, cls);
  treestore_nb_lst(tview, piter, cls.nbl);
}



/* Creates the pattern list tree.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - NB list
Output: - tree store updated
*/
static void
treestore_nb_lst(GtkTreeView *tview, GtkTreeIter piter, nb_lst nbl)
{
  int i;

  piter = treestore_node_nb_lst(tview, piter, nbl.cnt);

  for (i = 0; i < nbl.cnt; i++) {
    treestore_nb_def(tview, piter, nbl.nb[i], i);
  }
}



/* Creates the NB list tree node text.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - NB count
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_nb_lst(GtkTreeView *tview, GtkTreeIter piter, int nbl_cnt)
{
  char *summary;

  summary = treestore_summary_nb_lst(nbl_cnt);

  piter = treestore_node_string(tview, piter, "", "(NB list)", 
				summary, XPLAB_MAINTENANCE_NODE_ID_NB_LIST, NULL);

  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - number of NB
Output: - pointer to summary text
*/
static gchar *
treestore_summary_nb_lst(int nb_cnt)
{
  char *summary = NULL;

  int_to_string("{", nb_cnt, "}", &summary);

  return summary;
}



/* Creates a pattern tree.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - NB
        - NB index
Output: - tree store updated
*/
static void
treestore_nb_def(GtkTreeView *tview, GtkTreeIter piter, nb_def nb, gint nidx)
{
  piter = treestore_node_nb_index(tview, piter, nb, nidx);

  treestore_pic_def(tview, piter, nb.pat);
  treestore_nb_chn_lst(tview, piter, nb.cnb, 1);
  treestore_nb_chn_lst(tview, piter, nb.nnb, 0);
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - area
        - number of CNB
        - number of NNB
Output: - pointer to summary text
*/
static gchar *
treestore_summary_nb_index(area_def ad, gint cnb_cnt, gint nnb_cnt)
{
  gchar *summary = NULL;

  add_string("{", &summary);
  area_def_to_string(ad, &summary);

  if (cnb_cnt) {
    int_to_string(", CNB: ", cnb_cnt, "", &summary);
  }

  if (nnb_cnt) {
    int_to_string(", NNB: ", nnb_cnt, "", &summary);
  }

  add_string("}", &summary);

  return summary;
}



/* Converts area to string.

Revision: 30.07.2005 E.Sack

Input : - area
Output: - pointer to string
Postcondition: pointer to deallocate
*/
static void
area_def_to_string(area_def ad, char **pattern_size)
{
  if (pattern_size) {
    int_to_string("", ad.width, 
		  XPLAB_MAINTENANCE_NODE_ID_PATTERN_SIZE_FORMAT_SEP, 
		  pattern_size);

    int_to_string("", ad.height, "", pattern_size);
  }
}



/* Creates a pattern chain list tree.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - NB chain list
        - flag (0: input refers to NNB, 1: input refers to CNB)
Output: - tree store updated
*/
static void
treestore_nb_chn_lst(GtkTreeView *tview, GtkTreeIter piter, nb_chn_lst chn, 
		     gint is_cnb)
{
  int i;

  if (!chn.cnt) {
    return;                                          /* no chain, save space */
  }

  piter = treestore_node_nb_chn_lst(tview, piter, chn.cnt, is_cnb);

  for (i = 0; i < chn.cnt; i++) {
    treestore_nb_chn_def(tview, piter, chn.cd[i], is_cnb, 
			 is_cnb?XPLAB_MAINTENANCE_NODE_ID_CNB_INDEX:
			 XPLAB_MAINTENANCE_NODE_ID_NNB_INDEX, i);
  }
}



/* Creates the NB chain list tree node text.

Revision: 19.12.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - chain count
        - flag (0: NNB, 1: CNB)
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_nb_chn_lst(GtkTreeView *tview, GtkTreeIter piter, gint chn_cnt, 
			  gint is_cnb)
{
  char *summary;

  summary = treestore_summary_nb_chn_lst(chn_cnt);

  piter = treestore_node_string(tview, piter, "", 
				is_cnb?"(CNB list)":"(NNB list)", summary,
				is_cnb?XPLAB_MAINTENANCE_NODE_ID_CNB_LIST:
				XPLAB_MAINTENANCE_NODE_ID_NNB_LIST, NULL);

  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 19.12.2005 E.Sack

Input : - number of NB chains
Output: - pointer to summary text
*/
static gchar *
treestore_summary_nb_chn_lst(gint chn_cnt)
{
  gchar *summary = NULL;

  if (!chn_cnt) {
    return summary;                                 /* no chain, save space */
  }

  int_to_string("{", chn_cnt, "}", &summary);

  return summary;
}



/* Creates a pattern chain leaf.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - NB chain
        - flag (0: input refers to NNB, 1: input refers to CNB)
        - node ID
        - NB chain index
Output: - tree store updated
*/
static void
treestore_nb_chn_def(GtkTreeView *tview, GtkTreeIter piter, nb_chn_def cd, 
		     gint is_cnb, gint node_id, gint chidx)
{
  char *buf = NULL;

  piter = treestore_node_nb_chn_def(tview, piter, node_id, chidx, cd, is_cnb);

  int_to_string("", cd.cdx, "", &buf);
  treestore_leaf_string(tview, piter, buf, "(chain class index)", "",
			XPLAB_MAINTENANCE_NODE_ID_CHAIN_CLASS_INDEX,
			XPLAB_MAINTENANCE_EDIT_CHAIN_CLASS_INDEX &
			XPLAB_MAINTENANCE_EDITABLE);
  FREEMEM(buf);

  int_to_string("", cd.ndx, "", &buf);
  treestore_leaf_string(tview, piter, buf, "(chain NB index)", "",
			XPLAB_MAINTENANCE_NODE_ID_CHAIN_NB_INDEX,
			XPLAB_MAINTENANCE_EDIT_CHAIN_NB_INDEX &
			XPLAB_MAINTENANCE_EDITABLE);
  FREEMEM(buf);

  int_to_string("", cd.x, "", &buf);
  treestore_leaf_string(tview, piter, buf, "[pixel] (x offset)", "",
			XPLAB_MAINTENANCE_NODE_ID_CHAIN_X_OFFSET,
			XPLAB_MAINTENANCE_EDIT_CHAIN_X_OFFSET &
			XPLAB_MAINTENANCE_EDITABLE);
  FREEMEM(buf);

  int_to_string("", cd.y, "", &buf);
  treestore_leaf_string(tview, piter, buf, "[pixel] (y offset)", "",
			XPLAB_MAINTENANCE_NODE_ID_CHAIN_Y_OFFSET,
			XPLAB_MAINTENANCE_EDIT_CHAIN_Y_OFFSET &
			XPLAB_MAINTENANCE_EDITABLE);
  FREEMEM(buf);
}



/* Creates the NB chain tree node text.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - node ID
        - NB chain index
        - NB chain
        - flag (0: NNB, 1: CNB)
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_nb_chn_def(GtkTreeView *tview, GtkTreeIter piter, gint node_id,
			  gint chidx, nb_chn_def cd, gint is_cnb)
{
  char *index = NULL, *summary;

  int_to_string("", chidx, "", &index);

  summary = treestore_summary_nb_chn_def(cd.cdx, cd.ndx, cd.x, cd.y, is_cnb);

  piter = treestore_node_string(tview, piter, index, 
				is_cnb?"(CNB index)":"(NNB index)", 
				summary, node_id, NULL);

  FREEMEM(index);
  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - chain class index
        - chain NB index
        - chain x offset
        - chain y offset
        - flag (0: input refers to NNB, 1: input refers to CNB)
Output: - pointer to summary text
*/
static gchar *
treestore_summary_nb_chn_def(gint cd_cdx, gint cd_ndx, gint cd_x, gint cd_y, 
			     gint is_cnb)
{
  gchar *summary = NULL;

  int_to_string("{(", cd_cdx, ",", &summary);
  int_to_string("", cd_ndx, "), (", &summary);
  int_to_string("", cd_x, ",", &summary);
  int_to_string("", cd_y, ")}", &summary);

  return summary;
}



/* Creates an image tree.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - image definition
Output: - tree store updated
*/
static void
treestore_pic_def(GtkTreeView *tview, GtkTreeIter piter, pic_def img)
{
  char *buf = NULL;

  int_to_string("", img.ar.width, "x", &buf);
  int_to_string("", img.ar.height, "", &buf);

  treestore_leaf_string(tview, piter, buf, "[pixel] (pattern size)", "", 
			XPLAB_MAINTENANCE_NODE_ID_PATTERN,
			XPLAB_MAINTENANCE_EDIT_PATTERN &
			XPLAB_MAINTENANCE_EDITABLE);
  FREEMEM(buf);
}



/* Creates the NB index tree.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - NB
        - NB index
        - pointer to query set
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_nb_index(GtkTreeView *tview, GtkTreeIter piter, nb_def nb, 
			gint nidx)
{
  char *index = NULL, *summary;

  int_to_string("", nidx, "", &index);

  summary = treestore_summary_nb_index(nb.pat.ar, nb.cnb.cnt, nb.nnb.cnt);

  piter = treestore_node_string(tview, piter, index, "(NB index)", summary, 
				XPLAB_MAINTENANCE_NODE_ID_NB_INDEX, NULL);

  FREEMEM(index);
  FREEMEM(summary);

  return piter;
}



/* Creates the class index tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - class
        - class index
Output: - tree store updated
*/
static GtkTreeIter
treestore_node_cls_index(GtkTreeView *tview, GtkTreeIter piter, cls_def cls, 
			 gint cidx)
{
  char *index = NULL;
  char *summary;

  int_to_string("", cidx, "", &index);

  summary = treestore_summary_cls_index(cls.att.class, cls.att.alignment,
					cls.nbl.cnt);

  piter = treestore_node_string(tview, piter, index, "(class index)", summary, 
				XPLAB_MAINTENANCE_NODE_ID_CLASS_INDEX, NULL);

  FREEMEM(index);
  FREEMEM(summary);

  return piter;
}



/* Sets the node text.

Revision: 30.07.2005 E.Sack

Input : - pointer to class name
        - alignment
        - number of NB
Output: - pointer to summary text
*/
static gchar *
treestore_summary_cls_index(gchar *class_name, ALIGMNT class_alignment, 
			    gint nb_cnt)
{
  char *astring = NULL, *summary = NULL;

  aligmnt_to_string(class_alignment, &astring);

  cat_string("{", class_name, ", ", &summary);
  add_string(astring, &summary);
  int_to_string(", ", nb_cnt, "}", &summary);

  FREEMEM(astring);

  return summary;
}



/* Creates the class attributes tree.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - class
Output: - tree store updated
*/
static void
treestore_class_attributes(GtkTreeView *tview, GtkTreeIter piter, cls_def cls)
{
  piter = treestore_node_string(tview, piter, "", "(class attributes)", "", 
				XPLAB_MAINTENANCE_NODE_ID_CLASS_ATTRIBUTES,NULL);

  treestore_class_name(tview, piter, cls.att.class);
  treestore_class_translation(tview, piter, cls.att.sign);
  treestore_class_comment(tview, piter, cls.att.crem);
  treestore_class_alignment(tview, piter, cls.att.alignment);
  treestore_class_printable(tview, piter, cls.att.printable);
  treestore_class_fixed(tview, piter, cls.att.fixed);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - pointer to class name string
Output: - tree store updated
*/
static void
treestore_class_name(GtkTreeView *tview, GtkTreeIter piter, gchar *class_name)
{
  treestore_leaf_string(tview, piter, class_name, "(class name)", "", 
			XPLAB_MAINTENANCE_NODE_ID_CLASS_NAME,
			XPLAB_MAINTENANCE_EDIT_CLASS_NAME &
			XPLAB_MAINTENANCE_EDITABLE);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - pointer to class translation string
Output: - tree store updated
*/
static void
treestore_class_translation(GtkTreeView *tview, GtkTreeIter piter,
			    gchar *class_translation)
{
  treestore_leaf_string(tview, piter, class_translation, 
			"(class translation)", "", 
			XPLAB_MAINTENANCE_NODE_ID_CLASS_TRANSLATION,
			XPLAB_MAINTENANCE_EDIT_CLASS_TRANSLATION & 
			XPLAB_MAINTENANCE_EDITABLE);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - pointer to class comment string
Output: - tree store updated
*/
static void
treestore_class_comment(GtkTreeView *tview, GtkTreeIter piter,
			gchar *class_comment)
{
  treestore_leaf_string(tview, piter, class_comment, "(comment)", "", 
			XPLAB_MAINTENANCE_NODE_ID_CLASS_COMMENT,
			XPLAB_MAINTENANCE_EDIT_CLASS_COMMENT & 
			XPLAB_MAINTENANCE_EDITABLE);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - class alignment
Output: - tree store updated
*/
static void
treestore_class_alignment(GtkTreeView *tview, GtkTreeIter piter,
			  ALIGMNT class_alignment)
{
  char *astring = NULL;

  aligmnt_to_string(class_alignment, &astring);

  treestore_leaf_string(tview, piter, astring, "(alignment)", "", 
			XPLAB_MAINTENANCE_NODE_ID_CLASS_ALIGNMENT,
			XPLAB_MAINTENANCE_EDIT_CLASS_ALIGNMENT &
			XPLAB_MAINTENANCE_EDITABLE);

  FREEMEM(astring);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - class printable
Output: - tree store updated
*/
static void
treestore_class_printable(GtkTreeView *tview, GtkTreeIter piter, int printable)
{
  treestore_leaf_string(tview, piter, printable?"TRUE":"FALSE", 
			"(printable)", "", 
			XPLAB_MAINTENANCE_NODE_ID_CLASS_PRINTABLE,
			XPLAB_MAINTENANCE_EDIT_CLASS_PRINTABLE &
			XPLAB_MAINTENANCE_EDITABLE);
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - class fixed
Output: - tree store updated
*/
static void
treestore_class_fixed(GtkTreeView *tview, GtkTreeIter piter, int fixed)
{
  treestore_leaf_string(tview, piter, fixed?"TRUE":"FALSE", 
			"(fixed)", "", 
			XPLAB_MAINTENANCE_NODE_ID_CLASS_FIXED,
			XPLAB_MAINTENANCE_EDIT_CLASS_FIXED &
			XPLAB_MAINTENANCE_EDITABLE);
}



/* Scans an iter for children having flag set.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree model
        - pointer to parent iter
Output: - parent iter updated
        - node flag set
Postcondition: - if returned false, parent iter unchanged, else parent iter
                 set to node set child iter.
*/
static gboolean
scan_rows(GtkTreeModel *tmodel, GtkTreeIter *parent)
{
  gint        node_flag, n = 0;
  gboolean    node_flag_set = FALSE, is_iter;
  GtkTreeIter child;

  while (1) {
    is_iter =                                     /* scan overwriteable rows */
      gtk_tree_model_iter_nth_child(tmodel, &child, parent, n++);

    if (is_iter == TRUE) {
      gtk_tree_model_get(tmodel, &child,
			 XPLAB_MAINTENANCE_TREE_STORE_NODE_FLAG, &node_flag,
			 -1);

      if (node_flag & XPLAB_MAINTENANCE_NODE_FLAG_SEL) {
	*parent = child;
	node_flag_set = TRUE;
	break;
      }
    }
    else {
      break;
    }
  }

  return node_flag_set;
}



/* Creates a tree node.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - pointer to graphic string
        - pointer to comment string
        - pointer to summary string
        - node ID
        - pointer to path string
Output: - pointer to child iter
Precondition : - a non-NULL path string initiates a root node and therefore
                 should contain only one number. If it doesn't fit, a root 
                 node is created, anyway.
Postcondition: - iter returned is child iter of parent iter.
Usage : The iter returned is not neccessarily a new iter but a usable one.
        This ensures to reload a DB into an existing tree store.
*/
static GtkTreeIter
treestore_node_string(GtkTreeView *tview, GtkTreeIter piter, char *graphic_txt,
		      char *comment_txt, char *summary_txt, int node_id,
		      gchar *path_string)
{
  gboolean     iter_valid = FALSE;
  GtkTreeIter  *parent_iter = &piter, first_iter;
  GtkTreePath  *path = NULL;
  GtkTreeStore *tstore;

  tstore = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(tview)));
  assert(GTK_IS_TREE_MODEL(tstore) == TRUE);

  if (path_string) {
    iter_valid = 
      gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(tstore),
					  &first_iter, path_string);

    if (iter_valid == FALSE) {                         /* node doesn't exist */
      parent_iter = NULL;                                    /* default node */
    }
    else {
      piter = first_iter;                                   /* existing node */
    }
  }
  else {                                               /* check availability */
    iter_valid = scan_rows(GTK_TREE_MODEL(tstore), &piter);
  }

  if (iter_valid == FALSE) {
    gtk_tree_store_append(tstore, &piter, parent_iter);
  }

  path = gtk_tree_model_get_path(GTK_TREE_MODEL(tstore), &piter);
  assert(path);

  maintenance_store_set(tview, piter, graphic_txt, comment_txt, summary_txt,
			node_id, node_icon(tview, GTK_TREE_MODEL(tstore), 
					   path, &piter));

  gtk_tree_path_free(path);

  return piter;
}



/* Creates a database tree leaf.

Revision: 30.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - pointer to graphic string
        - pointer to comment string
        - pointer to summary string
        - node ID
        - flag (0: node not selectable, else selectable) 
Output: - tree leaf
*/
static void
treestore_leaf_string(GtkTreeView *tview, GtkTreeIter piter, char *graphic_txt,
		      char *comment_txt, char *summary_txt, int node_id, 
		      int is_editable)
{
  gint         d = 0;
  gboolean     tree_valid;
  GdkPixbuf    *icon;
  GtkTreeIter  parent_iter = piter;
  GtkTreeStore *tstore;

  tstore = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(tview)));
  assert(GTK_IS_TREE_MODEL(tstore) == TRUE);

  if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button)) == TRUE) {
    d = 1;
  }

  if (is_editable) {
    icon = d?mt.mpixbuf.toedit_write:mt.mpixbuf.toedit_read;
  }
  else {
    icon = mt.mpixbuf.unedit;
  }

  tree_valid = scan_rows(GTK_TREE_MODEL(tstore), &piter);

  if (tree_valid == FALSE) {                             /* no existing row */
    gtk_tree_store_append(tstore, &piter, &parent_iter);
  }

  maintenance_store_set(tview, piter, graphic_txt, comment_txt, summary_txt,
			node_id, icon);
}



/* Sets a row in the maintenance store.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree store
        - parent iter
        - pointer to graphic string
        - pointer to comment string
        - pointer to summary string
        - node ID
        - pointer to icon
Output: - tree store updated
*/
static void
maintenance_store_set(GtkTreeView *tview, GtkTreeIter piter, 
		      gchar *graphic_txt, gchar *comment_txt, gchar *summary_txt,
		      gint node_id, GdkPixbuf *pixbuf)
{
  gint         cols;
  gchar        *sum;
  gboolean     valid_utf;
  GtkTreeStore *tstore;

  assert(graphic_txt);
  assert(comment_txt);
  assert(summary_txt);
  assert(node_id);

  tstore = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(tview)));
  assert(GTK_IS_TREE_MODEL(tstore) == TRUE);

  for (cols = 0; cols < XPLAB_MAINTENANCE_TREE_STORE_COLUMNS; cols++) {
    switch (cols) {
    case XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC:

      if (pixbuf) {
	gtk_tree_store_set(tstore, &piter, cols,   /* Pixbuf to unref by GTK */
			   pixbuf, -1);
      }

      break;

    case XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT:
      valid_utf = g_utf8_validate(graphic_txt, -1, NULL);

      sum = (valid_utf == TRUE) ? 
                            g_strdup(graphic_txt) : xplab_to_utf(graphic_txt);

      gtk_tree_store_set(tstore, &piter, cols, 
			 sum?sum:XPLAB_UTF_CONVERSION_ERROR, -1);
      g_free(sum);
      break;

    case XPLAB_MAINTENANCE_TREE_STORE_COMMENT:
      valid_utf = g_utf8_validate(comment_txt, -1, NULL);

      sum = (valid_utf == TRUE) ? 
                            g_strdup(comment_txt) : xplab_to_utf(comment_txt);

      gtk_tree_store_set(tstore, &piter, cols,
			 sum?sum:XPLAB_UTF_CONVERSION_ERROR, -1);
      g_free(sum);
      break;

    case XPLAB_MAINTENANCE_TREE_STORE_SUMMARY:
      valid_utf = g_utf8_validate(summary_txt, -1, NULL);

      sum = (valid_utf == TRUE) ? 
                            g_strdup(summary_txt) : xplab_to_utf(summary_txt);

      gtk_tree_store_set(tstore, &piter, cols,
			 sum?sum:XPLAB_UTF_CONVERSION_ERROR, -1);
      g_free(sum);

      break;

    case XPLAB_MAINTENANCE_TREE_STORE_NODE_ID:
      gtk_tree_store_set(tstore, &piter, cols, node_id, -1);
      break;

    case XPLAB_MAINTENANCE_TREE_STORE_NODE_FLAG:
      gtk_tree_store_set(tstore, &piter, cols,
			 XPLAB_MAINTENANCE_NODE_FLAG_DEFAULT, -1);
      break;

    default:
      assert(0);
      break;
    }
  }
}



/* Sets flag for all NB references.

Revision: 10.12.2005 E.Sack

Input : - pointer to tree model
        - TPS index
        - class index
        - NB index
        - pointer to flag (0: don't center focus in window, else center)
        - pointer to flag (0: don't collapse tree view, else collapse)
Output: - tree model updated
*/
static int
reference_nb(GtkTreeModel *model, gint tdx, gint cdx, gint ndx, gint *moveto, 
	     gint *link_reset)
{
  query_chain qc;

  init_query_chain(&qc);

  qc.tdx = tdx;
  qc.cdx = cdx;
  qc.ndx = ndx;
  qc.aux = *moveto;                                /* linked cursor set flag */
  qc.aux2 = *link_reset;                           /* linked collapse flag   */

  gtk_tree_model_foreach(model, (GtkTreeModelForeachFunc)reference_nb_node,
			 (gpointer)&qc);

  *moveto = qc.aux;
  *link_reset = qc.aux2;

  return qc.sum;
}



/* Sets color flag for a NB reference node.

Revision: 10.12.2005 E.Sack

Input : - GTK convention
Output: - tree view updated
*/
static gboolean
reference_nb_node(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN)
{
  gint        tps_index, class_index, nb_index, node_id;
  GtkTreeIter tps_iter, cls_iter, nb_iter;
  query_chain *qc = (query_chain *)data;

  STSFY_GCC(path);

  assert(qc);
  assert(qc->tdx>=0);
  assert(qc->cdx>=0);

  gtk_tree_model_get(model, iter,
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

  if (node_id == XPLAB_MAINTENANCE_NODE_ID_NB_INDEX) {
    tps_index = child_has_tps_index(model, *iter, &tps_iter);

    if (tps_index < 0 || tps_index != qc->tdx) {       /* TPS always queried */
      return FALSE;
    }

    class_index = child_has_class_index(model, *iter, &cls_iter);

    if (class_index < 0 || class_index != qc->cdx) { /* class always queried */
      return FALSE;
    }

    if (qc->ndx >= 0) {               /* negative values switch off NB query */
      nb_index = child_has_nb_index(model, *iter, &nb_iter);

      if (nb_index < 0 || nb_index != qc->ndx) {
	return FALSE;
      }
    }

    qc->sum++;                                        /* class/NB pair found */

    if (qc->aux2) {                         /* tree view collapsed only once */
      qc->aux2 = 0;

      gtk_tree_view_collapse_all(GTK_TREE_VIEW(mt.treeview));
    }

    expand_to_root(GTK_TREE_VIEW(mt.treeview), *iter, FALSE);

    if (qc->aux) {                                   /* cursor set only once */
      qc->aux = 0;
      tree_view_set_cursor(GTK_TREE_VIEW(mt.treeview), *iter);
    }
  }

  return FALSE;                                           /* trace all nodes */
}



/* Wrapper for gtk_tree_view_set_cursor.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree view
Output: - iter for cursor
*/
static void
tree_view_set_cursor(GtkTreeView *tree_view, GtkTreeIter iter)
{
  GtkTreePath       *path;
  GtkTreeModel      *cmodel;
  GtkTreeViewColumn *fcolumn = NULL;

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(tree_view));

  path = gtk_tree_model_get_path(cmodel, &iter);
  assert(path);

  gtk_tree_view_set_cursor(tree_view, path, fcolumn, FALSE);

  gtk_tree_path_free(path);
}



/* Expands all parent nodes to the root.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree view
        - iter
        - flag (include iter for expanding)
Output: - tree view updated
*/
static void
expand_to_root(GtkTreeView *tree_view, GtkTreeIter iter, gboolean including)
{
  gboolean     has_parent;
  GtkTreeIter  child;
  GtkTreePath  *path;
  GtkTreeModel *cmodel;

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(tree_view));

  path = gtk_tree_model_get_path(cmodel, &iter);

  if (including == FALSE) {
    has_parent = gtk_tree_model_iter_parent(cmodel, &child, &iter);

    if (has_parent == TRUE) {
      gtk_tree_path_free(path);
      path = gtk_tree_model_get_path(cmodel, &child);
    }
  }

  assert(path);
  gtk_tree_view_expand_to_path(tree_view, path);
  gtk_tree_path_free(path);

  return;
}



/* Sets color flag for all chain references.

Revision: 10.12.2005 E.Sack

Input : - pointer to tree model
        - flag (0: NNB query, 1: CNB query)
        - TPS index
        - chain class index
        - chain NB index
        - pointer to flag (0: don't center focus in window, else center)
        - pointer to flag (0: don't collapse tree view, else collapse)
Output: - tree model updated
        - sum of items
*/
static int
reference_chain(GtkTreeModel *model, int is_cnb, int tdx, int cdx, int ndx, 
		int *moveto, int *link_reset)
{
  query_chain qc;

  init_query_chain(&qc);

  qc.is_cnb = is_cnb;
  qc.tdx = tdx;
  qc.cdx = cdx;
  qc.ndx = ndx;
  qc.aux = *moveto;                                /* linked cursor set flag */
  qc.aux2 = *link_reset;                           /* linked collapse flag   */

  gtk_tree_model_foreach(model, (GtkTreeModelForeachFunc)reference_chain_node,
			 (gpointer)&qc);

  *moveto = qc.aux;
  *link_reset = qc.aux2;

  return qc.sum;
}



/* Sets color flag for a chain reference node.

Revision: 10.12.2005 E.Sack

Input : - GTK convention
Output: - tree view updated
*/
static gboolean
reference_chain_node(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN)
{
  gint        tps_index, chain_class_index, chain_nb_index, node_id;
  query_chain *qc = (query_chain *)data;
  GtkTreeIter tps_iter;

  STSFY_GCC(path);

  assert(qc);
  assert(qc->tdx>=0);
  assert(qc->cdx>=0);
  assert(qc->is_cnb==1 || !qc->is_cnb);

  gtk_tree_model_get(model, iter,
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

  if (node_id == (qc->is_cnb?XPLAB_MAINTENANCE_NODE_ID_CNB_INDEX:
		             XPLAB_MAINTENANCE_NODE_ID_NNB_INDEX)) {
    tps_index = child_has_tps_index(model, *iter, &tps_iter);

    if (tps_index < 0 || tps_index != qc->tdx) {       /* TPS always queried */
      return FALSE;
    }

    chain_class_index = parent_has_chain_class_index(model, *iter);

    if (chain_class_index < 0 || chain_class_index != qc->cdx) {/* class query always used */
      return FALSE;
    }

    if (qc->ndx >= 0) {               /* negative values switch off NB query */
      chain_nb_index = parent_has_chain_nb_index(model, *iter);

      if (chain_nb_index < 0 || chain_nb_index != qc->ndx) {
	return FALSE;
      }
    }

    qc->sum++;                                                 /* query hits */

    if (qc->aux2) {                         /* tree view collapsed only once */
      qc->aux2 = 0;

      gtk_tree_view_collapse_all(GTK_TREE_VIEW(mt.treeview));
    }

    expand_to_root(GTK_TREE_VIEW(mt.treeview), *iter, FALSE);

    if (qc->aux) {                                   /* cursor set only once */
      qc->aux = 0;
      tree_view_set_cursor(GTK_TREE_VIEW(mt.treeview), *iter);
    }
  }

  return FALSE;                                           /* trace all nodes */
}



/* Sets color flag for chain references of all classes.

Revision: 10.12.2005 E.Sack

Input : - pointer to tree model
        - TPS index
        - class list
        - pointer to flag (0: don't center focus in window, else center)
        - pointer to flag (0: don't collapse tree view, else collapse)
Output: - tree model updated
*/
static void
reference_cls_lst(GtkTreeModel *model, gint tps_idx, cls_lst clst, int *moveto, 
		  gint *link_reset)
{
  int cls_idx;

  for (cls_idx = 0; cls_idx < clst.cnt; cls_idx++) {
    reference_nb_chain(model, tps_idx, cls_idx, -1, moveto, link_reset);
  }
}



/* Sets color flag for chain references of all NB of a list.

Revision: 10.12.2005 E.Sack

Input : - pointer to tree model
        - TPS index
        - class list
        - class index
        - pointer to flag (0: don't center focus in window, else center)
        - pointer to flag (0: don't collapse tree view, else collapse)
Output: - tree model updated
*/
static void
reference_nb_lst(GtkTreeModel *model, gint tps_idx, cls_lst clst, int cls_idx, 
		 int *moveto, int *link_reset)
{
  int nb_idx;

  for (nb_idx = 0; nb_idx < clst.cls[cls_idx].nbl.cnt; nb_idx++) {
    reference_nb_chain(model, tps_idx, cls_idx, nb_idx, moveto, link_reset);
  }
}



/* Marks rows for all NB chain references.

Revision: 10.12.2005 E.Sack

Input : - pointer to tree model
        - TPS index
        - chain class index
        - chain NB index
        - pointer to flag (0: don't center focus in window, else center)
        - pointer to flag (0: don't collapse tree view, else collapse)
Output: - tree model updated
        - sum of items
*/
static int
reference_nb_chain(GtkTreeModel *model, int tps_idx, int cls_idx, int nb_idx,
		   int *moveto, int *link_reset)
{
  int  cnb_sum, nnb_sum, sum;
  char *buf = NULL;

  nnb_sum = reference_chain(model, 0, tps_idx, cls_idx, nb_idx, moveto, link_reset);
  cnb_sum = reference_chain(model, 1, tps_idx, cls_idx, nb_idx, moveto, link_reset);

  if (nb_idx >= 0) {
    int_to_string("class/NB index pair (", cls_idx, ",", &buf);
    int_to_string("", nb_idx, ") is ", &buf);
  }
  else {
    int_to_string("class index ", cls_idx, " is ", &buf);
  }

  sum = cnb_sum + nnb_sum;

  if (sum > 0) {
    add_string("referenced in ", &buf);

    if (cnb_sum>0) {
      int_to_string("", cnb_sum, " CNB list", &buf);
      add_string(cnb_sum==1?"":"s", &buf);

      if (nnb_sum>0) {
	add_string(" and in ", &buf);
      }
    }

    if (nnb_sum>0) {
      int_to_string("", nnb_sum, " NNB list", &buf);
      add_string(nnb_sum==1?"":"s", &buf);
    }

    add_string("\n", &buf);
  }
  else {
    add_string("nowhere referenced\n", &buf);
  }

  xplab_text_insert(main_log, buf);

  FREEMEM(buf);

  return sum;
}



/* Returns node ID and iter from cursor.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree view
        - pointer to iter
Output: - iter updated
        - node ID
*/
static gint
cursor_get_node_id(GtkTreeView *tview, GtkTreeIter *iter)
{
  gboolean          is_path;
  gint              node_id = XPLAB_MAINTENANCE_NODE_ID_UNDEF;
  GtkTreePath       *path;
  GtkTreeModel      *cmodel;
  GtkTreeViewColumn *fcolumn;

  gtk_tree_view_get_cursor(tview, &path, &fcolumn);

  if (path) {
    cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));
    is_path = gtk_tree_model_get_iter(cmodel, iter, path);
    assert(is_path == TRUE);
    gtk_tree_path_free(path);

    gtk_tree_model_get(cmodel, iter,
		       XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);
  }

  return node_id;
}



/* Searches for a NB index node in a tree model.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree model
        - TPS index
        - class index
        - NB index
Output: - index node
*/
static treenode_def
xsearch_nb_index(GtkTreeModel *cmodel, gint tps_idx, gint cls_idx, gint nb_idx)
{
  treenode_def tnode;

  init_treenode_def(&tnode);

  tnode = xsearch_class_index(cmodel, tps_idx, cls_idx);

  if (tnode.baux == TRUE) {
    tnode.baux = search_nb_index(cmodel, &tnode.piter, nb_idx);
  }

  return tnode;
}



/* Sets cursor to a database index node in a tree model.

Revision: 15.01.2006 E.Sack

Input : - pointer to tree model
        - TPS index
        - class index
        - NB index
Output: - index node
*/
static void
index_to_cursor(gint tps_idx, gint cls_idx, gint nb_idx)
{
  treenode_def treenode;
  GtkTreeModel *cmodel;

  cb_collapse_all(NULL, NULL);

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  treenode = xsearch_nb_index(cmodel, tps_idx, cls_idx, nb_idx);
  assert(treenode.baux == TRUE);

  expand_to_root(GTK_TREE_VIEW(mt.treeview), treenode.piter, FALSE);

  tree_view_set_cursor(GTK_TREE_VIEW(mt.treeview), treenode.piter);
}



/* Searches for a class index node in a tree model.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree model
        - TPS index
        - class index
Output: - index node
*/
static treenode_def
xsearch_class_index(GtkTreeModel *cmodel, int tps_idx, int cls_idx)
{
  treenode_def tnode;

  init_treenode_def(&tnode);

  tnode = xsearch_tps_index(cmodel, tps_idx);

  if (tnode.baux == TRUE) {
    tnode.baux = search_class_index(cmodel, &tnode.piter, cls_idx);
  }

  return tnode;
}



/* Searches for a TPS index node in a tree model.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree model
        - TPS index
Output: - index node
*/
static treenode_def
xsearch_tps_index(GtkTreeModel *cmodel, int tps_index)
{
  treenode_def tnode;

  init_treenode_def(&tnode);

  tnode.iaux = tps_index;

  gtk_tree_model_foreach(cmodel, (GtkTreeModelForeachFunc)search_tps_index,
			 (gpointer)&tnode);

  return tnode;
}



/* Searches for a TPS index node.

Revision: 31.07.2005 E.Sack

Input : - GTK convention
Output: - search results
*/
static gboolean
search_tps_index(GTK_CONVENTION_GtkTreeModelForeachFunc_RTS_GBOOLEAN)
{
  gint         node_id, tps_index;
  gchar        *row_text;
  treenode_def *tnode = (treenode_def *)data;

  STSFY_GCC(path);

  assert(tnode);

  tnode->baux = FALSE;

  gtk_tree_model_get(model, iter, 
		     XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

  if (node_id == XPLAB_MAINTENANCE_NODE_ID_TPS_INDEX) {
    gtk_tree_model_get(model, iter,
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, 
		       &row_text, -1);
    assert(row_text);
    tps_index = atoi(row_text);
    g_free(row_text);
    assert(tps_index >= 0);

    if (tps_index == tnode->iaux) {
      tnode->piter = *iter;
      tnode->baux = TRUE;
    }
  }

  return tnode->baux;
}



/* Searches for a class index node.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree model
        - pointer to iter
        - class index
Output: - iter updated
        - flag (FALSE: not found, else found)
Postcondition: Search below or down the given iter.
*/
static gboolean
search_class_index(GtkTreeModel *model, GtkTreeIter *io_iter, gint cls_idx)
{
  gint        node_id, cls_index;
  gchar       *row_text;
  gboolean    found = FALSE, has_child, has_next;
  GtkTreeIter citer = *io_iter, child;

  while (1) {
    gtk_tree_model_get(model, &citer,
		       XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

    if (node_id == XPLAB_MAINTENANCE_NODE_ID_CLASS_INDEX) {
      gtk_tree_model_get(model, &citer,
			 XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, &row_text,
			 -1);
      assert(row_text);
      cls_index = atoi(row_text);
      g_free(row_text);
      assert(cls_index >= 0);

      if (cls_index == cls_idx) {
	*io_iter = citer;
	found = TRUE;
	break;
      }
    }

    has_child = gtk_tree_model_iter_children(model, &child, &citer);

    if (has_child == TRUE) {
      found = search_class_index(model, &child, cls_idx);

      if (found == TRUE) {
	*io_iter = child;
	break;
      }
    }

    has_next = gtk_tree_model_iter_next(model, &citer);

    if (has_next == FALSE) {
      break;
    }
  } 

  return found;
}



/* Searches for a NB index node.

Revision: 31.07.2005 E.Sack

Input : - pointer to tree model
        - pointer to iter
        - NB index
Output: - iter updated
        - flag (FALSE: not found, else found)
Postcondition: Search below or down the given iter.
*/
static gboolean
search_nb_index(GtkTreeModel *model, GtkTreeIter *io_iter, gint nb_idx)
{
  gboolean found = FALSE, has_child, has_next;
  gint node_id, nb_index;
  GtkTreeIter citer = *io_iter, child;
  gchar *row_text;

  while (1) {
    gtk_tree_model_get(model, &citer,
		       XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

    if (node_id == XPLAB_MAINTENANCE_NODE_ID_NB_INDEX) {
      gtk_tree_model_get(model, &citer,
			 XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, &row_text,
			 -1);
      assert(row_text);
      nb_index = atoi(row_text);
      g_free(row_text);
      assert(nb_index >= 0);

      if (nb_index == nb_idx) {
	*io_iter = citer;
	found = TRUE;
	break;
      }
    }

    has_child = gtk_tree_model_iter_children(model, &child, &citer);

    if (has_child == TRUE) {
      found = search_nb_index(model, &child, nb_idx);

      if (found == TRUE) {
	*io_iter = child;
	break;
      }
    }

    has_next = gtk_tree_model_iter_next(model, &citer);

    if (has_next == FALSE) {
      break;
    }
  } 

  return found;
}



/* Creates a dialog or performs an action directly for maintenance.

Revision: 20.08.2005 E.Sack

Input : - pointer to dialog settings
        - query flag
Output: - dialog
Usage : Either a dialog is created using a common interface or an action is
        performed directly in cases, a dialog is not necessary. 
*/
static void
dialog_row_edit(dialog_edit_def *edt, int do_action)
{
  gboolean raised;

  assert(edt);
  assert(edt->eset.func);

  if (do_action != XPLAB_EDIT_DIALOG) {   /* edit directly, no common dialog */
    edt->eset.func(edt->eset.piter, NULL, do_action);
  }
  else {                                              /* use a common dialog */
    raised = xplab_window_raise(&edt->edialog, 1);

    if (raised == FALSE) {                        /* create dialog only once */
      edt->edialog = gtk_dialog_new();              /* set the common dialog */
      gtk_window_set_modal(GTK_WINDOW(edt->edialog), TRUE);

      gtk_button_box_set_layout(GTK_BUTTON_BOX(GTK_DIALOG(edt->edialog)->action_area),
				GTK_BUTTONBOX_SPREAD);

      /* set the maintenance dialog title */
      gtk_window_set_title(GTK_WINDOW(edt->edialog), 
			   XPLAB_MAINTENANCE_EDIT_TITLE);

      dialog_row_edit_upper(edt);        /* set the maintenance dialog input */
      dialog_row_edit_lower(edt);       /* set the maintenance dialog action */

      gtk_widget_show_all(GTK_WIDGET(edt->edialog));

      g_signal_connect(G_OBJECT(edt->edialog), "delete-event", 
		       G_CALLBACK(cb_dialog_delete_edit), NULL);
    }
  }
}



/* Creates the edit dialog upper part.

Revision: 04.01.2004 E.Sack

Input : - pointer to container box
Output: - content updated
*/
static void
dialog_row_edit_upper(dialog_edit_def *edt)
{
  GtkWidget *vbox, *frame, *fbox;

  /* create a vertical box */
  vbox = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(edt->edialog)->vbox), vbox);

  /* put into vertical box a frame */
  frame = gtk_frame_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  /* put into frame a vertical box */
  fbox = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER(fbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(frame), fbox);

  /* put into vertical box a label */
  edt->einfoline = gtk_label_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), edt->einfoline, FALSE, FALSE, 0);

  edt->eset.func(edt->eset.piter, frame, XPLAB_EDIT_DIALOG_FRAME);
  edt->eset.func(edt->eset.piter, fbox, XPLAB_EDIT_DIALOG_UPPER);
}



/* Creates the dialog lower part.

Revision: 20.08.2005 E.Sack

Input : - pointer to edit set
Output: - -
*/
static void
dialog_row_edit_lower(dialog_edit_def *edt)
{
  guint16     rows, cols;
  gbutton_def gd;
  GtkWidget   *table;

  /* put into container a table                                              */
  table = gtk_table_new(1, 1, FALSE);                         /* start table */
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(edt->edialog)->action_area), table);

  mt.cancel_hook_func = NULL;

  /* set up individual dialog buttons                                        */
  edt->eset.func(edt->eset.piter, table, XPLAB_EDIT_DIALOG_LOWER);

  rows = GTK_TABLE(table)->nrows;
  cols = GTK_TABLE(table)->ncols;

  gtk_table_resize(GTK_TABLE(table), rows+1, cols);          /* expand table */

  /* put into table a common cancel button                                   */
  gd = xset_clicked_button(NULL, NULL, DIALOG_BUTTON_CANCEL_BIG, 
			   DESC1_CANCEL_ACTION, DESC2_CANCEL_ACTION, NULL);
  edt->cancel_button = gd.button;

  gtk_table_attach(GTK_TABLE(table), gd.ebox, 0, cols, rows, 
		   rows+1, GTK_FILL, GTK_FILL, 0, 0);

  if (mt.cancel_hook_func != NULL) {           /* individual CB called first */
    g_signal_connect(G_OBJECT(edt->cancel_button), "clicked", 
		     G_CALLBACK(mt.cancel_hook_func), 
		     mt.cancel_hook_data);
  }

  g_signal_connect(G_OBJECT(edt->cancel_button), "clicked", 
		   G_CALLBACK(cb_dialog_edit_cancel), (gpointer)edt);
}



/* Deletes top window.

Revision: 04.01.2004 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_dialog_delete_edit(GTK_CONVENTION_DIALOG_RTS_GBOOLEAN)
{
  STSFY_GCC(event);
  STSFY_GCC(widget);

  assert(!user_data);

  cb_dialog_edit_cancel(NULL, (gpointer)&mt.dialog.edt);
  cb_dialog_edit_cancel(NULL, (gpointer)&mt.dialog.ccf.edt);

  return FALSE;
}



/* Destroys dialog.

Revision: 31.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_edit_cancel(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  dialog_edit_def *edt;

  STSFY_GCC(button);

  assert(user_data);

  edt = (dialog_edit_def *)user_data;

  if (GTK_IS_WIDGET(edt->edialog) == TRUE) {
    xplab_window_raise(&mt.top_maintenance, xctrl.cfg.raise);
    xplab_widget_destroy(&edt->edialog);
  }
}



/* Edit function.

Revision: 31.07.2005 E.Sack

Input : - XPLAB convention
Output: - tree view updated
*/
static void
edit_func_tps_def(XPLAB_MAINTENANCE_EDIT_IF)
{
  gint         tps_idx, printable;
  gchar        *buf = NULL;
  gboolean     db_locked;
  GtkWidget    *label;
  GtkTreeIter  tps_iter;
  GtkTreeModel *cmodel;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_TPS_INDEX);

  tps_idx = child_has_tps_index(cmodel, piter, &tps_iter);
  assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    /* put into vertical box a label */
    int_to_string("TPS index: ", tps_idx, " (test pattern sets: ", &buf);
    int_to_string("", mt.mdb.tpl.cnt, ")", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    /* put into vertical box a label */
    cat_string("TPS ID: ", mt.mdb.tpl.tps[tps_idx].ident, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    /* put into vertical box a label */
    int_to_string("classes: ", mt.mdb.tpl.tps[tps_idx].cl.cnt, "", &buf);
    printable = getstr_printable_classes(mt.mdb.tpl.tps[tps_idx].cl);
    int_to_string(" (", printable, " printable)", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_erase), 
			       DIALOG_BUTTON_ERASE, DESC1_MAINTENANCE_TPS_ERASE,
			       DESC2_MAINTENANCE_TPS_ERASE,
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "TPS");

    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_ERASE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, db_locked);    /* use writeable DB */

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_erase), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* erase pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_ERASE:
    erase_tps_def(&mt.mdb.tpl, tps_idx, mt.mlabel);
    maintenance_tree_store_fill(mt.treeview, TRUE);
    mt.items_deleted++;

    break;

  case XPLAB_EDIT_ACTION_EXTRACT:
    xextract_tps_def(mt.mdb, tps_idx);

    break;

  case XPLAB_QUERY_IMAGE:
    create_image_window(tps_idx, XPLAB_IDX_NONE, XPLAB_IDX_NONE);
    break;

  case XPLAB_EDIT_ACTION_DUMP:
    xdump_all_classes(mt.mdb, tps_idx, xctrl.cfg.nb_border_x, 
		      xctrl.cfg.nb_border_y);

    break;

  case XPLAB_QUERY_ACTION_DUMP:      /* how many files will be dumped to HDD */
    {
      GList *child;

      child = gtk_container_get_children(GTK_CONTAINER(mt.row_dump));
      xplab_label_set_text(GTK_WIDGET(child->data), MENU_BUTTON_DUMP " (1)");

      g_list_free(child);
    }

    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, FALSE);   /* no references */

    gtk_widget_set_sensitive(mt.row_edit, db_locked);    /* use writeable DB */

    gtk_widget_set_sensitive(mt.row_extract, TRUE);       /* always possible */
    gtk_widget_set_sensitive(mt.row_dump, TRUE);          /* always possible */

    break;

  default: break;
  }
}



/* Builds an image window.

Revision: 11.01.2006 E.Sack

Input : - TPS index
        - class index
        - NB index
Output: - -
*/
static void
create_image_window(gint tps_idx, gint cls_idx, gint nb_idx)
{
  gchar      *window_title = NULL;
  gboolean   need_new = TRUE;
  query_mimg img_found;
  mimg_def   *mdef = NULL;
  area_def   far;
  img_ext    img_tmp;

  assert(tps_idx >= 0);

  init_query_mimg(&img_found);
  init_area_def(&far);
  init_img_ext(&img_tmp);

  img_found.tps_idx = tps_idx;                                /* setup query */
  img_found.cls_idx = cls_idx;
  img_found.nb_idx  = nb_idx;

  g_list_foreach(mt.mimg, search_image_window, &img_found);         /* query */

  if (img_found.aux == TRUE) {
    return;                                /* image window already displayed */
  }

  img_tmp = assemble_image(tps_idx, cls_idx, nb_idx,      /* original window */
			   &window_title);

  if (img_tmp.eimg.col == NULL) {
    free_img_ext(&img_tmp);
    string_to_log(main_log, "cannot create image", "", "\n");
  }
  else {
    if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.image_static)) == TRUE) {
      GList    *glist_tmp;
      mimg_def *mdef_tmp = NULL;

      glist_tmp = g_list_last(mt.mimg);                    /* get last entry */

      if (glist_tmp) {                                     /* list not empty */
	mdef_tmp = (mimg_def *)glist_tmp->data;         /* last entry member */ 
	assert(mdef_tmp);

	if (GTK_IS_WIDGET(mdef_tmp->mtop) == TRUE) {
	  free_img_ext(&mdef_tmp->pimg);     /* free old image, nothing else */
	  mdef = mdef_tmp;                      /* pointer copy - don't free */
	  xplab_window_raise(&mdef->mtop, 1);    /* old window may be hidden */
	  gtk_window_set_title(GTK_WINDOW(mdef->mtop), window_title);
	  need_new = FALSE;
	}
      }
    }

    if (need_new == TRUE) {
      CALLOC(mimg_def, 1, mdef);                /* create a new GList member */
      init_mimg_def(mdef);                                    /* init member */
    }

    mdef->tps_idx = tps_idx;                                /* setup indexes */
    mdef->cls_idx = cls_idx;
    mdef->nb_idx  = nb_idx;
    mdef->pimg    = img_tmp;                    /* pointer copy - don't free */

    if (need_new == TRUE) {
      mdef->mtop =                                 /* setup new image window */
	xplab_image_window(window_title, NULL, NULL, 
			   &mdef->scroll, mdef->pimg.eimg.ar,
			   GTK_SIGNAL_FUNC(cb_image_actions), NULL,
			   GTK_SIGNAL_FUNC(cb_image_expose),
			   GTK_SIGNAL_FUNC(cb_image_delete), 0, far);

      g_object_set_data(G_OBJECT(mdef->mtop), XPLAB_ASSN_KEY_0, 
			(gpointer)mdef);         /* data bound to top window */

      gtk_widget_show_all(mdef->mtop);

      mt.mimg = g_list_append(mt.mimg, (gpointer)mdef);    /* append to list */
    }
    else {
      rebuild_image_window(mdef,                  /* re-use old image window */
			   GTK_SIGNAL_FUNC(cb_image_actions), 
			   GTK_SIGNAL_FUNC(cb_image_expose));
    }
  }

  FREEMEM(window_title);
}



/* Validates non-negative indexes.

Revision: 12.01.2006 E.Sack

Input : - TPS list
        - TPS index
        - class index
        - NB index
Output: - flag (TRUE: valid indexes, else invalid indexes detected)
Precondition: - the first negative index stops validation.
*/
static XPLAB_INDEX_TYPE
db_indexes(tps_lst tpl, gint tps_idx, gint cls_idx, gint nb_idx)
{
  if (tps_idx < 0) {
    return XPLAB_IDX_NONE;                             /* no valid TPS index */
  }

  assert(tpl.cnt >= 0);                                 /* consistency check */

  if (tpl.cnt <= tps_idx) {
    return XPLAB_IDX_NONE;                             /* no valid TPS index */
  }

  if (cls_idx < 0) {
    return XPLAB_IDX_TPS;                                 /* valid TPS index */
  }

  assert(tpl.tps[tps_idx].cl.cnt >= 0);                 /* consistency check */

  if (tpl.tps[tps_idx].cl.cnt <= cls_idx) {
    return XPLAB_IDX_NONE;                           /* no valid class index */
  }

  if (nb_idx < 0) {
    return XPLAB_IDX_CLS;                               /* valid class index */
  }

  assert(tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt >= 0);/* consistency check */

  if (tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt <= nb_idx) {
    return XPLAB_IDX_NONE;                           /* no valid class index */
  }

  return XPLAB_IDX_NB;                                     /* valid NB index */
}



/* Returns the original image according to valid indexes.

Revision: 11.01.2006 E.Sack

Input : - TPS index
        - class index
        - NB index
        - pointer to string
Output: - image
        - string updated
*/
static img_ext
assemble_image(gint tps_idx, gint cls_idx, gint nb_idx, gchar **window_title)
{
  img_ext          img_tmp;
  XPLAB_INDEX_TYPE valid_idx;

  init_img_ext(&img_tmp);

  valid_idx = db_indexes(mt.mdb.tpl, tps_idx, cls_idx, nb_idx);/* index check*/

  switch (valid_idx) {                            /* different image sources */
  case XPLAB_IDX_TPS:                                                 /* TPS */
    if (!mt.mdb.tpl.tps[tps_idx].cl.cnt) {               /* empty class list */
      return img_tmp;
    }

    img_tmp = tps_image(mt.mdb.tpl.tps[tps_idx].cl, 0,
			xctrl.cfg.nb_border_x, xctrl.cfg.nb_border_y);

    if (!AREA_SIZE(img_tmp.eimg.ar)) {
      return img_tmp;
    }

    check_message_tps("", tps_idx, window_title);

    break;

  case XPLAB_IDX_CLS:                                               /* class */
    if (!mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt) { /* empty NB list */
      return img_tmp;
    }

    img_tmp.eimg = xplab_sample_image(mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx], 0,
				      xctrl.cfg.nb_border_x, 
				      xctrl.cfg.nb_border_y);

    check_message_class("", tps_idx, cls_idx, window_title);

    break;

  case XPLAB_IDX_NB:                                                   /* NB */
    copy_pic_def(&mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].pat,
		 &img_tmp.eimg);

    check_message_nb("", tps_idx, cls_idx, nb_idx, window_title);

    break;

  default: break;                                          /* no valid index */
  }

  convert_to_color(&img_tmp.eimg, TRUE);       /* color image for rendering, */
                                               /* BMP image for marking.     */

  return img_tmp;
}



/* Edits DB items.

Revision: 05.01.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_edit_erase(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  dialog_edit_def edt;

  STSFY_GCC(button);

  assert(user_data);

  edt = *(dialog_edit_def*)user_data;                          /* use copy */

  dialog_row_edit_action(edt, XPLAB_EDIT_ACTION_ERASE);
}



/* Triggers edit actions.

Revision: 01.01.2006 E.Sack

Input : - edit setting
        - action flag
Output: - -
*/
static void
dialog_row_edit_action(dialog_edit_def edt, int do_action)
{
  if (GTK_IS_TREE_VIEW(mt.treeview) == TRUE) {
    assert(edt.eset.func);
    edt.eset.func(edt.eset.piter, NULL, do_action);
  }
}



/* Edits DB items.

Revision: 05.01.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_edit_create(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  dialog_edit_def edt;

  STSFY_GCC(button);

  assert(user_data);

  edt = *(dialog_edit_def*)user_data;                          /* use copy */

  dialog_row_edit_action(edt, XPLAB_EDIT_ACTION_CREATE);
}



/* Dumps CCF results.

Revision: 06.03.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_ccf_dump(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  dialog_edit_def ccf;

  STSFY_GCC(button);

  assert(user_data);

  ccf = *(dialog_edit_def*)user_data;                          /* use copy */

  dialog_row_edit_action(ccf, XPLAB_EDIT_ACTION_DUMP);
}



/* Stops CCF results printing.

Revision: 20.03.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_ccf_stop(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  dialog_edit_def ccf;

  STSFY_GCC(button);

  assert(user_data);

  ccf = *(dialog_edit_def*)user_data;                          /* use copy */

  dialog_row_edit_action(ccf, XPLAB_EDIT_ACTION_STOP);
}



/* Edits DB items.

Revision: 20.02.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_edit_copy(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  dialog_edit_def edt;

  STSFY_GCC(button);

  assert(user_data);

  edt = *(dialog_edit_def*)user_data;                          /* use copy */

  dialog_row_edit_action(edt, XPLAB_EDIT_ACTION_COPY);
}



/* Edits DB items.

Revision: 29.02.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_edit_sort_name(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  dialog_edit_def edt;

  STSFY_GCC(button);

  assert(user_data);

  edt = *(dialog_edit_def*)user_data;                          /* use sort */

  dialog_row_edit_action(edt, XPLAB_EDIT_ACTION_SORT_BY_NAME);
}



/* Edits DB items.

Revision: 29.02.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_edit_sort_trans(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  dialog_edit_def edt;

  STSFY_GCC(button);

  assert(user_data);

  edt = *(dialog_edit_def*)user_data;                          /* use sort */

  dialog_row_edit_action(edt, XPLAB_EDIT_ACTION_SORT_BY_TRANS);
}



/* Edits DB items.

Revision: 21.02.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dialog_edit_paste(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  dialog_edit_def edt;

  STSFY_GCC(button);

  assert(user_data);

  edt = *(dialog_edit_def*)user_data;                          /* use copy */

  dialog_row_edit_action(edt, XPLAB_EDIT_ACTION_PASTE);
}



/* Edit function.

Revision: 31.07.2005 E.Sack

Input : - XPLAB convention
Output: - tree store updated
        - flag according to case (-1: no case found)
Usage:  Definition (only reported here): An edit function shall provide all 
        individual procedures for a certain database tree node instead having 
        individual functions for that a lot. The main goal is to lower 
        code maintenance because the edit function contains all procedures in
        one place in opposition to individual functions for these procedures,
        which are distributed in the C-modul and are nevertheless quite similar.
*/
static void
edit_func_tps_lst(XPLAB_MAINTENANCE_EDIT_IF)
{
  gchar        *buf = NULL;
  gboolean     db_locked;
  GtkWidget    *label;
  GtkTreeModel *cmodel;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_TPS_LIST);

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    /* put into vertical box a label */
    int_to_string("test pattern sets: ", mt.mdb.tpl.cnt, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_create), 
			       DIALOG_BUTTON_CREATE, 
			       DESC1_MAINTENANCE_TPS_CREATE,
			       DESC2_MAINTENANCE_TPS_CREATE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "TPS list");

    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_CREATE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, db_locked);    /* use writeable DB */

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_create), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* clear pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_CREATE:                      /* create an empty TPS */
    {
      gint         tps_idx;
      treenode_def tpsnode;
      GtkTreePath  *path;

      FREEMEM(buf);                                  /* force default TPS ID */

      tps_idx = create_tps_def(&mt.mdb.tpl, &buf);       /* create empty TPS */

      if (tps_idx < 0) {
	break;                                       /* something went wrong */
      }

      xdefault_media_att(&mt.mdb.tpl.tps[tps_idx].media);  /* default values */

      maintenance_tree_store_fill(mt.treeview, TRUE);

      tpsnode = xsearch_tps_index(cmodel, tps_idx);
      assert(tpsnode.baux == TRUE);                  /* there must be a node */

      expand_to_root(GTK_TREE_VIEW(mt.treeview), tpsnode.piter, FALSE);

      path = gtk_tree_model_get_path(cmodel, &tpsnode.piter);
      gtk_tree_view_expand_row(GTK_TREE_VIEW(mt.treeview), path, TRUE);
      gtk_tree_path_free(path);

      mt.items_created++;
      FREEMEM(buf);
    }
    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, FALSE);   /* no references */

    gtk_widget_set_sensitive(mt.row_edit, db_locked);

    gtk_widget_set_sensitive(mt.row_extract, FALSE);
    gtk_widget_set_sensitive(mt.row_dump, FALSE);

    break;

  default: break;
  }
}



/* Edit function.

Revision: 31.07.2005 E.Sack

Input : - XPLAB convention
Output: - tree view updated
        - flag according to case (-1: no case found)
*/
static void
edit_func_cls_def(XPLAB_MAINTENANCE_EDIT_IF)
{
  gint         d, tps_idx, cls_idx;
  gchar        *buf = NULL;
  gboolean     found, db_locked;
  GtkWidget    *label;
  GtkTreeIter  tps_iter, cls_iter;
  GtkTreeModel *cmodel;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_CLASS_INDEX);

  tps_idx = child_has_tps_index(cmodel, piter, &tps_iter);
  assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

  cls_idx = child_has_class_index(cmodel, piter, &cls_iter);
  assert(cls_idx >= 0 && cls_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    int_to_string("TPS index: ", tps_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    /* put into vertical box a label */
    int_to_string("class index: ", cls_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    /* put into vertical box a label */
    add_string("class name: ", &buf);
    add_string(mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.class, &buf);

    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);
    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gtk_table_resize(GTK_TABLE(fbox), 1, 2);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_erase), 
			       DIALOG_BUTTON_ERASE, DESC1_MAINTENANCE_CSS_ERASE,
			       DESC2_MAINTENANCE_CSS_ERASE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_paste), 
			       DIALOG_BUTTON_PASTE, DESC1_MAINTENANCE_CSS_PASTE,
			       DESC2_MAINTENANCE_CSS_PASTE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 1, 2, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, mt.copy_nbl.cnt ? TRUE : FALSE);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "class");
    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_ERASE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, db_locked);    /* use writeable DB */

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_erase), (gpointer)&mt.dialog.edt);

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_PASTE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && mt.copy_nbl.cnt) ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_paste), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* clear pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_PASTE:
    {
      found = search_below(cmodel, &piter, XPLAB_MAINTENANCE_NODE_ID_NB_LIST);
      assert(found == TRUE);

      edit_func_nb_lst(piter, NULL, XPLAB_EDIT_ACTION_PASTE);
    }

    break;

  case XPLAB_EDIT_ACTION_ERASE:
    d = erase_cls_def(&mt.mdb.tpl.tps[tps_idx].cl, cls_idx, mt.mlabel);

    if (d) {
      maintenance_tree_store_fill(mt.treeview, TRUE);
      mt.items_deleted++;
    }

    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    {
      int moveto = 1, link_reset = 1;

      reference_nb_chain(cmodel, tps_idx, cls_idx, -1, &moveto, &link_reset);
    }

    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, TRUE);

    gtk_widget_set_sensitive(mt.row_edit, db_locked);

    gtk_widget_set_sensitive(mt.row_extract, TRUE);
    gtk_widget_set_sensitive(mt.row_dump, TRUE);

    break;

  case XPLAB_EDIT_ACTION_EXTRACT:
    xextract_cls_def(mt.mdb, tps_idx, cls_idx);
    break;

  case XPLAB_QUERY_IMAGE:
    create_image_window(tps_idx, cls_idx, XPLAB_IDX_NONE);
    break;

  case XPLAB_EDIT_ACTION_DUMP:
    xdump_all_nb(mt.mdb, tps_idx, cls_idx, xctrl.cfg.nb_border_x, 
		 xctrl.cfg.nb_border_y);
    break;

  case XPLAB_QUERY_ACTION_DUMP:      /* how many files will be dumped to HDD */
    {
      GList *child;

      child = gtk_container_get_children(GTK_CONTAINER(mt.row_dump));
      xplab_label_set_text(GTK_WIDGET(child->data), MENU_BUTTON_DUMP " (1)");

      g_list_free(child);
    }

    break;

  default: break;
  }
}



/* Edit function.

Revision: 10.12.2005 E.Sack

Input : - XPLAB convention
Output: - tree view updated
        - flag according to case (-1: no case found)
*/
static void
edit_func_cls_lst(XPLAB_MAINTENANCE_EDIT_IF)
{
  gint         d, tps_idx;
  gchar        *buf = NULL;
  gboolean     db_locked;
  GtkWidget    *label;
  GtkTreeIter  tps_iter;
  GtkTreeModel *cmodel;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_CLASS_LIST);

  tps_idx = child_has_tps_index(cmodel, piter, &tps_iter);
  assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    /* put into vertical box a label */
    int_to_string("classes: ", mt.mdb.tpl.tps[tps_idx].cl.cnt, "", &buf);

    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gtk_table_resize(GTK_TABLE(fbox), 2, 2);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_erase), 
			       DIALOG_BUTTON_ERASE, DESC1_MAINTENANCE_CLS_ERASE,
			       DESC2_MAINTENANCE_CLS_ERASE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, 
			       mt.mdb.tpl.tps[tps_idx].cl.cnt
			       ? TRUE : FALSE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_create), 
			       DIALOG_BUTTON_CREATE, 
			       DESC1_MAINTENANCE_CLS_CREATE,
			       DESC2_MAINTENANCE_CLS_CREATE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 1, 2, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_sort_name),
			       DIALOG_BUTTON_SORT_NAME, 
			       DESC1_MAINTENANCE_CLS_SORT_NAME,
			       DESC2_MAINTENANCE_CLS_SORT_NAME, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 1, 2,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, 
			       mt.mdb.tpl.tps[tps_idx].cl.cnt
			       ? TRUE : FALSE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_sort_trans), 
			       DIALOG_BUTTON_SORT_TRANS, 
			       DESC1_MAINTENANCE_CLS_SORT_TRANS,
			       DESC2_MAINTENANCE_CLS_SORT_TRANS, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 1, 2, 1, 2,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, 
			       mt.mdb.tpl.tps[tps_idx].cl.cnt
			       ? TRUE : FALSE);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "class list");
    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_ERASE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && mt.mdb.tpl.tps[tps_idx].cl.cnt)
			       ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_erase), (gpointer)&mt.dialog.edt);

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_CREATE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, db_locked);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_create), (gpointer)&mt.dialog.edt);

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_SORT_NAME);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && mt.mdb.tpl.tps[tps_idx].cl.cnt)
			       ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_sort_name), (gpointer)&mt.dialog.edt);

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_SORT_TRANS);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && mt.mdb.tpl.tps[tps_idx].cl.cnt)
			       ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_sort_trans), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* clear pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_CREATE:                    /* create an empty class */
    {
      int          cls_idx;
      treenode_def cls_node;
      GtkTreePath  *path;

      cls_idx = create_cls_def(&mt.mdb.tpl.tps[tps_idx].cl, NULL);

      if (cls_idx < 0) {
	break;                                       /* something went wrong */
      }

      maintenance_tree_store_fill(mt.treeview, TRUE);

      cls_node = xsearch_class_index(cmodel, tps_idx, cls_idx);
      assert(cls_node.baux == TRUE);                 /* there must be a node */

      expand_to_root(GTK_TREE_VIEW(mt.treeview), cls_node.piter, FALSE);

      path = gtk_tree_model_get_path(cmodel, &cls_node.piter);
      gtk_tree_view_expand_row(GTK_TREE_VIEW(mt.treeview), path, TRUE);
      gtk_tree_path_free(path);

      mt.items_created++;
    }

    break;

  case XPLAB_EDIT_ACTION_ERASE:
    free_cls_lst(&mt.mdb.tpl.tps[tps_idx].cl);
    maintenance_tree_store_fill(mt.treeview, TRUE);

    mt.items_deleted++;
    break;

  case XPLAB_EDIT_ACTION_SORT_BY_NAME:
    if (mt.mdb.tpl.tps[tps_idx].cl.cnt > 0) {
      d = class_sort(mt.mdb.tpl.tps[tps_idx].cl.cnt, 
		     mt.mdb.tpl.tps[tps_idx].cl.cls, 0);
      if (d) {
	maintenance_tree_store_fill(mt.treeview, TRUE);
	mt.items_changed++;
      }
    }
    else {
      string_to_log(main_log, "empty class list\n", "", "");
    }

    break;

  case XPLAB_EDIT_ACTION_SORT_BY_TRANS:
    if (mt.mdb.tpl.tps[tps_idx].cl.cnt > 0) {
      d = class_sort(mt.mdb.tpl.tps[tps_idx].cl.cnt, 
		     mt.mdb.tpl.tps[tps_idx].cl.cls, 1);
      if (d) {
	maintenance_tree_store_fill(mt.treeview, TRUE);
	mt.items_changed++;
      }
    }
    else {
      string_to_log(main_log, "empty class list\n", "", "");
    }

    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    {
      int moveto = 1, link_reset = 1;

      reference_cls_lst(cmodel, tps_idx, mt.mdb.tpl.tps[tps_idx].cl, &moveto,
			&link_reset);
    }

    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, TRUE);

    gtk_widget_set_sensitive(mt.row_edit, db_locked);

    gtk_widget_set_sensitive(mt.row_extract, FALSE);
    gtk_widget_set_sensitive(mt.row_dump, TRUE);

    break;

  case XPLAB_EDIT_ACTION_DUMP:
    {
      int c;

      for (c = 0; c < mt.mdb.tpl.tps[tps_idx].cl.cnt; c++) {
	xdump_all_nb(mt.mdb, tps_idx, c, xctrl.cfg.nb_border_x, 
		     xctrl.cfg.nb_border_y);
      }
    }

    break;

  case XPLAB_QUERY_ACTION_DUMP:      /* how many files will be dumped to HDD */
    {
      GList *child;

      child = gtk_container_get_children(GTK_CONTAINER(mt.row_dump));
      assert(GTK_IS_LABEL(child->data) == TRUE);

      int_to_string(MENU_BUTTON_DUMP " (", 
		    mt.mdb.tpl.tps[tps_idx].cl.cnt, ")", &buf);

      xplab_label_set_text(GTK_WIDGET(child->data), buf);
      g_list_free(child);
      FREEMEM(buf);
    }

    break;

  default: break;
  }
}



/* Edit function.

Revision: 10.12.2005 E.Sack

Input : - XPLAB convention
Output: - tree view updated
        - flag according to case (-1: no case found)
*/
static void
edit_func_nb_lst(XPLAB_MAINTENANCE_EDIT_IF)
{
  gint         tps_idx, cls_idx;
  gchar        *buf = NULL;
  gboolean     db_locked;
  GtkWidget    *label;
  GtkTreeIter  cls_iter, tps_iter;
  GtkTreePath  *path;
  GtkTreeModel *cmodel;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_NB_LIST);

  tps_idx = child_has_tps_index(cmodel, piter, &tps_iter);
  assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

  cls_idx = child_has_class_index(cmodel, piter, &cls_iter);
  assert(cls_idx >= 0 && cls_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    /* put into vertical box a label */
    int_to_string("NBs: ", mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);
    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gtk_table_resize(GTK_TABLE(fbox), 1, 3);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_erase), 
			       DIALOG_BUTTON_ERASE, DESC1_MAINTENANCE_NBL_ERASE,
			       DESC2_MAINTENANCE_NBL_ERASE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, 
			       mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt
			       ? TRUE : FALSE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_create), 
			       DIALOG_BUTTON_CREATE, 
			       DESC1_MAINTENANCE_NBL_CREATE,
			       DESC2_MAINTENANCE_NBL_CREATE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 1, 2, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_paste), 
			       DIALOG_BUTTON_PASTE, DESC1_MAINTENANCE_NBL_PASTE,
			       DESC2_MAINTENANCE_NBL_PASTE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 2, 3, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, mt.copy_nbl.cnt ? TRUE : FALSE);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "NB list");
    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_ERASE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt)
			       ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_erase), (gpointer)&mt.dialog.edt);

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_CREATE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, db_locked);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_create), (gpointer)&mt.dialog.edt);

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_PASTE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && mt.copy_nbl.cnt) ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_paste), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* clear pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_PASTE:
    {
      idx_def idef;

      idef = paste_nb(cmodel, mt.copy_nbl, piter, mt.mdb.tpl);

      if (idef.tps_idx >= 0) {
	treenode_def treenode;

	assert(idef.cls_idx >= 0);
	assert(idef.nb_idx >= 0);

	maintenance_tree_store_fill(mt.treeview, TRUE);

	treenode = xsearch_nb_index(cmodel, idef.tps_idx, idef.cls_idx, 
				    idef.nb_idx);
	assert(treenode.baux == TRUE);

	expand_to_root(GTK_TREE_VIEW(mt.treeview), treenode.piter, FALSE);

	path = gtk_tree_model_get_path(cmodel, &treenode.piter);
	gtk_tree_view_expand_row(GTK_TREE_VIEW(mt.treeview), path, TRUE);
	gtk_tree_path_free(path);

	mt.items_created++;
      }
      else {
	xplab_label_set_text(mt.dialog.edt.einfoline, "paste failed");
	return;
      }
    }

    break;

  case XPLAB_EDIT_ACTION_CREATE:/* makes no sense to create without training */
    string_to_log(main_log, "new NB to be trained separately (--> training phase)", "", "\n");

    break;

  case XPLAB_EDIT_ACTION_ERASE:
    erase_nb_lst(mt.mdb.tpl.tps[tps_idx].cl, cls_idx, 1);
    maintenance_tree_store_fill(mt.treeview, TRUE);
    mt.items_deleted++;

    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    {
      int moveto = 1, link_reset = 1;

      reference_nb_lst(cmodel, tps_idx, mt.mdb.tpl.tps[tps_idx].cl, cls_idx, 
		       &moveto, &link_reset);
    }

    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, TRUE);

    gtk_widget_set_sensitive(mt.row_edit, db_locked);

    gtk_widget_set_sensitive(mt.row_extract, FALSE);
    gtk_widget_set_sensitive(mt.row_dump, TRUE);

    break;

  case XPLAB_EDIT_ACTION_DUMP:
    {
      int n;

      for (n = 0; n < mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt; n++) {
	xdump_nb_def(mt.mdb, tps_idx, cls_idx, n, xctrl.cfg.nb_border_x, 
		     xctrl.cfg.nb_border_y);
      }
    }

    break;

  case XPLAB_QUERY_ACTION_DUMP:      /* how many files will be dumped to HDD */
    {
      GList *child;

      child = gtk_container_get_children(GTK_CONTAINER(mt.row_dump));
      assert(GTK_IS_LABEL(child->data) == TRUE);

      int_to_string(MENU_BUTTON_DUMP " (", 
		    mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt, ")", &buf);

      xplab_label_set_text(GTK_WIDGET(child->data), buf);
      g_list_free(child);
      FREEMEM(buf);
    }

    break;

  default: break;
  }

  return;
}



/* Edit function.

Revision: 22.12.2005 E.Sack

Input : - XPLAB convention
Output: - tree view updated
        - flag according to case (-1: no case found)
*/
static void
edit_func_nb_def(XPLAB_MAINTENANCE_EDIT_IF)
{
  gint         tps_idx, cls_idx, nb_idx, dst_nnr;
  gchar        *idx_coord = NULL, *buf = NULL;
  gboolean     db_locked;
  GtkWidget    *label;
  GtkTreeModel *cmodel;
  GtkTreeIter  tps_iter, cls_iter, nb_iter;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_NB_INDEX);

  tps_idx = child_has_tps_index(cmodel, piter, &tps_iter);
  assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

  cls_idx = child_has_class_index(cmodel, piter, &cls_iter);
  assert(cls_idx >= 0 && cls_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

  nb_idx = child_has_nb_index(cmodel, piter, &nb_iter);
  assert(nb_idx >= 0 && nb_idx < mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt);

  int_to_string("NB[", tps_idx, "]", &idx_coord);
  int_to_string("[", cls_idx, "]", &idx_coord);
  int_to_string("[", nb_idx, "]", &idx_coord);

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    label = gtk_label_new(idx_coord);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gtk_table_resize(GTK_TABLE(fbox), 1, 2);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_erase), 
			       DIALOG_BUTTON_ERASE, DESC1_MAINTENANCE_NB_ERASE,
			       DESC2_MAINTENANCE_NB_ERASE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, 
			       mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt ?
			       TRUE : FALSE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_copy), 
			       DIALOG_BUTTON_COPY, DESC1_MAINTENANCE_NB_COPY,
			       DESC2_MAINTENANCE_NB_COPY, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 1, 2, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "NB");

    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_ERASE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt)
			       ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_erase), (gpointer)&mt.dialog.edt);

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_COPY);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, TRUE);   /* copy doesn't change DB */

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_copy), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* clear pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_COPY:
    free_nb_lst(&mt.copy_nbl);

    dst_nnr = cat_nb_def(mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx], 
			 &mt.copy_nbl, TRUE);        /* NB looses all chains */
    assert(dst_nnr == 0);

    string_to_log(main_log, "copied for paste: ", idx_coord, " (class name '");
    string_to_log(main_log, mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].att.class, 
		  "", "')\n");

    break;

  case XPLAB_EDIT_ACTION_ERASE:
    erase_nb_def(mt.mdb.tpl.tps[tps_idx].cl, cls_idx, nb_idx, NULL, 1);
    maintenance_tree_store_fill(mt.treeview, TRUE);
    mt.items_deleted++;

    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    {
      int moveto = 1, link_reset = 1;

      reference_nb_chain(cmodel, tps_idx, cls_idx, nb_idx, &moveto, &link_reset);
    }

    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, TRUE);

    gtk_widget_set_sensitive(mt.row_edit, db_locked);

    gtk_widget_set_sensitive(mt.row_extract, TRUE);
    gtk_widget_set_sensitive(mt.row_dump, TRUE);

    break;

  case XPLAB_EDIT_ACTION_EXTRACT:
    xextract_nb_def(mt.mdb, tps_idx, cls_idx, nb_idx);

    break;

  case XPLAB_QUERY_IMAGE:
    create_image_window(tps_idx, cls_idx, nb_idx);
    break;

  case XPLAB_EDIT_ACTION_DUMP:
    xdump_nb_def(mt.mdb, tps_idx, cls_idx, nb_idx, xctrl.cfg.nb_border_x, 
		 xctrl.cfg.nb_border_y);
    break;

  case XPLAB_QUERY_ACTION_DUMP:      /* how many files will be dumped to HDD */
    {
      GList *child;

      child = gtk_container_get_children(GTK_CONTAINER(mt.row_dump));
      assert(GTK_IS_LABEL(child->data) == TRUE);

      xplab_label_set_text(GTK_WIDGET(child->data), MENU_BUTTON_DUMP " (1)");
      g_list_free(child);
    }

    break;

  default: break;
  }

  FREEMEM(buf);
  FREEMEM(idx_coord);
}



/* Provides (mouse-) actions on a window.

Revision: 09.01.2006 E.Sack

Input : - GTK convention
Output: - propagation flag
Usage.  - CB meant for image windows.
*/
static gboolean
cb_image_actions(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN)
{
  area_def  far;
  gpointer  image_data_path = NULL;
  GtkWidget *top_level;

  STSFY_GCC(user_data);

  init_area_def(&far);

  if (GTK_IS_WIDGET(widget) == FALSE) {
    return TRUE;
  }

  if (event->type != GDK_BUTTON_PRESS) {
    return TRUE;
  }

  top_level = gtk_widget_get_toplevel(widget);

  if (GTK_IS_OBJECT(top_level) == FALSE) {
    return FALSE;
  }

  image_data_path = g_object_get_data(G_OBJECT(top_level), XPLAB_ASSN_KEY_0);

  if (image_data_path == NULL) {
    return TRUE;
  }

  assert(((mimg_def *)image_data_path)->mtop == top_level);

  switch (event->button) {
  case 1: 
    image_window_popup(event, image_data_path);
    break;

  case 3: 
    image_cursor(event, (mimg_def *)image_data_path);
    break;

  default: break;
  }

  return TRUE;
}



/* Highlights a selected NB in the image window by the mouse pointer.

Revision: 15.01.2006 E.Sack

Input : - pointer to GDK event
        - pointer to data path
Output: - -
*/
static void
image_cursor(GdkEventButton *event, mimg_def *mdef)
{
  int              cnr, nnr, tnr, n, within;
  c_def            mouse_ptr;
  sector_def       sec;
  query_mimg       query;
  XPLAB_INDEX_TYPE valid_idx;

  assert(event);
  assert(mdef);

  init_query_mimg(&query);

  tnr = mdef->tps_idx;                                        /* short names */
  cnr = mdef->cls_idx;
  nnr = mdef->nb_idx;

  AREA_TO_SECTOR(mdef->pimg.eimg.ar, sec);                    /* whole image */

  within = image_coordinates(event, &mouse_ptr, sec, mdef->magn);

  valid_idx = db_indexes(mt.mdb.tpl, tnr, cnr, nnr);

  switch (valid_idx) {                             /* different list sources */
  case XPLAB_IDX_TPS:                                                 /* TPS */
    if (!within) {
      break;                                         /* clear all background */
    }

    for (n = 0; n < mdef->pimg.il.cnt; n++) {
      if (WITHIN_SECTOR(mouse_ptr.x, mouse_ptr.y, mdef->pimg.il.dat[n].sec)) {
	index_to_cursor(tnr, mdef->pimg.il.dat[n].cidx, 
			mdef->pimg.il.dat[n].nidx);
	query.tps_idx = tnr;                                  /* setup query */
	query.cls_idx = mdef->pimg.il.dat[n].cidx;
	query.nb_idx  = mdef->pimg.il.dat[n].nidx;
	break;
      }
    }

    break;

  case XPLAB_IDX_CLS:                                               /* class */
    if (!within) {
      query.tps_idx = tnr;                          /* highlight whole class */
      query.cls_idx = cnr;
      break;
    }

    for (n = 0; n < mt.mdb.tpl.tps[tnr].cl.cls[cnr].nbl.cnt; n++) {
      if (WITHIN_SECTOR(mouse_ptr.x, mouse_ptr.y, 
			mt.mdb.tpl.tps[tnr].cl.cls[cnr].nbl.nb[n].dat.sec)) {
	index_to_cursor(tnr, cnr, n);
	query.tps_idx = tnr;                                  /* setup query */
	query.cls_idx = cnr;
	query.nb_idx  = n;
	break;
      }
    }
    break;

  case XPLAB_IDX_NB:                                                   /* NB */
    index_to_cursor(tnr, cnr, nnr);

    query.tps_idx = tnr;                                      /* setup query */
    query.cls_idx = cnr;
    query.nb_idx  = nnr;
    break;

  default: return; break;                                  /* no valid index */
  }

  g_list_foreach(mt.mimg, color_image_window, &query);           /* coloring */
}


/* Colors or clears background of an image window.

Revision: 15.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
color_image_window(GTK_CONVENTION_GFunc_RTS_VOID) 
{
  int              cnr, nnr, tnr, n;
  gboolean         multiple_nb;
  sector_def       sec;
  sector_lst       secl;              /* multiple sectors can be highlighted */
  mimg_def         *memb = (mimg_def *)data;       /* list-member definition */
  query_mimg       *query = (query_mimg *)user_data;
  XPLAB_INDEX_TYPE valid_idx;

  init_sector_lst(&secl);
  init_sector_def(&sec);                       /* initialized to size nought */

  if (GTK_IS_WIDGET(memb->mtop) == FALSE) {
    return;
  }

  tnr = memb->tps_idx;                                        /* short names */
  cnr = memb->cls_idx;
  nnr = memb->nb_idx;
  multiple_nb = (query->nb_idx == XPLAB_IDX_NONE) ? TRUE : FALSE;

  /* sector of size nought clears background                                 */
  xprint_rb(memb->pimg.eimg.col, memb->pimg.eimg, mt.pc, sec, 0);

  valid_idx = db_indexes(mt.mdb.tpl, tnr, cnr, nnr);    /* member image type */

  switch (valid_idx) {
  case XPLAB_IDX_TPS:                                                 /* TPS */
    if (query->tps_idx != tnr) {
      break;                                                /* different TPS */
    }

    for (n = 0; n < memb->pimg.il.cnt; n++) {
      if (query->cls_idx == memb->pimg.il.dat[n].cidx &&
	  (query->nb_idx  == memb->pimg.il.dat[n].nidx || 
	   multiple_nb == TRUE)) {
	put_sector_lst(&secl, memb->pimg.il.dat[n].sec);/* NB sector in image */

	if (multiple_nb == FALSE) {
	  break;
	}
      }
    }

    break;

  case XPLAB_IDX_CLS:                                               /* class */
    if (query->tps_idx != tnr || 
	query->cls_idx != cnr) {
      break;                                              /* different class */
    }

    for (n = 0; n < mt.mdb.tpl.tps[tnr].cl.cls[cnr].nbl.cnt; n++) {
      if (query->nb_idx  == n || multiple_nb == TRUE) {
	put_sector_lst(&secl, mt.mdb.tpl.tps[tnr].cl.cls[cnr].nbl.nb[n].dat.sec);/* NB sector in image */

	if (multiple_nb == FALSE) {
	  break;
	}
      }
    }

    break;

  case XPLAB_IDX_NB:                                                   /* NB */
    if (query->tps_idx == tnr && 
	query->cls_idx == cnr &&
	(query->nb_idx  == nnr || multiple_nb == TRUE)) {
      AREA_TO_SECTOR(mt.mdb.tpl.tps[tnr].cl.cls[cnr].nbl.nb[nnr].pat.ar, sec);
      put_sector_lst(&secl, sec);                      /* NB sector in image */
    }                                          /* else different class or NB */

    break;

  default: break;                                          /* no valid index */
  }

  for (n = 0; n < secl.cnt; n++) {    /* single or multiple NBs to highlight */
    xprint_rb(memb->pimg.eimg.col, memb->pimg.eimg, mt.pc, secl.sdl[n], 1);
  }

  mimage_to_state(memb, memb->magn != XPLAB_MAGN_X1_VALUE); /* re-draw image */

  free_sector_lst(&secl);
}



/* Creates a pop-up menu for an image window.

Revision: 09.01.2006 E.Sack

Input : - pointer to mouse button event
        - pointer to imaga data path
Output: - -
*/
static void
image_window_popup(GdkEventButton *event, gpointer *image_data_path)
{
  GtkWidget *menu_item, *popmenu;

  assert(event);
  assert(image_data_path);

  popmenu = gtk_menu_new();                 /* standalone, maintained by GTK */

  menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_QUIT);
  gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);

  g_signal_connect(G_OBJECT(menu_item), "activate", 
		   G_CALLBACK(cb_image_quit), image_data_path);

  xplab_menu_magn(popmenu, GTK_SIGNAL_FUNC(cb_image_magnitude), 
		  image_data_path, ((mimg_def *)image_data_path)->magn);

  gtk_widget_show_all(popmenu);

  gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		 event->button, event->time);
}



/* Sets the magnitude of backing image pixels. The complete image buffer is
   simulated by allocating only its visible part.

Revision: 13.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_magnitude(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  mimg_def    *image_data_path;
  const gchar *name;

  if (gtk_check_menu_item_get_active(checkmenuitem) == TRUE) {
    gboolean is_act;
    GSList   *radio_grp;

    image_data_path = (mimg_def *)user_data;

    if (image_data_path == NULL) {
      return;
    }

    radio_grp =                                      /* don't free this list */
      gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(checkmenuitem));
    assert(radio_grp);

    do {                                               /* scan radio buttons */
      is_act = 
	gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(radio_grp->data));

      if (is_act == TRUE) {
	break;
      }

      radio_grp = radio_grp->next;
    } while (radio_grp);

    assert(is_act == TRUE);

    name =                                         /* don't free this string */
      gtk_label_get_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(radio_grp->data))));

    if (!strcmp(name, XPLAB_MAGN_X1_NAME)) {
      image_data_path->magn = XPLAB_MAGN_X1_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_X2_NAME)) {
      image_data_path->magn = XPLAB_MAGN_X2_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_X4_NAME)) {
      image_data_path->magn = XPLAB_MAGN_X4_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_D2_NAME)) {
      image_data_path->magn = XPLAB_MAGN_D2_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_D4_NAME)) {
      image_data_path->magn = XPLAB_MAGN_D4_VALUE;
    }
    else {
      return;                                     /* magnitude not supported */
    }

    rebuild_image_window(image_data_path, 
			 GTK_SIGNAL_FUNC(cb_image_actions), 
			 GTK_SIGNAL_FUNC(cb_image_expose));
  }
}



/* Redraws the original image window.

Revision: 08.01.2006 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_image_expose(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN)
{
  mimg_def  *image_data_path = NULL;
  GtkWidget *top_level;

  STSFY_GCC(event);

  assert(user_data==NULL);
  assert(widget);

  if (GTK_IS_WIDGET(widget) == FALSE) {
    return FALSE;
  }

  top_level = gtk_widget_get_toplevel(widget);

  if (GTK_IS_OBJECT(top_level) == FALSE) {
    return FALSE;
  }

  image_data_path = (mimg_def *)g_object_get_data(G_OBJECT(top_level), 
						  XPLAB_ASSN_KEY_0);

  if (image_data_path == NULL) {
    return FALSE;
  }

  mimage_to_state(image_data_path, image_data_path->magn > 0.0);

  return FALSE;
}



/* Creates backing image.

Revision: 08.01.2006 E.Sack

Input : - pointer to image data
        - flag (0: don't reallocate image buffer, 1: reallocate)
Output: - image buffer and window updated
*/
static void
mimage_to_state(mimg_def *image_data_path, gboolean redraw_buf)
{
  if (GTK_IS_WIDGET(image_data_path->mtop) == FALSE) {
    return;
  }

  if (image_data_path->pimg.eimg.col == NULL) {
    return;
  }

  if (redraw_buf == TRUE) {
    magnitude_allocation(image_data_path->pimg.eimg.col, 
			 image_data_path->pimg.eimg.ar, 
			 image_data_path->scroll.h_adj->value, 
			 image_data_path->scroll.v_adj->value, 
			 image_data_path->magn,
			 &image_data_path->magn_img,
			 image_data_path->scroll.h_adj->page_size, 
			 image_data_path->scroll.v_adj->page_size, 
			 image_data_path->scroll.sdrawing->allocation.width,
			 image_data_path->scroll.sdrawing->allocation.height, 
			 1);

    image_data_path->magn_old = image_data_path->magn;
  }

  xplab_draw_image(image_data_path->pimg.eimg.col,
		   image_data_path->pimg.eimg.ar,
		   image_data_path->magn_img,
		   image_data_path->magn,
		   image_data_path->scroll.h_adj,
		   image_data_path->scroll.v_adj,
		   image_data_path->scroll.sdrawing);
}



/* Deletes image window.

Revision: 08.01.2006 E.Sack

Input : - GTK convention
Output: - propagation flag
Usage:  - the delete callback sometimes needs user data because of widget
          deletion and initialization problem. That means the normal
          widget pointer delivered by the GTK interface cannot be set to NULL
          because of copy-by-value. Use user data instead or (as here) 
          data bound to widget.
*/
static gboolean
cb_image_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN)
{
  mimg_def  *image_data_path = NULL;
  GtkWidget *top_level;

  STSFY_GCC(event);
  STSFY_GCC(user_data);

  if (GTK_IS_WIDGET(widget) == FALSE) {
    return FALSE;
  }

  top_level = gtk_widget_get_toplevel(widget);
  assert(GTK_IS_OBJECT(top_level) == TRUE);

  image_data_path = (mimg_def *)g_object_get_data(G_OBJECT(top_level),
						  XPLAB_ASSN_KEY_0);

  assert(image_data_path);
  assert(image_data_path->mtop == top_level);
  assert(image_data_path->mtop == widget);
  assert(GTK_IS_WIDGET(image_data_path->mtop) == TRUE);

  cb_image_quit(NULL, (gpointer)image_data_path);

  return FALSE;
}



/* Quits image window.

Revision: 09.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_quit(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  mimg_def *image_data_path = NULL;

  STSFY_GCC(button);

  image_data_path = (mimg_def *)user_data;

  assert(image_data_path);
  assert(GTK_IS_WIDGET(image_data_path->mtop) == TRUE);

  mt.mimg = g_list_remove(mt.mimg, (gconstpointer)image_data_path);/* remove from list */
  delete_image_window((gpointer)image_data_path, NULL);

  if (mt.mimg == NULL) {
    xplab_window_raise(&mt.top_maintenance, xctrl.cfg.raise);
  }
}



/* Searches for and raises a maintenance image window.

Revision: 07.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
search_image_window(GTK_CONVENTION_GFunc_RTS_VOID) 
{
  mimg_def   *mdef = (mimg_def *)data;                  /* member definition */
  query_mimg *img_found = (query_mimg *)user_data;

  if (GTK_IS_WIDGET(mdef->mtop) == FALSE || img_found->aux == TRUE) {
    return;
  }

  if (img_found->tps_idx == mdef->tps_idx &&
      img_found->cls_idx == mdef->cls_idx &&
      img_found->nb_idx  == mdef->nb_idx) {
    img_found->aux = TRUE;
    xplab_window_raise(&mdef->mtop, 1);
  }
}



/* Wrapper to update all image windows.

Revision: 12.01.2006 E.Sack

Input : - -
Output: - -
*/
static void
update_image_window_all(void)
{
  g_list_foreach(mt.mimg, update_image_window, NULL);
}



/* Updates an image window.

Revision: 12.01.2006 E.Sack

Input : - GTK convention
Output: - -
Postcondition: - leaves image window empty if no new image can be created. 
*/
static void 
update_image_window(GTK_CONVENTION_GFunc_RTS_VOID)
{
  mimg_def *mdef = (mimg_def *)data;                    /* member definition */
  gchar    *window_title = NULL;

  STSFY_GCC(user_data);

  if (GTK_IS_WIDGET(mdef->mtop) == FALSE) {
    return;
  }

  free_img_ext(&mdef->pimg);                               /* free old image */
  mdef->pimg = assemble_image(mdef->tps_idx, mdef->cls_idx, mdef->nb_idx,
			      &window_title);            /* create new image */
  FREEMEM(window_title);                      /* window title already exists */

  rebuild_image_window(mdef, 
		       GTK_SIGNAL_FUNC(cb_image_actions), 
		       GTK_SIGNAL_FUNC(cb_image_expose));
}



/* Exchanges the image in an image window.

Revision: 12.01.2006 E.Sack

Input : - pointer to data path
        - pointer to mouse CB
        - pointer to expose CB
Output: - -
*/
static void
rebuild_image_window(mimg_def *image_data_path, GtkSignalFunc mouse_cb, 
		     GtkSignalFunc expose_cb)
{
  area_def area;

  assert(image_data_path);

  if (GTK_IS_WIDGET(image_data_path->mtop) == FALSE) {
    return;
  }

  assert(GTK_IS_WIDGET(image_data_path->scroll.h_viewport) == TRUE);

  /* destroy viewport (drawing container) and associated signal handlers     */
  xplab_widget_destroy(&image_data_path->scroll.h_viewport);

  assert(GTK_IS_WIDGET(image_data_path->scroll.v_viewport) == TRUE);

  area = image_data_path->pimg.eimg.ar;                     /* original size */

  if (image_data_path->magn > 0.0) {
    area.width  *= image_data_path->magn;                 /* zooming defined */
    area.height *= image_data_path->magn;
  }

  image_data_path->scroll.sdrawing =   /* allocate new drawing in old window */
    horizontal_viewport(image_data_path->scroll.v_viewport,
			&image_data_path->scroll.h_viewport, 
			area, mouse_cb, expose_cb);
}



/* Edit function.

Revision: 10.12.2005 E.Sack

Input : - XPLAB convention
Output: - tree view updated
        - flag according to case (-1: no case found)
*/
static void
edit_func_cnb_lst(XPLAB_MAINTENANCE_EDIT_IF)
{
  gint         tps_idx, cls_idx, nb_idx;
  gchar        *buf = NULL;
  gboolean     db_locked;
  nb_chn_lst   parent;
  GtkWidget    *label;
  GtkTreeModel *cmodel;
  GtkTreeIter  tps_iter, cls_iter, nb_iter;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_CNB_LIST);

  tps_idx = child_has_tps_index(cmodel, piter, &tps_iter);
  assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

  cls_idx = child_has_class_index(cmodel, piter, &cls_iter);
  assert(cls_idx >= 0 && cls_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

  nb_idx = child_has_nb_index(cmodel, piter, &nb_iter);
  assert(nb_idx >= 0 && nb_idx < mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt);

  /* list of all parents (CNBs) of NB */
  parent = mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].cnb;

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    /* put into vertical box a label */
    int_to_string("TPS index: ", tps_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    int_to_string("class index: ", cls_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    int_to_string("NB index: ", nb_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    int_to_string("CNBs: ", parent.cnt, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);
    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_erase), 
			       DIALOG_BUTTON_ERASE, DESC1_MAINTENANCE_CNBL_ERASE,
			       DESC2_MAINTENANCE_CNBL_ERASE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, parent.cnt ? TRUE : FALSE);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "CNB list");

    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_ERASE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && parent.cnt) ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_erase), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* clear pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_ERASE:
    {
      erase_nb_chn_lst(mt.mdb.tpl.tps[tps_idx].cl, cls_idx, nb_idx,
		       mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].cnb.cd, 1);

      maintenance_tree_store_fill(mt.treeview, TRUE);

      mt.items_deleted++;
    }

    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    {
      int moveto = 1, link_reset = 1;

      reference_nb_chn_lst(cmodel, tps_idx, parent, &moveto, &link_reset);
    }

    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, TRUE);

    gtk_widget_set_sensitive(mt.row_edit, db_locked);

    gtk_widget_set_sensitive(mt.row_extract, FALSE);
    gtk_widget_set_sensitive(mt.row_dump, FALSE);

    break;

  default: break;
  }
}



/* Edit function.

Revision: 10.12.2005 E.Sack

Input : - XPLAB convention
Output: - tree store items updated
        - flag according to case (-1: no case found)
*/
static void
edit_func_cnb_def(XPLAB_MAINTENANCE_EDIT_IF)
{
  int          tps_idx, cls_idx, nb_idx, chn_idx;
  char         *buf = NULL;
  gboolean     db_locked;
  nb_chn_lst   parent;
  GtkWidget    *label;
  GtkTreeModel *cmodel;
  GtkTreeIter  tps_iter, cls_iter, nb_iter, cnb_iter;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_CNB_INDEX);

  tps_idx = child_has_tps_index(cmodel, piter, &tps_iter);
  assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

  cls_idx = child_has_class_index(cmodel, piter, &cls_iter);
  assert(cls_idx >= 0 && cls_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

  nb_idx = child_has_nb_index(cmodel, piter, &nb_iter);
  assert(nb_idx >= 0 && nb_idx < mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt);

  chn_idx = child_has_cnb_index(cmodel, piter, &cnb_iter);
  assert(chn_idx >= 0 && chn_idx < mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].cnb.cnt);

  /* list of all parents (CNBs) of NB */
  parent = mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].cnb;

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    /* put into vertical box a label */
    int_to_string("NB index: ", nb_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);
    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_erase), 
			       DIALOG_BUTTON_ERASE, DESC1_MAINTENANCE_CNB_ERASE,
			       DESC2_MAINTENANCE_CNB_ERASE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, parent.cnt ? TRUE : FALSE);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "CNB");

    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_ERASE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && parent.cnt) ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_erase), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* clear pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_ERASE:
    erase_nb_chn_def(mt.mdb.tpl.tps[tps_idx].cl, cls_idx, nb_idx, chn_idx,
		     mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].cnb.cd, 1);

    maintenance_tree_store_fill(mt.treeview, TRUE);

    mt.items_deleted++;

    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    {
      int moveto = 1, link_reset = 1;

      reference_class_nb(cmodel, tps_idx, parent.cd[chn_idx].cdx,
			 parent.cd[chn_idx].ndx, &moveto, &link_reset);
    }

    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, TRUE);

    gtk_widget_set_sensitive(mt.row_edit, db_locked);

    gtk_widget_set_sensitive(mt.row_extract, FALSE);
    gtk_widget_set_sensitive(mt.row_dump, FALSE);

    break;

  default: break;
  }
}



/* Sets color flag for all class/NB chain references.

Revision: 10.12.2005 E.Sack

Input : - pointer to tree model
        - TPS index
        - NB chain list
        - pointer to flag (0: don't center focus in window, else center)
        - pointer to flag (0: don't collapse tree view, else collapse)
Output: - tree model updated
*/
static void
reference_nb_chn_lst(GtkTreeModel *model, int tps_idx, nb_chn_lst chain_list, 
		     int *moveto, int *link_reset)
{
  int chn_idx;

  for (chn_idx = 0; chn_idx < chain_list.cnt; chn_idx++) {
    reference_class_nb(model, tps_idx, chain_list.cd[chn_idx].cdx,
		       chain_list.cd[chn_idx].ndx, moveto, link_reset);
  }
}



/* Sets color flag for all class/NB references.

Revision: 10.12.2005 E.Sack

Input : - pointer to tree model
        - TPS index
        - class index
        - NB index
        - pointer to flag (0: don't center focus in window, else center)
        - pointer to flag (0: don't collapse tree view, else collapse)
Output: - tree model updated
        - sum of items
*/
static int
reference_class_nb(GtkTreeModel *model, int tps_idx, int cls_idx, int nb_idx,
		   int *moveto, int *link_reset)
{
  int  nb_sum;
  char *buf = NULL;

  nb_sum = reference_nb(model, tps_idx, cls_idx, nb_idx, moveto, link_reset);
  assert(nb_sum<=1);

  switch (nb_sum) {
  case 0:
    int_to_string("no references for class/NB index pair (", cls_idx, ",", &buf);
    int_to_string("", nb_idx, ")\n", &buf);
    break;
  case 1:
    int_to_string("show class/NB index pair (", cls_idx, ",", &buf);
    int_to_string("", nb_idx, ")\n", &buf);
    break;
  default: assert(0);
    break;
  }

  xplab_text_insert(main_log, buf);

  FREEMEM(buf);

  return nb_sum;
}



/* Edit function.

Revision: 10.12.2005 E.Sack

Input : - XPLAB convention
Output: - tree view updated
        - flag according to case (-1: no case found)
*/
static void
edit_func_nnb_lst(XPLAB_MAINTENANCE_EDIT_IF)
{
  gint         tps_idx, cls_idx, nb_idx;
  gchar        *buf = NULL;
  gboolean     db_locked;
  nb_chn_lst   children;
  GtkWidget    *label;
  GtkTreeModel *cmodel;
  GtkTreeIter  tps_iter, cls_iter, nb_iter;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_NNB_LIST);

  tps_idx = child_has_tps_index(cmodel, piter, &tps_iter);
  assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

  cls_idx = child_has_class_index(cmodel, piter, &cls_iter);
  assert(cls_idx >= 0 && cls_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

  nb_idx = child_has_nb_index(cmodel, piter, &nb_iter);
  assert(nb_idx >= 0 && nb_idx < mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt);

  /* list of all children (NNBs) of NB */
  children = mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].nnb;

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    /* put into vertical box a label */
    int_to_string("TPS index: ", tps_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    int_to_string("class index: ", cls_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    int_to_string("NB index: ", nb_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);

    int_to_string("NNBs: ", children.cnt, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);
    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_erase), 
			       DIALOG_BUTTON_ERASE, DESC1_MAINTENANCE_NNBL_ERASE,
			       DESC2_MAINTENANCE_NNBL_ERASE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, children.cnt ? TRUE : FALSE);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "NNB list");
    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_ERASE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && children.cnt) ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_erase), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* clear pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_ERASE:
    {
      erase_nb_chn_lst(mt.mdb.tpl.tps[tps_idx].cl, cls_idx, nb_idx,
		       mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].nnb.cd, 1);

    maintenance_tree_store_fill(mt.treeview, TRUE);

      mt.items_deleted++;
    }

    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    {
      int moveto = 1, link_reset = 1;

      reference_nb_chn_lst(cmodel, tps_idx, children, &moveto, &link_reset);
    }

    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, TRUE);

    gtk_widget_set_sensitive(mt.row_edit, db_locked);

    gtk_widget_set_sensitive(mt.row_extract, FALSE);
    gtk_widget_set_sensitive(mt.row_dump, FALSE);

    break;

  default: break;
  }
}



/* Edit function.

Revision: 10.12.2005 E.Sack

Input : - XPLAB convention
Output: - tree view updated
        - flag according to case (-1: no case found)
*/
static void
edit_func_nnb_def(XPLAB_MAINTENANCE_EDIT_IF)
{
  gint         tps_idx, cls_idx, nb_idx, chn_idx;
  gchar        *buf = NULL;
  gboolean     db_locked;
  nb_chn_lst   children;
  GtkWidget    *label;
  GtkTreeModel *cmodel;
  GtkTreeIter  tps_iter, cls_iter, nb_iter, nnb_iter;

  db_locked = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt.lock_button));

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  assert_node_id(cmodel, piter, XPLAB_MAINTENANCE_NODE_ID_NNB_INDEX);

  tps_idx = child_has_tps_index(cmodel, piter, &tps_iter);
  assert(tps_idx >= 0 && tps_idx < mt.mdb.tpl.cnt);

  cls_idx = child_has_class_index(cmodel, piter, &cls_iter);
  assert(cls_idx >= 0 && cls_idx < mt.mdb.tpl.tps[tps_idx].cl.cnt);

  nb_idx = child_has_nb_index(cmodel, piter, &nb_iter);
  assert(nb_idx >= 0 && nb_idx < mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.cnt);

  chn_idx = child_has_nnb_index(cmodel, piter, &nnb_iter);
  assert(chn_idx >= 0 && 
	 chn_idx < mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].nnb.cnt);

  /* list of all children (CNBs) of NB */
  children = mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].nnb;

  switch (do_action) {
  case XPLAB_EDIT_DIALOG_UPPER:
    assert(GTK_IS_BOX(fbox) == TRUE);

    /* put into vertical box a label */
    int_to_string("NB index: ", nb_idx, "", &buf);
    label = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(fbox), label, FALSE, FALSE, 0);
    FREEMEM(buf);
    break;

  case XPLAB_EDIT_DIALOG_LOWER:
    {
      gbutton_def gd;

      assert(GTK_IS_TABLE(fbox) == TRUE);

      gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_dialog_edit_erase), 
			       DIALOG_BUTTON_ERASE, DESC1_MAINTENANCE_NNB_ERASE,
			       DESC2_MAINTENANCE_NNB_ERASE, 
			       (gpointer)&mt.dialog.edt);

      gtk_table_attach(GTK_TABLE(fbox), gd.ebox, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);

      g_signal_connect(G_OBJECT(gd.button), "clicked", 
		       G_CALLBACK(cb_dialog_edit_cancel), 
		       (gpointer)&mt.dialog.edt);

      gtk_widget_set_sensitive(gd.button, children.cnt ? TRUE : FALSE);
    }

    break;

  case XPLAB_EDIT_DIALOG_FRAME:
    assert(GTK_IS_FRAME(fbox) == TRUE);
    gtk_frame_set_label(GTK_FRAME(fbox), "NNB");

    break;

  case XPLAB_EDIT_MENU:
    if (mt.popevent) {
      GtkWidget *menu_item, *popmenu = NULL;

      popmenu = gtk_menu_new();

      menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_ERASE);
      gtk_menu_shell_append(GTK_MENU_SHELL(popmenu), menu_item);
      gtk_widget_set_sensitive(menu_item, 
			       (db_locked && children.cnt) ? TRUE : FALSE);

      g_signal_connect(G_OBJECT(menu_item), "activate", 
		       G_CALLBACK(cb_dialog_edit_erase), (gpointer)&mt.dialog.edt);

      gtk_widget_show_all(popmenu);

      gtk_menu_popup(GTK_MENU(popmenu), NULL, NULL, NULL, NULL,
		     mt.popevent->button.button, mt.popevent->button.time);

      mt.popevent = NULL;                              /* clear pointer copy */
    }

    break;

  case XPLAB_EDIT_ACTION_ERASE:
    erase_nb_chn_def(mt.mdb.tpl.tps[tps_idx].cl, cls_idx, nb_idx, chn_idx,
		     mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl.nb[nb_idx].nnb.cd, 1);

    maintenance_tree_store_fill(mt.treeview, TRUE);

    mt.items_deleted++;
    break;

  case XPLAB_EDIT_ACTION_WHERE_REFERENCED:
    {
      int moveto = 1, link_reset = 1;

      reference_class_nb(cmodel, tps_idx, children.cd[chn_idx].cdx,
			 children.cd[chn_idx].ndx, &moveto, &link_reset);
    }

    break;

  case XPLAB_QUERY_SENSITIVITY:
    gtk_widget_set_sensitive(mt.where_referenced, TRUE);

    gtk_widget_set_sensitive(mt.row_edit, db_locked);

    gtk_widget_set_sensitive(mt.row_extract, FALSE);
    gtk_widget_set_sensitive(mt.row_dump, FALSE);

    break;

  default: break;
  }
}



/* Returns NB index.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - child iter
        - pointer to parent iter
Output: - NB index (-1: no index found)
        - parent iter updated
Postcondition: - index iter validated by non-negative NB index
Usage : The parent index iter is searched from the child iter .
*/
static int
child_has_nb_index(GtkTreeModel *model, GtkTreeIter child, GtkTreeIter *parent)
{
  gint     nb_index = -1;
  gchar    *row_text;
  gboolean found;

  found = search_above(model, &child, XPLAB_MAINTENANCE_NODE_ID_NB_INDEX);

  if (found == TRUE) {
    gtk_tree_model_get(model, &child, 
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, 
		       &row_text, -1);
    assert(row_text);
    nb_index = atoi(row_text);
    g_free(row_text);
    assert(nb_index >= 0);

    if (parent) {                                       /* if parent defined */
      *parent = child;
    }
  }

  return nb_index;
}



/* Returns class index.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - child iter
        - pointer to parent iter
Output: - class index (-1: no index found)
        - parent iter updated
Postcondition: - index iter validated by non-negative NB index
Usage : The parent index iter is searched from the child iter .
*/
static int
child_has_class_index(GtkTreeModel *model, GtkTreeIter child, 
		      GtkTreeIter *parent)
{
  gint     class_index = -1;
  gchar    *row_text;
  gboolean found;

  found = search_above(model, &child, XPLAB_MAINTENANCE_NODE_ID_CLASS_INDEX);

  if (found == TRUE) {
    gtk_tree_model_get(model, &child,
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, &row_text,
		       -1);
    assert(row_text);
    class_index = atoi(row_text);
    g_free(row_text);
    assert(class_index >= 0);

    if (parent) {                                       /* if parent defined */
      *parent = child;
    }
  }

  return class_index;
}



/* Returns TPS index.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - child iter
        - pointer to parent iter
Output: - TPS index (-1: no index found)
        - parent iter updated
Postcondition: - index iter validated by non-negative NB index
Usage : The parent index iter is searched from the child iter .
*/
static int
child_has_tps_index(GtkTreeModel *model, GtkTreeIter child, GtkTreeIter *parent)
{
  gint     tps_index = -1;
  gchar    *row_text;
  gboolean found;

  found = search_above(model, &child, XPLAB_MAINTENANCE_NODE_ID_TPS_INDEX);

  if (found == TRUE) {
    gtk_tree_model_get(model, &child,
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, 
		       &row_text, -1);
    assert(row_text);
    tps_index = atoi(row_text);
    g_free(row_text);
    assert(tps_index >= 0);

    if (parent) {                                       /* if parent defined */
      *parent = child;
    }
  }

  return tps_index;
}



/* Returns iter from node ID above start iter.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - pointer to start iter
        - node ID
Output: - start iter updated
        - flag (node ID found)
*/
static gboolean
search_above(GtkTreeModel *model, GtkTreeIter *io_iter, gint search_id)
{
  gint        node_id;
  gboolean    found = FALSE, has_parent;
  GtkTreeIter tmp_iter = *io_iter, child;

  while (1) {
    gtk_tree_model_get(model, &tmp_iter,
		       XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

    if (node_id == search_id) {
      *io_iter = tmp_iter;
      found = TRUE;
      break;
    }

    has_parent = gtk_tree_model_iter_parent(model, &child, &tmp_iter);

    if (has_parent == FALSE) {
      break;
    }

    tmp_iter = child;
  } 

  return found;
}



/* Returns TPS index.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - pointer to iter
        - node ID
Output: - iter updated
Usage : The child ID iter is searched from the input iter .
*/
static gboolean
search_below(GtkTreeModel *model, GtkTreeIter *io_iter, gint search_id)
{
  gint        node_id;
  gboolean    found = FALSE, has_child, has_next;
  GtkTreeIter tmp_iter = *io_iter, child;

  while (1) {
    gtk_tree_model_get(model, &tmp_iter,
		       XPLAB_MAINTENANCE_TREE_STORE_NODE_ID, &node_id, -1);

    if (node_id == search_id) {
      *io_iter = tmp_iter;
      found = TRUE;
      break;
    }

    has_child = gtk_tree_model_iter_children(model, &child, &tmp_iter);

    if (has_child == TRUE) {
      found = search_below(model, &child, search_id);

      if (found == TRUE) {
	*io_iter = child;
	break;
      }
    }

    has_next = gtk_tree_model_iter_next(model, &tmp_iter);

    if (has_next == FALSE) {
      break;
    }
  } 

  return found;
}



/* Returns NNB index.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - child iter
        - pointer to parent iter
Output: - NNB index (-1: no index found)
        - parent iter updated
Postcondition: - index iter validated by non-negative NB index
Usage : The parent index iter is searched from the child iter .
*/
static int
child_has_nnb_index(GtkTreeModel *model, GtkTreeIter child, GtkTreeIter *parent)
{
  gint     nnb_index = -1;
  gchar    *row_text;
  gboolean found;

  found = search_above(model, &child, XPLAB_MAINTENANCE_NODE_ID_NNB_INDEX);

  if (found == TRUE) {
    gtk_tree_model_get(model, &child,
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, 
		       &row_text, -1);

    assert(row_text);
    nnb_index = atoi(row_text);
    g_free(row_text);
    assert(nnb_index >= 0);

    if (parent) {                                       /* if parent defined */
      *parent = child;
    }
  }

  return nnb_index;
}



/* Returns CNB index.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - child iter
        - pointer to parent iter
Output: - CNB index (-1: no index found)
        - parent iter updated
Postcondition: - index iter validated by non-negative NB index
Usage : The parent index iter is searched from the child iter .
*/
static int
child_has_cnb_index(GtkTreeModel *model, GtkTreeIter child, GtkTreeIter *parent)
{
  gint     cnb_index = -1;
  gchar    *row_text;
  gboolean found;

  found = search_above(model, &child, XPLAB_MAINTENANCE_NODE_ID_CNB_INDEX);

  if (found == TRUE) {
    gtk_tree_model_get(model, &child,
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, 
		       &row_text, -1);
    assert(row_text);
    cnb_index = atoi(row_text);
    g_free(row_text);
    assert(cnb_index >= 0);

    if (parent) {                                       /* if parent defined */
      *parent = child;
    }
  }

  return cnb_index;
}



/* Returns chain class index.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - parent iter
Output: - chain class index (-1: no index found)
*/
static int
parent_has_chain_class_index(GtkTreeModel *model, GtkTreeIter parent)
{
  gint     chain_class_index = -1;
  gchar    *row_text;
  gboolean found;

  found = search_below(model, &parent, XPLAB_MAINTENANCE_NODE_ID_CHAIN_CLASS_INDEX);

  if (found == TRUE) {
    gtk_tree_model_get(model, &parent,
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, &row_text,
		       -1);
    assert(row_text);
    chain_class_index = atoi(row_text);
    g_free(row_text);
    assert(chain_class_index >= 0);
  }

  return chain_class_index;
}



/* Returns chain NB index.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - parent iter
Output: - chain NB index (-1: no index found)
*/
static int
parent_has_chain_nb_index(GtkTreeModel *model, GtkTreeIter parent)
{
  gint     chain_nb_index = -1;
  gchar    *row_text;
  gboolean found;

  found = search_below(model, &parent, XPLAB_MAINTENANCE_NODE_ID_CHAIN_NB_INDEX);

  if (found == TRUE) {
    gtk_tree_model_get(model, &parent,
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, &row_text,
		       -1);
    assert(row_text);
    chain_nb_index = atoi(row_text);
    g_free(row_text);
    assert(chain_nb_index >= 0);
  }

  return chain_nb_index;
}


/* Returns class index.

Revision: 01.08.2005 E.Sack

Input : - pointer to tree model
        - parent iter
Output: - class index (-1: no index found)
*/
static int
parent_has_class_index(GtkTreeModel *model, GtkTreeIter parent, 
		       GtkTreeIter *child)
{
  gint     class_index = -1;
  gchar    *row_text;
  gboolean found;

  found = search_below(model, &parent, XPLAB_MAINTENANCE_NODE_ID_CLASS_INDEX);

  if (found == TRUE) {
    gtk_tree_model_get(model, &parent,
		       XPLAB_MAINTENANCE_TREE_STORE_GRAPHIC_TXT, &row_text,
		       -1);
    assert(row_text);
    class_index = atoi(row_text);
    g_free(row_text);
    assert(class_index >= 0);

    if (child) {                                         /* if child defined */
      *child = parent;
    }
  }

  return class_index;
}


/* Scales an image.

Revision: 30.01.2004 E.Sack

Input : - image
        - scale factor
Output: - image
Precondition: - a MULT/ADD operation cannot be optimized in modern prozessors
                that means, it is as fast as an ADD operation.
*/
static pic_def
scale_bmp(pic_def in_img, double vsi)
{
  double  ix, iy, hd, wd, oyd, oydd, oxd, oxdd, vsx, vsy;
  int     ox, oxi, oxii, oy, oyi, oyii, oyy, iyy, hi, wi, sum;
  pic_def out_img;
  PICT    pxl;

  assert(vsi > 0.0);

  init_pic_def(&out_img);

  hd = (double)in_img.ar.height;
  wd = (double)in_img.ar.width;
  hi = (int)(hd * vsi);
  wi = (int)(wd * vsi);

  if (wi == 0 || hi == 0) {                     /* image has subpixel size */
    CALLOC(PICT, 1, out_img.mem);      /* at least one pixel is foreground */
    out_img.mem[0] = FOREGROUND;
    return out_img;
  }

  vsx = (double)wi/wd;                 /* adapted scale value for abscissa */
  vsy = (double)hi/hd;                 /* adapted scale value for ordinate */

  if ((vsi == 1.0) || (vsx + vsy == 2.0)) {
    out_img = xdup_pic_def(in_img);                            /* copy 1:1 */
    return out_img;
  }

  out_img.ar.width = wi;
  out_img.ar.height = hi;
  CALLOC(PICT, AREA_SIZE(out_img.ar), out_img.mem);  /* at least one pixel */

  if (vsi > 1.0) {
    for (iy = 0.0; iy < hd; iy+=1.0) {
      oyi = (int)(iy*vsy);
      oyii = (int)(iy*vsy+vsy);
      iyy = (int)(iy*wd);

      for (ix = 0.0; ix < wd; ix+=1.0) {
	oxi = (int)(ix*vsx);
	oxii = (int)(ix*vsx+vsx);
	pxl = in_img.mem[(int)ix+iyy];

	for (oy = oyi; oy < oyii; oy++) {               /* sub-scan output */
	  oyy = oy*out_img.ar.width;

	  for (ox = oxi; ox < oxii; ox++) {
	    out_img.mem[ox + oyy] = pxl;
	  }
	}
      }
    }
  }
  else {
    vsx = 1/vsx;
    vsy = 1/vsy;

    for (oy = 0; oy < hi; oy++) {
      oyy = oy*out_img.ar.width;
      oyd = (double)oy * vsy;
      oydd = (double)oy * vsy + vsy;

      for (ox = 0; ox < wi; ox++) {
	oxd = (double)ox * vsx;
	oxdd = (double)ox * vsx + vsx;
	sum = 0;

	for (iy = oyd; iy < oydd; iy+=1.0) {             /* sub-scan input */
	  iyy = (int)iy*in_img.ar.width;

	  for (ix = oxd; ix < oxdd; ix+=1.0) {
	    sum += (int)in_img.mem[(int)ix+iyy];
	  }
	}

	out_img.mem[ox + oyy] = sum?FOREGROUND:BACKGROUND;     /* BMP only */
      }
    }
  }

  return out_img;
}



/* Gives back a default setting.

Revision: 15.02.2004 E.Sack

Input : - pointer to class name
Output: - default setting
*/
static void
xdefault_cls_att(cls_att *catt, char *cname)
{
  int d;

  assert(catt);

  d = check_name(cname);

  if (d) {
    add_string(XPLAB_DEF_CLASS_NAME, &catt->class);
  }
  else {
    add_string(cname, &catt->class);
  }

  add_string(XPLAB_DEF_CLASS_TRANS, &catt->sign);
  add_string(XPLAB_DEF_CLASS_REMARK, &catt->crem);

  catt->alignment = BASELINE;
  catt->printable = 1;
  catt->fixed     = 1;
}



/* Creates the CCF setting box.

Revision: 02.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
        - CCF buttons definition
*/
static ccf_def
ccf_setting(GtkWidget *fbox)
{
  ccf_def   ccfd;
  GSList    *item_group = NULL;
  GtkWidget *hbox, *vboxl, *vboxr;

  init_ccf_def(&ccfd);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(fbox), hbox, FALSE, FALSE, 0);

  vboxl = gtk_vbutton_box_new();
  vboxr = gtk_vbutton_box_new();

  ccfd.intra_button = gtk_radio_button_new_with_label(item_group, 
						      CCF_INTRA_BUTTON);
  ccfd.inter_button = 
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(ccfd.intra_button),
						CCF_INTER_BUTTON);
    
  gtk_box_pack_start(GTK_BOX(vboxl), ccfd.intra_button, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vboxl), ccfd.inter_button, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vboxl, FALSE, FALSE, 0);

  g_signal_connect(G_OBJECT(ccfd.intra_button), "toggled", 
		   G_CALLBACK(cb_dialog_ccf_setting), NULL);

  g_signal_connect(G_OBJECT(ccfd.inter_button), "toggled", 
		   G_CALLBACK(cb_dialog_ccf_setting), NULL);

  item_group = NULL;
  ccfd.sort_nb_button = gtk_radio_button_new_with_label(item_group, 
						      CCF_INDEX_BUTTON);
  ccfd.sort_cv_button = 
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(ccfd.sort_nb_button),
						CCF_VALUE_BUTTON);
    
  gtk_box_pack_start(GTK_BOX(vboxr), ccfd.sort_nb_button, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vboxr), ccfd.sort_cv_button, FALSE, FALSE, 0);
  gtk_box_pack_end(GTK_BOX(hbox), vboxr, FALSE, FALSE, 0);

  return ccfd;
}



/* Checks the CCF buttons sensitivity.

Revision: 04.01.2006 E.Sack

Input : - -
Output: - CCF buttons updated
*/
static void
ccf_sensitivity(void)
{
  gint              cls_idx = -1, tps_idx = -1, pairs = 0;
  gchar             *buf = NULL;
  gboolean          is_path, has_iter;
  GtkTreeModel      *cmodel;
  GtkTreeIter       iter, cls_iter, tps_iter;
  GtkTreePath       *path;
  GtkTreeViewColumn *fcolumn;

  if (mt.dialog.ccf.edt.edialog == NULL) {/* only for dialog update on the fly */
    return;
  }

  if (GTK_IS_TREE_VIEW(mt.treeview) == FALSE) {
    return;
  }

  gtk_tree_view_get_cursor(GTK_TREE_VIEW(mt.treeview), &path, &fcolumn);

  if (path == NULL) {
    return;
  }

  g_signal_handler_block((gpointer)mt.treeview, mt.cursor_changed_id);

  cmodel = GTK_TREE_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(mt.treeview)));

  is_path = gtk_tree_model_get_iter(cmodel, &iter, path);
  assert(is_path == TRUE);
  gtk_tree_path_free(path);
  path = NULL;

  cls_idx = child_has_class_index(cmodel, iter, &cls_iter);  /* search above */

  if (cls_idx < 0) {        /* no class found above in tree, so search below */
    cls_idx = parent_has_class_index(cmodel, iter, &cls_iter);

    if (cls_idx < 0) {          /* still no class found, so search from root */
      has_iter = gtk_tree_model_get_iter_first(cmodel, &iter);

      if (has_iter == TRUE) {                            /* first iter found */
	cls_idx = parent_has_class_index(cmodel, iter, &cls_iter);
      }                                                /* else tree is empty */
    }
  }

  if (cls_idx < 0) {                         /* tree doesn't contain a class */
    xplab_label_set_text(mt.dialog.ccf.edt.einfoline, "no class list found");
    gtk_widget_set_sensitive(mt.dialog.ccf.cbt.intra_button, FALSE);
    gtk_widget_set_sensitive(mt.dialog.ccf.cbt.inter_button, FALSE);
  }
  else {                                            /* class found somewhere */
    tps_idx = child_has_tps_index(cmodel, cls_iter, &tps_iter);/*search above*/
    assert(tps_idx >= 0);                          /* class must be in a TPS */

    int_to_string("TPS[", tps_idx, "]", &buf);

    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mt.dialog.ccf.cbt.intra_button)) == TRUE) {
      int_to_string(", class[", cls_idx, "]", &buf);
    }

    if (GTK_WIDGET_IS_SENSITIVE(mt.dialog.ccf.cbt.intra_button) == TRUE &&
	gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mt.dialog.ccf.cbt.intra_button)) == TRUE) {
      expand_to_root(GTK_TREE_VIEW(mt.treeview), cls_iter, FALSE);/* show row */
      tree_view_set_cursor(GTK_TREE_VIEW(mt.treeview), cls_iter);

      intra_corr(mt.mdb.tpl.tps[tps_idx].cl.cls[cls_idx].nbl, &pairs, NULL);/* preview */
    }
    else if
      (GTK_WIDGET_IS_SENSITIVE(mt.dialog.ccf.cbt.inter_button) == TRUE &&
       gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mt.dialog.ccf.cbt.inter_button)) == TRUE) {
      expand_to_root(GTK_TREE_VIEW(mt.treeview), tps_iter, FALSE);/* show row */
      tree_view_set_cursor(GTK_TREE_VIEW(mt.treeview), tps_iter);

      inter_corr(mt.mdb.tpl.tps[tps_idx].cl, &pairs, NULL);       /* preview */
    }

    if (pairs) {                  /* show preview pairs number of CCF matrix */
      int_to_string(", ", pairs, pairs==1?" pair":" pairs", &buf);
    }
    else {
      add_string(", no pairs", &buf);
    }

    xplab_label_set_text(mt.dialog.ccf.edt.einfoline, buf);
  }

  FREEMEM(buf);

  g_signal_handler_unblock((gpointer)mt.treeview, mt.cursor_changed_id);
}
