/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_image.h"
#include "xplab_training_fe.h"
#include "xplab_training_be.h"
#include "xplab_tprint.h"
#include "xplab_recall_be.h"
#include "xplab_maintenance_be.h"
#include "xplab_algorithm.h"
#include "xplab_dbio.h"
#include "xplab_preferences_fe.h"
#include "xplab_cluster.h"
#include "xplab_batch.h"

/* local definitions *******************************************************/

#define XPLAB_TRAINING_TITLE_ITEM         "training"
#define XPLAB_TRAINING_IMAGE_TITLE_ITEM   "training image"
#define XPLAB_TRAINING_DIALOG_PARSE_LABEL "parse"
#define XPLAB_TRAINING_DIALOG_TRAIN_LABEL "TRAIN"
#define XPLAB_TRAINING_PAGE_SELECT "Select"
#define XPLAB_TRAINING_PAGE_TRACE  "Trace"
#define XPLAB_IMG_SOURCE_IS_EMPTY_FILE "empty"
#define XPLAB_IMG_SOURCE_IS_EMPTY_MSG  "empty"           /* status message */
#define XPLAB_STATUS_MSG_SEPARATOR  ", "
#define XPLAB_STATUS_MSG_SUFFIX_LEFT    "["
#define XPLAB_STATUS_MSG_SUFFIX_RIGHT   "] "
#define XPLAB_FRAME_SPACING 10
#define XPLAB_WHITE_COLOR {255,255,255}
#define XPLAB_RB_FRAMES_COLOR {0, 255,0}
#define XPLAB_FOCUS_COLOR {255,0,0}
#define XPLAB_RB_STEP_INCREMENT 1.0
#define XPLAB_RB_PAGE_INCREMENT 1.0
#define XPLAB_RB_PAGE_SIZE 0.0
#define XPLAB_NNB_PARSING_ERROR "NNB parsing error - class names must be different"
#define XPLAB_PATTERN_PARSING_ERROR   "pattern parsing error - not trainable"
#define XPLAB_FRACT_PARSING_WARNING "WARNING: training fractions as single NB"
#define XPLAB_CNB_PARSING_ERROR  "CNB parsing error - not enaugh NNBs trained"
#define XPLAB_PATTERN_CORR_ERROR  "no selected patterns"
#define XPLAB_NO_TRAINABLE_PIXEL_MSG "no more trainable pattern present in image"
#define XPLAB_SELECT_TRAINED   "trained pattern not selectable"
#define XPLAB_DESELECT_TRAINED "trained pattern not deselectable"
#define XPLAB_CHANGE_TRAINED "trained pattern label not changeable"
#define XPLAB_FOCUS_SLIDER_MIN 1
#define XPLAB_MOUSE_BUTTON_RIGHT 3
#define XPLAB_TRACE_PATTERN_SPACING 3
#define XPLAB_START_RB_NO 0
#define XPLAB_START_NNB_NO 0
#define XPLAB_SINGLE_CLASS 0
#define XPLAB_TREEVIEW_MOVE_TO_COLUMN 0
#define XPLAB_PAGE_SELECT_CV_SEP " / "
#define TREE_VIEW_SEARCH_COL "search column: "

typedef enum {
  XPLAB_TAB_UNDEF = 42,
  XPLAB_TRAINING_PAGE_IS_SELECT,
  XPLAB_TRAINING_PAGE_IS_TRACE
} PAGE_TABS;

enum {
  XPLAB_DB_TREE_STORE_COLUMN_IDX,
  XPLAB_DB_TREE_STORE_COLUMN_CLS,
  XPLAB_DB_TREE_STORE_COLUMN_CLS_old,
  XPLAB_DB_TREE_STORE_COLUMN_TRS,
  XPLAB_DB_TREE_STORE_COLUMN_CMT,
  XPLAB_DB_TREE_STORE_COLUMN_ALT,
  XPLAB_DB_TREE_STORE_COLUMN_PRT,
  XPLAB_DB_TREE_STORE_COLUMN_FIX,
  XPLAB_DB_TREE_STORE_COLUMN_CNT,                              /* auxiliary */
  XPLAB_DB_TREE_STORE_COLUMN_CNT_old,                          /* auxiliary */
  XPLAB_DB_TREE_STORE_COLUMN_CVL,                              /* auxiliary */
  XPLAB_DB_TREE_STORE_COLUMN_children,                         /* auxiliary */
  XPLAB_DB_TREE_STORE_COLUMNS
};

enum {
  XPLAB_PAGE_SELECT_VIEW_COLUMN_IDX,
  XPLAB_PAGE_SELECT_VIEW_COLUMN_CLS,
  XPLAB_PAGE_SELECT_VIEW_COLUMN_TRS,
  XPLAB_PAGE_SELECT_VIEW_COLUMN_CMT,
  XPLAB_PAGE_SELECT_VIEW_COLUMN_ALT,
  XPLAB_PAGE_SELECT_VIEW_COLUMN_PRT,
  XPLAB_PAGE_SELECT_VIEW_COLUMN_FIX,
  XPLAB_PAGE_SELECT_VIEW_COLUMN_CNT,
  XPLAB_PAGE_SELECT_VIEW_COLUMN_CVL,
  XPLAB_PAGE_SELECT_VIEW_COLUMNS
};

#define XPLAB_PAGE_SELECT_VIEW_TITLE_IDX "No. "
#define XPLAB_PAGE_SELECT_VIEW_TITLE_CLS "class"
#define XPLAB_PAGE_SELECT_VIEW_TITLE_TRS "translation"
#define XPLAB_PAGE_SELECT_VIEW_TITLE_CMT "comment"
#define XPLAB_PAGE_SELECT_VIEW_TITLE_ALT "alignment"
#define XPLAB_PAGE_SELECT_VIEW_TITLE_PRT "printable"
#define XPLAB_PAGE_SELECT_VIEW_TITLE_FIX "fixed"
#define XPLAB_PAGE_SELECT_VIEW_TITLE_CNT "NBs "
#define XPLAB_PAGE_SELECT_VIEW_TITLE_CVL "CV max" XPLAB_PAGE_SELECT_CV_SEP "min"

enum {
  XPLAB_TR_TREE_STORE_COLUMN_IDX,
  XPLAB_TR_TREE_STORE_COLUMN_RBN,
  XPLAB_TR_TREE_STORE_COLUMN_CVL,
  XPLAB_TR_TREE_STORE_COLUMN_CLS,
  XPLAB_TR_TREE_STORE_COLUMN_NBN,
  XPLAB_TR_TREE_STORE_COLUMN_DUM,    /* empty column looks better in list */
  XPLAB_TR_TREE_STORE_COLUMNS
};

enum {
  XPLAB_PAGE_TRACE_VIEW_COLUMN_IDX,
  XPLAB_PAGE_TRACE_VIEW_COLUMN_RBN,
  XPLAB_PAGE_TRACE_VIEW_COLUMN_CVL,
  XPLAB_PAGE_TRACE_VIEW_COLUMN_CLS,
  XPLAB_PAGE_TRACE_VIEW_COLUMN_NBN,
  XPLAB_PAGE_TRACE_VIEW_COLUMN_DUM,
  XPLAB_PAGE_TRACE_VIEW_COLUMNS
};

#define XPLAB_PAGE_TRACE_VIEW_TITLE_IDX " No. "
#define XPLAB_PAGE_TRACE_VIEW_TITLE_RBN " RB "
#define XPLAB_PAGE_TRACE_VIEW_TITLE_CVL " CV "
#define XPLAB_PAGE_TRACE_VIEW_TITLE_CLS " class "
#define XPLAB_PAGE_TRACE_VIEW_TITLE_NBN " NB "
#define XPLAB_PAGE_TRACE_VIEW_TITLE_DUM " "

typedef enum {
  TR_ENUMS_UNDEF,
  IMG_SOURCE_IS_EMPTY,
  IMG_WITHOUT_RB_FRAMES,
  IMG_NOT_REDRAWED,
  IMG_REDRAWED,
  IMG_NOT_UPDATED,
  IMG_UPDATED,
  TR_ENUMS_MAX,
  DC_OPT_MARKED,
  DC_OPT_ORG,
  DC_OPT_CLEAN,
  DC_OPT_RB_FRAME_BMP
} XPLAB_TR_ENUMS;

typedef struct {
  GtkAdjustment *sadjustmnt;
  GtkWidget     *sbutton;
  GtkEntry      *senty;
} slider_def;

typedef struct {
  XPLAB_TR_ENUMS source, rb_frames;
  gfloat         magnitude;
} img_state;

typedef struct {
  GtkWidget     *status_bar, *mark_pattern;
  gint          context_id;
  scroll_def    scroll;
  img_state     state, old_state;
  pattern_color pcol;
} timg_def;

typedef struct {
  GdkColor cfnd;
} list_col;

typedef struct {
  GtkTreeView *tview;
  cls_lst     *cls;
} list_def;

typedef struct {
  scroll_def pscroll;
  pic_def    sample;
  sector_def cfoc;
} pattern_def;

typedef struct {
  cls_lst     sclass;
  pattern_def spattn;
} sample_def;

typedef struct {
  gint        chno, parse_status, lptr;
  cand_lst    hit_lst;    /* candidates hit list of the trained class        */
  int_lst     lbls;       /* traced NB label state list                      */
  hash_def    thash;      /* list containing NB relatives relations          */
  sample_def  trained;    /* items derived from the database and traced ones */
  sample_def  traced;     /* items built from the hit list                   */
} trace_data_def;

typedef struct {
  GtkWidget      *tparent, *tchild, *infoline, *all_button, *select_button;
  GtkWidget      *deselect_button, *left_button, *right_button;
  GtkTreeView    *tview;
  gint           tidx, cursor_changed_id;
  trace_data_def tdata;
} trace_def;

typedef struct {
  GtkWidget     *parent, *nb_button, *cnb_button, *nnb_button;
  GtkWidget     *nnb_lbutton, *nnb_rbutton, *propose_button, *trace_button;
  GtkWidget     *untrained_rb_item, *find_info, *pts_label, *nbhood;
  slider_def    focus_slider, rb_slider;
  gfloat        focus, focus_old;
  gint          focus_rb_no, rb_no, rb_no_old, blob_cnt, purpose_id, stop_id;
  gint          direction;
  gulong        focus_id, rb_id;
  sector_def    focus_sector, focus_sector_old;
  entry_def     find_entry;
  cls_att_entry satt;
} select_def;

typedef struct {
  GtkWidget *rparent, *label;
} revert_def;

typedef struct {
  GtkWidget  *notebook, *parse_button, *save_button, *save_quit_button;
  select_def select;
  trace_def  trace;
  revert_def revert;
} tr_pages_def;

typedef struct {
  char    *img_name;
  int     tps_idx;
  img_typ mimg;                                   /* original image buffer */
  db_def  tdb;
} session_def;

typedef struct {
  GtkWidget    *top_training, *dialog, **caller, *rb_frames_button;
  GtkWidget    *view_original;
  char         *status_suffix;
  col_def      *rgbbuf;                   /* whole current processed image */
  kern_def     ckernel;
  int          cluster_thold, single_line, single_nb;
  int          image_key, db_key;
  gboolean     cactive, cstop;
  session_def  session;
  rb_lst       trbl;
  list_def     list_old, list_new;
  list_col     lcol;
  timg_def     timg;
  tr_pages_def page;
  match_def    tad;
  run_def      trun, srun;
  pic_def      magn_img;                            /* scaled image buffer */
} tr_def;

/* local data **************************************************************/

static tr_def tr;                                    /* training structure */

/* local prototypes ********************************************************/

static void init_tr_def(tr_def *td);
static void init_session_def(session_def *sd);
static void init_list_def(list_def *ld);
static void init_list_col(list_col *lc);
static void init_img_state(img_state *state);
static void init_timg_def(timg_def *td);
static void init_tr_pages_def(tr_pages_def *pd);
static void init_slider_def(slider_def *sd);
static void init_select_def(select_def *td);
static void init_trace_def(trace_def *td);
static void init_trace_data_def(trace_data_def *td);
static void init_sample_def(sample_def *sd);
static void init_revert_def(revert_def *rd);
static void init_pattern_def(pattern_def *pd);
static void free_tr_def(tr_def *td);
static void free_session_def(session_def *sd);
static void free_list_def(list_def *ld);
static void free_list_col(list_col *lc);
static void free_img_state(img_state *state);
static void free_timg_def(timg_def *td);
static void free_tr_pages_def(tr_pages_def *pd);
static void free_slider_def(slider_def *sd);
static void free_select_def(select_def *sd);
static void free_trace_def(trace_def *td);
static void free_trace_data_def(trace_data_def *tdd);
static void free_sample_def(sample_def *sd);
static void free_revert_def(revert_def *rd);
static void free_pattern_def(pattern_def *pd);
static void clear_list_def(list_def *ld);
static void clear_tree_store(GtkTreeView *tview);

/* Functions assembling widgets */
static void xplab_training(match_def mat, char *filename, char *tps_id, rb_lst rbl, kern_def ckernel, int cluster_thold, int single_line, int single_nb);
static void default_select_def(select_def *sdef, rb_lst rbl);
static void default_list_col(list_col *lcol);
static void default_timg_def(timg_def *tig);
static void training_dialog(void);
static void training_dialog_upper(GtkContainer *container);
static GtkWidget *training_dialog_page_select(void);
static GtkWidget *training_dialog_page_trace(void);
static GtkWidget *training_dialog_page_revert(void);
static GtkWidget *training_treeview(GtkTreeView **tview, GdkColor traceable_class_color);
static GtkTreeModel *select_sort_model_of_tree_store_init(void);
static GtkTreeView *select_tree_view(GdkColor traceable_class_color);
static GtkTreeModel *trace_sort_model_of_tree_store_init(void);
static GtkTreeView *trace_tree_view(void);
static void training_dialog_select_lists(GtkWidget *vbox);
static void training_dialog_trace_list(GtkWidget *box);
static void training_dialog_trace_pattern(GtkWidget *fbox);
static void trace_pattern_selection(GtkWidget *cbox);
static void training_dialog_trace_control(GtkWidget *fbox);
static void trace_control_table(GtkWidget *vbox);
static void training_dialog_lower(GtkContainer *container);
static void training_window(void);
static void training_status_bar(GtkWidget *vbox);
static void training_image_menu_bar(GtkWidget *vbox);
static void training_image_menu_file(GtkWidget *menu_bar);
static void training_image_menu_save_quit(GtkWidget *file_menu);
static void training_image_menu_quit(GtkWidget *file_menu);
static void training_image_menu_dump(GtkWidget *file_menu);
static void training_image_menu_train(GtkWidget *file_menu);
static void training_image_menu_view(GtkWidget *menu_bar);
static void view_menu_item_focus(GtkWidget *view_menu);
static void view_menu_item_rb_frames(GtkWidget *view_menu);
static void view_menu_item_mark(GSList **image_group, GtkWidget *view_menu);
static void view_menu_item_orig(GSList **image_group, GtkWidget *view_menu);
static void view_menu_item_binar(GSList **image_group, GtkWidget *view_menu);
static void view_menu_item_empty(GSList **image_group, GtkWidget *view_menu);
static void separator_item(GtkWidget *menu);
static void create_rb_slider(GtkWidget *hbox);
static void create_focus_slider(GtkWidget *hbox);
static slider_def hslider(GtkWidget *in_box, GtkWidget *mbox, char *label, gint label_width, gint min_range, gint max_range, int is_box, gint init_slider_val, const gchar *tip_text, const gchar *tip_private);
static void find_tentry(GtkWidget *in_box, GtkTreeView *tview);
static void trace_sensitiveness(void);
static void save_training_session(void);
static void trace_parse(GtkButton *button);
static void trace_train(GtkButton *button);
static void select_parse(GtkButton *button);
static void select_train(GtkButton *button);
static void reset_trace_data(trace_data_def *tdata, GtkTreeView *tview);

/* Handlers/Callbacks */
static gboolean cb_training_phase(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN);
static void cb_training_page_switch(GTK_CONVENTION_NOTEBOOK_RTS_VOID);
static gboolean cb_training_dialog_delete(GTK_CONVENTION_DIALOG_RTS_GBOOLEAN);
static void cb_training_dialog_quit(GTK_CONVENTION_BUTTON_RTS_VOID );
static void cb_training_dialog_save_quit(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_training_save(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_trace_row_select(GTK_CONVENTION_TREEVIEW_CURSOR_CHANGED_RTS_VOID);
static void cb_trace_select(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_trace_deselect(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_trace_left_arrow(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_trace_right_arrow(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_trace_rb_button(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_trace_focus_button(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_trace_stop_button(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_parse_train(GTK_CONVENTION_BUTTON_RTS_VOID);
static gboolean cb_expose_trace_old(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN);
static gboolean cb_expose_trace_new(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN);
static void cb_select_stop_button(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_select_nb_button(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID);
static void cb_select_cnb_button(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID);
static void cb_select_nnb_button(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID);
static void cb_select_left_arrow(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_select_right_arrow(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_select_nbhood(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_select_search_column(GTK_CONVENTION_COMBO_CHANGED_RTS_VOID);
static void cb_select_find_trace_button(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID);
static void cb_select_find_proposal(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_select_rb_slider(GTK_CONVENTION_ADJUSTMENT_RTS_VOID);
static void cb_select_rb_slider_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_select_untrained_rb(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static void cb_select_focus_slider(GTK_CONVENTION_ADJUSTMENT_RTS_VOID);
static void cb_select_focus_slider_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_select_focus_reset(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_leave_train(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_select_train_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_select_row_select(GTK_CONVENTION_TREEVIEW_CURSOR_CHANGED_RTS_VOID);
static void cb_select_dump_rb(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_select_focus_dump(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_image_dump(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_image_focus(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_image_slider(GTK_CONVENTION_ADJUSTMENT_RTS_VOID);
static void cb_image_magnitude(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static void cb_rb_frames(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static void cb_image_mark(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static void cb_image_binarized(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static void cb_image_original(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static void cb_image_empty(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static gboolean cb_image_expose_event(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN);
static gboolean cb_mouse_training_image(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN);
static gboolean cb_mouse_pattern_traced(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN);
static gboolean cb_mouse_pattern_trained(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN);

/* Functions assembling handlers/callbacks */
static void binarized_input_image(pic_def *in_image);
static XPLAB_TR_ENUMS background_image(XPLAB_TR_ENUMS old_source, gint force);
static XPLAB_TR_ENUMS empty_image(area_def ar);
static XPLAB_TR_ENUMS patch_rb_frames(rb_lst rbl);
static void image_to_state(gint redraw, gint force);
static XPLAB_TR_ENUMS image_source(void);
static XPLAB_TR_ENUMS overlay_image(img_state *old_state, XPLAB_TR_ENUMS img_redrawed);
static char *name_of_image(char *image_filename, img_state istate, char **status_msg);
static char *xp_base_name(char *filename, char *type);
static gint assemble_status_message(gchar **status_msg, gchar *string, gint use_separator);
static gfloat focus_image(gfloat offset, gfloat image_size, gfloat page_size);
static void center_image(timg_def timg, sector_def sec);
static int xcreate_snb(int mark, sector_def fr, cls_lst *new);
static int xcreate_cnb(cls_lst *new, cls_att_entry entr);
static void training_pattern_type_selection(GtkWidget *box);
static void training_pattern_selection(GtkWidget *box);
static void rb_update(void);
static void sector_from_focus(void);
static gint max_focus_aspect(sector_def rsc, gfloat *focus);
static void limit_focus(gint diff, ival_def *old, ival_def *current, int max);
static void mark_nnb(int bot_up, int n4);
static void append_cand_lst_to_model(cand_lst clst, GtkTreeView *tview, gchar *class_name);
static void append_cls_lst_to_model(cls_lst clst, GtkTreeView *tview, int start_idx, int moveto, gboolean init_mirror);
static void treeview_moveto(GtkTreeView *tview, gint child_idx, gboolean be_quiet);
static int attributes_identic(cls_def cls, cls_att_entry att);
static int class_traceable(cls_def cls);
static gint class_has_children(nb_lst nbl);
static void reset_parse_status(cls_lst cls);
static void set_parse_status(cls_lst cls_new, PARSE_STAT parse_status);
static int get_parse_status(cls_lst cls_old);
static void update_lists(list_def *list_new, list_def list_old, int mborder, int maskweight, int moveto);
static void calculate_blob_number(int both_keys);
static int xsearch_rb_masked(PIXEL_KEY key, int dir, gboolean flip);
static void copy_row_to_attributes(GtkTreeView *treeview, GtkTreePath *path);
static int select_sort_func(GTK_CONVENTION_GtkTreeIterCompareFunc_RTS_INT);
static int trace_sort_func(GTK_CONVENTION_GtkTreeIterCompareFunc_RTS_INT);
static int xfind_row_from_index(GtkTreeView *tview, int cidx, GtkTreeIter *siter);
static void page_trace_start(void);
static void page_trace_stop(void);
static void page_select_start(void);
static void page_select_stop(void);
static int create_pattern_drawing(cls_def cls, pattern_def *pdef, GtkSignalFunc drawing_cb, GtkSignalFunc mouse_cb);
static void pattern_drawing(pattern_def pdef, pattern_color col, int outside_unchanged);
static void orphan_cls_def(cls_def ocls);
static cls_lst xtrace_generation(cand_lst hit, pic_def rpic, pic_def rorg, cls_lst old, hash_def *hash, kern_def kernel, int cluster_thold);
static void xtrace(sector_def bsec);
static sector_def color_reference(pic_def bref, sector_def rbs, pic_def cimg, int lx, int ly, int to);
static sector_def color_nb_def(nb_def nb, pic_def mimg, pic_def simg, int to);
static void color_cand_lst(cand_lst cndl, cls_lst clst, pic_def simg);
static void color_indexed_cls_lst(cls_lst clst, int_lst idx, pic_def mimg, pic_def simg);
static int select_child(trace_data_def *tdata, GtkTreeView *tview);
static int select_pattern(cls_lst clst, sector_def *sec, int *nnr);
static cand_def search_candidate_of_child_id(hash_def hash, int cid);
static int_lst search_child_indexes_of_parent_id(hash_def hash, int pid, cls_lst traced_children);
static int_lst setup_selected_pattern(int cnt);
static int parse_class_attributes_to_train(select_def *pselect, list_def list_old, list_def list_new);
static int select_parse_pattern(void);
static int trace_parse_pattern(void);
static int xremove_nb_def(cls_lst clst, int mark);
static int xmarked_pattern(cls_lst clst, int *cnr, int *nnr, int mark);
static void selected_candidates_message(int_lst lbls, GtkWidget *label);
static void trace_train_message(gint marked, gint trained, GtkWidget *label);
static gint get_candidate_index(GtkTreeView *treeview, GtkTreePath *path);
static int xproposal(list_def plist);
static void xdump_rb(void);
static void xdump_focus(void);
static void save_session_images(img_typ mimg, pic_def bimg, char *filename);
static void training_image_label(GdkEventButton *event);
static void training_image_focus(GdkEventButton *event);
static void training_image_boxes(GdkEventButton *event);
static void training_image_rb(c_def start);
static void training_image_kernel_select(c_def start);
static void training_image_default(c_def start);
static void traced_cursor(int nb_idx);
static void traced_pattern_cursor(GdkEventButton *event, cls_lst clst);
static void trained_pattern_cursor(GdkEventButton *event, cls_lst clst);
static int select_children(GtkTreeView *tview, trace_data_def *tdata, int parent);
static void traced_pattern_label(GdkEventButton *event, cls_lst clst);
static void training_image_kernel_label(GdkEventButton *event);
static int rb_from_coordinate(rb_lst rbl, c_def start);
static PAGE_TABS xplab_get_page_enum(GtkNotebook *notebook, int page_num);
static int cid_from_coordinate(hash_def hash, c_def start);
static int nb_from_coordinate(hash_def hash, nb_lst nbl, c_def start);
static void traced_nb_label(int nb_idx, int center);
static void cell_func_select(GTK_CONVENTION_TREE_CELL_DATA_FUNC_RTS_VOID);
static int check_class_attributes(cls_att_entry catt, gboolean is_nb, gboolean is_cnb);
static int xplab_check_alignment(entry_def alignment);

/* Functions ***************************************************************/

static void
init_tr_def(tr_def *td)
{
  assert(td);

  td->top_training = NULL;                                   /* zero pointer */
  td->dialog = NULL;                                         /* zero pointer */
  td->caller = NULL;                                         /* zero pointer */
  td->rb_frames_button = NULL;                               /* zero pointer */
  td->view_original = NULL;                                  /* zero pointer */

  td->status_suffix = NULL;                                  /* zero pointer */
  td->rgbbuf = NULL;                                         /* zero pointer */

  td->cluster_thold = td->single_line = td->single_nb = 0;
  td->image_key = td->db_key = 0;
  td->cactive = td->cstop = FALSE;

  init_session_def(&td->session);
  init_rb_lst(&td->trbl);
  init_kern_def(td->ckernel);
  init_list_def(&td->list_old);
  init_list_def(&td->list_new);
  init_list_col(&td->lcol);
  init_timg_def(&td->timg);
  init_tr_pages_def(&td->page);
  init_match_def(&td->tad);
  init_run_def(&td->trun);
  init_run_def(&td->srun);
  init_pic_def(&td->magn_img);
  init_c_def(&td->magn_img.org);                        /* special treatment */
}

static void
init_session_def(session_def *sd)
{
  assert(sd);

  sd->img_name = NULL;                                       /* zero pointer */

  sd->tps_idx = XPLAB_IDX_NONE;

  init_img_typ(&sd->mimg);
  init_db_def(&sd->tdb);
}

static void
init_list_def(list_def *ld)
{
  assert(ld);

  ld->tview = NULL;                                          /* zero pointer */
  ld->cls = NULL;                                            /* zero pointer */
}

static void
init_list_col(list_col *lc)
{
  assert(lc);
}

static void 
init_img_state(img_state *state)
{
  assert(state);

  state->source    = TR_ENUMS_UNDEF;
  state->rb_frames = TR_ENUMS_UNDEF;

  state->magnitude = 0.0;
}

static void 
init_timg_def(timg_def *td)
{
  assert(td);

  td->status_bar = NULL;                                     /* zero pointer */
  td->mark_pattern = NULL;                                   /* zero pointer */

  td->context_id = 0;
  init_scroll_def(&td->scroll);
  init_img_state(&td->state);
  init_img_state(&td->old_state);
  init_pattern_color(&td->pcol);
}

static void 
init_tr_pages_def(tr_pages_def *pd)
{
  assert(pd);

  pd->notebook = NULL;                                       /* zero pointer */
  pd->parse_button = NULL;                                   /* zero pointer */
  pd->save_button = NULL;                                    /* zero pointer */
  pd->save_quit_button = NULL;                               /* zero pointer */

  init_select_def(&pd->select);
  init_trace_def(&pd->trace);
  init_revert_def(&pd->revert);
}

static void 
init_pattern_def(pattern_def *pd)
{
  assert(pd);

  init_scroll_def(&pd->pscroll);
  init_pic_def(&pd->sample);
  init_sector_def(&pd->cfoc);
}

static void 
init_slider_def(slider_def *sd)
{
  assert(sd);

  sd->sadjustmnt = NULL;                                     /* zero pointer */
  sd->sbutton = NULL;                                        /* zero pointer */
  sd->senty = NULL;                                          /* zero pointer */
}

static void 
init_select_def(select_def *sd)
{
  assert(sd);

  sd->parent = NULL;                                         /* zero pointer */
  sd->nb_button = NULL;                                      /* zero pointer */
  sd->cnb_button = NULL;                                     /* zero pointer */
  sd->nnb_button = NULL;                                     /* zero pointer */
  sd->nnb_lbutton = NULL;                                    /* zero pointer */
  sd->nnb_rbutton = NULL;                                    /* zero pointer */
  sd->propose_button = NULL;                                 /* zero pointer */
  sd->trace_button = NULL;                                   /* zero pointer */
  sd->untrained_rb_item = NULL;                              /* zero pointer */
  sd->find_info = NULL;                                      /* zero pointer */
  sd->pts_label = NULL;                                      /* zero pointer */
  sd->nbhood = NULL;                                         /* zero pointer */

  init_slider_def(&sd->focus_slider);
  init_slider_def(&sd->rb_slider);

  sd->focus = sd->focus_old = 0.0;
  sd->rb_no = sd->rb_no_old = sd->focus_rb_no = sd->blob_cnt = 0;
  sd->purpose_id = sd->stop_id = 0;
  sd->direction = 0;

  sd->focus_id = sd->rb_id = 0;

  init_sector_def(&sd->focus_sector);
  init_sector_def(&sd->focus_sector_old);

  init_entry_def(&sd->find_entry, &sd->find_info);             /* elaboration*/

  init_cls_att_entry(&sd->satt);
}

static void 
init_trace_def(trace_def *td)
{
  assert(td);

  td->tparent = NULL;                                        /* zero pointer */
  td->tchild = NULL;                                         /* zero pointer */
  td->infoline = NULL;                                       /* zero pointer */
  td->all_button = NULL;                                     /* zero pointer */
  td->select_button = NULL;                                  /* zero pointer */
  td->deselect_button = NULL;                                /* zero pointer */
  td->left_button = NULL;                                    /* zero pointer */
  td->right_button = NULL;                                   /* zero pointer */
  td->tview = NULL;                                          /* zero pointer */

  td->tidx = 0;
  td->cursor_changed_id = 0;

  init_trace_data_def(&td->tdata);
}

static void 
init_trace_data_def(trace_data_def *tdd)
{
  assert(tdd);

  tdd->chno = 0;
  tdd->parse_status = XPLAB_PARSE_STAT_DEF;
  tdd->lptr = XPLAB_START_NNB_NO;

  init_sample_def(&tdd->trained);
  init_sample_def(&tdd->traced);

  init_cand_lst(&tdd->hit_lst);
  init_int_lst(&tdd->lbls);
  init_hash_def(&tdd->thash);
}

static void 
init_sample_def(sample_def *sd)
{
  assert(sd);

  init_cls_lst(&sd->sclass);
  init_pattern_def(&sd->spattn);
}

static void 
init_revert_def(revert_def *rd)
{
  assert(rd);

  rd->rparent = NULL;                                        /* zero pointer */
  rd->label = NULL;                                          /* zero pointer */
}

static void 
free_img_state(img_state *state)
{
  assert(state);

  init_img_state(state);
}

static void
free_tr_def(tr_def *td)
{
  assert(td);
                           /* don't free GTK stuff */

  FREEMEM(td->status_suffix);
  FREEMEM(td->rgbbuf);

  free_match_def(&td->tad);
  free_rb_lst(&td->trbl);
  free_session_def(&td->session);
  free_list_def(&td->list_old);
  free_list_def(&td->list_new);
  free_list_col(&td->lcol);
  free_timg_def(&td->timg);
  free_tr_pages_def(&td->page);
  free_pic_def(&td->magn_img);

  init_tr_def(td);
}



/* Wrapper for GTK function.

Revision: 24.07.2005 E.Sack

Input : - pointer to tree view
Output: - store cleared
*/
static void
clear_tree_store(GtkTreeView *tview)
{
  GtkTreeModel *tstore, *smodel;

  smodel = gtk_tree_view_get_model(GTK_TREE_VIEW(tview));
  assert(GTK_IS_TREE_MODEL(smodel) == TRUE);
  assert(GTK_IS_TREE_MODEL_SORT(smodel) == TRUE);

  tstore = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(smodel));

  gtk_tree_store_clear(GTK_TREE_STORE(tstore));              /* empty store */
}



/* Clears list.

Revision: 24.07.2005 E.Sack

Input : - pointer to list
Output: - list cleared
*/
static void
clear_list_def(list_def *ld) 
{
  assert(ld);

  clear_tree_store(ld->tview);

  if (ld->cls) {
    free_cls_lst(ld->cls);
  }
}



static void
free_session_def(session_def *sd)
{
  assert(sd);

  FREEMEM(sd->img_name);

  free_img_typ(&sd->mimg);
  free_db_def(&sd->tdb);

  init_session_def(sd);
}

static void
free_list_col(list_col *lc) 
{
  assert(lc);
                           /* don't free GTK stuff */
}

static void
free_list_def(list_def *ld) 
{
  assert(ld);
                           /* don't free GTK stuff */

  if (ld->cls) {
    free_cls_lst(ld->cls);
    FREEMEM(ld->cls);
  }

  init_list_def(ld); 
}

static void 
free_timg_def(timg_def *td)
{
  assert(td);
                           /* don't free GTK stuff */

  free_scroll_def(&td->scroll);
  free_img_state(&td->state);
  free_img_state(&td->old_state);

  init_timg_def(td);
}

static void 
free_tr_pages_def(tr_pages_def *pd)
{
  assert(pd);
                           /* don't free GTK stuff */

  free_select_def(&pd->select);
  free_trace_def(&pd->trace);
  free_revert_def(&pd->revert);

  init_tr_pages_def(pd);
}

static void 
free_slider_def(slider_def *sd)
{
  assert(sd);
                           /* don't free GTK stuff */

  init_slider_def(sd);
}

static void 
free_select_def(select_def *sd)
{
  assert(sd);
                           /* don't free GTK stuff */

  free_slider_def(&sd->focus_slider);
  free_slider_def(&sd->rb_slider);
  free_entry_def(&sd->find_entry);
  free_cls_att_entry(&sd->satt);

  init_select_def(sd);
}

static void 
free_trace_def(trace_def *td)
{
  assert(td);
                           /* don't free GTK stuff */

  free_trace_data_def(&td->tdata);

  init_trace_def(td);
}

static void 
free_trace_data_def(trace_data_def *tdd)
{
  assert(tdd);
                           /* don't free GTK stuff */

  free_cand_lst(&tdd->hit_lst);

  free_sample_def(&tdd->trained);
  free_sample_def(&tdd->traced);

  free_int_lst(&tdd->lbls);
  free_hash_def(&tdd->thash);

  init_trace_data_def(tdd);
}

static void 
free_sample_def(sample_def *sd)
{
  assert(sd);
                           /* don't free GTK stuff */

  free_cls_lst(&sd->sclass);
  free_pattern_def(&sd->spattn);
}

static void 
free_revert_def(revert_def *rd)
{
  assert(rd);
                           /* don't free GTK stuff */

  init_revert_def(rd);
}

static void 
free_pattern_def(pattern_def *pd)
{
  assert(pd);

  free_scroll_def(&pd->pscroll);
  free_pic_def(&pd->sample);

  init_pattern_def(pd);
}



/* Assembles base name for documentation images.

Revision: 03.06.2003 E.Sack

Input : - input filename
Output: - DC base name
*/
static char *
xp_base_name(char *filename, char *type)
{
  char *dc_name = NULL;

  cat_string(filename, XP_NAME_EXT, type, &dc_name);

  return dc_name;
}



/* Assembles status message.

Revision: 04.06.2003 E.Sack

Input : - pointer to message
        - pointer to string
        - flag (0: no separation, >0: separation needed)
Output: - message updated as flagged
        - flag (0: nothing changed, 1: string added)
*/
static gint 
assemble_status_message(gchar **status_msg, gchar *string, gint use_separator)
{
  gint check;

  if (use_separator) {
    check = add_string(XPLAB_STATUS_MSG_SEPARATOR, status_msg);
    assert(check);
  }

  use_separator = add_string(string, status_msg);
  assert(use_separator);

  return use_separator;
}



/* Builds name and message of the backing image according to its current 
   state.

Revision: 13.02.2004 E.Sack

Input : - input image filename
Output: - backing image name
*/
static char *
name_of_image(gchar *image_filename, img_state istate, gchar **status_msg) 
{
  guint  use_separator = 0, char_added = 0;
  gchar  *tname = 0;

  tname = xp_base_name(image_filename, DC_EXT_2D);

  if (tr.status_suffix != NULL) {
    assemble_status_message(status_msg, tr.status_suffix, 0);
  }

  switch (istate.source) {
  case IMG_SOURCE_IS_EMPTY:
    use_separator += assemble_status_message(status_msg, XPLAB_IMG_SOURCE_IS_EMPTY_MSG, use_separator);
    break;

  case DC_OPT_ORG:
    char_added += add_string(DC_OPT_ORG_FILE, &tname);
    use_separator += assemble_status_message(status_msg, XP_OPT_ORG_MSG, use_separator);
    break;

  case DC_OPT_CLEAN:
    char_added += add_string(DC_OPT_CLEAN_FILE, &tname);
    use_separator += assemble_status_message(status_msg, XP_OPT_CLEAN_MSG, use_separator);
    break;

  case DC_OPT_MARKED:
    char_added += add_string(DC_OPT_MARKED_FILE, &tname);
    use_separator += assemble_status_message(status_msg, XP_OPT_MARKED_MSG, use_separator);
    break;

  default:
    assert(0);                                     /* something went wrong */
    break;
  }

  switch (istate.rb_frames) {
  case IMG_WITHOUT_RB_FRAMES:
    break;                                            /* no extension used */

  case DC_OPT_RB_FRAME_BMP:
    char_added += add_string(DC_OPT_RB_FRAME_BMP_FILE, &tname);
    use_separator += assemble_status_message(status_msg, XP_OPT_RB_FRAME_BMP_MSG, use_separator);
    break;

  default:
    assert(0);                                     /* something went wrong */
    break;
  }

  if (!char_added) {                                      /* nothing added */
    add_string(XPLAB_IMG_SOURCE_IS_EMPTY_FILE, &tname);
  }

  return tname;
}



/* Redraws backing image from source.

Revision: 02.11.2004 E.Sack

Input : - pointer to general parameters
Output: - backing image updated as flagged
        - redraw flag 
*/
static XPLAB_TR_ENUMS
image_source(void)
{
  XPLAB_TR_ENUMS img_redrawed = IMG_NOT_REDRAWED;

  switch (tr.timg.state.source) {
  case DC_OPT_ORG:
    {
      pic_def ptmp, ptmpcol;

      init_pic_def(&ptmpcol);
      init_pic_def(&ptmp);

      FREEMEM(tr.rgbbuf);
      img_redrawed = IMG_REDRAWED;

      ptmp = xplab_request_image(PNM_FORMAT_ORIGINAL, NULL, NULL, 0, NULL).pimg;

      convert_to_color(&ptmp, FALSE);

      assert(ptmp.col);
      tr.rgbbuf = ptmp.col;                           /* pointer hand-over */
      ptmp.col = NULL;                              /* avoid bogus pointer */

      free_pic_def(&ptmpcol);
      free_pic_def(&ptmp);

      if (GTK_CHECK_MENU_ITEM(tr.rb_frames_button)->active == TRUE) {
	patch_rb_frames(tr.trbl);
      }
    }

    break;

  case IMG_SOURCE_IS_EMPTY:
    img_redrawed = empty_image(tr.session.mimg.pimg.ar);
    break;

  case DC_OPT_CLEAN:
    {
      pic_def tmp;

      tmp = xplab_request_image(P1, NULL, NULL, 0, NULL).pimg;/* image copy */

      binarized_input_image(&tmp);
      img_redrawed = IMG_REDRAWED;

      free_pic_def(&tmp);
    }

    break;

  case DC_OPT_MARKED:
    xprint_rb(tr.rgbbuf, tr.session.mimg.pimg, tr.timg.pcol, 
	      tr.page.select.focus_sector, 0);
    img_redrawed = IMG_REDRAWED;
    break;

  default:
    assert(0);                                     /* something went wrong */
    break;
  }

  return img_redrawed;
}



/* Creates backing image according to state.

Revision: 06.01.2006 E.Sack

Input : - flag (0: refresh window, 1: redraw window)
- flag (0: don't force redraw, 1: force redraw window)
Output: - backing image updated
*/
static void
image_to_state(gint redraw, gint force)
{
  gchar *image_name = NULL, *status_msg = NULL;

  if (GTK_IS_WIDGET(tr.top_training) == FALSE) {       /* deferred mapping */
    return;
  }

  assert(tr.rgbbuf);

  if (redraw) {
    XPLAB_TR_ENUMS img_redrawed_bgnd, img_redrawed_ovl;

    /* discards overlays */
    img_redrawed_bgnd = background_image(tr.timg.old_state.source, force);

    /* patch overlays */
    img_redrawed_ovl = overlay_image(&tr.timg.old_state, img_redrawed_bgnd);

    force = force || 
            img_redrawed_bgnd != IMG_NOT_REDRAWED ||
            img_redrawed_ovl != IMG_NOT_REDRAWED ||
            tr.timg.old_state.magnitude != tr.timg.state.magnitude;

    magnitude_allocation(tr.rgbbuf, tr.session.mimg.pimg.ar, 
			 tr.timg.scroll.h_adj->value, 
			 tr.timg.scroll.v_adj->value, 
			 tr.timg.state.magnitude, &tr.magn_img, 
			 tr.timg.scroll.h_adj->page_size, 
			 tr.timg.scroll.v_adj->page_size, 
			 tr.timg.scroll.sdrawing->allocation.width,
			 tr.timg.scroll.sdrawing->allocation.height, force);

    tr.timg.old_state = tr.timg.state;
  }

  xplab_draw_image(tr.rgbbuf, tr.session.mimg.pimg.ar, tr.magn_img, 
		   tr.timg.state.magnitude, tr.timg.scroll.h_adj, 
		   tr.timg.scroll.v_adj, tr.timg.scroll.sdrawing);

  image_name = name_of_image(tr.session.img_name, tr.timg.state, &status_msg);
  gtk_statusbar_push(GTK_STATUSBAR(tr.timg.status_bar), tr.timg.context_id, status_msg);

  FREEMEM(image_name);
  FREEMEM(status_msg);

  tr.timg.old_state.magnitude = tr.timg.state.magnitude;
}



/* Refreshs overlays.

Revision: 06.06.2003 E.Sack

Input : - pointer to training settings
        - pointer to old training settings
        - flag (0: overlays not discarded, 1: overlays discarded)
Output: - flag (0: backing image not updated, 1: backing image updated)
*/
static XPLAB_TR_ENUMS
overlay_image(img_state *old_state, XPLAB_TR_ENUMS img_redrawed)
{
  XPLAB_TR_ENUMS img_redrawed_tmp = IMG_NOT_REDRAWED;

  switch (tr.timg.state.rb_frames) {
  case IMG_WITHOUT_RB_FRAMES:
    if (tr.timg.state.rb_frames != old_state->rb_frames &&
	img_redrawed == IMG_NOT_REDRAWED) {
      img_redrawed_tmp = background_image(old_state->source, 1);/*clear overlays*/
    }
    break;

  case DC_OPT_RB_FRAME_BMP:
    if (tr.timg.state.rb_frames != old_state->rb_frames ||
	img_redrawed != IMG_NOT_REDRAWED) {
      img_redrawed_tmp = patch_rb_frames(tr.trbl);       /* patch overlays */
    }
    break;

  default:
    assert(0);                                      /* something went wrong */
    break;
  }

  return img_redrawed_tmp;
}



/* Refreshs backing image.

Revision: 06.06.2003 E.Sack

Input : - pointer to training settings
        - pointer to old training settings
        - flag (0: respect old settings, 1: force drawing)
Output: - flag (0: backing image not updated, 1: backing image updated)
*/
static XPLAB_TR_ENUMS
background_image(XPLAB_TR_ENUMS old_source, gint force)
{
  XPLAB_TR_ENUMS img_redrawed = IMG_NOT_REDRAWED;

  if (tr.timg.state.source != old_source || force) {
    img_redrawed = image_source();
  }

  return img_redrawed;
}



/* Converts bitmap to color.

Revision: 31.05.2003 E.Sack

Input : - input image
Output: - backing image
*/
static void
binarized_input_image(pic_def *in_image)
{
  int    i;
  guchar *pos = (guchar *)tr.rgbbuf, pix;

  assert(pos);

  for (i = 0; i < AREA_SIZE(in_image->ar); i++) {
    pix = ~((guchar)in_image->mem[i]*255);
    *pos++ = pix;
    *pos++ = pix;
    *pos++ = pix;
  }
}



/* Empties backing image.

Revision: 02.06.2003 E.Sack

Input : - -
Output: - backing image
*/
static XPLAB_TR_ENUMS
empty_image(area_def ar)
{
  int     i;
  col_def *pos = tr.rgbbuf, white = XPLAB_WHITE_COLOR;

  assert(pos);

  for (i = 0; i < AREA_SIZE(ar); i++) {
    *pos++ = white;
  }

  return IMG_REDRAWED;
}



/* Patches RB frames to backing image.

Revision: 01.06.2003 E.Sack

Input : - RB list
Output: - backing image updated
*/
static XPLAB_TR_ENUMS
patch_rb_frames(rb_lst rbl)
{
  gint    number_rb, x_start, x_end, y_start, y_end, y, x;
  gint    left, right, upper, lower;
  col_def rb_frame_color = XPLAB_RB_FRAMES_COLOR;

  for (number_rb = 0; number_rb < rbl.cnt; number_rb++) {
    x_start = rbl.rb[number_rb].sc.column.start;
    x_end   = rbl.rb[number_rb].sc.column.end;
    y_start = rbl.rb[number_rb].sc.line.start;
    y_end   = rbl.rb[number_rb].sc.line.end;

    for (y = y_start+1; y < y_end; y++) {                    /* plot frame */
      left = x_start + y*rbl.rar.width;                 /* left frame side */
      right = x_end + y*rbl.rar.width;                 /* right frame side */

      tr.rgbbuf[left] = tr.rgbbuf[right] = rb_frame_color;
    }

    y_start *= rbl.rar.width;
    y_end   *= rbl.rar.width;

    for (x = x_start; x <= x_end; x++) {                     /* plot frame */
      upper = x + y_start;                             /* upper frame side */
      lower = x + y_end;                               /* lower frame side */

      tr.rgbbuf[upper] = tr.rgbbuf[lower] = rb_frame_color;
    }
  }

  return IMG_UPDATED;
}



/* Builds an entry for a menu bar.

Revision: 27.10.2003 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
void
xplab_training_menu(GtkWidget *menu_bar, GtkWidget **caller)
{
  GtkWidget *train_entry;

  train_entry = gtk_menu_item_new_with_label(XPLAB_TRAINING_TITLE_ITEM);

  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), train_entry);

  g_signal_connect(G_OBJECT(train_entry), "event",
		   G_CALLBACK(cb_training_phase), (gpointer)caller);
}



/* Starts training phase.

Revision: 17.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static gboolean
cb_training_phase(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN)
{
  gboolean filtered, raised;

  assert(widget);
  assert(user_data);

  filtered = filter_event_type(event, GDK_BUTTON_PRESS);

  if (filtered == FALSE) {
    return FALSE;                      /* propagate further event processing */
  }

  raised = xplab_window_raise(&tr.top_training, 1);

  if (raised == FALSE) {
    xplab_free_training();      /* must free here because of sudden run stop */
    tr.caller = (GtkWidget **)user_data;              /* last succeeded call */

    xplab_training(xctrl.cfg.mad, xctrl.set.image_filename, xctrl.cfg.tps_id,
	           xctrl.set.rblst, xctrl.cfg.ckernel,
		   xctrl.cfg.cluster_thold, xctrl.cfg.single_line, 
		   xctrl.cfg.single_nb);
  }

  return TRUE;                      /* prevent from further event processing */
}



/* Switches between training notebook pages.

Revision: 01.10.2004 E.Sack

Input : - GTK convention
utput: - -
*/
static void
cb_training_page_switch(GTK_CONVENTION_NOTEBOOK_RTS_VOID)
{
  PAGE_TABS current_page, page_tab;

  STSFY_GCC(user_data);

  assert(notebook);
  assert(GTK_IS_NOTEBOOK(notebook));
  assert(page);

  if (gtk_notebook_get_current_page(notebook) < 0) {
    return;
  }

  current_page = xplab_get_page_enum(GTK_NOTEBOOK(notebook), -1);

  if (current_page == XPLAB_TAB_UNDEF) {
    return;
  }

  /* convert page number to page enumeration                               */
  page_tab = xplab_get_page_enum(GTK_NOTEBOOK(notebook), page_num);

  if (current_page == XPLAB_TRAINING_PAGE_IS_SELECT && 
      page_tab == XPLAB_TRAINING_PAGE_IS_TRACE) {
    page_select_stop();
    page_trace_start();
  }
  else if (current_page == XPLAB_TRAINING_PAGE_IS_TRACE && 
	   page_tab == XPLAB_TRAINING_PAGE_IS_SELECT) {
    page_trace_stop();
    page_select_start();
  }
  else {
    assert(0);                                      /* page selection fault */
  }

  assert(G_OBJECT(tr.page.trace.tchild)->ref_count <= 2);
  assert(G_OBJECT(tr.page.revert.rparent)->ref_count == 1);

  trace_sensitiveness();
}



/* Stops the trace page.

Revision: 26.08.2005 E.Sack

Input : - pointer to training settings
Output: - trace page stopped
*/
static void
page_trace_stop(void)
{
  area_def dsize = {0,0};

  cb_trace_stop_button(NULL, NULL);                  /* stop pattern tracing */

  /* clear drawing area                                                      */
  gtk_widget_set_size_request
    (GTK_WIDGET(tr.page.trace.tdata.traced.spattn.pscroll.sdrawing),
     dsize.width, dsize.height);
}



/* Stops the select page.

Revision: 20.09.2003 E.Sack

Input : - pointer to training settings
Output: - select page stopped
*/
static void
page_select_stop(void)
{
  gtk_widget_hide(tr.page.parse_button);

  cb_select_stop_button(NULL, NULL);            /* stop pattern correlation */
}



/* Starts the select page.

Revision: 17.08.2003 E.Sack

Input : - pointer to training settings
Output: - select page started
*/
static void
page_select_start(void)
{
  rb_update();

  gtk_widget_show(tr.page.parse_button);
}



/* Starts the trace page.

Revision: 26.08.2005 E.Sack

Input : - pointer to training settings
Output: - trace page started
*/
static void
page_trace_start(void)
{
  int        cidx = -1, flag, identic_class = 0, traceable_class = 0;
  gchar      *txt = NULL;          
  sector_def msec;
  GtkEntry   *centry;

  flag = xplab_check_entry(tr.page.select.satt.class, MAX_LEN_CLASS_NAME);

  if (flag != XPLAB_STRING_NOT_USABLE) {
    gchar *class;

    assert(GTK_IS_COMBO_BOX(tr.page.select.satt.class.combo)==TRUE);
    centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(tr.page.select.satt.class.combo))->child);
    class = xplab_entry_get_text(centry);                    /* fetch string */
    assert(class);

    /* search class in trained list                                          */
    cidx = search_class(class, *tr.list_old.cls);

    FREEMEM(class);
  }

  if (cidx >= 0) {                                            /* class found */
    identic_class = attributes_identic(tr.list_old.cls->cls[cidx], tr.page.select.satt);
  }

  if (!identic_class) {                         /* no trained class selected */
    xplab_label_set_text(tr.page.revert.label, TRAINING_WARNING_10);
  }
  else {
    traceable_class = class_traceable(tr.list_old.cls->cls[cidx]);

    if (!traceable_class) {
      xplab_label_set_text(tr.page.revert.label, TRAINING_WARNING_11);
    }
    else {                               /* create the trained pattern image */
      traceable_class = 
	create_pattern_drawing(tr.list_old.cls->cls[cidx],  
			       &tr.page.trace.tdata.trained.spattn, 
			       GTK_SIGNAL_FUNC(cb_expose_trace_old),
			       GTK_SIGNAL_FUNC(cb_mouse_pattern_trained));
    }
  }

  if (!traceable_class) {                             /* class not traceable */
    /* exchange trace page with reversionary page if not already done        */
    if (GTK_WIDGET(tr.page.trace.tparent) == 
	GTK_WIDGET(tr.page.trace.tchild)->parent) {
      assert(GTK_WIDGET(tr.page.revert.rparent)->parent == NULL);/* not connected */

      gtk_widget_ref(tr.page.trace.tchild);                    /* keep alive */
      gtk_container_remove(GTK_CONTAINER(tr.page.trace.tparent), 
			   tr.page.trace.tchild);
      gtk_container_add(GTK_CONTAINER(tr.page.trace.tparent), 
			tr.page.revert.rparent);
      gtk_widget_show_all(tr.page.revert.rparent);

      if (G_OBJECT(tr.page.revert.rparent)->ref_count == 2) {
	gtk_widget_unref(tr.page.revert.rparent);/* don't know why not working for the first time*/
      }
    }
    else {                                          /* right child connected */
      assert(GTK_WIDGET(tr.page.trace.tparent) == 
	     GTK_WIDGET(tr.page.revert.rparent)->parent);
    }
  }
  else {   /* exchange reversionary page with trace page if not already done */
    if (GTK_WIDGET(tr.page.trace.tparent) == 
	GTK_WIDGET(tr.page.revert.rparent)->parent) {
      assert(GTK_WIDGET(tr.page.trace.tchild)->parent == NULL); /* not connected */

      gtk_widget_ref(tr.page.revert.rparent);                 /* keep alive */
      gtk_container_remove(GTK_CONTAINER(tr.page.trace.tparent), 
			   tr.page.revert.rparent);
      gtk_container_add(GTK_CONTAINER(tr.page.trace.tparent), 
			tr.page.trace.tchild);
      assert(G_OBJECT(tr.page.trace.tchild)->ref_count == 2);
      gtk_widget_unref(tr.page.trace.tchild);

      gtk_widget_show_all(tr.page.trace.tchild);
    }
    else {                                          /* right child connected */
      assert(GTK_WIDGET(tr.page.trace.tparent) == 
	     GTK_WIDGET(tr.page.trace.tchild)->parent);
    }

    /* setup trace page parameters */
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tr.page.select.nb_button),
				 TRUE);                  /* discard NNB list */
    tr.page.trace.tidx = cidx;

    int_to_string("", tr.list_old.cls->cls[cidx].nbl.cnt,
		  " trained candidate", &txt);
    add_string(tr.list_old.cls->cls[cidx].nbl.cnt==1?"":"s", &txt);

    xplab_label_set_text(tr.page.trace.infoline, txt);
    FREEMEM(txt);

    xplab_label_set_text(tr.trun.plabel, "");

    reset_trace_data(&tr.page.trace.tdata, tr.page.trace.tview);

    AREA_TO_SECTOR(tr.session.mimg.pimg.ar, msec);
    set_pixel_masked(msec, (PICT)~XPLAB_TRAINED_PIXEL, FOREGROUND, 
		     tr.session.mimg.pimg, FALSE);

    image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);

    gtk_widget_show(tr.page.parse_button);
  }
}



/* Recreates a pattern drawing (RGB) from a class.

Revision: 26.09.2004 E.Sack

Input : - pointer to training settings
        - class
        - pointer to image composition
        - image event handler
Output: - image composition updated
        - flag (0: no image built, 1: image built)
Postcondition: - image contains logical and drawable color pixels.
*/
static int
create_pattern_drawing(cls_def cls, pattern_def *pdef, 
		       GtkSignalFunc drawing_cb, GtkSignalFunc mouse_cb)
{
  int netto_h;

  free_pic_def(&pdef->sample);                        /* deallocate old data */
  init_sector_def(&pdef->cfoc);                         /* reset image focus */

  netto_h = (int)(pdef->pscroll.h_adj->page_size - (gfloat)XPLAB_IMG_BORDER_SIZE);

  pdef->sample = xplab_sample_image(cls, netto_h,      /* NB auxiliaries set */
				    XPLAB_TRACE_PATTERN_SPACING, 
				    XPLAB_TRACE_PATTERN_SPACING); 
  assert(!pdef->sample.col);                      /* only binary NB patterns */

  if (!AREA_SIZE(pdef->sample.ar)) {
    return 0;
  }
  else {
    /* create drawable image, which is filled later using logical image      */
    CALLOC(col_def, AREA_SIZE(pdef->sample.ar), pdef->sample.col);

    xplab_widget_destroy(&pdef->pscroll.h_viewport);        /* CBs destroyed */

    pdef->pscroll.sdrawing = 
      horizontal_viewport(pdef->pscroll.v_viewport, &pdef->pscroll.h_viewport, 
			  pdef->sample.ar, mouse_cb, drawing_cb);
  }

  return 1;
}



/* Redraws trace page drawings.

Revision: 01.10.2004 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_expose_trace_old(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);

  assert(widget);
  assert(event);

  pattern_drawing(tr.page.trace.tdata.trained.spattn, tr.timg.pcol, 0);

  return TRUE;
}



/* Redraws trace page drawings.

Revision: 01.10.2004 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_expose_trace_new(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);

  assert(widget);
  assert(event);

  pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);

  return TRUE;
}



/* Draws pattern images to screen.

Revision: 01.10.2004 E.Sack

Input : - image composition
Output: - image composition drawed
*/
static void
pattern_drawing(pattern_def pdef, pattern_color col, int outside_unchanged)
{
  if (!pdef.sample.mem) {                               /* consistency check */
    assert(!pdef.sample.col);
    return;                                           /* image not available */
  }
  else {
    assert(pdef.sample.col);
  }

  xprint_rb(pdef.sample.col, pdef.sample, col, pdef.cfoc, outside_unchanged);

  gdk_draw_rgb_image(pdef.pscroll.sdrawing->window, 
		     pdef.pscroll.sdrawing->style->fg_gc[GTK_STATE_NORMAL],
		     0, 0, 
		     (gint)pdef.sample.ar.width, (gint)pdef.sample.ar.height,
		     GDK_RGB_DITHER_NONE,
		     (guchar *)pdef.sample.col,                     /* image */
		     (gint)pdef.sample.ar.width * 3);
}



/* Redraws the screen from the backing image.

Revision: 16.08.2005 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_image_expose_event(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN)
{
  STSFY_GCC(widget);
  STSFY_GCC(event);

  assert(user_data==NULL);

  if (!tr.rgbbuf) {             /* image buffer in backing store of server */
    if (AREA_SIZE(tr.session.mimg.pimg.ar)) {                /* does exist */
      CALLOC(col_def, AREA_SIZE(tr.session.mimg.pimg.ar), tr.rgbbuf);
      image_to_state(1, 0);
      center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
    }
  }
  else {
    image_to_state(0, 0);
  }

  return FALSE;
}



/* Deletes dialog data.

Revision: 03.10.2003 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_training_dialog_delete(GTK_CONVENTION_DIALOG_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);
  STSFY_GCC(event);
  STSFY_GCC(widget);

  cb_training_dialog_quit(NULL, NULL);                  /* event to signal */

  return FALSE;
}



/* Checks the image slider values.

Revision: 14.06.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_slider(GTK_CONVENTION_ADJUSTMENT_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(adjustment);

  if (tr.timg.state.magnitude == XPLAB_MAGN_X1_VALUE) {   /* original size */
    return;
  }

  image_to_state(1, 0);                                /* ...and display it */
}



/* Reset the focus values.

Revision: 12.07.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_focus_reset(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(menuitem);

  tr.page.select.rb_no_old = -1;

  cb_select_rb_slider(NULL, NULL);                  /* direct handler call */
}



/* Looks for next RB with marked pixels. A flag indicates wether the
   search shall start over on the opposite range border to get the
   complete loop.

Revision: 11.06.2004 E.Sack

Input : - mark
        - direction
        - flag (FALSE: search without loop, TRUE: search in a loop)
Output: - number of rb (-1: no RB found)
Usage: Since this function may be used by an user controlled slider which 
       can perform confusing jumps at the slide endings, the flag prevents
       this behavior. For instance, if an entry is used for input there is
       maximal one (not confusing) jump and the flag can be set. There may 
       be still jumps if the difference of old and new slider values is too
       big, however, this is a subject to the caller.
Postcondition: - no changes to global parameters.
*/
static int
xsearch_rb_masked(PIXEL_KEY key, int dir, gboolean flip)
{
  int found, rbn, first = 0;

  assert(tr.page.select.rb_no >= 0);
  assert(tr.page.select.rb_no < tr.trbl.cnt);

  dir = (dir >= 0) ? 1 : -1;

  rbn = tr.page.select.rb_no;

  while (1) {
    found = set_pixel_masked(tr.trbl.rb[rbn].sc, key, key, 
			     tr.session.mimg.pimg, TRUE);    /* count mask */

    if (found) {
      break;                                         /* masked pixel found */
    }

    if (rbn == tr.page.select.rb_no && first) {
      rbn = -1;
      break;                /* no masked pixel found after a complete loop */
    }

    rbn += dir;

    if (flip == FALSE) {
      if ((rbn >= tr.trbl.cnt) || (rbn < 0)) {               /* don't flip */
	rbn = -1;                /* no masked pixel found in one direction */
	break;
      }
    }
    else {
      if (rbn >= tr.trbl.cnt) {                     /* flip around borders */
	rbn = 0;
      }
      else if (rbn < 0) {
	rbn = tr.trbl.cnt-1;
      }
    }

    first = 1;
  }

  return rbn;
}



/* Wraps interface.

Revision: 16.03.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_rb_slider_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  cb_select_rb_slider(NULL, NULL);
}



/* Wraps interface.

Revision: 16.03.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_focus_slider_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  cb_select_focus_slider(NULL, NULL);
}



/* Checks the slider values.

Revision: 24.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_untrained_rb(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  STSFY_GCC(checkmenuitem);
  STSFY_GCC(user_data);

  if (GTK_CHECK_MENU_ITEM(tr.page.select.untrained_rb_item)->active == TRUE) {
    cb_select_rb_slider(NULL, NULL);
  }
}



/* Checks the slider values.

Revision: 11.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_rb_slider(GTK_CONVENTION_ADJUSTMENT_RTS_VOID)
{
  gchar *cbuf = NULL;
  gint  rbn, dir;

  STSFY_GCC(user_data);
  STSFY_GCC(adjustment);              /* may be NULL because of direct call */

  /* slider to RB numbers */
  tr.page.select.rb_no = (gint)tr.page.select.rb_slider.sadjustmnt->value;
  assert(tr.page.select.rb_no<tr.trbl.cnt);
  assert(tr.page.select.rb_no>=0);

  dir = tr.page.select.rb_no - tr.page.select.rb_no_old;

  if (dir) {                                        /* query first source */
    if (abs(dir) > 1) {                               /* fit slider value */
      rbn = tr.page.select.rb_no - ((dir > 0) ? 1 : -1);

      if ((rbn < tr.trbl.cnt) && (rbn >= 0)) {
	tr.page.select.rb_no_old = rbn;
      }
    }

    if (GTK_CHECK_MENU_ITEM(tr.page.select.untrained_rb_item)->active == TRUE) {
      /* New search for untrained RB overwrites initial RB index
         change, which serves as a trigger. Special action is made if
         no trainable RBs are left.                                        */

      rbn = xsearch_rb_masked(~XPLAB_TRAINED_PIXEL, dir, FALSE);/* search further */

      if (rbn == -1) {                     /* no trainable RB in direction */
	rbn = xsearch_rb_masked(~XPLAB_TRAINED_PIXEL, -dir, FALSE);/* search other */

	if (rbn == -1) {                         /* no trainable RB at all */
	  xplab_label_set_text(tr.page.select.satt.infoline, 
			       XPLAB_NO_TRAINABLE_PIXEL_MSG);
	}

	tr.page.select.rb_no = tr.page.select.rb_no_old;              /* reset to old state */
      }
      else {                                         /* trainable RB found */
	if (tr.page.select.rb_no_old == rbn) {  /* keep state consistent:  */
	  tr.page.select.rb_no_old =     /* the last starting search index */
	    tr.page.select.rb_no;        /* has to be the old one.         */
	}

	tr.page.select.rb_no = rbn;        /* initial index is overwritten */
      }

      /* block RB handler to avoid superfluous call back trigger.          */
      g_signal_handler_block((gpointer)tr.page.select.rb_slider.sadjustmnt, tr.page.select.rb_id);

      /* adjust slider */
      gtk_adjustment_set_value(GTK_ADJUSTMENT(tr.page.select.rb_slider.sadjustmnt), tr.page.select.rb_no);

      /* unblock rb handler */
      g_signal_handler_unblock((gpointer)tr.page.select.rb_slider.sadjustmnt, tr.page.select.rb_id);
    }
  }
  else {                                              /* query second source */
    gchar *buf;

    buf = xplab_entry_get_text(tr.page.select.rb_slider.senty);
    assert(buf);

    if (!strlen(buf)) {
      tr.page.select.rb_no = 0;
    }
    else {
      tr.page.select.rb_no = atoi(buf);                  /* text to RB index */
    }

    FREEMEM(buf);

    tr.page.select.rb_no = XPMAX(0, tr.page.select.rb_no);/* discard negative numbers */
    tr.page.select.rb_no = XPMIN(tr.trbl.cnt-1, tr.page.select.rb_no);/* discard large numbers */

    if (GTK_CHECK_MENU_ITEM(tr.page.select.untrained_rb_item)->active == TRUE) {
      rbn = xsearch_rb_masked(~XPLAB_TRAINED_PIXEL, dir, TRUE);/* search further */

      if (rbn == -1) {                           /* no trainable RB at all */
	xplab_label_set_text(tr.page.select.satt.infoline, 
			     XPLAB_NO_TRAINABLE_PIXEL_MSG);
      }
      else {
	tr.page.select.rb_no = rbn;
      }
    }

    dir = tr.page.select.rb_no - tr.page.select.rb_no_old;

    if (dir) {
      /* block rb handler to avoid superfluous call back */
      g_signal_handler_block((gpointer)tr.page.select.rb_slider.sadjustmnt, tr.page.select.rb_id);

      /* adjust slider */
      gtk_adjustment_set_value(GTK_ADJUSTMENT(tr.page.select.rb_slider.sadjustmnt), tr.page.select.rb_no);

      /* unblock rb handler */
      g_signal_handler_unblock((gpointer)tr.page.select.rb_slider.sadjustmnt, tr.page.select.rb_id);
    }
  }

  /* redisplay in textline, doesn't need to block because call back not triggered */
  int_to_string("", tr.page.select.rb_no, "", &cbuf);
  xplab_entry_set_text(tr.page.select.rb_slider.senty, cbuf);

  FREEMEM(cbuf);

  /* update focus */
  cb_select_focus_slider(NULL, NULL);               /* direct handler call */

  tr.page.select.rb_no_old = tr.page.select.rb_no;
}


/* Checks the slider values.

Revision: 30.06.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_focus_slider(GTK_CONVENTION_ADJUSTMENT_RTS_VOID)
{
  char *cbuf = NULL;
  gint max_focus = (gint)GTK_ADJUSTMENT(tr.page.select.focus_slider.sadjustmnt)->upper;

  STSFY_GCC(user_data);
  STSFY_GCC(adjustment);              /* may be NULL because of direct call */

  /* slider to focus */
  tr.page.select.focus = (gint)gtk_adjustment_get_value(tr.page.select.focus_slider.sadjustmnt);

  if (tr.page.select.focus_old != tr.page.select.focus) {/* query first source */
    sector_from_focus();
  }
  else {                                              /* query second source */
    gchar *buf;

    buf = xplab_entry_get_text(tr.page.select.focus_slider.senty);
    assert(buf);

    if (!strlen(buf)) {
      tr.page.select.focus = XPLAB_FOCUS_SLIDER_MIN;
    }
    else {
      tr.page.select.focus = atof(buf);                     /* text to focus */
    }

    FREEMEM(buf);

    tr.page.select.focus = XPMAX(XPLAB_FOCUS_SLIDER_MIN, tr.page.select.focus);/* discard negatives*/
    tr.page.select.focus = XPMIN(max_focus, tr.page.select.focus);/* discard large numbers */

    if (tr.page.select.focus_old != tr.page.select.focus) {
      sector_from_focus();

      /* block focus handler to avoid superfluous call back */
      g_signal_handler_block((gpointer)tr.page.select.focus_slider.sadjustmnt, tr.page.select.focus_id);

      /* adjust slider */
      gtk_adjustment_set_value(GTK_ADJUSTMENT(tr.page.select.focus_slider.sadjustmnt), (gfloat)tr.page.select.focus);

      /* unblock focus handler */
      g_signal_handler_unblock((gpointer)tr.page.select.focus_slider.sadjustmnt, tr.page.select.focus_id);
    }
    else if (tr.page.select.rb_no_old != tr.page.select.rb_no) {/* query third source */
      /* reset focus to RB frame */
      tr.page.select.focus_sector = tr.page.select.focus_sector_old = tr.trbl.rb[tr.page.select.rb_no].sc;
      tr.page.select.focus_rb_no = tr.page.select.rb_no;

      GTK_ADJUSTMENT(tr.page.select.focus_slider.sadjustmnt)->upper =
	(gfloat)max_focus_aspect(tr.trbl.rb[tr.page.select.focus_rb_no].sc, 
				 &tr.page.select.focus);

      if (tr.page.select.focus_old != tr.page.select.focus) {
	/* block focus handler to avoid superfluous call back */
        g_signal_handler_block((gpointer)tr.page.select.focus_slider.sadjustmnt, tr.page.select.focus_id);
	gtk_adjustment_set_value(GTK_ADJUSTMENT(tr.page.select.focus_slider.sadjustmnt), tr.page.select.focus);
 
	/* unblock focus handler */
        g_signal_handler_unblock((gpointer)tr.page.select.focus_slider.sadjustmnt, tr.page.select.focus_id);
      }
    }
  }

  if (tr.page.select.rb_no_old == tr.page.select.rb_no &&
      tr.page.select.focus_old == tr.page.select.focus) {
    return;
  }

  clear_list_def(&tr.list_new);                        /* discard NNB list */

  /* redisplay in textline doesn't need to be blocked because not activated */
  int_to_string("", (int)tr.page.select.focus, "", &cbuf);
  xplab_entry_set_text(tr.page.select.focus_slider.senty, cbuf);

  FREEMEM(cbuf);

  rb_update();

  tr.page.select.focus_old = tr.page.select.focus;
}



/* NB marking.

Revision: 13.09.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_nb_button(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(togglebutton);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nb_button)) == FALSE) {
    set_pixel_masked(tr.page.select.focus_sector, (PICT)~XPLAB_TRAINED_PIXEL, 
		     FOREGROUND, tr.session.mimg.pimg, FALSE);
    return;
  }

  gtk_widget_set_sensitive(tr.page.select.nbhood, TRUE);

  clear_list_def(&tr.list_new);                          /* discard NNB list */

  /* mark all pixel in focus for NB training                                 */
  set_pixel_masked(tr.page.select.focus_sector, (PICT)~XPLAB_TRAINED_PIXEL, 
		   XPLAB_SELECTED_PIXEL, tr.session.mimg.pimg, FALSE);

  center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);
}



/* Prepares data for CNB training. Reset pixel marking.

Revision: 13.09.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_cnb_button(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(togglebutton);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.cnb_button)) == FALSE) {
    set_pixel_masked(tr.page.select.focus_sector, XPLAB_SELECTED_PIXEL, 
		     XPLAB_RESERVED_PIXEL, tr.session.mimg.pimg, FALSE);
    return;
  }

  gtk_widget_set_sensitive(tr.page.select.nbhood, FALSE);

  /* reset all pixel in focus for CNB training */
  set_pixel_masked(tr.page.select.focus_sector,   /* pixels not selectable */
		   (PICT)~(XPLAB_TRAINED_PIXEL|XPLAB_RESERVED_PIXEL),
		   FOREGROUND, tr.session.mimg.pimg, FALSE);

  set_pixel_masked(tr.page.select.focus_sector, XPLAB_RESERVED_PIXEL,
		   XPLAB_SELECTED_PIXEL, tr.session.mimg.pimg, FALSE);

  center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);
}



/* Searches NNBs from bottom to top.

Revision: 14.07.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_left_arrow(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  int n4;

  STSFY_GCC(user_data);

  assert(button);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nnb_button)) == FALSE) {
    return;
  }

  n4 = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nbhood)) == TRUE?
    1:0;
  mark_nnb(1, n4);              /* mark pattern in focus for NNB training */

  center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);
}



/* Selects a pattern in trace selection format.

Revision: 18.108.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_trace_select(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  int cnr = XPLAB_SINGLE_CLASS, nnr;

  STSFY_GCC(user_data);
  STSFY_GCC(button);

  if (!tr.page.trace.tdata.lbls.nr) {                   /* consistency check */
    assert(!tr.page.trace.tdata.lbls.cnt);
    return;
  }
  else {
    assert(tr.page.trace.tdata.lbls.cnt>0);
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.trace.all_button)) == TRUE) {
    /* select all traced pattern                                             */
    set_int_lst(tr.page.trace.tdata.lbls, XPLAB_SELECTED_PIXEL,
		XPLAB_TRAINED_PIXEL);

    /* coloring selected traced patterns                                     */
    color_indexed_cls_lst(tr.page.trace.tdata.traced.sclass, 
			  tr.page.trace.tdata.lbls, 
			  tr.session.mimg.pimg, 
			  tr.page.trace.tdata.traced.spattn.sample);

    pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);
  }
  else {
    nnr = tr.page.trace.tdata.chno;
    assert(nnr < tr.page.trace.tdata.lbls.cnt);

    if (tr.page.trace.tdata.lbls.nr[nnr] == XPLAB_TRAINED_PIXEL) {
      xplab_label_set_text(tr.trun.plabel, XPLAB_SELECT_TRAINED);
    }
    else {                                   /* select single traced pattern */
      tr.page.trace.tdata.lbls.nr[nnr] = XPLAB_SELECTED_PIXEL;

      /* coloring selected traced pattern                                    */
      color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr], 
		   tr.session.mimg.pimg, 
		   tr.page.trace.tdata.traced.spattn.sample, 
		   XPLAB_SELECTED_PIXEL);

      select_child(&tr.page.trace.tdata, tr.page.trace.tview);
    }
  }

  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);

  selected_candidates_message(tr.page.trace.tdata.lbls, tr.trun.plabel);
}



/* Prints a message.

Revision: 20.09.2003 E.Sack

Input : - integer list
Output: - message
*/
static void
selected_candidates_message(int_lst lbls, GtkWidget *label)
{
  gint  i, d;
  gchar *buf = NULL;

  for (i = d = 0; i < lbls.cnt; i++) {
    d += (lbls.nr[i] == XPLAB_SELECTED_PIXEL) ? 1:0;
  }

  int_to_string("", d, " selected candidate", &buf);

  add_string(d==1?"":"s", &buf);

  xplab_label_set_text(label, buf);

  FREEMEM(buf);
}



/* Deselects a pattern in trace selection format.

Revision: 18.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_trace_deselect(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  int cnr = XPLAB_SINGLE_CLASS, nnr;

  STSFY_GCC(user_data);

  assert(button);

  if (!tr.page.trace.tdata.lbls.nr) {                   /* consistency check */
    assert(!tr.page.trace.tdata.lbls.cnt);
    return;
  }
  else {
    assert(tr.page.trace.tdata.lbls.cnt>0);
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.trace.all_button)) == TRUE) {
    /* deselect all traced pattern                                           */
    set_int_lst(tr.page.trace.tdata.lbls, XPLAB_SELECTBL_PIXEL,
		XPLAB_TRAINED_PIXEL);

    /* coloring deselected traced patterns                                   */
    color_indexed_cls_lst(tr.page.trace.tdata.traced.sclass, 
			  tr.page.trace.tdata.lbls, 
			  tr.session.mimg.pimg, 
			  tr.page.trace.tdata.traced.spattn.sample);
    pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);
  }
  else {
    nnr = tr.page.trace.tdata.chno;
    assert(nnr < tr.page.trace.tdata.lbls.cnt);

    if (tr.page.trace.tdata.lbls.nr[nnr] == XPLAB_TRAINED_PIXEL) {
      xplab_label_set_text(tr.trun.plabel, XPLAB_DESELECT_TRAINED);
    }
    else {                                 /* deselect single traced pattern */
      tr.page.trace.tdata.lbls.nr[nnr] = XPLAB_SELECTBL_PIXEL;

      /* coloring deselected traced pattern                                  */
      color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr], 
		   tr.session.mimg.pimg, 
		   tr.page.trace.tdata.traced.spattn.sample,
		   XPLAB_SELECTBL_PIXEL);

      select_child(&tr.page.trace.tdata, tr.page.trace.tview);
    }
  }

  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);

  selected_candidates_message(tr.page.trace.tdata.lbls, tr.trun.plabel);
  return;
}



/* (De-)Selects a pattern in trace selection format by the mouse.

Revision: 02.10.2004 E.Sack

Input : - pointer to GDK event
- class list
Output: - -
*/
static void
traced_pattern_label(GdkEventButton *event, cls_lst clst)
{
  int cnr = XPLAB_SINGLE_CLASS, nnr;

  if (!tr.page.trace.tdata.lbls.nr) {                   /* consistency check */
    assert(!tr.page.trace.tdata.lbls.cnt);
    return;
  }
  else {
    assert(tr.page.trace.tdata.lbls.cnt>0);
  }

  nnr = coordinates_to_index(clst.cls[cnr], (int)event->x, (int)event->y);

  if (nnr >= 0) {
    traced_nb_label(nnr, 1);
  }                          /* else mouse coordinates not within NB image */
}



/* (De-)Selects a traced NB.

Revision: 02.10.2004 E.Sack

Input : - traced NB index
Output: - traced NB label changed
*/
static void
traced_nb_label(int nb_idx, int center)
{
  int        cnr = XPLAB_SINGLE_CLASS;
  sector_def msec;

  assert(nb_idx >= 0 && nb_idx < tr.page.trace.tdata.lbls.cnt);

  switch (tr.page.trace.tdata.lbls.nr[nb_idx]) {
  case XPLAB_TRAINED_PIXEL:
    xplab_label_set_text(tr.trun.plabel, XPLAB_CHANGE_TRAINED);
    break;

  case XPLAB_SELECTED_PIXEL:             /* deselect single traced pattern */
    tr.page.trace.tdata.lbls.nr[nb_idx] = XPLAB_SELECTBL_PIXEL;

    /* coloring deselected traced pattern */
    msec = color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nb_idx],
			tr.session.mimg.pimg, 
			tr.page.trace.tdata.traced.spattn.sample,
			XPLAB_SELECTBL_PIXEL);

    if (center) {
      center_image(tr.timg, msec);
    }

    pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);
    break;

  case XPLAB_SELECTBL_PIXEL:               /* select single traced pattern */
    tr.page.trace.tdata.lbls.nr[nb_idx] = XPLAB_SELECTED_PIXEL;

    /* coloring selected traced pattern */
    msec = color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nb_idx],
			tr.session.mimg.pimg, 
			tr.page.trace.tdata.traced.spattn.sample,
			XPLAB_SELECTED_PIXEL);

    if (center) {
      center_image(tr.timg, msec);
    }

    pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);
    break;

  default: break;
  }

  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);

  selected_candidates_message(tr.page.trace.tdata.lbls, tr.trun.plabel);
}



/* Selects a left pattern in trace selection format.

Revision: 01.10.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_trace_left_arrow(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(button);

  tr.page.trace.tdata.chno--;
  select_child(&tr.page.trace.tdata, tr.page.trace.tview);     /* left child */

  pattern_drawing(tr.page.trace.tdata.trained.spattn, tr.timg.pcol, 0);
  pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);
}



/* Selects a right pattern in trace selection format.

Revision: 01.10.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_trace_right_arrow(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  tr.page.trace.tdata.chno++;
  select_child(&tr.page.trace.tdata, tr.page.trace.tview);    /* right child */

  pattern_drawing(tr.page.trace.tdata.trained.spattn, tr.timg.pcol, 0);
  pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);
}



/* Returns the text index from the tree view by the row index.

Revision: 20.09.2003 E.Sack

Usage: Whenever row indexes of a tree view are used, they should converted to
       their corresponding TPS indexes (e.g. class indexes) because the
       latter are fixed in the training phase. The row indexes depend on 
       their sorting.

Input : - pointer to tree view
        - pointer to path
Output: - text index
*/
static gint
get_candidate_index(GtkTreeView *treeview, GtkTreePath *path)
{
  gint         idx;
  gboolean     row_found;
  GtkTreeIter  iter;
  GtkTreeModel *model;

  model = gtk_tree_view_get_model(treeview);
  assert(XPLAB_TR_TREE_STORE_COLUMNS == gtk_tree_model_get_n_columns(model));

  row_found = gtk_tree_model_get_iter(model, &iter, path);
  assert(row_found == TRUE);                            /* index must be in */

  gtk_tree_model_get(model, &iter, XPLAB_TR_TREE_STORE_COLUMN_IDX, &idx, -1);

  return idx;
}



/* Searches NNBs from top to bottom.

Revision: 14.07.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_right_arrow(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  int n4;

  STSFY_GCC(user_data);

  assert(button);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nnb_button)) == FALSE) {
    return;
  }

  n4 = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nbhood)) == TRUE?
    1:0;
  mark_nnb(0, n4);               /* mark pattern in focus for NNB training */

  center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);
}



/* Resets pixel marking according to neighborhood.

Revision: 20.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_nbhood(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(button);

  clear_list_def(&tr.list_new);                          /* discard NNB list */

  rb_update();
}



/* Enables NNB pattern training.

Revision: 13.09.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_nnb_button(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID)
{
  int n4;

  STSFY_GCC(user_data);
  STSFY_GCC(togglebutton);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nnb_button)) == FALSE) {
    gtk_widget_set_sensitive(tr.page.select.nnb_lbutton, FALSE);
    gtk_widget_set_sensitive(tr.page.select.nnb_rbutton, FALSE);
    set_pixel_masked(tr.page.select.focus_sector, /* pixels not selectable */
		     (PICT)~(XPLAB_TRAINED_PIXEL|XPLAB_RESERVED_PIXEL),
		     FOREGROUND, tr.session.mimg.pimg, FALSE);
    return;
  }

  gtk_widget_set_sensitive(tr.page.select.nnb_lbutton, TRUE);
  gtk_widget_set_sensitive(tr.page.select.nnb_rbutton, TRUE);
  gtk_widget_set_sensitive(tr.page.select.nbhood, TRUE);

  set_pixel_masked(tr.page.select.focus_sector,       /* pixels selectable */
		   (PICT)~(XPLAB_TRAINED_PIXEL|XPLAB_RESERVED_PIXEL),
		   XPLAB_SELECTBL_PIXEL, tr.session.mimg.pimg, FALSE);

  n4 = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nbhood)) == TRUE?
    1:0;
  mark_nnb(0, n4);               /* mark pattern in focus for NNB training */
 
  center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);
}



/* Cycles NNB pattern in focus.

Revision: 13.09.2004 E.Sack

Input : - pointer to training settings
        - neighborhood
Output: - NNB pattern updated
*/
static void
mark_nnb(int bot_up, int n4)
{
  int found, ignore;

  if (tr.page.select.direction != bot_up) {              /* reverse search */
    /* reset old selected pattern */
    set_pixel(tr.page.select.focus_sector, XPLAB_SELECTED_PIXEL,
	      XPLAB_SELECTBL_PIXEL, tr.session.mimg.pimg, 0);

    while (1) {                             /* select last ignored pattern */
      found = xcontour_mark(tr.session.mimg.pimg, tr.page.select.focus_sector, 
			    XPLAB_IGNORED_PIXEL, XPLAB_SELECTED_PIXEL, 
			    tr.page.select.direction, n4);
      if (!found) {
	tr.page.select.direction = bot_up;
	break;                                              /* no reversal */
      }

      found = set_pixel(tr.page.select.focus_sector, XPLAB_IGNORED_PIXEL, 
			XPLAB_IGNORED_PIXEL, tr.session.mimg.pimg, 0);
      if (!found) {
	set_pixel(tr.page.select.focus_sector, FOREGROUND,
		  XPLAB_IGNORED_PIXEL, tr.session.mimg.pimg, 0);
	return;                                           /* last reversal */
      }

      xcontour_mark(tr.session.mimg.pimg, tr.page.select.focus_sector, 
		    XPLAB_SELECTED_PIXEL, FOREGROUND, 
		    tr.page.select.direction, n4);    /* not last reversal */
    }
  }

  /* ignore old selected pattern */
  set_pixel(tr.page.select.focus_sector, XPLAB_SELECTED_PIXEL,
	    XPLAB_IGNORED_PIXEL, tr.session.mimg.pimg, 0);

  /* select next pattern */
  found = xcontour_mark(tr.session.mimg.pimg, tr.page.select.focus_sector, 
			XPLAB_SELECTBL_PIXEL, XPLAB_SELECTED_PIXEL, bot_up, 
			n4);

  if (!found) {                                   /* reset ignored pattern */
    ignore = set_pixel(tr.page.select.focus_sector, XPLAB_IGNORED_PIXEL,
		       XPLAB_SELECTBL_PIXEL, tr.session.mimg.pimg, 0);

    if (ignore) {                               /* restart pattern cycling */
      found = xcontour_mark(tr.session.mimg.pimg, tr.page.select.focus_sector,
			    XPLAB_SELECTBL_PIXEL, XPLAB_SELECTED_PIXEL, 
			    bot_up, n4);
    }
    else {
      xplab_label_set_text(tr.page.select.satt.infoline, 
			   "no more unmarked pattern in focus");
    }
  }
}



/* Calculates maximal focus slider value. The result depends on the focus
   location in the image and therefore can be different for the same focus
   size.

Revision: 04.07.2003 E.Sack

Input : - pointer to training settings
Output: - focus size updated
*/
static gint
max_focus_aspect(sector_def rsc, gfloat *focus)
{
  gint fwidth, fheight, max_to_border, len;

  max_to_border = XPMAX(rsc.column.start, rsc.line.start);

  max_to_border = XPMAX(max_to_border, 
			tr.session.mimg.pimg.ar.width - rsc.column.end - 1);

  max_to_border = XPMAX(max_to_border, 
			tr.session.mimg.pimg.ar.height - rsc.line.end - 1);

  fwidth = INTERVAL_SIZE(rsc.column);
  fheight = INTERVAL_SIZE(rsc.line);

  len = XPMAX(fwidth, fheight);

  if (len & 1) {       
    len = (len-1) >> 1;      /* odd */
  }
  else {
    len = (len >> 1) - 1;    /* even */
  }

  len = len == 0 ? XPLAB_FOCUS_SLIDER_MIN : len;/* at least one pixel width */

  *focus = (gfloat)len;                 /* default focus value comprises RB */

  return len + max_to_border;
}



/* Calculates the focus size according to the focus slider. The range of the
   smallest focus size is from 1x1 to 2x2 pixels with all permutations.

Revision: 04.07.2003 E.Sack

Input : - pointer to training settings
Output: - focus size updated
*/
static void
sector_from_focus(void)
{
  gint diff;

  /* negative values means smaller focus sizes */
  diff = tr.page.select.focus - tr.page.select.focus_old;

  if (!diff) {
    return;
  }

  limit_focus(diff, &tr.page.select.focus_sector_old.column,
	      &tr.page.select.focus_sector.column,
	      tr.session.mimg.pimg.ar.width);

  limit_focus(diff, &tr.page.select.focus_sector_old.line,
	      &tr.page.select.focus_sector.line,
	      tr.session.mimg.pimg.ar.height);
}



/* Calculates a centered interval keeping the center of a original interval.

Revision: 04.07.2003 E.Sack

Input : - symmetric increment/decrement
- pointer to symmetric interval
- pointer to drawable interval
- drawable maximum
Output: - drawable interval updated
*/
static void
limit_focus(gint diff, ival_def *backstore, ival_def *current, int max)
{
  gint     d_max = 0, d_backstore;
  ival_def last_backstore = *backstore;

  /* backup sector to keep aspect ratio without any conditions */
  backstore->start -= diff;
  backstore->end   += diff;

  d_backstore = INTERVAL_SIZE(*backstore);                     /* forecast */

  if (d_backstore < 1) {                     /* new backstore not drawable */
    d_max = INTERVAL_SIZE(last_backstore) - 1;

    if (d_max >= 0) {           /* last backstore at least partly drawable */
      d_max = XPMIN(-2*diff, d_max)/2;                 /* integer division */
      current->start = last_backstore.start + d_max;
      current->end = last_backstore.end - d_max;
    }                                               /* else size unchanged */
  }
  else {                         /* new backstore at least partly drawable */
    *current = *backstore;                                 /* running size */
  }

  /* maximal size is upper limit */
  current->start = XPMAX(0, XPMIN(current->start, max-1));
  current->end   = XPMAX(0, XPMIN(current->end,   max-1));

  assert(current->start <= current->end);                    /* algo fault */
}



/* Updates data according to new RB or focus selection, respectively.

Revision: 30.06.2003 E.Sack

Input : - pointer to training settings
Output: - data updated
*/
static void
rb_update(void)
{
  sector_def msector;

  xplab_label_set_text(tr.page.select.satt.infoline, ""); /* empty infoline */

  /* calculate new focus location */
  center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);

  /* reset all pixel */
  AREA_TO_SECTOR(tr.session.mimg.pimg.ar, msector);
  set_pixel_masked(msector, (PICT)~XPLAB_TRAINED_PIXEL, FOREGROUND, 
		   tr.session.mimg.pimg, FALSE);

  calculate_blob_number(1);

  if (tr.page.select.blob_cnt <= 1) {/* automatic selection for pattern type */
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nb_button)) == FALSE) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tr.page.select.nb_button), TRUE);
    }
    else {
      cb_select_nb_button(NULL, NULL);
    }
  }
  else {
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nnb_button)) == FALSE) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tr.page.select.nnb_button), TRUE);
    }
    else {
      cb_select_nnb_button(NULL, NULL);
    }
  }

  return;
}



/* Calculate pattern number in focus.

Revision: 13.09.2003 E.Sack

Input : - pointer to training settings
- flag (0: calculate trained number, 1: calculate all pattern numbers)
Output: - pattern numbers output
Precondition: - best usage: flag=1 if there are only both used keys
*/
static void
calculate_blob_number(int both_keys)
{
  gint blob_trained, n4;
  char *buf = NULL;

  n4 = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nbhood)) == TRUE ?
    1:0;

  blob_trained = xcount_contours(tr.session.mimg.pimg, 
				 tr.page.select.focus_sector, 
				 XPLAB_TRAINED_PIXEL, n4);

  if (both_keys) {                           /* calculate foreground blobs */
    tr.page.select.blob_cnt = 
      xcount_contours(tr.session.mimg.pimg, tr.page.select.focus_sector, 
		      FOREGROUND, n4);
  }

  int_to_string("pattern in focus: ", tr.page.select.blob_cnt+blob_trained, 
		" ", &buf);
  int_to_string("(trained: ", blob_trained, ")", &buf);

  xplab_label_set_text(tr.page.select.satt.statusline3, buf);
  FREEMEM(buf);
}



/* Puts/discards RB frames to/from the backing image.

Revision: 30.05.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rb_frames(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  if (checkmenuitem->active == TRUE) {
    tr.timg.state.rb_frames = DC_OPT_RB_FRAME_BMP;
  }
  else {
    tr.timg.state.rb_frames = IMG_WITHOUT_RB_FRAMES;
  }

  image_to_state(1, 0);
}



/* Puts original/binarized input image to the backing image.

Revision: 19.09.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_original(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  if (checkmenuitem->active == TRUE) {
    tr.timg.state.source = DC_OPT_ORG;
    image_to_state(1, 0);
  }
}



/* Puts original/binarized input image to the backing image.

Revision: 19.09.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_binarized(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  if (checkmenuitem->active == TRUE) {
    tr.timg.state.source = DC_OPT_CLEAN;
    image_to_state(1, 0);
  }
}



/* Callback for window mark.

Revision: 19.09.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_mark(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  if (checkmenuitem->active == TRUE) {
    center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
    tr.timg.state.source = DC_OPT_MARKED;
    image_to_state(1, 0);
  }
}



/* Empties the backing image.

Revision: 19.09.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_empty(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  if (checkmenuitem->active == TRUE) {
    tr.timg.state.source = IMG_SOURCE_IS_EMPTY;
    image_to_state(1, 0);
  }
}



/* Sets the magnitude of backing image pixels. The whole image buffer is
   simulated, not allocated.

Revision: 26.09.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_magnitude(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  const gchar *name;
  area_def area;

  STSFY_GCC(user_data);

  if (gtk_check_menu_item_get_active(checkmenuitem) == TRUE) {
    gboolean is_act;
    GSList   *radio_grp;

    radio_grp =                                      /* don't free this list */
      gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(checkmenuitem));
    assert(radio_grp);

    do {                                               /* scan radio buttons */
      is_act = 
	gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(radio_grp->data));

      if (is_act == TRUE) {
	break;
      }

      radio_grp = radio_grp->next;
    } while (radio_grp);

    assert(is_act == TRUE);

    name =                                         /* don't free this string */
      gtk_label_get_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(radio_grp->data))));

    if (!strcmp(name, XPLAB_MAGN_X1_NAME)) {
      tr.timg.state.magnitude = XPLAB_MAGN_X1_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_X2_NAME)) {
      tr.timg.state.magnitude = XPLAB_MAGN_X2_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_X4_NAME)) {
      tr.timg.state.magnitude = XPLAB_MAGN_X4_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_D2_NAME)) {
      tr.timg.state.magnitude = XPLAB_MAGN_D2_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_D4_NAME)) {
      tr.timg.state.magnitude = XPLAB_MAGN_D4_VALUE;
    }
    else {
      assert(0);                               /* magnitude not supported */
    }

    xplab_widget_destroy(&tr.timg.scroll.h_viewport);/* signal handlers destroyed */
    area.width = tr.session.mimg.pimg.ar.width*tr.timg.state.magnitude;
    area.height = tr.session.mimg.pimg.ar.height*tr.timg.state.magnitude;

    tr.timg.scroll.sdrawing = 
      horizontal_viewport(tr.timg.scroll.v_viewport, &tr.timg.scroll.h_viewport, 
			  area, GTK_SIGNAL_FUNC(cb_mouse_training_image),
			  GTK_SIGNAL_FUNC(cb_image_expose_event));
  }
}



/* Dumps window content to file.

Revision: 03.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_dump(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  pic_def src;
  char    *image_name = NULL, *status_msg = NULL;

  STSFY_GCC(user_data);

  assert(menuitem);

  init_pic_def(&src);

  if (!tr.rgbbuf) {
    string_to_log(main_log, TRAINING_WARNING_12, "", "");
    return;
  }

  src.ar = tr.session.mimg.pimg.ar;
  src.col = tr.rgbbuf;

  assert(AREA_SIZE(src.ar) > 0);

  image_name = name_of_image(tr.session.img_name, tr.timg.state, &status_msg);

  dump_pic_def(src, image_name, src.ar, 0, 0, 1, PNM_FORMAT_UNDEF);

  FREEMEM(image_name);
  FREEMEM(status_msg);
}



/* Starts pattern tracing in focus.

Revision: 08.08.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_trace_focus_button(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(button);

  if (*tr.trun.pactive == TRUE) {
    return;
  }

  start_run_def(&tr.trun);

  xtrace(tr.page.select.focus_sector);                      /* trace focus */

  stop_run_def(&tr.trun);

  trace_sensitiveness();
}



/* Starts pattern tracing in RBs.

Revision: 15.08.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_trace_rb_button(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  sector_def bsec;

  STSFY_GCC(user_data);

  assert(button);
  assert(tr.trun.pactive);

  if (*tr.trun.pactive == TRUE) {
    return;
  }

  start_run_def(&tr.trun);

  AREA_TO_SECTOR(tr.session.mimg.pimg.ar, bsec);
  xtrace(bsec);                                         /* trace whole image */

  stop_run_def(&tr.trun);

  trace_sensitiveness();
}



/* Sets the sensitiveness in the trace page.

Revision: 16.05.2004 E.Sack

Input : - -
Output: - sensitiveness updated
*/
static void
trace_sensitiveness(void)
{
  gboolean is_true;

  if (tr.dialog) {
    if (GTK_IS_WIDGET(tr.dialog) == TRUE)
      {
	is_true = tr.page.trace.tdata.traced.sclass.cnt ? TRUE : FALSE;

	gtk_widget_set_sensitive(tr.page.trace.all_button, is_true);
	gtk_widget_set_sensitive(tr.page.trace.select_button, is_true);
	gtk_widget_set_sensitive(tr.page.trace.deselect_button, is_true);
	gtk_widget_set_sensitive(tr.page.trace.left_button, is_true);
	gtk_widget_set_sensitive(tr.page.trace.right_button, is_true);
      }
  }
}



/* Resets the trace page data.

Revision: 26.08.2005 E.Sack

Input : - pointer to trace data
        - pointer to trace store
Output: - trace data reset
Usage : The function resets all data to their initial state, loosing their
        contents. 
*/
static void
reset_trace_data(trace_data_def *tdata, GtkTreeView *tview)
{
  free_cand_lst(&tdata->hit_lst);
  free_cls_lst(&tdata->trained.sclass);
  free_cls_lst(&tdata->traced.sclass);
  free_int_lst(&tdata->lbls);
  free_hash_def(&tdata->thash);

  init_sector_def(&tdata->traced.spattn.cfoc);
  init_sector_def(&tdata->trained.spattn.cfoc);

  tdata->chno = XPLAB_START_NNB_NO;                         /* default child */

  clear_tree_store(tview);                                    /* empty store */
}



/* Traces patterns of a class.

Revision: 29.07.2005 E.Sack

Input : - sector of pattern image
        - pointer to training settings
Output: - training settings updated
Usage:  - this function can generate extensive calculations and consequently 
          retarded event handling.
Postcondition: - on success (some NBs traced) all lists in the trace 
                 definition trace_def are consistently filled up. The lists
                 contain different aspects of the used NBs, especially
                 the relation between trained (parent) and traced (children)
                 NBs.
*/
static void
xtrace(sector_def bsec)
{
  int        new_cnr, new_nnr, traced = 0;
  cls_def    cls_new;
  pic_def    mdup, tmp;
  sector_def msec;

  init_cls_def(&cls_new);
  init_pic_def(&mdup);

  reset_trace_data(&tr.page.trace.tdata, tr.page.trace.tview);

  /* reset pixels in trained pattern drawing                                 */
  set_all_pixel(tr.page.trace.tdata.trained.spattn.sample, 
		XPLAB_TRAINED_PIXEL, FOREGROUND);

  /* pick up selected class                                                  */
  copy_cls_def(&tr.list_old.cls->cls[tr.page.trace.tidx], &cls_new, FALSE);
  orphan_cls_def(cls_new);                         /* no siblings calculable */

  cat_cls_def(cls_new, &tr.page.trace.tdata.trained.sclass, -1, &new_cnr,
	      &new_nnr);
  free_cls_def(&cls_new);
  assert(new_cnr == 0);
  assert(new_nnr == 0);

  /* reset pixels in training image                                          */
  AREA_TO_SECTOR(tr.session.mimg.pimg.ar, msec);
  set_pixel_masked(msec, (PICT)~XPLAB_TRAINED_PIXEL, FOREGROUND, 
		   tr.session.mimg.pimg, FALSE);       /* all unless trained */
  mdup = xdup_pic_def(tr.session.mimg.pimg);              /* duplicate image */
  set_pixel_masked(msec, XPLAB_TRAINED_PIXEL, BACKGROUND, mdup, FALSE);/* erase trained*/
  set_pixel(bsec, FOREGROUND, BACKGROUND, mdup, 1); /* clear sector environmt*/

  reset_rb_lst(tr.trbl);                                   /* clear RB flags */

  tr.page.trace.tdata.hit_lst =                     /* extensive calculation */ 
    xpattern_matching(tr.tad, tr.single_line, tr.trbl, 
		      tr.page.trace.tdata.trained.sclass,
		      &mdup, 0, &tr.trun, tr.single_nb, 0);

  tmp = xplab_request_image(P1, NULL, NULL, 0, NULL).pimg;     /* image copy */
  tr.page.trace.tdata.traced.sclass = 
    xtrace_generation(tr.page.trace.tdata.hit_lst, mdup, tmp, 
		      tr.page.trace.tdata.trained.sclass, 
		      &tr.page.trace.tdata.thash,
		      tr.ckernel, tr.cluster_thold);

  free_pic_def(&tmp);
  free_pic_def(&mdup);

  gen_class_references(tr.page.trace.tdata.traced.sclass, tr.tad.mborder, 
		       tr.tad.maskweight);

  if (tr.page.trace.tdata.traced.sclass.cnt > 0) {/* create the traced pattern image */
    assert(tr.page.trace.tdata.traced.sclass.cnt == 1);/* only one class traced */

    traced = 
      create_pattern_drawing(tr.page.trace.tdata.traced.sclass.cls[XPLAB_SINGLE_CLASS],
			     &tr.page.trace.tdata.traced.spattn, 
			     GTK_SIGNAL_FUNC(cb_expose_trace_new),
			     GTK_SIGNAL_FUNC(cb_mouse_pattern_traced));
  }

  if (traced) {                                            /* image coloring */
    char *txt = NULL;

    int_to_string("", tr.page.trace.tdata.traced.sclass.cls[XPLAB_SINGLE_CLASS].nbl.cnt, " traced candidate", &txt);
    add_string(tr.page.trace.tdata.traced.sclass.cls[XPLAB_SINGLE_CLASS].nbl.cnt==1?"":"s", &txt);

    xplab_label_set_text(tr.trun.plabel, txt);
    FREEMEM(txt);

    /* coloring parents once                                                 */
    color_cand_lst(tr.page.trace.tdata.hit_lst, 
		   tr.page.trace.tdata.trained.sclass, 
		   tr.page.trace.tdata.trained.spattn.sample);

    /* setup selected index list for traced patterns                         */
    tr.page.trace.tdata.lbls = 
      setup_selected_pattern(tr.page.trace.tdata.traced.sclass.cls[XPLAB_SINGLE_CLASS].nbl.cnt);

    /* coloring traced pattern                                               */
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tr.page.trace.all_button), TRUE);
    cb_trace_select(NULL, NULL);                      /* direct handler call */

    /* setup trace model                                                     */
    append_cand_lst_to_model(tr.page.trace.tdata.hit_lst, tr.page.trace.tview, 
			     tr.page.trace.tdata.traced.sclass.cls[XPLAB_SINGLE_CLASS].att.class);

    select_child(&tr.page.trace.tdata, tr.page.trace.tview);/* highlight first child */
  }
  else {
    area_def dsize = {0,0};

    gtk_widget_set_size_request(GTK_WIDGET(tr.page.trace.tdata.traced.spattn.pscroll.sdrawing),
				dsize.width, dsize.height);

    xplab_label_set_text(tr.trun.plabel, "no traced candidates");
  }

  pattern_drawing(tr.page.trace.tdata.trained.spattn, tr.timg.pcol, 0);
  pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);
  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);
}



/* Builds a class list with patterns from a candidate list using clustering
   rules. The classes retains their references from the candidate list.

   Revision: 16.08.2003 E.Sack

   Input : - candidate list
   - working pattern image
   - original pattern image
   - candidate class list
   - pointer to hash list
   - pointer to clustering kernel
   - clustering threshold
   Output: - new class list
   - candidate list marked
*/
static cls_lst 
xtrace_generation(cand_lst hit, pic_def rpic, pic_def rorg, cls_lst old,
	          hash_def *hash, kern_def kernel, int cluster_thold)
{
  int     i, no, co, cn, nn;
  nb_def  nbz;
  cls_lst new;

  init_nb_def(&nbz);
  init_cls_lst(&new);

  if (!hit.cnt) {
    return new;
  }

  for (i = 0; i < hit.cnt; i++) {
    if (hit.c[i].n < 0) {                    /* only recognized candidates */
      continue;
    }

    nbz.pat = cut_mask(rpic, rorg, hit.c[i], old, kernel,/* get NB pattern */
		       cluster_thold);
    clear_nb(rpic, rorg, hit.c[i], old);                /* erase candidate */

    if (!nbz.pat.mem) {                  /* NB empty, nothing to take over */
      free_nb_def(&nbz);                                    /* be pedantic */
      continue;
    }

    co = hit.c[i].c;                       /* short cut for parent indexes */
    no = hit.c[i].n;

    nbz.aux = gen_hash_entry(hash, hit.c[i], &old.cls[co].nbl.nb[no].aux);
    cn = search_class(old.cls[co].att.class, new);

    if (cn >= 0) {                            /* class already in new list */
      REALLOC(nb_def, new.cls[cn].nbl.cnt+1, new.cls[cn].nbl.nb);
    }
    else {                                  /* class not found in new list */
      cn = new.cnt++;

      if (cn) {
	REALLOC(cls_def, cn+1, new.cls);         /* new list was not empty */
      }
      else {
	CALLOC(cls_def, cn+1, new.cls);              /* new list was empty */
      }

      init_cls_def(&new.cls[cn]);            /* take over class attributes */
      copy_cls_att(old.cls[co].att, &new.cls[cn].att);

      CALLOC(nb_def, new.cls[cn].nbl.cnt+1, new.cls[cn].nbl.nb);
    }

    nn = new.cls[cn].nbl.cnt++;                               /* short cut */
    init_nb_def(&new.cls[cn].nbl.nb[nn]);

    move_nb_def(&nbz, &new.cls[cn].nbl.nb[nn], FALSE);/* enroll NB pattern */
  }

  return new;
}



/* Sets up the background focus of a child pattern and its parent.

Revision: 29.07.2005 E.Sack

Input : - pointer to trace settings
        - pointer to child index
Output: - child index updated
        - flag (0: no child found, 1: child found)
Postcondition: - child index wrapped against number of children
*/
static int
select_child(trace_data_def *tdata, GtkTreeView *tview)
{
  int        cid, parent, old_parent;
  sector_def msec;

  cid = select_pattern(tdata->traced.sclass, &tdata->traced.spattn.cfoc,
		       &tdata->chno);

  if (cid == -1 ) {
    return 0;                                              /* no child found */
  }

  assert(cid >= XPLAB_INIT_HASH_ID);                        /* child hash id */

  parent = old_parent = search_candidate_of_child_id(tdata->thash, cid).n;

  select_pattern(tdata->trained.sclass, &tdata->trained.spattn.cfoc, &parent);
  assert(parent==old_parent);

  treeview_moveto(tview, tdata->chno, FALSE);

  /* center the training image                                               */
  msec = 
    color_nb_def(tdata->traced.sclass.cls[XPLAB_SINGLE_CLASS].nbl.nb[tdata->chno],
		 tr.session.mimg.pimg, tdata->traced.spattn.sample,
		 -1);                                   /* don't touch image */
  center_image(tr.timg, msec);

  return 1;                                                   /* child found */
}



/* Sets up the background focus of a parent pattern and its children.

Revision: 29.07.2005 E.Sack

Input : - pointer to trace settings
        - parent index
Output: - flag (0: no child found, 1: child found)
*/
static int
select_children(GtkTreeView *tview, trace_data_def *tdata, int parent)
{
  int        pid, child, reset;
  int_lst    child_lst;
  sector_def test_sec;

  init_int_lst(&child_lst);

  /* trained class only contains parent IDs                                  */
  pid = select_pattern(tdata->trained.sclass, &test_sec, &parent);

  if (pid < XPLAB_INIT_HASH_ID) {
    return 0;
  }

  child_lst = search_child_indexes_of_parent_id(tdata->thash, pid, 
						tdata->traced.sclass);

  if (!child_lst.cnt) {
    assert(!child_lst.nr);
    return 0;
  }

  reset = 1;                                       /* reset last highligthed */
  for (child = 0; child < child_lst.cnt; child++) {
    tdata->chno = child_lst.nr[child];
    select_pattern(tdata->traced.sclass, &tdata->traced.spattn.cfoc, 
		   &tdata->chno);
    pattern_drawing(tdata->traced.spattn, tr.timg.pcol, !reset);/* highlight children */
    reset = 0;
  }

  tdata->chno = child_lst.nr[0];                                /* any child */

  free_int_lst(&child_lst);

  /* set foci                                                                */
  select_pattern(tdata->traced.sclass, &tdata->traced.spattn.cfoc, &tdata->chno);
  select_pattern(tdata->trained.sclass, &tdata->trained.spattn.cfoc, &parent);

  treeview_moveto(tview, tdata->chno, TRUE);

  return 1;                                                /* children found */
}



/* Sets up the background focus of a NB pattern.

Revision: 25.08.2005 E.Sack

Input : - class list
        - pointer to pattern index
Output: - pattern index updated
        - flag (-1: class list empty, >=0: hash ID of NB)
Precondition:  - NB initialized with sample image setting
Postcondition: - pattern index wrapped against number of NB
*/
static int
select_pattern(cls_lst clst, sector_def *sec, int *nnr)
{
  int cnr = XPLAB_SINGLE_CLASS;

  if (!clst.cls) {                                      /* consistency check */
    assert(!clst.cnt);
    return -1;                                           /* class list empty */
  }
  else {
    assert(clst.cnt > 0);
  }

  xwrap_idx(nnr, clst.cls[cnr].nbl.cnt);

  *sec = clst.cls[cnr].nbl.nb[*nnr].dat.sec;

  return clst.cls[cnr].nbl.nb[*nnr].aux;                       /* NB hash ID */
}



/* Searches the parent of a child using the child hash ID.

Revision: 28.08.2003 E.Sack

Input : - generation reference
        - child hash ID
Output: - parent hash ID
*/
static cand_def
search_candidate_of_child_id(hash_def hash, int cid)
{
  int      pair;
  cand_def cand;

  init_cand_def(&cand);

  assert(cid > 0);                              /* positive child hash IDs */

  for (pair = 0; pair < hash.cnt; pair++) {
    if (hash.cid[pair] == cid) {                         /* child ID found */
      assert(hash.pid[pair] > 0);              /* positive parent hash IDs */
      assert(hash.kr[pair].n >= 0);                   /* consistency check */
      assert(hash.kr[pair].c >= 0);
      cand = hash.kr[pair];                       /* candidate of child id */
      break;
    }
  }

  return cand;
}



/* Searches the indexes of the traced patterns using the parent hash ID.

Revision: 01.10.2004 E.Sack

Input : - hash list
        - parent hash ID
        - traced pattern class list
Output: - traced pattern index list
*/
static int_lst
search_child_indexes_of_parent_id(hash_def hash, int pid, 
				  cls_lst traced_children)
{
  int     child, child_found, pair, cnr = XPLAB_SINGLE_CLASS;
  int_lst child_lst;

  init_int_lst(&child_lst);

  assert(pid >= XPLAB_INIT_HASH_ID);            /* positive parent hash ID */

  for (pair = 0; pair < hash.cnt; pair++) {
    if (hash.pid[pair] == pid) {                        /* parent ID found */
      assert(hash.cid[pair] >= XPLAB_INIT_HASH_ID);/*positive child hash ID */
      assert(hash.kr[pair].n >= 0);                   /* consistency check */
      assert(hash.kr[pair].c >= 0); /* indexes of parent list (trained NB) */

      child_found = -1;
      for (child = 0; child < traced_children.cls[cnr].nbl.cnt; child++) {
	if (hash.cid[pair] == traced_children.cls[cnr].nbl.nb[child].aux) {
	  child_found = child;        /* id of child (traced NB) is unique */
	  break;           
	}
      }

      if (child_found >= 0) {
	if (child_lst.cnt++) {
	  REALLOC(int, child_lst.cnt, child_lst.nr);
	}
	else {
	  CALLOC(int, child_lst.cnt, child_lst.nr);
	}

	child_lst.nr[child_lst.cnt-1] = child_found;
      }
    }
  }

  return child_lst;
}



/* Marks sample image by a candidate list.

Revision: 26.08.2005 E.Sack

Input : - candidate list
        - class list
        - sample image 
Output: - image updated
Precondition: - sample image and NB auxiliary references must be consistent.
*/
static void
color_cand_lst(cand_lst cndl, cls_lst clst, pic_def simg)
{
  int        cand, cnr, nnr;
  sector_def roi;

  for (cand = 0; cand < cndl.cnt; cand++) {
    cnr = cndl.c[cand].c;
    nnr = cndl.c[cand].n;

    if (nnr < 0) {                             /* only recognized candidates */
      continue;
    }

    if (simg.mem) {                            /* coloring patterns in image */
      AREA_TO_SECTOR(clst.cls[cnr].nbl.nb[nnr].pat.ar, roi);

      color_reference(clst.cls[cnr].nbl.nb[nnr].pat, roi, simg,
		      clst.cls[cnr].nbl.nb[nnr].dat.afs.x, 
		      clst.cls[cnr].nbl.nb[nnr].dat.afs.y, XPLAB_TRAINED_PIXEL);
    }
  }
}



/* Initializes all patterns as selected.

Revision: 28.08.2003 E.Sack

Input : - class list
        - image 
        - mark
Output: - image updated
*/
static int_lst
setup_selected_pattern(int cnt)
{
  int_lst lbls;

  assert(cnt >= 0);

  init_int_lst(&lbls);

  if (!cnt) {
    return lbls;
  }

  lbls.cnt = cnt;
  CALLOC(int, lbls.cnt, lbls.nr);

  set_int_lst(lbls, XPLAB_SELECTED_PIXEL, -1);

  return lbls;
}



/* Marks an image by an indexed pattern list.

Revision: 26.08.2005 E.Sack

Input : - class list
        - index list
        - source image 
        - sample image 
Output: - images updated
*/
static void
color_indexed_cls_lst(cls_lst clst, int_lst idx, pic_def mimg, pic_def simg)
{
  int cnr = XPLAB_SINGLE_CLASS, nnr;

  if (!clst.cls) {                                    /* consistency check */
    assert(!clst.cnt);
    return;                                            /* class list empty */
  }
  else {
    assert(clst.cnt > 0);
    assert(idx.cnt == clst.cls[cnr].nbl.cnt);
  }

  for (nnr = 0; nnr < clst.cls[cnr].nbl.cnt; nnr++) {
    color_nb_def(clst.cls[cnr].nbl.nb[nnr], mimg, simg, idx.nr[nnr]);
  }
}



/* Marks referenced images by a NB.

Revision: 26.08.2005 E.Sack

Input : - NB
        - source image 
        - sample image 
        - mark
Output: - image sector
        - image updated
*/
static sector_def
color_nb_def(nb_def nb, pic_def mimg, pic_def simg, int to)
{
  sector_def roi, msec;

  AREA_TO_SECTOR(nb.pat.ar, roi);

  /* coloring patterns in source image                                       */
  msec = color_reference(nb.pat, roi, mimg, nb.pat.org.x, nb.pat.org.y, to);

  if (simg.mem) {                       /* coloring patterns in sample image */
    color_reference(nb.pat, roi, simg, nb.dat.afs.x, nb.dat.afs.y, to);
  }

  return msec;
}



/* Marks image foreground pixels by reference foreground pixels.

Revision: 18.08.2005 E.Sack

Input : - binary image reference
        - sector of image
        - image
        - lag of sector
        - mark
Output: - updated image
        - sector of update
Precondition: - if mark is negative, image is left untouched.
              - the image and lags must be consistent
Usage : - The binary image (normally a NB) is compared against an image. Hence
          the reference should have reasonable offsets to the image.
        - The sector on the image compared against the reference is returned.
          It may be different (smaller) than the reference size.
*/
static sector_def
color_reference(pic_def bref, sector_def rbs, pic_def cimg, int lx, int ly,
		int to)
{
  int        sx, sy, syy, sxmin, symin, tx, ty, tyy;
  int        xstart, xend, ystart, yend;
  sector_def sec;

  xstart = XPMAX(0, rbs.column.start+lx);
  ystart = XPMAX(0, rbs.line.start+ly);
  xend = XPMIN(cimg.ar.width-1, rbs.column.start+lx+bref.ar.width-1);
  yend = XPMIN(cimg.ar.height-1, rbs.line.start+ly+bref.ar.height-1);

  lx = rbs.column.start+lx-xstart;                           /* correct lags */
  ly = rbs.line.start+ly-ystart;

  sec.column.start = XPMAX(0, lx) + xstart;
  sec.column.end = XPMIN(xend, lx+bref.ar.width-1 + xstart);
  sec.line.start = XPMAX(0, ly) + ystart;
  sec.line.end = XPMIN(yend, ly+bref.ar.height-1 + ystart);

  if (to >= 0 ) {                                              /* comparison */
    sxmin = XPMAX(0,-lx);
    symin = XPMAX(0,-ly);

    for (ty = sec.line.start, sy = symin; ty <= sec.line.end; ty++, sy++) {
      tyy = ty*cimg.ar.width;
      syy = sy*bref.ar.width;
      sx = sxmin;

      for (tx = sec.column.start; tx <= sec.column.end; tx++, sx++) {
	if (bref.mem[sx + syy] != BACKGROUND && /*foreground reference pixel */
	    cimg.mem[tx + tyy] != BACKGROUND) {    /* foreground image pixel */
	  cimg.mem[tx + tyy] = to;                     /* both pixels marked */
	}
      }
    }
  }

  return sec;                                 /* image sector for comparison */
}



/* Erases all information of children/parents from the class.

Revision: 08.08.2003 E.Sack

Input : - class
Output: - class updated
*/
static void
orphan_cls_def(cls_def ocls)
{
  int nidx;

  for (nidx = 0; nidx < ocls.nbl.cnt; nidx++) {
    free_nb_chn_lst(&ocls.nbl.nb[nidx].cnb);
    free_nb_chn_lst(&ocls.nbl.nb[nidx].nnb);
  }
}



/* Stops pattern tracing.

Revision: 08.08.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_trace_stop_button(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  stop_run_def(&tr.trun);

  trace_sensitiveness();
}



/* Stops pattern correlation.

Revision: 21.09.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_stop_button(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  stop_run_def(&tr.srun);
}



/* Sets the column in the view tree to be searched.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_search_column(GTK_CONVENTION_COMBO_CHANGED_RTS_VOID)
{
  gchar *buf;
  GtkTreeView *tview = (GtkTreeView *)user_data;

  assert(combo_box);
  assert(user_data);
  assert(GTK_IS_TREE_VIEW(tview) == TRUE);
  assert(GTK_IS_COMBO_BOX(combo_box) == TRUE);

  buf = gtk_combo_box_get_active_text(GTK_COMBO_BOX(combo_box));

  assert(buf);

  if (strstr(buf, XPLAB_PAGE_SELECT_VIEW_TITLE_CLS) != NULL) {
    gtk_tree_view_set_search_column(GTK_TREE_VIEW(tview), 
				    XPLAB_PAGE_SELECT_VIEW_COLUMN_CLS);
  }
  else if (strstr(buf, XPLAB_PAGE_SELECT_VIEW_TITLE_TRS) != NULL) {
    gtk_tree_view_set_search_column(GTK_TREE_VIEW(tview), 
				    XPLAB_PAGE_SELECT_VIEW_COLUMN_TRS);
  }
  else if (strstr(buf, XPLAB_PAGE_SELECT_VIEW_TITLE_CMT) != NULL) {
    gtk_tree_view_set_search_column(GTK_TREE_VIEW(tview), 
				    XPLAB_PAGE_SELECT_VIEW_COLUMN_CMT);
  }
  else {
    assert(0);
  }
}



/* Correlates all NBs for the selected pattern in image to find the best
   fitted class.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_find_proposal(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  int               cnr_max, pattern_cnt, n4;
  GtkTreeViewColumn *vcol;

  STSFY_GCC(user_data);

  assert(button);

  xplab_label_set_text(tr.page.select.find_info, "");
  start_run_def(&tr.srun);

  n4 = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nbhood)) == TRUE ?
    1:0;

  /* check for pattern in image to correlate                               */
  pattern_cnt = xcount_contours(tr.session.mimg.pimg, 
				tr.page.select.focus_sector, 
				XPLAB_SELECTED_PIXEL, n4);

  if (pattern_cnt == 0) {
    xplab_label_set_text(tr.page.select.find_info, XPLAB_PATTERN_CORR_ERROR);
    return;                                   /* no selected pattern found */
  }

  gtk_button_leave(GTK_BUTTON(GTK_OBJECT(button)));/*just for display taste*/

  /* set button from Purpose to Stop */
  gtk_button_set_label(button, "S");

  /* block purpose handler to avoid multiple activations */
  g_signal_handler_block((gpointer)button, tr.page.select.purpose_id);

  /* connect to stop handler to enable interrupt long runs */
  tr.page.select.stop_id = 
    g_signal_connect(G_OBJECT(button), "clicked", 
		     G_CALLBACK(cb_select_stop_button), NULL);

  /* discard stop signal from activating immediately */
  g_signal_stop_emission_by_name((gpointer)button, "clicked");

  /* setup widgets                                                         */
  gtk_widget_hide(tr.page.select.find_info);                 /* hide label */
  gtk_widget_show(tr.srun.pbar);                      /* show progress bar */

  vcol = gtk_tree_view_get_column(tr.list_old.tview, XPLAB_PAGE_SELECT_VIEW_COLUMN_CVL);
  gtk_tree_view_column_set_visible(vcol, TRUE);

  cnr_max = xproposal(tr.list_old);            /* calculate class proposal */

  if (tr.srun.pbar != NULL) {                      /* check user interrupt */
    /* set button from Stop to Purpose */
    gtk_button_set_label(button, "P");

    /* disconnect Stop handler */
    g_signal_handler_disconnect(button, tr.page.select.stop_id);

    /* discard signal */
    g_signal_stop_emission_by_name((gpointer)button, "clicked");

    gtk_widget_hide(tr.srun.pbar);                    /* hide progress bar */
    gtk_widget_show(tr.page.select.find_info);               /* show label */

    if (cnr_max >= 0) {
      treeview_moveto(tr.list_old.tview, cnr_max, FALSE);
    }
    else {
      xplab_label_set_text(tr.page.select.find_info, "database empty");
      gtk_tree_view_column_set_visible(vcol, TRUE);
    }
  }

  /* unblock purpose handler */
  g_signal_handler_unblock((gpointer)button, tr.page.select.purpose_id);
}



/* Proposes a class for selected patterns by correlation with the trained DB.

Revision: 06.08.2005 E.Sack

Input : - pointer to training settings
Output: - training settings updated
        - proposed class index (-1: no proposal found)
Usage:  - this function can generate extensive calculations and consequently 
          retarded event handling.
*/
static int
xproposal(list_def plist)
{
  int          mark_found, nnr, cnr, cnr_max = -1;
  double       cv, cv_nb_min, cv_nb_max, cv_max = -1.0;
  sector_def   msec;
  pic_def      inew;
  gboolean     found;
  GtkTreeModel *tstore, *smodel;
  GtkTreePath  *child_path;

  init_pic_def(&inew);

  smodel = gtk_tree_view_get_model(plist.tview);
  assert(GTK_IS_TREE_MODEL_SORT(smodel) == TRUE);
  tstore = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(smodel));

  SECTOR_TO_AREA(tr.page.select.focus_sector, inew.ar);
  CALLOC(PICT, AREA_SIZE(inew.ar), inew.mem);

  copy_bmps(tr.session.mimg.pimg, inew, inew.ar, /* copy pattern temporarily */
	    tr.page.select.focus_sector.column.start, 
	    tr.page.select.focus_sector.line.start, 0, 0);

  AREA_TO_SECTOR(inew.ar, msec);     /* separate & prepare pattern for corr. */
  set_pixel_masked(msec, ~XPLAB_SELECTED_PIXEL, BACKGROUND, inew, FALSE);
  mark_found = set_pixel_masked(msec, XPLAB_SELECTED_PIXEL, FOREGROUND, inew, FALSE);
  assert(mark_found);

  mark_found = xcopy_and_fit_pic(inew, &inew);   /* important for used corr. */
  assert(mark_found);

  xplab_progress_configure(&tr.srun, NULL);
  xplab_progress_set_value(&tr.srun, (gdouble)1.0, (gdouble)plist.cls->cnt);

  /* correlate pattern with DB                                               */
  for (cnr = 0; cnr < plist.cls->cnt && *tr.srun.pactive==TRUE; cnr++) {
    cv_nb_min = 2.0;
    cv_nb_max = -1.0;

    for (nnr = 0; 
	 nnr < plist.cls->cls[cnr].nbl.cnt && *tr.srun.pactive==TRUE; 
	 nnr++) {
      xplab_progress_set_value(&tr.srun, (gdouble)(cnr+1), (gdouble)plist.cls->cnt);

      cv = full_correlation1(plist.cls->cls[cnr].nbl.nb[nnr].pat,
			     inew, &tr.srun).cv;   /* input order irrelevant */

      if (cv_max < cv) {
	cv_max = cv;
	cnr_max = cnr;
      }

      cv_nb_min = XPMIN(cv_nb_min, cv);
      cv_nb_max = XPMAX(cv_nb_max, cv);
    }

    if (tr.srun.pbar != NULL) {                      /* check user interrupt */
      gchar       *buf = NULL;
      GtkTreeIter child_iter;

      double_to_string("", cv_nb_max, XPLAB_PAGE_SELECT_CV_SEP, &buf);
      double_to_string("", cv_nb_min, "", &buf);       /* correlation values */

      child_path = gtk_tree_path_new_from_indices(cnr, -1);
      found = gtk_tree_model_get_iter(GTK_TREE_MODEL(tstore), 
				      &child_iter, child_path);
      assert(found == TRUE);
      gtk_tree_path_free(child_path);

      gtk_tree_store_set(GTK_TREE_STORE(tstore), &child_iter,
			 XPLAB_DB_TREE_STORE_COLUMN_CVL, buf, -1);

      FREEMEM(buf);
    }
  }

  free_pic_def(&inew);

  return cnr_max;
}



/* Finds traceable NBs in TPS.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_find_trace_button(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID)
{
  int  d, traceable, cidx;
  char *txt = NULL;

  STSFY_GCC(user_data);

  assert(togglebutton);

  assert(GTK_IS_WIDGET(tr.page.select.parent) == TRUE);  /* list container */

  /* trigger for updating display by GtkTreeCellDataFunc */
  gdk_window_invalidate_rect(GTK_WIDGET(tr.page.select.parent)->window,
			     NULL, TRUE);

  gdk_window_process_updates(GTK_WIDGET(tr.page.select.parent)->window, TRUE);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.trace_button)) == TRUE) {
    for (cidx = traceable = 0; cidx < tr.list_old.cls->cnt; cidx++) {
      d = class_traceable(tr.list_old.cls->cls[cidx]);

      if (d) {                                           /* class traceable */
	traceable++;
      }
    }

    int_to_string("traceable: ", traceable, "", &txt);
    xplab_label_set_text(tr.page.select.find_info, txt);
  }
  else {
    xplab_label_set_text(tr.page.select.find_info, "");
  }

  FREEMEM(txt);
}



/* Copies trained/untrained row data to attributes mask (mouse button
   double click).

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_row_select(GTK_CONVENTION_TREEVIEW_CURSOR_CHANGED_RTS_VOID)
{
  GtkTreePath       *path;
  GtkTreeViewColumn *fcolumn;

  STSFY_GCC(user_data);

  assert(treeview);

  gtk_tree_view_get_cursor(treeview, &path, &fcolumn);

  if (path) {
    copy_row_to_attributes(treeview, path);
    gtk_tree_path_free(path);
  }
}



/* Selects traced pattern.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_trace_row_select(GTK_CONVENTION_TREEVIEW_CURSOR_CHANGED_RTS_VOID)
{
  GtkTreePath       *path;
  GtkTreeViewColumn *fcolumn;

  STSFY_GCC(user_data);

  assert(treeview);

  gtk_tree_view_get_cursor(treeview, &path, &fcolumn);

  if (path) {
    g_signal_handler_block((gpointer)treeview, tr.page.trace.cursor_changed_id);

    tr.page.trace.tdata.chno = get_candidate_index(treeview, path);
    gtk_tree_path_free(path);

    assert(tr.page.trace.tdata.chno >= 0 && 
	   tr.page.trace.tdata.chno < tr.page.trace.tdata.lbls.cnt);

    select_child(&tr.page.trace.tdata, tr.page.trace.tview);

    pattern_drawing(tr.page.trace.tdata.trained.spattn, tr.timg.pcol, 0);
    pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);

    g_signal_handler_unblock((gpointer)treeview, tr.page.trace.cursor_changed_id);
  }
}



/* Copies tree view data to attributes mask.

Revision: 29.07.2005 E.Sack

Input : - pointer to tree view
        - pointer to path
        - row index
Output: - attributes mask updated
*/
static void
copy_row_to_attributes(GtkTreeView *treeview, GtkTreePath *path)
{
  gchar        *row_text = NULL;
  gboolean     row_found;
  GtkEntry     *centry;
  GtkTreeIter  iter;
  GtkTreeModel *tmodel;

  assert(treeview);
  assert(path);
  assert(GTK_IS_TREE_VIEW(treeview) == TRUE);

  tmodel = gtk_tree_view_get_model(GTK_TREE_VIEW(treeview));
  assert(GTK_IS_TREE_MODEL(tmodel) == TRUE);
  assert(XPLAB_DB_TREE_STORE_COLUMNS == gtk_tree_model_get_n_columns(GTK_TREE_MODEL(tmodel)));

  row_found = gtk_tree_model_get_iter(tmodel, &iter, path);
  assert(row_found == TRUE);                            /* class must be in */

  /* get class */
  gtk_tree_model_get(tmodel, &iter, XPLAB_DB_TREE_STORE_COLUMN_CLS, 
		     &row_text, -1);
  assert(row_text);

  /* set class */
  assert(GTK_IS_COMBO_BOX(tr.page.select.satt.class.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(tr.page.select.satt.class.combo))->child);
  xplab_entry_set_text(centry, row_text);
  g_free(row_text);

  /* get translation */
  gtk_tree_model_get(tmodel, &iter, XPLAB_DB_TREE_STORE_COLUMN_TRS,
		     &row_text, -1);
  assert(row_text);

  /* set translation */
  assert(GTK_IS_COMBO_BOX(tr.page.select.satt.translatn.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(tr.page.select.satt.translatn.combo))->child);
  xplab_entry_set_text(centry, row_text);
  g_free(row_text);

  /* get comment */
  gtk_tree_model_get(tmodel, &iter, XPLAB_DB_TREE_STORE_COLUMN_CMT, 
		     &row_text, -1);
  assert(row_text);

  /* set comment */
  assert(GTK_IS_COMBO_BOX(tr.page.select.satt.commnt.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(tr.page.select.satt.commnt.combo))->child);
  xplab_entry_set_text(centry, row_text);
  g_free(row_text);

  /* get alignment */
  gtk_tree_model_get(tmodel, &iter, XPLAB_DB_TREE_STORE_COLUMN_ALT, 
		     &row_text, -1);
  assert(row_text);

  /* set alignment */
  assert(GTK_IS_COMBO_BOX(tr.page.select.satt.alignment.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(tr.page.select.satt.alignment.combo))->child);
  xplab_entry_set_text(centry, row_text);
  g_free(row_text);

  /* get printable */
  gtk_tree_model_get(tmodel, &iter, XPLAB_DB_TREE_STORE_COLUMN_PRT, 
		     &row_text, -1);
  assert(row_text);

  /* set printable */
  assert(GTK_IS_COMBO_BOX(tr.page.select.satt.printable.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(tr.page.select.satt.printable.combo))->child);
  xplab_entry_set_text(centry, row_text);
  g_free(row_text);

  /* get fixed */
  gtk_tree_model_get(tmodel, &iter, XPLAB_DB_TREE_STORE_COLUMN_FIX, 
		     &row_text, -1);
  assert(row_text);

  /* set fixed */
  assert(GTK_IS_COMBO_BOX(tr.page.select.satt.fixed.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(tr.page.select.satt.fixed.combo))->child);
  xplab_entry_set_text(centry, row_text);
  g_free(row_text);
}



/* Checks wether a class is traceable.
   Rule to extend class: - NB is not a CNB (has no children),
                         - NB is printable (not a incomplete fragment).

Revision: 29.07.2005 E.Sack

Input : - class
Output: - flag (0: not traceable, 1: traceable)
*/
static int
class_traceable(cls_def cls)
{
  int children;

  children = class_has_children(cls.nbl);

  return !children && cls.att.printable;
}



/* Checks wether a class is a CNB (has children). A class is meant a CNB
   if at least one of its members is a CNB.

Revision: 06.08.2005 E.Sack

Input : - class
Output: - number of children
*/
static gint
class_has_children(nb_lst nbl)
{
  int nidx, children;

  for (nidx = children = 0; nidx < nbl.cnt; nidx++) {
    children += nbl.nb[nidx].nnb.cnt;             /* CNB has children as NNB */
  }

  return children;                        /* return total amount of children */
}



/* Compares class attributes against the class attributes dialog content.

Revision: 29.07.2005 E.Sack

Input : - class
        - attribute entrys
Output: - flag (0: not identic, 1: identic)
*/
static int
attributes_identic(cls_def cls, cls_att_entry att)
{
  gint      flag;
  gchar    *buf;
  ALIGMNT  alg;
  GtkEntry *centry;

  assert(GTK_IS_COMBO_BOX(att.class.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.class.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  flag = check_name(buf);

  if (flag) {
    FREEMEM(buf);
    return 0;
  }

  if (strcmp(buf, cls.att.class)) {
    FREEMEM(buf);
    return 0;
  }

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.translatn.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.translatn.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  flag = check_name(buf);

  if (flag) {
    FREEMEM(buf);
    return 0;
  }

  if (strcmp(buf, cls.att.sign)) {
    FREEMEM(buf);
    return 0;
  }

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.commnt.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.commnt.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  flag = check_name(buf);

  if (flag) {
    FREEMEM(buf);
    return 0;
  }

  if ((!cls.att.crem && buf) || (cls.att.crem && !buf)) {
    FREEMEM(buf);
    return 0;
  }

  if (cls.att.crem && buf) {
    if (strcmp(buf, cls.att.crem)) {
      FREEMEM(buf);
      return 0;
    }
  }

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.alignment.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.alignment.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  flag = check_name(buf);

  if (flag) {
    FREEMEM(buf);
    return 0;
  }

  flag = string_to_aligmnt(buf, &alg);

  if (!flag || alg != cls.att.alignment) {
    FREEMEM(buf);
    return 0;
  }

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.printable.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.printable.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  flag = check_name(buf);

  if (flag) {
    FREEMEM(buf);
    return 0;
  }

  if (!strcmp(buf, "TRUE")) {
    flag = 1;
  }
  else if (!strcmp(buf, "FALSE")) {
    flag = 0;
  }
  else {
    FREEMEM(buf);
    return 0;
  }

  if (flag != cls.att.printable) {
    FREEMEM(buf);
    return 0;
  }

  FREEMEM(buf);

  assert(GTK_IS_COMBO_BOX(att.fixed.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(att.fixed.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  flag = check_name(buf);

  if (flag) {
    FREEMEM(buf);
    return 0;
  }

  if (!strcmp(buf, "TRUE")) {
    flag = 1;
  }
  else if (!strcmp(buf, "FALSE")) {
    flag = 0;
  }
  else {
    FREEMEM(buf);
    return 0;
  }

  FREEMEM(buf);

  if (flag != cls.att.fixed) {
    return 0;
  }

  return 1;
}



/* Sums the trained NB items.

Revision: 20.07.2003 E.Sack

Input : - class list
Output: - number of trained NB
*/
static int
get_parse_status(cls_lst cls_old)
{
  int cidx, nidx, parse_status = 0;

  for (cidx = 0; cidx < cls_old.cnt; cidx++) {
    for (nidx = 0; nidx < cls_old.cls[cidx].nbl.cnt; nidx++) {
      if ((PARSE_STAT)cls_old.cls[cidx].nbl.nb[nidx].aux & (XPLAB_PARSE_STAT_EXT|XPLAB_PARSE_STAT_NEW)) {
	parse_status++;
      }
    }
  }

  return parse_status;
}



/* Enrolls the parse status to the trained NB.

Revision: 17.07.2003 E.Sack

Input : - class list
        - parse status
Output: - NB updated
Usage : Status ist used for coloring tree view.
*/
static void
set_parse_status(cls_lst cls_new, PARSE_STAT parse_status)
{
  int cidx, nidx;

  assert(parse_status > 0);

  for (cidx = 0; cidx < cls_new.cnt; cidx++) {
    for (nidx = 0; nidx < cls_new.cls[cidx].nbl.cnt; nidx++) {
      if ((PARSE_STAT)cls_new.cls[cidx].nbl.nb[nidx].aux & (XPLAB_PARSE_STAT_EXT|XPLAB_PARSE_STAT_NEW)) {
	continue;                      /* already set in an earlier stage */
      }

      cls_new.cls[cidx].nbl.nb[nidx].aux = (int)parse_status;
    }
  }
}



/* Resets the parse status.

Revision: 17.07.2003 E.Sack

Input : - pointer to training settings
Output: - NB updated
*/
static void
reset_parse_status(cls_lst cls)
{
  int cidx, nidx;

  for (cidx = 0; cidx < cls.cnt; cidx++) {
    for (nidx = 0; nidx < cls.cls[cidx].nbl.cnt; nidx++) {
      cls.cls[cidx].nbl.nb[nidx].aux = (int)XPLAB_PARSE_STAT_DEF;
    }
  }
}



/* Callback function for parsing.

Revision: 13.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_parse_train(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  PAGE_TABS   current_page;
  const gchar *label_text;
  gint        is_train;

  STSFY_GCC(user_data);

  assert(button);

  current_page = xplab_get_page_enum(GTK_NOTEBOOK(tr.page.notebook), -1);

  label_text = gtk_button_get_label(button);
  is_train = strcmp(label_text, XPLAB_TRAINING_DIALOG_TRAIN_LABEL) == 0;

  switch (current_page) {
  case XPLAB_TRAINING_PAGE_IS_SELECT:

    is_train ? select_train(button) : select_parse(button);

    break;

  case XPLAB_TRAINING_PAGE_IS_TRACE:

    is_train ? trace_train(button) : trace_parse(button);

    break;

  case XPLAB_TAB_UNDEF:
    break;
  }
}



/* Checks the user training values of the trace page format.

Revision: 13.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
trace_parse(GtkButton *button)
{
  int att_changes, pattern_warning;

  tr.page.trace.tdata.parse_status = XPLAB_PARSE_STAT_DEF;

  pattern_warning = trace_parse_pattern();

  if (pattern_warning == -1) {
    return;                                       /* parse faults detected */
  }

  /* attributes parsing                                                    */
  att_changes = parse_class_attributes_to_train(&tr.page.select, tr.list_old,
						tr.list_new);

  if (att_changes == -1) {
    return;                                       /* parse faults detected */
  }

  tr.page.trace.tdata.parse_status = XPLAB_PARSE_STAT_EXT;

  /* set button to train */
  gtk_button_set_label(button, XPLAB_TRAINING_DIALOG_TRAIN_LABEL);
}



/* Checks the user training values of the select page format.

Revision: 13.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
select_parse(GtkButton *button)
{
  int att_changes, pattern_warning;

  tr.page.select.satt.parse_status = XPLAB_PARSE_STAT_DEF;

  pattern_warning = select_parse_pattern();             /* pattern parsing */

  if (pattern_warning == -1) {
    return;                                       /* parse faults detected */
  }

  /* attributes parsing                                                    */
  att_changes = parse_class_attributes_to_train(&tr.page.select, tr.list_old,
						tr.list_new);

  if (att_changes == -1) {
    return;                                       /* parse faults detected */
  }

  switch (att_changes) {
  case 1:
    xplab_label_set_text(tr.page.select.satt.infoline, XPLAB_ATTRIBUTES_PARSING_SUCCESS_CHANGED);
    break;

  case 2:
    break;

  default:
    if (pattern_warning) {
      xplab_label_set_text(tr.page.select.satt.infoline, XPLAB_FRACT_PARSING_WARNING);
    }
    else {
      xplab_label_set_text(tr.page.select.satt.infoline, XPLAB_ATTRIBUTES_PARSING_SUCCESS);
    }
    break;
  }

  /* set button to train */
  gtk_button_set_label(button, XPLAB_TRAINING_DIALOG_TRAIN_LABEL);

  update_history_cls_att_entry(&tr.page.select.satt);
}



/* Parses patterns used by trace page format.

Revision: 29.07.2005 E.Sack

Input : - pointer to training settings
Output: - flag (0: no parse faults, -1: parse faults, 1: parse corrections)
*/
static int
trace_parse_pattern(void)
{
  int      cnr = XPLAB_SINGLE_CLASS, pattern_warning = 0, identic = 0;
  int      trainable = 0, nnr, cid, d, broken = 0;
  char     *buf = NULL;
  cand_def cand;

  if (!tr.page.trace.tdata.lbls.cnt) {
    xplab_label_set_text(tr.trun.plabel, TRAINING_WARNING_13);
    return -1;
  }

  /* pattern parsing rules                                                 */
  for (nnr = 0; nnr < tr.page.trace.tdata.lbls.cnt; nnr++) {
    if ((tr.page.trace.tdata.lbls.nr[nnr] != XPLAB_SELECTED_PIXEL) &&
	(tr.page.trace.tdata.lbls.nr[nnr] != XPLAB_IGNORED_PIXEL)) {
      continue;
    }

    trainable++;

    cid = tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr].aux;
    cand = search_candidate_of_child_id(tr.page.trace.tdata.thash, cid);

    if (cand.m.cv == 1.0) {          /* candidate identic to its reference */
      tr.page.trace.tdata.lbls.nr[nnr] = XPLAB_IGNORED_PIXEL;
      color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr], 
		   tr.session.mimg.pimg, tr.page.trace.tdata.traced.spattn.sample,
		   XPLAB_IGNORED_PIXEL);
      identic++;
      continue;
    }

    d = search_nb(tr.page.trace.tdata.traced.sclass.cls[cnr].att.class, 
		  tr.page.trace.tdata.traced.sclass.cls[cnr].att.alignment, 
		  tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr],
		  *tr.list_old.cls);              /* search in source list */

    if (d) {                        /* candidate identic to a class member */
      tr.page.trace.tdata.lbls.nr[nnr] = XPLAB_IGNORED_PIXEL;
      color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr], 
		   tr.session.mimg.pimg, 
		   tr.page.trace.tdata.traced.spattn.sample,
		   XPLAB_IGNORED_PIXEL);
      identic++;
      continue;
    }

    d = search_identic_nb(tr.page.trace.tdata.traced.sclass.cls[cnr].att.class, 
			  tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr], 
			  *tr.list_old.cls);      /* search in source list */

    if (d) {           /* never take over identic pattern of another class */
      tr.page.trace.tdata.lbls.nr[nnr] = XPLAB_IGNORED_PIXEL;
      color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr], 
		   tr.session.mimg.pimg, tr.page.trace.tdata.traced.spattn.sample,
		   XPLAB_IGNORED_PIXEL);
      identic++;
      continue;
    }

    d = check_part_count(tr.page.trace.tdata.trained.sclass.cls[cand.c].nbl.nb[cand.n].pat,
			 tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr].pat);

    if (d > 0) {      /* never take over more fractions than reference has */
      tr.page.trace.tdata.lbls.nr[nnr] = XPLAB_SELECTBL_PIXEL;
      color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr], 
		   tr.session.mimg.pimg, tr.page.trace.tdata.traced.spattn.sample,
		   XPLAB_SELECTBL_PIXEL);
      broken++;
      trainable--;                                        /* not trainable */
      continue;
    }
  }

  if (!trainable) {
    xplab_label_set_text(tr.trun.plabel, TRAINING_WARNING_14);
    return -1;
  }

  if (identic) {
    int_to_string("", identic, " identic pattern", &buf);
    add_string(identic==1?"":"s", &buf);
    add_string(" in class list", &buf);

    pattern_warning = 1;
  }

  if (broken) {
    int_to_string("", broken, " broken pattern", &buf);
    add_string(broken==1?"":"s", &buf);

    pattern_warning = 1;
  }

  if (buf != NULL) {
    xplab_label_set_text(tr.trun.plabel, buf);
    FREEMEM(buf);
  }

  pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);
  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);

  return pattern_warning;
}



/* Parses patterns used by select page format.

Revision: 13.09.2004 E.Sack

Input : - pointer to training settings
Output: - flag (0: no parse faults, -1: parse faults, 1: parse corrections)
*/
static int
select_parse_pattern(void)
{
  int pattern_warning = 0, pattern_cnt, n4;

  n4 = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nbhood)) == TRUE ?
    1:0;

  pattern_cnt = xcount_contours(tr.session.mimg.pimg, 
				tr.page.select.focus_sector, 
				XPLAB_SELECTED_PIXEL, n4);

  /* validate pattern in focus */
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nb_button)) == TRUE ||
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nnb_button)) == TRUE) {
    if (pattern_cnt == 0) {
      xplab_label_set_text(tr.page.select.satt.infoline, XPLAB_PATTERN_PARSING_ERROR);
      return -1;
    }

    if (pattern_cnt > 1) {
      pattern_warning = 1;
    }
  }
  else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.cnb_button)) == TRUE) {
    if (pattern_cnt < 2) {
      xplab_label_set_text(tr.page.select.satt.infoline, XPLAB_CNB_PARSING_ERROR);
      return -1;
    }
  }
  else {
    assert(0);                                       /* check button fault */
  }

  return pattern_warning;
}



/* Parses attributes for a new or an existing class. A new class will be
   created if there is no other class with the same name and all the
   attributes are valid. If the class already exists, all attributes have
   to be identic (unless its pattern). In all other cases the input will
   be rejected.

Revision: 29.07.2005 E.Sack

Input : - pointer to training settings
Output: - flag (0: no faults, no corrections, -1: faults, 1: corrections,
                2: warnings printed)
*/
static int
parse_class_attributes_to_train(select_def *pselect, list_def list_old,
				list_def list_new)
{
  int      err, flag, att_changes = 0, cidx;
  int      identic_class;
  gchar    *class_name;          
  GtkEntry *centry;

  /* first check class name                                                */
  flag = xplab_check_entry(pselect->satt.class, MAX_LEN_CLASS_NAME);

  if (flag == XPLAB_STRING_NOT_USABLE) {
    xplab_label_set_text(pselect->satt.infoline, XPLAB_CLASS_PARSING_ERROR);
    return -1;
  }
  else if (flag == XPLAB_STRING_WITH_CHANGES) {
    att_changes = 1;
  }

  assert(GTK_IS_COMBO_BOX(pselect->satt.class.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(pselect->satt.class.combo))->child);
  class_name = xplab_entry_get_text(centry);                 /* fetch string */

  err = check_name(class_name);
  assert(!err);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pselect->nnb_button)) == TRUE ||
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pselect->cnb_button)) == TRUE) {
    /* check class name against new NNB list */
    flag = search_class(class_name, *list_new.cls);/* search class in NNB list */

    if (flag >= 0) {
      xplab_label_set_text(pselect->satt.infoline, XPLAB_NNB_PARSING_ERROR);
      FREEMEM(class_name);
      return -1;
    }
  }

  cidx = search_class(class_name, *list_old.cls);/* search class in trained list */

  FREEMEM(class_name);

  if (cidx >= 0) {                                            /* class found */
    identic_class = attributes_identic(list_old.cls->cls[cidx], pselect->satt);

    if (!identic_class) {    /* identic class name with different attributes */
      treeview_moveto(list_old.tview, cidx, FALSE);

      xplab_label_set_text(pselect->satt.infoline, XPLAB_CLASS_NAME_PARSING_ERROR);
      return -1;
    }

    pselect->satt.parse_status = XPLAB_PARSE_STAT_EXT;/* extend trained class */
  }
  else {
    pselect->satt.parse_status = XPLAB_PARSE_STAT_NEW;
  }

  att_changes =                                          /* check the rest */
    check_class_attributes(pselect->satt,
			   gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nb_button)),
			   gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.cnb_button)));

  return att_changes;
}



/* Parses class attributes except class name.

Revision: 29.07.2005 E.Sack

Input : - attribute settings
        - flag (TRUE: attributes are for a NB,
                FALSE: attributes are not for a NB)
        - flag (TRUE: attributes are for a CNB,
                FALSE: attributes are not for a CNB)
Output: - flag (0: no faults, no corrections, -1: faults, 1: corrections,
                2: warnings printed)
*/
static int
check_class_attributes(cls_att_entry catt, gboolean is_a_nb, 
		       gboolean is_a_cnb)
{
  int      flag, changes = 0;
  gchar    *buf = NULL;          
  GtkEntry *centry;

  /* check translation entry                                               */
  flag = xplab_check_entry(catt.translatn, MAX_LEN_CLASS_TRANS);

  if (flag == XPLAB_STRING_NOT_USABLE) {
    xplab_label_set_text(catt.infoline, XPLAB_CLASS_TRANSLATION_PARSING_ERROR);
    return -1;
  }
  else if (flag == XPLAB_STRING_WITH_CHANGES) {
    changes = 1;
  }

  /* check comment entry                                                   */
  flag = xplab_check_entry(catt.commnt, MAX_LEN_CLASS_COMMENT);

  if (flag == XPLAB_STRING_NOT_USABLE) {
    xplab_label_set_text(catt.infoline, XPLAB_CLASS_COMMENT_PARSING_ERROR);
    return -1;
  }
  else if (flag == XPLAB_STRING_WITH_CHANGES) {
    changes = 1;
  }

  /* check alignment entry                                                 */
  flag = xplab_check_alignment(catt.alignment);

  if (!flag) {
    return -1;
  }

  /* check printable entry                                                 */
  assert(GTK_IS_COMBO_BOX(catt.printable.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(catt.printable.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  assert(buf);

  if (!strlen(buf) || (strcmp(buf, "TRUE") && strcmp(buf, "FALSE"))) {
    xplab_label_set_text(catt.infoline, XPLAB_CLASS_PRINTABLE_PARSING_ERROR);
    FREEMEM(buf);
    return -1;
  }
  else if (strcmp(buf, "TRUE")) {                         /* not printable */
    if (is_a_cnb == TRUE) {                                /* has priority */
      xplab_label_set_text(catt.infoline, XPLAB_CLASS_PRINTABLE_CNB_WARNING);
      changes = 2;
    }
    else if (is_a_nb == TRUE) {
      xplab_label_set_text(catt.infoline, XPLAB_CLASS_PRINTABLE_NB_WARNING);
      changes = 2;
    }
  }

  FREEMEM(buf);

  /* check fixed entry                                                     */
  assert(GTK_IS_COMBO_BOX(catt.fixed.combo)==TRUE);
  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(catt.fixed.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  assert(buf);

  if (!strlen(buf) || (strcmp(buf, "TRUE") && strcmp(buf, "FALSE"))) {
    xplab_label_set_text(catt.infoline, XPLAB_CLASS_FIXED_PARSING_ERROR);
    FREEMEM(buf);
    return -1;
  }

  FREEMEM(buf);

  return changes;
}



/* Checks the user alignment string input.

Revision: 29.07.2005 E.Sack

Input : - pointer to training settings
Output: - flag (0: error, 1: success)
*/
static int 
xplab_check_alignment(entry_def alignment)
{
  gint     flag;
  gchar    *buf;
  ALIGMNT  enum_out;
  GtkEntry *centry;

  assert(GTK_IS_COMBO_BOX(alignment.combo)==TRUE);

  centry = GTK_ENTRY(GTK_BIN(GTK_COMBO_BOX(alignment.combo))->child);
  buf = xplab_entry_get_text(centry);                      /* fetch string */

  flag = string_to_aligmnt(buf, &enum_out);

  FREEMEM(buf);

  xplab_label_set_text(*alignment.info_pt, 
		     flag?XPLAB_ALIGNMENT_CONVERSION_SUCCESS_MSG:
		     XPLAB_ALIGNMENT_CONVERSION_FAILED_MSG);

  return flag;
}



/* Takes over train data from trace page format to DB.

Revision: 19.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
trace_train(GtkButton *button)
{
  gint nnr, cnr = XPLAB_SINGLE_CLASS, marked = 0, trained = 0;

  assert(tr.page.trace.tdata.parse_status == XPLAB_PARSE_STAT_EXT);/* must be parsed */

  /* save hash IDs in traced list                                            */
  copy_cls_lst(&tr.page.trace.tdata.traced.sclass, tr.list_new.cls, FALSE);

  for (nnr = 0; nnr < tr.page.trace.tdata.lbls.cnt; nnr++) {
    tr.list_new.cls->cls[cnr].nbl.nb[nnr].aux = tr.page.trace.tdata.lbls.nr[nnr];

    switch (tr.page.trace.tdata.lbls.nr[nnr]) {    /* hash IDs in new list deleted */
    case XPLAB_SELECTED_PIXEL:                  /* train and mark as trained */ 
      tr.page.trace.tdata.lbls.nr[nnr] = XPLAB_TRAINED_PIXEL;
      color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr], 
		   tr.session.mimg.pimg, tr.page.trace.tdata.traced.spattn.sample,
		   XPLAB_TRAINED_PIXEL);
      trained++;
      marked++;
      break;

    case XPLAB_IGNORED_PIXEL:                        /* mark as trained only */
      tr.page.trace.tdata.lbls.nr[nnr] = XPLAB_TRAINED_PIXEL;
      color_nb_def(tr.page.trace.tdata.traced.sclass.cls[cnr].nbl.nb[nnr], 
		   tr.session.mimg.pimg, tr.page.trace.tdata.traced.spattn.sample,
		   XPLAB_TRAINED_PIXEL);
      marked++;
      break;

    case XPLAB_SELECTBL_PIXEL:      /* not trained and not marked as trained */
      break;

    case XPLAB_TRAINED_PIXEL:                             /* already trained */
      break;

    default: assert(0); break;
    }
  }

  /* extract trainable pattern                                               */
  xremove_nb_def(*tr.list_new.cls, XPLAB_IGNORED_PIXEL);
  xremove_nb_def(*tr.list_new.cls, XPLAB_SELECTBL_PIXEL);
  xremove_nb_def(*tr.list_new.cls, XPLAB_TRAINED_PIXEL);

  assert(trained==tr.list_new.cls->cls[cnr].nbl.cnt);       /* check extract */

  trace_train_message(marked, trained, tr.trun.plabel);

  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);
  pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);

  reset_parse_status(*tr.list_new.cls);
  set_parse_status(*tr.list_new.cls, tr.page.trace.tdata.parse_status);
  update_lists(&tr.list_new, tr.list_old, tr.tad.mborder, 
	       tr.tad.maskweight, 1);              /* add new NB to old list */

  cb_leave_train(button, NULL);                       /* direct handler call */

  gtk_widget_set_sensitive(tr.page.save_button, tr.list_old.cls->cnt?TRUE:FALSE);
  gtk_widget_set_sensitive(tr.page.save_quit_button, tr.list_old.cls->cnt?TRUE:FALSE);
}



/* Prints a message.

Revision: 12.09.2003 E.Sack

Input : - numbers to be output
        - widget
Output: - message
*/
static void
trace_train_message(gint marked, gint trained, GtkWidget *label)
{
  char *buf = NULL;

  int_to_string("patterns ", trained, " trained, ", &buf);
  int_to_string("", marked, " marked", &buf);

  xplab_label_set_text(label, buf);

  FREEMEM(buf);
}



/* Deletes marked NB in a class list.

Revision: 12.09.2003 E.Sack

Input : - class list
Output: - class list updated
        - number of erased NBs
*/
static int
xremove_nb_def(cls_lst clst, int mark) 
{
  int erased = 0, cnr, err1, nnr;

  while (1) {
    err1 = xmarked_pattern(clst, &cnr, &nnr, mark);

    if (err1) {
      erase_nb_def(clst, cnr, nnr, NULL, 0);                   /* erase NB */
      erased++;
    }
    else {
      break;
    }
  }

  return erased;
}



/* Detects in a class list marked NBs.

Revision: 12.09.2003 E.Sack

Input : - class list
        - pointer to indexes
Output: - flag (0: no mark found, 1: mark found)
        - indexes updated as flagged
*/
static int 
xmarked_pattern(cls_lst clst, int *cnr, int *nnr, int mark) 
{
  for (*cnr = 0; *cnr < clst.cnt; (*cnr)++) {               /* all classes */
    for (*nnr = 0; *nnr < clst.cls[*cnr].nbl.cnt; (*nnr)++) {   /* all NBs */
      if (clst.cls[*cnr].nbl.nb[*nnr].aux == mark) {
	return 1;
      }
    }
  }

  return 0;
}



/* Takes over trained data from select page format to DB.

Revision: 20.01.2006 E.Sack

Input : - GTK convention
Output: - -
*/
static void
select_train(GtkButton *button)
{
  int      d, found, cnt_new;
  gboolean nb_is, cnb_is, nnb_is;

  assert(tr.page.select.satt.parse_status == XPLAB_PARSE_STAT_NEW || /* must be parsed */
         tr.page.select.satt.parse_status == XPLAB_PARSE_STAT_EXT);

  cnt_new = tr.list_new.cls->cnt;
  nb_is = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nb_button));
  cnb_is = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.cnb_button));
  nnb_is = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nnb_button));

  if (nb_is == TRUE) {                          /* mark pattern as trained */
    d = xcreate_snb(XPLAB_SELECTED_PIXEL, tr.page.select.focus_sector, tr.list_new.cls);
    found = set_pixel(tr.page.select.focus_sector, XPLAB_SELECTED_PIXEL, 
		      XPLAB_TRAINED_PIXEL, tr.session.mimg.pimg, 0);
  }
  else if (cnb_is == TRUE) {
    d = xcreate_cnb(tr.list_new.cls, tr.page.select.satt);
    found = set_pixel(tr.page.select.focus_sector, XPLAB_SELECTED_PIXEL, 
		      XPLAB_TRAINED_PIXEL, tr.session.mimg.pimg, 0);
  }
  else if (nnb_is == TRUE) {
    d = xcreate_snb(XPLAB_SELECTED_PIXEL, tr.page.select.focus_sector, tr.list_new.cls);
    found = set_pixel(tr.page.select.focus_sector, XPLAB_SELECTED_PIXEL, 
		      XPLAB_RESERVED_PIXEL, tr.session.mimg.pimg, 0);
  }
  else {
    assert(0);
  }

  assert(found);
  assert(d >= 0);

  set_parse_status(*tr.list_new.cls, tr.page.select.satt.parse_status);

  center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
  image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);

  if (nb_is == TRUE || cnb_is == TRUE) {         /* add new NB to old list */
    update_lists(&tr.list_new, tr.list_old, tr.tad.mborder, tr.tad.maskweight, 1);
  }
  else {                            /* don't copy NNB to old list directly */
    append_cls_lst_to_model(*tr.list_new.cls, tr.list_new.tview, cnt_new, 1, FALSE);
  }

  cb_leave_train(button, NULL);                       /* direct handler call */

  if (cnb_is == TRUE) {
    rb_update();
  }
  else {
    calculate_blob_number(0);
  }

  gtk_widget_set_sensitive(tr.page.select.propose_button, tr.list_old.cls->cnt?TRUE:FALSE);
  gtk_widget_set_sensitive(tr.page.select.trace_button, tr.list_old.cls->cnt?TRUE:FALSE);

  found = get_parse_status(*tr.list_old.cls);

  if (found) {                                           /* new data to save */
    gtk_widget_set_sensitive(tr.page.save_button, TRUE);
    gtk_widget_set_sensitive(tr.page.save_quit_button, TRUE);
  }
}



/* Copies new trained class list to old trained class list and updates the
   model accordingly.

Revision: 10.09.2003 E.Sack

Input : - pointer to input list
        - output list
        - mask border
        - mask weight
        - row flag
Output: - class lists updated
*/
static void
update_lists(list_def *list_new, list_def list_old, int mborder, 
	     int maskweight, int moveto)
{
  int cnt_old = list_old.cls->cnt;

  if (list_new->cls->cnt) {                     /* trained NB copied to DB */
    gen_class_references(*list_new->cls, mborder, maskweight);/*new NB refs*/

    cat_cls_lst(*list_new->cls, list_old.cls);         /* merge class list */

    clear_list_def(list_new);

    append_cls_lst_to_model(*list_old.cls, list_old.tview, cnt_old, moveto, FALSE);
  }
}



/* Assembles the NB attributes and pattern into a class list, which is
   extended accordingly.

Revision: 28.09.2003 E.Sack

Input : - pixel marking
        - sector affected
        - pointer to class list
        - pointer to training settings
Output: - class list updated
        - new class index 
*/
static int 
xcreate_snb(int mark, sector_def fr, cls_lst *new)
{
  int        cls_idx, d;
  cls_def    zcls;                                      /* tentative class */
  sector_def sd;

  assert(new);

  init_cls_def(&zcls);

  zcls.att = xget_entry_cls_att(tr.page.select.satt);/*get trained attributes*/

  zcls.nbl.cnt = 1;                                    /* list with one NB */
  CALLOC(nb_def, zcls.nbl.cnt, zcls.nbl.nb);
  init_nb_def(zcls.nbl.nb); 

  d = contour_frame((PICT)mark, &fr, tr.session.mimg.pimg, 0);/* minimal NB frame */
  assert(d);

  SECTOR_TO_AREA(fr, zcls.nbl.nb->pat.ar);       /* over all NB frame size */

  CALLOC(PICT, AREA_SIZE(zcls.nbl.nb->pat.ar), zcls.nbl.nb->pat.mem);
  copy_bmps(tr.session.mimg.pimg, zcls.nbl.nb->pat, zcls.nbl.nb->pat.ar, 
	    fr.column.start, fr.line.start, 0, 0);   /* copy pattern to NB */
  zcls.nbl.nb->pat.org.x = fr.column.start;          /* save source origin */
  zcls.nbl.nb->pat.org.y = fr.line.start;

  AREA_TO_SECTOR(zcls.nbl.nb->pat.ar, sd);            /* convert to bitmap */
  unset_pixel(sd, (PICT)mark, BACKGROUND, zcls.nbl.nb->pat);
  set_pixel(sd, (PICT)mark, FOREGROUND, zcls.nbl.nb->pat, 0);

  cls_idx = new->cnt++;                                /* enroll new class */

  if (cls_idx) {                                /* extend input class list */
    REALLOC(cls_def, new->cnt, new->cls);
  }
  else {
    CALLOC(cls_def, new->cnt, new->cls);
  }

  init_cls_def(&new->cls[cls_idx]);                       /* enroll new NB */
  move_cls_def(&zcls, &new->cls[cls_idx], FALSE);

  return cls_idx;
}



/* Builds a CNB from a NB list and enrolls it.

Revision: 28.09.2003 E.Sack

Input : - pointer to class list
Output: - class list updated
        - CNB class index (-1: empty class list)
*/
static int
xcreate_cnb(cls_lst *new, cls_att_entry entr) 
{
  int        cnr, cstart, lstart, width, height, cnb_idx;
  sector_def cfr;
  cls_def    zcls;                                      /* tentative class */

  assert(new);

  init_cls_def(&zcls);

  if (!new->cls) {                                       /* NNB list empty */
    assert(!new->cnt);
    return -1;
  }
  else {
    assert(new->cnt > 0);
  }

  for (cnr = 0; cnr < new->cnt; cnr++) {/*CNB built only of single-NB-class*/
    assert(new->cls[cnr].nbl.cnt==1);/* NNB classes contain exactly one NB */
  }

  zcls.att = xget_entry_cls_att(entr);           /* get trained attributes */

  zcls.nbl.cnt = 1;                                    /* list with one NB */
  CALLOC(nb_def, zcls.nbl.cnt, zcls.nbl.nb);
  init_nb_def(zcls.nbl.nb); 

  cfr.column.start = new->cls[0].nbl.nb[0].pat.org.x; /* init with class 0 */
  cfr.line.start   = new->cls[0].nbl.nb[0].pat.org.y;
  cfr.column.end   = cfr.column.start + 
                     new->cls[0].nbl.nb[0].pat.ar.width - 1;
  cfr.line.end     = cfr.line.start + 
                     new->cls[0].nbl.nb[0].pat.ar.height - 1;

  for (cnr = 1; cnr < new->cnt; cnr++) {              /* minimal CNB frame */
    cstart = new->cls[cnr].nbl.nb[0].pat.org.x;
    lstart = new->cls[cnr].nbl.nb[0].pat.org.y;
    width  = new->cls[cnr].nbl.nb[0].pat.ar.width;
    height = new->cls[cnr].nbl.nb[0].pat.ar.height;

    cfr.column.start = XPMIN(cfr.column.start, cstart);
    cfr.line.start   = XPMIN(cfr.line.start,   lstart);
    cfr.column.end   = XPMAX(cfr.column.end, cstart + width - 1);
    cfr.line.end     = XPMAX(cfr.line.end,   lstart + height - 1);
  }

  zcls.nbl.nb[0].pat.org.x = cfr.column.start;   /* save CNB source origin */
  zcls.nbl.nb[0].pat.org.y = cfr.line.start;

  SECTOR_TO_AREA(cfr, zcls.nbl.nb[0].pat.ar);   /* over all CNB frame size */

  CALLOC(PICT, AREA_SIZE(zcls.nbl.nb[0].pat.ar), 
	 zcls.nbl.nb[0].pat.mem);            /* assemble CNB pattern later */

  cnb_idx = new->cnt++;                /* index of CNB class in input list */

  if (cnb_idx) {                                /* extend input class list */
    REALLOC(cls_def, new->cnt, new->cls);
  }
  else {
    CALLOC(cls_def, new->cnt, new->cls);
  }

  init_cls_def(&new->cls[cnb_idx]);
  move_cls_def(&zcls, &new->cls[cnb_idx], FALSE);      /* enroll CNB class */

  for (cnr = 0; cnr < cnb_idx; cnr++) {           /* patch & reference NNB */
    new->cls[cnr].nbl.nb[0].cnb.cnt = 1;        /* enroll chain to one CNB */
    CALLOC(nb_chn_def, 1, new->cls[cnr].nbl.nb[0].cnb.cd);

    new->cls[cnr].nbl.nb[0].cnb.cd[0].cdx = cnb_idx;       /* chain to CNB */
    new->cls[cnr].nbl.nb[0].cnb.cd[0].ndx = 0;          /* NB-index of CNB */

    cstart = new->cls[cnr].nbl.nb[0].pat.org.x - cfr.column.start;
    lstart = new->cls[cnr].nbl.nb[0].pat.org.y - cfr.line.start;

    new->cls[cnr].nbl.nb[0].cnb.cd[0].x = cstart;/* relat. position in CNB */
    new->cls[cnr].nbl.nb[0].cnb.cd[0].y = lstart;

    copy_bmps(new->cls[cnr].nbl.nb[0].pat,         /* assemble CNB pattern */
	      new->cls[cnb_idx].nbl.nb[0].pat,
	      new->cls[cnr].nbl.nb[0].pat.ar, 0, 0, cstart, lstart);
  }

  gen_back_refs(*new);

  return cnb_idx;
}



/* Resets train button. The only function to leave train button.

Revision: 06.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_leave_train(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  PAGE_TABS   current_page;
  const gchar *label_text;
  gint        is_train;

  STSFY_GCC(user_data);

  assert(button);

  label_text = gtk_button_get_label(button);

  is_train = strcmp(label_text, XPLAB_TRAINING_DIALOG_TRAIN_LABEL) == 0;

  if (!is_train) {
    return;
  }

  current_page = xplab_get_page_enum(GTK_NOTEBOOK(tr.page.notebook), -1);

  switch (current_page) {
  case XPLAB_TRAINING_PAGE_IS_SELECT:

    tr.page.select.satt.parse_status = XPLAB_PARSE_STAT_DEF;
    xplab_label_set_text(tr.page.select.satt.infoline, "");/* empty infoline */

    break;

  case XPLAB_TRAINING_PAGE_IS_TRACE:

    tr.page.trace.tdata.parse_status = XPLAB_PARSE_STAT_DEF;

    break;

  case XPLAB_TAB_UNDEF:
    break;
  }

  /* reset button to parse                                                   */
  gtk_button_set_label(button, XPLAB_TRAINING_DIALOG_PARSE_LABEL);
}



/* Starts training dialog.

Revision: 19.09.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_train_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gboolean raised;

  STSFY_GCC(user_data);

  assert(menuitem);

  /* default setting before training */
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(tr.timg.mark_pattern), 
				 TRUE);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(tr.rb_frames_button), 
				 FALSE);

  raised = xplab_window_raise(&tr.dialog, 1);

  if (raised == FALSE) {                        /* create dialog only once */
    training_dialog();                           /* create training dialog */
  }
}



/* Calculates center of image displayed.

Revision: 09.06.2003 E.Sack

Input : - coordinate to center
        - image to display
        - window size
Output: - window offset
*/
static gfloat
focus_image(gfloat offset, gfloat image_size, gfloat page_size)
{

  if (image_size < page_size) {
    offset = 0.0;                                              /* most left */
  }
  else if (offset < page_size/2.0) {
    offset = 0.0;                                              /* most left */
  }
  else if (image_size - offset < page_size/2.0) {
    offset = image_size - page_size;                          /* most right */
  }
  else {
    offset -= page_size/2.0;                                      /* center */
  }

  return offset;
}



/* Centers window content to sector.

Revision: 08.09.2003 E.Sack

Input : - pointer to training settings
Output: - -
*/
static void
center_image(timg_def timg, sector_def sec)
{
  gfloat x_offset, y_offset;

  assert(timg.scroll.h_adj);
  assert(timg.scroll.v_adj);

  /* absolute center of sector */
  x_offset = (gfloat)(sec.column.start + sec.column.end)/2.0;
  y_offset = (gfloat)(sec.line.start + sec.line.end)/2.0;

  /* correct to rendering size */
  x_offset *= timg.state.magnitude;
  y_offset *= timg.state.magnitude;

  /* absolute adjustment of center of RB relative to clipping */
  x_offset = focus_image(x_offset, (gfloat)(timg.scroll.sdrawing->allocation.width+2*XPLAB_IMG_BORDER_SIZE), timg.scroll.h_adj->page_size);
  y_offset = focus_image(y_offset, (gfloat)(timg.scroll.sdrawing->allocation.height+2*XPLAB_IMG_BORDER_SIZE), timg.scroll.v_adj->page_size);

  gtk_adjustment_set_value(GTK_ADJUSTMENT(timg.scroll.h_adj), x_offset);
  gtk_adjustment_set_value(GTK_ADJUSTMENT(timg.scroll.v_adj), y_offset);
}



/* Callback for window focus.

Revision: 10.06.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_image_focus(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(menuitem);

  center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
}



/* Destroys training dialog.

Revision: 28.08.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_training_dialog_quit(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  int        parse_status;
  char       *buf = NULL;
  sector_def msec;

  STSFY_GCC(user_data);
  STSFY_GCC(button);

  if (GTK_IS_WIDGET(tr.dialog) == TRUE) {
    cb_trace_stop_button(NULL, NULL);              /* stop pattern tracing */
    cb_select_stop_button(NULL, NULL);         /* stop pattern correlation */

    xplab_window_raise(&tr.top_training, xctrl.cfg.raise); /* raise caller */

    clear_list_def(&tr.list_new);                      /* discard NNB list */

    AREA_TO_SECTOR(tr.session.mimg.pimg.ar, msec);
    set_pixel_masked(msec, (PICT)~XPLAB_TRAINED_PIXEL, FOREGROUND, 
		     tr.session.mimg.pimg, FALSE);

    image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);

    parse_status = get_parse_status(*tr.list_old.cls);

    int_to_string("", parse_status, " NB", &buf);
    cat_string(parse_status==1?"":"s", " trained", "", &buf);

    gtk_statusbar_push(GTK_STATUSBAR(tr.timg.status_bar), tr.timg.context_id,
		       buf);
    FREEMEM(buf);

    xplab_widget_destroy(&tr.dialog);

    tr.trun.pbar = NULL;
    tr.srun.pbar = NULL;
  }
}



/* Dumps RB to disk in displayed format.

Revision: 28.09.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_dump_rb(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(menuitem);

  xdump_rb();
}



/* Dumps image focus to disk in displayed format.

Revision: 28.09.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_select_focus_dump(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(menuitem);

  xdump_focus();
}



/* Dumps a RB sector to harddisc.

Revision: 28.09.2003 E.Sack

Input : - pointer to training settings
Output: - RB sector dumped
*/
static void 
xdump_rb(void) 
{
  int      rbn = tr.page.select.rb_no;
  char     *tmp_fname = NULL;
  area_def ard;
  pic_def  temp_img;

  init_pic_def(&temp_img);

  cat_string(tr.session.img_name, ".", RBNAME_DEFAULT, &tmp_fname);
  int_to_string("", rbn, "", &tmp_fname);

  temp_img.ar = tr.session.mimg.pimg.ar;
  temp_img.col = tr.rgbbuf;                             /* assemble format */

  ASSERT_SECTOR(tr.trbl.rb[rbn].sc, tr.session.mimg.pimg.ar);
  SECTOR_TO_AREA(tr.trbl.rb[rbn].sc, ard);

  dump_pic_def(temp_img, tmp_fname, ard, 
	       tr.trbl.rb[rbn].sc.column.start, 
	       tr.trbl.rb[rbn].sc.line.start, 
	       1, PNM_FORMAT_UNDEF);

  FREEMEM(tmp_fname);
}



/* Dumps image focus sector to harddisc.

Revision: 28.09.2003 E.Sack

Input : - pointer to training settings
Output: - focus sector dumped
*/
static void 
xdump_focus(void) 
{
  char     *tmp_fname = NULL;
  area_def ard;
  pic_def  temp_img;

  init_pic_def(&temp_img);

  add_string(tr.session.img_name, &tmp_fname);
  int_to_string(FOCUS_NAME_DEFAULT, 
		tr.page.select.focus_sector.column.start, "-", &tmp_fname);
  int_to_string("", tr.page.select.focus_sector.column.end, "x", &tmp_fname);
  int_to_string("", tr.page.select.focus_sector.line.start, "-", &tmp_fname);
  int_to_string("", tr.page.select.focus_sector.line.end, "", &tmp_fname);

  temp_img.ar = tr.session.mimg.pimg.ar;
  temp_img.col = tr.rgbbuf;                             /* assemble format */

  ASSERT_SECTOR(tr.page.select.focus_sector, tr.session.mimg.pimg.ar);
  SECTOR_TO_AREA(tr.page.select.focus_sector, ard);

  dump_pic_def(temp_img, tmp_fname, ard, 
	       tr.page.select.focus_sector.column.start, 
	       tr.page.select.focus_sector.line.start, 
	       1, PNM_FORMAT_UNDEF);

  FREEMEM(tmp_fname);
}



/* Saves trained data to disk.

Revision: 05.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_training_save(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  if (GTK_IS_WIDGET(tr.top_training) == TRUE) {
    save_training_session();
  }
}



/* Wrapper for button.

Revision: 05.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_training_dialog_save_quit(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  cb_training_save(NULL, NULL);                       /* direct handler call */
  cb_training_dialog_quit(NULL, NULL);                /* direct handler call */
}



/* Quits training phase and saves trained data to disk.

Revision: 07.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_training_save_quit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  if (GTK_IS_WIDGET(tr.top_training) == TRUE) {
    cb_training_save(NULL, NULL);                   /* at first save session */
    cb_training_quit(NULL, NULL);
  }
}



/* Quits GTK main loop.

Revision: 01.11.2004 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_training_quit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  int err;

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  cb_training_dialog_quit(NULL, NULL);

  if (GTK_IS_WIDGET(tr.top_training) == TRUE) {
    err = xplab_release_db(&tr.db_key, 1);                      /* unlock DB */
    assert(!err);

    xplab_update_session(tr.session.mimg, tr.image_key, tr.session.img_name);
    err = xplab_release_image(PXPLAB2, &tr.image_key, 1);    /* unlock image */
    assert(!err);

    xplab_window_raise(tr.caller, xctrl.cfg.raise);

    xplab_widget_destroy(&tr.top_training);

    /* cannot free data here because of sudden run stop                      */

  }
}



/* Deletes top window.

Revision: 17.10.2003 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
gboolean
cb_training_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);

  assert(event);
  assert(widget);

  cb_training_quit(NULL, NULL);

  return FALSE;
}



/* Initializes the training phase.

Revision: 09.05.2004 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_init_training(void)
{
  init_tr_def(&tr);

  tr.srun.pactive = &tr.cactive;
  tr.srun.pstop = &tr.cstop;
  tr.trun.pactive = &tr.cactive;
  tr.trun.pstop = &tr.cstop;
}



/* Frees the training phase.

Revision: 09.05.2004 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_free_training(void)
{
  tr.list_old.cls = NULL;                            /* protect input source */
  free_tr_def(&tr);

  xplab_init_training();
}



/* Training phase of pattern matching methods.

Revision: 01.11.2004 E.Sack

Input : - matching parameters
        - pointer to input image filename
        - pointer to TPS ID string
        - RB list
        - clustering kernel
        - clustering extension level
        - flag (0: multi line, 1: single line)
        - flag (0: multi NB, 1: single NB)
Output: - DB updated
Precondition: - global variables must be initialized.
*/
static void
xplab_training(match_def mat, char *filename, char *tps_id, rb_lst rbl, 
	       kern_def ckernel, int cluster_thold, int single_line, 
	       int single_nb)
{
  tr.session.tdb = xplab_request_db(XPLAB_TRAINING_TITLE_ITEM, &tr.db_key, 
				    XPLAB_DB_RW, NULL);

  tr.session.mimg = xplab_request_image(PXPLAB2, XPLAB_TRAINING_TITLE_ITEM,
					&tr.image_key, 1, NULL);

  tr.session.tps_idx = check_training_resources(tr.session.tdb, tps_id, 
						filename, tr.session.mimg.pimg,
						rbl);

  if (tr.session.tps_idx < 0) {
    xplab_release_db(&tr.db_key, 1);                            /* unlock DB */
    xplab_release_image(PXPLAB2, &tr.image_key, 1);  /* unlock session image */
    return;
  }

  add_string(filename, &tr.session.img_name);                  /* copy input */

  tr.cluster_thold = cluster_thold;                            /* copy input */
  tr.single_line = single_line;                                /* copy input */
  tr.single_nb = single_nb;                                    /* copy input */

  default_timg_def(&tr.timg);                       /* training image window */
  default_list_col(&tr.lcol);                                 /* list colors */
  default_select_def(&tr.page.select, rbl);            /* dialog page select */

  copy_match_def(&mat, &tr.tad);                               /* copy input */
  copy_rb_lst(&rbl, &tr.trbl);
  copy_kern_def(ckernel, tr.ckernel);

  CALLOC(cls_lst, 1, tr.list_new.cls);          /* temporary training buffer */
  init_cls_lst(tr.list_new.cls);

  cat_string(XPLAB_STATUS_MSG_SUFFIX_LEFT, tps_id,
	     XPLAB_STATUS_MSG_SUFFIX_RIGHT, &tr.status_suffix);

  tr.list_old.cls =                               /* short cut to local list */
    &tr.session.tdb.tpl.tps[tr.session.tps_idx].cl;

  training_window();                                        /* setup widgets */

  if (tr.session.mimg.typ == PXPLAB2 || tr.session.mimg.typ == PXPLAB5) {
    gtk_widget_set_sensitive(GTK_WIDGET(tr.view_original), FALSE);
  }
  else {
    gtk_widget_set_sensitive(GTK_WIDGET(tr.view_original), TRUE);
  }
}



/* Saves the training session to disc.

Revision: 19.08.2005 E.Sack

Input : - -
Output: - session saved to disc
*/
static void
save_training_session(void)
{
  int     err, parse_status;
  pic_def tmp;

  assert(tr.session.tdb.tpl.tps);                       /* check consistency */
  assert(tr.session.tps_idx>=0);
  assert(tr.session.tps_idx<tr.session.tdb.tpl.cnt);

  tmp = xplab_request_image(P1, NULL, NULL, 0, NULL).pimg;
  save_session_images(tr.session.mimg, tmp, tr.session.img_name);
  free_pic_def(&tmp);

  parse_status = get_parse_status(tr.session.tdb.tpl.tps[tr.session.tps_idx].cl);

  if (!parse_status) {
    return;
  }

  err = xplab_update_db(tr.session.tdb, tr.db_key);             /* update DB */

  if (err) {
    xplab_text_insert(main_log, TRAINING_WARNING_15);
  }
  else {
    xplab_text_insert(main_log, TRAINING_INFO_1);
  }

  reset_parse_status(tr.session.tdb.tpl.tps[tr.session.tps_idx].cl);

  gtk_widget_set_sensitive(tr.page.save_button, FALSE);
  gtk_widget_set_sensitive(tr.page.save_quit_button, FALSE);
}



/* Saves session images.

Revision: 09.07.2004 E.Sack

Input : - marked image
        - binary image
        - image filename
Output: - session saved
Postcondition: - input images left unchanged
*/
static void
save_session_images(img_typ mimg, pic_def bimg, char *filename)
{
  char    *buf = NULL, pxl_found;
  pic_def mtmp;

  assert(filename);

  mtmp = xdup_pic_def(mimg.pimg);                 /* duplicate input image */

  pxl_found = session_to_untrained(mtmp);

  if (pxl_found) {                                   /* dump session image */
    xplab_update_session(mimg, tr.image_key, filename);
  }

  init_img_typ(&mimg);                          /* inhibit input image use */

  if (xctrl.cfg.overwrite_image) {
    cat_string(filename, "", "",&buf);             /* dump untrained image */
    dump_pic_def(mtmp, buf, mtmp.ar, 0, 0, 1, PNM_FORMAT_UNDEF);
    FREEMEM(buf);
  }

  if (xctrl.cfg.training_write) {
    cat_string(filename, FILE_TO_TRAIN, "", &buf); /* dump untrained image */
    dump_pic_def(mtmp, buf, mtmp.ar, 0, 0, 1, PNM_FORMAT_UNDEF);
    FREEMEM(buf);

    /* trained = original - untrained                                      */
    mark_diff(bimg, mtmp, bimg.ar, 0, 0, 0, 0, FOREGROUND);

    cat_string(filename, FILE_TRAINED, "", &buf);    /* dump trained image */
    dump_pic_def(mtmp, buf, mtmp.ar, 0, 0, 1, PNM_FORMAT_UNDEF);
  }

  FREEMEM(buf);
  free_pic_def(&mtmp);
}



/* Gives back a default setting.

Revision: 25.06.2004 E.Sack

Input : - pointer to set
        - RB list
Output: - set updated
*/
static void
default_select_def(select_def *sdef, rb_lst rbl)
{
  assert(rbl.cnt);
  assert(XPLAB_START_RB_NO >= 0 && XPLAB_START_RB_NO < rbl.cnt);

  sdef->rb_no_old = -1;
  sdef->rb_no = XPLAB_START_RB_NO;

  sdef->focus_sector = sdef->focus_sector_old = rbl.rb[sdef->rb_no].sc;
  sdef->focus_rb_no = sdef->rb_no;
}



/* Gives back a default setting.

Revision: 22.08.2003 E.Sack

Input : - pointer to set
Output: - set updated
*/
static void
default_list_col(list_col *lcol)
{
  GdkColor xplab_class_fnd_color = XPLAB_CLASS_FND_COLOR;

  lcol->cfnd = xplab_class_fnd_color;
}



/* Gives back a default setting.

Revision: 22.08.2003 E.Sack

Input : - pointer to set
Output: - set updated
*/
static void
default_timg_def(timg_def *tig)
{
  tig->state.source    = DC_OPT_MARKED;
  tig->state.rb_frames = IMG_WITHOUT_RB_FRAMES;
  tig->state.magnitude = XPLAB_MAGN_X1_VALUE;

  tig->pcol = default_pattern_color();
}



/* Creates a dialog for training.

Revision: 06.08.2005 E.Sack

Input : - -
Output: - dialog created
*/
static void
training_dialog(void)
{
  gchar *title_string = NULL;

  tr.dialog = gtk_dialog_new();            /* set the training dialog widget */
  xplab_set_window_icon(tr.dialog);                           /* window icon */

  gtk_button_box_set_layout(GTK_BUTTON_BOX(GTK_DIALOG(tr.dialog)->action_area),
			    GTK_BUTTONBOX_SPREAD);

  /* set the training dialog title                                           */
  cat_string(XPLAB_TRAINING_TITLE_ITEM " ", tr.session.img_name, "", &title_string);
  gtk_window_set_title(GTK_WINDOW(tr.dialog), title_string);
  FREEMEM(title_string);

  /* set the training dialog input                                           */
  training_dialog_upper(GTK_CONTAINER(GTK_DIALOG(tr.dialog)->vbox));

  /* set the training dialog action                                          */
  training_dialog_lower(GTK_CONTAINER(GTK_DIALOG(tr.dialog)->action_area));

  gtk_widget_show_all(GTK_WIDGET(tr.dialog));
  gtk_widget_hide(tr.srun.pbar);                        /* hide progress bar */

  tr.page.select.rb_no_old = -1;             /* force start from the scratch */
  cb_select_rb_slider(NULL, NULL);                             /* init focus */

  g_signal_connect(G_OBJECT(tr.dialog), "delete-event",
		   G_CALLBACK(cb_training_dialog_delete), NULL);
}



/* Creates non editable DB lists.

Revision: 27.06.2003 E.Sack

Input : - pointer to container box
        - pointer to training settings
Output: - container box updated
*/
static void
training_dialog_select_lists(GtkWidget *box)
{
  GtkWidget *pan_vbox, *panwin;

  /* put into container a vertical box */
  pan_vbox = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(pan_vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(box), pan_vbox);

  /* put into vertical box a pan window */
  panwin = gtk_hpaned_new();
  gtk_box_pack_start(GTK_BOX(pan_vbox), panwin, TRUE, TRUE, 0);

  /* put into pan window two scrolled tree views */
  gtk_paned_add1(GTK_PANED(panwin), training_treeview(&tr.list_new.tview, tr.lcol.cfnd));
  gtk_paned_add2(GTK_PANED(panwin), training_treeview(&tr.list_old.tview, tr.lcol.cfnd));

  /* fill into model the trained attributes list */
  append_cls_lst_to_model(*tr.list_old.cls, tr.list_old.tview, 0, 0, TRUE);

  /* put into vertical box the find entry box */
  find_tentry(pan_vbox, tr.list_old.tview);
}



/* GTK cell function for row colors.

Revision: 06.08.2005 E.Sack

Input : - GTK convention
Output: - cell renderer updated
Usage : - Keep in mind that GTK doesn't like to change child models here.
*/
static void
cell_func_select(GTK_CONVENTION_TREE_CELL_DATA_FUNC_RTS_VOID)
{
  gint     d, children, cnt_new, cnt_old;
  gchar    *gbuf = NULL, *gbuf1 = NULL;
  gboolean is_printable, is_traceable, no_children;

  assert(tree_column);
  assert(GTK_IS_CELL_RENDERER(cell) == TRUE);
  assert(GTK_IS_TREE_MODEL(tree_model) == TRUE);
  assert(GTK_IS_TREE_MODEL_SORT(tree_model) == TRUE);
  assert(GTK_IS_TOGGLE_BUTTON(tr.page.select.trace_button) == TRUE);

  switch (GPOINTER_TO_INT(data)) {                     /* specific to column */
  case XPLAB_PAGE_SELECT_VIEW_COLUMN_CLS:              /* look for new class */
    gtk_tree_model_get(GTK_TREE_MODEL(tree_model), iter, 
		       XPLAB_DB_TREE_STORE_COLUMN_CLS_old, &gbuf, -1);

    if (gbuf) {                                         /* check consistency */
      gtk_tree_model_get(GTK_TREE_MODEL(tree_model), iter, 
			 XPLAB_DB_TREE_STORE_COLUMN_CLS, &gbuf1, -1);
      assert(gbuf1);

      d = strcmp(gbuf1, gbuf);
      assert(!d);                            /* class names cannot be edited */

      g_object_set(cell, "foreground-set", FALSE, NULL);          /* default */

      g_free(gbuf);
      g_free(gbuf1);
    }
    else {                                                /* new class found */
      g_object_set(cell, "foreground-set", TRUE, NULL);   /* only new string */
    }

    break;

  case XPLAB_PAGE_SELECT_VIEW_COLUMN_CNT:                /* check for new NB */
    gtk_tree_model_get(GTK_TREE_MODEL(tree_model), iter, 
		       XPLAB_DB_TREE_STORE_COLUMN_CNT, &cnt_new, -1);
    gtk_tree_model_get(GTK_TREE_MODEL(tree_model), iter, 
		       XPLAB_DB_TREE_STORE_COLUMN_CNT_old, &cnt_old, -1);

    if (cnt_new - cnt_old) {/* indicates a change against the original value */
      assert(cnt_new>cnt_old);                    /* a new NB has been added */

      g_object_set(cell, "foreground-set", TRUE, NULL);
    }
    else {                                                  /* no any change */
      g_object_set(cell, "foreground-set", FALSE, NULL);          /* default */
    }

    break;
  }

  /* the following is common to all cells                                    */
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.trace_button)) == FALSE) {
    g_object_set(cell, "background-set", FALSE, NULL);
    return;
  }

  /* look for traceable NB                                                   */
  gtk_tree_model_get(GTK_TREE_MODEL(tree_model), iter, 
		     XPLAB_DB_TREE_STORE_COLUMN_children, &children, -1);

  no_children = (children == 0);                    /* true means: not a CNB */

  gtk_tree_model_get(GTK_TREE_MODEL(tree_model), iter,
		     XPLAB_DB_TREE_STORE_COLUMN_PRT, &gbuf, -1);
  assert(gbuf);
  is_printable = (strcmp(gbuf, "TRUE") == 0) ? TRUE : FALSE; 
  g_free(gbuf);                                     /* free allocated memory */

  is_traceable = is_printable && /* 1st rule for definition of traceable object */
                 no_children;    /* 2nd rule for definition of traceable object */

  /* highlight row if it contains a traceable object */
  g_object_set(cell, "background-set", is_traceable, NULL);
}



/* Creates the basic tree store contained in a sort model.

Revision: 30.04.2005 E.Sack

Input : - -
Output: - pointer to sort model
*/
static GtkTreeModel *
select_sort_model_of_tree_store_init(void)
{
  gint         idx;
  GType        gtypes[XPLAB_DB_TREE_STORE_COLUMNS];
  GtkTreeStore *tstore;
  GtkTreeModel *smodel;

  for (idx = 0; idx < XPLAB_DB_TREE_STORE_COLUMNS; idx++) {
    switch (idx) {                             /* init store column types */
    case XPLAB_DB_TREE_STORE_COLUMN_IDX:
    case XPLAB_DB_TREE_STORE_COLUMN_CNT:
    case XPLAB_DB_TREE_STORE_COLUMN_CNT_old:
    case XPLAB_DB_TREE_STORE_COLUMN_children:
      gtypes[idx] = G_TYPE_INT;
      break;

    case XPLAB_DB_TREE_STORE_COLUMN_CLS:
    case XPLAB_DB_TREE_STORE_COLUMN_CLS_old:
    case XPLAB_DB_TREE_STORE_COLUMN_TRS:
    case XPLAB_DB_TREE_STORE_COLUMN_CMT:
    case XPLAB_DB_TREE_STORE_COLUMN_ALT:
    case XPLAB_DB_TREE_STORE_COLUMN_PRT:
    case XPLAB_DB_TREE_STORE_COLUMN_FIX:
    case XPLAB_DB_TREE_STORE_COLUMN_CVL:
      gtypes[idx] = G_TYPE_STRING;
      break;

    default:
      assert(0);                                  /* something is missing */
      break;
    }
  }
 
  tstore = gtk_tree_store_newv(XPLAB_DB_TREE_STORE_COLUMNS, gtypes);

  smodel = gtk_tree_model_sort_new_with_model(GTK_TREE_MODEL(tstore));
  g_object_unref(tstore);                 /* get rid of our own reference */

  for (idx = 0; idx < XPLAB_DB_TREE_STORE_COLUMNS; idx++) {
    gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(smodel), idx, 
				    select_sort_func, 
				    (gpointer)GTK_TREE_SORTABLE(smodel), 
				    NULL);
  }

  return smodel;
}



/* Creates the tree view with an empty store for the select page.

Revision: 06.08.2005 E.Sack

Input : - color for traceable NBs
Output: - pointer to tree view
*/
static GtkTreeView *
select_tree_view(GdkColor traceable_class_color)
{
  gint              cols;
  view_def          vcols[XPLAB_PAGE_SELECT_VIEW_COLUMNS];
  GtkTreeView       *tview;
  GtkTreeSelection  *tselctn;
  GtkTreeModel      *smodel;
  GtkTreeViewColumn *tv_column;
  GtkCellRenderer   *crenderer;
  GdkColor          gdk_red = XPLAB_RED_COLOR;

  /* We need a model, even if it is empty, else no headers are displayed.    */
  smodel = select_sort_model_of_tree_store_init();

  assert(XPLAB_PAGE_SELECT_VIEW_COLUMNS <= XPLAB_DB_TREE_STORE_COLUMNS);

  for (cols = 0; cols < XPLAB_PAGE_SELECT_VIEW_COLUMNS; cols++) {
    init_view_def(&vcols[cols]);
  }

  /* assign view headers */
  add_string(XPLAB_PAGE_SELECT_VIEW_TITLE_IDX, &vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_IDX].title);
  add_string(XPLAB_PAGE_SELECT_VIEW_TITLE_CLS, &vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_CLS].title);
  add_string(XPLAB_PAGE_SELECT_VIEW_TITLE_TRS, &vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_TRS].title);
  add_string(XPLAB_PAGE_SELECT_VIEW_TITLE_CMT, &vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_CMT].title);
  add_string(XPLAB_PAGE_SELECT_VIEW_TITLE_ALT, &vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_ALT].title);
  add_string(XPLAB_PAGE_SELECT_VIEW_TITLE_PRT, &vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_PRT].title);
  add_string(XPLAB_PAGE_SELECT_VIEW_TITLE_FIX, &vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_FIX].title);
  add_string(XPLAB_PAGE_SELECT_VIEW_TITLE_CNT, &vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_CNT].title);
  add_string(XPLAB_PAGE_SELECT_VIEW_TITLE_CVL, &vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_CVL].title);

  /* sort-column-ID translation table for the view.                        */
  vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_IDX].mcol = XPLAB_DB_TREE_STORE_COLUMN_IDX;
  vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_CLS].mcol = XPLAB_DB_TREE_STORE_COLUMN_CLS;
  vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_TRS].mcol = XPLAB_DB_TREE_STORE_COLUMN_TRS;
  vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_CMT].mcol = XPLAB_DB_TREE_STORE_COLUMN_CMT;
  vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_ALT].mcol = XPLAB_DB_TREE_STORE_COLUMN_ALT;
  vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_PRT].mcol = XPLAB_DB_TREE_STORE_COLUMN_PRT;
  vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_FIX].mcol = XPLAB_DB_TREE_STORE_COLUMN_FIX;
  vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_CNT].mcol = XPLAB_DB_TREE_STORE_COLUMN_CNT;
  vcols[XPLAB_PAGE_SELECT_VIEW_COLUMN_CVL].mcol = XPLAB_DB_TREE_STORE_COLUMN_CVL;

  tview = GTK_TREE_VIEW(gtk_tree_view_new_with_model(GTK_TREE_MODEL(smodel)));
  g_object_unref(G_OBJECT(smodel));          /* get rid of our own reference */

  tselctn = gtk_tree_view_get_selection(tview);
  gtk_tree_selection_set_mode(GTK_TREE_SELECTION(tselctn), GTK_SELECTION_SINGLE);

  for (cols = 0; cols < XPLAB_PAGE_SELECT_VIEW_COLUMNS; cols++) {
    assert(vcols[cols].title);    /* every column shall have a header string */

    crenderer = gtk_cell_renderer_text_new();/* text renderer for all view columns*/

    tv_column = gtk_tree_view_column_new_with_attributes
      (vcols[cols].title,
       crenderer, 
       "text", vcols[cols].mcol, /* model column: the real link to the store */
       NULL);

    gtk_tree_view_column_set_cell_data_func(tv_column, crenderer,
					    (GtkTreeCellDataFunc)cell_func_select,
					    GINT_TO_POINTER(cols),/* assuming columns not reordered*/
					    NULL);

    if (cols == XPLAB_PAGE_SELECT_VIEW_COLUMN_CVL) {
      gtk_tree_view_column_set_visible(GTK_TREE_VIEW_COLUMN(tv_column), FALSE);
    }
    else {
      gtk_tree_view_column_set_visible(GTK_TREE_VIEW_COLUMN(tv_column), TRUE);
    }

    g_object_set(crenderer,                                /* set to default */
		 "background-gdk", &traceable_class_color,
		 "background-set", FALSE, 
		 "foreground-gdk", &gdk_red,
		 "foreground-set", FALSE, 
		 NULL);

    gtk_tree_view_append_column(tview, GTK_TREE_VIEW_COLUMN(tv_column));

    /* Assign the sort-column-ID to the view. This is due to the
       fact, that view and model columns usually don't are
       identic. Furthermore, as the view columns contain data of the
       model, it has to use the model sort-column-IDs.                    */
    gtk_tree_view_column_set_sort_column_id(GTK_TREE_VIEW_COLUMN(tv_column),
					    vcols[cols].mcol);
  }

  gtk_tree_view_set_enable_search(tview, TRUE);
  gtk_tree_view_set_search_column(tview,                          /* default */
				  XPLAB_PAGE_SELECT_VIEW_COLUMN_CLS);

  g_signal_connect(G_OBJECT(tview), "cursor-changed",
		   G_CALLBACK(cb_select_row_select), NULL);

  for (cols = 0; cols < XPLAB_PAGE_SELECT_VIEW_COLUMNS; cols++) {
    free_view_def(&vcols[cols]);
  }

  return tview;
}



/* Creates a non editable view with an empty list model.

Revision: 06.08.2005 E.Sack

Input : - pointer to tree view pointer
        - color for traceable NBs
Output: - scrolled window
        - tree view updated
*/
static GtkWidget *
training_treeview(GtkTreeView **tview, GdkColor traceable_class_color)
{
  GtkWidget *scroll;

  assert(tview);

  *tview = select_tree_view(traceable_class_color);

  /* create a scrolled window                                                */
  scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), 
				 GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);

  /* put into scrolled window the tree view                                  */
  gtk_container_add(GTK_CONTAINER(scroll), GTK_WIDGET(*tview));

  return scroll;
}



/* Enrolls a class list into the tree view model.

Revision: 06.08.2005 E.Sack

Input : - class list
        - pointer to tree view
        - start index of class list
        - flag (0: let display unchanged, 1: move to updated row)
        - flag (initialization flag)
Output: - tree model updated
*/
static void
append_cls_lst_to_model(cls_lst clst, GtkTreeView *tview, int start_idx,
			int moveto, gboolean init_mirror)
{
  gint         cols, cidx, d, last_cidx = -1;
  gchar        *alignment = NULL, true[] = "TRUE", false[] = "FALSE";
  gchar        *sum, *txt0;
  guint        chdn;
  gboolean     ridx, valid_utf;
  GtkTreeIter  siter;
  GtkTreeModel *tstore, *smodel;

  assert(start_idx>=0);
  assert(GTK_IS_TREE_VIEW(tview) == TRUE);

  smodel = gtk_tree_view_get_model(GTK_TREE_VIEW(tview));
  assert(GTK_IS_TREE_MODEL(smodel) == TRUE);
  assert(GTK_IS_TREE_MODEL_SORT(smodel) == TRUE);
  assert(XPLAB_DB_TREE_STORE_COLUMNS == gtk_tree_model_get_n_columns(smodel));

  tstore = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(smodel));

  /* update changed values in model rows                                     */
  for (cidx = 0; cidx < start_idx; cidx++) {
    /* convert class index to row index                                      */
    ridx = xfind_row_from_index(tview, cidx, &siter);
    assert(ridx == TRUE);                                /* class must be in */

    gtk_tree_model_get(GTK_TREE_MODEL(smodel), &siter,
	 	       (gint)XPLAB_DB_TREE_STORE_COLUMN_CNT, &d, -1);

    if (clst.cls[cidx].nbl.cnt != d) { /* compare class list with tree model */
      GtkTreeIter child_iter;

      assert(clst.cls[cidx].nbl.cnt > d);/* difference can be extension only */

      gtk_tree_model_sort_convert_iter_to_child_iter
	(GTK_TREE_MODEL_SORT(smodel), &child_iter, &siter);

      gtk_tree_store_set(GTK_TREE_STORE(tstore),        /* update tree store */
			 &child_iter, XPLAB_DB_TREE_STORE_COLUMN_CNT, 
			 clst.cls[cidx].nbl.cnt, -1);

      last_cidx = cidx;                                  /* extended classes */
    }
  }

  /* append new classes to the (unsorted) model                              */
  for (cidx = start_idx; cidx < clst.cnt; cidx++) {
    gtk_tree_store_append(GTK_TREE_STORE(tstore), /* acquire a list iterator */
			  &siter, NULL);

    for (cols = 0; cols < XPLAB_DB_TREE_STORE_COLUMNS; cols++) {
      switch (cols) {
      case XPLAB_DB_TREE_STORE_COLUMN_IDX:
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, cidx, -1);
	break;

      case XPLAB_DB_TREE_STORE_COLUMN_CLS_old:
	if (init_mirror == FALSE) {
	  break;                                         /* attention please */
	}

      case XPLAB_DB_TREE_STORE_COLUMN_CLS:
	txt0 = clst.cls[cidx].att.class;                     /* pointer copy */

	valid_utf = g_utf8_validate(txt0, -1, NULL);

	sum = (valid_utf == TRUE) ? g_strdup(txt0) : xplab_to_utf(txt0);

	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   sum?sum:XPLAB_UTF_CONVERSION_ERROR, -1);
	g_free(sum);
	break;

      case XPLAB_DB_TREE_STORE_COLUMN_TRS:
	txt0 = clst.cls[cidx].att.sign;                      /* pointer copy */

	valid_utf = g_utf8_validate(txt0, -1, NULL);

	sum = (valid_utf == TRUE) ? g_strdup(txt0) : xplab_to_utf(txt0);

	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   sum?sum:XPLAB_UTF_CONVERSION_ERROR, -1);
	g_free(sum);
	break;

      case XPLAB_DB_TREE_STORE_COLUMN_CMT:
	txt0 = clst.cls[cidx].att.crem;                      /* pointer copy */

	valid_utf = g_utf8_validate(txt0, -1, NULL);

	sum = (valid_utf == TRUE) ? g_strdup(txt0) : xplab_to_utf(txt0);

	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   sum?sum:XPLAB_UTF_CONVERSION_ERROR, -1);
	g_free(sum);
	break;

      case XPLAB_DB_TREE_STORE_COLUMN_ALT:
	d = aligmnt_to_string(clst.cls[cidx].att.alignment, &alignment);
	assert(d==1);

	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, alignment, -1);
	FREEMEM(alignment);
	break;

      case XPLAB_DB_TREE_STORE_COLUMN_PRT:
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   clst.cls[cidx].att.printable?true:false, -1);
	break;

      case XPLAB_DB_TREE_STORE_COLUMN_FIX:
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   clst.cls[cidx].att.fixed?true:false, -1);
	break;

      case XPLAB_DB_TREE_STORE_COLUMN_CNT_old:
	if (init_mirror == FALSE) {
	  break;
	}

      case XPLAB_DB_TREE_STORE_COLUMN_CNT:
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   clst.cls[cidx].nbl.cnt, -1);
	break;

      case XPLAB_DB_TREE_STORE_COLUMN_CVL:
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, "---", -1);
	break;

      case XPLAB_DB_TREE_STORE_COLUMN_children:                 /* Ctrl Bgnd */
	chdn = class_has_children(clst.cls[cidx].nbl);
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, chdn, -1);
	break;

      default:
	assert(0);
	break;
      }
    }

    last_cidx = cidx;
  }

  if (last_cidx >= 0 && moveto) {                      /* rows can be sorted */
    treeview_moveto(tview, last_cidx, FALSE);
  }
}



/* Sets the cursor to a row in the tree view.

Revision: 06.08.2005 E.Sack

Input : - pointer to tree view
        - index of class list
        - flag (block curser signal)
Output: - tree view updated
*/
static void
treeview_moveto(GtkTreeView *tview, gint idx, gboolean be_quiet)
{
  gboolean          found;
  GtkTreePath       *path;
  GtkTreeIter       siter;
  GtkTreeModel      *smodel;
  GtkTreeViewColumn *vcol;

  found = xfind_row_from_index(GTK_TREE_VIEW(tview), idx, &siter);
  assert(found == TRUE);

  smodel = gtk_tree_view_get_model(GTK_TREE_VIEW(tview));
  assert(GTK_IS_TREE_MODEL(smodel) == TRUE);
  assert(GTK_IS_TREE_MODEL_SORT(smodel) == TRUE);

  path = gtk_tree_model_get_path(GTK_TREE_MODEL(smodel), &siter);

  vcol = gtk_tree_view_get_column(GTK_TREE_VIEW(tview), 
				  XPLAB_TREEVIEW_MOVE_TO_COLUMN);

  gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(tview), path, vcol, TRUE,
			       0.5, 0.0);

  if (be_quiet == TRUE) {
    g_signal_handler_block((gpointer)tview, tr.page.trace.cursor_changed_id);
  }

  gtk_tree_view_set_cursor(GTK_TREE_VIEW(tview), path, NULL, FALSE);

  if (be_quiet == TRUE) {
    g_signal_handler_unblock((gpointer)tview, tr.page.trace.cursor_changed_id);
  }

  gtk_tree_path_free(path);
}



/* Creates the basic tree store contained in a sort model.

Revision: 30.04.2005 E.Sack

Input : - -
Output: - pointer to sort model
*/
static GtkTreeModel *
trace_sort_model_of_tree_store_init(void)
{
  gint         idx;
  GType        gtypes[XPLAB_TR_TREE_STORE_COLUMNS];
  GtkTreeStore *tstore;
  GtkTreeModel *smodel;

  for (idx = 0; idx < XPLAB_TR_TREE_STORE_COLUMNS; idx++) {
    switch (idx) {                             /* init store column types */
    case XPLAB_TR_TREE_STORE_COLUMN_IDX:
    case XPLAB_TR_TREE_STORE_COLUMN_RBN:
    case XPLAB_TR_TREE_STORE_COLUMN_NBN:
      gtypes[idx] = G_TYPE_INT;
      break;

    case XPLAB_TR_TREE_STORE_COLUMN_CVL:
    case XPLAB_TR_TREE_STORE_COLUMN_CLS:
    case XPLAB_TR_TREE_STORE_COLUMN_DUM:
      gtypes[idx] = G_TYPE_STRING;
      break;

    default:
      assert(0);                                  /* something is missing */
      break;
    }
  }
 
  tstore = gtk_tree_store_newv(XPLAB_TR_TREE_STORE_COLUMNS, gtypes);

  smodel = gtk_tree_model_sort_new_with_model(GTK_TREE_MODEL(tstore));
  g_object_unref(tstore);                 /* get rid of our own reference */

  for (idx = 0; idx < XPLAB_TR_TREE_STORE_COLUMNS; idx++) {
    gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(smodel), idx, 
				    trace_sort_func, 
				    (gpointer)GTK_TREE_SORTABLE(smodel), 
				    NULL);
  }

  return smodel;
}



/* Creates the tree view.

Revision: 06.08.2005 E.Sack

Input : - -
Output: - pointer to sort model
*/
static GtkTreeView *
trace_tree_view(void)
{
  gint              cols;
  view_def          vcols[XPLAB_PAGE_TRACE_VIEW_COLUMNS];
  GtkTreeView       *tview;
  GtkTreeModel      *smodel;
  GtkCellRenderer   *crenderer;
  GtkTreeSelection  *tselctn;
  GtkTreeViewColumn *tv_column;

  smodel = trace_sort_model_of_tree_store_init();

  assert(XPLAB_PAGE_TRACE_VIEW_COLUMNS <= XPLAB_TR_TREE_STORE_COLUMNS);

  for (cols = 0; cols < XPLAB_PAGE_TRACE_VIEW_COLUMNS; cols++) {
    init_view_def(&vcols[cols]);
  }

  /* assign view headers                                                     */
  add_string(XPLAB_PAGE_TRACE_VIEW_TITLE_IDX, &vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_IDX].title);
  add_string(XPLAB_PAGE_TRACE_VIEW_TITLE_RBN, &vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_RBN].title);
  add_string(XPLAB_PAGE_TRACE_VIEW_TITLE_CVL, &vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_CVL].title);
  add_string(XPLAB_PAGE_TRACE_VIEW_TITLE_CLS, &vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_CLS].title);
  add_string(XPLAB_PAGE_TRACE_VIEW_TITLE_NBN, &vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_NBN].title);
  add_string(XPLAB_PAGE_TRACE_VIEW_TITLE_DUM, &vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_DUM].title);

  /* sort-column-ID translation table for the view.                          */
  vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_IDX].mcol = XPLAB_TR_TREE_STORE_COLUMN_IDX;
  vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_RBN].mcol = XPLAB_TR_TREE_STORE_COLUMN_RBN;
  vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_CVL].mcol = XPLAB_TR_TREE_STORE_COLUMN_CVL;
  vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_CLS].mcol = XPLAB_TR_TREE_STORE_COLUMN_CLS;
  vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_NBN].mcol = XPLAB_TR_TREE_STORE_COLUMN_NBN;
  vcols[XPLAB_PAGE_TRACE_VIEW_COLUMN_DUM].mcol = XPLAB_TR_TREE_STORE_COLUMN_DUM;

  tview = GTK_TREE_VIEW(gtk_tree_view_new_with_model(GTK_TREE_MODEL(smodel)));
  g_object_unref(G_OBJECT(smodel));          /* get rid of our own reference */

  tselctn = gtk_tree_view_get_selection(tview);
  gtk_tree_selection_set_mode(GTK_TREE_SELECTION(tselctn), GTK_SELECTION_SINGLE);

  for (cols = 0; cols < XPLAB_PAGE_TRACE_VIEW_COLUMNS; cols++) {
    assert(vcols[cols].title);    /* every column shall have a header string */

    crenderer = gtk_cell_renderer_text_new();/* text renderer for all view columns*/

    tv_column = gtk_tree_view_column_new_with_attributes
      (vcols[cols].title,
       crenderer, 
       "text", vcols[cols].mcol, /* model column: the real link to the store */
       NULL);

    gtk_tree_view_append_column(tview, GTK_TREE_VIEW_COLUMN(tv_column));

    gtk_tree_view_column_set_sort_column_id(GTK_TREE_VIEW_COLUMN(tv_column),
					    vcols[cols].mcol);
  }

  gtk_tree_view_set_enable_search(tview, FALSE);

  tr.page.trace.cursor_changed_id = 
    g_signal_connect(G_OBJECT(tview), "cursor-changed",
		   G_CALLBACK(cb_trace_row_select), NULL);

  for (cols = 0; cols < XPLAB_PAGE_TRACE_VIEW_COLUMNS; cols++) {
    free_view_def(&vcols[cols]);
  }

  return tview;
}



/* Creates non editable correlation result list.

Revision: 06.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
training_dialog_trace_list(GtkWidget *box)
{
  GtkWidget *vbox, *scroll;

  tr.page.trace.tview = trace_tree_view();

  /* put into container a vertical box                                       */
  vbox = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(box), vbox);

  /* put into vertical box a scrolled window                                 */
  scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), 
				 GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(vbox), scroll, TRUE, TRUE, 0);

  gtk_container_add(GTK_CONTAINER(scroll), GTK_WIDGET(tr.page.trace.tview));
}



/* Bundle of compare functions for sorting tree lists.

Revision: 06.08.2005 E.Sack

Input : - GTK convention
Output: - tree model updated
Usage: every tree model column has its own comparison, therefore this
       function does only fit to its associated tree model.
*/
static gint
select_sort_func(GTK_CONVENTION_GtkTreeIterCompareFunc_RTS_INT)
{
  gint             ia, ib, ret = 0;
  gint             sort_column_id;
  gchar            *texta, *textb;
  gboolean         rv;
  GtkSortType      order;
  GtkTreeModelSort *sort_model = (GtkTreeModelSort *)user_data;

  assert(a && b);
  assert(user_data);
  assert(GTK_IS_TREE_MODEL(model) == TRUE);                  /* child model  */
  assert(GTK_IS_TREE_MODEL_SORT(sort_model) == TRUE);        /* parent model */

  rv = gtk_tree_sortable_get_sort_column_id(GTK_TREE_SORTABLE(sort_model),
					    &sort_column_id, &order);

  assert(rv == TRUE);

  switch (sort_column_id) {
  case XPLAB_DB_TREE_STORE_COLUMN_IDX:
  case XPLAB_DB_TREE_STORE_COLUMN_CNT:
    gtk_tree_model_get(GTK_TREE_MODEL(model), a, sort_column_id, &ia, -1);
    gtk_tree_model_get(GTK_TREE_MODEL(model), b, sort_column_id, &ib, -1);

    ret = ((ia < ib) ? -1 : (ia > ib) ? 1 : 0);

    break;

  default:
    gtk_tree_model_get(GTK_TREE_MODEL(model), a, sort_column_id, &texta, -1);
    gtk_tree_model_get(GTK_TREE_MODEL(model), b, sort_column_id, &textb, -1);

    if (texta == NULL || textb == NULL) { 
      if (texta == NULL && textb == NULL) {
	break;                                      /* both equal => ret = 0 */
      }

      ret = (texta == NULL) ? -1 : 1;
    } 
    else {
      ret = g_utf8_collate(texta, textb);
    }

    g_free(texta);
    g_free(textb);

    break;
  }

  return ret;
}



/* Bundle of compare functions for sorting tree lists.

Revision: 06.08.2005 E.Sack

Input : - GTK convention
Output: - tree model updated
Usage: every tree model column has its own comparison, therefore this
       function does only fit to its associated tree model.
*/
static gint
trace_sort_func(GTK_CONVENTION_GtkTreeIterCompareFunc_RTS_INT)
{
  gint             ia, ib, ret = 0;
  gint             sort_column_id;
  gchar            *texta, *textb;
  gboolean         rv;
  GtkSortType      order;
  GtkTreeModelSort *sort_model = (GtkTreeModelSort *)user_data;

  assert(a && b);
  assert(user_data);
  assert(GTK_IS_TREE_MODEL(model) == TRUE);                  /* child model  */
  assert(GTK_IS_TREE_MODEL_SORT(sort_model) == TRUE);        /* parent model */

  rv = gtk_tree_sortable_get_sort_column_id(GTK_TREE_SORTABLE(sort_model),
					    &sort_column_id, &order);

  assert(rv == TRUE);

  switch (sort_column_id) {
  case XPLAB_TR_TREE_STORE_COLUMN_IDX:
  case XPLAB_TR_TREE_STORE_COLUMN_RBN:
  case XPLAB_TR_TREE_STORE_COLUMN_NBN:
    gtk_tree_model_get(GTK_TREE_MODEL(model), a, sort_column_id, &ia, -1);
    gtk_tree_model_get(GTK_TREE_MODEL(model), b, sort_column_id, &ib, -1);

    ret = ((ia < ib) ? -1 : (ia > ib) ? 1 : 0);

    break;

  default:
    gtk_tree_model_get(GTK_TREE_MODEL(model), a, sort_column_id, &texta, -1);
    gtk_tree_model_get(GTK_TREE_MODEL(model), b, sort_column_id, &textb, -1);

    if (texta == NULL || textb == NULL) { 
      if (texta == NULL && textb == NULL) {
	break;                                      /* both equal => ret = 0 */
      }

      ret = (texta == NULL) ? -1 : 1;
    } 
    else {
      ret = g_utf8_collate(texta, textb);
    }

    g_free(texta);
    g_free(textb);

    break;
  }

  return ret;
}



/* Creates the find entry.

Revision: 07.08.2005 E.Sack

Input : - pointer to input box
        - pointer to tree view
Output: - input box updated
*/
static void
find_tentry(GtkWidget *in_box, GtkTreeView *tview)
{
  GtkWidget   *hbox, *tmp_box;
  gbutton_def bdef;

  if (tr.page.select.find_entry.elist.items == NULL) {/* deallocated at the very end */
    tr.page.select.find_entry.elist.items = g_list_append(tr.page.select.find_entry.elist.items, TREE_VIEW_SEARCH_COL XPLAB_PAGE_SELECT_VIEW_TITLE_CLS);
    tr.page.select.find_entry.elist.items = g_list_append(tr.page.select.find_entry.elist.items, TREE_VIEW_SEARCH_COL XPLAB_PAGE_SELECT_VIEW_TITLE_TRS);
    tr.page.select.find_entry.elist.items = g_list_append(tr.page.select.find_entry.elist.items, TREE_VIEW_SEARCH_COL XPLAB_PAGE_SELECT_VIEW_TITLE_CMT);
    add_string(CTRL_FLOW_1, &tr.page.select.find_entry.elist.fail);
  }

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(in_box), hbox, FALSE, FALSE, 0);

  tmp_box = combo_list_find(&tr.page.select.find_entry, 
			    XPLAB_CLASS_ENTRY_LENGTH,
			    GTK_SIGNAL_FUNC(cb_select_search_column),
			    (gpointer)tview);

  tr.page.select.trace_button = 
    xset_toggled_button(tmp_box,
			GTK_SIGNAL_FUNC(cb_select_find_trace_button),
			"T", DESC1_TRAINING_TRACBLE_BUTTON,
			DESC2_TRAINING_TRACBLE_BUTTON, NULL).button;

  bdef = xset_clicked_button(tmp_box, 
			     GTK_SIGNAL_FUNC(cb_select_find_proposal),
			     "P", DESC1_TRAINING_PROPOSE_BUTTON,
			     DESC2_TRAINING_PROPOSE_BUTTON, NULL);
  tr.page.select.purpose_id = bdef.id;
  tr.page.select.propose_button = bdef.button;

  gtk_box_pack_start(GTK_BOX(hbox), tmp_box, FALSE, FALSE, 0);

  tr.page.select.find_info = gtk_label_new("");
  gtk_box_pack_start(GTK_BOX(hbox), tr.page.select.find_info, TRUE, TRUE, 0);
  gtk_misc_set_padding(GTK_MISC(tr.page.select.find_info), 5, 0);

  tr.srun.pbar = xplab_progress(hbox);
}



/* Enrolls a candidate list into store as text.

Revision: 07.08.2005 E.Sack

Input : - candidate list
        - pointer to tree view
        - pointer to class name string
Output: - tree view model updated
*/
static void
append_cand_lst_to_model(cand_lst hlst, GtkTreeView *tview, gchar *class_name)
{
  gint         cols, rnr, hidx;
  gchar        *sum, *txt0, *buf = NULL;
  gboolean     valid_utf;
  GtkTreeIter  siter;
  GtkTreeModel *tstore, *smodel;

  smodel = gtk_tree_view_get_model(tview);
  assert(GTK_IS_TREE_MODEL(smodel) == TRUE);
  assert(GTK_IS_TREE_MODEL_SORT(smodel) == TRUE);
  assert(XPLAB_TR_TREE_STORE_COLUMNS == gtk_tree_model_get_n_columns(GTK_TREE_MODEL(smodel)));

  tstore = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(smodel));

  /* append candidates to the (unsorted) model                               */
  for (hidx = rnr = 0; hidx < hlst.cnt; hidx++) {
    if (hlst.c[hidx].c < 0) {                  /* only recognized candidates */
      continue;
    }

    gtk_tree_store_append(GTK_TREE_STORE(tstore), &siter, NULL);/* acquire a list iterator */

    for (cols = 0; cols < XPLAB_TR_TREE_STORE_COLUMNS; cols++) {
      switch (cols) {
      case XPLAB_TR_TREE_STORE_COLUMN_IDX:
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, rnr++, -1);
	break;

      case XPLAB_TR_TREE_STORE_COLUMN_RBN:
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   hlst.c[hidx].rbn, -1);
	break;

      case XPLAB_TR_TREE_STORE_COLUMN_CVL:
	double_to_string("", hlst.c[hidx].m.cv, "", &buf);
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   buf, -1);
	FREEMEM(buf);
	break;

      case XPLAB_TR_TREE_STORE_COLUMN_CLS:
	txt0 = class_name;                                   /* pointer copy */

	valid_utf = g_utf8_validate(txt0, -1, NULL);

	sum = (valid_utf == TRUE) ? g_strdup(txt0) : xplab_to_utf(txt0);

	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   sum?sum:XPLAB_UTF_CONVERSION_ERROR, -1);
	g_free(sum);
	break;

      case XPLAB_TR_TREE_STORE_COLUMN_NBN:
	gtk_tree_store_set(GTK_TREE_STORE(tstore), &siter, cols, 
			   hlst.c[hidx].n, -1);
	break;

      case XPLAB_TR_TREE_STORE_COLUMN_DUM:   /* used for better delimitation */
	break;

      default:
	assert(0);
	break;
      }
    }
  }
}



/* Returns the iter from sorted model of the class index of the store.

Revision: 07.08.2005 E.Sack

Input : - pointer to tree view
        - class index
        - pointer to sorted iter
Output: - flag (index iter found)
Usage : - this function is used to convert class indexes to row indexes, which
          shall be the same always in the underlying tree store.
*/
static gboolean
xfind_row_from_index(GtkTreeView *tview, int cidx, GtkTreeIter *siter)
{
  gboolean     index_found;
  GtkTreePath  *child_path, *spath;
  GtkTreeModel *smodel;

  child_path = gtk_tree_path_new_from_indices(cidx, -1);        /* init path */

  smodel = gtk_tree_view_get_model(GTK_TREE_VIEW(tview));
  assert(GTK_IS_TREE_MODEL(smodel) == TRUE); 
  assert(GTK_IS_TREE_MODEL_SORT(smodel) == TRUE);

  spath =
    gtk_tree_model_sort_convert_child_path_to_path(GTK_TREE_MODEL_SORT(smodel), 
						   child_path);

  gtk_tree_path_free(child_path);

  index_found = gtk_tree_model_get_iter(smodel, siter, spath);

  gtk_tree_path_free(spath);

  return index_found;
}



/* Creates the training dialog upper part as a notebook with pages.

Revision: 07.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
training_dialog_upper(GtkContainer *vbox_in)
{
  gint pidx;

  tr.page.notebook = gtk_notebook_new();
  gtk_container_add(GTK_CONTAINER(vbox_in), tr.page.notebook);

  /* put the first page into the notebook, automatically displayed           */
  tr.page.select.parent = training_dialog_page_select();
  pidx = gtk_notebook_append_page(GTK_NOTEBOOK(tr.page.notebook), 
				  tr.page.select.parent, NULL);
  assert(pidx>=0);

  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(tr.page.notebook), 
				  tr.page.select.parent,
				  XPLAB_TRAINING_PAGE_SELECT);

  /* put the second page into the notebook                                   */
  tr.page.trace.tparent = training_dialog_page_trace();

  pidx = gtk_notebook_append_page(GTK_NOTEBOOK(tr.page.notebook), 
				  tr.page.trace.tparent, NULL);
  assert(pidx>=0);

  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(tr.page.notebook), 
				  tr.page.trace.tparent,
				  XPLAB_TRAINING_PAGE_TRACE);

  /* create a reversionary page with reference count = 1                     */
  tr.page.revert.rparent = training_dialog_page_revert();

  g_signal_connect(G_OBJECT(tr.page.notebook), "switch-page",
		   G_CALLBACK(cb_training_page_switch), NULL);
}



/* Creates the training page selection.

Revision: 20.06.2003 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget *
training_dialog_page_select(void)
{
  GtkWidget *vbox, *frame;

  /* create a vertical box */
  vbox = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);

  /* put into vertical box a frame */
  frame = gtk_frame_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  edit_class_attributes(frame, &tr.page.select.satt);         /* fill frame */

  /* put into vertical box another frame */
  frame = gtk_frame_new("trained attribute lists");
  gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  training_dialog_select_lists(frame);                       /* fill frame */

  /* put into vertical box another frame */
  frame = gtk_frame_new("pattern type selection");
  gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  training_pattern_type_selection(frame);                    /* fill frame */

  /* put into vertical box another frame */
  frame = gtk_frame_new("pattern selection");
  gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  training_pattern_selection(frame);                         /* fill frame */

  gtk_widget_set_sensitive(tr.page.select.propose_button, tr.list_old.cls->cnt?TRUE:FALSE);
  gtk_widget_set_sensitive(tr.page.select.trace_button, tr.list_old.cls->cnt?TRUE:FALSE);

  return vbox;
}



/* Creates the training page trace.

Revision: 01.08.2003 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget *
training_dialog_page_trace(void)
{
  GtkWidget *vbox, *frame;

  /* create a parent vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);

  /* put into parent a vertical box as child */
  tr.page.trace.tchild = gtk_vbox_new(FALSE, XPLAB_FRAME_SPACING);
  gtk_box_pack_start(GTK_BOX(vbox), tr.page.trace.tchild, TRUE, TRUE, 0);

  /* put into vertical box a frame */
  frame = gtk_frame_new("trace pattern selection");
  gtk_box_pack_start(GTK_BOX(tr.page.trace.tchild), frame, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  training_dialog_trace_pattern(frame);                      /* fill frame */

  /* put into vertical box another frame */
  frame = gtk_frame_new("traced attribute list");
  gtk_box_pack_start(GTK_BOX(tr.page.trace.tchild), frame, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  training_dialog_trace_list(frame);                         /* fill frame */

  /* put into vertical box another frame */
  frame = gtk_frame_new("trace control");
  gtk_box_pack_start(GTK_BOX(tr.page.trace.tchild), frame, FALSE, FALSE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  training_dialog_trace_control(frame);                      /* fill frame */

  return vbox;
}



/* Creates the reversionary page.

Revision: 02.08.2003 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget *
training_dialog_page_revert(void)
{
  GtkWidget *vbox, *frame;

  /* create a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 0);

  /* put into vertical box a frame */
  frame = gtk_frame_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  tr.page.revert.label = gtk_label_new ("reversionary page");
  gtk_container_add(GTK_CONTAINER(frame), tr.page.revert.label);

  return vbox;
}


/* Creates the trace pattern selection.

Revision: 26.09.2004 E.Sack

Input : - pointer to container box
        - pointer to training settings
Output: - container box updated
*/
static void
training_dialog_trace_pattern(GtkWidget *fbox)
{
  GtkWidget *pan_vbox, *panwin, *scroll_new, *scroll_old;
  area_def  dsize = {0,0};                                        /* reserve */

  /* put into container a vertical box                                       */
  pan_vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER(pan_vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(fbox), pan_vbox);

  /* put into vertical box a pan window                                      */
  panwin = gtk_vpaned_new();
  gtk_box_pack_start(GTK_BOX(pan_vbox), panwin, TRUE, TRUE, 0);

  /* put into upper box one scrolled drawing                                 */
  scroll_old = scrolled_drawing(&tr.page.trace.tdata.trained.spattn.pscroll, 
				dsize, NULL, NULL);

  gtk_paned_add1(GTK_PANED(panwin), scroll_old);

  /* put into lower box one scrolled drawing                                 */
  scroll_new = scrolled_drawing(&tr.page.trace.tdata.traced.spattn.pscroll, 
				dsize, NULL, NULL);

  gtk_paned_add2(GTK_PANED(panwin), scroll_new);

  /* put into vertical box the pattern selection                             */
  trace_pattern_selection(pan_vbox);
}



/* Creates the trace pattern selection format.

Revision: 07.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
trace_pattern_selection(GtkWidget *cbox)
{
  GtkWidget *hbox;

  /* put into container a horizontal box                                     */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 0);

  /* put into horizontal box a button                                        */
  tr.page.trace.all_button = 
    xset_toggled_button(hbox, NULL, "A", DESC1_TRAINING_PATTERN_ALL,
			DESC2_TRAINING_PATTERN_ALL, NULL).button;

  /* put into horizontal box a button                                        */
  tr.page.trace.select_button = 
    xset_clicked_button(hbox, GTK_SIGNAL_FUNC(cb_trace_select), 
			SINGLE_BUTTON_SELECT,
			DESC1_TRAINING_PATTERN_SELECT, 
			DESC2_TRAINING_PATTERN_SELECT, NULL).button;

  /* put into horizontal box a button                                        */
  tr.page.trace.deselect_button = 
    xset_clicked_button(hbox, GTK_SIGNAL_FUNC(cb_trace_deselect), 
			SINGLE_BUTTON_DESELECT,
			DESC1_TRAINING_PATTERN_DESLCT, 
			DESC2_TRAINING_PATTERN_DESLCT, NULL).button;

  /* put into horizontal box a button                                        */
  tr.page.trace.left_button = 
    xset_clicked_button(hbox, GTK_SIGNAL_FUNC(cb_trace_left_arrow), "<", 
			DESC1_TRAINING_PATTERN_BACKWD, 
			DESC2_TRAINING_PATTERN_BACKWD, NULL).button;

  /* put into horizontal box a button                                        */
  tr.page.trace.right_button = 
    xset_clicked_button(hbox, GTK_SIGNAL_FUNC(cb_trace_right_arrow), ">", 
			DESC1_TRAINING_PATTERN_FORWD, 
			DESC2_TRAINING_PATTERN_FORWD, NULL).button;

  tr.page.trace.infoline = gtk_label_new ("");
  gtk_box_pack_start(GTK_BOX(hbox), tr.page.trace.infoline, TRUE, TRUE, 0);
}



/* Creates the trace pattern control format.

Revision: 01.08.2003 E.Sack

Input : - pointer to container box
        - pointer to training settings
Output: - container box updated
*/
static void
training_dialog_trace_control(GtkWidget *fbox)
{
  GtkWidget *vbox, *hbox;

  /* put into container a vertical box */
  vbox = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(fbox), vbox);

  /* put into vertical box a horizontal box */
  hbox = gtk_hbox_new(FALSE, XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(vbox), hbox);

  /* put into horizontal box the button table */
  trace_control_table(hbox);

  /* put into horizontal box a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);

  /* put into vertical box a label */
  tr.trun.plabel = gtk_label_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), tr.trun.plabel, TRUE, TRUE, 0);

  /* put into vertical box the progress bar */
  tr.trun.pbar = xplab_progress(vbox);
}



/* Creates the trace pattern control format buttons.

Revision: 07.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
trace_control_table(GtkWidget *vbox)
{
  GtkWidget *table, *ebox;

  /* put into container a table                                              */
  table = gtk_table_new(2, 2, TRUE);
  gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 0);

  /* put into table a button                                                 */
  ebox = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_trace_rb_button), 
			     "RB", DESC1_TRAINING_PATTERN_TRACE, 
			     DESC2_TRAINING_PATTERN_TRACE, NULL).ebox;
  gtk_table_attach(GTK_TABLE(table), ebox, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

  /* put into table a button                                                 */
  ebox = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_trace_focus_button), 
			     SINGLE_BUTTON_FOCUS, DESC1_TRAINING_PATTERN_TRACEF, 
			     DESC2_TRAINING_PATTERN_TRACEF, NULL).ebox;
  gtk_table_attach(GTK_TABLE(table), ebox, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);

  /* put into table a button                                                 */
  ebox = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_trace_stop_button), 
			     SINGLE_BUTTON_STOP, DESC1_STOP_ACTION, 
			     DESC2_STOP_ACTION, NULL).ebox;
  gtk_table_attach(GTK_TABLE(table), ebox, 1, 2, 0, 2, GTK_FILL, GTK_FILL, 0, 0);
}



/* Creates the pattern type selection dialog.

Revision: 07.08.2005 E.Sack

Input : - pointer to container box
        - pointer to training settings
Output: - container box updated
*/
static void
training_pattern_type_selection(GtkWidget *box)
{
  GSList    *nb_group = NULL;
  GtkWidget *hbox, *hbox1, *event_box;

  /* put into vertical box a horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(box), hbox);

  /* put into horizontal box a NB button */
  tr.page.select.nb_button = gtk_radio_button_new_with_label(nb_group, "NB");
  event_box = gtk_event_box_new();
  gtk_box_pack_start(GTK_BOX(hbox), event_box, FALSE, FALSE, 0);
  gtk_container_add(GTK_CONTAINER(event_box), tr.page.select.nb_button);

  nb_group = 
    gtk_radio_button_get_group(GTK_RADIO_BUTTON(tr.page.select.nb_button));

  g_signal_connect(G_OBJECT(tr.page.select.nb_button), "toggled", 
		   G_CALLBACK(cb_select_nb_button), NULL);

  xplab_tooltips_set_tip(event_box, DESC1_TRAINING_PATTERN_NB,
			 DESC2_TRAINING_PATTERN_NB);

  /* put into horizontal box a CNB button                                    */
  tr.page.select.cnb_button = gtk_radio_button_new_with_label(nb_group,"CNB");
  event_box = gtk_event_box_new();
  gtk_box_pack_start(GTK_BOX(hbox), event_box, FALSE, FALSE, 0);
  gtk_container_add(GTK_CONTAINER(event_box), tr.page.select.cnb_button);

  nb_group = 
    gtk_radio_button_get_group(GTK_RADIO_BUTTON(tr.page.select.cnb_button));

  g_signal_connect(G_OBJECT(tr.page.select.cnb_button), "toggled", 
		   G_CALLBACK(cb_select_cnb_button), NULL);

  xplab_tooltips_set_tip(event_box, DESC1_TRAINING_PATTERN_CNB,
			 DESC2_TRAINING_PATTERN_CNB);

  /* put into horizontal box a NNB button                                    */
  tr.page.select.nnb_button = gtk_radio_button_new_with_label(nb_group, "NNB");

  event_box = gtk_event_box_new();
  gtk_box_pack_start(GTK_BOX(hbox), event_box, FALSE, FALSE, 0);
  gtk_container_add(GTK_CONTAINER(event_box), tr.page.select.nnb_button);

  nb_group = 
    gtk_radio_button_get_group(GTK_RADIO_BUTTON(tr.page.select.nnb_button));

  g_signal_connect(G_OBJECT(tr.page.select.nnb_button), "toggled", 
		   G_CALLBACK(cb_select_nnb_button), NULL);

  xplab_tooltips_set_tip(event_box, DESC1_TRAINING_PATTERN_NNB,
			 DESC2_TRAINING_PATTERN_NNB);

  /* put into horizontal box a horizontal box                                */
  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(hbox1), 5);
  gtk_box_pack_start(GTK_BOX(hbox), hbox1, FALSE, FALSE, 0);

  /* put into horizontal box a left arrow                                    */
  tr.page.select.nnb_lbutton = 
    xset_clicked_button(hbox1, GTK_SIGNAL_FUNC(cb_select_left_arrow), 
			"<", DESC1_TRAINING_PATTERN_NNBBWD, 
			DESC2_TRAINING_PATTERN_NNBBWD, NULL).button;

  /* put into horizontal box a right arrow                                   */
  tr.page.select.nnb_rbutton = 
    xset_clicked_button(hbox1, GTK_SIGNAL_FUNC(cb_select_right_arrow), 
			">", DESC1_TRAINING_PATTERN_NNBFWD, 
			DESC2_TRAINING_PATTERN_NNBFWD, NULL).button;

  /* put into horizontal box a neighborhood selection                        */
  tr.page.select.nbhood = 
    xset_toggled_button(hbox1, GTK_SIGNAL_FUNC(cb_select_nbhood), 
			" 4-N ", DESC1_TRAINING_PATTERN_NBHOOD, 
			DESC2_TRAINING_PATTERN_NBHOOD, NULL).button;

  /* put into horizontal box a label                                         */
  tr.page.select.pts_label = gtk_label_new("");
  gtk_box_pack_start(GTK_BOX(hbox1), tr.page.select.pts_label, FALSE, FALSE, 0);

  gtk_widget_set_sensitive(tr.page.select.nnb_lbutton, FALSE);
  gtk_widget_set_sensitive(tr.page.select.nnb_rbutton, FALSE);
  gtk_widget_set_sensitive(tr.page.select.nbhood, FALSE);
}



/* Creates the training dialog lower part.

Revision: 07.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
Bug: Cannot use leave-signal and event-box together, hence no displaying tips.
*/
static void
training_dialog_lower(GtkContainer *action)
{
  guint16     rows = 2, cols = 3;
  GtkWidget   *table;
  gbutton_def gd;

  table = gtk_table_new(cols, rows, FALSE);    /* put into container a table */
  gtk_container_add(action, table);

  /* put into table a parse button                                           */
  tr.page.parse_button = 
    gtk_button_new_with_label(XPLAB_TRAINING_DIALOG_PARSE_LABEL);

  gtk_table_attach_defaults(GTK_TABLE(table), tr.page.parse_button, 0, 3, 0, 1);

  g_signal_connect(G_OBJECT(tr.page.parse_button), "clicked",
		   G_CALLBACK(cb_parse_train), NULL);

  g_signal_connect(G_OBJECT(tr.page.parse_button), "leave", 
		   G_CALLBACK(cb_leave_train), NULL);

  xplab_tooltips_set_tip(tr.page.parse_button, DESC1_TRAINING_PARSE_BUTTON,
			 DESC2_TRAINING_PARSE_BUTTON);

  /* put into table a save button                                            */
  gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_training_save), 
			   DIALOG_BUTTON_SAVE, DESC1_TRAINING_SAVE_BUTTON, 
			   DESC2_TRAINING_SAVE_BUTTON, NULL);
  tr.page.save_button = gd.button;

  gtk_table_attach_defaults(GTK_TABLE(table), gd.ebox, 0, 1, 1, 2);

  /* put into table a save & quit button                                     */
  gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_training_dialog_save_quit), 
			   DIALOG_BUTTON_SAVE_QUIT, DESC1_TRAINING_SAVE_QUIT_BUTTON, 
			   DESC2_TRAINING_SAVE_QUIT_BUTTON, NULL);
  tr.page.save_quit_button = gd.button;

  gtk_table_attach_defaults(GTK_TABLE(table), gd.ebox, 1, 2, 1, 2);

  /* put into table a quit button                                            */
  gd = xset_clicked_button(NULL, GTK_SIGNAL_FUNC(cb_training_dialog_quit), 
			   DIALOG_BUTTON_QUIT, DESC1_TRAINING_QUIT_BUTTON, 
			   DESC2_TRAINING_QUIT_BUTTON, NULL);

  gtk_table_attach_defaults(GTK_TABLE(table), gd.ebox, 2, 3, 1, 2);

  gtk_widget_set_sensitive(tr.page.save_button, FALSE);
  gtk_widget_set_sensitive(tr.page.save_quit_button, FALSE);
}



/* Creates the pattern selection dialog.

Revision: 20.06.2003 E.Sack

Input : - pointer to container box
        - pointer to training settings
Output: - container box updated
*/
static void
training_pattern_selection(GtkWidget *box)
{
  GtkWidget *hbox, *vbox;

  /* put into container a vertical box */
  vbox = gtk_vbox_new(FALSE, XPLAB_IMG_BORDER_SIZE);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(box), vbox);

  /* put into vertical box a horizontal button box */
  hbox = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  create_rb_slider(hbox);         /* put into horizontal box the RB slider */
  create_focus_slider(hbox);    /*put into horizontal box the focus slider */

  /* put into vertical box a label */
  tr.page.select.satt.statusline3 = gtk_label_new ("");
  gtk_box_pack_start(GTK_BOX(vbox), tr.page.select.satt.statusline3, FALSE, FALSE, 0);
}



/* Creates the RB slider.

Revision: 07.08.2005 E.Sack

Input : - pointer to horizontal box
Output: - horizontal box updated
*/
static void
create_rb_slider(GtkWidget *hbox)
{
  GtkWidget *menu_bar, *rb_entry, *rb_entry_menu, *rb_dump;

  /* Create a focus menu                                                     */
  menu_bar = gtk_menu_bar_new();

  rb_entry = gtk_menu_item_new_with_label("RB:");
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), rb_entry);

  rb_entry_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(rb_entry), rb_entry_menu);

  tr.page.select.untrained_rb_item = gtk_check_menu_item_new_with_label("untrained RB");
  gtk_menu_shell_append(GTK_MENU_SHELL(rb_entry_menu), tr.page.select.untrained_rb_item);

  g_signal_connect(G_OBJECT(tr.page.select.untrained_rb_item), "toggled", 
		   G_CALLBACK(cb_select_untrained_rb), NULL);

  rb_dump = gtk_menu_item_new_with_label("dump RB");
  gtk_menu_shell_append(GTK_MENU_SHELL(rb_entry_menu), rb_dump);

  /* put into horizontal box a horizontal RB slider                          */
  tr.page.select.rb_slider = hslider(hbox, menu_bar, NULL, 30, 0, 
				     tr.trbl.cnt-1, 1, tr.page.select.rb_no,
				     DESC1_TRAINING_RB_SLIDER, 
				     DESC2_TRAINING_RB_SLIDER);

  g_signal_connect(G_OBJECT(tr.page.select.rb_slider.sadjustmnt), "changed", 
		   G_CALLBACK(cb_select_rb_slider), NULL);

  tr.page.select.rb_id = 
    g_signal_connect(G_OBJECT(tr.page.select.rb_slider.sadjustmnt), "value-changed",
		   G_CALLBACK(cb_select_rb_slider), NULL);

  g_signal_connect(G_OBJECT(tr.page.select.rb_slider.senty), "activate", 
		   G_CALLBACK(cb_select_rb_slider_entry), NULL);

  g_signal_connect(G_OBJECT(rb_dump), "activate", 
		   G_CALLBACK(cb_select_dump_rb), NULL);
}



/* Creates the focus slider.

Revision: 07.08.2005 E.Sack

Input : - pointer to horizontal box
Output: - horizontal box updated
*/
static void
create_focus_slider(GtkWidget *hbox)
{
  gint      focus_range;
  GtkWidget *menubar, *focus_entry, *focus_entry_menu, *focus_reset, *focus_dump;

  /* Create a focus menu                                                     */
  menubar = gtk_menu_bar_new();

  focus_entry = gtk_menu_item_new_with_label("focus:");
  gtk_menu_shell_append(GTK_MENU_SHELL(menubar), focus_entry);

  focus_entry_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(focus_entry), focus_entry_menu);

  focus_reset = gtk_menu_item_new_with_label("reset focus");
  gtk_menu_shell_append(GTK_MENU_SHELL(focus_entry_menu), focus_reset);

  focus_dump = gtk_menu_item_new_with_label("dump focus");
  gtk_menu_shell_append(GTK_MENU_SHELL(focus_entry_menu), focus_dump);

  g_signal_connect(G_OBJECT(focus_reset), "activate", 
		   G_CALLBACK(cb_select_focus_reset), NULL);

  g_signal_connect(G_OBJECT(focus_dump), "activate", 
		   G_CALLBACK(cb_select_focus_dump), NULL);

  /* put into horizontal box a horizontal focus slider with a menu           */
  focus_range = max_focus_aspect(tr.trbl.rb[tr.page.select.focus_rb_no].sc, 
				 &tr.page.select.focus);       /* init focus */
  tr.page.select.focus_old = tr.page.select.focus;

  tr.page.select.focus_slider = hslider(hbox, menubar, NULL, 50, 
					XPLAB_FOCUS_SLIDER_MIN, focus_range, 
					1, tr.page.select.focus,
					DESC1_TRAINING_FOCUS_SLIDER,
					DESC2_TRAINING_FOCUS_SLIDER);

  g_signal_connect(G_OBJECT(tr.page.select.focus_slider.sadjustmnt), "changed", 
		   G_CALLBACK(cb_select_focus_slider), NULL);

  tr.page.select.focus_id = 
    g_signal_connect(G_OBJECT(tr.page.select.focus_slider.sadjustmnt), "value-changed",
		     G_CALLBACK(cb_select_focus_slider), NULL);

  g_signal_connect(G_OBJECT(tr.page.select.focus_slider.senty), "activate", 
		   G_CALLBACK(cb_select_focus_slider_entry), NULL);
}



/* Creates a general horizontal slider.

Revision: 07.08.2005 E.Sack

Input : - pointer to container
        - pointer to label box
        - pointer to label string
        - label width
        - slider minimum
        - slider maximum
        - flag (0: box is to be packed, 1: box is container)
        - initial slider value
        - pointer to tooltip strings
Output: - slider
        - container updated
*/
static slider_def
hslider(GtkWidget *in_box, GtkWidget *mbox, char *label, gint label_width, 
	gint min_range, gint max_range, int is_box, gint init_slider_val, 
	const gchar *tip_text, const gchar *tip_private)
{
  char       cbuf[16];
  GtkWidget  *hbox1, *slider_label, *slider_ebox;
  slider_def slidr;

  init_slider_def(&slidr);

  assert(max_range >= 0);

  if (label || mbox) {
    assert(label_width > 0);
  }
  else {
    label_width = 0;
  }

  /* put into container a horizontal box                                     */
  hbox1 = gtk_hbox_new(FALSE, 5);

  if (is_box) {
    gtk_box_pack_start(GTK_BOX(in_box), hbox1, TRUE, TRUE, 0);
  }
  else {
    gtk_container_add(GTK_CONTAINER(in_box), hbox1);
  }

  if (mbox) {
    gtk_box_pack_start(GTK_BOX(hbox1), mbox, FALSE, FALSE, 0);
  }

  /* put into horizontal box a label                                         */
  if (label) {
    slider_label = gtk_label_new(label);

    gtk_box_pack_start(GTK_BOX(hbox1), slider_label, FALSE, FALSE, 0);
  }

  /* put into horizontal box a horizontal scale                              */
  slider_ebox = gtk_event_box_new();

  gtk_box_pack_start(GTK_BOX(hbox1), slider_ebox, TRUE, TRUE, 0);

  slidr.sadjustmnt = 
    GTK_ADJUSTMENT(gtk_adjustment_new(0.0, (gfloat)min_range, (gfloat)max_range,
				      XPLAB_RB_STEP_INCREMENT, 
				      XPLAB_RB_PAGE_INCREMENT, 
				      XPLAB_RB_PAGE_SIZE));

  slidr.sbutton = gtk_hscale_new(slidr.sadjustmnt);
  gtk_scale_set_draw_value(GTK_SCALE(slidr.sbutton), FALSE);
  gtk_widget_set_size_request(GTK_WIDGET(slidr.sbutton), 100, -1);

  gtk_container_add(GTK_CONTAINER(slider_ebox), slidr.sbutton);

  /* put into horizontal box an editable text line                           */
  slidr.senty = GTK_ENTRY(gtk_entry_new());
  gtk_entry_set_max_length(slidr.senty, 5);
  gtk_widget_set_size_request(GTK_WIDGET(slidr.senty), 40, 20);

  gtk_box_pack_start(GTK_BOX(hbox1), GTK_WIDGET(slidr.senty), FALSE, FALSE, 0);

  if (tip_text) {
    xplab_tooltips_set_tip(slider_ebox, tip_text, tip_private);
    xplab_tooltips_set_tip(GTK_WIDGET(slidr.senty), tip_text, tip_private);
  }

  g_snprintf(cbuf, sizeof(cbuf), "%d", init_slider_val);      /* init slider */
  xplab_entry_set_text(slidr.senty, cbuf);
  gtk_adjustment_set_value(slidr.sadjustmnt, (gfloat)init_slider_val);

  return slidr;
}



/* Creates top window for training.

Revision: 06.11.2004 E.Sack

Input : - -
Output: - -
*/
static void
training_window(void)
{
  area_def far;

  init_area_def(&far);

  /* set the training top window */
  tr.top_training = 
    xplab_image_window(XPLAB_TRAINING_IMAGE_TITLE_ITEM,
		       training_image_menu_bar,
		       training_status_bar, 
		       &tr.timg.scroll, tr.session.mimg.pimg.ar,
		       GTK_SIGNAL_FUNC(cb_mouse_training_image),
		       GTK_SIGNAL_FUNC(cb_image_slider),
		       GTK_SIGNAL_FUNC(cb_image_expose_event),
		       GTK_SIGNAL_FUNC(cb_training_delete),
		       XPLAB_IMG_EXTRA_SIZE, far);

  gtk_widget_show_all(tr.top_training);
}



/* Builds a status bar.

Revision: 07.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
training_status_bar(GtkWidget *vbox)
{
  gchar *image_name = NULL, *status_msg = NULL;

  /* put into container one status bar */
  tr.timg.status_bar = gtk_statusbar_new();
  gtk_box_pack_start(GTK_BOX (vbox), tr.timg.status_bar, FALSE, FALSE, 0);

  /* put into the status bar a message */
  tr.timg.context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(tr.timg.status_bar), "Statusbar binarized Input Image");
  image_name = name_of_image(tr.session.img_name, tr.timg.state, &status_msg);
  gtk_statusbar_push(GTK_STATUSBAR(tr.timg.status_bar), tr.timg.context_id, status_msg);

  FREEMEM(image_name);
  FREEMEM(status_msg);
}



/* Starts mouse button actions on the training image drawing.

Revision: 05.11.2004 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_mouse_training_image(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN)
{
  STSFY_GCC(widget);
  STSFY_GCC(user_data);

  if (tr.timg.scroll.sdrawing == NULL || 
      GTK_IS_WIDGET(tr.timg.scroll.sdrawing) != TRUE) {
    return TRUE;
  }

  switch (event->button) {
  case 1: training_image_boxes(event); break;
  case 2: training_image_label(event); break;
  }

  return TRUE;
}



/* Starts mouse button actions on the traced pattern drawing.

Revision: 05.11.2004 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_mouse_pattern_traced(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN)
{
  STSFY_GCC(widget);
  STSFY_GCC(user_data);

  if (tr.page.trace.tdata.traced.spattn.pscroll.sdrawing == NULL || 
      GTK_IS_WIDGET(tr.page.trace.tdata.traced.spattn.pscroll.sdrawing) != TRUE) {
    return TRUE;
  }

  switch (event->button) {
  case 1: 
    traced_pattern_cursor(event, tr.page.trace.tdata.traced.sclass);
    break;

  case 2: 
    traced_pattern_label(event, tr.page.trace.tdata.traced.sclass);
    break;
  }

  return TRUE;
}



/* Starts mouse button actions on the trained pattern drawing.

Revision: 05.11.2004 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_mouse_pattern_trained(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN)
{
  STSFY_GCC(widget);
  STSFY_GCC(user_data);

  if (tr.page.trace.tdata.trained.spattn.pscroll.sdrawing == NULL || 
      GTK_IS_WIDGET(tr.page.trace.tdata.trained.spattn.pscroll.sdrawing) != TRUE) {
    return FALSE;
  }

  switch (event->button) {
  case 1: 
    trained_pattern_cursor(event, tr.page.trace.tdata.trained.sclass);
    break;
  }

  return TRUE;
}



/* Moves background cursor in the trained pattern image according to the 
   mouse pointer coordinates. A NB parent can have more than one NB 
   child, so all children are highlighted consequently.

Revision: 07.08.2005 E.Sack

Input : - pointer to GDK event
        - class list
Output: - -
Postcondition: - traced child index may be changed
*/
static void
trained_pattern_cursor(GdkEventButton *event, cls_lst clst)
{
  int parent_found, nnr, cnr = XPLAB_SINGLE_CLASS;

  if (!clst.cls) {                                      /* consistency check */
    assert(!clst.cnt);
    return;
  }
  else {
    assert(clst.cnt == 1);                         /* single class mandatory */
  }

                                       /* use coordinates from mouse pointer */
  nnr = coordinates_to_index(clst.cls[cnr], (int)event->x, (int)event->y);

  if (nnr >= 0) {                                  /* move background cursor */
    parent_found = select_children(tr.page.trace.tview, &tr.page.trace.tdata, 
				   nnr);

    if (!parent_found) {
      return;    /* trained pattern found not a parent and hence no children */
    }

    pattern_drawing(tr.page.trace.tdata.trained.spattn, tr.timg.pcol, 0);
    pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 1);
  }
}



/* Moves background cursor in the traced pattern image according to the 
   mouse pointer coordinates.

Revision: 02.10.2004 E.Sack

Input : - pointer to GDK event
        - class list
Output: - -
Postcondition: - traced child index may be changed
*/
static void
traced_pattern_cursor(GdkEventButton *event, cls_lst clst)
{
  int nnr, cnr = XPLAB_SINGLE_CLASS;

  if (!clst.cls) {                                    /* consistency check */
    assert(!clst.cnt);
    return;
  }
  else {
    assert(clst.cnt == 1);                       /* single class mandatory */
  }

                                         /* coordinates from mouse pointer */
  nnr = coordinates_to_index(clst.cls[cnr], (int)event->x, (int)event->y);

  if (nnr >= 0) {                                /* move background cursor */
    traced_cursor(nnr);
  }
}



/* Moves the background highlighting cursor in the traced pattern image.

Revision: 07.08.2005 E.Sack

Input : - NB index
Output: - curser updated
*/
static void
traced_cursor(int nb_idx)
{
  assert(nb_idx >= 0 && nb_idx < tr.page.trace.tdata.lbls.cnt);/*check consistency */

  tr.page.trace.tdata.chno = nb_idx;

  select_child(&tr.page.trace.tdata, tr.page.trace.tview);

  pattern_drawing(tr.page.trace.tdata.trained.spattn, tr.timg.pcol, 0);
  pattern_drawing(tr.page.trace.tdata.traced.spattn, tr.timg.pcol, 0);
}



/* Performs actions triggered by mouse events on the training image.

Revision: 01.10.2004 E.Sack

Input : - pointer to GDK event
Output: - -
*/
static void
training_image_boxes(GdkEventButton *event)
{
  int        within;
  c_def      start;
  sector_def sec;

  AREA_TO_SECTOR(tr.session.mimg.pimg.ar, sec);    /* whole training image */

  within = image_coordinates(event, &start, sec, tr.timg.state.magnitude);

  if (!within) {            /* preselection: must be inside training image */
    return;
  }

  if (GTK_IS_WIDGET(tr.dialog) == TRUE) {
    PAGE_TABS current_page;

    current_page = xplab_get_page_enum(GTK_NOTEBOOK(tr.page.notebook), -1);

    switch (current_page) {
    case XPLAB_TRAINING_PAGE_IS_SELECT:
      training_image_rb(start);
      break;

    case XPLAB_TRAINING_PAGE_IS_TRACE:
      training_image_kernel_select(start);
      break;

    case XPLAB_TAB_UNDEF:
      break;
    }
  }
  else {        /* mouse actions on training image without training dialog */
    training_image_default(start);
  }
}



/* Changes RB-focus in the training image.

Revision: 07.08.2005 E.Sack

Input : - coordinate
Output: - RB index updated
Usage:  - training dialog activated.
*/
static void
training_image_rb(c_def start)
{
  int rb_idx, rb_idx_old;

  rb_idx = rb_from_coordinate(tr.trbl, start);

  if (rb_idx < 0) {
    return;                              /* coordinate not covered by a RB */
  }

  rb_idx_old = tr.page.select.rb_no;
  tr.page.select.rb_no = rb_idx;

  gtk_adjustment_set_value(GTK_ADJUSTMENT(tr.page.select.rb_slider.sadjustmnt),
			   tr.page.select.rb_no);

  if (rb_idx_old == rb_idx) {                    /* needs an extra trigger */
    tr.page.select.rb_no_old = -1;
    gtk_adjustment_value_changed(GTK_ADJUSTMENT(tr.page.select.rb_slider.sadjustmnt));
  }
}



/* Changes RB in the training image.

Revision: 02.10.2004 E.Sack

Input : - coordinate
Output: - RB selected
Usage:  - training dialog deactivated.
*/
static void
training_image_default(c_def start)
{
  int rb_idx = rb_from_coordinate(tr.trbl, start);

  if (rb_idx < 0) {
    return;                              /* coordinate not covered by a RB */
  }

  tr.page.select.rb_no = rb_idx;
  tr.page.select.focus_sector = tr.page.select.focus_sector_old = tr.trbl.rb[tr.page.select.rb_no].sc;
  tr.page.select.focus_rb_no = tr.page.select.rb_no;

  center_image(tr.timg, tr.trbl.rb[tr.page.select.rb_no].sc);
  image_to_state(1, 1);
}



/* Returns the enumeration of the current page tab of a notebook. This 
   allows the pages to be in an arbitrary order in the notebook.

Revision: 07.08.2005 E.Sack

Input : - pointer to notebook
        - page number (-1: current page)
Output: - page tab enum
*/
static PAGE_TABS
xplab_get_page_enum(GtkNotebook *notebook, int page_num)
{
  int         err;
  PAGE_TABS   tabs_enum = XPLAB_TAB_UNDEF;
  GtkWidget   *cpage, *clab;
  const gchar *str;

  if (!notebook) {
    return XPLAB_TAB_UNDEF;
  }

  assert(GTK_IS_NOTEBOOK(notebook) == TRUE);

  page_num = page_num < 0 ? gtk_notebook_get_current_page(notebook) : 
                            page_num;

  cpage = gtk_notebook_get_nth_page(notebook, page_num);

  if (!cpage) {
    return XPLAB_TAB_UNDEF;                /* page number is out of bounds */
  }

  clab = gtk_notebook_get_tab_label(notebook, cpage);

  if (!cpage) {
    return XPLAB_TAB_UNDEF;                                  /* not a page */
  }

  assert(GTK_IS_LABEL(clab) == TRUE);

  str = gtk_label_get_text(GTK_LABEL(clab));

  err = check_name(str);

  if (err) {
    return XPLAB_TAB_UNDEF;                            /* no label defined */
  }

  if (!strcmp(str, XPLAB_TRAINING_PAGE_SELECT)) {
    tabs_enum = XPLAB_TRAINING_PAGE_IS_SELECT;
  }
  else if (!strcmp(str, XPLAB_TRAINING_PAGE_TRACE)) {
    tabs_enum = XPLAB_TRAINING_PAGE_IS_TRACE;
  }

  return tabs_enum;
}



/* Searches the RB's index covering a given coordinate.

Revision: 02.10.2004 E.Sack

Input : - RB list
        - coordinate
Output: - RB index (-1: no RB found)
*/
static int
rb_from_coordinate(rb_lst rbl, c_def start)
{
  int        rdx;
  sector_def sec;

  if (!rbl.rb) {                                      /* consistency check */
    assert(!rbl.cnt);
    return -1;
  }
  else {
    assert(rbl.cnt>0);
  }

  AREA_TO_SECTOR(rbl.rar, sec);

  if (!WITHIN_SECTOR(start.x, start.y, sec)) {      /* must be inside area */
    return -1;
  }

  for (rdx = 0; rdx < rbl.cnt; rdx++) {
    if (WITHIN_SECTOR(start.x, start.y, rbl.rb[rdx].sc)) {
      return rdx;
    }
  }

  return -1;
}



/* Searches the child (traced NB) ID whose kernel covers a given coordinate.

Revision: 02.10.2004 E.Sack

Input : - hash list
        - coordinate
Output: - child ID (-1: no child found)
*/
static int
cid_from_coordinate(hash_def hash, c_def start)
{
  int hidx, hidx_found = -1;

  if (!hash.kr) {                                      /* consistency check */
    assert(!hash.cnt);
    assert(!hash.pid);
    assert(!hash.cid);
    return -1;
  }
  else {
    assert(hash.cnt>0);
    assert(hash.pid);
    assert(hash.cid);
  }

  for (hidx = 0; hidx < hash.cnt; hidx++) {
    if (WITHIN_SECTOR(start.x, start.y, hash.kr[hidx].knl)) {
      hidx_found = hidx;
      break;
    }
  }

  if (hidx_found < 0) {
    return -1;                    /* no child kernel for coordinates found */
  }

  return hash.cid[hidx_found];
}



/* Searches the traced NB whose kernel covers a given coordinate.

Revision: 02.10.2004 E.Sack

Input : - hash list
        - NB list
        - coordinate
Output: - NB index (-1: no child found)
*/
static int
nb_from_coordinate(hash_def hash, nb_lst nbl, c_def start)
{
  int cid, nb_idx;

  cid = cid_from_coordinate(hash, start);

  if (cid < XPLAB_INIT_HASH_ID) {
    return -1;                    /* coordinate not covered by a traced NB */
  }

  nb_idx = search_aux_nb_def(nbl, cid);

  assert(nb_idx >= 0);/*consistency fault, no traced NB for child ID found */

  return nb_idx;
}



/* Performs actions triggered by mouse events on the training image.

Revision: 02.10.2004 E.Sack

Input : - pointer to GDK event
Output: - -
*/
static void
training_image_label(GdkEventButton *event)
{
  PAGE_TABS current_page;

  if (GTK_IS_WIDGET(tr.dialog) == FALSE) {     /* training dialog required */
    return;
  }

  current_page = xplab_get_page_enum(GTK_NOTEBOOK(tr.page.notebook), -1);

  switch (current_page) {
  case XPLAB_TRAINING_PAGE_IS_SELECT:
    training_image_focus(event);
    break;

  case XPLAB_TRAINING_PAGE_IS_TRACE:
    training_image_kernel_label(event);
    break;

  case XPLAB_TAB_UNDEF:
    break;
  }
}



/* Labels a traced NB in the training image by the mouse pointer.

Revision: 02.10.2004 E.Sack

Input : - pointer to GDK event
Output: - -
*/
static void
training_image_kernel_label(GdkEventButton *event)
{
  int        within, nb_idx, cnr = XPLAB_SINGLE_CLASS;
  c_def      start;
  sector_def sec;

  if (!tr.page.trace.tdata.lbls.nr) {                   /* consistency check */
    assert(!tr.page.trace.tdata.lbls.cnt);
    return;
  }
  else {
    assert(tr.page.trace.tdata.lbls.cnt>0);
  }

  AREA_TO_SECTOR(tr.session.mimg.pimg.ar, sec);      /* whole training image */

  within = image_coordinates(event, &start, sec, tr.timg.state.magnitude);

  if (!within) {              /* preselection: must be inside training image */
    return;
  }

  nb_idx = nb_from_coordinate(tr.page.trace.tdata.thash, 
			      tr.page.trace.tdata.traced.sclass.cls[cnr].nbl,
			      start);

  if (nb_idx >= 0) {
    traced_nb_label(nb_idx, 0);
  }
}



/* Selects traced NB in the training image.

Revision: 02.10.2004 E.Sack

Input : - mouse pointer coordinate
Output: - -
*/
static void
training_image_kernel_select(c_def start)
{
  int nb_idx, cnr = XPLAB_SINGLE_CLASS;

  if (!tr.page.trace.tdata.traced.sclass.cls) {         /* consistency check */
    assert(!tr.page.trace.tdata.traced.sclass.cnt);
    return;
  }
  else {
    assert(tr.page.trace.tdata.traced.sclass.cnt>0);
  }

  nb_idx = nb_from_coordinate(tr.page.trace.tdata.thash, 
			      tr.page.trace.tdata.traced.sclass.cls[cnr].nbl,
			      start);

  if (nb_idx >= 0) {
    traced_cursor(nb_idx);                         /* move background cursor */
  }
}


/* Changes pixel labels in the training image.

Revision: 02.10.2004 E.Sack

Input : - pointer to GDK event
Output: - -
*/
static void
training_image_focus(GdkEventButton *event)
{
  int   pxl, n4, within;
  c_def start;

  within = image_coordinates(event, &start, tr.page.select.focus_sector, 
			     tr.timg.state.magnitude);

  if (!within) {                                             /* focus only */
    return;
  }

  /* get pixel value of coordinate                                         */
  pxl = tr.session.mimg.pimg.mem[start.x+start.y*tr.session.mimg.pimg.ar.width];

  if ((pxl == XPLAB_SELECTED_PIXEL) || (pxl == XPLAB_SELECTBL_PIXEL)) {
    n4 = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tr.page.select.nbhood)) == TRUE ?
      1:0;

    xcontour_mark_from_start(tr.session.mimg.pimg, tr.page.select.focus_sector, 
			     pxl==XPLAB_SELECTED_PIXEL ?
			     XPLAB_SELECTBL_PIXEL:XPLAB_SELECTED_PIXEL, 
			     start, n4);

    image_to_state(1, tr.timg.state.source == DC_OPT_MARKED);
  }
}



/* Builds a menu bar.

Revision: 07.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
training_image_menu_bar(GtkWidget *vbox)
{
  GtkWidget *menu_bar;

  menu_bar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox), menu_bar, FALSE, FALSE, 0);

  training_image_menu_file(menu_bar);                           /* menu file */
  training_image_menu_view(menu_bar);                           /* menu view */
  xplab_menu_magn(menu_bar, GTK_SIGNAL_FUNC(cb_image_magnitude),/* menu magnitude */
		  NULL, XPLAB_MAGN_X1_VALUE);
  xplab_menu_help(menu_bar, &tr.top_training);                  /* menu help */
}



/* Builds an entry for a menu bar.

Revision: 07.08.2005 E.Sack

Input : - menu bar
Output: - menu bar updated
*/
static void
training_image_menu_file(GtkWidget *menu_bar)
{
  GtkWidget *file_entry, *file_menu, *sep;

  file_entry = gtk_menu_item_new_with_label(MENU_BUTTON_FILE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_entry);

  file_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_entry), file_menu);

  xplab_preferences_menu(file_menu, &tr.top_training);   /* item preferences */
  training_image_menu_train(file_menu);                        /* item train */
  training_image_menu_dump(file_menu);                          /* item dump */

  sep = gtk_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), sep);

  training_image_menu_save_quit(file_menu);                     /* item save */
  training_image_menu_quit(file_menu);                          /* item quit */
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
training_image_menu_save_quit(GtkWidget *file_menu)
{
  GtkWidget *file_menu_save_quit;

  file_menu_save_quit = gtk_menu_item_new_with_label(DIALOG_BUTTON_SAVE_QUIT);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_save_quit);

  g_signal_connect(G_OBJECT(file_menu_save_quit), "activate", 
		   G_CALLBACK(cb_training_save_quit), NULL);
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
training_image_menu_quit(GtkWidget *file_menu)
{
  GtkWidget *file_menu_quit;

  file_menu_quit = gtk_menu_item_new_with_label(DIALOG_BUTTON_QUIT);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_quit);

  g_signal_connect(G_OBJECT(file_menu_quit), "activate", 
		   G_CALLBACK(cb_training_quit), NULL);
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
training_image_menu_dump(GtkWidget *file_menu)
{
  GtkWidget *file_menu_dump;

  file_menu_dump = gtk_menu_item_new_with_label("dump training image");
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_dump);

  g_signal_connect(G_OBJECT(file_menu_dump), "activate",
		   G_CALLBACK(cb_image_dump), NULL);
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
training_image_menu_train(GtkWidget *file_menu)
{
  GtkWidget *file_menu_train;

  file_menu_train = gtk_menu_item_new_with_label(XPLAB_TRAINING_TITLE_ITEM " dialog...");
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_train);

  g_signal_connect(G_OBJECT(file_menu_train), "activate",
		   G_CALLBACK(cb_select_train_dialog), NULL);
}



/* Builds an entry for a menu bar.

Revision: 07.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
training_image_menu_view(GtkWidget *menu_bar)
{
  GSList    *image_group = NULL;
  GtkWidget *view_entry, *view_menu;

  view_entry = gtk_menu_item_new_with_label("View");
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), view_entry);

  view_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(view_entry), view_menu);

  view_menu_item_focus(view_menu);                            /* item focus */
  separator_item(view_menu);                             /* separation line */
  view_menu_item_rb_frames(view_menu);                    /* item rb frames */
  separator_item(view_menu);                             /* separation line */
  view_menu_item_mark(&image_group, view_menu);        /* item mark pattern */
  view_menu_item_orig(&image_group, view_menu);    /* item original pattern */
  view_menu_item_binar(&image_group, view_menu);  /* item binarized pattern */
  view_menu_item_empty(&image_group, view_menu);        /* item empty image */
}



/* Adds a separation line to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
separator_item(GtkWidget *menu)
{
  GtkWidget *separator;

  separator = gtk_menu_item_new();

  gtk_menu_shell_append(GTK_MENU_SHELL(menu), separator);
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
view_menu_item_focus(GtkWidget *view_menu)
{
  GtkWidget *focus_item;

  focus_item = gtk_menu_item_new_with_label("focus");
  gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), focus_item);

  g_signal_connect(G_OBJECT(focus_item), "activate",
		   G_CALLBACK(cb_image_focus), NULL);
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
view_menu_item_rb_frames(GtkWidget *view_menu)
{
  tr.rb_frames_button = gtk_check_menu_item_new_with_label("RB frames");
  gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), tr.rb_frames_button);

  g_signal_connect(G_OBJECT(tr.rb_frames_button), "toggled",
		   G_CALLBACK(cb_rb_frames), NULL);
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to widget group pointer
        - pointer to menu
Output: - group pointer updated
        - menu updated
*/
static void
view_menu_item_mark(GSList **image_group, GtkWidget *view_menu)
{
  tr.timg.mark_pattern = 
    gtk_radio_menu_item_new_with_label(*image_group, "mark pattern");

  *image_group = 
    gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(tr.timg.mark_pattern));

  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(tr.timg.mark_pattern), 
				 TRUE);

  gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), tr.timg.mark_pattern);

  g_signal_connect(tr.timg.mark_pattern, "toggled",
		   G_CALLBACK(cb_image_mark), NULL);
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to widget group pointer
        - pointer to menu
Output: - group pointer updated
        - menu updated
*/
static void
view_menu_item_orig(GSList **image_group, GtkWidget *view_menu)
{
  tr.view_original = gtk_radio_menu_item_new_with_label(*image_group, 
							"original input image");

  *image_group =
    gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(tr.view_original));

  gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), tr.view_original);

  g_signal_connect(tr.view_original, "toggled",
		   G_CALLBACK(cb_image_original), NULL);
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to widget group pointer
        - pointer to menu
Output: - group pointer updated
        - menu updated
*/
static void
view_menu_item_binar(GSList **image_group, GtkWidget *view_menu)
{
  GtkWidget *view_item;

  view_item = gtk_radio_menu_item_new_with_label(*image_group, 
						 "binarized input image");

  *image_group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(view_item));

  gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), view_item);

  g_signal_connect(view_item, "toggled",
		   G_CALLBACK(cb_image_binarized), NULL);
}



/* Adds an item to a menu.

Revision: 07.08.2005 E.Sack

Input : - pointer to widget group pointer
        - pointer to menu
Output: - group pointer updated
        - menu updated
*/
static void
view_menu_item_empty(GSList **image_group, GtkWidget *view_menu)
{
  GtkWidget *view_item;

  view_item = gtk_radio_menu_item_new_with_label(*image_group, "empty image");

  *image_group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(view_item));

  gtk_menu_shell_append(GTK_MENU_SHELL(view_menu), view_item);

  g_signal_connect(view_item, "toggled",
		   G_CALLBACK(cb_image_empty), NULL);
}
